:- module(iana, [iana/1,iana/2,iana/3]).

/** <module> IANA defined language list.
 *
 * All IANA language tags to be used in literal language assignment
 *
 * Taken from: "https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry",
 * File-Date: "2022-08-08",
 *
 */
:- use_module(library(apply)).
:- use_module(library(lists)).

/* iana(?Language_Entry) is nondet.
 *
 * Language_Entry is a dict, with (some) of the following fields:
 *
 * Language_Entry = tag{
 *   type: "Type, % one of: language, script, region, redundant, variant",
 *   subtag: "Tag,",
 *   tag: "Tag,",
 *   description: "Description(s),",
 *   added: "Date",
 *   suppress_script: "Script",
 *   scope: "Scope,",
 *   prefix: "Prefix(s),",
 *   comments: "Comment,",
 *   preferred_value: "Tag,",
 *   deprecated: "Date",
 * }
 *
 * A language tag is of the form:
 * language-extlang-script-region-variant-extension-privateuse
 */

% iana(+Lang) is semidet.
%
% Check if this language exists.
iana(Lang) :-
    iana(Lang,_).

% iana(+LANG,-Record) is semidet.
%
% If language exists, return the records.
iana(Tag,[Record]) :-
    iana(grandfathered,Tag,Record),
    !.
iana(Tag,Records) :-
    split_string(Tag,'-','',List),
    maplist(atom_string,Parts,List),
    Types = [language,extlang,script,region,variant,extension,privateuse],
    select_iana(Parts,Types,Records).

select_iana([],_,[]) .
select_iana([Tag|Tags],Types,[Record|Records]) :-
    select(Type,Types,Remaining),
    iana(Type,Tag,Record),
    select_iana(Tags,Remaining,Records).

iana(language,aa,tag{
         type: "language",
         subtag: "aa",
         description: "Afar",
         added: "2005-10-16"}).
iana(language,ab,tag{
         type: "language",
         subtag: "ab",
         description: "Abkhazian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,ae,tag{
         type: "language",
         subtag: "ae",
         description: "Avestan",
         added: "2005-10-16"}).
iana(language,af,tag{
         type: "language",
         subtag: "af",
         description: "Afrikaans",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ak,tag{
         type: "language",
         subtag: "ak",
         description: "Akan",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,am,tag{
         type: "language",
         subtag: "am",
         description: "Amharic",
         added: "2005-10-16",
         suppress_script: "Ethi"}).
iana(language,an,tag{
         type: "language",
         subtag: "an",
         description: "Aragonese",
         added: "2005-10-16"}).
iana(language,ar,tag{
         type: "language",
         subtag: "ar",
         description: "Arabic",
         added: "2005-10-16",
         suppress_script: "Arab",
         scope: "macrolanguage"}).
iana(language,as,tag{
         type: "language",
         subtag: "as",
         description: "Assamese",
         added: "2005-10-16",
         suppress_script: "Beng"}).
iana(language,av,tag{
         type: "language",
         subtag: "av",
         description: "Avaric",
         added: "2005-10-16"}).
iana(language,ay,tag{
         type: "language",
         subtag: "ay",
         description: "Aymara",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,az,tag{
         type: "language",
         subtag: "az",
         description: "Azerbaijani",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,ba,tag{
         type: "language",
         subtag: "ba",
         description: "Bashkir",
         added: "2005-10-16"}).
iana(language,be,tag{
         type: "language",
         subtag: "be",
         description: "Belarusian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,bg,tag{
         type: "language",
         subtag: "bg",
         description: "Bulgarian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,bh,tag{
         type: "language",
         subtag: "bh",
         description: "Bihari languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,bi,tag{
         type: "language",
         subtag: "bi",
         description: "Bislama",
         added: "2005-10-16"}).
iana(language,bm,tag{
         type: "language",
         subtag: "bm",
         description: "Bambara",
         added: "2005-10-16"}).
iana(language,bn,tag{
         type: "language",
         subtag: "bn",
         description: ["Bengali","Bangla"],
         added: "2005-10-16",
         suppress_script: "Beng"}).
iana(language,bo,tag{
         type: "language",
         subtag: "bo",
         description: "Tibetan",
         added: "2005-10-16"}).
iana(language,br,tag{
         type: "language",
         subtag: "br",
         description: "Breton",
         added: "2005-10-16"}).
iana(language,bs,tag{
         type: "language",
         subtag: "bs",
         description: "Bosnian",
         added: "2005-10-16",
         suppress_script: "Latn",
         macrolanguage: "sh"}).
iana(language,ca,tag{
         type: "language",
         subtag: "ca",
         description: ["Catalan","Valencian"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ce,tag{
         type: "language",
         subtag: "ce",
         description: "Chechen",
         added: "2005-10-16"}).
iana(language,ch,tag{
         type: "language",
         subtag: "ch",
         description: "Chamorro",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,co,tag{
         type: "language",
         subtag: "co",
         description: "Corsican",
         added: "2005-10-16"}).
iana(language,cr,tag{
         type: "language",
         subtag: "cr",
         description: "Cree",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,cs,tag{
         type: "language",
         subtag: "cs",
         description: "Czech",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,cu,tag{
         type: "language",
         subtag: "cu",
         description: ["Church Slavic","Church Slavonic","Old Bulgarian","Old Church Slavonic","Old Slavonic"],
         added: "2005-10-16"}).
iana(language,cv,tag{
         type: "language",
         subtag: "cv",
         description: "Chuvash",
         added: "2005-10-16"}).
iana(language,cy,tag{
         type: "language",
         subtag: "cy",
         description: "Welsh",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,da,tag{
         type: "language",
         subtag: "da",
         description: "Danish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,de,tag{
         type: "language",
         subtag: "de",
         description: "German",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,dv,tag{
         type: "language",
         subtag: "dv",
         description: ["Dhivehi","Divehi","Maldivian"],
         added: "2005-10-16",
         suppress_script: "Thaa"}).
iana(language,dz,tag{
         type: "language",
         subtag: "dz",
         description: "Dzongkha",
         added: "2005-10-16",
         suppress_script: "Tibt"}).
iana(language,ee,tag{
         type: "language",
         subtag: "ee",
         description: "Ewe",
         added: "2005-10-16"}).
iana(language,el,tag{
         type: "language",
         subtag: "el",
         description: "Modern Greek (1453-)",
         added: "2005-10-16",
         suppress_script: "Grek"}).
iana(language,en,tag{
         type: "language",
         subtag: "en",
         description: "English",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,eo,tag{
         type: "language",
         subtag: "eo",
         description: "Esperanto",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,es,tag{
         type: "language",
         subtag: "es",
         description: ["Spanish","Castilian"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,et,tag{
         type: "language",
         subtag: "et",
         description: "Estonian",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,eu,tag{
         type: "language",
         subtag: "eu",
         description: "Basque",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,fa,tag{
         type: "language",
         subtag: "fa",
         description: "Persian",
         added: "2005-10-16",
         suppress_script: "Arab",
         scope: "macrolanguage"}).
iana(language,ff,tag{
         type: "language",
         subtag: "ff",
         description: "Fulah",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,fi,tag{
         type: "language",
         subtag: "fi",
         description: "Finnish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,fj,tag{
         type: "language",
         subtag: "fj",
         description: "Fijian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,fo,tag{
         type: "language",
         subtag: "fo",
         description: "Faroese",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,fr,tag{
         type: "language",
         subtag: "fr",
         description: "French",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,fy,tag{
         type: "language",
         subtag: "fy",
         description: "Western Frisian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ga,tag{
         type: "language",
         subtag: "ga",
         description: "Irish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,gd,tag{
         type: "language",
         subtag: "gd",
         description: ["Scottish Gaelic","Gaelic"],
         added: "2005-10-16"}).
iana(language,gl,tag{
         type: "language",
         subtag: "gl",
         description: "Galician",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,gn,tag{
         type: "language",
         subtag: "gn",
         description: "Guarani",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,gu,tag{
         type: "language",
         subtag: "gu",
         description: "Gujarati",
         added: "2005-10-16",
         suppress_script: "Gujr"}).
iana(language,gv,tag{
         type: "language",
         subtag: "gv",
         description: "Manx",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ha,tag{
         type: "language",
         subtag: "ha",
         description: "Hausa",
         added: "2005-10-16"}).
iana(language,he,tag{
         type: "language",
         subtag: "he",
         description: "Hebrew",
         added: "2005-10-16",
         suppress_script: "Hebr"}).
iana(language,hi,tag{
         type: "language",
         subtag: "hi",
         description: "Hindi",
         added: "2005-10-16",
         suppress_script: "Deva"}).
iana(language,ho,tag{
         type: "language",
         subtag: "ho",
         description: "Hiri Motu",
         added: "2005-10-16"}).
iana(language,hr,tag{
         type: "language",
         subtag: "hr",
         description: "Croatian",
         added: "2005-10-16",
         suppress_script: "Latn",
         macrolanguage: "sh"}).
iana(language,ht,tag{
         type: "language",
         subtag: "ht",
         description: ["Haitian","Haitian Creole"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,hu,tag{
         type: "language",
         subtag: "hu",
         description: "Hungarian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,hy,tag{
         type: "language",
         subtag: "hy",
         description: "Armenian",
         added: "2005-10-16",
         suppress_script: "Armn",
         comments: "see also hyw"}).
iana(language,hz,tag{
         type: "language",
         subtag: "hz",
         description: "Herero",
         added: "2005-10-16"}).
iana(language,ia,tag{
         type: "language",
         subtag: "ia",
         description: "Interlingua (International Auxiliary Language association)",
         added: "2005-10-16"}).
iana(language,id,tag{
         type: "language",
         subtag: "id",
         description: "Indonesian",
         added: "2005-10-16",
         suppress_script: "Latn",
         macrolanguage: "ms"}).
iana(language,ie,tag{
         type: "language",
         subtag: "ie",
         description: ["Interlingue","Occidental"],
         added: "2005-10-16"}).
iana(language,ig,tag{
         type: "language",
         subtag: "ig",
         description: "Igbo",
         added: "2005-10-16"}).
iana(language,ii,tag{
         type: "language",
         subtag: "ii",
         description: ["Sichuan Yi","Nuosu"],
         added: "2005-10-16"}).
iana(language,ik,tag{
         type: "language",
         subtag: "ik",
         description: "Inupiaq",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,in,tag{
         type: "language",
         subtag: "in",
         description: "Indonesian",
         added: "2005-10-16",
         deprecated: "1989-01-01",
         preferred_value: "id",
         suppress_script: "Latn",
         macrolanguage: "ms"}).
iana(language,io,tag{
         type: "language",
         subtag: "io",
         description: "Ido",
         added: "2005-10-16"}).
iana(language,is,tag{
         type: "language",
         subtag: "is",
         description: "Icelandic",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,it,tag{
         type: "language",
         subtag: "it",
         description: "Italian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,iu,tag{
         type: "language",
         subtag: "iu",
         description: "Inuktitut",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,iw,tag{
         type: "language",
         subtag: "iw",
         description: "Hebrew",
         added: "2005-10-16",
         deprecated: "1989-01-01",
         preferred_value: "he",
         suppress_script: "Hebr"}).
iana(language,ja,tag{
         type: "language",
         subtag: "ja",
         description: "Japanese",
         added: "2005-10-16",
         suppress_script: "Jpan"}).
iana(language,ji,tag{
         type: "language",
         subtag: "ji",
         description: "Yiddish",
         added: "2005-10-16",
         deprecated: "1989-01-01",
         preferred_value: "yi"}).
iana(language,jv,tag{
         type: "language",
         subtag: "jv",
         description: "Javanese",
         added: "2005-10-16"}).
iana(language,jw,tag{
         type: "language",
         subtag: "jw",
         description: "Javanese",
         added: "2005-10-16",
         deprecated: "2001-08-13",
         preferred_value: "jv",
         comments: "published by error in Table 1 of ISO 639:1988"}).
iana(language,ka,tag{
         type: "language",
         subtag: "ka",
         description: "Georgian",
         added: "2005-10-16",
         suppress_script: "Geor"}).
iana(language,kg,tag{
         type: "language",
         subtag: "kg",
         description: "Kongo",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,ki,tag{
         type: "language",
         subtag: "ki",
         description: ["Kikuyu","Gikuyu"],
         added: "2005-10-16"}).
iana(language,kj,tag{
         type: "language",
         subtag: "kj",
         description: ["Kuanyama","Kwanyama"],
         added: "2005-10-16"}).
iana(language,kk,tag{
         type: "language",
         subtag: "kk",
         description: "Kazakh",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,kl,tag{
         type: "language",
         subtag: "kl",
         description: ["Kalaallisut","Greenlandic"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,km,tag{
         type: "language",
         subtag: "km",
         description: ["Khmer","Central Khmer"],
         added: "2005-10-16",
         suppress_script: "Khmr"}).
iana(language,kn,tag{
         type: "language",
         subtag: "kn",
         description: "Kannada",
         added: "2005-10-16",
         suppress_script: "Knda"}).
iana(language,ko,tag{
         type: "language",
         subtag: "ko",
         description: "Korean",
         added: "2005-10-16",
         suppress_script: "Kore"}).
iana(language,kr,tag{
         type: "language",
         subtag: "kr",
         description: "Kanuri",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,ks,tag{
         type: "language",
         subtag: "ks",
         description: "Kashmiri",
         added: "2005-10-16"}).
iana(language,ku,tag{
         type: "language",
         subtag: "ku",
         description: "Kurdish",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,kv,tag{
         type: "language",
         subtag: "kv",
         description: "Komi",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,kw,tag{
         type: "language",
         subtag: "kw",
         description: "Cornish",
         added: "2005-10-16"}).
iana(language,ky,tag{
         type: "language",
         subtag: "ky",
         description: ["Kirghiz","Kyrgyz"],
         added: "2005-10-16"}).
iana(language,la,tag{
         type: "language",
         subtag: "la",
         description: "Latin",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,lb,tag{
         type: "language",
         subtag: "lb",
         description: ["Luxembourgish","Letzeburgesch"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,lg,tag{
         type: "language",
         subtag: "lg",
         description: ["Ganda","Luganda"],
         added: "2005-10-16"}).
iana(language,li,tag{
         type: "language",
         subtag: "li",
         description: ["Limburgan","Limburger","Limburgish"],
         added: "2005-10-16"}).
iana(language,ln,tag{
         type: "language",
         subtag: "ln",
         description: "Lingala",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,lo,tag{
         type: "language",
         subtag: "lo",
         description: "Lao",
         added: "2005-10-16",
         suppress_script: "Laoo"}).
iana(language,lt,tag{
         type: "language",
         subtag: "lt",
         description: "Lithuanian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,lu,tag{
         type: "language",
         subtag: "lu",
         description: "Luba-Katanga",
         added: "2005-10-16"}).
iana(language,lv,tag{
         type: "language",
         subtag: "lv",
         description: "Latvian",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,mg,tag{
         type: "language",
         subtag: "mg",
         description: "Malagasy",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,mh,tag{
         type: "language",
         subtag: "mh",
         description: "Marshallese",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,mi,tag{
         type: "language",
         subtag: "mi",
         description: "Maori",
         added: "2005-10-16"}).
iana(language,mk,tag{
         type: "language",
         subtag: "mk",
         description: "Macedonian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,ml,tag{
         type: "language",
         subtag: "ml",
         description: "Malayalam",
         added: "2005-10-16",
         suppress_script: "Mlym"}).
iana(language,mn,tag{
         type: "language",
         subtag: "mn",
         description: "Mongolian",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,mo,tag{
         type: "language",
         subtag: "mo",
         description: ["Moldavian","Moldovan"],
         added: "2005-10-16",
         deprecated: "2008-11-22",
         preferred_value: "ro",
         suppress_script: "Latn"}).
iana(language,mr,tag{
         type: "language",
         subtag: "mr",
         description: "Marathi",
         added: "2005-10-16",
         suppress_script: "Deva"}).
iana(language,ms,tag{
         type: "language",
         subtag: "ms",
         description: "Malay (macrolanguage)",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,mt,tag{
         type: "language",
         subtag: "mt",
         description: "Maltese",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,my,tag{
         type: "language",
         subtag: "my",
         description: "Burmese",
         added: "2005-10-16",
         suppress_script: "Mymr"}).
iana(language,na,tag{
         type: "language",
         subtag: "na",
         description: "Nauru",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,nb,tag{
         type: "language",
         subtag: "nb",
         description: "Norwegian Bokmål",
         added: "2005-10-16",
         suppress_script: "Latn",
         macrolanguage: "no"}).
iana(language,nd,tag{
         type: "language",
         subtag: "nd",
         description: "North Ndebele",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ne,tag{
         type: "language",
         subtag: "ne",
         description: "Nepali (macrolanguage)",
         added: "2005-10-16",
         suppress_script: "Deva",
         scope: "macrolanguage"}).
iana(language,ng,tag{
         type: "language",
         subtag: "ng",
         description: "Ndonga",
         added: "2005-10-16"}).
iana(language,nl,tag{
         type: "language",
         subtag: "nl",
         description: ["Dutch","Flemish"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,nn,tag{
         type: "language",
         subtag: "nn",
         description: "Norwegian Nynorsk",
         added: "2005-10-16",
         suppress_script: "Latn",
         macrolanguage: "no"}).
iana(language,no,tag{
         type: "language",
         subtag: "no",
         description: "Norwegian",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,nr,tag{
         type: "language",
         subtag: "nr",
         description: "South Ndebele",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,nv,tag{
         type: "language",
         subtag: "nv",
         description: ["Navajo","Navaho"],
         added: "2005-10-16"}).
iana(language,ny,tag{
         type: "language",
         subtag: "ny",
         description: ["Nyanja","Chewa","Chichewa"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,oc,tag{
         type: "language",
         subtag: "oc",
         description: "Occitan (post 1500)",
         added: "2005-10-16"}).
iana(language,oj,tag{
         type: "language",
         subtag: "oj",
         description: "Ojibwa",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,om,tag{
         type: "language",
         subtag: "om",
         description: "Oromo",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,or,tag{
         type: "language",
         subtag: "or",
         description: ["Oriya (macrolanguage)","Odia (macrolanguage)"],
         added: "2005-10-16",
         suppress_script: "Orya",
         scope: "macrolanguage"}).
iana(language,os,tag{
         type: "language",
         subtag: "os",
         description: ["Ossetian","Ossetic"],
         added: "2005-10-16"}).
iana(language,pa,tag{
         type: "language",
         subtag: "pa",
         description: ["Panjabi","Punjabi"],
         added: "2005-10-16",
         suppress_script: "Guru"}).
iana(language,pi,tag{
         type: "language",
         subtag: "pi",
         description: "Pali",
         added: "2005-10-16"}).
iana(language,pl,tag{
         type: "language",
         subtag: "pl",
         description: "Polish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ps,tag{
         type: "language",
         subtag: "ps",
         description: ["Pushto","Pashto"],
         added: "2005-10-16",
         suppress_script: "Arab",
         scope: "macrolanguage"}).
iana(language,pt,tag{
         type: "language",
         subtag: "pt",
         description: "Portuguese",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,qu,tag{
         type: "language",
         subtag: "qu",
         description: "Quechua",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,rm,tag{
         type: "language",
         subtag: "rm",
         description: "Romansh",
         suppress_script: "Latn",
         added: "2005-10-16"}).
iana(language,rn,tag{
         type: "language",
         subtag: "rn",
         description: "Rundi",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ro,tag{
         type: "language",
         subtag: "ro",
         description: ["Romanian","Moldavian","Moldovan"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ru,tag{
         type: "language",
         subtag: "ru",
         description: "Russian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,rw,tag{
         type: "language",
         subtag: "rw",
         description: "Kinyarwanda",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sa,tag{
         type: "language",
         subtag: "sa",
         description: "Sanskrit",
         added: "2005-10-16"}).
iana(language,sc,tag{
         type: "language",
         subtag: "sc",
         description: "Sardinian",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,sd,tag{
         type: "language",
         subtag: "sd",
         description: "Sindhi",
         added: "2005-10-16"}).
iana(language,se,tag{
         type: "language",
         subtag: "se",
         description: "Northern Sami",
         added: "2005-10-16"}).
iana(language,sg,tag{
         type: "language",
         subtag: "sg",
         description: "Sango",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sh,tag{
         type: "language",
         subtag: "sh",
         description: "Serbo-Croatian",
         added: "2005-10-16",
         scope: "macrolanguage",
         comments: "sr, hr, bs are preferred for most modern uses"}).
iana(language,si,tag{
         type: "language",
         subtag: "si",
         description: ["Sinhala","Sinhalese"],
         added: "2005-10-16",
         suppress_script: "Sinh"}).
iana(language,sk,tag{
         type: "language",
         subtag: "sk",
         description: "Slovak",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sl,tag{
         type: "language",
         subtag: "sl",
         description: "Slovenian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sm,tag{
         type: "language",
         subtag: "sm",
         description: "Samoan",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sn,tag{
         type: "language",
         subtag: "sn",
         description: "Shona",
         added: "2005-10-16"}).
iana(language,so,tag{
         type: "language",
         subtag: "so",
         description: "Somali",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sq,tag{
         type: "language",
         subtag: "sq",
         description: "Albanian",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,sr,tag{
         type: "language",
         subtag: "sr",
         description: "Serbian",
         added: "2005-10-16",
         macrolanguage: "sh",
         comments: "see cnr for Montenegrin"}).
iana(language,ss,tag{
         type: "language",
         subtag: "ss",
         description: "Swati",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,st,tag{
         type: "language",
         subtag: "st",
         description: "Southern Sotho",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,su,tag{
         type: "language",
         subtag: "su",
         description: "Sundanese",
         added: "2005-10-16"}).
iana(language,sv,tag{
         type: "language",
         subtag: "sv",
         description: "Swedish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,sw,tag{
         type: "language",
         subtag: "sw",
         description: "Swahili (macrolanguage)",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,ta,tag{
         type: "language",
         subtag: "ta",
         description: "Tamil",
         added: "2005-10-16",
         suppress_script: "Taml"}).
iana(language,te,tag{
         type: "language",
         subtag: "te",
         description: "Telugu",
         added: "2005-10-16",
         suppress_script: "Telu"}).
iana(language,tg,tag{
         type: "language",
         subtag: "tg",
         description: "Tajik",
         added: "2005-10-16"}).
iana(language,th,tag{
         type: "language",
         subtag: "th",
         description: "Thai",
         added: "2005-10-16",
         suppress_script: "Thai"}).
iana(language,ti,tag{
         type: "language",
         subtag: "ti",
         description: "Tigrinya",
         added: "2005-10-16",
         suppress_script: "Ethi"}).
iana(language,tk,tag{
         type: "language",
         subtag: "tk",
         description: "Turkmen",
         added: "2005-10-16"}).
iana(language,tl,tag{
         type: "language",
         subtag: "tl",
         description: "Tagalog",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tn,tag{
         type: "language",
         subtag: "tn",
         description: "Tswana",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,to,tag{
         type: "language",
         subtag: "to",
         description: "Tonga (Tonga Islands)",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tr,tag{
         type: "language",
         subtag: "tr",
         description: "Turkish",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ts,tag{
         type: "language",
         subtag: "ts",
         description: "Tsonga",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tt,tag{
         type: "language",
         subtag: "tt",
         description: "Tatar",
         added: "2005-10-16"}).
iana(language,tw,tag{
         type: "language",
         subtag: "tw",
         description: "Twi",
         added: "2005-10-16",
         macrolanguage: "ak"}).
iana(language,ty,tag{
         type: "language",
         subtag: "ty",
         description: "Tahitian",
         added: "2005-10-16"}).
iana(language,ug,tag{
         type: "language",
         subtag: "ug",
         description: ["Uighur","Uyghur"],
         added: "2005-10-16"}).
iana(language,uk,tag{
         type: "language",
         subtag: "uk",
         description: "Ukrainian",
         added: "2005-10-16",
         suppress_script: "Cyrl"}).
iana(language,ur,tag{
         type: "language",
         subtag: "ur",
         description: "Urdu",
         added: "2005-10-16",
         suppress_script: "Arab"}).
iana(language,uz,tag{
         type: "language",
         subtag: "uz",
         description: "Uzbek",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,ve,tag{
         type: "language",
         subtag: "ve",
         description: "Venda",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,vi,tag{
         type: "language",
         subtag: "vi",
         description: "Vietnamese",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,vo,tag{
         type: "language",
         subtag: "vo",
         description: "Volapük",
         added: "2005-10-16"}).
iana(language,wa,tag{
         type: "language",
         subtag: "wa",
         description: "Walloon",
         added: "2005-10-16"}).
iana(language,wo,tag{
         type: "language",
         subtag: "wo",
         description: "Wolof",
         added: "2005-10-16"}).
iana(language,xh,tag{
         type: "language",
         subtag: "xh",
         description: "Xhosa",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,yi,tag{
         type: "language",
         subtag: "yi",
         description: "Yiddish",
         added: "2005-10-16",
         suppress_script: "Hebr",
         scope: "macrolanguage"}).
iana(language,yo,tag{
         type: "language",
         subtag: "yo",
         description: "Yoruba",
         added: "2005-10-16"}).
iana(language,za,tag{
         type: "language",
         subtag: "za",
         description: ["Zhuang","Chuang"],
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,zh,tag{
         type: "language",
         subtag: "zh",
         description: "Chinese",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,zu,tag{
         type: "language",
         subtag: "zu",
         description: "Zulu",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,aaa,tag{
         type: "language",
         subtag: "aaa",
         description: "Ghotuo",
         added: "2009-07-29"}).
iana(language,aab,tag{
         type: "language",
         subtag: "aab",
         description: "Alumu-Tesu",
         added: "2009-07-29"}).
iana(language,aac,tag{
         type: "language",
         subtag: "aac",
         description: "Ari",
         added: "2009-07-29"}).
iana(language,aad,tag{
         type: "language",
         subtag: "aad",
         description: "Amal",
         added: "2009-07-29"}).
iana(language,aae,tag{
         type: "language",
         subtag: "aae",
         description: "Arbëreshë Albanian",
         added: "2009-07-29",
         macrolanguage: "sq"}).
iana(language,aaf,tag{
         type: "language",
         subtag: "aaf",
         description: "Aranadan",
         added: "2009-07-29"}).
iana(language,aag,tag{
         type: "language",
         subtag: "aag",
         description: "Ambrak",
         added: "2009-07-29"}).
iana(language,aah,tag{
         type: "language",
         subtag: "aah",
         description: "Abu' Arapesh",
         added: "2009-07-29"}).
iana(language,aai,tag{
         type: "language",
         subtag: "aai",
         description: "Arifama-Miniafia",
         added: "2009-07-29"}).
iana(language,aak,tag{
         type: "language",
         subtag: "aak",
         description: "Ankave",
         added: "2009-07-29"}).
iana(language,aal,tag{
         type: "language",
         subtag: "aal",
         description: "Afade",
         added: "2009-07-29"}).
iana(language,aam,tag{
         type: "language",
         subtag: "aam",
         description: "Aramanik",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "aas"}).
iana(language,aan,tag{
         type: "language",
         subtag: "aan",
         description: "Anambé",
         added: "2009-07-29"}).
iana(language,aao,tag{
         type: "language",
         subtag: "aao",
         description: "Algerian Saharan Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,aap,tag{
         type: "language",
         subtag: "aap",
         description: "Pará Arára",
         added: "2009-07-29"}).
iana(language,aaq,tag{
         type: "language",
         subtag: "aaq",
         description: "Eastern Abnaki",
         added: "2009-07-29"}).
iana(language,aas,tag{
         type: "language",
         subtag: "aas",
         description: "Aasáx",
         added: "2009-07-29"}).
iana(language,aat,tag{
         type: "language",
         subtag: "aat",
         description: "Arvanitika Albanian",
         added: "2009-07-29",
         macrolanguage: "sq"}).
iana(language,aau,tag{
         type: "language",
         subtag: "aau",
         description: "Abau",
         added: "2009-07-29"}).
iana(language,aav,tag{
         type: "language",
         subtag: "aav",
         description: "Austro-Asiatic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,aaw,tag{
         type: "language",
         subtag: "aaw",
         description: "Solong",
         added: "2009-07-29"}).
iana(language,aax,tag{
         type: "language",
         subtag: "aax",
         description: "Mandobo Atas",
         added: "2009-07-29"}).
iana(language,aaz,tag{
         type: "language",
         subtag: "aaz",
         description: "Amarasi",
         added: "2009-07-29"}).
iana(language,aba,tag{
         type: "language",
         subtag: "aba",
         description: "Abé",
         added: "2009-07-29"}).
iana(language,abb,tag{
         type: "language",
         subtag: "abb",
         description: "Bankon",
         added: "2009-07-29"}).
iana(language,abc,tag{
         type: "language",
         subtag: "abc",
         description: "Ambala Ayta",
         added: "2009-07-29"}).
iana(language,abd,tag{
         type: "language",
         subtag: "abd",
         description: "Manide",
         added: "2009-07-29"}).
iana(language,abe,tag{
         type: "language",
         subtag: "abe",
         description: "Western Abnaki",
         added: "2009-07-29"}).
iana(language,abf,tag{
         type: "language",
         subtag: "abf",
         description: "Abai Sungai",
         added: "2009-07-29"}).
iana(language,abg,tag{
         type: "language",
         subtag: "abg",
         description: "Abaga",
         added: "2009-07-29"}).
iana(language,abh,tag{
         type: "language",
         subtag: "abh",
         description: "Tajiki Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,abi,tag{
         type: "language",
         subtag: "abi",
         description: "Abidji",
         added: "2009-07-29"}).
iana(language,abj,tag{
         type: "language",
         subtag: "abj",
         description: "Aka-Bea",
         added: "2009-07-29"}).
iana(language,abl,tag{
         type: "language",
         subtag: "abl",
         description: "Lampung Nyo",
         added: "2009-07-29"}).
iana(language,abm,tag{
         type: "language",
         subtag: "abm",
         description: "Abanyom",
         added: "2009-07-29"}).
iana(language,abn,tag{
         type: "language",
         subtag: "abn",
         description: "Abua",
         added: "2009-07-29"}).
iana(language,abo,tag{
         type: "language",
         subtag: "abo",
         description: "Abon",
         added: "2009-07-29"}).
iana(language,abp,tag{
         type: "language",
         subtag: "abp",
         description: "Abellen Ayta",
         added: "2009-07-29"}).
iana(language,abq,tag{
         type: "language",
         subtag: "abq",
         description: "Abaza",
         added: "2009-07-29"}).
iana(language,abr,tag{
         type: "language",
         subtag: "abr",
         description: "Abron",
         added: "2009-07-29"}).
iana(language,abs,tag{
         type: "language",
         subtag: "abs",
         description: "Ambonese Malay",
         added: "2009-07-29"}).
iana(language,abt,tag{
         type: "language",
         subtag: "abt",
         description: "Ambulas",
         added: "2009-07-29"}).
iana(language,abu,tag{
         type: "language",
         subtag: "abu",
         description: "Abure",
         added: "2009-07-29"}).
iana(language,abv,tag{
         type: "language",
         subtag: "abv",
         description: "Baharna Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,abw,tag{
         type: "language",
         subtag: "abw",
         description: "Pal",
         added: "2009-07-29"}).
iana(language,abx,tag{
         type: "language",
         subtag: "abx",
         description: "Inabaknon",
         added: "2009-07-29"}).
iana(language,aby,tag{
         type: "language",
         subtag: "aby",
         description: "Aneme Wake",
         added: "2009-07-29"}).
iana(language,abz,tag{
         type: "language",
         subtag: "abz",
         description: "Abui",
         added: "2009-07-29"}).
iana(language,aca,tag{
         type: "language",
         subtag: "aca",
         description: "Achagua",
         added: "2009-07-29"}).
iana(language,acb,tag{
         type: "language",
         subtag: "acb",
         description: "Áncá",
         added: "2009-07-29"}).
iana(language,acd,tag{
         type: "language",
         subtag: "acd",
         description: "Gikyode",
         added: "2009-07-29"}).
iana(language,ace,tag{
         type: "language",
         subtag: "ace",
         description: "Achinese",
         added: "2005-10-16"}).
iana(language,acf,tag{
         type: "language",
         subtag: "acf",
         description: "Saint Lucian Creole French",
         added: "2009-07-29"}).
iana(language,ach,tag{
         type: "language",
         subtag: "ach",
         description: "Acoli",
         added: "2005-10-16"}).
iana(language,aci,tag{
         type: "language",
         subtag: "aci",
         description: "Aka-Cari",
         added: "2009-07-29"}).
iana(language,ack,tag{
         type: "language",
         subtag: "ack",
         description: "Aka-Kora",
         added: "2009-07-29"}).
iana(language,acl,tag{
         type: "language",
         subtag: "acl",
         description: "Akar-Bale",
         added: "2009-07-29"}).
iana(language,acm,tag{
         type: "language",
         subtag: "acm",
         description: "Mesopotamian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,acn,tag{
         type: "language",
         subtag: "acn",
         description: "Achang",
         added: "2009-07-29"}).
iana(language,acp,tag{
         type: "language",
         subtag: "acp",
         description: "Eastern Acipa",
         added: "2009-07-29"}).
iana(language,acq,tag{
         type: "language",
         subtag: "acq",
         description: "Ta'izzi-Adeni Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,acr,tag{
         type: "language",
         subtag: "acr",
         description: "Achi",
         added: "2009-07-29"}).
iana(language,acs,tag{
         type: "language",
         subtag: "acs",
         description: "Acroá",
         added: "2009-07-29"}).
iana(language,act,tag{
         type: "language",
         subtag: "act",
         description: "Achterhoeks",
         added: "2009-07-29"}).
iana(language,acu,tag{
         type: "language",
         subtag: "acu",
         description: "Achuar-Shiwiar",
         added: "2009-07-29"}).
iana(language,acv,tag{
         type: "language",
         subtag: "acv",
         description: "Achumawi",
         added: "2009-07-29"}).
iana(language,acw,tag{
         type: "language",
         subtag: "acw",
         description: "Hijazi Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,acx,tag{
         type: "language",
         subtag: "acx",
         description: "Omani Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,acy,tag{
         type: "language",
         subtag: "acy",
         description: "Cypriot Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,acz,tag{
         type: "language",
         subtag: "acz",
         description: "Acheron",
         added: "2009-07-29"}).
iana(language,ada,tag{
         type: "language",
         subtag: "ada",
         description: "Adangme",
         added: "2005-10-16"}).
iana(language,adb,tag{
         type: "language",
         subtag: "adb",
         description: "Atauran",
         added: "2009-07-29"}).
iana(language,add,tag{
         type: "language",
         subtag: "add",
         description: ["Lidzonka","Dzodinka"],
         added: "2009-07-29"}).
iana(language,ade,tag{
         type: "language",
         subtag: "ade",
         description: "Adele",
         added: "2009-07-29"}).
iana(language,adf,tag{
         type: "language",
         subtag: "adf",
         description: "Dhofari Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,adg,tag{
         type: "language",
         subtag: "adg",
         description: "Andegerebinha",
         added: "2009-07-29"}).
iana(language,adh,tag{
         type: "language",
         subtag: "adh",
         description: "Adhola",
         added: "2009-07-29"}).
iana(language,adi,tag{
         type: "language",
         subtag: "adi",
         description: "Adi",
         added: "2009-07-29"}).
iana(language,adj,tag{
         type: "language",
         subtag: "adj",
         description: "Adioukrou",
         added: "2009-07-29"}).
iana(language,adl,tag{
         type: "language",
         subtag: "adl",
         description: "Galo",
         added: "2009-07-29"}).
iana(language,adn,tag{
         type: "language",
         subtag: "adn",
         description: "Adang",
         added: "2009-07-29"}).
iana(language,ado,tag{
         type: "language",
         subtag: "ado",
         description: "Abu",
         added: "2009-07-29"}).
iana(language,adp,tag{
         type: "language",
         subtag: "adp",
         description: "Adap",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "dz"}).
iana(language,adq,tag{
         type: "language",
         subtag: "adq",
         description: "Adangbe",
         added: "2009-07-29"}).
iana(language,adr,tag{
         type: "language",
         subtag: "adr",
         description: "Adonara",
         added: "2009-07-29"}).
iana(language,ads,tag{
         type: "language",
         subtag: "ads",
         description: "Adamorobe Sign Language",
         added: "2009-07-29"}).
iana(language,adt,tag{
         type: "language",
         subtag: "adt",
         description: "Adnyamathanha",
         added: "2009-07-29"}).
iana(language,adu,tag{
         type: "language",
         subtag: "adu",
         description: "Aduge",
         added: "2009-07-29"}).
iana(language,adw,tag{
         type: "language",
         subtag: "adw",
         description: "Amundava",
         added: "2009-07-29"}).
iana(language,adx,tag{
         type: "language",
         subtag: "adx",
         description: "Amdo Tibetan",
         added: "2009-07-29"}).
iana(language,ady,tag{
         type: "language",
         subtag: "ady",
         description: ["Adyghe","Adygei"],
         added: "2005-10-16"}).
iana(language,adz,tag{
         type: "language",
         subtag: "adz",
         description: "Adzera",
         added: "2009-07-29"}).
iana(language,aea,tag{
         type: "language",
         subtag: "aea",
         description: "Areba",
         added: "2009-07-29"}).
iana(language,aeb,tag{
         type: "language",
         subtag: "aeb",
         description: "Tunisian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,aec,tag{
         type: "language",
         subtag: "aec",
         description: "Saidi Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,aed,tag{
         type: "language",
         subtag: "aed",
         description: "Argentine Sign Language",
         added: "2009-07-29"}).
iana(language,aee,tag{
         type: "language",
         subtag: "aee",
         description: ["Northeast Pashai","Northeast Pashayi"],
         added: "2009-07-29"}).
iana(language,aek,tag{
         type: "language",
         subtag: "aek",
         description: "Haeke",
         added: "2009-07-29"}).
iana(language,ael,tag{
         type: "language",
         subtag: "ael",
         description: "Ambele",
         added: "2009-07-29"}).
iana(language,aem,tag{
         type: "language",
         subtag: "aem",
         description: "Arem",
         added: "2009-07-29"}).
iana(language,aen,tag{
         type: "language",
         subtag: "aen",
         description: "Armenian Sign Language",
         added: "2009-07-29"}).
iana(language,aeq,tag{
         type: "language",
         subtag: "aeq",
         description: "Aer",
         added: "2009-07-29"}).
iana(language,aer,tag{
         type: "language",
         subtag: "aer",
         description: "Eastern Arrernte",
         added: "2009-07-29"}).
iana(language,aes,tag{
         type: "language",
         subtag: "aes",
         description: "Alsea",
         added: "2009-07-29"}).
iana(language,aeu,tag{
         type: "language",
         subtag: "aeu",
         description: "Akeu",
         added: "2009-07-29"}).
iana(language,aew,tag{
         type: "language",
         subtag: "aew",
         description: "Ambakich",
         added: "2009-07-29"}).
iana(language,aey,tag{
         type: "language",
         subtag: "aey",
         description: "Amele",
         added: "2009-07-29"}).
iana(language,aez,tag{
         type: "language",
         subtag: "aez",
         description: "Aeka",
         added: "2009-07-29"}).
iana(language,afa,tag{
         type: "language",
         subtag: "afa",
         description: "Afro-Asiatic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,afb,tag{
         type: "language",
         subtag: "afb",
         description: "Gulf Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,afd,tag{
         type: "language",
         subtag: "afd",
         description: "Andai",
         added: "2009-07-29"}).
iana(language,afe,tag{
         type: "language",
         subtag: "afe",
         description: "Putukwam",
         added: "2009-07-29"}).
iana(language,afg,tag{
         type: "language",
         subtag: "afg",
         description: "Afghan Sign Language",
         added: "2009-07-29"}).
iana(language,afh,tag{
         type: "language",
         subtag: "afh",
         description: "Afrihili",
         added: "2005-10-16"}).
iana(language,afi,tag{
         type: "language",
         subtag: "afi",
         description: ["Akrukay","Chini"],
         added: "2009-07-29"}).
iana(language,afk,tag{
         type: "language",
         subtag: "afk",
         description: "Nanubae",
         added: "2009-07-29"}).
iana(language,afn,tag{
         type: "language",
         subtag: "afn",
         description: "Defaka",
         added: "2009-07-29"}).
iana(language,afo,tag{
         type: "language",
         subtag: "afo",
         description: "Eloyi",
         added: "2009-07-29"}).
iana(language,afp,tag{
         type: "language",
         subtag: "afp",
         description: "Tapei",
         added: "2009-07-29"}).
iana(language,afs,tag{
         type: "language",
         subtag: "afs",
         description: "Afro-Seminole Creole",
         added: "2009-07-29"}).
iana(language,aft,tag{
         type: "language",
         subtag: "aft",
         description: "Afitti",
         added: "2009-07-29"}).
iana(language,afu,tag{
         type: "language",
         subtag: "afu",
         description: "Awutu",
         added: "2009-07-29"}).
iana(language,afz,tag{
         type: "language",
         subtag: "afz",
         description: "Obokuitai",
         added: "2009-07-29"}).
iana(language,aga,tag{
         type: "language",
         subtag: "aga",
         description: "Aguano",
         added: "2009-07-29"}).
iana(language,agb,tag{
         type: "language",
         subtag: "agb",
         description: "Legbo",
         added: "2009-07-29"}).
iana(language,agc,tag{
         type: "language",
         subtag: "agc",
         description: "Agatu",
         added: "2009-07-29"}).
iana(language,agd,tag{
         type: "language",
         subtag: "agd",
         description: "Agarabi",
         added: "2009-07-29"}).
iana(language,age,tag{
         type: "language",
         subtag: "age",
         description: "Angal",
         added: "2009-07-29"}).
iana(language,agf,tag{
         type: "language",
         subtag: "agf",
         description: "Arguni",
         added: "2009-07-29"}).
iana(language,agg,tag{
         type: "language",
         subtag: "agg",
         description: "Angor",
         added: "2009-07-29"}).
iana(language,agh,tag{
         type: "language",
         subtag: "agh",
         description: "Ngelima",
         added: "2009-07-29"}).
iana(language,agi,tag{
         type: "language",
         subtag: "agi",
         description: "Agariya",
         added: "2009-07-29"}).
iana(language,agj,tag{
         type: "language",
         subtag: "agj",
         description: "Argobba",
         added: "2009-07-29"}).
iana(language,agk,tag{
         type: "language",
         subtag: "agk",
         description: "Isarog Agta",
         added: "2009-07-29"}).
iana(language,agl,tag{
         type: "language",
         subtag: "agl",
         description: "Fembe",
         added: "2009-07-29"}).
iana(language,agm,tag{
         type: "language",
         subtag: "agm",
         description: "Angaataha",
         added: "2009-07-29"}).
iana(language,agn,tag{
         type: "language",
         subtag: "agn",
         description: "Agutaynen",
         added: "2009-07-29"}).
iana(language,ago,tag{
         type: "language",
         subtag: "ago",
         description: "Tainae",
         added: "2009-07-29"}).
iana(language,agp,tag{
         type: "language",
         subtag: "agp",
         description: "Paranan",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see apf, prf"}).
iana(language,agq,tag{
         type: "language",
         subtag: "agq",
         description: "Aghem",
         added: "2009-07-29"}).
iana(language,agr,tag{
         type: "language",
         subtag: "agr",
         description: "Aguaruna",
         added: "2009-07-29"}).
iana(language,ags,tag{
         type: "language",
         subtag: "ags",
         description: "Esimbi",
         added: "2009-07-29"}).
iana(language,agt,tag{
         type: "language",
         subtag: "agt",
         description: "Central Cagayan Agta",
         added: "2009-07-29"}).
iana(language,agu,tag{
         type: "language",
         subtag: "agu",
         description: "Aguacateco",
         added: "2009-07-29"}).
iana(language,agv,tag{
         type: "language",
         subtag: "agv",
         description: "Remontado Dumagat",
         added: "2009-07-29"}).
iana(language,agw,tag{
         type: "language",
         subtag: "agw",
         description: "Kahua",
         added: "2009-07-29"}).
iana(language,agx,tag{
         type: "language",
         subtag: "agx",
         description: "Aghul",
         added: "2009-07-29"}).
iana(language,agy,tag{
         type: "language",
         subtag: "agy",
         description: "Southern Alta",
         added: "2009-07-29"}).
iana(language,agz,tag{
         type: "language",
         subtag: "agz",
         description: "Mt. Iriga Agta",
         added: "2009-07-29"}).
iana(language,aha,tag{
         type: "language",
         subtag: "aha",
         description: "Ahanta",
         added: "2009-07-29"}).
iana(language,ahb,tag{
         type: "language",
         subtag: "ahb",
         description: "Axamb",
         added: "2009-07-29"}).
iana(language,ahg,tag{
         type: "language",
         subtag: "ahg",
         description: "Qimant",
         added: "2009-07-29"}).
iana(language,ahh,tag{
         type: "language",
         subtag: "ahh",
         description: "Aghu",
         added: "2009-07-29"}).
iana(language,ahi,tag{
         type: "language",
         subtag: "ahi",
         description: "Tiagbamrin Aizi",
         added: "2009-07-29"}).
iana(language,ahk,tag{
         type: "language",
         subtag: "ahk",
         description: "Akha",
         added: "2009-07-29"}).
iana(language,ahl,tag{
         type: "language",
         subtag: "ahl",
         description: "Igo",
         added: "2009-07-29"}).
iana(language,ahm,tag{
         type: "language",
         subtag: "ahm",
         description: "Mobumrin Aizi",
         added: "2009-07-29"}).
iana(language,ahn,tag{
         type: "language",
         subtag: "ahn",
         description: "Àhàn",
         added: "2009-07-29"}).
iana(language,aho,tag{
         type: "language",
         subtag: "aho",
         description: "Ahom",
         added: "2009-07-29"}).
iana(language,ahp,tag{
         type: "language",
         subtag: "ahp",
         description: "Aproumu Aizi",
         added: "2009-07-29"}).
iana(language,ahr,tag{
         type: "language",
         subtag: "ahr",
         description: "Ahirani",
         added: "2009-07-29"}).
iana(language,ahs,tag{
         type: "language",
         subtag: "ahs",
         description: "Ashe",
         added: "2009-07-29"}).
iana(language,aht,tag{
         type: "language",
         subtag: "aht",
         description: "Ahtena",
         added: "2009-07-29"}).
iana(language,aia,tag{
         type: "language",
         subtag: "aia",
         description: "Arosi",
         added: "2009-07-29"}).
iana(language,aib,tag{
         type: "language",
         subtag: "aib",
         description: "Ainu (China)",
         added: "2009-07-29"}).
iana(language,aic,tag{
         type: "language",
         subtag: "aic",
         description: "Ainbai",
         added: "2009-07-29"}).
iana(language,aid,tag{
         type: "language",
         subtag: "aid",
         description: "Alngith",
         added: "2009-07-29"}).
iana(language,aie,tag{
         type: "language",
         subtag: "aie",
         description: "Amara",
         added: "2009-07-29"}).
iana(language,aif,tag{
         type: "language",
         subtag: "aif",
         description: "Agi",
         added: "2009-07-29"}).
iana(language,aig,tag{
         type: "language",
         subtag: "aig",
         description: "Antigua and Barbuda Creole English",
         added: "2009-07-29"}).
iana(language,aih,tag{
         type: "language",
         subtag: "aih",
         description: "Ai-Cham",
         added: "2009-07-29"}).
iana(language,aii,tag{
         type: "language",
         subtag: "aii",
         description: "Assyrian Neo-Aramaic",
         added: "2009-07-29",
         macrolanguage: "syr"}).
iana(language,aij,tag{
         type: "language",
         subtag: "aij",
         description: "Lishanid Noshan",
         added: "2009-07-29"}).
iana(language,aik,tag{
         type: "language",
         subtag: "aik",
         description: "Ake",
         added: "2009-07-29"}).
iana(language,ail,tag{
         type: "language",
         subtag: "ail",
         description: "Aimele",
         added: "2009-07-29"}).
iana(language,aim,tag{
         type: "language",
         subtag: "aim",
         description: "Aimol",
         added: "2009-07-29"}).
iana(language,ain,tag{
         type: "language",
         subtag: "ain",
         description: "Ainu (Japan)",
         added: "2005-10-16"}).
iana(language,aio,tag{
         type: "language",
         subtag: "aio",
         description: "Aiton",
         added: "2009-07-29"}).
iana(language,aip,tag{
         type: "language",
         subtag: "aip",
         description: "Burumakok",
         added: "2009-07-29"}).
iana(language,aiq,tag{
         type: "language",
         subtag: "aiq",
         description: "Aimaq",
         added: "2009-07-29"}).
iana(language,air,tag{
         type: "language",
         subtag: "air",
         description: "Airoran",
         added: "2009-07-29"}).
iana(language,ais,tag{
         type: "language",
         subtag: "ais",
         description: "Nataoran Amis",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         comments: "see ami, szy"}).
iana(language,ait,tag{
         type: "language",
         subtag: "ait",
         description: "Arikem",
         added: "2009-07-29"}).
iana(language,aiw,tag{
         type: "language",
         subtag: "aiw",
         description: "Aari",
         added: "2009-07-29"}).
iana(language,aix,tag{
         type: "language",
         subtag: "aix",
         description: "Aighon",
         added: "2009-07-29"}).
iana(language,aiy,tag{
         type: "language",
         subtag: "aiy",
         description: "Ali",
         added: "2009-07-29"}).
iana(language,aja,tag{
         type: "language",
         subtag: "aja",
         description: "Aja (South Sudan)",
         added: "2009-07-29"}).
iana(language,ajg,tag{
         type: "language",
         subtag: "ajg",
         description: "Aja (Benin)",
         added: "2009-07-29"}).
iana(language,aji,tag{
         type: "language",
         subtag: "aji",
         description: "Ajië",
         added: "2009-07-29"}).
iana(language,ajn,tag{
         type: "language",
         subtag: "ajn",
         description: "Andajin",
         added: "2012-08-12"}).
iana(language,ajp,tag{
         type: "language",
         subtag: "ajp",
         description: "South Levantine Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ajs,tag{
         type: "language",
         subtag: "ajs",
         description: "Algerian Jewish Sign Language",
         added: "2022-02-25"}).
iana(language,ajt,tag{
         type: "language",
         subtag: "ajt",
         description: "Judeo-Tunisian Arabic",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         preferred_value: "aeb",
         macrolanguage: "jrb"}).
iana(language,aju,tag{
         type: "language",
         subtag: "aju",
         description: "Judeo-Moroccan Arabic",
         added: "2009-07-29",
         macrolanguage: "jrb"}).
iana(language,ajw,tag{
         type: "language",
         subtag: "ajw",
         description: "Ajawa",
         added: "2009-07-29"}).
iana(language,ajz,tag{
         type: "language",
         subtag: "ajz",
         description: "Amri Karbi",
         added: "2009-07-29"}).
iana(language,akb,tag{
         type: "language",
         subtag: "akb",
         description: "Batak Angkola",
         added: "2009-07-29"}).
iana(language,akc,tag{
         type: "language",
         subtag: "akc",
         description: "Mpur",
         added: "2009-07-29"}).
iana(language,akd,tag{
         type: "language",
         subtag: "akd",
         description: "Ukpet-Ehom",
         added: "2009-07-29"}).
iana(language,ake,tag{
         type: "language",
         subtag: "ake",
         description: "Akawaio",
         added: "2009-07-29"}).
iana(language,akf,tag{
         type: "language",
         subtag: "akf",
         description: "Akpa",
         added: "2009-07-29"}).
iana(language,akg,tag{
         type: "language",
         subtag: "akg",
         description: "Anakalangu",
         added: "2009-07-29"}).
iana(language,akh,tag{
         type: "language",
         subtag: "akh",
         description: "Angal Heneng",
         added: "2009-07-29"}).
iana(language,aki,tag{
         type: "language",
         subtag: "aki",
         description: "Aiome",
         added: "2009-07-29"}).
iana(language,akj,tag{
         type: "language",
         subtag: "akj",
         description: "Aka-Jeru",
         added: "2009-07-29"}).
iana(language,akk,tag{
         type: "language",
         subtag: "akk",
         description: "Akkadian",
         added: "2005-10-16"}).
iana(language,akl,tag{
         type: "language",
         subtag: "akl",
         description: "Aklanon",
         added: "2009-07-29"}).
iana(language,akm,tag{
         type: "language",
         subtag: "akm",
         description: "Aka-Bo",
         added: "2009-07-29"}).
iana(language,ako,tag{
         type: "language",
         subtag: "ako",
         description: "Akurio",
         added: "2009-07-29"}).
iana(language,akp,tag{
         type: "language",
         subtag: "akp",
         description: "Siwu",
         added: "2009-07-29"}).
iana(language,akq,tag{
         type: "language",
         subtag: "akq",
         description: "Ak",
         added: "2009-07-29"}).
iana(language,akr,tag{
         type: "language",
         subtag: "akr",
         description: "Araki",
         added: "2009-07-29"}).
iana(language,aks,tag{
         type: "language",
         subtag: "aks",
         description: "Akaselem",
         added: "2009-07-29"}).
iana(language,akt,tag{
         type: "language",
         subtag: "akt",
         description: "Akolet",
         added: "2009-07-29"}).
iana(language,aku,tag{
         type: "language",
         subtag: "aku",
         description: "Akum",
         added: "2009-07-29"}).
iana(language,akv,tag{
         type: "language",
         subtag: "akv",
         description: "Akhvakh",
         added: "2009-07-29"}).
iana(language,akw,tag{
         type: "language",
         subtag: "akw",
         description: "Akwa",
         added: "2009-07-29"}).
iana(language,akx,tag{
         type: "language",
         subtag: "akx",
         description: "Aka-Kede",
         added: "2009-07-29"}).
iana(language,aky,tag{
         type: "language",
         subtag: "aky",
         description: "Aka-Kol",
         added: "2009-07-29"}).
iana(language,akz,tag{
         type: "language",
         subtag: "akz",
         description: "Alabama",
         added: "2009-07-29"}).
iana(language,ala,tag{
         type: "language",
         subtag: "ala",
         description: "Alago",
         added: "2009-07-29"}).
iana(language,alc,tag{
         type: "language",
         subtag: "alc",
         description: "Qawasqar",
         added: "2009-07-29"}).
iana(language,ald,tag{
         type: "language",
         subtag: "ald",
         description: "Alladian",
         added: "2009-07-29"}).
iana(language,ale,tag{
         type: "language",
         subtag: "ale",
         description: "Aleut",
         added: "2005-10-16"}).
iana(language,alf,tag{
         type: "language",
         subtag: "alf",
         description: "Alege",
         added: "2009-07-29"}).
iana(language,alg,tag{
         type: "language",
         subtag: "alg",
         description: "Algonquian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,alh,tag{
         type: "language",
         subtag: "alh",
         description: "Alawa",
         added: "2009-07-29"}).
iana(language,ali,tag{
         type: "language",
         subtag: "ali",
         description: "Amaimon",
         added: "2009-07-29"}).
iana(language,alj,tag{
         type: "language",
         subtag: "alj",
         description: "Alangan",
         added: "2009-07-29"}).
iana(language,alk,tag{
         type: "language",
         subtag: "alk",
         description: "Alak",
         added: "2009-07-29"}).
iana(language,all,tag{
         type: "language",
         subtag: "all",
         description: "Allar",
         added: "2009-07-29"}).
iana(language,alm,tag{
         type: "language",
         subtag: "alm",
         description: "Amblong",
         added: "2009-07-29"}).
iana(language,aln,tag{
         type: "language",
         subtag: "aln",
         description: "Gheg Albanian",
         added: "2009-07-29",
         macrolanguage: "sq"}).
iana(language,alo,tag{
         type: "language",
         subtag: "alo",
         description: "Larike-Wakasihu",
         added: "2009-07-29"}).
iana(language,alp,tag{
         type: "language",
         subtag: "alp",
         description: "Alune",
         added: "2009-07-29"}).
iana(language,alq,tag{
         type: "language",
         subtag: "alq",
         description: "Algonquin",
         added: "2009-07-29"}).
iana(language,alr,tag{
         type: "language",
         subtag: "alr",
         description: "Alutor",
         added: "2009-07-29"}).
iana(language,als,tag{
         type: "language",
         subtag: "als",
         description: "Tosk Albanian",
         added: "2009-07-29",
         macrolanguage: "sq"}).
iana(language,alt,tag{
         type: "language",
         subtag: "alt",
         description: "Southern Altai",
         added: "2005-10-16"}).
iana(language,alu,tag{
         type: "language",
         subtag: "alu",
         description: "'Are'are",
         added: "2009-07-29"}).
iana(language,alv,tag{
         type: "language",
         subtag: "alv",
         description: "Atlantic-Congo languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,alw,tag{
         type: "language",
         subtag: "alw",
         description: ["Alaba-K’abeena","Wanbasana"],
         added: "2009-07-29"}).
iana(language,alx,tag{
         type: "language",
         subtag: "alx",
         description: "Amol",
         added: "2009-07-29"}).
iana(language,aly,tag{
         type: "language",
         subtag: "aly",
         description: "Alyawarr",
         added: "2009-07-29"}).
iana(language,alz,tag{
         type: "language",
         subtag: "alz",
         description: "Alur",
         added: "2009-07-29"}).
iana(language,ama,tag{
         type: "language",
         subtag: "ama",
         description: "Amanayé",
         added: "2009-07-29"}).
iana(language,amb,tag{
         type: "language",
         subtag: "amb",
         description: "Ambo",
         added: "2009-07-29"}).
iana(language,amc,tag{
         type: "language",
         subtag: "amc",
         description: "Amahuaca",
         added: "2009-07-29"}).
iana(language,ame,tag{
         type: "language",
         subtag: "ame",
         description: "Yanesha'",
         added: "2009-07-29"}).
iana(language,amf,tag{
         type: "language",
         subtag: "amf",
         description: "Hamer-Banna",
         added: "2009-07-29"}).
iana(language,amg,tag{
         type: "language",
         subtag: "amg",
         description: "Amurdak",
         added: "2009-07-29"}).
iana(language,ami,tag{
         type: "language",
         subtag: "ami",
         description: "Amis",
         added: "2009-07-29"}).
iana(language,amj,tag{
         type: "language",
         subtag: "amj",
         description: "Amdang",
         added: "2009-07-29"}).
iana(language,amk,tag{
         type: "language",
         subtag: "amk",
         description: "Ambai",
         added: "2009-07-29"}).
iana(language,aml,tag{
         type: "language",
         subtag: "aml",
         description: "War-Jaintia",
         added: "2009-07-29"}).
iana(language,amm,tag{
         type: "language",
         subtag: "amm",
         description: "Ama (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,amn,tag{
         type: "language",
         subtag: "amn",
         description: "Amanab",
         added: "2009-07-29"}).
iana(language,amo,tag{
         type: "language",
         subtag: "amo",
         description: "Amo",
         added: "2009-07-29"}).
iana(language,amp,tag{
         type: "language",
         subtag: "amp",
         description: "Alamblak",
         added: "2009-07-29"}).
iana(language,amq,tag{
         type: "language",
         subtag: "amq",
         description: "Amahai",
         added: "2009-07-29"}).
iana(language,amr,tag{
         type: "language",
         subtag: "amr",
         description: "Amarakaeri",
         added: "2009-07-29"}).
iana(language,ams,tag{
         type: "language",
         subtag: "ams",
         description: "Southern Amami-Oshima",
         added: "2009-07-29"}).
iana(language,amt,tag{
         type: "language",
         subtag: "amt",
         description: "Amto",
         added: "2009-07-29"}).
iana(language,amu,tag{
         type: "language",
         subtag: "amu",
         description: "Guerrero Amuzgo",
         added: "2009-07-29"}).
iana(language,amv,tag{
         type: "language",
         subtag: "amv",
         description: "Ambelau",
         added: "2009-07-29"}).
iana(language,amw,tag{
         type: "language",
         subtag: "amw",
         description: "Western Neo-Aramaic",
         added: "2009-07-29"}).
iana(language,amx,tag{
         type: "language",
         subtag: "amx",
         description: "Anmatyerre",
         added: "2009-07-29"}).
iana(language,amy,tag{
         type: "language",
         subtag: "amy",
         description: "Ami",
         added: "2009-07-29"}).
iana(language,amz,tag{
         type: "language",
         subtag: "amz",
         description: "Atampaya",
         added: "2009-07-29"}).
iana(language,ana,tag{
         type: "language",
         subtag: "ana",
         description: "Andaqui",
         added: "2009-07-29"}).
iana(language,anb,tag{
         type: "language",
         subtag: "anb",
         description: "Andoa",
         added: "2009-07-29"}).
iana(language,anc,tag{
         type: "language",
         subtag: "anc",
         description: "Ngas",
         added: "2009-07-29"}).
iana(language,and,tag{
         type: "language",
         subtag: "and",
         description: "Ansus",
         added: "2009-07-29"}).
iana(language,ane,tag{
         type: "language",
         subtag: "ane",
         description: "Xârâcùù",
         added: "2009-07-29"}).
iana(language,anf,tag{
         type: "language",
         subtag: "anf",
         description: "Animere",
         added: "2009-07-29"}).
iana(language,ang,tag{
         type: "language",
         subtag: "ang",
         description: "Old English (ca. 450-1100)",
         added: "2005-10-16"}).
iana(language,anh,tag{
         type: "language",
         subtag: "anh",
         description: "Nend",
         added: "2009-07-29"}).
iana(language,ani,tag{
         type: "language",
         subtag: "ani",
         description: "Andi",
         added: "2009-07-29"}).
iana(language,anj,tag{
         type: "language",
         subtag: "anj",
         description: "Anor",
         added: "2009-07-29"}).
iana(language,ank,tag{
         type: "language",
         subtag: "ank",
         description: "Goemai",
         added: "2009-07-29"}).
iana(language,anl,tag{
         type: "language",
         subtag: "anl",
         description: "Anu-Hkongso Chin",
         added: "2009-07-29"}).
iana(language,anm,tag{
         type: "language",
         subtag: "anm",
         description: "Anal",
         added: "2009-07-29"}).
iana(language,ann,tag{
         type: "language",
         subtag: "ann",
         description: "Obolo",
         added: "2009-07-29"}).
iana(language,ano,tag{
         type: "language",
         subtag: "ano",
         description: "Andoque",
         added: "2009-07-29"}).
iana(language,anp,tag{
         type: "language",
         subtag: "anp",
         description: "Angika",
         added: "2006-03-08"}).
iana(language,anq,tag{
         type: "language",
         subtag: "anq",
         description: "Jarawa (India)",
         added: "2009-07-29"}).
iana(language,anr,tag{
         type: "language",
         subtag: "anr",
         description: "Andh",
         added: "2009-07-29"}).
iana(language,ans,tag{
         type: "language",
         subtag: "ans",
         description: "Anserma",
         added: "2009-07-29"}).
iana(language,ant,tag{
         type: "language",
         subtag: "ant",
         description: ["Antakarinya","Antikarinya"],
         added: "2009-07-29"}).
iana(language,anu,tag{
         type: "language",
         subtag: "anu",
         description: "Anuak",
         added: "2009-07-29"}).
iana(language,anv,tag{
         type: "language",
         subtag: "anv",
         description: "Denya",
         added: "2009-07-29"}).
iana(language,anw,tag{
         type: "language",
         subtag: "anw",
         description: "Anaang",
         added: "2009-07-29"}).
iana(language,anx,tag{
         type: "language",
         subtag: "anx",
         description: "Andra-Hus",
         added: "2009-07-29"}).
iana(language,any,tag{
         type: "language",
         subtag: "any",
         description: "Anyin",
         added: "2009-07-29"}).
iana(language,anz,tag{
         type: "language",
         subtag: "anz",
         description: "Anem",
         added: "2009-07-29"}).
iana(language,aoa,tag{
         type: "language",
         subtag: "aoa",
         description: "Angolar",
         added: "2009-07-29"}).
iana(language,aob,tag{
         type: "language",
         subtag: "aob",
         description: "Abom",
         added: "2009-07-29"}).
iana(language,aoc,tag{
         type: "language",
         subtag: "aoc",
         description: "Pemon",
         added: "2009-07-29"}).
iana(language,aod,tag{
         type: "language",
         subtag: "aod",
         description: "Andarum",
         added: "2009-07-29"}).
iana(language,aoe,tag{
         type: "language",
         subtag: "aoe",
         description: "Angal Enen",
         added: "2009-07-29"}).
iana(language,aof,tag{
         type: "language",
         subtag: "aof",
         description: "Bragat",
         added: "2009-07-29"}).
iana(language,aog,tag{
         type: "language",
         subtag: "aog",
         description: "Angoram",
         added: "2009-07-29"}).
iana(language,aoh,tag{
         type: "language",
         subtag: "aoh",
         description: "Arma",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,aoi,tag{
         type: "language",
         subtag: "aoi",
         description: "Anindilyakwa",
         added: "2009-07-29"}).
iana(language,aoj,tag{
         type: "language",
         subtag: "aoj",
         description: "Mufian",
         added: "2009-07-29"}).
iana(language,aok,tag{
         type: "language",
         subtag: "aok",
         description: "Arhö",
         added: "2009-07-29"}).
iana(language,aol,tag{
         type: "language",
         subtag: "aol",
         description: "Alor",
         added: "2009-07-29"}).
iana(language,aom,tag{
         type: "language",
         subtag: "aom",
         description: "Ömie",
         added: "2009-07-29"}).
iana(language,aon,tag{
         type: "language",
         subtag: "aon",
         description: "Bumbita Arapesh",
         added: "2009-07-29"}).
iana(language,aor,tag{
         type: "language",
         subtag: "aor",
         description: "Aore",
         added: "2009-07-29"}).
iana(language,aos,tag{
         type: "language",
         subtag: "aos",
         description: "Taikat",
         added: "2009-07-29"}).
iana(language,aot,tag{
         type: "language",
         subtag: "aot",
         description: ["Atong (India)","A'tong"],
         added: "2009-07-29"}).
iana(language,aou,tag{
         type: "language",
         subtag: "aou",
         description: "A'ou",
         added: "2012-08-12"}).
iana(language,aox,tag{
         type: "language",
         subtag: "aox",
         description: "Atorada",
         added: "2009-07-29"}).
iana(language,aoz,tag{
         type: "language",
         subtag: "aoz",
         description: "Uab Meto",
         added: "2009-07-29"}).
iana(language,apa,tag{
         type: "language",
         subtag: "apa",
         description: "Apache languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,apb,tag{
         type: "language",
         subtag: "apb",
         description: "Sa'a",
         added: "2009-07-29"}).
iana(language,apc,tag{
         type: "language",
         subtag: "apc",
         description: "North Levantine Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,apd,tag{
         type: "language",
         subtag: "apd",
         description: "Sudanese Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ape,tag{
         type: "language",
         subtag: "ape",
         description: "Bukiyip",
         added: "2009-07-29"}).
iana(language,apf,tag{
         type: "language",
         subtag: "apf",
         description: "Pahanan Agta",
         added: "2010-03-11"}).
iana(language,apg,tag{
         type: "language",
         subtag: "apg",
         description: "Ampanang",
         added: "2009-07-29"}).
iana(language,aph,tag{
         type: "language",
         subtag: "aph",
         description: "Athpariya",
         added: "2009-07-29"}).
iana(language,api,tag{
         type: "language",
         subtag: "api",
         description: "Apiaká",
         added: "2009-07-29"}).
iana(language,apj,tag{
         type: "language",
         subtag: "apj",
         description: "Jicarilla Apache",
         added: "2009-07-29"}).
iana(language,apk,tag{
         type: "language",
         subtag: "apk",
         description: "Kiowa Apache",
         added: "2009-07-29"}).
iana(language,apl,tag{
         type: "language",
         subtag: "apl",
         description: "Lipan Apache",
         added: "2009-07-29"}).
iana(language,apm,tag{
         type: "language",
         subtag: "apm",
         description: "Mescalero-Chiricahua Apache",
         added: "2009-07-29"}).
iana(language,apn,tag{
         type: "language",
         subtag: "apn",
         description: "Apinayé",
         added: "2009-07-29"}).
iana(language,apo,tag{
         type: "language",
         subtag: "apo",
         description: "Ambul",
         added: "2009-07-29"}).
iana(language,app,tag{
         type: "language",
         subtag: "app",
         description: "Apma",
         added: "2009-07-29"}).
iana(language,apq,tag{
         type: "language",
         subtag: "apq",
         description: "A-Pucikwar",
         added: "2009-07-29"}).
iana(language,apr,tag{
         type: "language",
         subtag: "apr",
         description: "Arop-Lokep",
         added: "2009-07-29"}).
iana(language,aps,tag{
         type: "language",
         subtag: "aps",
         description: "Arop-Sissano",
         added: "2009-07-29"}).
iana(language,apt,tag{
         type: "language",
         subtag: "apt",
         description: "Apatani",
         added: "2009-07-29"}).
iana(language,apu,tag{
         type: "language",
         subtag: "apu",
         description: "Apurinã",
         added: "2009-07-29"}).
iana(language,apv,tag{
         type: "language",
         subtag: "apv",
         description: "Alapmunte",
         added: "2009-07-29"}).
iana(language,apw,tag{
         type: "language",
         subtag: "apw",
         description: "Western Apache",
         added: "2009-07-29"}).
iana(language,apx,tag{
         type: "language",
         subtag: "apx",
         description: "Aputai",
         added: "2009-07-29"}).
iana(language,apy,tag{
         type: "language",
         subtag: "apy",
         description: "Apalaí",
         added: "2009-07-29"}).
iana(language,apz,tag{
         type: "language",
         subtag: "apz",
         description: "Safeyoka",
         added: "2009-07-29"}).
iana(language,aqa,tag{
         type: "language",
         subtag: "aqa",
         description: "Alacalufan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,aqc,tag{
         type: "language",
         subtag: "aqc",
         description: "Archi",
         added: "2009-07-29"}).
iana(language,aqd,tag{
         type: "language",
         subtag: "aqd",
         description: "Ampari Dogon",
         added: "2011-08-16"}).
iana(language,aqg,tag{
         type: "language",
         subtag: "aqg",
         description: "Arigidi",
         added: "2009-07-29"}).
iana(language,aqk,tag{
         type: "language",
         subtag: "aqk",
         description: "Aninka",
         added: "2021-02-20"}).
iana(language,aql,tag{
         type: "language",
         subtag: "aql",
         description: "Algic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,aqm,tag{
         type: "language",
         subtag: "aqm",
         description: "Atohwaim",
         added: "2009-07-29"}).
iana(language,aqn,tag{
         type: "language",
         subtag: "aqn",
         description: "Northern Alta",
         added: "2009-07-29"}).
iana(language,aqp,tag{
         type: "language",
         subtag: "aqp",
         description: "Atakapa",
         added: "2009-07-29"}).
iana(language,aqr,tag{
         type: "language",
         subtag: "aqr",
         description: "Arhâ",
         added: "2009-07-29"}).
iana(language,aqt,tag{
         type: "language",
         subtag: "aqt",
         description: "Angaité",
         added: "2014-02-28"}).
iana(language,aqz,tag{
         type: "language",
         subtag: "aqz",
         description: "Akuntsu",
         added: "2010-03-11"}).
iana(language,arb,tag{
         type: "language",
         subtag: "arb",
         description: "Standard Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,arc,tag{
         type: "language",
         subtag: "arc",
         description: ["Official Aramaic (700-300 BCE)","Imperial Aramaic (700-300 BCE)"],
         added: "2005-10-16"}).
iana(language,ard,tag{
         type: "language",
         subtag: "ard",
         description: "Arabana",
         added: "2009-07-29"}).
iana(language,are,tag{
         type: "language",
         subtag: "are",
         description: "Western Arrarnta",
         added: "2009-07-29"}).
iana(language,arh,tag{
         type: "language",
         subtag: "arh",
         description: "Arhuaco",
         added: "2009-07-29"}).
iana(language,ari,tag{
         type: "language",
         subtag: "ari",
         description: "Arikara",
         added: "2009-07-29"}).
iana(language,arj,tag{
         type: "language",
         subtag: "arj",
         description: "Arapaso",
         added: "2009-07-29"}).
iana(language,ark,tag{
         type: "language",
         subtag: "ark",
         description: "Arikapú",
         added: "2009-07-29"}).
iana(language,arl,tag{
         type: "language",
         subtag: "arl",
         description: "Arabela",
         added: "2009-07-29"}).
iana(language,arn,tag{
         type: "language",
         subtag: "arn",
         description: ["Mapudungun","Mapuche"],
         added: "2005-10-16"}).
iana(language,aro,tag{
         type: "language",
         subtag: "aro",
         description: "Araona",
         added: "2009-07-29"}).
iana(language,arp,tag{
         type: "language",
         subtag: "arp",
         description: "Arapaho",
         added: "2005-10-16"}).
iana(language,arq,tag{
         type: "language",
         subtag: "arq",
         description: "Algerian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,arr,tag{
         type: "language",
         subtag: "arr",
         description: "Karo (Brazil)",
         added: "2009-07-29"}).
iana(language,ars,tag{
         type: "language",
         subtag: "ars",
         description: "Najdi Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,art,tag{
         type: "language",
         subtag: "art",
         description: "Artificial languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,aru,tag{
         type: "language",
         subtag: "aru",
         description: ["Aruá (Amazonas State)","Arawá"],
         added: "2009-07-29"}).
iana(language,arv,tag{
         type: "language",
         subtag: "arv",
         description: "Arbore",
         added: "2009-07-29"}).
iana(language,arw,tag{
         type: "language",
         subtag: "arw",
         description: "Arawak",
         added: "2005-10-16"}).
iana(language,arx,tag{
         type: "language",
         subtag: "arx",
         description: "Aruá (Rodonia State)",
         added: "2009-07-29"}).
iana(language,ary,tag{
         type: "language",
         subtag: "ary",
         description: "Moroccan Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,arz,tag{
         type: "language",
         subtag: "arz",
         description: "Egyptian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,asa,tag{
         type: "language",
         subtag: "asa",
         description: "Asu (Tanzania)",
         added: "2009-07-29"}).
iana(language,asb,tag{
         type: "language",
         subtag: "asb",
         description: "Assiniboine",
         added: "2009-07-29"}).
iana(language,asc,tag{
         type: "language",
         subtag: "asc",
         description: "Casuarina Coast Asmat",
         added: "2009-07-29"}).
iana(language,asd,tag{
         type: "language",
         subtag: "asd",
         description: "Asas",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         preferred_value: "snz"}).
iana(language,ase,tag{
         type: "language",
         subtag: "ase",
         description: "American Sign Language",
         added: "2009-07-29"}).
iana(language,asf,tag{
         type: "language",
         subtag: "asf",
         description: ["Auslan","Australian Sign Language"],
         added: "2009-07-29"}).
iana(language,asg,tag{
         type: "language",
         subtag: "asg",
         description: "Cishingini",
         added: "2009-07-29"}).
iana(language,ash,tag{
         type: "language",
         subtag: "ash",
         description: "Abishira",
         added: "2009-07-29"}).
iana(language,asi,tag{
         type: "language",
         subtag: "asi",
         description: "Buruwai",
         added: "2009-07-29"}).
iana(language,asj,tag{
         type: "language",
         subtag: "asj",
         description: "Sari",
         added: "2009-07-29"}).
iana(language,ask,tag{
         type: "language",
         subtag: "ask",
         description: "Ashkun",
         added: "2009-07-29"}).
iana(language,asl,tag{
         type: "language",
         subtag: "asl",
         description: "Asilulu",
         added: "2009-07-29"}).
iana(language,asn,tag{
         type: "language",
         subtag: "asn",
         description: "Xingú Asuriní",
         added: "2009-07-29"}).
iana(language,aso,tag{
         type: "language",
         subtag: "aso",
         description: "Dano",
         added: "2009-07-29"}).
iana(language,asp,tag{
         type: "language",
         subtag: "asp",
         description: "Algerian Sign Language",
         added: "2009-07-29"}).
iana(language,asq,tag{
         type: "language",
         subtag: "asq",
         description: "Austrian Sign Language",
         added: "2009-07-29"}).
iana(language,asr,tag{
         type: "language",
         subtag: "asr",
         description: "Asuri",
         added: "2009-07-29"}).
iana(language,ass,tag{
         type: "language",
         subtag: "ass",
         description: "Ipulo",
         added: "2009-07-29"}).
iana(language,ast,tag{
         type: "language",
         subtag: "ast",
         description: ["Asturian","Asturleonese","Bable","Leonese"],
         added: "2005-10-16"}).
iana(language,asu,tag{
         type: "language",
         subtag: "asu",
         description: "Tocantins Asurini",
         added: "2009-07-29"}).
iana(language,asv,tag{
         type: "language",
         subtag: "asv",
         description: "Asoa",
         added: "2009-07-29"}).
iana(language,asw,tag{
         type: "language",
         subtag: "asw",
         description: "Australian Aborigines Sign Language",
         added: "2009-07-29"}).
iana(language,asx,tag{
         type: "language",
         subtag: "asx",
         description: "Muratayak",
         added: "2009-07-29"}).
iana(language,asy,tag{
         type: "language",
         subtag: "asy",
         description: "Yaosakor Asmat",
         added: "2009-07-29"}).
iana(language,asz,tag{
         type: "language",
         subtag: "asz",
         description: "As",
         added: "2009-07-29"}).
iana(language,ata,tag{
         type: "language",
         subtag: "ata",
         description: "Pele-Ata",
         added: "2009-07-29"}).
iana(language,atb,tag{
         type: "language",
         subtag: "atb",
         description: "Zaiwa",
         added: "2009-07-29"}).
iana(language,atc,tag{
         type: "language",
         subtag: "atc",
         description: "Atsahuaca",
         added: "2009-07-29"}).
iana(language,atd,tag{
         type: "language",
         subtag: "atd",
         description: "Ata Manobo",
         added: "2009-07-29"}).
iana(language,ate,tag{
         type: "language",
         subtag: "ate",
         description: "Atemble",
         added: "2009-07-29"}).
iana(language,atg,tag{
         type: "language",
         subtag: "atg",
         description: "Ivbie North-Okpela-Arhe",
         added: "2009-07-29"}).
iana(language,ath,tag{
         type: "language",
         subtag: "ath",
         description: "Athapascan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ati,tag{
         type: "language",
         subtag: "ati",
         description: "Attié",
         added: "2009-07-29"}).
iana(language,atj,tag{
         type: "language",
         subtag: "atj",
         description: "Atikamekw",
         added: "2009-07-29"}).
iana(language,atk,tag{
         type: "language",
         subtag: "atk",
         description: "Ati",
         added: "2009-07-29"}).
iana(language,atl,tag{
         type: "language",
         subtag: "atl",
         description: "Mt. Iraya Agta",
         added: "2009-07-29"}).
iana(language,atm,tag{
         type: "language",
         subtag: "atm",
         description: "Ata",
         added: "2009-07-29"}).
iana(language,atn,tag{
         type: "language",
         subtag: "atn",
         description: "Ashtiani",
         added: "2009-07-29"}).
iana(language,ato,tag{
         type: "language",
         subtag: "ato",
         description: "Atong (Cameroon)",
         added: "2009-07-29"}).
iana(language,atp,tag{
         type: "language",
         subtag: "atp",
         description: "Pudtol Atta",
         added: "2009-07-29"}).
iana(language,atq,tag{
         type: "language",
         subtag: "atq",
         description: "Aralle-Tabulahan",
         added: "2009-07-29"}).
iana(language,atr,tag{
         type: "language",
         subtag: "atr",
         description: "Waimiri-Atroari",
         added: "2009-07-29"}).
iana(language,ats,tag{
         type: "language",
         subtag: "ats",
         description: "Gros Ventre",
         added: "2009-07-29"}).
iana(language,att,tag{
         type: "language",
         subtag: "att",
         description: "Pamplona Atta",
         added: "2009-07-29"}).
iana(language,atu,tag{
         type: "language",
         subtag: "atu",
         description: "Reel",
         added: "2009-07-29"}).
iana(language,atv,tag{
         type: "language",
         subtag: "atv",
         description: "Northern Altai",
         added: "2009-07-29"}).
iana(language,atw,tag{
         type: "language",
         subtag: "atw",
         description: "Atsugewi",
         added: "2009-07-29"}).
iana(language,atx,tag{
         type: "language",
         subtag: "atx",
         description: "Arutani",
         added: "2009-07-29"}).
iana(language,aty,tag{
         type: "language",
         subtag: "aty",
         description: "Aneityum",
         added: "2009-07-29"}).
iana(language,atz,tag{
         type: "language",
         subtag: "atz",
         description: "Arta",
         added: "2009-07-29"}).
iana(language,aua,tag{
         type: "language",
         subtag: "aua",
         description: "Asumboa",
         added: "2009-07-29"}).
iana(language,aub,tag{
         type: "language",
         subtag: "aub",
         description: "Alugu",
         added: "2009-07-29"}).
iana(language,auc,tag{
         type: "language",
         subtag: "auc",
         description: "Waorani",
         added: "2009-07-29"}).
iana(language,aud,tag{
         type: "language",
         subtag: "aud",
         description: "Anuta",
         added: "2009-07-29"}).
iana(language,aue,tag{
         type: "language",
         subtag: "aue",
         description: "ǂKxʼauǁʼein",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "ktz"}).
iana(language,auf,tag{
         type: "language",
         subtag: "auf",
         description: "Arauan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,aug,tag{
         type: "language",
         subtag: "aug",
         description: "Aguna",
         added: "2009-07-29"}).
iana(language,auh,tag{
         type: "language",
         subtag: "auh",
         description: "Aushi",
         added: "2009-07-29"}).
iana(language,aui,tag{
         type: "language",
         subtag: "aui",
         description: "Anuki",
         added: "2009-07-29"}).
iana(language,auj,tag{
         type: "language",
         subtag: "auj",
         description: "Awjilah",
         added: "2009-07-29"}).
iana(language,auk,tag{
         type: "language",
         subtag: "auk",
         description: "Heyo",
         added: "2009-07-29"}).
iana(language,aul,tag{
         type: "language",
         subtag: "aul",
         description: "Aulua",
         added: "2009-07-29"}).
iana(language,aum,tag{
         type: "language",
         subtag: "aum",
         description: "Asu (Nigeria)",
         added: "2009-07-29"}).
iana(language,aun,tag{
         type: "language",
         subtag: "aun",
         description: "Molmo One",
         added: "2009-07-29"}).
iana(language,auo,tag{
         type: "language",
         subtag: "auo",
         description: "Auyokawa",
         added: "2009-07-29"}).
iana(language,aup,tag{
         type: "language",
         subtag: "aup",
         description: "Makayam",
         added: "2009-07-29"}).
iana(language,auq,tag{
         type: "language",
         subtag: "auq",
         description: ["Anus","Korur"],
         added: "2009-07-29"}).
iana(language,aur,tag{
         type: "language",
         subtag: "aur",
         description: "Aruek",
         added: "2009-07-29"}).
iana(language,aus,tag{
         type: "language",
         subtag: "aus",
         description: "Australian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,aut,tag{
         type: "language",
         subtag: "aut",
         description: "Austral",
         added: "2009-07-29"}).
iana(language,auu,tag{
         type: "language",
         subtag: "auu",
         description: "Auye",
         added: "2009-07-29"}).
iana(language,auw,tag{
         type: "language",
         subtag: "auw",
         description: "Awyi",
         added: "2009-07-29"}).
iana(language,aux,tag{
         type: "language",
         subtag: "aux",
         description: "Aurá",
         added: "2009-07-29"}).
iana(language,auy,tag{
         type: "language",
         subtag: "auy",
         description: "Awiyaana",
         added: "2009-07-29"}).
iana(language,auz,tag{
         type: "language",
         subtag: "auz",
         description: "Uzbeki Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,avb,tag{
         type: "language",
         subtag: "avb",
         description: "Avau",
         added: "2009-07-29"}).
iana(language,avd,tag{
         type: "language",
         subtag: "avd",
         description: "Alviri-Vidari",
         added: "2009-07-29"}).
iana(language,avi,tag{
         type: "language",
         subtag: "avi",
         description: "Avikam",
         added: "2009-07-29"}).
iana(language,avk,tag{
         type: "language",
         subtag: "avk",
         description: "Kotava",
         added: "2009-07-29"}).
iana(language,avl,tag{
         type: "language",
         subtag: "avl",
         description: "Eastern Egyptian Bedawi Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,avm,tag{
         type: "language",
         subtag: "avm",
         description: "Angkamuthi",
         added: "2012-08-12"}).
iana(language,avn,tag{
         type: "language",
         subtag: "avn",
         description: "Avatime",
         added: "2009-07-29"}).
iana(language,avo,tag{
         type: "language",
         subtag: "avo",
         description: "Agavotaguerra",
         added: "2009-07-29"}).
iana(language,avs,tag{
         type: "language",
         subtag: "avs",
         description: "Aushiri",
         added: "2009-07-29"}).
iana(language,avt,tag{
         type: "language",
         subtag: "avt",
         description: "Au",
         added: "2009-07-29"}).
iana(language,avu,tag{
         type: "language",
         subtag: "avu",
         description: "Avokaya",
         added: "2009-07-29"}).
iana(language,avv,tag{
         type: "language",
         subtag: "avv",
         description: "Avá-Canoeiro",
         added: "2009-07-29"}).
iana(language,awa,tag{
         type: "language",
         subtag: "awa",
         description: "Awadhi",
         added: "2005-10-16"}).
iana(language,awb,tag{
         type: "language",
         subtag: "awb",
         description: "Awa (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,awc,tag{
         type: "language",
         subtag: "awc",
         description: "Cicipu",
         added: "2009-07-29"}).
iana(language,awd,tag{
         type: "language",
         subtag: "awd",
         description: "Arawakan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,awe,tag{
         type: "language",
         subtag: "awe",
         description: "Awetí",
         added: "2009-07-29"}).
iana(language,awg,tag{
         type: "language",
         subtag: "awg",
         description: "Anguthimri",
         added: "2012-08-12"}).
iana(language,awh,tag{
         type: "language",
         subtag: "awh",
         description: "Awbono",
         added: "2009-07-29"}).
iana(language,awi,tag{
         type: "language",
         subtag: "awi",
         description: "Aekyom",
         added: "2009-07-29"}).
iana(language,awk,tag{
         type: "language",
         subtag: "awk",
         description: "Awabakal",
         added: "2009-07-29"}).
iana(language,awm,tag{
         type: "language",
         subtag: "awm",
         description: "Arawum",
         added: "2009-07-29"}).
iana(language,awn,tag{
         type: "language",
         subtag: "awn",
         description: "Awngi",
         added: "2009-07-29"}).
iana(language,awo,tag{
         type: "language",
         subtag: "awo",
         description: "Awak",
         added: "2009-07-29"}).
iana(language,awr,tag{
         type: "language",
         subtag: "awr",
         description: "Awera",
         added: "2009-07-29"}).
iana(language,aws,tag{
         type: "language",
         subtag: "aws",
         description: "South Awyu",
         added: "2009-07-29"}).
iana(language,awt,tag{
         type: "language",
         subtag: "awt",
         description: "Araweté",
         added: "2009-07-29"}).
iana(language,awu,tag{
         type: "language",
         subtag: "awu",
         description: "Central Awyu",
         added: "2009-07-29"}).
iana(language,awv,tag{
         type: "language",
         subtag: "awv",
         description: "Jair Awyu",
         added: "2009-07-29"}).
iana(language,aww,tag{
         type: "language",
         subtag: "aww",
         description: "Awun",
         added: "2009-07-29"}).
iana(language,awx,tag{
         type: "language",
         subtag: "awx",
         description: "Awara",
         added: "2009-07-29"}).
iana(language,awy,tag{
         type: "language",
         subtag: "awy",
         description: "Edera Awyu",
         added: "2009-07-29"}).
iana(language,axb,tag{
         type: "language",
         subtag: "axb",
         description: "Abipon",
         added: "2009-07-29"}).
iana(language,axe,tag{
         type: "language",
         subtag: "axe",
         description: "Ayerrerenge",
         added: "2012-08-12"}).
iana(language,axg,tag{
         type: "language",
         subtag: "axg",
         description: "Mato Grosso Arára",
         added: "2009-07-29"}).
iana(language,axk,tag{
         type: "language",
         subtag: "axk",
         description: "Yaka (Central African Republic)",
         added: "2009-07-29"}).
iana(language,axl,tag{
         type: "language",
         subtag: "axl",
         description: "Lower Southern Aranda",
         added: "2013-09-10"}).
iana(language,axm,tag{
         type: "language",
         subtag: "axm",
         description: "Middle Armenian",
         added: "2009-07-29"}).
iana(language,axx,tag{
         type: "language",
         subtag: "axx",
         description: "Xârâgurè",
         added: "2009-07-29"}).
iana(language,aya,tag{
         type: "language",
         subtag: "aya",
         description: "Awar",
         added: "2009-07-29"}).
iana(language,ayb,tag{
         type: "language",
         subtag: "ayb",
         description: "Ayizo Gbe",
         added: "2009-07-29"}).
iana(language,ayc,tag{
         type: "language",
         subtag: "ayc",
         description: "Southern Aymara",
         added: "2009-07-29",
         macrolanguage: "ay"}).
iana(language,ayd,tag{
         type: "language",
         subtag: "ayd",
         description: "Ayabadhu",
         added: "2009-07-29"}).
iana(language,aye,tag{
         type: "language",
         subtag: "aye",
         description: "Ayere",
         added: "2009-07-29"}).
iana(language,ayg,tag{
         type: "language",
         subtag: "ayg",
         description: "Ginyanga",
         added: "2009-07-29"}).
iana(language,ayh,tag{
         type: "language",
         subtag: "ayh",
         description: "Hadrami Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ayi,tag{
         type: "language",
         subtag: "ayi",
         description: "Leyigha",
         added: "2009-07-29"}).
iana(language,ayk,tag{
         type: "language",
         subtag: "ayk",
         description: "Akuku",
         added: "2009-07-29"}).
iana(language,ayl,tag{
         type: "language",
         subtag: "ayl",
         description: "Libyan Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ayn,tag{
         type: "language",
         subtag: "ayn",
         description: "Sanaani Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ayo,tag{
         type: "language",
         subtag: "ayo",
         description: "Ayoreo",
         added: "2009-07-29"}).
iana(language,ayp,tag{
         type: "language",
         subtag: "ayp",
         description: "North Mesopotamian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ayq,tag{
         type: "language",
         subtag: "ayq",
         description: "Ayi (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,ayr,tag{
         type: "language",
         subtag: "ayr",
         description: "Central Aymara",
         added: "2009-07-29",
         macrolanguage: "ay"}).
iana(language,ays,tag{
         type: "language",
         subtag: "ays",
         description: "Sorsogon Ayta",
         added: "2009-07-29"}).
iana(language,ayt,tag{
         type: "language",
         subtag: "ayt",
         description: "Magbukun Ayta",
         added: "2009-07-29"}).
iana(language,ayu,tag{
         type: "language",
         subtag: "ayu",
         description: "Ayu",
         added: "2009-07-29"}).
iana(language,ayx,tag{
         type: "language",
         subtag: "ayx",
         description: "Ayi (China)",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         preferred_value: "nun"}).
iana(language,ayy,tag{
         type: "language",
         subtag: "ayy",
         description: "Tayabas Ayta",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,ayz,tag{
         type: "language",
         subtag: "ayz",
         description: "Mai Brat",
         added: "2009-07-29"}).
iana(language,aza,tag{
         type: "language",
         subtag: "aza",
         description: "Azha",
         added: "2009-07-29"}).
iana(language,azb,tag{
         type: "language",
         subtag: "azb",
         description: "South Azerbaijani",
         added: "2009-07-29",
         macrolanguage: "az"}).
iana(language,azc,tag{
         type: "language",
         subtag: "azc",
         description: "Uto-Aztecan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,azd,tag{
         type: "language",
         subtag: "azd",
         description: "Eastern Durango Nahuatl",
         added: "2012-08-12"}).
iana(language,azg,tag{
         type: "language",
         subtag: "azg",
         description: "San Pedro Amuzgos Amuzgo",
         added: "2009-07-29"}).
iana(language,azj,tag{
         type: "language",
         subtag: "azj",
         description: "North Azerbaijani",
         added: "2009-07-29",
         macrolanguage: "az"}).
iana(language,azm,tag{
         type: "language",
         subtag: "azm",
         description: "Ipalapa Amuzgo",
         added: "2009-07-29"}).
iana(language,azn,tag{
         type: "language",
         subtag: "azn",
         description: "Western Durango Nahuatl",
         added: "2012-08-12"}).
iana(language,azo,tag{
         type: "language",
         subtag: "azo",
         description: "Awing",
         added: "2009-07-29"}).
iana(language,azt,tag{
         type: "language",
         subtag: "azt",
         description: "Faire Atta",
         added: "2009-07-29"}).
iana(language,azz,tag{
         type: "language",
         subtag: "azz",
         description: "Highland Puebla Nahuatl",
         added: "2009-07-29"}).
iana(language,baa,tag{
         type: "language",
         subtag: "baa",
         description: "Babatana",
         added: "2009-07-29"}).
iana(language,bab,tag{
         type: "language",
         subtag: "bab",
         description: "Bainouk-Gunyuño",
         added: "2009-07-29"}).
iana(language,bac,tag{
         type: "language",
         subtag: "bac",
         description: "Badui",
         added: "2009-07-29"}).
iana(language,bad,tag{
         type: "language",
         subtag: "bad",
         description: "Banda languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,bae,tag{
         type: "language",
         subtag: "bae",
         description: "Baré",
         added: "2009-07-29"}).
iana(language,baf,tag{
         type: "language",
         subtag: "baf",
         description: "Nubaca",
         added: "2009-07-29"}).
iana(language,bag,tag{
         type: "language",
         subtag: "bag",
         description: "Tuki",
         added: "2009-07-29"}).
iana(language,bah,tag{
         type: "language",
         subtag: "bah",
         description: "Bahamas Creole English",
         added: "2009-07-29"}).
iana(language,bai,tag{
         type: "language",
         subtag: "bai",
         description: "Bamileke languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,baj,tag{
         type: "language",
         subtag: "baj",
         description: "Barakai",
         added: "2009-07-29"}).
iana(language,bal,tag{
         type: "language",
         subtag: "bal",
         description: "Baluchi",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,ban,tag{
         type: "language",
         subtag: "ban",
         description: "Balinese",
         added: "2005-10-16"}).
iana(language,bao,tag{
         type: "language",
         subtag: "bao",
         description: "Waimaha",
         added: "2009-07-29"}).
iana(language,bap,tag{
         type: "language",
         subtag: "bap",
         description: "Bantawa",
         added: "2009-07-29"}).
iana(language,bar,tag{
         type: "language",
         subtag: "bar",
         description: "Bavarian",
         added: "2009-07-29"}).
iana(language,bas,tag{
         type: "language",
         subtag: "bas",
         description: "Basa (Cameroon)",
         added: "2005-10-16"}).
iana(language,bat,tag{
         type: "language",
         subtag: "bat",
         description: "Baltic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,bau,tag{
         type: "language",
         subtag: "bau",
         description: "Bada (Nigeria)",
         added: "2009-07-29"}).
iana(language,bav,tag{
         type: "language",
         subtag: "bav",
         description: "Vengo",
         added: "2009-07-29"}).
iana(language,baw,tag{
         type: "language",
         subtag: "baw",
         description: "Bambili-Bambui",
         added: "2009-07-29"}).
iana(language,bax,tag{
         type: "language",
         subtag: "bax",
         description: "Bamun",
         added: "2009-07-29"}).
iana(language,bay,tag{
         type: "language",
         subtag: "bay",
         description: "Batuley",
         added: "2009-07-29"}).
iana(language,baz,tag{
         type: "language",
         subtag: "baz",
         description: "Tunen",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see nvo, tvu"}).
iana(language,bba,tag{
         type: "language",
         subtag: "bba",
         description: "Baatonum",
         added: "2009-07-29"}).
iana(language,bbb,tag{
         type: "language",
         subtag: "bbb",
         description: "Barai",
         added: "2009-07-29"}).
iana(language,bbc,tag{
         type: "language",
         subtag: "bbc",
         description: "Batak Toba",
         added: "2009-07-29"}).
iana(language,bbd,tag{
         type: "language",
         subtag: "bbd",
         description: "Bau",
         added: "2009-07-29"}).
iana(language,bbe,tag{
         type: "language",
         subtag: "bbe",
         description: "Bangba",
         added: "2009-07-29"}).
iana(language,bbf,tag{
         type: "language",
         subtag: "bbf",
         description: "Baibai",
         added: "2009-07-29"}).
iana(language,bbg,tag{
         type: "language",
         subtag: "bbg",
         description: "Barama",
         added: "2009-07-29"}).
iana(language,bbh,tag{
         type: "language",
         subtag: "bbh",
         description: "Bugan",
         added: "2009-07-29"}).
iana(language,bbi,tag{
         type: "language",
         subtag: "bbi",
         description: "Barombi",
         added: "2009-07-29"}).
iana(language,bbj,tag{
         type: "language",
         subtag: "bbj",
         description: "Ghomálá'",
         added: "2009-07-29"}).
iana(language,bbk,tag{
         type: "language",
         subtag: "bbk",
         description: "Babanki",
         added: "2009-07-29"}).
iana(language,bbl,tag{
         type: "language",
         subtag: "bbl",
         description: "Bats",
         added: "2009-07-29"}).
iana(language,bbm,tag{
         type: "language",
         subtag: "bbm",
         description: "Babango",
         added: "2009-07-29"}).
iana(language,bbn,tag{
         type: "language",
         subtag: "bbn",
         description: "Uneapa",
         added: "2009-07-29"}).
iana(language,bbo,tag{
         type: "language",
         subtag: "bbo",
         description: ["Northern Bobo Madaré","Konabéré"],
         added: "2009-07-29"}).
iana(language,bbp,tag{
         type: "language",
         subtag: "bbp",
         description: "West Central Banda",
         added: "2009-07-29"}).
iana(language,bbq,tag{
         type: "language",
         subtag: "bbq",
         description: "Bamali",
         added: "2009-07-29"}).
iana(language,bbr,tag{
         type: "language",
         subtag: "bbr",
         description: "Girawa",
         added: "2009-07-29"}).
iana(language,bbs,tag{
         type: "language",
         subtag: "bbs",
         description: "Bakpinka",
         added: "2009-07-29"}).
iana(language,bbt,tag{
         type: "language",
         subtag: "bbt",
         description: "Mburku",
         added: "2009-07-29"}).
iana(language,bbu,tag{
         type: "language",
         subtag: "bbu",
         description: "Kulung (Nigeria)",
         added: "2009-07-29"}).
iana(language,bbv,tag{
         type: "language",
         subtag: "bbv",
         description: "Karnai",
         added: "2009-07-29"}).
iana(language,bbw,tag{
         type: "language",
         subtag: "bbw",
         description: "Baba",
         added: "2009-07-29"}).
iana(language,bbx,tag{
         type: "language",
         subtag: "bbx",
         description: "Bubia",
         added: "2009-07-29"}).
iana(language,bby,tag{
         type: "language",
         subtag: "bby",
         description: "Befang",
         added: "2009-07-29"}).
iana(language,bbz,tag{
         type: "language",
         subtag: "bbz",
         description: "Babalia Creole Arabic",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         macrolanguage: "ar"}).
iana(language,bca,tag{
         type: "language",
         subtag: "bca",
         description: "Central Bai",
         added: "2009-07-29"}).
iana(language,bcb,tag{
         type: "language",
         subtag: "bcb",
         description: "Bainouk-Samik",
         added: "2009-07-29"}).
iana(language,bcc,tag{
         type: "language",
         subtag: "bcc",
         description: "Southern Balochi",
         added: "2009-07-29",
         macrolanguage: "bal"}).
iana(language,bcd,tag{
         type: "language",
         subtag: "bcd",
         description: "North Babar",
         added: "2009-07-29"}).
iana(language,bce,tag{
         type: "language",
         subtag: "bce",
         description: "Bamenyam",
         added: "2009-07-29"}).
iana(language,bcf,tag{
         type: "language",
         subtag: "bcf",
         description: "Bamu",
         added: "2009-07-29"}).
iana(language,bcg,tag{
         type: "language",
         subtag: "bcg",
         description: "Baga Pokur",
         added: "2009-07-29"}).
iana(language,bch,tag{
         type: "language",
         subtag: "bch",
         description: "Bariai",
         added: "2009-07-29"}).
iana(language,bci,tag{
         type: "language",
         subtag: "bci",
         description: "Baoulé",
         added: "2009-07-29"}).
iana(language,bcj,tag{
         type: "language",
         subtag: "bcj",
         description: "Bardi",
         added: "2009-07-29"}).
iana(language,bck,tag{
         type: "language",
         subtag: "bck",
         description: "Bunuba",
         added: "2009-07-29"}).
iana(language,bcl,tag{
         type: "language",
         subtag: "bcl",
         description: "Central Bikol",
         added: "2009-07-29",
         macrolanguage: "bik"}).
iana(language,bcm,tag{
         type: "language",
         subtag: "bcm",
         description: "Bannoni",
         added: "2009-07-29"}).
iana(language,bcn,tag{
         type: "language",
         subtag: "bcn",
         description: "Bali (Nigeria)",
         added: "2009-07-29"}).
iana(language,bco,tag{
         type: "language",
         subtag: "bco",
         description: "Kaluli",
         added: "2009-07-29"}).
iana(language,bcp,tag{
         type: "language",
         subtag: "bcp",
         description: "Bali (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,bcq,tag{
         type: "language",
         subtag: "bcq",
         description: "Bench",
         added: "2009-07-29"}).
iana(language,bcr,tag{
         type: "language",
         subtag: "bcr",
         description: "Babine",
         added: "2009-07-29"}).
iana(language,bcs,tag{
         type: "language",
         subtag: "bcs",
         description: "Kohumono",
         added: "2009-07-29"}).
iana(language,bct,tag{
         type: "language",
         subtag: "bct",
         description: "Bendi",
         added: "2009-07-29"}).
iana(language,bcu,tag{
         type: "language",
         subtag: "bcu",
         description: "Awad Bing",
         added: "2009-07-29"}).
iana(language,bcv,tag{
         type: "language",
         subtag: "bcv",
         description: "Shoo-Minda-Nye",
         added: "2009-07-29"}).
iana(language,bcw,tag{
         type: "language",
         subtag: "bcw",
         description: "Bana",
         added: "2009-07-29"}).
iana(language,bcy,tag{
         type: "language",
         subtag: "bcy",
         description: "Bacama",
         added: "2009-07-29"}).
iana(language,bcz,tag{
         type: "language",
         subtag: "bcz",
         description: "Bainouk-Gunyaamolo",
         added: "2009-07-29"}).
iana(language,bda,tag{
         type: "language",
         subtag: "bda",
         description: "Bayot",
         added: "2009-07-29"}).
iana(language,bdb,tag{
         type: "language",
         subtag: "bdb",
         description: "Basap",
         added: "2009-07-29"}).
iana(language,bdc,tag{
         type: "language",
         subtag: "bdc",
         description: "Emberá-Baudó",
         added: "2009-07-29"}).
iana(language,bdd,tag{
         type: "language",
         subtag: "bdd",
         description: "Bunama",
         added: "2009-07-29"}).
iana(language,bde,tag{
         type: "language",
         subtag: "bde",
         description: "Bade",
         added: "2009-07-29"}).
iana(language,bdf,tag{
         type: "language",
         subtag: "bdf",
         description: "Biage",
         added: "2010-03-11"}).
iana(language,bdg,tag{
         type: "language",
         subtag: "bdg",
         description: "Bonggi",
         added: "2009-07-29"}).
iana(language,bdh,tag{
         type: "language",
         subtag: "bdh",
         description: "Baka (South Sudan)",
         added: "2009-07-29"}).
iana(language,bdi,tag{
         type: "language",
         subtag: "bdi",
         description: "Burun",
         added: "2009-07-29"}).
iana(language,bdj,tag{
         type: "language",
         subtag: "bdj",
         description: ["Bai (South Sudan)","Bai"],
         added: "2009-07-29"}).
iana(language,bdk,tag{
         type: "language",
         subtag: "bdk",
         description: "Budukh",
         added: "2009-07-29"}).
iana(language,bdl,tag{
         type: "language",
         subtag: "bdl",
         description: "Indonesian Bajau",
         added: "2009-07-29"}).
iana(language,bdm,tag{
         type: "language",
         subtag: "bdm",
         description: "Buduma",
         added: "2009-07-29"}).
iana(language,bdn,tag{
         type: "language",
         subtag: "bdn",
         description: "Baldemu",
         added: "2009-07-29"}).
iana(language,bdo,tag{
         type: "language",
         subtag: "bdo",
         description: "Morom",
         added: "2009-07-29"}).
iana(language,bdp,tag{
         type: "language",
         subtag: "bdp",
         description: "Bende",
         added: "2009-07-29"}).
iana(language,bdq,tag{
         type: "language",
         subtag: "bdq",
         description: "Bahnar",
         added: "2009-07-29"}).
iana(language,bdr,tag{
         type: "language",
         subtag: "bdr",
         description: "West Coast Bajau",
         added: "2009-07-29"}).
iana(language,bds,tag{
         type: "language",
         subtag: "bds",
         description: "Burunge",
         added: "2009-07-29"}).
iana(language,bdt,tag{
         type: "language",
         subtag: "bdt",
         description: "Bokoto",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,bdu,tag{
         type: "language",
         subtag: "bdu",
         description: "Oroko",
         added: "2009-07-29"}).
iana(language,bdv,tag{
         type: "language",
         subtag: "bdv",
         description: "Bodo Parja",
         added: "2009-07-29"}).
iana(language,bdw,tag{
         type: "language",
         subtag: "bdw",
         description: "Baham",
         added: "2009-07-29"}).
iana(language,bdx,tag{
         type: "language",
         subtag: "bdx",
         description: "Budong-Budong",
         added: "2009-07-29"}).
iana(language,bdy,tag{
         type: "language",
         subtag: "bdy",
         description: "Bandjalang",
         added: "2009-07-29"}).
iana(language,bdz,tag{
         type: "language",
         subtag: "bdz",
         description: "Badeshi",
         added: "2009-07-29"}).
iana(language,bea,tag{
         type: "language",
         subtag: "bea",
         description: "Beaver",
         added: "2009-07-29"}).
iana(language,beb,tag{
         type: "language",
         subtag: "beb",
         description: "Bebele",
         added: "2009-07-29"}).
iana(language,bec,tag{
         type: "language",
         subtag: "bec",
         description: "Iceve-Maci",
         added: "2009-07-29"}).
iana(language,bed,tag{
         type: "language",
         subtag: "bed",
         description: "Bedoanas",
         added: "2009-07-29"}).
iana(language,bee,tag{
         type: "language",
         subtag: "bee",
         description: "Byangsi",
         added: "2009-07-29"}).
iana(language,bef,tag{
         type: "language",
         subtag: "bef",
         description: "Benabena",
         added: "2009-07-29"}).
iana(language,beg,tag{
         type: "language",
         subtag: "beg",
         description: "Belait",
         added: "2009-07-29"}).
iana(language,beh,tag{
         type: "language",
         subtag: "beh",
         description: "Biali",
         added: "2009-07-29"}).
iana(language,bei,tag{
         type: "language",
         subtag: "bei",
         description: "Bekati'",
         added: "2009-07-29"}).
iana(language,bej,tag{
         type: "language",
         subtag: "bej",
         description: ["Beja","Bedawiyet"],
         added: "2005-10-16"}).
iana(language,bek,tag{
         type: "language",
         subtag: "bek",
         description: "Bebeli",
         added: "2009-07-29"}).
iana(language,bem,tag{
         type: "language",
         subtag: "bem",
         description: "Bemba (Zambia)",
         added: "2005-10-16"}).
iana(language,beo,tag{
         type: "language",
         subtag: "beo",
         description: "Beami",
         added: "2009-07-29"}).
iana(language,bep,tag{
         type: "language",
         subtag: "bep",
         description: "Besoa",
         added: "2009-07-29"}).
iana(language,beq,tag{
         type: "language",
         subtag: "beq",
         description: "Beembe",
         added: "2009-07-29"}).
iana(language,ber,tag{
         type: "language",
         subtag: "ber",
         description: "Berber languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,bes,tag{
         type: "language",
         subtag: "bes",
         description: "Besme",
         added: "2009-07-29"}).
iana(language,bet,tag{
         type: "language",
         subtag: "bet",
         description: "Guiberoua Béte",
         added: "2009-07-29"}).
iana(language,beu,tag{
         type: "language",
         subtag: "beu",
         description: "Blagar",
         added: "2009-07-29"}).
iana(language,bev,tag{
         type: "language",
         subtag: "bev",
         description: "Daloa Bété",
         added: "2009-07-29"}).
iana(language,bew,tag{
         type: "language",
         subtag: "bew",
         description: "Betawi",
         added: "2009-07-29"}).
iana(language,bex,tag{
         type: "language",
         subtag: "bex",
         description: "Jur Modo",
         added: "2009-07-29"}).
iana(language,bey,tag{
         type: "language",
         subtag: "bey",
         description: "Beli (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,bez,tag{
         type: "language",
         subtag: "bez",
         description: "Bena (Tanzania)",
         added: "2009-07-29"}).
iana(language,bfa,tag{
         type: "language",
         subtag: "bfa",
         description: "Bari",
         added: "2009-07-29"}).
iana(language,bfb,tag{
         type: "language",
         subtag: "bfb",
         description: "Pauri Bareli",
         added: "2009-07-29"}).
iana(language,bfc,tag{
         type: "language",
         subtag: "bfc",
         description: ["Panyi Bai","Northern Bai"],
         added: "2009-07-29"}).
iana(language,bfd,tag{
         type: "language",
         subtag: "bfd",
         description: "Bafut",
         added: "2009-07-29"}).
iana(language,bfe,tag{
         type: "language",
         subtag: "bfe",
         description: ["Betaf","Tena"],
         added: "2009-07-29"}).
iana(language,bff,tag{
         type: "language",
         subtag: "bff",
         description: "Bofi",
         added: "2009-07-29"}).
iana(language,bfg,tag{
         type: "language",
         subtag: "bfg",
         description: "Busang Kayan",
         added: "2009-07-29"}).
iana(language,bfh,tag{
         type: "language",
         subtag: "bfh",
         description: "Blafe",
         added: "2009-07-29"}).
iana(language,bfi,tag{
         type: "language",
         subtag: "bfi",
         description: "British Sign Language",
         added: "2009-07-29"}).
iana(language,bfj,tag{
         type: "language",
         subtag: "bfj",
         description: "Bafanji",
         added: "2009-07-29"}).
iana(language,bfk,tag{
         type: "language",
         subtag: "bfk",
         description: "Ban Khor Sign Language",
         added: "2009-07-29"}).
iana(language,bfl,tag{
         type: "language",
         subtag: "bfl",
         description: "Banda-Ndélé",
         added: "2009-07-29"}).
iana(language,bfm,tag{
         type: "language",
         subtag: "bfm",
         description: "Mmen",
         added: "2009-07-29"}).
iana(language,bfn,tag{
         type: "language",
         subtag: "bfn",
         description: "Bunak",
         added: "2009-07-29"}).
iana(language,bfo,tag{
         type: "language",
         subtag: "bfo",
         description: "Malba Birifor",
         added: "2009-07-29"}).
iana(language,bfp,tag{
         type: "language",
         subtag: "bfp",
         description: "Beba",
         added: "2009-07-29"}).
iana(language,bfq,tag{
         type: "language",
         subtag: "bfq",
         description: "Badaga",
         added: "2009-07-29"}).
iana(language,bfr,tag{
         type: "language",
         subtag: "bfr",
         description: "Bazigar",
         added: "2009-07-29"}).
iana(language,bfs,tag{
         type: "language",
         subtag: "bfs",
         description: "Southern Bai",
         added: "2009-07-29"}).
iana(language,bft,tag{
         type: "language",
         subtag: "bft",
         description: "Balti",
         added: "2009-07-29"}).
iana(language,bfu,tag{
         type: "language",
         subtag: "bfu",
         description: "Gahri",
         added: "2009-07-29"}).
iana(language,bfw,tag{
         type: "language",
         subtag: "bfw",
         description: "Bondo",
         added: "2009-07-29"}).
iana(language,bfx,tag{
         type: "language",
         subtag: "bfx",
         description: "Bantayanon",
         added: "2010-03-11"}).
iana(language,bfy,tag{
         type: "language",
         subtag: "bfy",
         description: "Bagheli",
         added: "2009-07-29"}).
iana(language,bfz,tag{
         type: "language",
         subtag: "bfz",
         description: "Mahasu Pahari",
         added: "2009-07-29"}).
iana(language,bga,tag{
         type: "language",
         subtag: "bga",
         description: "Gwamhi-Wuri",
         added: "2009-07-29"}).
iana(language,bgb,tag{
         type: "language",
         subtag: "bgb",
         description: "Bobongko",
         added: "2009-07-29"}).
iana(language,bgc,tag{
         type: "language",
         subtag: "bgc",
         description: "Haryanvi",
         added: "2009-07-29"}).
iana(language,bgd,tag{
         type: "language",
         subtag: "bgd",
         description: "Rathwi Bareli",
         added: "2009-07-29"}).
iana(language,bge,tag{
         type: "language",
         subtag: "bge",
         description: "Bauria",
         added: "2009-07-29"}).
iana(language,bgf,tag{
         type: "language",
         subtag: "bgf",
         description: "Bangandu",
         added: "2009-07-29"}).
iana(language,bgg,tag{
         type: "language",
         subtag: "bgg",
         description: "Bugun",
         added: "2009-07-29"}).
iana(language,bgi,tag{
         type: "language",
         subtag: "bgi",
         description: "Giangan",
         added: "2009-07-29"}).
iana(language,bgj,tag{
         type: "language",
         subtag: "bgj",
         description: "Bangolan",
         added: "2009-07-29"}).
iana(language,bgk,tag{
         type: "language",
         subtag: "bgk",
         description: ["Bit","Buxinhua"],
         added: "2009-07-29"}).
iana(language,bgl,tag{
         type: "language",
         subtag: "bgl",
         description: "Bo (Laos)",
         added: "2009-07-29"}).
iana(language,bgm,tag{
         type: "language",
         subtag: "bgm",
         description: "Baga Mboteni",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "bcg"}).
iana(language,bgn,tag{
         type: "language",
         subtag: "bgn",
         description: "Western Balochi",
         added: "2009-07-29",
         macrolanguage: "bal"}).
iana(language,bgo,tag{
         type: "language",
         subtag: "bgo",
         description: "Baga Koga",
         added: "2009-07-29"}).
iana(language,bgp,tag{
         type: "language",
         subtag: "bgp",
         description: "Eastern Balochi",
         added: "2009-07-29",
         macrolanguage: "bal"}).
iana(language,bgq,tag{
         type: "language",
         subtag: "bgq",
         description: "Bagri",
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,bgr,tag{
         type: "language",
         subtag: "bgr",
         description: "Bawm Chin",
         added: "2009-07-29"}).
iana(language,bgs,tag{
         type: "language",
         subtag: "bgs",
         description: "Tagabawa",
         added: "2009-07-29"}).
iana(language,bgt,tag{
         type: "language",
         subtag: "bgt",
         description: "Bughotu",
         added: "2009-07-29"}).
iana(language,bgu,tag{
         type: "language",
         subtag: "bgu",
         description: "Mbongno",
         added: "2009-07-29"}).
iana(language,bgv,tag{
         type: "language",
         subtag: "bgv",
         description: "Warkay-Bipim",
         added: "2009-07-29"}).
iana(language,bgw,tag{
         type: "language",
         subtag: "bgw",
         description: "Bhatri",
         added: "2009-07-29"}).
iana(language,bgx,tag{
         type: "language",
         subtag: "bgx",
         description: "Balkan Gagauz Turkish",
         added: "2009-07-29"}).
iana(language,bgy,tag{
         type: "language",
         subtag: "bgy",
         description: "Benggoi",
         added: "2009-07-29"}).
iana(language,bgz,tag{
         type: "language",
         subtag: "bgz",
         description: "Banggai",
         added: "2009-07-29"}).
iana(language,bha,tag{
         type: "language",
         subtag: "bha",
         description: "Bharia",
         added: "2009-07-29"}).
iana(language,bhb,tag{
         type: "language",
         subtag: "bhb",
         description: "Bhili",
         added: "2009-07-29"}).
iana(language,bhc,tag{
         type: "language",
         subtag: "bhc",
         description: "Biga",
         added: "2009-07-29"}).
iana(language,bhd,tag{
         type: "language",
         subtag: "bhd",
         description: "Bhadrawahi",
         added: "2009-07-29"}).
iana(language,bhe,tag{
         type: "language",
         subtag: "bhe",
         description: "Bhaya",
         added: "2009-07-29"}).
iana(language,bhf,tag{
         type: "language",
         subtag: "bhf",
         description: "Odiai",
         added: "2009-07-29"}).
iana(language,bhg,tag{
         type: "language",
         subtag: "bhg",
         description: "Binandere",
         added: "2009-07-29"}).
iana(language,bhh,tag{
         type: "language",
         subtag: "bhh",
         description: "Bukharic",
         added: "2009-07-29"}).
iana(language,bhi,tag{
         type: "language",
         subtag: "bhi",
         description: "Bhilali",
         added: "2009-07-29"}).
iana(language,bhj,tag{
         type: "language",
         subtag: "bhj",
         description: "Bahing",
         added: "2009-07-29"}).
iana(language,bhk,tag{
         type: "language",
         subtag: "bhk",
         description: "Albay Bicolano",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         macrolanguage: "bik",
         comments: "see fbl, lbl, rbl, ubl"}).
iana(language,bhl,tag{
         type: "language",
         subtag: "bhl",
         description: "Bimin",
         added: "2009-07-29"}).
iana(language,bhm,tag{
         type: "language",
         subtag: "bhm",
         description: "Bathari",
         added: "2009-07-29"}).
iana(language,bhn,tag{
         type: "language",
         subtag: "bhn",
         description: "Bohtan Neo-Aramaic",
         added: "2009-07-29"}).
iana(language,bho,tag{
         type: "language",
         subtag: "bho",
         description: "Bhojpuri",
         added: "2005-10-16"}).
iana(language,bhp,tag{
         type: "language",
         subtag: "bhp",
         description: "Bima",
         added: "2009-07-29"}).
iana(language,bhq,tag{
         type: "language",
         subtag: "bhq",
         description: "Tukang Besi South",
         added: "2009-07-29"}).
iana(language,bhr,tag{
         type: "language",
         subtag: "bhr",
         description: "Bara Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,bhs,tag{
         type: "language",
         subtag: "bhs",
         description: "Buwal",
         added: "2009-07-29"}).
iana(language,bht,tag{
         type: "language",
         subtag: "bht",
         description: "Bhattiyali",
         added: "2009-07-29"}).
iana(language,bhu,tag{
         type: "language",
         subtag: "bhu",
         description: "Bhunjia",
         added: "2009-07-29"}).
iana(language,bhv,tag{
         type: "language",
         subtag: "bhv",
         description: "Bahau",
         added: "2009-07-29"}).
iana(language,bhw,tag{
         type: "language",
         subtag: "bhw",
         description: "Biak",
         added: "2009-07-29"}).
iana(language,bhx,tag{
         type: "language",
         subtag: "bhx",
         description: "Bhalay",
         added: "2009-07-29"}).
iana(language,bhy,tag{
         type: "language",
         subtag: "bhy",
         description: "Bhele",
         added: "2009-07-29"}).
iana(language,bhz,tag{
         type: "language",
         subtag: "bhz",
         description: "Bada (Indonesia)",
         added: "2009-07-29"}).
iana(language,bia,tag{
         type: "language",
         subtag: "bia",
         description: "Badimaya",
         added: "2009-07-29"}).
iana(language,bib,tag{
         type: "language",
         subtag: "bib",
         description: ["Bissa","Bisa"],
         added: "2009-07-29"}).
iana(language,bic,tag{
         type: "language",
         subtag: "bic",
         description: "Bikaru",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         preferred_value: "bir"}).
iana(language,bid,tag{
         type: "language",
         subtag: "bid",
         description: "Bidiyo",
         added: "2009-07-29"}).
iana(language,bie,tag{
         type: "language",
         subtag: "bie",
         description: "Bepour",
         added: "2009-07-29"}).
iana(language,bif,tag{
         type: "language",
         subtag: "bif",
         description: "Biafada",
         added: "2009-07-29"}).
iana(language,big,tag{
         type: "language",
         subtag: "big",
         description: "Biangai",
         added: "2009-07-29"}).
iana(language,bij,tag{
         type: "language",
         subtag: "bij",
         description: "Vaghat-Ya-Bijim-Legeri",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see dkg, jbm, tyy"}).
iana(language,bik,tag{
         type: "language",
         subtag: "bik",
         description: "Bikol",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,bil,tag{
         type: "language",
         subtag: "bil",
         description: "Bile",
         added: "2009-07-29"}).
iana(language,bim,tag{
         type: "language",
         subtag: "bim",
         description: "Bimoba",
         added: "2009-07-29"}).
iana(language,bin,tag{
         type: "language",
         subtag: "bin",
         description: ["Bini","Edo"],
         added: "2005-10-16"}).
iana(language,bio,tag{
         type: "language",
         subtag: "bio",
         description: "Nai",
         added: "2009-07-29"}).
iana(language,bip,tag{
         type: "language",
         subtag: "bip",
         description: "Bila",
         added: "2009-07-29"}).
iana(language,biq,tag{
         type: "language",
         subtag: "biq",
         description: "Bipi",
         added: "2009-07-29"}).
iana(language,bir,tag{
         type: "language",
         subtag: "bir",
         description: "Bisorio",
         added: "2009-07-29"}).
iana(language,bit,tag{
         type: "language",
         subtag: "bit",
         description: "Berinomo",
         added: "2009-07-29"}).
iana(language,biu,tag{
         type: "language",
         subtag: "biu",
         description: "Biete",
         added: "2009-07-29"}).
iana(language,biv,tag{
         type: "language",
         subtag: "biv",
         description: "Southern Birifor",
         added: "2009-07-29"}).
iana(language,biw,tag{
         type: "language",
         subtag: "biw",
         description: "Kol (Cameroon)",
         added: "2009-07-29"}).
iana(language,bix,tag{
         type: "language",
         subtag: "bix",
         description: "Bijori",
         added: "2009-07-29"}).
iana(language,biy,tag{
         type: "language",
         subtag: "biy",
         description: "Birhor",
         added: "2009-07-29"}).
iana(language,biz,tag{
         type: "language",
         subtag: "biz",
         description: "Baloi",
         added: "2009-07-29"}).
iana(language,bja,tag{
         type: "language",
         subtag: "bja",
         description: "Budza",
         added: "2009-07-29"}).
iana(language,bjb,tag{
         type: "language",
         subtag: "bjb",
         description: "Banggarla",
         added: "2009-07-29"}).
iana(language,bjc,tag{
         type: "language",
         subtag: "bjc",
         description: "Bariji",
         added: "2009-07-29"}).
iana(language,bjd,tag{
         type: "language",
         subtag: "bjd",
         description: "Bandjigali",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "drl"}).
iana(language,bje,tag{
         type: "language",
         subtag: "bje",
         description: "Biao-Jiao Mien",
         added: "2009-07-29"}).
iana(language,bjf,tag{
         type: "language",
         subtag: "bjf",
         description: "Barzani Jewish Neo-Aramaic",
         added: "2009-07-29"}).
iana(language,bjg,tag{
         type: "language",
         subtag: "bjg",
         description: "Bidyogo",
         added: "2009-07-29"}).
iana(language,bjh,tag{
         type: "language",
         subtag: "bjh",
         description: "Bahinemo",
         added: "2009-07-29"}).
iana(language,bji,tag{
         type: "language",
         subtag: "bji",
         description: "Burji",
         added: "2009-07-29"}).
iana(language,bjj,tag{
         type: "language",
         subtag: "bjj",
         description: "Kanauji",
         added: "2009-07-29"}).
iana(language,bjk,tag{
         type: "language",
         subtag: "bjk",
         description: "Barok",
         added: "2009-07-29"}).
iana(language,bjl,tag{
         type: "language",
         subtag: "bjl",
         description: "Bulu (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,bjm,tag{
         type: "language",
         subtag: "bjm",
         description: "Bajelani",
         added: "2009-07-29"}).
iana(language,bjn,tag{
         type: "language",
         subtag: "bjn",
         description: "Banjar",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,bjo,tag{
         type: "language",
         subtag: "bjo",
         description: "Mid-Southern Banda",
         added: "2009-07-29"}).
iana(language,bjp,tag{
         type: "language",
         subtag: "bjp",
         description: "Fanamaket",
         added: "2013-09-10"}).
iana(language,bjq,tag{
         type: "language",
         subtag: "bjq",
         description: "Southern Betsimisaraka Malagasy",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         macrolanguage: "mg",
         comments: "see bzc, tkg"}).
iana(language,bjr,tag{
         type: "language",
         subtag: "bjr",
         description: "Binumarien",
         added: "2009-07-29"}).
iana(language,bjs,tag{
         type: "language",
         subtag: "bjs",
         description: "Bajan",
         added: "2009-07-29"}).
iana(language,bjt,tag{
         type: "language",
         subtag: "bjt",
         description: "Balanta-Ganja",
         added: "2009-07-29"}).
iana(language,bju,tag{
         type: "language",
         subtag: "bju",
         description: "Busuu",
         added: "2009-07-29"}).
iana(language,bjv,tag{
         type: "language",
         subtag: "bjv",
         description: "Bedjond",
         added: "2009-07-29"}).
iana(language,bjw,tag{
         type: "language",
         subtag: "bjw",
         description: "Bakwé",
         added: "2009-07-29"}).
iana(language,bjx,tag{
         type: "language",
         subtag: "bjx",
         description: "Banao Itneg",
         added: "2009-07-29"}).
iana(language,bjy,tag{
         type: "language",
         subtag: "bjy",
         description: "Bayali",
         added: "2009-07-29"}).
iana(language,bjz,tag{
         type: "language",
         subtag: "bjz",
         description: "Baruga",
         added: "2009-07-29"}).
iana(language,bka,tag{
         type: "language",
         subtag: "bka",
         description: "Kyak",
         added: "2009-07-29"}).
iana(language,bkb,tag{
         type: "language",
         subtag: "bkb",
         description: "Finallig",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see ebk, obk"}).
iana(language,bkc,tag{
         type: "language",
         subtag: "bkc",
         description: "Baka (Cameroon)",
         added: "2009-07-29"}).
iana(language,bkd,tag{
         type: "language",
         subtag: "bkd",
         description: ["Binukid","Talaandig"],
         added: "2009-07-29"}).
iana(language,bkf,tag{
         type: "language",
         subtag: "bkf",
         description: "Beeke",
         added: "2009-07-29"}).
iana(language,bkg,tag{
         type: "language",
         subtag: "bkg",
         description: "Buraka",
         added: "2009-07-29"}).
iana(language,bkh,tag{
         type: "language",
         subtag: "bkh",
         description: "Bakoko",
         added: "2009-07-29"}).
iana(language,bki,tag{
         type: "language",
         subtag: "bki",
         description: "Baki",
         added: "2009-07-29"}).
iana(language,bkj,tag{
         type: "language",
         subtag: "bkj",
         description: "Pande",
         added: "2009-07-29"}).
iana(language,bkk,tag{
         type: "language",
         subtag: "bkk",
         description: "Brokskat",
         added: "2009-07-29"}).
iana(language,bkl,tag{
         type: "language",
         subtag: "bkl",
         description: "Berik",
         added: "2009-07-29"}).
iana(language,bkm,tag{
         type: "language",
         subtag: "bkm",
         description: "Kom (Cameroon)",
         added: "2009-07-29"}).
iana(language,bkn,tag{
         type: "language",
         subtag: "bkn",
         description: "Bukitan",
         added: "2009-07-29"}).
iana(language,bko,tag{
         type: "language",
         subtag: "bko",
         description: "Kwa'",
         added: "2009-07-29"}).
iana(language,bkp,tag{
         type: "language",
         subtag: "bkp",
         description: "Boko (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,bkq,tag{
         type: "language",
         subtag: "bkq",
         description: "Bakairí",
         added: "2009-07-29"}).
iana(language,bkr,tag{
         type: "language",
         subtag: "bkr",
         description: "Bakumpai",
         added: "2009-07-29"}).
iana(language,bks,tag{
         type: "language",
         subtag: "bks",
         description: "Northern Sorsoganon",
         added: "2009-07-29"}).
iana(language,bkt,tag{
         type: "language",
         subtag: "bkt",
         description: "Boloki",
         added: "2009-07-29"}).
iana(language,bku,tag{
         type: "language",
         subtag: "bku",
         description: "Buhid",
         added: "2009-07-29"}).
iana(language,bkv,tag{
         type: "language",
         subtag: "bkv",
         description: "Bekwarra",
         added: "2009-07-29"}).
iana(language,bkw,tag{
         type: "language",
         subtag: "bkw",
         description: "Bekwel",
         added: "2009-07-29"}).
iana(language,bkx,tag{
         type: "language",
         subtag: "bkx",
         description: "Baikeno",
         added: "2009-07-29"}).
iana(language,bky,tag{
         type: "language",
         subtag: "bky",
         description: "Bokyi",
         added: "2009-07-29"}).
iana(language,bkz,tag{
         type: "language",
         subtag: "bkz",
         description: "Bungku",
         added: "2009-07-29"}).
iana(language,bla,tag{
         type: "language",
         subtag: "bla",
         description: "Siksika",
         added: "2005-10-16"}).
iana(language,blb,tag{
         type: "language",
         subtag: "blb",
         description: "Bilua",
         added: "2009-07-29"}).
iana(language,blc,tag{
         type: "language",
         subtag: "blc",
         description: "Bella Coola",
         added: "2009-07-29"}).
iana(language,bld,tag{
         type: "language",
         subtag: "bld",
         description: "Bolango",
         added: "2009-07-29"}).
iana(language,ble,tag{
         type: "language",
         subtag: "ble",
         description: "Balanta-Kentohe",
         added: "2009-07-29"}).
iana(language,blf,tag{
         type: "language",
         subtag: "blf",
         description: "Buol",
         added: "2009-07-29"}).
iana(language,blg,tag{
         type: "language",
         subtag: "blg",
         description: "Balau",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         preferred_value: "iba"}).
iana(language,blh,tag{
         type: "language",
         subtag: "blh",
         description: "Kuwaa",
         added: "2009-07-29"}).
iana(language,bli,tag{
         type: "language",
         subtag: "bli",
         description: "Bolia",
         added: "2009-07-29"}).
iana(language,blj,tag{
         type: "language",
         subtag: "blj",
         description: "Bolongan",
         added: "2009-07-29"}).
iana(language,blk,tag{
         type: "language",
         subtag: "blk",
         description: ["Pa'o Karen","Pa'O"],
         added: "2009-07-29"}).
iana(language,bll,tag{
         type: "language",
         subtag: "bll",
         description: "Biloxi",
         added: "2009-07-29"}).
iana(language,blm,tag{
         type: "language",
         subtag: "blm",
         description: "Beli (South Sudan)",
         added: "2009-07-29"}).
iana(language,bln,tag{
         type: "language",
         subtag: "bln",
         description: "Southern Catanduanes Bikol",
         added: "2009-07-29",
         macrolanguage: "bik"}).
iana(language,blo,tag{
         type: "language",
         subtag: "blo",
         description: "Anii",
         added: "2009-07-29"}).
iana(language,blp,tag{
         type: "language",
         subtag: "blp",
         description: "Blablanga",
         added: "2009-07-29"}).
iana(language,blq,tag{
         type: "language",
         subtag: "blq",
         description: "Baluan-Pam",
         added: "2009-07-29"}).
iana(language,blr,tag{
         type: "language",
         subtag: "blr",
         description: "Blang",
         added: "2009-07-29"}).
iana(language,bls,tag{
         type: "language",
         subtag: "bls",
         description: "Balaesang",
         added: "2009-07-29"}).
iana(language,blt,tag{
         type: "language",
         subtag: "blt",
         description: "Tai Dam",
         added: "2009-07-29"}).
iana(language,blv,tag{
         type: "language",
         subtag: "blv",
         description: ["Kibala","Bolo"],
         added: "2009-07-29"}).
iana(language,blw,tag{
         type: "language",
         subtag: "blw",
         description: "Balangao",
         added: "2009-07-29"}).
iana(language,blx,tag{
         type: "language",
         subtag: "blx",
         description: "Mag-Indi Ayta",
         added: "2009-07-29"}).
iana(language,bly,tag{
         type: "language",
         subtag: "bly",
         description: "Notre",
         added: "2009-07-29"}).
iana(language,blz,tag{
         type: "language",
         subtag: "blz",
         description: "Balantak",
         added: "2009-07-29"}).
iana(language,bma,tag{
         type: "language",
         subtag: "bma",
         description: "Lame",
         added: "2009-07-29"}).
iana(language,bmb,tag{
         type: "language",
         subtag: "bmb",
         description: "Bembe",
         added: "2009-07-29"}).
iana(language,bmc,tag{
         type: "language",
         subtag: "bmc",
         description: "Biem",
         added: "2009-07-29"}).
iana(language,bmd,tag{
         type: "language",
         subtag: "bmd",
         description: "Baga Manduri",
         added: "2009-07-29"}).
iana(language,bme,tag{
         type: "language",
         subtag: "bme",
         description: "Limassa",
         added: "2009-07-29"}).
iana(language,bmf,tag{
         type: "language",
         subtag: "bmf",
         description: "Bom-Kim",
         added: "2009-07-29"}).
iana(language,bmg,tag{
         type: "language",
         subtag: "bmg",
         description: "Bamwe",
         added: "2009-07-29"}).
iana(language,bmh,tag{
         type: "language",
         subtag: "bmh",
         description: "Kein",
         added: "2009-07-29"}).
iana(language,bmi,tag{
         type: "language",
         subtag: "bmi",
         description: "Bagirmi",
         added: "2009-07-29"}).
iana(language,bmj,tag{
         type: "language",
         subtag: "bmj",
         description: "Bote-Majhi",
         added: "2009-07-29"}).
iana(language,bmk,tag{
         type: "language",
         subtag: "bmk",
         description: "Ghayavi",
         added: "2009-07-29"}).
iana(language,bml,tag{
         type: "language",
         subtag: "bml",
         description: "Bomboli",
         added: "2009-07-29"}).
iana(language,bmm,tag{
         type: "language",
         subtag: "bmm",
         description: "Northern Betsimisaraka Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,bmn,tag{
         type: "language",
         subtag: "bmn",
         description: "Bina (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,bmo,tag{
         type: "language",
         subtag: "bmo",
         description: "Bambalang",
         added: "2009-07-29"}).
iana(language,bmp,tag{
         type: "language",
         subtag: "bmp",
         description: "Bulgebi",
         added: "2009-07-29"}).
iana(language,bmq,tag{
         type: "language",
         subtag: "bmq",
         description: "Bomu",
         added: "2009-07-29"}).
iana(language,bmr,tag{
         type: "language",
         subtag: "bmr",
         description: "Muinane",
         added: "2009-07-29"}).
iana(language,bms,tag{
         type: "language",
         subtag: "bms",
         description: "Bilma Kanuri",
         added: "2009-07-29"}).
iana(language,bmt,tag{
         type: "language",
         subtag: "bmt",
         description: "Biao Mon",
         added: "2009-07-29"}).
iana(language,bmu,tag{
         type: "language",
         subtag: "bmu",
         description: "Somba-Siawari",
         added: "2009-07-29"}).
iana(language,bmv,tag{
         type: "language",
         subtag: "bmv",
         description: "Bum",
         added: "2009-07-29"}).
iana(language,bmw,tag{
         type: "language",
         subtag: "bmw",
         description: "Bomwali",
         added: "2009-07-29"}).
iana(language,bmx,tag{
         type: "language",
         subtag: "bmx",
         description: "Baimak",
         added: "2009-07-29"}).
iana(language,bmy,tag{
         type: "language",
         subtag: "bmy",
         description: "Bemba (Democratic Republic of Congo)",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,bmz,tag{
         type: "language",
         subtag: "bmz",
         description: "Baramu",
         added: "2009-07-29"}).
iana(language,bna,tag{
         type: "language",
         subtag: "bna",
         description: "Bonerate",
         added: "2009-07-29"}).
iana(language,bnb,tag{
         type: "language",
         subtag: "bnb",
         description: "Bookan",
         added: "2009-07-29"}).
iana(language,bnc,tag{
         type: "language",
         subtag: "bnc",
         description: "Bontok",
         added: "2009-07-29",
         scope: "macrolanguage"}).
iana(language,bnd,tag{
         type: "language",
         subtag: "bnd",
         description: "Banda (Indonesia)",
         added: "2009-07-29"}).
iana(language,bne,tag{
         type: "language",
         subtag: "bne",
         description: "Bintauna",
         added: "2009-07-29"}).
iana(language,bnf,tag{
         type: "language",
         subtag: "bnf",
         description: "Masiwang",
         added: "2009-07-29"}).
iana(language,bng,tag{
         type: "language",
         subtag: "bng",
         description: "Benga",
         added: "2009-07-29"}).
iana(language,bni,tag{
         type: "language",
         subtag: "bni",
         description: "Bangi",
         added: "2009-07-29"}).
iana(language,bnj,tag{
         type: "language",
         subtag: "bnj",
         description: "Eastern Tawbuid",
         added: "2009-07-29"}).
iana(language,bnk,tag{
         type: "language",
         subtag: "bnk",
         description: "Bierebo",
         added: "2009-07-29"}).
iana(language,bnl,tag{
         type: "language",
         subtag: "bnl",
         description: "Boon",
         added: "2009-07-29"}).
iana(language,bnm,tag{
         type: "language",
         subtag: "bnm",
         description: "Batanga",
         added: "2009-07-29"}).
iana(language,bnn,tag{
         type: "language",
         subtag: "bnn",
         description: "Bunun",
         added: "2009-07-29"}).
iana(language,bno,tag{
         type: "language",
         subtag: "bno",
         description: "Bantoanon",
         added: "2009-07-29"}).
iana(language,bnp,tag{
         type: "language",
         subtag: "bnp",
         description: "Bola",
         added: "2009-07-29"}).
iana(language,bnq,tag{
         type: "language",
         subtag: "bnq",
         description: "Bantik",
         added: "2009-07-29"}).
iana(language,bnr,tag{
         type: "language",
         subtag: "bnr",
         description: "Butmas-Tur",
         added: "2009-07-29"}).
iana(language,bns,tag{
         type: "language",
         subtag: "bns",
         description: "Bundeli",
         added: "2009-07-29"}).
iana(language,bnt,tag{
         type: "language",
         subtag: "bnt",
         description: "Bantu languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,bnu,tag{
         type: "language",
         subtag: "bnu",
         description: "Bentong",
         added: "2009-07-29"}).
iana(language,bnv,tag{
         type: "language",
         subtag: "bnv",
         description: ["Bonerif","Beneraf","Edwas"],
         added: "2009-07-29"}).
iana(language,bnw,tag{
         type: "language",
         subtag: "bnw",
         description: "Bisis",
         added: "2009-07-29"}).
iana(language,bnx,tag{
         type: "language",
         subtag: "bnx",
         description: "Bangubangu",
         added: "2009-07-29"}).
iana(language,bny,tag{
         type: "language",
         subtag: "bny",
         description: "Bintulu",
         added: "2009-07-29"}).
iana(language,bnz,tag{
         type: "language",
         subtag: "bnz",
         description: "Beezen",
         added: "2009-07-29"}).
iana(language,boa,tag{
         type: "language",
         subtag: "boa",
         description: "Bora",
         added: "2009-07-29"}).
iana(language,bob,tag{
         type: "language",
         subtag: "bob",
         description: "Aweer",
         added: "2009-07-29"}).
iana(language,boe,tag{
         type: "language",
         subtag: "boe",
         description: "Mundabli",
         added: "2009-07-29"}).
iana(language,bof,tag{
         type: "language",
         subtag: "bof",
         description: "Bolon",
         added: "2009-07-29"}).
iana(language,bog,tag{
         type: "language",
         subtag: "bog",
         description: "Bamako Sign Language",
         added: "2009-07-29"}).
iana(language,boh,tag{
         type: "language",
         subtag: "boh",
         description: "Boma",
         added: "2009-07-29"}).
iana(language,boi,tag{
         type: "language",
         subtag: "boi",
         description: "Barbareño",
         added: "2009-07-29"}).
iana(language,boj,tag{
         type: "language",
         subtag: "boj",
         description: "Anjam",
         added: "2009-07-29"}).
iana(language,bok,tag{
         type: "language",
         subtag: "bok",
         description: "Bonjo",
         added: "2009-07-29"}).
iana(language,bol,tag{
         type: "language",
         subtag: "bol",
         description: "Bole",
         added: "2009-07-29"}).
iana(language,bom,tag{
         type: "language",
         subtag: "bom",
         description: "Berom",
         added: "2009-07-29"}).
iana(language,bon,tag{
         type: "language",
         subtag: "bon",
         description: "Bine",
         added: "2009-07-29"}).
iana(language,boo,tag{
         type: "language",
         subtag: "boo",
         description: "Tiemacèwè Bozo",
         added: "2009-07-29"}).
iana(language,bop,tag{
         type: "language",
         subtag: "bop",
         description: "Bonkiman",
         added: "2009-07-29"}).
iana(language,boq,tag{
         type: "language",
         subtag: "boq",
         description: "Bogaya",
         added: "2009-07-29"}).
iana(language,bor,tag{
         type: "language",
         subtag: "bor",
         description: "Borôro",
         added: "2009-07-29"}).
iana(language,bot,tag{
         type: "language",
         subtag: "bot",
         description: "Bongo",
         added: "2009-07-29"}).
iana(language,bou,tag{
         type: "language",
         subtag: "bou",
         description: "Bondei",
         added: "2009-07-29"}).
iana(language,bov,tag{
         type: "language",
         subtag: "bov",
         description: "Tuwuli",
         added: "2009-07-29"}).
iana(language,bow,tag{
         type: "language",
         subtag: "bow",
         description: "Rema",
         added: "2009-07-29"}).
iana(language,box,tag{
         type: "language",
         subtag: "box",
         description: "Buamu",
         added: "2009-07-29"}).
iana(language,boy,tag{
         type: "language",
         subtag: "boy",
         description: "Bodo (Central African Republic)",
         added: "2009-07-29"}).
iana(language,boz,tag{
         type: "language",
         subtag: "boz",
         description: "Tiéyaxo Bozo",
         added: "2009-07-29"}).
iana(language,bpa,tag{
         type: "language",
         subtag: "bpa",
         description: "Daakaka",
         added: "2009-07-29"}).
iana(language,bpb,tag{
         type: "language",
         subtag: "bpb",
         description: "Barbacoas",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,bpc,tag{
         type: "language",
         subtag: "bpc",
         description: "Mbuk",
         added: "2022-02-25"}).
iana(language,bpd,tag{
         type: "language",
         subtag: "bpd",
         description: "Banda-Banda",
         added: "2009-07-29"}).
iana(language,bpe,tag{
         type: "language",
         subtag: "bpe",
         description: "Bauni",
         added: "2021-02-20"}).
iana(language,bpg,tag{
         type: "language",
         subtag: "bpg",
         description: "Bonggo",
         added: "2009-07-29"}).
iana(language,bph,tag{
         type: "language",
         subtag: "bph",
         description: "Botlikh",
         added: "2009-07-29"}).
iana(language,bpi,tag{
         type: "language",
         subtag: "bpi",
         description: "Bagupi",
         added: "2009-07-29"}).
iana(language,bpj,tag{
         type: "language",
         subtag: "bpj",
         description: "Binji",
         added: "2009-07-29"}).
iana(language,bpk,tag{
         type: "language",
         subtag: "bpk",
         description: ["Orowe","'Ôrôê"],
         added: "2009-07-29"}).
iana(language,bpl,tag{
         type: "language",
         subtag: "bpl",
         description: "Broome Pearling Lugger Pidgin",
         added: "2009-07-29"}).
iana(language,bpm,tag{
         type: "language",
         subtag: "bpm",
         description: "Biyom",
         added: "2009-07-29"}).
iana(language,bpn,tag{
         type: "language",
         subtag: "bpn",
         description: "Dzao Min",
         added: "2009-07-29"}).
iana(language,bpo,tag{
         type: "language",
         subtag: "bpo",
         description: "Anasi",
         added: "2009-07-29"}).
iana(language,bpp,tag{
         type: "language",
         subtag: "bpp",
         description: "Kaure",
         added: "2009-07-29"}).
iana(language,bpq,tag{
         type: "language",
         subtag: "bpq",
         description: "Banda Malay",
         added: "2009-07-29"}).
iana(language,bpr,tag{
         type: "language",
         subtag: "bpr",
         description: "Koronadal Blaan",
         added: "2009-07-29"}).
iana(language,bps,tag{
         type: "language",
         subtag: "bps",
         description: "Sarangani Blaan",
         added: "2009-07-29"}).
iana(language,bpt,tag{
         type: "language",
         subtag: "bpt",
         description: "Barrow Point",
         added: "2009-07-29"}).
iana(language,bpu,tag{
         type: "language",
         subtag: "bpu",
         description: "Bongu",
         added: "2009-07-29"}).
iana(language,bpv,tag{
         type: "language",
         subtag: "bpv",
         description: "Bian Marind",
         added: "2009-07-29"}).
iana(language,bpw,tag{
         type: "language",
         subtag: "bpw",
         description: "Bo (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,bpx,tag{
         type: "language",
         subtag: "bpx",
         description: "Palya Bareli",
         added: "2009-07-29"}).
iana(language,bpy,tag{
         type: "language",
         subtag: "bpy",
         description: "Bishnupriya",
         added: "2009-07-29"}).
iana(language,bpz,tag{
         type: "language",
         subtag: "bpz",
         description: "Bilba",
         added: "2009-07-29"}).
iana(language,bqa,tag{
         type: "language",
         subtag: "bqa",
         description: "Tchumbuli",
         added: "2009-07-29"}).
iana(language,bqb,tag{
         type: "language",
         subtag: "bqb",
         description: "Bagusa",
         added: "2009-07-29"}).
iana(language,bqc,tag{
         type: "language",
         subtag: "bqc",
         description: ["Boko (Benin)","Boo"],
         added: "2009-07-29"}).
iana(language,bqd,tag{
         type: "language",
         subtag: "bqd",
         description: "Bung",
         added: "2009-07-29"}).
iana(language,bqf,tag{
         type: "language",
         subtag: "bqf",
         description: "Baga Kaloum",
         added: "2009-07-29"}).
iana(language,bqg,tag{
         type: "language",
         subtag: "bqg",
         description: "Bago-Kusuntu",
         added: "2009-07-29"}).
iana(language,bqh,tag{
         type: "language",
         subtag: "bqh",
         description: "Baima",
         added: "2009-07-29"}).
iana(language,bqi,tag{
         type: "language",
         subtag: "bqi",
         description: "Bakhtiari",
         added: "2009-07-29"}).
iana(language,bqj,tag{
         type: "language",
         subtag: "bqj",
         description: "Bandial",
         added: "2009-07-29"}).
iana(language,bqk,tag{
         type: "language",
         subtag: "bqk",
         description: "Banda-Mbrès",
         added: "2009-07-29"}).
iana(language,bql,tag{
         type: "language",
         subtag: "bql",
         description: "Bilakura",
         added: "2009-07-29"}).
iana(language,bqm,tag{
         type: "language",
         subtag: "bqm",
         description: "Wumboko",
         added: "2009-07-29"}).
iana(language,bqn,tag{
         type: "language",
         subtag: "bqn",
         description: "Bulgarian Sign Language",
         added: "2009-07-29"}).
iana(language,bqo,tag{
         type: "language",
         subtag: "bqo",
         description: "Balo",
         added: "2009-07-29"}).
iana(language,bqp,tag{
         type: "language",
         subtag: "bqp",
         description: "Busa",
         added: "2009-07-29"}).
iana(language,bqq,tag{
         type: "language",
         subtag: "bqq",
         description: "Biritai",
         added: "2009-07-29"}).
iana(language,bqr,tag{
         type: "language",
         subtag: "bqr",
         description: "Burusu",
         added: "2009-07-29"}).
iana(language,bqs,tag{
         type: "language",
         subtag: "bqs",
         description: "Bosngun",
         added: "2009-07-29"}).
iana(language,bqt,tag{
         type: "language",
         subtag: "bqt",
         description: "Bamukumbit",
         added: "2009-07-29"}).
iana(language,bqu,tag{
         type: "language",
         subtag: "bqu",
         description: "Boguru",
         added: "2009-07-29"}).
iana(language,bqv,tag{
         type: "language",
         subtag: "bqv",
         description: ["Koro Wachi","Begbere-Ejar"],
         added: "2009-07-29"}).
iana(language,bqw,tag{
         type: "language",
         subtag: "bqw",
         description: "Buru (Nigeria)",
         added: "2009-07-29"}).
iana(language,bqx,tag{
         type: "language",
         subtag: "bqx",
         description: "Baangi",
         added: "2009-07-29"}).
iana(language,bqy,tag{
         type: "language",
         subtag: "bqy",
         description: "Bengkala Sign Language",
         added: "2009-07-29"}).
iana(language,bqz,tag{
         type: "language",
         subtag: "bqz",
         description: "Bakaka",
         added: "2009-07-29"}).
iana(language,bra,tag{
         type: "language",
         subtag: "bra",
         description: "Braj",
         added: "2005-10-16"}).
iana(language,brb,tag{
         type: "language",
         subtag: "brb",
         description: ["Brao","Lave"],
         added: "2009-07-29"}).
iana(language,brc,tag{
         type: "language",
         subtag: "brc",
         description: "Berbice Creole Dutch",
         added: "2009-07-29"}).
iana(language,brd,tag{
         type: "language",
         subtag: "brd",
         description: "Baraamu",
         added: "2009-07-29"}).
iana(language,brf,tag{
         type: "language",
         subtag: "brf",
         description: "Bira",
         added: "2009-07-29"}).
iana(language,brg,tag{
         type: "language",
         subtag: "brg",
         description: "Baure",
         added: "2009-07-29"}).
iana(language,brh,tag{
         type: "language",
         subtag: "brh",
         description: "Brahui",
         added: "2009-07-29"}).
iana(language,bri,tag{
         type: "language",
         subtag: "bri",
         description: "Mokpwe",
         added: "2009-07-29"}).
iana(language,brj,tag{
         type: "language",
         subtag: "brj",
         description: "Bieria",
         added: "2009-07-29"}).
iana(language,brk,tag{
         type: "language",
         subtag: "brk",
         description: "Birked",
         added: "2009-07-29"}).
iana(language,brl,tag{
         type: "language",
         subtag: "brl",
         description: "Birwa",
         added: "2009-07-29"}).
iana(language,brm,tag{
         type: "language",
         subtag: "brm",
         description: "Barambu",
         added: "2009-07-29"}).
iana(language,brn,tag{
         type: "language",
         subtag: "brn",
         description: "Boruca",
         added: "2009-07-29"}).
iana(language,bro,tag{
         type: "language",
         subtag: "bro",
         description: "Brokkat",
         added: "2009-07-29"}).
iana(language,brp,tag{
         type: "language",
         subtag: "brp",
         description: "Barapasi",
         added: "2009-07-29"}).
iana(language,brq,tag{
         type: "language",
         subtag: "brq",
         description: "Breri",
         added: "2009-07-29"}).
iana(language,brr,tag{
         type: "language",
         subtag: "brr",
         description: "Birao",
         added: "2009-07-29"}).
iana(language,brs,tag{
         type: "language",
         subtag: "brs",
         description: "Baras",
         added: "2009-07-29"}).
iana(language,brt,tag{
         type: "language",
         subtag: "brt",
         description: "Bitare",
         added: "2009-07-29"}).
iana(language,bru,tag{
         type: "language",
         subtag: "bru",
         description: "Eastern Bru",
         added: "2009-07-29"}).
iana(language,brv,tag{
         type: "language",
         subtag: "brv",
         description: "Western Bru",
         added: "2009-07-29"}).
iana(language,brw,tag{
         type: "language",
         subtag: "brw",
         description: "Bellari",
         added: "2009-07-29"}).
iana(language,brx,tag{
         type: "language",
         subtag: "brx",
         description: "Bodo (India)",
         added: "2009-07-29"}).
iana(language,bry,tag{
         type: "language",
         subtag: "bry",
         description: "Burui",
         added: "2009-07-29"}).
iana(language,brz,tag{
         type: "language",
         subtag: "brz",
         description: "Bilbil",
         added: "2009-07-29"}).
iana(language,bsa,tag{
         type: "language",
         subtag: "bsa",
         description: "Abinomn",
         added: "2009-07-29"}).
iana(language,bsb,tag{
         type: "language",
         subtag: "bsb",
         description: "Brunei Bisaya",
         added: "2009-07-29"}).
iana(language,bsc,tag{
         type: "language",
         subtag: "bsc",
         description: ["Bassari","Oniyan"],
         added: "2009-07-29"}).
iana(language,bse,tag{
         type: "language",
         subtag: "bse",
         description: "Wushi",
         added: "2009-07-29"}).
iana(language,bsf,tag{
         type: "language",
         subtag: "bsf",
         description: "Bauchi",
         added: "2009-07-29"}).
iana(language,bsg,tag{
         type: "language",
         subtag: "bsg",
         description: "Bashkardi",
         added: "2009-07-29"}).
iana(language,bsh,tag{
         type: "language",
         subtag: "bsh",
         description: "Kati",
         added: "2009-07-29"}).
iana(language,bsi,tag{
         type: "language",
         subtag: "bsi",
         description: "Bassossi",
         added: "2009-07-29"}).
iana(language,bsj,tag{
         type: "language",
         subtag: "bsj",
         description: "Bangwinji",
         added: "2009-07-29"}).
iana(language,bsk,tag{
         type: "language",
         subtag: "bsk",
         description: "Burushaski",
         added: "2009-07-29"}).
iana(language,bsl,tag{
         type: "language",
         subtag: "bsl",
         description: "Basa-Gumna",
         added: "2009-07-29"}).
iana(language,bsm,tag{
         type: "language",
         subtag: "bsm",
         description: "Busami",
         added: "2009-07-29"}).
iana(language,bsn,tag{
         type: "language",
         subtag: "bsn",
         description: "Barasana-Eduria",
         added: "2009-07-29"}).
iana(language,bso,tag{
         type: "language",
         subtag: "bso",
         description: "Buso",
         added: "2009-07-29"}).
iana(language,bsp,tag{
         type: "language",
         subtag: "bsp",
         description: "Baga Sitemu",
         added: "2009-07-29"}).
iana(language,bsq,tag{
         type: "language",
         subtag: "bsq",
         description: "Bassa",
         added: "2009-07-29"}).
iana(language,bsr,tag{
         type: "language",
         subtag: "bsr",
         description: "Bassa-Kontagora",
         added: "2009-07-29"}).
iana(language,bss,tag{
         type: "language",
         subtag: "bss",
         description: "Akoose",
         added: "2009-07-29"}).
iana(language,bst,tag{
         type: "language",
         subtag: "bst",
         description: "Basketo",
         added: "2009-07-29"}).
iana(language,bsu,tag{
         type: "language",
         subtag: "bsu",
         description: "Bahonsuai",
         added: "2009-07-29"}).
iana(language,bsv,tag{
         type: "language",
         subtag: "bsv",
         description: "Baga Sobané",
         added: "2009-07-29"}).
iana(language,bsw,tag{
         type: "language",
         subtag: "bsw",
         description: "Baiso",
         added: "2009-07-29"}).
iana(language,bsx,tag{
         type: "language",
         subtag: "bsx",
         description: "Yangkam",
         added: "2009-07-29"}).
iana(language,bsy,tag{
         type: "language",
         subtag: "bsy",
         description: "Sabah Bisaya",
         added: "2009-07-29"}).
iana(language,bta,tag{
         type: "language",
         subtag: "bta",
         description: "Bata",
         added: "2009-07-29"}).
iana(language,btb,tag{
         type: "language",
         subtag: "btb",
         description: "Beti (Cameroon)",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see beb, bum, bxp, eto, ewo, fan, mct"}).
iana(language,btc,tag{
         type: "language",
         subtag: "btc",
         description: "Bati (Cameroon)",
         added: "2009-07-29"}).
iana(language,btd,tag{
         type: "language",
         subtag: "btd",
         description: "Batak Dairi",
         added: "2009-07-29"}).
iana(language,bte,tag{
         type: "language",
         subtag: "bte",
         description: "Gamo-Ningi",
         added: "2009-07-29"}).
iana(language,btf,tag{
         type: "language",
         subtag: "btf",
         description: "Birgit",
         added: "2009-07-29"}).
iana(language,btg,tag{
         type: "language",
         subtag: "btg",
         description: "Gagnoa Bété",
         added: "2009-07-29"}).
iana(language,bth,tag{
         type: "language",
         subtag: "bth",
         description: "Biatah Bidayuh",
         added: "2009-07-29"}).
iana(language,bti,tag{
         type: "language",
         subtag: "bti",
         description: "Burate",
         added: "2009-07-29"}).
iana(language,btj,tag{
         type: "language",
         subtag: "btj",
         description: "Bacanese Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,btk,tag{
         type: "language",
         subtag: "btk",
         description: "Batak languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,btl,tag{
         type: "language",
         subtag: "btl",
         description: "Bhatola",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,btm,tag{
         type: "language",
         subtag: "btm",
         description: "Batak Mandailing",
         added: "2009-07-29"}).
iana(language,btn,tag{
         type: "language",
         subtag: "btn",
         description: "Ratagnon",
         added: "2009-07-29"}).
iana(language,bto,tag{
         type: "language",
         subtag: "bto",
         description: "Rinconada Bikol",
         added: "2009-07-29",
         macrolanguage: "bik"}).
iana(language,btp,tag{
         type: "language",
         subtag: "btp",
         description: "Budibud",
         added: "2009-07-29"}).
iana(language,btq,tag{
         type: "language",
         subtag: "btq",
         description: "Batek",
         added: "2009-07-29"}).
iana(language,btr,tag{
         type: "language",
         subtag: "btr",
         description: "Baetora",
         added: "2009-07-29"}).
iana(language,bts,tag{
         type: "language",
         subtag: "bts",
         description: "Batak Simalungun",
         added: "2009-07-29"}).
iana(language,btt,tag{
         type: "language",
         subtag: "btt",
         description: "Bete-Bendi",
         added: "2009-07-29"}).
iana(language,btu,tag{
         type: "language",
         subtag: "btu",
         description: "Batu",
         added: "2009-07-29"}).
iana(language,btv,tag{
         type: "language",
         subtag: "btv",
         description: "Bateri",
         added: "2009-07-29"}).
iana(language,btw,tag{
         type: "language",
         subtag: "btw",
         description: "Butuanon",
         added: "2009-07-29"}).
iana(language,btx,tag{
         type: "language",
         subtag: "btx",
         description: "Batak Karo",
         added: "2009-07-29"}).
iana(language,bty,tag{
         type: "language",
         subtag: "bty",
         description: "Bobot",
         added: "2009-07-29"}).
iana(language,btz,tag{
         type: "language",
         subtag: "btz",
         description: "Batak Alas-Kluet",
         added: "2009-07-29"}).
iana(language,bua,tag{
         type: "language",
         subtag: "bua",
         description: "Buriat",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,bub,tag{
         type: "language",
         subtag: "bub",
         description: "Bua",
         added: "2009-07-29"}).
iana(language,buc,tag{
         type: "language",
         subtag: "buc",
         description: "Bushi",
         added: "2009-07-29"}).
iana(language,bud,tag{
         type: "language",
         subtag: "bud",
         description: "Ntcham",
         added: "2009-07-29"}).
iana(language,bue,tag{
         type: "language",
         subtag: "bue",
         description: "Beothuk",
         added: "2009-07-29"}).
iana(language,buf,tag{
         type: "language",
         subtag: "buf",
         description: "Bushoong",
         added: "2009-07-29"}).
iana(language,bug,tag{
         type: "language",
         subtag: "bug",
         description: "Buginese",
         added: "2005-10-16"}).
iana(language,buh,tag{
         type: "language",
         subtag: "buh",
         description: "Younuo Bunu",
         added: "2009-07-29"}).
iana(language,bui,tag{
         type: "language",
         subtag: "bui",
         description: "Bongili",
         added: "2009-07-29"}).
iana(language,buj,tag{
         type: "language",
         subtag: "buj",
         description: "Basa-Gurmana",
         added: "2009-07-29"}).
iana(language,buk,tag{
         type: "language",
         subtag: "buk",
         description: "Bugawac",
         added: "2009-07-29"}).
iana(language,bum,tag{
         type: "language",
         subtag: "bum",
         description: "Bulu (Cameroon)",
         added: "2009-07-29"}).
iana(language,bun,tag{
         type: "language",
         subtag: "bun",
         description: "Sherbro",
         added: "2009-07-29"}).
iana(language,buo,tag{
         type: "language",
         subtag: "buo",
         description: "Terei",
         added: "2009-07-29"}).
iana(language,bup,tag{
         type: "language",
         subtag: "bup",
         description: "Busoa",
         added: "2009-07-29"}).
iana(language,buq,tag{
         type: "language",
         subtag: "buq",
         description: "Brem",
         added: "2009-07-29"}).
iana(language,bus,tag{
         type: "language",
         subtag: "bus",
         description: "Bokobaru",
         added: "2009-07-29"}).
iana(language,but,tag{
         type: "language",
         subtag: "but",
         description: "Bungain",
         added: "2009-07-29"}).
iana(language,buu,tag{
         type: "language",
         subtag: "buu",
         description: "Budu",
         added: "2009-07-29"}).
iana(language,buv,tag{
         type: "language",
         subtag: "buv",
         description: "Bun",
         added: "2009-07-29"}).
iana(language,buw,tag{
         type: "language",
         subtag: "buw",
         description: "Bubi",
         added: "2009-07-29"}).
iana(language,bux,tag{
         type: "language",
         subtag: "bux",
         description: "Boghom",
         added: "2009-07-29"}).
iana(language,buy,tag{
         type: "language",
         subtag: "buy",
         description: "Bullom So",
         added: "2009-07-29"}).
iana(language,buz,tag{
         type: "language",
         subtag: "buz",
         description: "Bukwen",
         added: "2009-07-29"}).
iana(language,bva,tag{
         type: "language",
         subtag: "bva",
         description: "Barein",
         added: "2009-07-29"}).
iana(language,bvb,tag{
         type: "language",
         subtag: "bvb",
         description: "Bube",
         added: "2009-07-29"}).
iana(language,bvc,tag{
         type: "language",
         subtag: "bvc",
         description: "Baelelea",
         added: "2009-07-29"}).
iana(language,bvd,tag{
         type: "language",
         subtag: "bvd",
         description: "Baeggu",
         added: "2009-07-29"}).
iana(language,bve,tag{
         type: "language",
         subtag: "bve",
         description: "Berau Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,bvf,tag{
         type: "language",
         subtag: "bvf",
         description: "Boor",
         added: "2009-07-29"}).
iana(language,bvg,tag{
         type: "language",
         subtag: "bvg",
         description: "Bonkeng",
         added: "2009-07-29"}).
iana(language,bvh,tag{
         type: "language",
         subtag: "bvh",
         description: "Bure",
         added: "2009-07-29"}).
iana(language,bvi,tag{
         type: "language",
         subtag: "bvi",
         description: "Belanda Viri",
         added: "2009-07-29"}).
iana(language,bvj,tag{
         type: "language",
         subtag: "bvj",
         description: "Baan",
         added: "2009-07-29"}).
iana(language,bvk,tag{
         type: "language",
         subtag: "bvk",
         description: "Bukat",
         added: "2009-07-29"}).
iana(language,bvl,tag{
         type: "language",
         subtag: "bvl",
         description: "Bolivian Sign Language",
         added: "2009-07-29"}).
iana(language,bvm,tag{
         type: "language",
         subtag: "bvm",
         description: "Bamunka",
         added: "2009-07-29"}).
iana(language,bvn,tag{
         type: "language",
         subtag: "bvn",
         description: "Buna",
         added: "2009-07-29"}).
iana(language,bvo,tag{
         type: "language",
         subtag: "bvo",
         description: "Bolgo",
         added: "2009-07-29"}).
iana(language,bvp,tag{
         type: "language",
         subtag: "bvp",
         description: "Bumang",
         added: "2013-09-10"}).
iana(language,bvq,tag{
         type: "language",
         subtag: "bvq",
         description: "Birri",
         added: "2009-07-29"}).
iana(language,bvr,tag{
         type: "language",
         subtag: "bvr",
         description: "Burarra",
         added: "2009-07-29"}).
iana(language,bvt,tag{
         type: "language",
         subtag: "bvt",
         description: "Bati (Indonesia)",
         added: "2009-07-29"}).
iana(language,bvu,tag{
         type: "language",
         subtag: "bvu",
         description: "Bukit Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,bvv,tag{
         type: "language",
         subtag: "bvv",
         description: "Baniva",
         added: "2009-07-29"}).
iana(language,bvw,tag{
         type: "language",
         subtag: "bvw",
         description: "Boga",
         added: "2009-07-29"}).
iana(language,bvx,tag{
         type: "language",
         subtag: "bvx",
         description: "Dibole",
         added: "2009-07-29"}).
iana(language,bvy,tag{
         type: "language",
         subtag: "bvy",
         description: "Baybayanon",
         added: "2010-03-11"}).
iana(language,bvz,tag{
         type: "language",
         subtag: "bvz",
         description: "Bauzi",
         added: "2009-07-29"}).
iana(language,bwa,tag{
         type: "language",
         subtag: "bwa",
         description: "Bwatoo",
         added: "2009-07-29"}).
iana(language,bwb,tag{
         type: "language",
         subtag: "bwb",
         description: "Namosi-Naitasiri-Serua",
         added: "2009-07-29"}).
iana(language,bwc,tag{
         type: "language",
         subtag: "bwc",
         description: "Bwile",
         added: "2009-07-29"}).
iana(language,bwd,tag{
         type: "language",
         subtag: "bwd",
         description: "Bwaidoka",
         added: "2009-07-29"}).
iana(language,bwe,tag{
         type: "language",
         subtag: "bwe",
         description: "Bwe Karen",
         added: "2009-07-29"}).
iana(language,bwf,tag{
         type: "language",
         subtag: "bwf",
         description: "Boselewa",
         added: "2009-07-29"}).
iana(language,bwg,tag{
         type: "language",
         subtag: "bwg",
         description: "Barwe",
         added: "2009-07-29"}).
iana(language,bwh,tag{
         type: "language",
         subtag: "bwh",
         description: "Bishuo",
         added: "2009-07-29"}).
iana(language,bwi,tag{
         type: "language",
         subtag: "bwi",
         description: "Baniwa",
         added: "2009-07-29"}).
iana(language,bwj,tag{
         type: "language",
         subtag: "bwj",
         description: "Láá Láá Bwamu",
         added: "2009-07-29"}).
iana(language,bwk,tag{
         type: "language",
         subtag: "bwk",
         description: "Bauwaki",
         added: "2009-07-29"}).
iana(language,bwl,tag{
         type: "language",
         subtag: "bwl",
         description: "Bwela",
         added: "2009-07-29"}).
iana(language,bwm,tag{
         type: "language",
         subtag: "bwm",
         description: "Biwat",
         added: "2009-07-29"}).
iana(language,bwn,tag{
         type: "language",
         subtag: "bwn",
         description: "Wunai Bunu",
         added: "2009-07-29"}).
iana(language,bwo,tag{
         type: "language",
         subtag: "bwo",
         description: ["Boro (Ethiopia)","Borna (Ethiopia)"],
         added: "2009-07-29"}).
iana(language,bwp,tag{
         type: "language",
         subtag: "bwp",
         description: "Mandobo Bawah",
         added: "2009-07-29"}).
iana(language,bwq,tag{
         type: "language",
         subtag: "bwq",
         description: "Southern Bobo Madaré",
         added: "2009-07-29"}).
iana(language,bwr,tag{
         type: "language",
         subtag: "bwr",
         description: "Bura-Pabir",
         added: "2009-07-29"}).
iana(language,bws,tag{
         type: "language",
         subtag: "bws",
         description: "Bomboma",
         added: "2009-07-29"}).
iana(language,bwt,tag{
         type: "language",
         subtag: "bwt",
         description: "Bafaw-Balong",
         added: "2009-07-29"}).
iana(language,bwu,tag{
         type: "language",
         subtag: "bwu",
         description: "Buli (Ghana)",
         added: "2009-07-29"}).
iana(language,bww,tag{
         type: "language",
         subtag: "bww",
         description: "Bwa",
         added: "2009-07-29"}).
iana(language,bwx,tag{
         type: "language",
         subtag: "bwx",
         description: "Bu-Nao Bunu",
         added: "2009-07-29"}).
iana(language,bwy,tag{
         type: "language",
         subtag: "bwy",
         description: "Cwi Bwamu",
         added: "2009-07-29"}).
iana(language,bwz,tag{
         type: "language",
         subtag: "bwz",
         description: "Bwisi",
         added: "2009-07-29"}).
iana(language,bxa,tag{
         type: "language",
         subtag: "bxa",
         description: "Tairaha",
         added: "2009-07-29"}).
iana(language,bxb,tag{
         type: "language",
         subtag: "bxb",
         description: "Belanda Bor",
         added: "2009-07-29"}).
iana(language,bxc,tag{
         type: "language",
         subtag: "bxc",
         description: "Molengue",
         added: "2009-07-29"}).
iana(language,bxd,tag{
         type: "language",
         subtag: "bxd",
         description: "Pela",
         added: "2009-07-29"}).
iana(language,bxe,tag{
         type: "language",
         subtag: "bxe",
         description: "Birale",
         added: "2009-07-29"}).
iana(language,bxf,tag{
         type: "language",
         subtag: "bxf",
         description: ["Bilur","Minigir"],
         added: "2009-07-29",
         comments: "see also vmg"}).
iana(language,bxg,tag{
         type: "language",
         subtag: "bxg",
         description: "Bangala",
         added: "2009-07-29"}).
iana(language,bxh,tag{
         type: "language",
         subtag: "bxh",
         description: "Buhutu",
         added: "2009-07-29"}).
iana(language,bxi,tag{
         type: "language",
         subtag: "bxi",
         description: "Pirlatapa",
         added: "2009-07-29"}).
iana(language,bxj,tag{
         type: "language",
         subtag: "bxj",
         description: "Bayungu",
         added: "2009-07-29"}).
iana(language,bxk,tag{
         type: "language",
         subtag: "bxk",
         description: ["Bukusu","Lubukusu"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,bxl,tag{
         type: "language",
         subtag: "bxl",
         description: "Jalkunan",
         added: "2009-07-29"}).
iana(language,bxm,tag{
         type: "language",
         subtag: "bxm",
         description: "Mongolia Buriat",
         added: "2009-07-29",
         macrolanguage: "bua"}).
iana(language,bxn,tag{
         type: "language",
         subtag: "bxn",
         description: "Burduna",
         added: "2009-07-29"}).
iana(language,bxo,tag{
         type: "language",
         subtag: "bxo",
         description: "Barikanchi",
         added: "2009-07-29"}).
iana(language,bxp,tag{
         type: "language",
         subtag: "bxp",
         description: "Bebil",
         added: "2009-07-29"}).
iana(language,bxq,tag{
         type: "language",
         subtag: "bxq",
         description: "Beele",
         added: "2009-07-29"}).
iana(language,bxr,tag{
         type: "language",
         subtag: "bxr",
         description: "Russia Buriat",
         added: "2009-07-29",
         macrolanguage: "bua"}).
iana(language,bxs,tag{
         type: "language",
         subtag: "bxs",
         description: "Busam",
         added: "2009-07-29"}).
iana(language,bxu,tag{
         type: "language",
         subtag: "bxu",
         description: "China Buriat",
         added: "2009-07-29",
         macrolanguage: "bua"}).
iana(language,bxv,tag{
         type: "language",
         subtag: "bxv",
         description: "Berakou",
         added: "2009-07-29"}).
iana(language,bxw,tag{
         type: "language",
         subtag: "bxw",
         description: "Bankagooma",
         added: "2009-07-29"}).
iana(language,bxx,tag{
         type: "language",
         subtag: "bxx",
         description: "Borna (Democratic Republic of Congo)",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,bxz,tag{
         type: "language",
         subtag: "bxz",
         description: "Binahari",
         added: "2009-07-29"}).
iana(language,bya,tag{
         type: "language",
         subtag: "bya",
         description: "Batak",
         added: "2009-07-29"}).
iana(language,byb,tag{
         type: "language",
         subtag: "byb",
         description: "Bikya",
         added: "2009-07-29"}).
iana(language,byc,tag{
         type: "language",
         subtag: "byc",
         description: "Ubaghara",
         added: "2009-07-29"}).
iana(language,byd,tag{
         type: "language",
         subtag: "byd",
         description: "Benyadu'",
         added: "2009-07-29"}).
iana(language,bye,tag{
         type: "language",
         subtag: "bye",
         description: "Pouye",
         added: "2009-07-29"}).
iana(language,byf,tag{
         type: "language",
         subtag: "byf",
         description: "Bete",
         added: "2009-07-29"}).
iana(language,byg,tag{
         type: "language",
         subtag: "byg",
         description: "Baygo",
         added: "2009-07-29"}).
iana(language,byh,tag{
         type: "language",
         subtag: "byh",
         description: "Bhujel",
         added: "2009-07-29"}).
iana(language,byi,tag{
         type: "language",
         subtag: "byi",
         description: "Buyu",
         added: "2009-07-29"}).
iana(language,byj,tag{
         type: "language",
         subtag: "byj",
         description: "Bina (Nigeria)",
         added: "2009-07-29"}).
iana(language,byk,tag{
         type: "language",
         subtag: "byk",
         description: "Biao",
         added: "2009-07-29"}).
iana(language,byl,tag{
         type: "language",
         subtag: "byl",
         description: "Bayono",
         added: "2009-07-29"}).
iana(language,bym,tag{
         type: "language",
         subtag: "bym",
         description: "Bidjara",
         added: "2009-07-29"}).
iana(language,byn,tag{
         type: "language",
         subtag: "byn",
         description: ["Bilin","Blin"],
         added: "2005-10-16"}).
iana(language,byo,tag{
         type: "language",
         subtag: "byo",
         description: "Biyo",
         added: "2009-07-29"}).
iana(language,byp,tag{
         type: "language",
         subtag: "byp",
         description: "Bumaji",
         added: "2009-07-29"}).
iana(language,byq,tag{
         type: "language",
         subtag: "byq",
         description: "Basay",
         added: "2009-07-29"}).
iana(language,byr,tag{
         type: "language",
         subtag: "byr",
         description: ["Baruya","Yipma"],
         added: "2009-07-29"}).
iana(language,bys,tag{
         type: "language",
         subtag: "bys",
         description: "Burak",
         added: "2009-07-29"}).
iana(language,byt,tag{
         type: "language",
         subtag: "byt",
         description: "Berti",
         added: "2009-07-29"}).
iana(language,byv,tag{
         type: "language",
         subtag: "byv",
         description: "Medumba",
         added: "2009-07-29"}).
iana(language,byw,tag{
         type: "language",
         subtag: "byw",
         description: "Belhariya",
         added: "2009-07-29"}).
iana(language,byx,tag{
         type: "language",
         subtag: "byx",
         description: "Qaqet",
         added: "2009-07-29"}).
iana(language,byy,tag{
         type: "language",
         subtag: "byy",
         description: "Buya",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,byz,tag{
         type: "language",
         subtag: "byz",
         description: "Banaro",
         added: "2009-07-29"}).
iana(language,bza,tag{
         type: "language",
         subtag: "bza",
         description: "Bandi",
         added: "2009-07-29"}).
iana(language,bzb,tag{
         type: "language",
         subtag: "bzb",
         description: "Andio",
         added: "2009-07-29"}).
iana(language,bzc,tag{
         type: "language",
         subtag: "bzc",
         description: "Southern Betsimisaraka Malagasy",
         added: "2011-08-16",
         macrolanguage: "mg"}).
iana(language,bzd,tag{
         type: "language",
         subtag: "bzd",
         description: "Bribri",
         added: "2009-07-29"}).
iana(language,bze,tag{
         type: "language",
         subtag: "bze",
         description: "Jenaama Bozo",
         added: "2009-07-29"}).
iana(language,bzf,tag{
         type: "language",
         subtag: "bzf",
         description: "Boikin",
         added: "2009-07-29"}).
iana(language,bzg,tag{
         type: "language",
         subtag: "bzg",
         description: "Babuza",
         added: "2009-07-29"}).
iana(language,bzh,tag{
         type: "language",
         subtag: "bzh",
         description: "Mapos Buang",
         added: "2009-07-29"}).
iana(language,bzi,tag{
         type: "language",
         subtag: "bzi",
         description: "Bisu",
         added: "2009-07-29"}).
iana(language,bzj,tag{
         type: "language",
         subtag: "bzj",
         description: "Belize Kriol English",
         added: "2009-07-29"}).
iana(language,bzk,tag{
         type: "language",
         subtag: "bzk",
         description: "Nicaragua Creole English",
         added: "2009-07-29"}).
iana(language,bzl,tag{
         type: "language",
         subtag: "bzl",
         description: "Boano (Sulawesi)",
         added: "2009-07-29"}).
iana(language,bzm,tag{
         type: "language",
         subtag: "bzm",
         description: "Bolondo",
         added: "2009-07-29"}).
iana(language,bzn,tag{
         type: "language",
         subtag: "bzn",
         description: "Boano (Maluku)",
         added: "2009-07-29"}).
iana(language,bzo,tag{
         type: "language",
         subtag: "bzo",
         description: "Bozaba",
         added: "2009-07-29"}).
iana(language,bzp,tag{
         type: "language",
         subtag: "bzp",
         description: "Kemberano",
         added: "2009-07-29"}).
iana(language,bzq,tag{
         type: "language",
         subtag: "bzq",
         description: "Buli (Indonesia)",
         added: "2009-07-29"}).
iana(language,bzr,tag{
         type: "language",
         subtag: "bzr",
         description: "Biri",
         added: "2009-07-29"}).
iana(language,bzs,tag{
         type: "language",
         subtag: "bzs",
         description: "Brazilian Sign Language",
         added: "2009-07-29"}).
iana(language,bzt,tag{
         type: "language",
         subtag: "bzt",
         description: "Brithenig",
         added: "2009-07-29"}).
iana(language,bzu,tag{
         type: "language",
         subtag: "bzu",
         description: "Burmeso",
         added: "2009-07-29"}).
iana(language,bzv,tag{
         type: "language",
         subtag: "bzv",
         description: "Naami",
         added: "2009-07-29"}).
iana(language,bzw,tag{
         type: "language",
         subtag: "bzw",
         description: "Basa (Nigeria)",
         added: "2009-07-29"}).
iana(language,bzx,tag{
         type: "language",
         subtag: "bzx",
         description: "Kɛlɛngaxo Bozo",
         added: "2009-07-29"}).
iana(language,bzy,tag{
         type: "language",
         subtag: "bzy",
         description: "Obanliku",
         added: "2009-07-29"}).
iana(language,bzz,tag{
         type: "language",
         subtag: "bzz",
         description: "Evant",
         added: "2009-07-29"}).
iana(language,caa,tag{
         type: "language",
         subtag: "caa",
         description: "Chortí",
         added: "2009-07-29"}).
iana(language,cab,tag{
         type: "language",
         subtag: "cab",
         description: "Garifuna",
         added: "2009-07-29"}).
iana(language,cac,tag{
         type: "language",
         subtag: "cac",
         description: "Chuj",
         added: "2009-07-29"}).
iana(language,cad,tag{
         type: "language",
         subtag: "cad",
         description: "Caddo",
         added: "2005-10-16"}).
iana(language,cae,tag{
         type: "language",
         subtag: "cae",
         description: ["Lehar","Laalaa"],
         added: "2009-07-29"}).
iana(language,caf,tag{
         type: "language",
         subtag: "caf",
         description: "Southern Carrier",
         added: "2009-07-29"}).
iana(language,cag,tag{
         type: "language",
         subtag: "cag",
         description: "Nivaclé",
         added: "2009-07-29"}).
iana(language,cah,tag{
         type: "language",
         subtag: "cah",
         description: "Cahuarano",
         added: "2009-07-29"}).
iana(language,cai,tag{
         type: "language",
         subtag: "cai",
         description: "Central American Indian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,caj,tag{
         type: "language",
         subtag: "caj",
         description: "Chané",
         added: "2009-07-29"}).
iana(language,cak,tag{
         type: "language",
         subtag: "cak",
         description: ["Kaqchikel","Cakchiquel"],
         added: "2009-07-29"}).
iana(language,cal,tag{
         type: "language",
         subtag: "cal",
         description: "Carolinian",
         added: "2009-07-29"}).
iana(language,cam,tag{
         type: "language",
         subtag: "cam",
         description: "Cemuhî",
         added: "2009-07-29"}).
iana(language,can,tag{
         type: "language",
         subtag: "can",
         description: "Chambri",
         added: "2009-07-29"}).
iana(language,cao,tag{
         type: "language",
         subtag: "cao",
         description: "Chácobo",
         added: "2009-07-29"}).
iana(language,cap,tag{
         type: "language",
         subtag: "cap",
         description: "Chipaya",
         added: "2009-07-29"}).
iana(language,caq,tag{
         type: "language",
         subtag: "caq",
         description: "Car Nicobarese",
         added: "2009-07-29"}).
iana(language,car,tag{
         type: "language",
         subtag: "car",
         description: "Galibi Carib",
         added: "2005-10-16"}).
iana(language,cas,tag{
         type: "language",
         subtag: "cas",
         description: "Tsimané",
         added: "2009-07-29"}).
iana(language,cau,tag{
         type: "language",
         subtag: "cau",
         description: "Caucasian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cav,tag{
         type: "language",
         subtag: "cav",
         description: "Cavineña",
         added: "2009-07-29"}).
iana(language,caw,tag{
         type: "language",
         subtag: "caw",
         description: "Callawalla",
         added: "2009-07-29"}).
iana(language,cax,tag{
         type: "language",
         subtag: "cax",
         description: "Chiquitano",
         added: "2009-07-29"}).
iana(language,cay,tag{
         type: "language",
         subtag: "cay",
         description: "Cayuga",
         added: "2009-07-29"}).
iana(language,caz,tag{
         type: "language",
         subtag: "caz",
         description: "Canichana",
         added: "2009-07-29"}).
iana(language,cba,tag{
         type: "language",
         subtag: "cba",
         description: "Chibchan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,cbb,tag{
         type: "language",
         subtag: "cbb",
         description: "Cabiyarí",
         added: "2009-07-29"}).
iana(language,cbc,tag{
         type: "language",
         subtag: "cbc",
         description: "Carapana",
         added: "2009-07-29"}).
iana(language,cbd,tag{
         type: "language",
         subtag: "cbd",
         description: "Carijona",
         added: "2009-07-29"}).
iana(language,cbe,tag{
         type: "language",
         subtag: "cbe",
         description: "Chipiajes",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,cbg,tag{
         type: "language",
         subtag: "cbg",
         description: "Chimila",
         added: "2009-07-29"}).
iana(language,cbh,tag{
         type: "language",
         subtag: "cbh",
         description: "Cagua",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,cbi,tag{
         type: "language",
         subtag: "cbi",
         description: "Chachi",
         added: "2009-07-29"}).
iana(language,cbj,tag{
         type: "language",
         subtag: "cbj",
         description: "Ede Cabe",
         added: "2009-07-29"}).
iana(language,cbk,tag{
         type: "language",
         subtag: "cbk",
         description: "Chavacano",
         added: "2009-07-29"}).
iana(language,cbl,tag{
         type: "language",
         subtag: "cbl",
         description: "Bualkhaw Chin",
         added: "2009-07-29"}).
iana(language,cbn,tag{
         type: "language",
         subtag: "cbn",
         description: "Nyahkur",
         added: "2009-07-29"}).
iana(language,cbo,tag{
         type: "language",
         subtag: "cbo",
         description: "Izora",
         added: "2009-07-29"}).
iana(language,cbq,tag{
         type: "language",
         subtag: "cbq",
         description: ["Tsucuba","Cuba"],
         added: "2015-02-12"}).
iana(language,cbr,tag{
         type: "language",
         subtag: "cbr",
         description: "Cashibo-Cacataibo",
         added: "2009-07-29"}).
iana(language,cbs,tag{
         type: "language",
         subtag: "cbs",
         description: "Cashinahua",
         added: "2009-07-29"}).
iana(language,cbt,tag{
         type: "language",
         subtag: "cbt",
         description: "Chayahuita",
         added: "2009-07-29"}).
iana(language,cbu,tag{
         type: "language",
         subtag: "cbu",
         description: "Candoshi-Shapra",
         added: "2009-07-29"}).
iana(language,cbv,tag{
         type: "language",
         subtag: "cbv",
         description: "Cacua",
         added: "2009-07-29"}).
iana(language,cbw,tag{
         type: "language",
         subtag: "cbw",
         description: "Kinabalian",
         added: "2010-03-11"}).
iana(language,cby,tag{
         type: "language",
         subtag: "cby",
         description: "Carabayo",
         added: "2009-07-29"}).
iana(language,cca,tag{
         type: "language",
         subtag: "cca",
         description: "Cauca",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,ccc,tag{
         type: "language",
         subtag: "ccc",
         description: "Chamicuro",
         added: "2009-07-29"}).
iana(language,ccd,tag{
         type: "language",
         subtag: "ccd",
         description: "Cafundo Creole",
         added: "2009-07-29"}).
iana(language,cce,tag{
         type: "language",
         subtag: "cce",
         description: "Chopi",
         added: "2009-07-29"}).
iana(language,ccg,tag{
         type: "language",
         subtag: "ccg",
         description: "Samba Daka",
         added: "2009-07-29"}).
iana(language,cch,tag{
         type: "language",
         subtag: "cch",
         description: "Atsam",
         added: "2009-07-29"}).
iana(language,ccj,tag{
         type: "language",
         subtag: "ccj",
         description: "Kasanga",
         added: "2009-07-29"}).
iana(language,ccl,tag{
         type: "language",
         subtag: "ccl",
         description: "Cutchi-Swahili",
         added: "2009-07-29"}).
iana(language,ccm,tag{
         type: "language",
         subtag: "ccm",
         description: "Malaccan Creole Malay",
         added: "2009-07-29"}).
iana(language,ccn,tag{
         type: "language",
         subtag: "ccn",
         description: "North Caucasian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,cco,tag{
         type: "language",
         subtag: "cco",
         description: "Comaltepec Chinantec",
         added: "2009-07-29"}).
iana(language,ccp,tag{
         type: "language",
         subtag: "ccp",
         description: "Chakma",
         added: "2009-07-29"}).
iana(language,ccq,tag{
         type: "language",
         subtag: "ccq",
         description: "Chaungtha",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "rki"}).
iana(language,ccr,tag{
         type: "language",
         subtag: "ccr",
         description: "Cacaopera",
         added: "2009-07-29"}).
iana(language,ccs,tag{
         type: "language",
         subtag: "ccs",
         description: "South Caucasian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,cda,tag{
         type: "language",
         subtag: "cda",
         description: "Choni",
         added: "2009-07-29"}).
iana(language,cdc,tag{
         type: "language",
         subtag: "cdc",
         description: "Chadic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,cdd,tag{
         type: "language",
         subtag: "cdd",
         description: "Caddoan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,cde,tag{
         type: "language",
         subtag: "cde",
         description: "Chenchu",
         added: "2009-07-29"}).
iana(language,cdf,tag{
         type: "language",
         subtag: "cdf",
         description: "Chiru",
         added: "2009-07-29"}).
iana(language,cdg,tag{
         type: "language",
         subtag: "cdg",
         description: "Chamari",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,cdh,tag{
         type: "language",
         subtag: "cdh",
         description: "Chambeali",
         added: "2009-07-29"}).
iana(language,cdi,tag{
         type: "language",
         subtag: "cdi",
         description: "Chodri",
         added: "2009-07-29"}).
iana(language,cdj,tag{
         type: "language",
         subtag: "cdj",
         description: "Churahi",
         added: "2009-07-29"}).
iana(language,cdm,tag{
         type: "language",
         subtag: "cdm",
         description: "Chepang",
         added: "2009-07-29"}).
iana(language,cdn,tag{
         type: "language",
         subtag: "cdn",
         description: "Chaudangsi",
         added: "2009-07-29"}).
iana(language,cdo,tag{
         type: "language",
         subtag: "cdo",
         description: "Min Dong Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,cdr,tag{
         type: "language",
         subtag: "cdr",
         description: "Cinda-Regi-Tiyal",
         added: "2009-07-29"}).
iana(language,cds,tag{
         type: "language",
         subtag: "cds",
         description: "Chadian Sign Language",
         added: "2009-07-29"}).
iana(language,cdy,tag{
         type: "language",
         subtag: "cdy",
         description: "Chadong",
         added: "2009-07-29"}).
iana(language,cdz,tag{
         type: "language",
         subtag: "cdz",
         description: "Koda",
         added: "2009-07-29"}).
iana(language,cea,tag{
         type: "language",
         subtag: "cea",
         description: "Lower Chehalis",
         added: "2009-07-29"}).
iana(language,ceb,tag{
         type: "language",
         subtag: "ceb",
         description: "Cebuano",
         added: "2005-10-16"}).
iana(language,ceg,tag{
         type: "language",
         subtag: "ceg",
         description: "Chamacoco",
         added: "2009-07-29"}).
iana(language,cek,tag{
         type: "language",
         subtag: "cek",
         description: "Eastern Khumi Chin",
         added: "2012-08-12"}).
iana(language,cel,tag{
         type: "language",
         subtag: "cel",
         description: "Celtic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cen,tag{
         type: "language",
         subtag: "cen",
         description: "Cen",
         added: "2009-07-29"}).
iana(language,cet,tag{
         type: "language",
         subtag: "cet",
         description: "Centúúm",
         added: "2009-07-29"}).
iana(language,cey,tag{
         type: "language",
         subtag: "cey",
         description: "Ekai Chin",
         added: "2019-04-16"}).
iana(language,cfa,tag{
         type: "language",
         subtag: "cfa",
         description: "Dijim-Bwilim",
         added: "2009-07-29"}).
iana(language,cfd,tag{
         type: "language",
         subtag: "cfd",
         description: "Cara",
         added: "2009-07-29"}).
iana(language,cfg,tag{
         type: "language",
         subtag: "cfg",
         description: "Como Karim",
         added: "2009-07-29"}).
iana(language,cfm,tag{
         type: "language",
         subtag: "cfm",
         description: "Falam Chin",
         added: "2009-07-29"}).
iana(language,cga,tag{
         type: "language",
         subtag: "cga",
         description: "Changriwa",
         added: "2009-07-29"}).
iana(language,cgc,tag{
         type: "language",
         subtag: "cgc",
         description: "Kagayanen",
         added: "2009-07-29"}).
iana(language,cgg,tag{
         type: "language",
         subtag: "cgg",
         description: "Chiga",
         added: "2009-07-29"}).
iana(language,cgk,tag{
         type: "language",
         subtag: "cgk",
         description: "Chocangacakha",
         added: "2009-07-29"}).
iana(language,chb,tag{
         type: "language",
         subtag: "chb",
         description: "Chibcha",
         added: "2005-10-16"}).
iana(language,chc,tag{
         type: "language",
         subtag: "chc",
         description: "Catawba",
         added: "2009-07-29"}).
iana(language,chd,tag{
         type: "language",
         subtag: "chd",
         description: "Highland Oaxaca Chontal",
         added: "2009-07-29"}).
iana(language,chf,tag{
         type: "language",
         subtag: "chf",
         description: "Tabasco Chontal",
         added: "2009-07-29"}).
iana(language,chg,tag{
         type: "language",
         subtag: "chg",
         description: "Chagatai",
         added: "2005-10-16"}).
iana(language,chh,tag{
         type: "language",
         subtag: "chh",
         description: "Chinook",
         added: "2009-07-29"}).
iana(language,chj,tag{
         type: "language",
         subtag: "chj",
         description: "Ojitlán Chinantec",
         added: "2009-07-29"}).
iana(language,chk,tag{
         type: "language",
         subtag: "chk",
         description: "Chuukese",
         added: "2005-10-16"}).
iana(language,chl,tag{
         type: "language",
         subtag: "chl",
         description: "Cahuilla",
         added: "2009-07-29"}).
iana(language,chm,tag{
         type: "language",
         subtag: "chm",
         description: "Mari (Russia)",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,chn,tag{
         type: "language",
         subtag: "chn",
         description: "Chinook jargon",
         added: "2005-10-16"}).
iana(language,cho,tag{
         type: "language",
         subtag: "cho",
         description: "Choctaw",
         added: "2005-10-16"}).
iana(language,chp,tag{
         type: "language",
         subtag: "chp",
         description: ["Chipewyan","Dene Suline"],
         added: "2005-10-16"}).
iana(language,chq,tag{
         type: "language",
         subtag: "chq",
         description: "Quiotepec Chinantec",
         added: "2009-07-29"}).
iana(language,chr,tag{
         type: "language",
         subtag: "chr",
         description: "Cherokee",
         added: "2005-10-16"}).
iana(language,cht,tag{
         type: "language",
         subtag: "cht",
         description: "Cholón",
         added: "2009-07-29"}).
iana(language,chw,tag{
         type: "language",
         subtag: "chw",
         description: "Chuwabu",
         added: "2009-07-29"}).
iana(language,chx,tag{
         type: "language",
         subtag: "chx",
         description: "Chantyal",
         added: "2009-07-29"}).
iana(language,chy,tag{
         type: "language",
         subtag: "chy",
         description: "Cheyenne",
         added: "2005-10-16"}).
iana(language,chz,tag{
         type: "language",
         subtag: "chz",
         description: "Ozumacín Chinantec",
         added: "2009-07-29"}).
iana(language,cia,tag{
         type: "language",
         subtag: "cia",
         description: "Cia-Cia",
         added: "2009-07-29"}).
iana(language,cib,tag{
         type: "language",
         subtag: "cib",
         description: "Ci Gbe",
         added: "2009-07-29"}).
iana(language,cic,tag{
         type: "language",
         subtag: "cic",
         description: "Chickasaw",
         added: "2009-07-29"}).
iana(language,cid,tag{
         type: "language",
         subtag: "cid",
         description: "Chimariko",
         added: "2009-07-29"}).
iana(language,cie,tag{
         type: "language",
         subtag: "cie",
         description: "Cineni",
         added: "2009-07-29"}).
iana(language,cih,tag{
         type: "language",
         subtag: "cih",
         description: "Chinali",
         added: "2009-07-29"}).
iana(language,cik,tag{
         type: "language",
         subtag: "cik",
         description: "Chitkuli Kinnauri",
         added: "2009-07-29"}).
iana(language,cim,tag{
         type: "language",
         subtag: "cim",
         description: "Cimbrian",
         added: "2009-07-29"}).
iana(language,cin,tag{
         type: "language",
         subtag: "cin",
         description: "Cinta Larga",
         added: "2009-07-29"}).
iana(language,cip,tag{
         type: "language",
         subtag: "cip",
         description: "Chiapanec",
         added: "2009-07-29"}).
iana(language,cir,tag{
         type: "language",
         subtag: "cir",
         description: ["Tiri","Haméa","Méa"],
         added: "2009-07-29"}).
iana(language,ciw,tag{
         type: "language",
         subtag: "ciw",
         description: "Chippewa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,ciy,tag{
         type: "language",
         subtag: "ciy",
         description: "Chaima",
         added: "2009-07-29"}).
iana(language,cja,tag{
         type: "language",
         subtag: "cja",
         description: "Western Cham",
         added: "2009-07-29"}).
iana(language,cje,tag{
         type: "language",
         subtag: "cje",
         description: "Chru",
         added: "2009-07-29"}).
iana(language,cjh,tag{
         type: "language",
         subtag: "cjh",
         description: "Upper Chehalis",
         added: "2009-07-29"}).
iana(language,cji,tag{
         type: "language",
         subtag: "cji",
         description: "Chamalal",
         added: "2009-07-29"}).
iana(language,cjk,tag{
         type: "language",
         subtag: "cjk",
         description: "Chokwe",
         added: "2009-07-29"}).
iana(language,cjm,tag{
         type: "language",
         subtag: "cjm",
         description: "Eastern Cham",
         added: "2009-07-29"}).
iana(language,cjn,tag{
         type: "language",
         subtag: "cjn",
         description: "Chenapian",
         added: "2009-07-29"}).
iana(language,cjo,tag{
         type: "language",
         subtag: "cjo",
         description: "Ashéninka Pajonal",
         added: "2009-07-29"}).
iana(language,cjp,tag{
         type: "language",
         subtag: "cjp",
         description: "Cabécar",
         added: "2009-07-29"}).
iana(language,cjr,tag{
         type: "language",
         subtag: "cjr",
         description: "Chorotega",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "mom"}).
iana(language,cjs,tag{
         type: "language",
         subtag: "cjs",
         description: "Shor",
         added: "2009-07-29"}).
iana(language,cjv,tag{
         type: "language",
         subtag: "cjv",
         description: "Chuave",
         added: "2009-07-29"}).
iana(language,cjy,tag{
         type: "language",
         subtag: "cjy",
         description: "Jinyu Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,cka,tag{
         type: "language",
         subtag: "cka",
         description: "Khumi Awa Chin",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "cmr"}).
iana(language,ckb,tag{
         type: "language",
         subtag: "ckb",
         description: "Central Kurdish",
         added: "2009-07-29",
         macrolanguage: "ku"}).
iana(language,ckh,tag{
         type: "language",
         subtag: "ckh",
         description: "Chak",
         added: "2009-07-29"}).
iana(language,ckl,tag{
         type: "language",
         subtag: "ckl",
         description: "Cibak",
         added: "2009-07-29"}).
iana(language,ckm,tag{
         type: "language",
         subtag: "ckm",
         description: "Chakavian",
         added: "2020-03-28"}).
iana(language,ckn,tag{
         type: "language",
         subtag: "ckn",
         description: "Kaang Chin",
         added: "2013-09-10"}).
iana(language,cko,tag{
         type: "language",
         subtag: "cko",
         description: "Anufo",
         added: "2009-07-29"}).
iana(language,ckq,tag{
         type: "language",
         subtag: "ckq",
         description: "Kajakse",
         added: "2009-07-29"}).
iana(language,ckr,tag{
         type: "language",
         subtag: "ckr",
         description: "Kairak",
         added: "2009-07-29"}).
iana(language,cks,tag{
         type: "language",
         subtag: "cks",
         description: "Tayo",
         added: "2009-07-29"}).
iana(language,ckt,tag{
         type: "language",
         subtag: "ckt",
         description: "Chukot",
         added: "2009-07-29"}).
iana(language,cku,tag{
         type: "language",
         subtag: "cku",
         description: "Koasati",
         added: "2009-07-29"}).
iana(language,ckv,tag{
         type: "language",
         subtag: "ckv",
         description: "Kavalan",
         added: "2009-07-29"}).
iana(language,ckx,tag{
         type: "language",
         subtag: "ckx",
         description: "Caka",
         added: "2009-07-29"}).
iana(language,cky,tag{
         type: "language",
         subtag: "cky",
         description: "Cakfem-Mushere",
         added: "2009-07-29"}).
iana(language,ckz,tag{
         type: "language",
         subtag: "ckz",
         description: "Cakchiquel-Quiché Mixed Language",
         added: "2009-07-29"}).
iana(language,cla,tag{
         type: "language",
         subtag: "cla",
         description: "Ron",
         added: "2009-07-29"}).
iana(language,clc,tag{
         type: "language",
         subtag: "clc",
         description: "Chilcotin",
         added: "2009-07-29"}).
iana(language,cld,tag{
         type: "language",
         subtag: "cld",
         description: "Chaldean Neo-Aramaic",
         added: "2009-07-29",
         macrolanguage: "syr"}).
iana(language,cle,tag{
         type: "language",
         subtag: "cle",
         description: "Lealao Chinantec",
         added: "2009-07-29"}).
iana(language,clh,tag{
         type: "language",
         subtag: "clh",
         description: "Chilisso",
         added: "2009-07-29"}).
iana(language,cli,tag{
         type: "language",
         subtag: "cli",
         description: "Chakali",
         added: "2009-07-29"}).
iana(language,clj,tag{
         type: "language",
         subtag: "clj",
         description: "Laitu Chin",
         added: "2012-08-12"}).
iana(language,clk,tag{
         type: "language",
         subtag: "clk",
         description: "Idu-Mishmi",
         added: "2009-07-29"}).
iana(language,cll,tag{
         type: "language",
         subtag: "cll",
         description: "Chala",
         added: "2009-07-29"}).
iana(language,clm,tag{
         type: "language",
         subtag: "clm",
         description: "Clallam",
         added: "2009-07-29"}).
iana(language,clo,tag{
         type: "language",
         subtag: "clo",
         description: "Lowland Oaxaca Chontal",
         added: "2009-07-29"}).
iana(language,clt,tag{
         type: "language",
         subtag: "clt",
         description: "Lautu Chin",
         added: "2012-08-12"}).
iana(language,clu,tag{
         type: "language",
         subtag: "clu",
         description: "Caluyanun",
         added: "2009-07-29"}).
iana(language,clw,tag{
         type: "language",
         subtag: "clw",
         description: "Chulym",
         added: "2009-07-29"}).
iana(language,cly,tag{
         type: "language",
         subtag: "cly",
         description: "Eastern Highland Chatino",
         added: "2009-07-29"}).
iana(language,cma,tag{
         type: "language",
         subtag: "cma",
         description: "Maa",
         added: "2009-07-29"}).
iana(language,cmc,tag{
         type: "language",
         subtag: "cmc",
         description: "Chamic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cme,tag{
         type: "language",
         subtag: "cme",
         description: "Cerma",
         added: "2009-07-29"}).
iana(language,cmg,tag{
         type: "language",
         subtag: "cmg",
         description: "Classical Mongolian",
         added: "2009-07-29"}).
iana(language,cmi,tag{
         type: "language",
         subtag: "cmi",
         description: "Emberá-Chamí",
         added: "2009-07-29"}).
iana(language,cmk,tag{
         type: "language",
         subtag: "cmk",
         description: "Chimakum",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "xch"}).
iana(language,cml,tag{
         type: "language",
         subtag: "cml",
         description: "Campalagian",
         added: "2009-07-29"}).
iana(language,cmm,tag{
         type: "language",
         subtag: "cmm",
         description: "Michigamea",
         added: "2009-07-29"}).
iana(language,cmn,tag{
         type: "language",
         subtag: "cmn",
         description: "Mandarin Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,cmo,tag{
         type: "language",
         subtag: "cmo",
         description: "Central Mnong",
         added: "2009-07-29"}).
iana(language,cmr,tag{
         type: "language",
         subtag: "cmr",
         description: "Mro-Khimi Chin",
         added: "2009-07-29"}).
iana(language,cms,tag{
         type: "language",
         subtag: "cms",
         description: "Messapic",
         added: "2009-07-29"}).
iana(language,cmt,tag{
         type: "language",
         subtag: "cmt",
         description: "Camtho",
         added: "2009-07-29"}).
iana(language,cna,tag{
         type: "language",
         subtag: "cna",
         description: "Changthang",
         added: "2009-07-29"}).
iana(language,cnb,tag{
         type: "language",
         subtag: "cnb",
         description: "Chinbon Chin",
         added: "2009-07-29"}).
iana(language,cnc,tag{
         type: "language",
         subtag: "cnc",
         description: "Côông",
         added: "2009-07-29"}).
iana(language,cng,tag{
         type: "language",
         subtag: "cng",
         description: "Northern Qiang",
         added: "2009-07-29"}).
iana(language,cnh,tag{
         type: "language",
         subtag: "cnh",
         description: ["Hakha Chin","Haka Chin"],
         added: "2009-07-29"}).
iana(language,cni,tag{
         type: "language",
         subtag: "cni",
         description: "Asháninka",
         added: "2009-07-29"}).
iana(language,cnk,tag{
         type: "language",
         subtag: "cnk",
         description: "Khumi Chin",
         added: "2009-07-29"}).
iana(language,cnl,tag{
         type: "language",
         subtag: "cnl",
         description: "Lalana Chinantec",
         added: "2009-07-29"}).
iana(language,cno,tag{
         type: "language",
         subtag: "cno",
         description: "Con",
         added: "2009-07-29"}).
iana(language,cnp,tag{
         type: "language",
         subtag: "cnp",
         description: ["Northern Ping Chinese","Northern Pinghua"],
         added: "2020-03-28",
         macrolanguage: "zh"}).
iana(language,cnq,tag{
         type: "language",
         subtag: "cnq",
         description: "Chung",
         added: "2022-02-25"}).
iana(language,cnr,tag{
         type: "language",
         subtag: "cnr",
         description: "Montenegrin",
         added: "2018-01-23",
         macrolanguage: "sh",
         comments: "see sr for Serbian"}).
iana(language,cns,tag{
         type: "language",
         subtag: "cns",
         description: "Central Asmat",
         added: "2009-07-29"}).
iana(language,cnt,tag{
         type: "language",
         subtag: "cnt",
         description: "Tepetotutla Chinantec",
         added: "2009-07-29"}).
iana(language,cnu,tag{
         type: "language",
         subtag: "cnu",
         description: "Chenoua",
         added: "2009-07-29"}).
iana(language,cnw,tag{
         type: "language",
         subtag: "cnw",
         description: "Ngawn Chin",
         added: "2009-07-29"}).
iana(language,cnx,tag{
         type: "language",
         subtag: "cnx",
         description: "Middle Cornish",
         added: "2009-07-29"}).
iana(language,coa,tag{
         type: "language",
         subtag: "coa",
         description: "Cocos Islands Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,cob,tag{
         type: "language",
         subtag: "cob",
         description: "Chicomuceltec",
         added: "2009-07-29"}).
iana(language,coc,tag{
         type: "language",
         subtag: "coc",
         description: "Cocopa",
         added: "2009-07-29"}).
iana(language,cod,tag{
         type: "language",
         subtag: "cod",
         description: "Cocama-Cocamilla",
         added: "2009-07-29"}).
iana(language,coe,tag{
         type: "language",
         subtag: "coe",
         description: "Koreguaje",
         added: "2009-07-29"}).
iana(language,cof,tag{
         type: "language",
         subtag: "cof",
         description: "Colorado",
         added: "2009-07-29"}).
iana(language,cog,tag{
         type: "language",
         subtag: "cog",
         description: "Chong",
         added: "2009-07-29"}).
iana(language,coh,tag{
         type: "language",
         subtag: "coh",
         description: ["Chonyi-Dzihana-Kauma","Chichonyi-Chidzihana-Chikauma"],
         added: "2009-07-29"}).
iana(language,coj,tag{
         type: "language",
         subtag: "coj",
         description: "Cochimi",
         added: "2009-07-29"}).
iana(language,cok,tag{
         type: "language",
         subtag: "cok",
         description: "Santa Teresa Cora",
         added: "2009-07-29"}).
iana(language,col,tag{
         type: "language",
         subtag: "col",
         description: "Columbia-Wenatchi",
         added: "2009-07-29"}).
iana(language,com,tag{
         type: "language",
         subtag: "com",
         description: "Comanche",
         added: "2009-07-29"}).
iana(language,con,tag{
         type: "language",
         subtag: "con",
         description: "Cofán",
         added: "2009-07-29"}).
iana(language,coo,tag{
         type: "language",
         subtag: "coo",
         description: "Comox",
         added: "2009-07-29"}).
iana(language,cop,tag{
         type: "language",
         subtag: "cop",
         description: "Coptic",
         added: "2005-10-16"}).
iana(language,coq,tag{
         type: "language",
         subtag: "coq",
         description: "Coquille",
         added: "2009-07-29"}).
iana(language,cot,tag{
         type: "language",
         subtag: "cot",
         description: "Caquinte",
         added: "2009-07-29"}).
iana(language,cou,tag{
         type: "language",
         subtag: "cou",
         description: "Wamey",
         added: "2009-07-29"}).
iana(language,cov,tag{
         type: "language",
         subtag: "cov",
         description: "Cao Miao",
         added: "2009-07-29"}).
iana(language,cow,tag{
         type: "language",
         subtag: "cow",
         description: "Cowlitz",
         added: "2009-07-29"}).
iana(language,cox,tag{
         type: "language",
         subtag: "cox",
         description: "Nanti",
         added: "2009-07-29"}).
iana(language,coy,tag{
         type: "language",
         subtag: "coy",
         description: "Coyaima",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "pij"}).
iana(language,coz,tag{
         type: "language",
         subtag: "coz",
         description: "Chochotec",
         added: "2009-07-29"}).
iana(language,cpa,tag{
         type: "language",
         subtag: "cpa",
         description: "Palantla Chinantec",
         added: "2009-07-29"}).
iana(language,cpb,tag{
         type: "language",
         subtag: "cpb",
         description: "Ucayali-Yurúa Ashéninka",
         added: "2009-07-29"}).
iana(language,cpc,tag{
         type: "language",
         subtag: "cpc",
         description: "Ajyíninka Apurucayali",
         added: "2009-07-29"}).
iana(language,cpe,tag{
         type: "language",
         subtag: "cpe",
         description: "English-based creoles and pidgins",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cpf,tag{
         type: "language",
         subtag: "cpf",
         description: "French-based creoles and pidgins",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cpg,tag{
         type: "language",
         subtag: "cpg",
         description: "Cappadocian Greek",
         added: "2009-07-29"}).
iana(language,cpi,tag{
         type: "language",
         subtag: "cpi",
         description: "Chinese Pidgin English",
         added: "2009-07-29"}).
iana(language,cpn,tag{
         type: "language",
         subtag: "cpn",
         description: "Cherepon",
         added: "2009-07-29"}).
iana(language,cpo,tag{
         type: "language",
         subtag: "cpo",
         description: "Kpeego",
         added: "2013-09-10"}).
iana(language,cpp,tag{
         type: "language",
         subtag: "cpp",
         description: "Portuguese-based creoles and pidgins",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cps,tag{
         type: "language",
         subtag: "cps",
         description: "Capiznon",
         added: "2009-07-29"}).
iana(language,cpu,tag{
         type: "language",
         subtag: "cpu",
         description: "Pichis Ashéninka",
         added: "2009-07-29"}).
iana(language,cpx,tag{
         type: "language",
         subtag: "cpx",
         description: "Pu-Xian Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,cpy,tag{
         type: "language",
         subtag: "cpy",
         description: "South Ucayali Ashéninka",
         added: "2009-07-29"}).
iana(language,cqd,tag{
         type: "language",
         subtag: "cqd",
         description: "Chuanqiandian Cluster Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,cqu,tag{
         type: "language",
         subtag: "cqu",
         description: "Chilean Quechua",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "quh",
         macrolanguage: "qu"}).
iana(language,cra,tag{
         type: "language",
         subtag: "cra",
         description: "Chara",
         added: "2009-07-29"}).
iana(language,crb,tag{
         type: "language",
         subtag: "crb",
         description: "Island Carib",
         added: "2009-07-29"}).
iana(language,crc,tag{
         type: "language",
         subtag: "crc",
         description: "Lonwolwol",
         added: "2009-07-29"}).
iana(language,crd,tag{
         type: "language",
         subtag: "crd",
         description: "Coeur d'Alene",
         added: "2009-07-29"}).
iana(language,crf,tag{
         type: "language",
         subtag: "crf",
         description: "Caramanta",
         added: "2009-07-29"}).
iana(language,crg,tag{
         type: "language",
         subtag: "crg",
         description: "Michif",
         added: "2009-07-29"}).
iana(language,crh,tag{
         type: "language",
         subtag: "crh",
         description: ["Crimean Tatar","Crimean Turkish"],
         added: "2005-10-16"}).
iana(language,cri,tag{
         type: "language",
         subtag: "cri",
         description: "Sãotomense",
         added: "2009-07-29"}).
iana(language,crj,tag{
         type: "language",
         subtag: "crj",
         description: "Southern East Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,crk,tag{
         type: "language",
         subtag: "crk",
         description: "Plains Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,crl,tag{
         type: "language",
         subtag: "crl",
         description: "Northern East Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,crm,tag{
         type: "language",
         subtag: "crm",
         description: "Moose Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,crn,tag{
         type: "language",
         subtag: "crn",
         description: "El Nayar Cora",
         added: "2009-07-29"}).
iana(language,cro,tag{
         type: "language",
         subtag: "cro",
         description: "Crow",
         added: "2009-07-29"}).
iana(language,crp,tag{
         type: "language",
         subtag: "crp",
         description: "Creoles and pidgins",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,crq,tag{
         type: "language",
         subtag: "crq",
         description: "Iyo'wujwa Chorote",
         added: "2009-07-29"}).
iana(language,crr,tag{
         type: "language",
         subtag: "crr",
         description: "Carolina Algonquian",
         added: "2009-07-29"}).
iana(language,crs,tag{
         type: "language",
         subtag: "crs",
         description: "Seselwa Creole French",
         added: "2009-07-29"}).
iana(language,crt,tag{
         type: "language",
         subtag: "crt",
         description: "Iyojwa'ja Chorote",
         added: "2009-07-29"}).
iana(language,crv,tag{
         type: "language",
         subtag: "crv",
         description: "Chaura",
         added: "2009-07-29"}).
iana(language,crw,tag{
         type: "language",
         subtag: "crw",
         description: "Chrau",
         added: "2009-07-29"}).
iana(language,crx,tag{
         type: "language",
         subtag: "crx",
         description: "Carrier",
         added: "2009-07-29"}).
iana(language,cry,tag{
         type: "language",
         subtag: "cry",
         description: "Cori",
         added: "2009-07-29"}).
iana(language,crz,tag{
         type: "language",
         subtag: "crz",
         description: "Cruzeño",
         added: "2009-07-29"}).
iana(language,csa,tag{
         type: "language",
         subtag: "csa",
         description: "Chiltepec Chinantec",
         added: "2009-07-29"}).
iana(language,csb,tag{
         type: "language",
         subtag: "csb",
         description: "Kashubian",
         added: "2005-10-16"}).
iana(language,csc,tag{
         type: "language",
         subtag: "csc",
         description: ["Catalan Sign Language","Lengua de señas catalana","Llengua de Signes Catalana"],
         added: "2009-07-29"}).
iana(language,csd,tag{
         type: "language",
         subtag: "csd",
         description: "Chiangmai Sign Language",
         added: "2009-07-29"}).
iana(language,cse,tag{
         type: "language",
         subtag: "cse",
         description: "Czech Sign Language",
         added: "2009-07-29"}).
iana(language,csf,tag{
         type: "language",
         subtag: "csf",
         description: "Cuba Sign Language",
         added: "2009-07-29"}).
iana(language,csg,tag{
         type: "language",
         subtag: "csg",
         description: "Chilean Sign Language",
         added: "2009-07-29"}).
iana(language,csh,tag{
         type: "language",
         subtag: "csh",
         description: "Asho Chin",
         added: "2009-07-29"}).
iana(language,csi,tag{
         type: "language",
         subtag: "csi",
         description: "Coast Miwok",
         added: "2009-07-29"}).
iana(language,csj,tag{
         type: "language",
         subtag: "csj",
         description: "Songlai Chin",
         added: "2013-09-10"}).
iana(language,csk,tag{
         type: "language",
         subtag: "csk",
         description: "Jola-Kasa",
         added: "2009-07-29"}).
iana(language,csl,tag{
         type: "language",
         subtag: "csl",
         description: "Chinese Sign Language",
         added: "2009-07-29"}).
iana(language,csm,tag{
         type: "language",
         subtag: "csm",
         description: "Central Sierra Miwok",
         added: "2009-07-29"}).
iana(language,csn,tag{
         type: "language",
         subtag: "csn",
         description: "Colombian Sign Language",
         added: "2009-07-29"}).
iana(language,cso,tag{
         type: "language",
         subtag: "cso",
         description: ["Sochiapam Chinantec","Sochiapan Chinantec"],
         added: "2009-07-29"}).
iana(language,csp,tag{
         type: "language",
         subtag: "csp",
         description: ["Southern Ping Chinese","Southern Pinghua"],
         added: "2020-03-28",
         macrolanguage: "zh"}).
iana(language,csq,tag{
         type: "language",
         subtag: "csq",
         description: "Croatia Sign Language",
         added: "2009-07-29"}).
iana(language,csr,tag{
         type: "language",
         subtag: "csr",
         description: "Costa Rican Sign Language",
         added: "2009-07-29"}).
iana(language,css,tag{
         type: "language",
         subtag: "css",
         description: "Southern Ohlone",
         added: "2009-07-29"}).
iana(language,cst,tag{
         type: "language",
         subtag: "cst",
         description: "Northern Ohlone",
         added: "2009-07-29"}).
iana(language,csu,tag{
         type: "language",
         subtag: "csu",
         description: "Central Sudanic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,csv,tag{
         type: "language",
         subtag: "csv",
         description: "Sumtu Chin",
         added: "2012-08-12"}).
iana(language,csw,tag{
         type: "language",
         subtag: "csw",
         description: "Swampy Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,csx,tag{
         type: "language",
         subtag: "csx",
         description: "Cambodian Sign Language",
         added: "2021-02-20"}).
iana(language,csy,tag{
         type: "language",
         subtag: "csy",
         description: "Siyin Chin",
         added: "2009-07-29"}).
iana(language,csz,tag{
         type: "language",
         subtag: "csz",
         description: "Coos",
         added: "2009-07-29"}).
iana(language,cta,tag{
         type: "language",
         subtag: "cta",
         description: "Tataltepec Chatino",
         added: "2009-07-29"}).
iana(language,ctc,tag{
         type: "language",
         subtag: "ctc",
         description: "Chetco",
         added: "2009-07-29"}).
iana(language,ctd,tag{
         type: "language",
         subtag: "ctd",
         description: "Tedim Chin",
         added: "2009-07-29"}).
iana(language,cte,tag{
         type: "language",
         subtag: "cte",
         description: "Tepinapa Chinantec",
         added: "2009-07-29"}).
iana(language,ctg,tag{
         type: "language",
         subtag: "ctg",
         description: "Chittagonian",
         added: "2009-07-29"}).
iana(language,cth,tag{
         type: "language",
         subtag: "cth",
         description: "Thaiphum Chin",
         added: "2012-08-12"}).
iana(language,ctl,tag{
         type: "language",
         subtag: "ctl",
         description: "Tlacoatzintepec Chinantec",
         added: "2009-07-29"}).
iana(language,ctm,tag{
         type: "language",
         subtag: "ctm",
         description: "Chitimacha",
         added: "2009-07-29"}).
iana(language,ctn,tag{
         type: "language",
         subtag: "ctn",
         description: "Chhintange",
         added: "2009-07-29"}).
iana(language,cto,tag{
         type: "language",
         subtag: "cto",
         description: "Emberá-Catío",
         added: "2009-07-29"}).
iana(language,ctp,tag{
         type: "language",
         subtag: "ctp",
         description: "Western Highland Chatino",
         added: "2009-07-29"}).
iana(language,cts,tag{
         type: "language",
         subtag: "cts",
         description: "Northern Catanduanes Bikol",
         added: "2009-07-29",
         macrolanguage: "bik"}).
iana(language,ctt,tag{
         type: "language",
         subtag: "ctt",
         description: "Wayanad Chetti",
         added: "2009-07-29"}).
iana(language,ctu,tag{
         type: "language",
         subtag: "ctu",
         description: "Chol",
         added: "2009-07-29"}).
iana(language,cty,tag{
         type: "language",
         subtag: "cty",
         description: "Moundadan Chetty",
         added: "2021-02-20"}).
iana(language,ctz,tag{
         type: "language",
         subtag: "ctz",
         description: "Zacatepec Chatino",
         added: "2009-07-29"}).
iana(language,cua,tag{
         type: "language",
         subtag: "cua",
         description: "Cua",
         added: "2009-07-29"}).
iana(language,cub,tag{
         type: "language",
         subtag: "cub",
         description: "Cubeo",
         added: "2009-07-29"}).
iana(language,cuc,tag{
         type: "language",
         subtag: "cuc",
         description: "Usila Chinantec",
         added: "2009-07-29"}).
iana(language,cug,tag{
         type: "language",
         subtag: "cug",
         description: ["Chungmboko","Cung"],
         added: "2009-07-29",
         deprecated: "2022-02-25",
         comments: "see bpc, cnq"}).
iana(language,cuh,tag{
         type: "language",
         subtag: "cuh",
         description: ["Chuka","Gichuka"],
         added: "2009-07-29"}).
iana(language,cui,tag{
         type: "language",
         subtag: "cui",
         description: "Cuiba",
         added: "2009-07-29"}).
iana(language,cuj,tag{
         type: "language",
         subtag: "cuj",
         description: "Mashco Piro",
         added: "2009-07-29"}).
iana(language,cuk,tag{
         type: "language",
         subtag: "cuk",
         description: "San Blas Kuna",
         added: "2009-07-29"}).
iana(language,cul,tag{
         type: "language",
         subtag: "cul",
         description: ["Culina","Kulina"],
         added: "2009-07-29"}).
iana(language,cum,tag{
         type: "language",
         subtag: "cum",
         description: "Cumeral",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,cuo,tag{
         type: "language",
         subtag: "cuo",
         description: "Cumanagoto",
         added: "2009-07-29"}).
iana(language,cup,tag{
         type: "language",
         subtag: "cup",
         description: "Cupeño",
         added: "2009-07-29"}).
iana(language,cuq,tag{
         type: "language",
         subtag: "cuq",
         description: "Cun",
         added: "2009-07-29"}).
iana(language,cur,tag{
         type: "language",
         subtag: "cur",
         description: "Chhulung",
         added: "2009-07-29"}).
iana(language,cus,tag{
         type: "language",
         subtag: "cus",
         description: "Cushitic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,cut,tag{
         type: "language",
         subtag: "cut",
         description: "Teutila Cuicatec",
         added: "2009-07-29"}).
iana(language,cuu,tag{
         type: "language",
         subtag: "cuu",
         description: "Tai Ya",
         added: "2009-07-29"}).
iana(language,cuv,tag{
         type: "language",
         subtag: "cuv",
         description: "Cuvok",
         added: "2009-07-29"}).
iana(language,cuw,tag{
         type: "language",
         subtag: "cuw",
         description: "Chukwa",
         added: "2009-07-29"}).
iana(language,cux,tag{
         type: "language",
         subtag: "cux",
         description: "Tepeuxila Cuicatec",
         added: "2009-07-29"}).
iana(language,cuy,tag{
         type: "language",
         subtag: "cuy",
         description: "Cuitlatec",
         added: "2018-03-08"}).
iana(language,cvg,tag{
         type: "language",
         subtag: "cvg",
         description: "Chug",
         added: "2009-07-29"}).
iana(language,cvn,tag{
         type: "language",
         subtag: "cvn",
         description: "Valle Nacional Chinantec",
         added: "2009-07-29"}).
iana(language,cwa,tag{
         type: "language",
         subtag: "cwa",
         description: "Kabwa",
         added: "2009-07-29"}).
iana(language,cwb,tag{
         type: "language",
         subtag: "cwb",
         description: "Maindo",
         added: "2009-07-29"}).
iana(language,cwd,tag{
         type: "language",
         subtag: "cwd",
         description: "Woods Cree",
         added: "2009-07-29",
         macrolanguage: "cr"}).
iana(language,cwe,tag{
         type: "language",
         subtag: "cwe",
         description: "Kwere",
         added: "2009-07-29"}).
iana(language,cwg,tag{
         type: "language",
         subtag: "cwg",
         description: ["Chewong","Cheq Wong"],
         added: "2009-07-29"}).
iana(language,cwt,tag{
         type: "language",
         subtag: "cwt",
         description: "Kuwaataay",
         added: "2009-07-29"}).
iana(language,cya,tag{
         type: "language",
         subtag: "cya",
         description: "Nopala Chatino",
         added: "2009-07-29"}).
iana(language,cyb,tag{
         type: "language",
         subtag: "cyb",
         description: "Cayubaba",
         added: "2009-07-29"}).
iana(language,cyo,tag{
         type: "language",
         subtag: "cyo",
         description: "Cuyonon",
         added: "2009-07-29"}).
iana(language,czh,tag{
         type: "language",
         subtag: "czh",
         description: "Huizhou Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,czk,tag{
         type: "language",
         subtag: "czk",
         description: "Knaanic",
         added: "2009-07-29"}).
iana(language,czn,tag{
         type: "language",
         subtag: "czn",
         description: "Zenzontepec Chatino",
         added: "2009-07-29"}).
iana(language,czo,tag{
         type: "language",
         subtag: "czo",
         description: "Min Zhong Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,czt,tag{
         type: "language",
         subtag: "czt",
         description: "Zotung Chin",
         added: "2009-07-29"}).
iana(language,daa,tag{
         type: "language",
         subtag: "daa",
         description: "Dangaléat",
         added: "2009-07-29"}).
iana(language,dac,tag{
         type: "language",
         subtag: "dac",
         description: "Dambi",
         added: "2009-07-29"}).
iana(language,dad,tag{
         type: "language",
         subtag: "dad",
         description: "Marik",
         added: "2009-07-29"}).
iana(language,dae,tag{
         type: "language",
         subtag: "dae",
         description: "Duupa",
         added: "2009-07-29"}).
iana(language,daf,tag{
         type: "language",
         subtag: "daf",
         description: "Dan",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see dnj, lda"}).
iana(language,dag,tag{
         type: "language",
         subtag: "dag",
         description: "Dagbani",
         added: "2009-07-29"}).
iana(language,dah,tag{
         type: "language",
         subtag: "dah",
         description: "Gwahatike",
         added: "2009-07-29"}).
iana(language,dai,tag{
         type: "language",
         subtag: "dai",
         description: "Day",
         added: "2009-07-29"}).
iana(language,daj,tag{
         type: "language",
         subtag: "daj",
         description: "Dar Fur Daju",
         added: "2009-07-29"}).
iana(language,dak,tag{
         type: "language",
         subtag: "dak",
         description: "Dakota",
         added: "2005-10-16"}).
iana(language,dal,tag{
         type: "language",
         subtag: "dal",
         description: "Dahalo",
         added: "2009-07-29"}).
iana(language,dam,tag{
         type: "language",
         subtag: "dam",
         description: "Damakawa",
         added: "2009-07-29"}).
iana(language,dao,tag{
         type: "language",
         subtag: "dao",
         description: "Daai Chin",
         added: "2009-07-29"}).
iana(language,dap,tag{
         type: "language",
         subtag: "dap",
         description: "Nisi (India)",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see njz, tgj"}).
iana(language,daq,tag{
         type: "language",
         subtag: "daq",
         description: "Dandami Maria",
         added: "2009-07-29"}).
iana(language,dar,tag{
         type: "language",
         subtag: "dar",
         description: "Dargwa",
         added: "2005-10-16"}).
iana(language,das,tag{
         type: "language",
         subtag: "das",
         description: "Daho-Doo",
         added: "2009-07-29"}).
iana(language,dau,tag{
         type: "language",
         subtag: "dau",
         description: "Dar Sila Daju",
         added: "2009-07-29"}).
iana(language,dav,tag{
         type: "language",
         subtag: "dav",
         description: ["Taita","Dawida"],
         added: "2009-07-29"}).
iana(language,daw,tag{
         type: "language",
         subtag: "daw",
         description: "Davawenyo",
         added: "2009-07-29"}).
iana(language,dax,tag{
         type: "language",
         subtag: "dax",
         description: "Dayi",
         added: "2009-07-29"}).
iana(language,day,tag{
         type: "language",
         subtag: "day",
         description: "Land Dayak languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,daz,tag{
         type: "language",
         subtag: "daz",
         description: "Dao",
         added: "2009-07-29"}).
iana(language,dba,tag{
         type: "language",
         subtag: "dba",
         description: "Bangime",
         added: "2009-07-29"}).
iana(language,dbb,tag{
         type: "language",
         subtag: "dbb",
         description: "Deno",
         added: "2009-07-29"}).
iana(language,dbd,tag{
         type: "language",
         subtag: "dbd",
         description: "Dadiya",
         added: "2009-07-29"}).
iana(language,dbe,tag{
         type: "language",
         subtag: "dbe",
         description: "Dabe",
         added: "2009-07-29"}).
iana(language,dbf,tag{
         type: "language",
         subtag: "dbf",
         description: "Edopi",
         added: "2009-07-29"}).
iana(language,dbg,tag{
         type: "language",
         subtag: "dbg",
         description: "Dogul Dom Dogon",
         added: "2009-07-29"}).
iana(language,dbi,tag{
         type: "language",
         subtag: "dbi",
         description: "Doka",
         added: "2009-07-29"}).
iana(language,dbj,tag{
         type: "language",
         subtag: "dbj",
         description: "Ida'an",
         added: "2009-07-29"}).
iana(language,dbl,tag{
         type: "language",
         subtag: "dbl",
         description: "Dyirbal",
         added: "2009-07-29"}).
iana(language,dbm,tag{
         type: "language",
         subtag: "dbm",
         description: "Duguri",
         added: "2009-07-29"}).
iana(language,dbn,tag{
         type: "language",
         subtag: "dbn",
         description: "Duriankere",
         added: "2009-07-29"}).
iana(language,dbo,tag{
         type: "language",
         subtag: "dbo",
         description: "Dulbu",
         added: "2009-07-29"}).
iana(language,dbp,tag{
         type: "language",
         subtag: "dbp",
         description: "Duwai",
         added: "2009-07-29"}).
iana(language,dbq,tag{
         type: "language",
         subtag: "dbq",
         description: "Daba",
         added: "2009-07-29"}).
iana(language,dbr,tag{
         type: "language",
         subtag: "dbr",
         description: "Dabarre",
         added: "2009-07-29"}).
iana(language,dbt,tag{
         type: "language",
         subtag: "dbt",
         description: "Ben Tey Dogon",
         added: "2012-08-12"}).
iana(language,dbu,tag{
         type: "language",
         subtag: "dbu",
         description: "Bondum Dom Dogon",
         added: "2009-07-29"}).
iana(language,dbv,tag{
         type: "language",
         subtag: "dbv",
         description: "Dungu",
         added: "2009-07-29"}).
iana(language,dbw,tag{
         type: "language",
         subtag: "dbw",
         description: "Bankan Tey Dogon",
         added: "2012-08-12"}).
iana(language,dby,tag{
         type: "language",
         subtag: "dby",
         description: "Dibiyaso",
         added: "2009-07-29"}).
iana(language,dcc,tag{
         type: "language",
         subtag: "dcc",
         description: "Deccan",
         added: "2009-07-29"}).
iana(language,dcr,tag{
         type: "language",
         subtag: "dcr",
         description: "Negerhollands",
         added: "2009-07-29"}).
iana(language,dda,tag{
         type: "language",
         subtag: "dda",
         description: "Dadi Dadi",
         added: "2012-08-12"}).
iana(language,ddd,tag{
         type: "language",
         subtag: "ddd",
         description: "Dongotono",
         added: "2009-07-29"}).
iana(language,dde,tag{
         type: "language",
         subtag: "dde",
         description: "Doondo",
         added: "2009-07-29"}).
iana(language,ddg,tag{
         type: "language",
         subtag: "ddg",
         description: "Fataluku",
         added: "2009-07-29"}).
iana(language,ddi,tag{
         type: "language",
         subtag: "ddi",
         description: "West Goodenough",
         added: "2009-07-29"}).
iana(language,ddj,tag{
         type: "language",
         subtag: "ddj",
         description: "Jaru",
         added: "2009-07-29"}).
iana(language,ddn,tag{
         type: "language",
         subtag: "ddn",
         description: "Dendi (Benin)",
         added: "2009-07-29"}).
iana(language,ddo,tag{
         type: "language",
         subtag: "ddo",
         description: "Dido",
         added: "2009-07-29"}).
iana(language,ddr,tag{
         type: "language",
         subtag: "ddr",
         description: "Dhudhuroa",
         added: "2012-08-12"}).
iana(language,dds,tag{
         type: "language",
         subtag: "dds",
         description: "Donno So Dogon",
         added: "2009-07-29"}).
iana(language,ddw,tag{
         type: "language",
         subtag: "ddw",
         description: "Dawera-Daweloor",
         added: "2009-07-29"}).
iana(language,dec,tag{
         type: "language",
         subtag: "dec",
         description: "Dagik",
         added: "2009-07-29"}).
iana(language,ded,tag{
         type: "language",
         subtag: "ded",
         description: "Dedua",
         added: "2009-07-29"}).
iana(language,dee,tag{
         type: "language",
         subtag: "dee",
         description: "Dewoin",
         added: "2009-07-29"}).
iana(language,def,tag{
         type: "language",
         subtag: "def",
         description: "Dezfuli",
         added: "2009-07-29"}).
iana(language,deg,tag{
         type: "language",
         subtag: "deg",
         description: "Degema",
         added: "2009-07-29"}).
iana(language,deh,tag{
         type: "language",
         subtag: "deh",
         description: "Dehwari",
         added: "2009-07-29"}).
iana(language,dei,tag{
         type: "language",
         subtag: "dei",
         description: "Demisa",
         added: "2009-07-29"}).
iana(language,dek,tag{
         type: "language",
         subtag: "dek",
         description: "Dek",
         added: "2009-07-29"}).
iana(language,del,tag{
         type: "language",
         subtag: "del",
         description: "Delaware",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,dem,tag{
         type: "language",
         subtag: "dem",
         description: "Dem",
         added: "2009-07-29"}).
iana(language,den,tag{
         type: "language",
         subtag: "den",
         description: "Slave (Athapascan)",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,dep,tag{
         type: "language",
         subtag: "dep",
         description: "Pidgin Delaware",
         added: "2009-07-29"}).
iana(language,deq,tag{
         type: "language",
         subtag: "deq",
         description: "Dendi (Central African Republic)",
         added: "2009-07-29"}).
iana(language,der,tag{
         type: "language",
         subtag: "der",
         description: "Deori",
         added: "2009-07-29"}).
iana(language,des,tag{
         type: "language",
         subtag: "des",
         description: "Desano",
         added: "2009-07-29"}).
iana(language,dev,tag{
         type: "language",
         subtag: "dev",
         description: "Domung",
         added: "2009-07-29"}).
iana(language,dez,tag{
         type: "language",
         subtag: "dez",
         description: "Dengese",
         added: "2009-07-29"}).
iana(language,dga,tag{
         type: "language",
         subtag: "dga",
         description: "Southern Dagaare",
         added: "2009-07-29"}).
iana(language,dgb,tag{
         type: "language",
         subtag: "dgb",
         description: "Bunoge Dogon",
         added: "2009-07-29"}).
iana(language,dgc,tag{
         type: "language",
         subtag: "dgc",
         description: "Casiguran Dumagat Agta",
         added: "2009-07-29"}).
iana(language,dgd,tag{
         type: "language",
         subtag: "dgd",
         description: "Dagaari Dioula",
         added: "2009-07-29"}).
iana(language,dge,tag{
         type: "language",
         subtag: "dge",
         description: "Degenan",
         added: "2009-07-29"}).
iana(language,dgg,tag{
         type: "language",
         subtag: "dgg",
         description: "Doga",
         added: "2009-07-29"}).
iana(language,dgh,tag{
         type: "language",
         subtag: "dgh",
         description: "Dghwede",
         added: "2009-07-29"}).
iana(language,dgi,tag{
         type: "language",
         subtag: "dgi",
         description: "Northern Dagara",
         added: "2009-07-29"}).
iana(language,dgk,tag{
         type: "language",
         subtag: "dgk",
         description: "Dagba",
         added: "2009-07-29"}).
iana(language,dgl,tag{
         type: "language",
         subtag: "dgl",
         description: ["Andaandi","Dongolawi"],
         added: "2012-08-12"}).
iana(language,dgn,tag{
         type: "language",
         subtag: "dgn",
         description: "Dagoman",
         added: "2009-07-29"}).
iana(language,dgo,tag{
         type: "language",
         subtag: "dgo",
         description: "Dogri (individual language)",
         added: "2009-07-29",
         macrolanguage: "doi"}).
iana(language,dgr,tag{
         type: "language",
         subtag: "dgr",
         description: ["Dogrib","Tłı̨chǫ"],
         added: "2005-10-16"}).
iana(language,dgs,tag{
         type: "language",
         subtag: "dgs",
         description: "Dogoso",
         added: "2009-07-29"}).
iana(language,dgt,tag{
         type: "language",
         subtag: "dgt",
         description: "Ndra'ngith",
         added: "2012-08-12"}).
iana(language,dgu,tag{
         type: "language",
         subtag: "dgu",
         description: "Degaru",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,dgw,tag{
         type: "language",
         subtag: "dgw",
         description: "Daungwurrung",
         added: "2012-08-12"}).
iana(language,dgx,tag{
         type: "language",
         subtag: "dgx",
         description: "Doghoro",
         added: "2009-07-29"}).
iana(language,dgz,tag{
         type: "language",
         subtag: "dgz",
         description: "Daga",
         added: "2009-07-29"}).
iana(language,dha,tag{
         type: "language",
         subtag: "dha",
         description: "Dhanwar (India)",
         added: "2009-07-29",
         deprecated: "2011-08-16"}).
iana(language,dhd,tag{
         type: "language",
         subtag: "dhd",
         description: "Dhundari",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,dhg,tag{
         type: "language",
         subtag: "dhg",
         description: ["Dhangu-Djangu","Dhangu","Djangu"],
         added: "2009-07-29"}).
iana(language,dhi,tag{
         type: "language",
         subtag: "dhi",
         description: "Dhimal",
         added: "2009-07-29"}).
iana(language,dhl,tag{
         type: "language",
         subtag: "dhl",
         description: "Dhalandji",
         added: "2009-07-29"}).
iana(language,dhm,tag{
         type: "language",
         subtag: "dhm",
         description: "Zemba",
         added: "2009-07-29"}).
iana(language,dhn,tag{
         type: "language",
         subtag: "dhn",
         description: "Dhanki",
         added: "2009-07-29"}).
iana(language,dho,tag{
         type: "language",
         subtag: "dho",
         description: "Dhodia",
         added: "2009-07-29"}).
iana(language,dhr,tag{
         type: "language",
         subtag: "dhr",
         description: "Dhargari",
         added: "2009-07-29"}).
iana(language,dhs,tag{
         type: "language",
         subtag: "dhs",
         description: "Dhaiso",
         added: "2009-07-29"}).
iana(language,dhu,tag{
         type: "language",
         subtag: "dhu",
         description: "Dhurga",
         added: "2009-07-29"}).
iana(language,dhv,tag{
         type: "language",
         subtag: "dhv",
         description: ["Dehu","Drehu"],
         added: "2009-07-29"}).
iana(language,dhw,tag{
         type: "language",
         subtag: "dhw",
         description: "Dhanwar (Nepal)",
         added: "2009-07-29"}).
iana(language,dhx,tag{
         type: "language",
         subtag: "dhx",
         description: "Dhungaloo",
         added: "2013-09-10"}).
iana(language,dia,tag{
         type: "language",
         subtag: "dia",
         description: "Dia",
         added: "2009-07-29"}).
iana(language,dib,tag{
         type: "language",
         subtag: "dib",
         description: "South Central Dinka",
         added: "2009-07-29",
         macrolanguage: "din"}).
iana(language,dic,tag{
         type: "language",
         subtag: "dic",
         description: "Lakota Dida",
         added: "2009-07-29"}).
iana(language,did,tag{
         type: "language",
         subtag: "did",
         description: "Didinga",
         added: "2009-07-29"}).
iana(language,dif,tag{
         type: "language",
         subtag: "dif",
         description: ["Dieri","Diyari"],
         added: "2009-07-29"}).
iana(language,dig,tag{
         type: "language",
         subtag: "dig",
         description: ["Digo","Chidigo"],
         added: "2009-07-29"}).
iana(language,dih,tag{
         type: "language",
         subtag: "dih",
         description: "Kumiai",
         added: "2009-07-29"}).
iana(language,dii,tag{
         type: "language",
         subtag: "dii",
         description: "Dimbong",
         added: "2009-07-29"}).
iana(language,dij,tag{
         type: "language",
         subtag: "dij",
         description: "Dai",
         added: "2009-07-29"}).
iana(language,dik,tag{
         type: "language",
         subtag: "dik",
         description: "Southwestern Dinka",
         added: "2009-07-29",
         macrolanguage: "din"}).
iana(language,dil,tag{
         type: "language",
         subtag: "dil",
         description: "Dilling",
         added: "2009-07-29"}).
iana(language,dim,tag{
         type: "language",
         subtag: "dim",
         description: "Dime",
         added: "2009-07-29"}).
iana(language,din,tag{
         type: "language",
         subtag: "din",
         description: "Dinka",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,dio,tag{
         type: "language",
         subtag: "dio",
         description: "Dibo",
         added: "2009-07-29"}).
iana(language,dip,tag{
         type: "language",
         subtag: "dip",
         description: "Northeastern Dinka",
         added: "2009-07-29",
         macrolanguage: "din"}).
iana(language,diq,tag{
         type: "language",
         subtag: "diq",
         description: "Dimli (individual language)",
         added: "2009-07-29",
         macrolanguage: "zza"}).
iana(language,dir,tag{
         type: "language",
         subtag: "dir",
         description: "Dirim",
         added: "2009-07-29"}).
iana(language,dis,tag{
         type: "language",
         subtag: "dis",
         description: "Dimasa",
         added: "2009-07-29"}).
iana(language,dit,tag{
         type: "language",
         subtag: "dit",
         description: "Dirari",
         added: "2009-07-29",
         deprecated: "2019-04-29",
         preferred_value: "dif"}).
iana(language,diu,tag{
         type: "language",
         subtag: "diu",
         description: "Diriku",
         added: "2009-07-29"}).
iana(language,diw,tag{
         type: "language",
         subtag: "diw",
         description: "Northwestern Dinka",
         added: "2009-07-29",
         macrolanguage: "din"}).
iana(language,dix,tag{
         type: "language",
         subtag: "dix",
         description: "Dixon Reef",
         added: "2009-07-29"}).
iana(language,diy,tag{
         type: "language",
         subtag: "diy",
         description: "Diuwe",
         added: "2009-07-29"}).
iana(language,diz,tag{
         type: "language",
         subtag: "diz",
         description: "Ding",
         added: "2009-07-29"}).
iana(language,dja,tag{
         type: "language",
         subtag: "dja",
         description: "Djadjawurrung",
         added: "2012-08-12"}).
iana(language,djb,tag{
         type: "language",
         subtag: "djb",
         description: "Djinba",
         added: "2009-07-29"}).
iana(language,djc,tag{
         type: "language",
         subtag: "djc",
         description: "Dar Daju Daju",
         added: "2009-07-29"}).
iana(language,djd,tag{
         type: "language",
         subtag: "djd",
         description: ["Djamindjung","Ngaliwurru"],
         added: "2009-07-29"}).
iana(language,dje,tag{
         type: "language",
         subtag: "dje",
         description: "Zarma",
         added: "2009-07-29"}).
iana(language,djf,tag{
         type: "language",
         subtag: "djf",
         description: "Djangun",
         added: "2009-07-29"}).
iana(language,dji,tag{
         type: "language",
         subtag: "dji",
         description: "Djinang",
         added: "2009-07-29"}).
iana(language,djj,tag{
         type: "language",
         subtag: "djj",
         description: "Djeebbana",
         added: "2009-07-29"}).
iana(language,djk,tag{
         type: "language",
         subtag: "djk",
         description: ["Eastern Maroon Creole","Businenge Tongo","Nenge"],
         added: "2009-07-29"}).
iana(language,djl,tag{
         type: "language",
         subtag: "djl",
         description: "Djiwarli",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see dze, iin"}).
iana(language,djm,tag{
         type: "language",
         subtag: "djm",
         description: "Jamsay Dogon",
         added: "2009-07-29"}).
iana(language,djn,tag{
         type: "language",
         subtag: "djn",
         description: ["Jawoyn","Djauan"],
         added: "2009-07-29"}).
iana(language,djo,tag{
         type: "language",
         subtag: "djo",
         description: "Jangkang",
         added: "2009-07-29"}).
iana(language,djr,tag{
         type: "language",
         subtag: "djr",
         description: "Djambarrpuyngu",
         added: "2009-07-29"}).
iana(language,dju,tag{
         type: "language",
         subtag: "dju",
         description: "Kapriman",
         added: "2009-07-29"}).
iana(language,djw,tag{
         type: "language",
         subtag: "djw",
         description: "Djawi",
         added: "2009-07-29"}).
iana(language,dka,tag{
         type: "language",
         subtag: "dka",
         description: "Dakpakha",
         added: "2009-07-29"}).
iana(language,dkg,tag{
         type: "language",
         subtag: "dkg",
         description: "Kadung",
         added: "2021-02-20"}).
iana(language,dkk,tag{
         type: "language",
         subtag: "dkk",
         description: "Dakka",
         added: "2009-07-29"}).
iana(language,dkl,tag{
         type: "language",
         subtag: "dkl",
         description: "Kolum So Dogon",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         comments: "see aqd, dmb"}).
iana(language,dkr,tag{
         type: "language",
         subtag: "dkr",
         description: "Kuijau",
         added: "2009-07-29"}).
iana(language,dks,tag{
         type: "language",
         subtag: "dks",
         description: "Southeastern Dinka",
         added: "2009-07-29",
         macrolanguage: "din"}).
iana(language,dkx,tag{
         type: "language",
         subtag: "dkx",
         description: "Mazagway",
         added: "2009-07-29"}).
iana(language,dlg,tag{
         type: "language",
         subtag: "dlg",
         description: "Dolgan",
         added: "2009-07-29"}).
iana(language,dlk,tag{
         type: "language",
         subtag: "dlk",
         description: "Dahalik",
         added: "2013-09-10"}).
iana(language,dlm,tag{
         type: "language",
         subtag: "dlm",
         description: "Dalmatian",
         added: "2009-07-29"}).
iana(language,dln,tag{
         type: "language",
         subtag: "dln",
         description: "Darlong",
         added: "2009-07-29"}).
iana(language,dma,tag{
         type: "language",
         subtag: "dma",
         description: "Duma",
         added: "2009-07-29"}).
iana(language,dmb,tag{
         type: "language",
         subtag: "dmb",
         description: "Mombo Dogon",
         added: "2011-08-16"}).
iana(language,dmc,tag{
         type: "language",
         subtag: "dmc",
         description: "Gavak",
         added: "2009-07-29"}).
iana(language,dmd,tag{
         type: "language",
         subtag: "dmd",
         description: "Madhi Madhi",
         added: "2012-08-12"}).
iana(language,dme,tag{
         type: "language",
         subtag: "dme",
         description: "Dugwor",
         added: "2009-07-29"}).
iana(language,dmf,tag{
         type: "language",
         subtag: "dmf",
         description: "Medefaidrin",
         added: "2020-03-28"}).
iana(language,dmg,tag{
         type: "language",
         subtag: "dmg",
         description: "Upper Kinabatangan",
         added: "2009-07-29"}).
iana(language,dmk,tag{
         type: "language",
         subtag: "dmk",
         description: "Domaaki",
         added: "2009-07-29"}).
iana(language,dml,tag{
         type: "language",
         subtag: "dml",
         description: "Dameli",
         added: "2009-07-29"}).
iana(language,dmm,tag{
         type: "language",
         subtag: "dmm",
         description: "Dama",
         added: "2009-07-29"}).
iana(language,dmn,tag{
         type: "language",
         subtag: "dmn",
         description: "Mande languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,dmo,tag{
         type: "language",
         subtag: "dmo",
         description: "Kemedzung",
         added: "2009-07-29"}).
iana(language,dmr,tag{
         type: "language",
         subtag: "dmr",
         description: "East Damar",
         added: "2009-07-29"}).
iana(language,dms,tag{
         type: "language",
         subtag: "dms",
         description: "Dampelas",
         added: "2009-07-29"}).
iana(language,dmu,tag{
         type: "language",
         subtag: "dmu",
         description: ["Dubu","Tebi"],
         added: "2009-07-29"}).
iana(language,dmv,tag{
         type: "language",
         subtag: "dmv",
         description: "Dumpas",
         added: "2009-07-29"}).
iana(language,dmw,tag{
         type: "language",
         subtag: "dmw",
         description: "Mudburra",
         added: "2013-09-10"}).
iana(language,dmx,tag{
         type: "language",
         subtag: "dmx",
         description: "Dema",
         added: "2009-07-29"}).
iana(language,dmy,tag{
         type: "language",
         subtag: "dmy",
         description: ["Demta","Sowari"],
         added: "2009-07-29"}).
iana(language,dna,tag{
         type: "language",
         subtag: "dna",
         description: "Upper Grand Valley Dani",
         added: "2009-07-29"}).
iana(language,dnd,tag{
         type: "language",
         subtag: "dnd",
         description: "Daonda",
         added: "2009-07-29"}).
iana(language,dne,tag{
         type: "language",
         subtag: "dne",
         description: "Ndendeule",
         added: "2009-07-29"}).
iana(language,dng,tag{
         type: "language",
         subtag: "dng",
         description: "Dungan",
         added: "2009-07-29"}).
iana(language,dni,tag{
         type: "language",
         subtag: "dni",
         description: "Lower Grand Valley Dani",
         added: "2009-07-29"}).
iana(language,dnj,tag{
         type: "language",
         subtag: "dnj",
         description: "Dan",
         added: "2013-09-10"}).
iana(language,dnk,tag{
         type: "language",
         subtag: "dnk",
         description: "Dengka",
         added: "2009-07-29"}).
iana(language,dnn,tag{
         type: "language",
         subtag: "dnn",
         description: "Dzùùngoo",
         added: "2009-07-29"}).
iana(language,dno,tag{
         type: "language",
         subtag: "dno",
         description: ["Ndrulo","Northern Lendu"],
         added: "2018-10-28"}).
iana(language,dnr,tag{
         type: "language",
         subtag: "dnr",
         description: "Danaru",
         added: "2009-07-29"}).
iana(language,dnt,tag{
         type: "language",
         subtag: "dnt",
         description: "Mid Grand Valley Dani",
         added: "2009-07-29"}).
iana(language,dnu,tag{
         type: "language",
         subtag: "dnu",
         description: "Danau",
         added: "2009-07-29"}).
iana(language,dnv,tag{
         type: "language",
         subtag: "dnv",
         description: "Danu",
         added: "2012-08-12"}).
iana(language,dnw,tag{
         type: "language",
         subtag: "dnw",
         description: "Western Dani",
         added: "2009-07-29"}).
iana(language,dny,tag{
         type: "language",
         subtag: "dny",
         description: "Dení",
         added: "2009-07-29"}).
iana(language,doa,tag{
         type: "language",
         subtag: "doa",
         description: "Dom",
         added: "2009-07-29"}).
iana(language,dob,tag{
         type: "language",
         subtag: "dob",
         description: "Dobu",
         added: "2009-07-29"}).
iana(language,doc,tag{
         type: "language",
         subtag: "doc",
         description: "Northern Dong",
         added: "2009-07-29"}).
iana(language,doe,tag{
         type: "language",
         subtag: "doe",
         description: "Doe",
         added: "2009-07-29"}).
iana(language,dof,tag{
         type: "language",
         subtag: "dof",
         description: "Domu",
         added: "2009-07-29"}).
iana(language,doh,tag{
         type: "language",
         subtag: "doh",
         description: "Dong",
         added: "2009-07-29"}).
iana(language,doi,tag{
         type: "language",
         subtag: "doi",
         description: "Dogri (macrolanguage)",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,dok,tag{
         type: "language",
         subtag: "dok",
         description: "Dondo",
         added: "2009-07-29"}).
iana(language,dol,tag{
         type: "language",
         subtag: "dol",
         description: "Doso",
         added: "2009-07-29"}).
iana(language,don,tag{
         type: "language",
         subtag: "don",
         description: "Toura (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,doo,tag{
         type: "language",
         subtag: "doo",
         description: "Dongo",
         added: "2009-07-29"}).
iana(language,dop,tag{
         type: "language",
         subtag: "dop",
         description: "Lukpa",
         added: "2009-07-29"}).
iana(language,doq,tag{
         type: "language",
         subtag: "doq",
         description: "Dominican Sign Language",
         added: "2009-07-29"}).
iana(language,dor,tag{
         type: "language",
         subtag: "dor",
         description: "Dori'o",
         added: "2009-07-29"}).
iana(language,dos,tag{
         type: "language",
         subtag: "dos",
         description: "Dogosé",
         added: "2009-07-29"}).
iana(language,dot,tag{
         type: "language",
         subtag: "dot",
         description: "Dass",
         added: "2009-07-29"}).
iana(language,dov,tag{
         type: "language",
         subtag: "dov",
         description: "Dombe",
         added: "2009-07-29"}).
iana(language,dow,tag{
         type: "language",
         subtag: "dow",
         description: "Doyayo",
         added: "2009-07-29"}).
iana(language,dox,tag{
         type: "language",
         subtag: "dox",
         description: "Bussa",
         added: "2009-07-29"}).
iana(language,doy,tag{
         type: "language",
         subtag: "doy",
         description: "Dompo",
         added: "2009-07-29"}).
iana(language,doz,tag{
         type: "language",
         subtag: "doz",
         description: "Dorze",
         added: "2009-07-29"}).
iana(language,dpp,tag{
         type: "language",
         subtag: "dpp",
         description: "Papar",
         added: "2009-07-29"}).
iana(language,dra,tag{
         type: "language",
         subtag: "dra",
         description: "Dravidian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,drb,tag{
         type: "language",
         subtag: "drb",
         description: "Dair",
         added: "2009-07-29"}).
iana(language,drc,tag{
         type: "language",
         subtag: "drc",
         description: "Minderico",
         added: "2011-08-16"}).
iana(language,drd,tag{
         type: "language",
         subtag: "drd",
         description: "Darmiya",
         added: "2009-07-29"}).
iana(language,dre,tag{
         type: "language",
         subtag: "dre",
         description: "Dolpo",
         added: "2009-07-29"}).
iana(language,drg,tag{
         type: "language",
         subtag: "drg",
         description: "Rungus",
         added: "2009-07-29"}).
iana(language,drh,tag{
         type: "language",
         subtag: "drh",
         description: "Darkhat",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "khk"}).
iana(language,dri,tag{
         type: "language",
         subtag: "dri",
         description: "C'Lela",
         added: "2009-07-29"}).
iana(language,drl,tag{
         type: "language",
         subtag: "drl",
         description: "Paakantyi",
         added: "2009-07-29"}).
iana(language,drn,tag{
         type: "language",
         subtag: "drn",
         description: "West Damar",
         added: "2009-07-29"}).
iana(language,dro,tag{
         type: "language",
         subtag: "dro",
         description: "Daro-Matu Melanau",
         added: "2009-07-29"}).
iana(language,drq,tag{
         type: "language",
         subtag: "drq",
         description: "Dura",
         added: "2009-07-29"}).
iana(language,drr,tag{
         type: "language",
         subtag: "drr",
         description: "Dororo",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "kzk"}).
iana(language,drs,tag{
         type: "language",
         subtag: "drs",
         description: "Gedeo",
         added: "2009-07-29"}).
iana(language,drt,tag{
         type: "language",
         subtag: "drt",
         description: "Drents",
         added: "2009-07-29"}).
iana(language,dru,tag{
         type: "language",
         subtag: "dru",
         description: "Rukai",
         added: "2009-07-29"}).
iana(language,drw,tag{
         type: "language",
         subtag: "drw",
         description: "Darwazi",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "prs"}).
iana(language,dry,tag{
         type: "language",
         subtag: "dry",
         description: "Darai",
         added: "2009-07-29"}).
iana(language,dsb,tag{
         type: "language",
         subtag: "dsb",
         description: "Lower Sorbian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,dse,tag{
         type: "language",
         subtag: "dse",
         description: "Dutch Sign Language",
         added: "2009-07-29"}).
iana(language,dsh,tag{
         type: "language",
         subtag: "dsh",
         description: "Daasanach",
         added: "2009-07-29"}).
iana(language,dsi,tag{
         type: "language",
         subtag: "dsi",
         description: "Disa",
         added: "2009-07-29"}).
iana(language,dsl,tag{
         type: "language",
         subtag: "dsl",
         description: "Danish Sign Language",
         added: "2009-07-29"}).
iana(language,dsn,tag{
         type: "language",
         subtag: "dsn",
         description: "Dusner",
         added: "2009-07-29"}).
iana(language,dso,tag{
         type: "language",
         subtag: "dso",
         description: "Desiya",
         added: "2009-07-29"}).
iana(language,dsq,tag{
         type: "language",
         subtag: "dsq",
         description: "Tadaksahak",
         added: "2009-07-29"}).
iana(language,dsz,tag{
         type: "language",
         subtag: "dsz",
         description: "Mardin Sign Language",
         added: "2022-02-25"}).
iana(language,dta,tag{
         type: "language",
         subtag: "dta",
         description: "Daur",
         added: "2009-07-29"}).
iana(language,dtb,tag{
         type: "language",
         subtag: "dtb",
         description: "Labuk-Kinabatangan Kadazan",
         added: "2009-07-29"}).
iana(language,dtd,tag{
         type: "language",
         subtag: "dtd",
         description: "Ditidaht",
         added: "2011-08-16"}).
iana(language,dth,tag{
         type: "language",
         subtag: "dth",
         description: "Adithinngithigh",
         added: "2012-08-12"}).
iana(language,dti,tag{
         type: "language",
         subtag: "dti",
         description: "Ana Tinga Dogon",
         added: "2009-07-29"}).
iana(language,dtk,tag{
         type: "language",
         subtag: "dtk",
         description: "Tene Kan Dogon",
         added: "2009-07-29"}).
iana(language,dtm,tag{
         type: "language",
         subtag: "dtm",
         description: "Tomo Kan Dogon",
         added: "2009-07-29"}).
iana(language,dtn,tag{
         type: "language",
         subtag: "dtn",
         description: "Daatsʼíin",
         added: "2016-05-30"}).
iana(language,dto,tag{
         type: "language",
         subtag: "dto",
         description: "Tommo So Dogon",
         added: "2012-08-12"}).
iana(language,dtp,tag{
         type: "language",
         subtag: "dtp",
         description: ["Kadazan Dusun","Central Dusun"],
         added: "2009-07-29"}).
iana(language,dtr,tag{
         type: "language",
         subtag: "dtr",
         description: "Lotud",
         added: "2009-07-29"}).
iana(language,dts,tag{
         type: "language",
         subtag: "dts",
         description: "Toro So Dogon",
         added: "2009-07-29"}).
iana(language,dtt,tag{
         type: "language",
         subtag: "dtt",
         description: "Toro Tegu Dogon",
         added: "2009-07-29"}).
iana(language,dtu,tag{
         type: "language",
         subtag: "dtu",
         description: "Tebul Ure Dogon",
         added: "2009-07-29"}).
iana(language,dty,tag{
         type: "language",
         subtag: "dty",
         description: "Dotyali",
         added: "2012-08-12",
         macrolanguage: "ne"}).
iana(language,dua,tag{
         type: "language",
         subtag: "dua",
         description: "Duala",
         added: "2005-10-16"}).
iana(language,dub,tag{
         type: "language",
         subtag: "dub",
         description: "Dubli",
         added: "2009-07-29"}).
iana(language,duc,tag{
         type: "language",
         subtag: "duc",
         description: "Duna",
         added: "2009-07-29"}).
iana(language,dud,tag{
         type: "language",
         subtag: "dud",
         description: "Hun-Saare",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         comments: "see uth, uss"}).
iana(language,due,tag{
         type: "language",
         subtag: "due",
         description: "Umiray Dumaget Agta",
         added: "2009-07-29"}).
iana(language,duf,tag{
         type: "language",
         subtag: "duf",
         description: ["Dumbea","Drubea"],
         added: "2009-07-29"}).
iana(language,dug,tag{
         type: "language",
         subtag: "dug",
         description: ["Duruma","Chiduruma"],
         added: "2009-07-29"}).
iana(language,duh,tag{
         type: "language",
         subtag: "duh",
         description: "Dungra Bhil",
         added: "2009-07-29"}).
iana(language,dui,tag{
         type: "language",
         subtag: "dui",
         description: "Dumun",
         added: "2009-07-29"}).
iana(language,duj,tag{
         type: "language",
         subtag: "duj",
         description: "Dhuwal",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         comments: "see dwu, dwy"}).
iana(language,duk,tag{
         type: "language",
         subtag: "duk",
         description: "Uyajitaya",
         added: "2009-07-29"}).
iana(language,dul,tag{
         type: "language",
         subtag: "dul",
         description: "Alabat Island Agta",
         added: "2009-07-29"}).
iana(language,dum,tag{
         type: "language",
         subtag: "dum",
         description: "Middle Dutch (ca. 1050-1350)",
         added: "2005-10-16"}).
iana(language,dun,tag{
         type: "language",
         subtag: "dun",
         description: "Dusun Deyah",
         added: "2009-07-29"}).
iana(language,duo,tag{
         type: "language",
         subtag: "duo",
         description: "Dupaninan Agta",
         added: "2009-07-29"}).
iana(language,dup,tag{
         type: "language",
         subtag: "dup",
         description: "Duano",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,duq,tag{
         type: "language",
         subtag: "duq",
         description: "Dusun Malang",
         added: "2009-07-29"}).
iana(language,dur,tag{
         type: "language",
         subtag: "dur",
         description: "Dii",
         added: "2009-07-29"}).
iana(language,dus,tag{
         type: "language",
         subtag: "dus",
         description: "Dumi",
         added: "2009-07-29"}).
iana(language,duu,tag{
         type: "language",
         subtag: "duu",
         description: "Drung",
         added: "2009-07-29"}).
iana(language,duv,tag{
         type: "language",
         subtag: "duv",
         description: "Duvle",
         added: "2009-07-29"}).
iana(language,duw,tag{
         type: "language",
         subtag: "duw",
         description: "Dusun Witu",
         added: "2009-07-29"}).
iana(language,dux,tag{
         type: "language",
         subtag: "dux",
         description: "Duungooma",
         added: "2009-07-29"}).
iana(language,duy,tag{
         type: "language",
         subtag: "duy",
         description: "Dicamay Agta",
         added: "2009-07-29"}).
iana(language,duz,tag{
         type: "language",
         subtag: "duz",
         description: "Duli-Gey",
         added: "2009-07-29"}).
iana(language,dva,tag{
         type: "language",
         subtag: "dva",
         description: "Duau",
         added: "2009-07-29"}).
iana(language,dwa,tag{
         type: "language",
         subtag: "dwa",
         description: "Diri",
         added: "2009-07-29"}).
iana(language,dwk,tag{
         type: "language",
         subtag: "dwk",
         description: "Dawik Kui",
         added: "2020-03-28"}).
iana(language,dwl,tag{
         type: "language",
         subtag: "dwl",
         description: "Walo Kumbe Dogon",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see dbt, dbw"}).
iana(language,dwr,tag{
         type: "language",
         subtag: "dwr",
         description: "Dawro",
         added: "2009-07-29"}).
iana(language,dws,tag{
         type: "language",
         subtag: "dws",
         description: "Dutton World Speedwords",
         added: "2009-07-29"}).
iana(language,dwu,tag{
         type: "language",
         subtag: "dwu",
         description: "Dhuwal",
         added: "2016-05-30"}).
iana(language,dww,tag{
         type: "language",
         subtag: "dww",
         description: "Dawawa",
         added: "2009-07-29"}).
iana(language,dwy,tag{
         type: "language",
         subtag: "dwy",
         description: "Dhuwaya",
         added: "2016-05-30"}).
iana(language,dwz,tag{
         type: "language",
         subtag: "dwz",
         description: "Dewas Rai",
         added: "2018-10-28"}).
iana(language,dya,tag{
         type: "language",
         subtag: "dya",
         description: "Dyan",
         added: "2009-07-29"}).
iana(language,dyb,tag{
         type: "language",
         subtag: "dyb",
         description: "Dyaberdyaber",
         added: "2009-07-29"}).
iana(language,dyd,tag{
         type: "language",
         subtag: "dyd",
         description: "Dyugun",
         added: "2009-07-29"}).
iana(language,dyg,tag{
         type: "language",
         subtag: "dyg",
         description: "Villa Viciosa Agta",
         added: "2009-07-29"}).
iana(language,dyi,tag{
         type: "language",
         subtag: "dyi",
         description: "Djimini Senoufo",
         added: "2009-07-29"}).
iana(language,dym,tag{
         type: "language",
         subtag: "dym",
         description: "Yanda Dom Dogon",
         added: "2009-07-29"}).
iana(language,dyn,tag{
         type: "language",
         subtag: "dyn",
         description: ["Dyangadi","Dhanggatti"],
         added: "2009-07-29"}).
iana(language,dyo,tag{
         type: "language",
         subtag: "dyo",
         description: "Jola-Fonyi",
         added: "2009-07-29"}).
iana(language,dyu,tag{
         type: "language",
         subtag: "dyu",
         description: "Dyula",
         added: "2005-10-16"}).
iana(language,dyy,tag{
         type: "language",
         subtag: "dyy",
         description: ["Djabugay","Dyaabugay"],
         added: "2009-07-29"}).
iana(language,dza,tag{
         type: "language",
         subtag: "dza",
         description: "Tunzu",
         added: "2009-07-29"}).
iana(language,dzd,tag{
         type: "language",
         subtag: "dzd",
         description: "Daza",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,dze,tag{
         type: "language",
         subtag: "dze",
         description: "Djiwarli",
         added: "2013-09-10"}).
iana(language,dzg,tag{
         type: "language",
         subtag: "dzg",
         description: "Dazaga",
         added: "2009-07-29"}).
iana(language,dzl,tag{
         type: "language",
         subtag: "dzl",
         description: "Dzalakha",
         added: "2009-07-29"}).
iana(language,dzn,tag{
         type: "language",
         subtag: "dzn",
         description: "Dzando",
         added: "2009-07-29"}).
iana(language,eaa,tag{
         type: "language",
         subtag: "eaa",
         description: "Karenggapa",
         added: "2013-09-10"}).
iana(language,ebc,tag{
         type: "language",
         subtag: "ebc",
         description: "Beginci",
         added: "2020-03-28"}).
iana(language,ebg,tag{
         type: "language",
         subtag: "ebg",
         description: "Ebughu",
         added: "2009-07-29"}).
iana(language,ebk,tag{
         type: "language",
         subtag: "ebk",
         description: "Eastern Bontok",
         added: "2010-03-11",
         macrolanguage: "bnc"}).
iana(language,ebo,tag{
         type: "language",
         subtag: "ebo",
         description: "Teke-Ebo",
         added: "2009-07-29"}).
iana(language,ebr,tag{
         type: "language",
         subtag: "ebr",
         description: "Ebrié",
         added: "2009-07-29"}).
iana(language,ebu,tag{
         type: "language",
         subtag: "ebu",
         description: ["Embu","Kiembu"],
         added: "2009-07-29"}).
iana(language,ecr,tag{
         type: "language",
         subtag: "ecr",
         description: "Eteocretan",
         added: "2009-07-29"}).
iana(language,ecs,tag{
         type: "language",
         subtag: "ecs",
         description: "Ecuadorian Sign Language",
         added: "2009-07-29"}).
iana(language,ecy,tag{
         type: "language",
         subtag: "ecy",
         description: "Eteocypriot",
         added: "2009-07-29"}).
iana(language,eee,tag{
         type: "language",
         subtag: "eee",
         description: "E",
         added: "2009-07-29"}).
iana(language,efa,tag{
         type: "language",
         subtag: "efa",
         description: "Efai",
         added: "2009-07-29"}).
iana(language,efe,tag{
         type: "language",
         subtag: "efe",
         description: "Efe",
         added: "2009-07-29"}).
iana(language,efi,tag{
         type: "language",
         subtag: "efi",
         description: "Efik",
         added: "2005-10-16"}).
iana(language,ega,tag{
         type: "language",
         subtag: "ega",
         description: "Ega",
         added: "2009-07-29"}).
iana(language,egl,tag{
         type: "language",
         subtag: "egl",
         description: "Emilian",
         added: "2009-07-29"}).
iana(language,egm,tag{
         type: "language",
         subtag: "egm",
         description: "Benamanga",
         added: "2022-02-25"}).
iana(language,ego,tag{
         type: "language",
         subtag: "ego",
         description: "Eggon",
         added: "2009-07-29"}).
iana(language,egx,tag{
         type: "language",
         subtag: "egx",
         description: "Egyptian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,egy,tag{
         type: "language",
         subtag: "egy",
         description: "Egyptian (Ancient)",
         added: "2005-10-16"}).
iana(language,ehs,tag{
         type: "language",
         subtag: "ehs",
         description: "Miyakubo Sign Language",
         added: "2021-02-20"}).
iana(language,ehu,tag{
         type: "language",
         subtag: "ehu",
         description: "Ehueun",
         added: "2009-07-29"}).
iana(language,eip,tag{
         type: "language",
         subtag: "eip",
         description: "Eipomek",
         added: "2009-07-29"}).
iana(language,eit,tag{
         type: "language",
         subtag: "eit",
         description: "Eitiep",
         added: "2009-07-29"}).
iana(language,eiv,tag{
         type: "language",
         subtag: "eiv",
         description: "Askopan",
         added: "2009-07-29"}).
iana(language,eja,tag{
         type: "language",
         subtag: "eja",
         description: "Ejamat",
         added: "2009-07-29"}).
iana(language,eka,tag{
         type: "language",
         subtag: "eka",
         description: "Ekajuk",
         added: "2005-10-16"}).
iana(language,ekc,tag{
         type: "language",
         subtag: "ekc",
         description: "Eastern Karnic",
         added: "2013-09-10",
         deprecated: "2020-03-28"}).
iana(language,eke,tag{
         type: "language",
         subtag: "eke",
         description: "Ekit",
         added: "2009-07-29"}).
iana(language,ekg,tag{
         type: "language",
         subtag: "ekg",
         description: "Ekari",
         added: "2009-07-29"}).
iana(language,eki,tag{
         type: "language",
         subtag: "eki",
         description: "Eki",
         added: "2009-07-29"}).
iana(language,ekk,tag{
         type: "language",
         subtag: "ekk",
         description: "Standard Estonian",
         added: "2009-07-29",
         macrolanguage: "et"}).
iana(language,ekl,tag{
         type: "language",
         subtag: "ekl",
         description: ["Kol (Bangladesh)","Kol"],
         added: "2009-07-29"}).
iana(language,ekm,tag{
         type: "language",
         subtag: "ekm",
         description: "Elip",
         added: "2009-07-29"}).
iana(language,eko,tag{
         type: "language",
         subtag: "eko",
         description: "Koti",
         added: "2009-07-29"}).
iana(language,ekp,tag{
         type: "language",
         subtag: "ekp",
         description: "Ekpeye",
         added: "2009-07-29"}).
iana(language,ekr,tag{
         type: "language",
         subtag: "ekr",
         description: "Yace",
         added: "2009-07-29"}).
iana(language,eky,tag{
         type: "language",
         subtag: "eky",
         description: "Eastern Kayah",
         added: "2009-07-29"}).
iana(language,ele,tag{
         type: "language",
         subtag: "ele",
         description: "Elepi",
         added: "2009-07-29"}).
iana(language,elh,tag{
         type: "language",
         subtag: "elh",
         description: "El Hugeirat",
         added: "2009-07-29"}).
iana(language,eli,tag{
         type: "language",
         subtag: "eli",
         description: "Nding",
         added: "2009-07-29"}).
iana(language,elk,tag{
         type: "language",
         subtag: "elk",
         description: "Elkei",
         added: "2009-07-29"}).
iana(language,elm,tag{
         type: "language",
         subtag: "elm",
         description: "Eleme",
         added: "2009-07-29"}).
iana(language,elo,tag{
         type: "language",
         subtag: "elo",
         description: "El Molo",
         added: "2009-07-29"}).
iana(language,elp,tag{
         type: "language",
         subtag: "elp",
         description: "Elpaputih",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see amq, plh"}).
iana(language,elu,tag{
         type: "language",
         subtag: "elu",
         description: "Elu",
         added: "2009-07-29"}).
iana(language,elx,tag{
         type: "language",
         subtag: "elx",
         description: "Elamite",
         added: "2005-10-16"}).
iana(language,ema,tag{
         type: "language",
         subtag: "ema",
         description: "Emai-Iuleha-Ora",
         added: "2009-07-29"}).
iana(language,emb,tag{
         type: "language",
         subtag: "emb",
         description: "Embaloh",
         added: "2009-07-29"}).
iana(language,eme,tag{
         type: "language",
         subtag: "eme",
         description: "Emerillon",
         added: "2009-07-29"}).
iana(language,emg,tag{
         type: "language",
         subtag: "emg",
         description: "Eastern Meohang",
         added: "2009-07-29"}).
iana(language,emi,tag{
         type: "language",
         subtag: "emi",
         description: "Mussau-Emira",
         added: "2009-07-29"}).
iana(language,emk,tag{
         type: "language",
         subtag: "emk",
         description: "Eastern Maninkakan",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,emm,tag{
         type: "language",
         subtag: "emm",
         description: "Mamulique",
         added: "2009-07-29"}).
iana(language,emn,tag{
         type: "language",
         subtag: "emn",
         description: "Eman",
         added: "2009-07-29"}).
iana(language,emo,tag{
         type: "language",
         subtag: "emo",
         description: "Emok",
         added: "2009-07-29",
         deprecated: "2014-02-28"}).
iana(language,emp,tag{
         type: "language",
         subtag: "emp",
         description: "Northern Emberá",
         added: "2009-07-29"}).
iana(language,emq,tag{
         type: "language",
         subtag: "emq",
         description: "Eastern Minyag",
         added: "2021-02-20"}).
iana(language,ems,tag{
         type: "language",
         subtag: "ems",
         description: "Pacific Gulf Yupik",
         added: "2009-07-29"}).
iana(language,emu,tag{
         type: "language",
         subtag: "emu",
         description: "Eastern Muria",
         added: "2009-07-29"}).
iana(language,emw,tag{
         type: "language",
         subtag: "emw",
         description: "Emplawas",
         added: "2009-07-29"}).
iana(language,emx,tag{
         type: "language",
         subtag: "emx",
         description: "Erromintxela",
         added: "2010-03-11"}).
iana(language,emy,tag{
         type: "language",
         subtag: "emy",
         description: "Epigraphic Mayan",
         added: "2009-07-29"}).
iana(language,emz,tag{
         type: "language",
         subtag: "emz",
         description: "Mbessa",
         added: "2021-02-20"}).
iana(language,ena,tag{
         type: "language",
         subtag: "ena",
         description: "Apali",
         added: "2009-07-29"}).
iana(language,enb,tag{
         type: "language",
         subtag: "enb",
         description: "Markweeta",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,enc,tag{
         type: "language",
         subtag: "enc",
         description: "En",
         added: "2009-07-29"}).
iana(language,end,tag{
         type: "language",
         subtag: "end",
         description: "Ende",
         added: "2009-07-29"}).
iana(language,enf,tag{
         type: "language",
         subtag: "enf",
         description: "Forest Enets",
         added: "2009-07-29"}).
iana(language,enh,tag{
         type: "language",
         subtag: "enh",
         description: "Tundra Enets",
         added: "2009-07-29"}).
iana(language,enl,tag{
         type: "language",
         subtag: "enl",
         description: "Enlhet",
         added: "2014-02-28"}).
iana(language,enm,tag{
         type: "language",
         subtag: "enm",
         description: "Middle English (1100-1500)",
         added: "2005-10-16"}).
iana(language,enn,tag{
         type: "language",
         subtag: "enn",
         description: "Engenni",
         added: "2009-07-29"}).
iana(language,eno,tag{
         type: "language",
         subtag: "eno",
         description: "Enggano",
         added: "2009-07-29"}).
iana(language,enq,tag{
         type: "language",
         subtag: "enq",
         description: "Enga",
         added: "2009-07-29"}).
iana(language,enr,tag{
         type: "language",
         subtag: "enr",
         description: ["Emumu","Emem"],
         added: "2009-07-29"}).
iana(language,enu,tag{
         type: "language",
         subtag: "enu",
         description: "Enu",
         added: "2009-07-29"}).
iana(language,env,tag{
         type: "language",
         subtag: "env",
         description: "Enwan (Edo State)",
         added: "2009-07-29"}).
iana(language,enw,tag{
         type: "language",
         subtag: "enw",
         description: "Enwan (Akwa Ibom State)",
         added: "2009-07-29"}).
iana(language,enx,tag{
         type: "language",
         subtag: "enx",
         description: "Enxet",
         added: "2014-02-28"}).
iana(language,eot,tag{
         type: "language",
         subtag: "eot",
         description: "Beti (Côte d'Ivoire)",
         added: "2009-07-29"}).
iana(language,epi,tag{
         type: "language",
         subtag: "epi",
         description: "Epie",
         added: "2009-07-29"}).
iana(language,era,tag{
         type: "language",
         subtag: "era",
         description: "Eravallan",
         added: "2009-07-29"}).
iana(language,erg,tag{
         type: "language",
         subtag: "erg",
         description: "Sie",
         added: "2009-07-29"}).
iana(language,erh,tag{
         type: "language",
         subtag: "erh",
         description: "Eruwa",
         added: "2009-07-29"}).
iana(language,eri,tag{
         type: "language",
         subtag: "eri",
         description: "Ogea",
         added: "2009-07-29"}).
iana(language,erk,tag{
         type: "language",
         subtag: "erk",
         description: "South Efate",
         added: "2009-07-29"}).
iana(language,ero,tag{
         type: "language",
         subtag: "ero",
         description: "Horpa",
         added: "2009-07-29"}).
iana(language,err,tag{
         type: "language",
         subtag: "err",
         description: "Erre",
         added: "2009-07-29"}).
iana(language,ers,tag{
         type: "language",
         subtag: "ers",
         description: "Ersu",
         added: "2009-07-29"}).
iana(language,ert,tag{
         type: "language",
         subtag: "ert",
         description: "Eritai",
         added: "2009-07-29"}).
iana(language,erw,tag{
         type: "language",
         subtag: "erw",
         description: "Erokwanas",
         added: "2009-07-29"}).
iana(language,ese,tag{
         type: "language",
         subtag: "ese",
         description: "Ese Ejja",
         added: "2009-07-29"}).
iana(language,esg,tag{
         type: "language",
         subtag: "esg",
         description: "Aheri Gondi",
         added: "2016-05-30",
         macrolanguage: "gon"}).
iana(language,esh,tag{
         type: "language",
         subtag: "esh",
         description: "Eshtehardi",
         added: "2009-07-29"}).
iana(language,esi,tag{
         type: "language",
         subtag: "esi",
         description: "North Alaskan Inupiatun",
         added: "2009-07-29",
         macrolanguage: "ik"}).
iana(language,esk,tag{
         type: "language",
         subtag: "esk",
         description: "Northwest Alaska Inupiatun",
         added: "2009-07-29",
         macrolanguage: "ik"}).
iana(language,esl,tag{
         type: "language",
         subtag: "esl",
         description: "Egypt Sign Language",
         added: "2009-07-29"}).
iana(language,esm,tag{
         type: "language",
         subtag: "esm",
         description: "Esuma",
         added: "2009-07-29"}).
iana(language,esn,tag{
         type: "language",
         subtag: "esn",
         description: "Salvadoran Sign Language",
         added: "2009-07-29"}).
iana(language,eso,tag{
         type: "language",
         subtag: "eso",
         description: "Estonian Sign Language",
         added: "2009-07-29"}).
iana(language,esq,tag{
         type: "language",
         subtag: "esq",
         description: "Esselen",
         added: "2009-07-29"}).
iana(language,ess,tag{
         type: "language",
         subtag: "ess",
         description: "Central Siberian Yupik",
         added: "2009-07-29"}).
iana(language,esu,tag{
         type: "language",
         subtag: "esu",
         description: "Central Yupik",
         added: "2009-07-29"}).
iana(language,esx,tag{
         type: "language",
         subtag: "esx",
         description: "Eskimo-Aleut languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,esy,tag{
         type: "language",
         subtag: "esy",
         description: "Eskayan",
         added: "2014-04-06"}).
iana(language,etb,tag{
         type: "language",
         subtag: "etb",
         description: "Etebi",
         added: "2009-07-29"}).
iana(language,etc,tag{
         type: "language",
         subtag: "etc",
         description: "Etchemin",
         added: "2009-07-29"}).
iana(language,eth,tag{
         type: "language",
         subtag: "eth",
         description: "Ethiopian Sign Language",
         added: "2009-07-29"}).
iana(language,etn,tag{
         type: "language",
         subtag: "etn",
         description: "Eton (Vanuatu)",
         added: "2009-07-29"}).
iana(language,eto,tag{
         type: "language",
         subtag: "eto",
         description: "Eton (Cameroon)",
         added: "2009-07-29"}).
iana(language,etr,tag{
         type: "language",
         subtag: "etr",
         description: "Edolo",
         added: "2009-07-29"}).
iana(language,ets,tag{
         type: "language",
         subtag: "ets",
         description: "Yekhee",
         added: "2009-07-29"}).
iana(language,ett,tag{
         type: "language",
         subtag: "ett",
         description: "Etruscan",
         added: "2009-07-29"}).
iana(language,etu,tag{
         type: "language",
         subtag: "etu",
         description: "Ejagham",
         added: "2009-07-29"}).
iana(language,etx,tag{
         type: "language",
         subtag: "etx",
         description: "Eten",
         added: "2009-07-29"}).
iana(language,etz,tag{
         type: "language",
         subtag: "etz",
         description: "Semimi",
         added: "2009-07-29"}).
iana(language,euq,tag{
         type: "language",
         subtag: "euq",
         description: "Basque (family)",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,eve,tag{
         type: "language",
         subtag: "eve",
         description: "Even",
         added: "2009-07-29"}).
iana(language,evh,tag{
         type: "language",
         subtag: "evh",
         description: "Uvbie",
         added: "2009-07-29"}).
iana(language,evn,tag{
         type: "language",
         subtag: "evn",
         description: "Evenki",
         added: "2009-07-29"}).
iana(language,ewo,tag{
         type: "language",
         subtag: "ewo",
         description: "Ewondo",
         added: "2005-10-16"}).
iana(language,ext,tag{
         type: "language",
         subtag: "ext",
         description: "Extremaduran",
         added: "2009-07-29"}).
iana(language,eya,tag{
         type: "language",
         subtag: "eya",
         description: "Eyak",
         added: "2009-07-29"}).
iana(language,eyo,tag{
         type: "language",
         subtag: "eyo",
         description: "Keiyo",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,eza,tag{
         type: "language",
         subtag: "eza",
         description: "Ezaa",
         added: "2013-09-10"}).
iana(language,eze,tag{
         type: "language",
         subtag: "eze",
         description: "Uzekwe",
         added: "2009-07-29"}).
iana(language,faa,tag{
         type: "language",
         subtag: "faa",
         description: "Fasu",
         added: "2009-07-29"}).
iana(language,fab,tag{
         type: "language",
         subtag: "fab",
         description: "Fa d'Ambu",
         added: "2009-07-29"}).
iana(language,fad,tag{
         type: "language",
         subtag: "fad",
         description: "Wagi",
         added: "2009-07-29"}).
iana(language,faf,tag{
         type: "language",
         subtag: "faf",
         description: "Fagani",
         added: "2009-07-29"}).
iana(language,fag,tag{
         type: "language",
         subtag: "fag",
         description: "Finongan",
         added: "2009-07-29"}).
iana(language,fah,tag{
         type: "language",
         subtag: "fah",
         description: "Baissa Fali",
         added: "2009-07-29"}).
iana(language,fai,tag{
         type: "language",
         subtag: "fai",
         description: "Faiwol",
         added: "2009-07-29"}).
iana(language,faj,tag{
         type: "language",
         subtag: "faj",
         description: "Faita",
         added: "2009-07-29"}).
iana(language,fak,tag{
         type: "language",
         subtag: "fak",
         description: "Fang (Cameroon)",
         added: "2009-07-29"}).
iana(language,fal,tag{
         type: "language",
         subtag: "fal",
         description: "South Fali",
         added: "2009-07-29"}).
iana(language,fam,tag{
         type: "language",
         subtag: "fam",
         description: "Fam",
         added: "2009-07-29"}).
iana(language,fan,tag{
         type: "language",
         subtag: "fan",
         description: "Fang (Equatorial Guinea)",
         added: "2005-10-16"}).
iana(language,fap,tag{
         type: "language",
         subtag: "fap",
         description: "Paloor",
         added: "2009-07-29"}).
iana(language,far,tag{
         type: "language",
         subtag: "far",
         description: "Fataleka",
         added: "2009-07-29"}).
iana(language,fat,tag{
         type: "language",
         subtag: "fat",
         description: "Fanti",
         added: "2005-10-16",
         macrolanguage: "ak"}).
iana(language,fau,tag{
         type: "language",
         subtag: "fau",
         description: "Fayu",
         added: "2009-07-29"}).
iana(language,fax,tag{
         type: "language",
         subtag: "fax",
         description: "Fala",
         added: "2009-07-29"}).
iana(language,fay,tag{
         type: "language",
         subtag: "fay",
         description: "Southwestern Fars",
         added: "2009-07-29"}).
iana(language,faz,tag{
         type: "language",
         subtag: "faz",
         description: "Northwestern Fars",
         added: "2009-07-29"}).
iana(language,fbl,tag{
         type: "language",
         subtag: "fbl",
         description: "West Albay Bikol",
         added: "2010-03-11",
         macrolanguage: "bik"}).
iana(language,fcs,tag{
         type: "language",
         subtag: "fcs",
         description: "Quebec Sign Language",
         added: "2009-07-29"}).
iana(language,fer,tag{
         type: "language",
         subtag: "fer",
         description: "Feroge",
         added: "2009-07-29"}).
iana(language,ffi,tag{
         type: "language",
         subtag: "ffi",
         description: "Foia Foia",
         added: "2009-07-29"}).
iana(language,ffm,tag{
         type: "language",
         subtag: "ffm",
         description: "Maasina Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fgr,tag{
         type: "language",
         subtag: "fgr",
         description: "Fongoro",
         added: "2009-07-29"}).
iana(language,fia,tag{
         type: "language",
         subtag: "fia",
         description: "Nobiin",
         added: "2009-07-29"}).
iana(language,fie,tag{
         type: "language",
         subtag: "fie",
         description: "Fyer",
         added: "2009-07-29"}).
iana(language,fif,tag{
         type: "language",
         subtag: "fif",
         description: "Faifi",
         added: "2020-06-08"}).
iana(language,fil,tag{
         type: "language",
         subtag: "fil",
         description: ["Filipino","Pilipino"],
         added: "2005-10-16"}).
iana(language,fip,tag{
         type: "language",
         subtag: "fip",
         description: "Fipa",
         added: "2009-07-29"}).
iana(language,fir,tag{
         type: "language",
         subtag: "fir",
         description: "Firan",
         added: "2009-07-29"}).
iana(language,fit,tag{
         type: "language",
         subtag: "fit",
         description: ["Tornedalen Finnish","Meänkieli"],
         added: "2009-07-29"}).
iana(language,fiu,tag{
         type: "language",
         subtag: "fiu",
         description: "Finno-Ugrian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,fiw,tag{
         type: "language",
         subtag: "fiw",
         description: "Fiwaga",
         added: "2009-07-29"}).
iana(language,fkk,tag{
         type: "language",
         subtag: "fkk",
         description: "Kirya-Konzəl",
         added: "2012-08-12"}).
iana(language,fkv,tag{
         type: "language",
         subtag: "fkv",
         description: "Kven Finnish",
         added: "2009-07-29"}).
iana(language,fla,tag{
         type: "language",
         subtag: "fla",
         description: "Kalispel-Pend d'Oreille",
         added: "2009-07-29"}).
iana(language,flh,tag{
         type: "language",
         subtag: "flh",
         description: "Foau",
         added: "2009-07-29"}).
iana(language,fli,tag{
         type: "language",
         subtag: "fli",
         description: "Fali",
         added: "2009-07-29"}).
iana(language,fll,tag{
         type: "language",
         subtag: "fll",
         description: "North Fali",
         added: "2009-07-29"}).
iana(language,fln,tag{
         type: "language",
         subtag: "fln",
         description: "Flinders Island",
         added: "2009-07-29"}).
iana(language,flr,tag{
         type: "language",
         subtag: "flr",
         description: "Fuliiru",
         added: "2009-07-29"}).
iana(language,fly,tag{
         type: "language",
         subtag: "fly",
         description: ["Flaaitaal","Tsotsitaal"],
         added: "2009-07-29"}).
iana(language,fmp,tag{
         type: "language",
         subtag: "fmp",
         description: "Fe'fe'",
         added: "2009-07-29"}).
iana(language,fmu,tag{
         type: "language",
         subtag: "fmu",
         description: "Far Western Muria",
         added: "2009-07-29"}).
iana(language,fnb,tag{
         type: "language",
         subtag: "fnb",
         description: "Fanbak",
         added: "2016-05-30"}).
iana(language,fng,tag{
         type: "language",
         subtag: "fng",
         description: "Fanagalo",
         added: "2009-07-29"}).
iana(language,fni,tag{
         type: "language",
         subtag: "fni",
         description: "Fania",
         added: "2009-07-29"}).
iana(language,fod,tag{
         type: "language",
         subtag: "fod",
         description: "Foodo",
         added: "2009-07-29"}).
iana(language,foi,tag{
         type: "language",
         subtag: "foi",
         description: "Foi",
         added: "2009-07-29"}).
iana(language,fom,tag{
         type: "language",
         subtag: "fom",
         description: "Foma",
         added: "2009-07-29"}).
iana(language,fon,tag{
         type: "language",
         subtag: "fon",
         description: "Fon",
         added: "2005-10-16"}).
iana(language,for,tag{
         type: "language",
         subtag: "for",
         description: "Fore",
         added: "2009-07-29"}).
iana(language,fos,tag{
         type: "language",
         subtag: "fos",
         description: "Siraya",
         added: "2009-07-29"}).
iana(language,fox,tag{
         type: "language",
         subtag: "fox",
         description: "Formosan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,fpe,tag{
         type: "language",
         subtag: "fpe",
         description: "Fernando Po Creole English",
         added: "2009-07-29"}).
iana(language,fqs,tag{
         type: "language",
         subtag: "fqs",
         description: "Fas",
         added: "2009-07-29"}).
iana(language,frc,tag{
         type: "language",
         subtag: "frc",
         description: "Cajun French",
         added: "2009-07-29"}).
iana(language,frd,tag{
         type: "language",
         subtag: "frd",
         description: "Fordata",
         added: "2009-07-29"}).
iana(language,frk,tag{
         type: "language",
         subtag: "frk",
         description: "Frankish",
         added: "2009-07-29"}).
iana(language,frm,tag{
         type: "language",
         subtag: "frm",
         description: "Middle French (ca. 1400-1600)",
         added: "2005-10-16"}).
iana(language,fro,tag{
         type: "language",
         subtag: "fro",
         description: "Old French (842-ca. 1400)",
         added: "2005-10-16"}).
iana(language,frp,tag{
         type: "language",
         subtag: "frp",
         description: ["Arpitan","Francoprovençal"],
         added: "2009-07-29"}).
iana(language,frq,tag{
         type: "language",
         subtag: "frq",
         description: "Forak",
         added: "2009-07-29"}).
iana(language,frr,tag{
         type: "language",
         subtag: "frr",
         description: "Northern Frisian",
         added: "2006-03-08",
         suppress_script: "Latn"}).
iana(language,frs,tag{
         type: "language",
         subtag: "frs",
         description: "Eastern Frisian",
         added: "2006-03-08",
         suppress_script: "Latn"}).
iana(language,frt,tag{
         type: "language",
         subtag: "frt",
         description: "Fortsenal",
         added: "2009-07-29"}).
iana(language,fse,tag{
         type: "language",
         subtag: "fse",
         description: "Finnish Sign Language",
         added: "2009-07-29"}).
iana(language,fsl,tag{
         type: "language",
         subtag: "fsl",
         description: "French Sign Language",
         added: "2009-07-29"}).
iana(language,fss,tag{
         type: "language",
         subtag: "fss",
         description: ["Finland-Swedish Sign Language","finlandssvenskt teckenspråk","suomenruotsalainen viittomakieli"],
         added: "2009-07-29"}).
iana(language,fub,tag{
         type: "language",
         subtag: "fub",
         description: "Adamawa Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fuc,tag{
         type: "language",
         subtag: "fuc",
         description: "Pulaar",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fud,tag{
         type: "language",
         subtag: "fud",
         description: "East Futuna",
         added: "2009-07-29"}).
iana(language,fue,tag{
         type: "language",
         subtag: "fue",
         description: "Borgu Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fuf,tag{
         type: "language",
         subtag: "fuf",
         description: "Pular",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fuh,tag{
         type: "language",
         subtag: "fuh",
         description: "Western Niger Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fui,tag{
         type: "language",
         subtag: "fui",
         description: "Bagirmi Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fuj,tag{
         type: "language",
         subtag: "fuj",
         description: "Ko",
         added: "2009-07-29"}).
iana(language,fum,tag{
         type: "language",
         subtag: "fum",
         description: "Fum",
         added: "2009-07-29"}).
iana(language,fun,tag{
         type: "language",
         subtag: "fun",
         description: "Fulniô",
         added: "2009-07-29"}).
iana(language,fuq,tag{
         type: "language",
         subtag: "fuq",
         description: "Central-Eastern Niger Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fur,tag{
         type: "language",
         subtag: "fur",
         description: "Friulian",
         added: "2005-10-16"}).
iana(language,fut,tag{
         type: "language",
         subtag: "fut",
         description: "Futuna-Aniwa",
         added: "2009-07-29"}).
iana(language,fuu,tag{
         type: "language",
         subtag: "fuu",
         description: "Furu",
         added: "2009-07-29"}).
iana(language,fuv,tag{
         type: "language",
         subtag: "fuv",
         description: "Nigerian Fulfulde",
         added: "2009-07-29",
         macrolanguage: "ff"}).
iana(language,fuy,tag{
         type: "language",
         subtag: "fuy",
         description: "Fuyug",
         added: "2009-07-29"}).
iana(language,fvr,tag{
         type: "language",
         subtag: "fvr",
         description: "Fur",
         added: "2009-07-29"}).
iana(language,fwa,tag{
         type: "language",
         subtag: "fwa",
         description: "Fwâi",
         added: "2009-07-29"}).
iana(language,fwe,tag{
         type: "language",
         subtag: "fwe",
         description: "Fwe",
         added: "2009-07-29"}).
iana(language,gaa,tag{
         type: "language",
         subtag: "gaa",
         description: "Ga",
         added: "2005-10-16"}).
iana(language,gab,tag{
         type: "language",
         subtag: "gab",
         description: "Gabri",
         added: "2009-07-29"}).
iana(language,gac,tag{
         type: "language",
         subtag: "gac",
         description: "Mixed Great Andamanese",
         added: "2010-03-11"}).
iana(language,gad,tag{
         type: "language",
         subtag: "gad",
         description: "Gaddang",
         added: "2009-07-29"}).
iana(language,gae,tag{
         type: "language",
         subtag: "gae",
         description: "Guarequena",
         added: "2009-07-29"}).
iana(language,gaf,tag{
         type: "language",
         subtag: "gaf",
         description: "Gende",
         added: "2009-07-29"}).
iana(language,gag,tag{
         type: "language",
         subtag: "gag",
         description: "Gagauz",
         added: "2009-07-29"}).
iana(language,gah,tag{
         type: "language",
         subtag: "gah",
         description: "Alekano",
         added: "2009-07-29"}).
iana(language,gai,tag{
         type: "language",
         subtag: "gai",
         description: "Borei",
         added: "2009-07-29"}).
iana(language,gaj,tag{
         type: "language",
         subtag: "gaj",
         description: "Gadsup",
         added: "2009-07-29"}).
iana(language,gak,tag{
         type: "language",
         subtag: "gak",
         description: "Gamkonora",
         added: "2009-07-29"}).
iana(language,gal,tag{
         type: "language",
         subtag: "gal",
         description: "Galolen",
         added: "2009-07-29"}).
iana(language,gam,tag{
         type: "language",
         subtag: "gam",
         description: "Kandawo",
         added: "2009-07-29"}).
iana(language,gan,tag{
         type: "language",
         subtag: "gan",
         description: "Gan Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,gao,tag{
         type: "language",
         subtag: "gao",
         description: "Gants",
         added: "2009-07-29"}).
iana(language,gap,tag{
         type: "language",
         subtag: "gap",
         description: "Gal",
         added: "2009-07-29"}).
iana(language,gaq,tag{
         type: "language",
         subtag: "gaq",
         description: "Gata'",
         added: "2009-07-29"}).
iana(language,gar,tag{
         type: "language",
         subtag: "gar",
         description: "Galeya",
         added: "2009-07-29"}).
iana(language,gas,tag{
         type: "language",
         subtag: "gas",
         description: "Adiwasi Garasia",
         added: "2009-07-29"}).
iana(language,gat,tag{
         type: "language",
         subtag: "gat",
         description: "Kenati",
         added: "2009-07-29"}).
iana(language,gau,tag{
         type: "language",
         subtag: "gau",
         description: "Mudhili Gadaba",
         added: "2009-07-29"}).
iana(language,gav,tag{
         type: "language",
         subtag: "gav",
         description: "Gabutamon",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "dev"}).
iana(language,gaw,tag{
         type: "language",
         subtag: "gaw",
         description: "Nobonob",
         added: "2009-07-29"}).
iana(language,gax,tag{
         type: "language",
         subtag: "gax",
         description: "Borana-Arsi-Guji Oromo",
         added: "2009-07-29",
         macrolanguage: "om"}).
iana(language,gay,tag{
         type: "language",
         subtag: "gay",
         description: "Gayo",
         added: "2005-10-16"}).
iana(language,gaz,tag{
         type: "language",
         subtag: "gaz",
         description: "West Central Oromo",
         added: "2009-07-29",
         macrolanguage: "om"}).
iana(language,gba,tag{
         type: "language",
         subtag: "gba",
         description: "Gbaya (Central African Republic)",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,gbb,tag{
         type: "language",
         subtag: "gbb",
         description: "Kaytetye",
         added: "2009-07-29"}).
iana(language,gbc,tag{
         type: "language",
         subtag: "gbc",
         description: "Garawa",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see wny, wrk"}).
iana(language,gbd,tag{
         type: "language",
         subtag: "gbd",
         description: "Karajarri",
         added: "2009-07-29"}).
iana(language,gbe,tag{
         type: "language",
         subtag: "gbe",
         description: "Niksek",
         added: "2009-07-29"}).
iana(language,gbf,tag{
         type: "language",
         subtag: "gbf",
         description: "Gaikundi",
         added: "2009-07-29"}).
iana(language,gbg,tag{
         type: "language",
         subtag: "gbg",
         description: "Gbanziri",
         added: "2009-07-29"}).
iana(language,gbh,tag{
         type: "language",
         subtag: "gbh",
         description: "Defi Gbe",
         added: "2009-07-29"}).
iana(language,gbi,tag{
         type: "language",
         subtag: "gbi",
         description: "Galela",
         added: "2009-07-29"}).
iana(language,gbj,tag{
         type: "language",
         subtag: "gbj",
         description: "Bodo Gadaba",
         added: "2009-07-29"}).
iana(language,gbk,tag{
         type: "language",
         subtag: "gbk",
         description: "Gaddi",
         added: "2009-07-29"}).
iana(language,gbl,tag{
         type: "language",
         subtag: "gbl",
         description: "Gamit",
         added: "2009-07-29"}).
iana(language,gbm,tag{
         type: "language",
         subtag: "gbm",
         description: "Garhwali",
         added: "2009-07-29"}).
iana(language,gbn,tag{
         type: "language",
         subtag: "gbn",
         description: "Mo'da",
         added: "2009-07-29"}).
iana(language,gbo,tag{
         type: "language",
         subtag: "gbo",
         description: "Northern Grebo",
         added: "2009-07-29",
         macrolanguage: "grb"}).
iana(language,gbp,tag{
         type: "language",
         subtag: "gbp",
         description: "Gbaya-Bossangoa",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,gbq,tag{
         type: "language",
         subtag: "gbq",
         description: "Gbaya-Bozoum",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,gbr,tag{
         type: "language",
         subtag: "gbr",
         description: "Gbagyi",
         added: "2009-07-29"}).
iana(language,gbs,tag{
         type: "language",
         subtag: "gbs",
         description: "Gbesi Gbe",
         added: "2009-07-29"}).
iana(language,gbu,tag{
         type: "language",
         subtag: "gbu",
         description: "Gagadu",
         added: "2009-07-29"}).
iana(language,gbv,tag{
         type: "language",
         subtag: "gbv",
         description: "Gbanu",
         added: "2009-07-29"}).
iana(language,gbw,tag{
         type: "language",
         subtag: "gbw",
         description: "Gabi-Gabi",
         added: "2013-09-10"}).
iana(language,gbx,tag{
         type: "language",
         subtag: "gbx",
         description: "Eastern Xwla Gbe",
         added: "2009-07-29"}).
iana(language,gby,tag{
         type: "language",
         subtag: "gby",
         description: "Gbari",
         added: "2009-07-29"}).
iana(language,gbz,tag{
         type: "language",
         subtag: "gbz",
         description: "Zoroastrian Dari",
         added: "2009-07-29"}).
iana(language,gcc,tag{
         type: "language",
         subtag: "gcc",
         description: "Mali",
         added: "2009-07-29"}).
iana(language,gcd,tag{
         type: "language",
         subtag: "gcd",
         description: "Ganggalida",
         added: "2009-07-29"}).
iana(language,gce,tag{
         type: "language",
         subtag: "gce",
         description: "Galice",
         added: "2009-07-29"}).
iana(language,gcf,tag{
         type: "language",
         subtag: "gcf",
         description: "Guadeloupean Creole French",
         added: "2009-07-29"}).
iana(language,gcl,tag{
         type: "language",
         subtag: "gcl",
         description: "Grenadian Creole English",
         added: "2009-07-29"}).
iana(language,gcn,tag{
         type: "language",
         subtag: "gcn",
         description: "Gaina",
         added: "2009-07-29"}).
iana(language,gcr,tag{
         type: "language",
         subtag: "gcr",
         description: "Guianese Creole French",
         added: "2009-07-29"}).
iana(language,gct,tag{
         type: "language",
         subtag: "gct",
         description: "Colonia Tovar German",
         added: "2009-07-29"}).
iana(language,gda,tag{
         type: "language",
         subtag: "gda",
         description: "Gade Lohar",
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,gdb,tag{
         type: "language",
         subtag: "gdb",
         description: "Pottangi Ollar Gadaba",
         added: "2009-07-29"}).
iana(language,gdc,tag{
         type: "language",
         subtag: "gdc",
         description: "Gugu Badhun",
         added: "2009-07-29"}).
iana(language,gdd,tag{
         type: "language",
         subtag: "gdd",
         description: "Gedaged",
         added: "2009-07-29"}).
iana(language,gde,tag{
         type: "language",
         subtag: "gde",
         description: "Gude",
         added: "2009-07-29"}).
iana(language,gdf,tag{
         type: "language",
         subtag: "gdf",
         description: "Guduf-Gava",
         added: "2009-07-29"}).
iana(language,gdg,tag{
         type: "language",
         subtag: "gdg",
         description: "Ga'dang",
         added: "2009-07-29"}).
iana(language,gdh,tag{
         type: "language",
         subtag: "gdh",
         description: ["Gadjerawang","Gajirrabeng"],
         added: "2009-07-29"}).
iana(language,gdi,tag{
         type: "language",
         subtag: "gdi",
         description: "Gundi",
         added: "2009-07-29"}).
iana(language,gdj,tag{
         type: "language",
         subtag: "gdj",
         description: "Gurdjar",
         added: "2009-07-29"}).
iana(language,gdk,tag{
         type: "language",
         subtag: "gdk",
         description: "Gadang",
         added: "2009-07-29"}).
iana(language,gdl,tag{
         type: "language",
         subtag: "gdl",
         description: "Dirasha",
         added: "2009-07-29"}).
iana(language,gdm,tag{
         type: "language",
         subtag: "gdm",
         description: "Laal",
         added: "2009-07-29"}).
iana(language,gdn,tag{
         type: "language",
         subtag: "gdn",
         description: "Umanakaina",
         added: "2009-07-29"}).
iana(language,gdo,tag{
         type: "language",
         subtag: "gdo",
         description: "Ghodoberi",
         added: "2009-07-29"}).
iana(language,gdq,tag{
         type: "language",
         subtag: "gdq",
         description: "Mehri",
         added: "2009-07-29"}).
iana(language,gdr,tag{
         type: "language",
         subtag: "gdr",
         description: "Wipi",
         added: "2009-07-29"}).
iana(language,gds,tag{
         type: "language",
         subtag: "gds",
         description: "Ghandruk Sign Language",
         added: "2012-08-12"}).
iana(language,gdt,tag{
         type: "language",
         subtag: "gdt",
         description: "Kungardutyi",
         added: "2013-09-10"}).
iana(language,gdu,tag{
         type: "language",
         subtag: "gdu",
         description: "Gudu",
         added: "2009-07-29"}).
iana(language,gdx,tag{
         type: "language",
         subtag: "gdx",
         description: "Godwari",
         added: "2009-07-29"}).
iana(language,gea,tag{
         type: "language",
         subtag: "gea",
         description: "Geruma",
         added: "2009-07-29"}).
iana(language,geb,tag{
         type: "language",
         subtag: "geb",
         description: "Kire",
         added: "2009-07-29"}).
iana(language,gec,tag{
         type: "language",
         subtag: "gec",
         description: "Gboloo Grebo",
         added: "2009-07-29",
         macrolanguage: "grb"}).
iana(language,ged,tag{
         type: "language",
         subtag: "ged",
         description: "Gade",
         added: "2009-07-29"}).
iana(language,gef,tag{
         type: "language",
         subtag: "gef",
         description: "Gerai",
         added: "2020-03-28"}).
iana(language,geg,tag{
         type: "language",
         subtag: "geg",
         description: "Gengle",
         added: "2009-07-29"}).
iana(language,geh,tag{
         type: "language",
         subtag: "geh",
         description: ["Hutterite German","Hutterisch"],
         added: "2009-07-29"}).
iana(language,gei,tag{
         type: "language",
         subtag: "gei",
         description: "Gebe",
         added: "2009-07-29"}).
iana(language,gej,tag{
         type: "language",
         subtag: "gej",
         description: "Gen",
         added: "2009-07-29"}).
iana(language,gek,tag{
         type: "language",
         subtag: "gek",
         description: "Ywom",
         added: "2009-07-29"}).
iana(language,gel,tag{
         type: "language",
         subtag: "gel",
         description: "ut-Ma'in",
         added: "2009-07-29"}).
iana(language,gem,tag{
         type: "language",
         subtag: "gem",
         description: "Germanic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,geq,tag{
         type: "language",
         subtag: "geq",
         description: "Geme",
         added: "2009-07-29"}).
iana(language,ges,tag{
         type: "language",
         subtag: "ges",
         description: "Geser-Gorom",
         added: "2009-07-29"}).
iana(language,gev,tag{
         type: "language",
         subtag: "gev",
         description: "Eviya",
         added: "2014-02-28"}).
iana(language,gew,tag{
         type: "language",
         subtag: "gew",
         description: "Gera",
         added: "2009-07-29"}).
iana(language,gex,tag{
         type: "language",
         subtag: "gex",
         description: "Garre",
         added: "2009-07-29"}).
iana(language,gey,tag{
         type: "language",
         subtag: "gey",
         description: "Enya",
         added: "2009-07-29"}).
iana(language,gez,tag{
         type: "language",
         subtag: "gez",
         description: "Geez",
         added: "2005-10-16"}).
iana(language,gfk,tag{
         type: "language",
         subtag: "gfk",
         description: "Patpatar",
         added: "2009-07-29"}).
iana(language,gft,tag{
         type: "language",
         subtag: "gft",
         description: "Gafat",
         added: "2009-07-29"}).
iana(language,gfx,tag{
         type: "language",
         subtag: "gfx",
         description: "Mangetti Dune ǃXung",
         added: "2012-08-12",
         deprecated: "2015-02-12",
         preferred_value: "vaj"}).
iana(language,gga,tag{
         type: "language",
         subtag: "gga",
         description: "Gao",
         added: "2009-07-29"}).
iana(language,ggb,tag{
         type: "language",
         subtag: "ggb",
         description: "Gbii",
         added: "2009-07-29"}).
iana(language,ggd,tag{
         type: "language",
         subtag: "ggd",
         description: "Gugadj",
         added: "2009-07-29"}).
iana(language,gge,tag{
         type: "language",
         subtag: "gge",
         description: "Gurr-goni",
         added: "2009-07-29"}).
iana(language,ggg,tag{
         type: "language",
         subtag: "ggg",
         description: "Gurgula",
         added: "2009-07-29"}).
iana(language,ggk,tag{
         type: "language",
         subtag: "ggk",
         description: "Kungarakany",
         added: "2009-07-29"}).
iana(language,ggl,tag{
         type: "language",
         subtag: "ggl",
         description: "Ganglau",
         added: "2009-07-29"}).
iana(language,ggn,tag{
         type: "language",
         subtag: "ggn",
         description: "Eastern Gurung",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "gvr"}).
iana(language,ggo,tag{
         type: "language",
         subtag: "ggo",
         description: "Southern Gondi",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         comments: "see esg, wsg"}).
iana(language,ggr,tag{
         type: "language",
         subtag: "ggr",
         description: "Aghu Tharnggalu",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see gtu, ikr"}).
iana(language,ggt,tag{
         type: "language",
         subtag: "ggt",
         description: "Gitua",
         added: "2009-07-29"}).
iana(language,ggu,tag{
         type: "language",
         subtag: "ggu",
         description: ["Gagu","Gban"],
         added: "2009-07-29"}).
iana(language,ggw,tag{
         type: "language",
         subtag: "ggw",
         description: "Gogodala",
         added: "2009-07-29"}).
iana(language,gha,tag{
         type: "language",
         subtag: "gha",
         description: "Ghadamès",
         added: "2009-07-29"}).
iana(language,ghc,tag{
         type: "language",
         subtag: "ghc",
         description: "Hiberno-Scottish Gaelic",
         added: "2009-07-29"}).
iana(language,ghe,tag{
         type: "language",
         subtag: "ghe",
         description: "Southern Ghale",
         added: "2009-07-29"}).
iana(language,ghh,tag{
         type: "language",
         subtag: "ghh",
         description: "Northern Ghale",
         added: "2009-07-29"}).
iana(language,ghk,tag{
         type: "language",
         subtag: "ghk",
         description: "Geko Karen",
         added: "2009-07-29"}).
iana(language,ghl,tag{
         type: "language",
         subtag: "ghl",
         description: "Ghulfan",
         added: "2009-07-29"}).
iana(language,ghn,tag{
         type: "language",
         subtag: "ghn",
         description: "Ghanongga",
         added: "2009-07-29"}).
iana(language,gho,tag{
         type: "language",
         subtag: "gho",
         description: "Ghomara",
         added: "2009-07-29"}).
iana(language,ghr,tag{
         type: "language",
         subtag: "ghr",
         description: "Ghera",
         added: "2009-07-29"}).
iana(language,ghs,tag{
         type: "language",
         subtag: "ghs",
         description: "Guhu-Samane",
         added: "2009-07-29"}).
iana(language,ght,tag{
         type: "language",
         subtag: "ght",
         description: ["Kuke","Kutang Ghale"],
         added: "2009-07-29"}).
iana(language,gia,tag{
         type: "language",
         subtag: "gia",
         description: "Kija",
         added: "2009-07-29"}).
iana(language,gib,tag{
         type: "language",
         subtag: "gib",
         description: "Gibanawa",
         added: "2009-07-29"}).
iana(language,gic,tag{
         type: "language",
         subtag: "gic",
         description: "Gail",
         added: "2009-07-29"}).
iana(language,gid,tag{
         type: "language",
         subtag: "gid",
         description: "Gidar",
         added: "2009-07-29"}).
iana(language,gie,tag{
         type: "language",
         subtag: "gie",
         description: ["Gaɓogbo","Guébie"],
         added: "2017-02-23"}).
iana(language,gig,tag{
         type: "language",
         subtag: "gig",
         description: "Goaria",
         added: "2009-07-29"}).
iana(language,gih,tag{
         type: "language",
         subtag: "gih",
         description: "Githabul",
         added: "2013-09-10"}).
iana(language,gii,tag{
         type: "language",
         subtag: "gii",
         description: "Girirra",
         added: "2021-02-20"}).
iana(language,gil,tag{
         type: "language",
         subtag: "gil",
         description: "Gilbertese",
         added: "2005-10-16"}).
iana(language,gim,tag{
         type: "language",
         subtag: "gim",
         description: "Gimi (Eastern Highlands)",
         added: "2009-07-29"}).
iana(language,gin,tag{
         type: "language",
         subtag: "gin",
         description: "Hinukh",
         added: "2009-07-29"}).
iana(language,gio,tag{
         type: "language",
         subtag: "gio",
         description: "Gelao",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see aou, gqu"}).
iana(language,gip,tag{
         type: "language",
         subtag: "gip",
         description: "Gimi (West New Britain)",
         added: "2009-07-29"}).
iana(language,giq,tag{
         type: "language",
         subtag: "giq",
         description: "Green Gelao",
         added: "2009-07-29"}).
iana(language,gir,tag{
         type: "language",
         subtag: "gir",
         description: "Red Gelao",
         added: "2009-07-29"}).
iana(language,gis,tag{
         type: "language",
         subtag: "gis",
         description: "North Giziga",
         added: "2009-07-29"}).
iana(language,git,tag{
         type: "language",
         subtag: "git",
         description: "Gitxsan",
         added: "2009-07-29"}).
iana(language,giu,tag{
         type: "language",
         subtag: "giu",
         description: "Mulao",
         added: "2012-08-12"}).
iana(language,giw,tag{
         type: "language",
         subtag: "giw",
         description: "White Gelao",
         added: "2009-07-29"}).
iana(language,gix,tag{
         type: "language",
         subtag: "gix",
         description: "Gilima",
         added: "2009-07-29"}).
iana(language,giy,tag{
         type: "language",
         subtag: "giy",
         description: "Giyug",
         added: "2009-07-29"}).
iana(language,giz,tag{
         type: "language",
         subtag: "giz",
         description: "South Giziga",
         added: "2009-07-29"}).
iana(language,gji,tag{
         type: "language",
         subtag: "gji",
         description: "Geji",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see gyz, zbu"}).
iana(language,gjk,tag{
         type: "language",
         subtag: "gjk",
         description: "Kachi Koli",
         added: "2009-07-29"}).
iana(language,gjm,tag{
         type: "language",
         subtag: "gjm",
         description: "Gunditjmara",
         added: "2013-09-10"}).
iana(language,gjn,tag{
         type: "language",
         subtag: "gjn",
         description: "Gonja",
         added: "2009-07-29"}).
iana(language,gjr,tag{
         type: "language",
         subtag: "gjr",
         description: "Gurindji Kriol",
         added: "2016-05-30"}).
iana(language,gju,tag{
         type: "language",
         subtag: "gju",
         description: "Gujari",
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,gka,tag{
         type: "language",
         subtag: "gka",
         description: "Guya",
         added: "2009-07-29"}).
iana(language,gkd,tag{
         type: "language",
         subtag: "gkd",
         description: "Magɨ (Madang Province)",
         added: "2018-03-08"}).
iana(language,gke,tag{
         type: "language",
         subtag: "gke",
         description: "Ndai",
         added: "2009-07-29"}).
iana(language,gkn,tag{
         type: "language",
         subtag: "gkn",
         description: "Gokana",
         added: "2009-07-29"}).
iana(language,gko,tag{
         type: "language",
         subtag: "gko",
         description: "Kok-Nar",
         added: "2012-08-12"}).
iana(language,gkp,tag{
         type: "language",
         subtag: "gkp",
         description: "Guinea Kpelle",
         added: "2009-07-29",
         macrolanguage: "kpe"}).
iana(language,gku,tag{
         type: "language",
         subtag: "gku",
         description: "ǂUngkue",
         added: "2015-02-12"}).
iana(language,glb,tag{
         type: "language",
         subtag: "glb",
         description: "Belning",
         added: "2021-02-20"}).
iana(language,glc,tag{
         type: "language",
         subtag: "glc",
         description: "Bon Gula",
         added: "2009-07-29"}).
iana(language,gld,tag{
         type: "language",
         subtag: "gld",
         description: "Nanai",
         added: "2009-07-29"}).
iana(language,glh,tag{
         type: "language",
         subtag: "glh",
         description: ["Northwest Pashai","Northwest Pashayi"],
         added: "2009-07-29"}).
iana(language,gli,tag{
         type: "language",
         subtag: "gli",
         description: "Guliguli",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "kzk"}).
iana(language,glj,tag{
         type: "language",
         subtag: "glj",
         description: "Gula Iro",
         added: "2009-07-29"}).
iana(language,glk,tag{
         type: "language",
         subtag: "glk",
         description: "Gilaki",
         added: "2009-07-29"}).
iana(language,gll,tag{
         type: "language",
         subtag: "gll",
         description: "Garlali",
         added: "2013-09-10"}).
iana(language,glo,tag{
         type: "language",
         subtag: "glo",
         description: "Galambu",
         added: "2009-07-29"}).
iana(language,glr,tag{
         type: "language",
         subtag: "glr",
         description: "Glaro-Twabo",
         added: "2009-07-29"}).
iana(language,glu,tag{
         type: "language",
         subtag: "glu",
         description: "Gula (Chad)",
         added: "2009-07-29"}).
iana(language,glw,tag{
         type: "language",
         subtag: "glw",
         description: "Glavda",
         added: "2009-07-29"}).
iana(language,gly,tag{
         type: "language",
         subtag: "gly",
         description: "Gule",
         added: "2009-07-29"}).
iana(language,gma,tag{
         type: "language",
         subtag: "gma",
         description: "Gambera",
         added: "2009-07-29"}).
iana(language,gmb,tag{
         type: "language",
         subtag: "gmb",
         description: "Gula'alaa",
         added: "2009-07-29"}).
iana(language,gmd,tag{
         type: "language",
         subtag: "gmd",
         description: "Mághdì",
         added: "2009-07-29"}).
iana(language,gme,tag{
         type: "language",
         subtag: "gme",
         description: "East Germanic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,gmg,tag{
         type: "language",
         subtag: "gmg",
         description: "Magɨyi",
         added: "2014-02-28"}).
iana(language,gmh,tag{
         type: "language",
         subtag: "gmh",
         description: "Middle High German (ca. 1050-1500)",
         added: "2005-10-16"}).
iana(language,gml,tag{
         type: "language",
         subtag: "gml",
         description: "Middle Low German",
         added: "2009-07-29"}).
iana(language,gmm,tag{
         type: "language",
         subtag: "gmm",
         description: "Gbaya-Mbodomo",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,gmn,tag{
         type: "language",
         subtag: "gmn",
         description: "Gimnime",
         added: "2009-07-29"}).
iana(language,gmq,tag{
         type: "language",
         subtag: "gmq",
         description: "North Germanic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,gmr,tag{
         type: "language",
         subtag: "gmr",
         description: ["Mirning","Mirniny"],
         added: "2020-03-28"}).
iana(language,gmu,tag{
         type: "language",
         subtag: "gmu",
         description: "Gumalu",
         added: "2009-07-29"}).
iana(language,gmv,tag{
         type: "language",
         subtag: "gmv",
         description: "Gamo",
         added: "2009-07-29"}).
iana(language,gmw,tag{
         type: "language",
         subtag: "gmw",
         description: "West Germanic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,gmx,tag{
         type: "language",
         subtag: "gmx",
         description: "Magoma",
         added: "2009-07-29"}).
iana(language,gmy,tag{
         type: "language",
         subtag: "gmy",
         description: "Mycenaean Greek",
         added: "2009-07-29"}).
iana(language,gmz,tag{
         type: "language",
         subtag: "gmz",
         description: "Mgbolizhia",
         added: "2013-09-10"}).
iana(language,gna,tag{
         type: "language",
         subtag: "gna",
         description: "Kaansa",
         added: "2009-07-29"}).
iana(language,gnb,tag{
         type: "language",
         subtag: "gnb",
         description: "Gangte",
         added: "2009-07-29"}).
iana(language,gnc,tag{
         type: "language",
         subtag: "gnc",
         description: "Guanche",
         added: "2009-07-29"}).
iana(language,gnd,tag{
         type: "language",
         subtag: "gnd",
         description: "Zulgo-Gemzek",
         added: "2009-07-29"}).
iana(language,gne,tag{
         type: "language",
         subtag: "gne",
         description: "Ganang",
         added: "2009-07-29"}).
iana(language,gng,tag{
         type: "language",
         subtag: "gng",
         description: "Ngangam",
         added: "2009-07-29"}).
iana(language,gnh,tag{
         type: "language",
         subtag: "gnh",
         description: "Lere",
         added: "2009-07-29"}).
iana(language,gni,tag{
         type: "language",
         subtag: "gni",
         description: "Gooniyandi",
         added: "2009-07-29"}).
iana(language,gnj,tag{
         type: "language",
         subtag: "gnj",
         description: "Ngen",
         added: "2018-03-08"}).
iana(language,gnk,tag{
         type: "language",
         subtag: "gnk",
         description: "ǁGana",
         added: "2009-07-29"}).
iana(language,gnl,tag{
         type: "language",
         subtag: "gnl",
         description: "Gangulu",
         added: "2009-07-29"}).
iana(language,gnm,tag{
         type: "language",
         subtag: "gnm",
         description: "Ginuman",
         added: "2009-07-29"}).
iana(language,gnn,tag{
         type: "language",
         subtag: "gnn",
         description: "Gumatj",
         added: "2009-07-29"}).
iana(language,gno,tag{
         type: "language",
         subtag: "gno",
         description: "Northern Gondi",
         added: "2009-07-29",
         macrolanguage: "gon"}).
iana(language,gnq,tag{
         type: "language",
         subtag: "gnq",
         description: "Gana",
         added: "2009-07-29"}).
iana(language,gnr,tag{
         type: "language",
         subtag: "gnr",
         description: "Gureng Gureng",
         added: "2009-07-29"}).
iana(language,gnt,tag{
         type: "language",
         subtag: "gnt",
         description: "Guntai",
         added: "2009-07-29"}).
iana(language,gnu,tag{
         type: "language",
         subtag: "gnu",
         description: "Gnau",
         added: "2009-07-29"}).
iana(language,gnw,tag{
         type: "language",
         subtag: "gnw",
         description: "Western Bolivian Guaraní",
         added: "2009-07-29",
         macrolanguage: "gn"}).
iana(language,gnz,tag{
         type: "language",
         subtag: "gnz",
         description: "Ganzi",
         added: "2009-07-29"}).
iana(language,goa,tag{
         type: "language",
         subtag: "goa",
         description: "Guro",
         added: "2009-07-29"}).
iana(language,gob,tag{
         type: "language",
         subtag: "gob",
         description: "Playero",
         added: "2009-07-29"}).
iana(language,goc,tag{
         type: "language",
         subtag: "goc",
         description: "Gorakor",
         added: "2009-07-29"}).
iana(language,god,tag{
         type: "language",
         subtag: "god",
         description: "Godié",
         added: "2009-07-29"}).
iana(language,goe,tag{
         type: "language",
         subtag: "goe",
         description: "Gongduk",
         added: "2009-07-29"}).
iana(language,gof,tag{
         type: "language",
         subtag: "gof",
         description: "Gofa",
         added: "2009-07-29"}).
iana(language,gog,tag{
         type: "language",
         subtag: "gog",
         description: "Gogo",
         added: "2009-07-29"}).
iana(language,goh,tag{
         type: "language",
         subtag: "goh",
         description: "Old High German (ca. 750-1050)",
         added: "2005-10-16"}).
iana(language,goi,tag{
         type: "language",
         subtag: "goi",
         description: "Gobasi",
         added: "2009-07-29"}).
iana(language,goj,tag{
         type: "language",
         subtag: "goj",
         description: "Gowlan",
         added: "2009-07-29"}).
iana(language,gok,tag{
         type: "language",
         subtag: "gok",
         description: "Gowli",
         added: "2009-07-29"}).
iana(language,gol,tag{
         type: "language",
         subtag: "gol",
         description: "Gola",
         added: "2009-07-29"}).
iana(language,gom,tag{
         type: "language",
         subtag: "gom",
         description: "Goan Konkani",
         added: "2009-07-29",
         macrolanguage: "kok"}).
iana(language,gon,tag{
         type: "language",
         subtag: "gon",
         description: "Gondi",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,goo,tag{
         type: "language",
         subtag: "goo",
         description: "Gone Dau",
         added: "2009-07-29"}).
iana(language,gop,tag{
         type: "language",
         subtag: "gop",
         description: "Yeretuar",
         added: "2009-07-29"}).
iana(language,goq,tag{
         type: "language",
         subtag: "goq",
         description: "Gorap",
         added: "2009-07-29"}).
iana(language,gor,tag{
         type: "language",
         subtag: "gor",
         description: "Gorontalo",
         added: "2005-10-16"}).
iana(language,gos,tag{
         type: "language",
         subtag: "gos",
         description: "Gronings",
         added: "2009-07-29"}).
iana(language,got,tag{
         type: "language",
         subtag: "got",
         description: "Gothic",
         added: "2005-10-16"}).
iana(language,gou,tag{
         type: "language",
         subtag: "gou",
         description: "Gavar",
         added: "2009-07-29"}).
iana(language,gov,tag{
         type: "language",
         subtag: "gov",
         description: "Goo",
         added: "2022-02-25"}).
iana(language,gow,tag{
         type: "language",
         subtag: "gow",
         description: "Gorowa",
         added: "2009-07-29"}).
iana(language,gox,tag{
         type: "language",
         subtag: "gox",
         description: "Gobu",
         added: "2009-07-29"}).
iana(language,goy,tag{
         type: "language",
         subtag: "goy",
         description: "Goundo",
         added: "2009-07-29"}).
iana(language,goz,tag{
         type: "language",
         subtag: "goz",
         description: "Gozarkhani",
         added: "2009-07-29"}).
iana(language,gpa,tag{
         type: "language",
         subtag: "gpa",
         description: "Gupa-Abawa",
         added: "2009-07-29"}).
iana(language,gpe,tag{
         type: "language",
         subtag: "gpe",
         description: "Ghanaian Pidgin English",
         added: "2012-08-12"}).
iana(language,gpn,tag{
         type: "language",
         subtag: "gpn",
         description: "Taiap",
         added: "2009-07-29"}).
iana(language,gqa,tag{
         type: "language",
         subtag: "gqa",
         description: "Ga'anda",
         added: "2009-07-29"}).
iana(language,gqi,tag{
         type: "language",
         subtag: "gqi",
         description: "Guiqiong",
         added: "2009-07-29"}).
iana(language,gqn,tag{
         type: "language",
         subtag: "gqn",
         description: "Guana (Brazil)",
         added: "2009-07-29"}).
iana(language,gqr,tag{
         type: "language",
         subtag: "gqr",
         description: "Gor",
         added: "2009-07-29"}).
iana(language,gqu,tag{
         type: "language",
         subtag: "gqu",
         description: "Qau",
         added: "2012-08-12"}).
iana(language,gra,tag{
         type: "language",
         subtag: "gra",
         description: "Rajput Garasia",
         added: "2009-07-29"}).
iana(language,grb,tag{
         type: "language",
         subtag: "grb",
         description: "Grebo",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,grc,tag{
         type: "language",
         subtag: "grc",
         description: "Ancient Greek (to 1453)",
         added: "2005-10-16"}).
iana(language,grd,tag{
         type: "language",
         subtag: "grd",
         description: "Guruntum-Mbaaru",
         added: "2009-07-29"}).
iana(language,grg,tag{
         type: "language",
         subtag: "grg",
         description: "Madi",
         added: "2009-07-29"}).
iana(language,grh,tag{
         type: "language",
         subtag: "grh",
         description: "Gbiri-Niragu",
         added: "2009-07-29"}).
iana(language,gri,tag{
         type: "language",
         subtag: "gri",
         description: "Ghari",
         added: "2009-07-29"}).
iana(language,grj,tag{
         type: "language",
         subtag: "grj",
         description: "Southern Grebo",
         added: "2009-07-29",
         macrolanguage: "grb"}).
iana(language,grk,tag{
         type: "language",
         subtag: "grk",
         description: "Greek languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,grm,tag{
         type: "language",
         subtag: "grm",
         description: "Kota Marudu Talantang",
         added: "2009-07-29"}).
iana(language,gro,tag{
         type: "language",
         subtag: "gro",
         description: "Groma",
         added: "2009-07-29"}).
iana(language,grq,tag{
         type: "language",
         subtag: "grq",
         description: "Gorovu",
         added: "2009-07-29"}).
iana(language,grr,tag{
         type: "language",
         subtag: "grr",
         description: "Taznatit",
         added: "2009-07-29"}).
iana(language,grs,tag{
         type: "language",
         subtag: "grs",
         description: "Gresi",
         added: "2009-07-29"}).
iana(language,grt,tag{
         type: "language",
         subtag: "grt",
         description: "Garo",
         added: "2009-07-29"}).
iana(language,gru,tag{
         type: "language",
         subtag: "gru",
         description: "Kistane",
         added: "2009-07-29"}).
iana(language,grv,tag{
         type: "language",
         subtag: "grv",
         description: "Central Grebo",
         added: "2009-07-29",
         macrolanguage: "grb"}).
iana(language,grw,tag{
         type: "language",
         subtag: "grw",
         description: "Gweda",
         added: "2009-07-29"}).
iana(language,grx,tag{
         type: "language",
         subtag: "grx",
         description: "Guriaso",
         added: "2009-07-29"}).
iana(language,gry,tag{
         type: "language",
         subtag: "gry",
         description: "Barclayville Grebo",
         added: "2009-07-29",
         macrolanguage: "grb"}).
iana(language,grz,tag{
         type: "language",
         subtag: "grz",
         description: "Guramalum",
         added: "2009-07-29"}).
iana(language,gse,tag{
         type: "language",
         subtag: "gse",
         description: "Ghanaian Sign Language",
         added: "2009-07-29"}).
iana(language,gsg,tag{
         type: "language",
         subtag: "gsg",
         description: "German Sign Language",
         added: "2009-07-29"}).
iana(language,gsl,tag{
         type: "language",
         subtag: "gsl",
         description: "Gusilay",
         added: "2009-07-29"}).
iana(language,gsm,tag{
         type: "language",
         subtag: "gsm",
         description: "Guatemalan Sign Language",
         added: "2009-07-29"}).
iana(language,gsn,tag{
         type: "language",
         subtag: "gsn",
         description: ["Nema","Gusan"],
         added: "2009-07-29"}).
iana(language,gso,tag{
         type: "language",
         subtag: "gso",
         description: "Southwest Gbaya",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,gsp,tag{
         type: "language",
         subtag: "gsp",
         description: "Wasembo",
         added: "2009-07-29"}).
iana(language,gss,tag{
         type: "language",
         subtag: "gss",
         description: "Greek Sign Language",
         added: "2009-07-29"}).
iana(language,gsw,tag{
         type: "language",
         subtag: "gsw",
         description: ["Swiss German","Alemannic","Alsatian"],
         added: "2006-03-08",
         suppress_script: "Latn"}).
iana(language,gta,tag{
         type: "language",
         subtag: "gta",
         description: "Guató",
         added: "2009-07-29"}).
iana(language,gti,tag{
         type: "language",
         subtag: "gti",
         description: "Gbati-ri",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "nyc"}).
iana(language,gtu,tag{
         type: "language",
         subtag: "gtu",
         description: "Aghu-Tharnggala",
         added: "2013-09-10"}).
iana(language,gua,tag{
         type: "language",
         subtag: "gua",
         description: "Shiki",
         added: "2009-07-29"}).
iana(language,gub,tag{
         type: "language",
         subtag: "gub",
         description: "Guajajára",
         added: "2009-07-29"}).
iana(language,guc,tag{
         type: "language",
         subtag: "guc",
         description: "Wayuu",
         added: "2009-07-29"}).
iana(language,gud,tag{
         type: "language",
         subtag: "gud",
         description: "Yocoboué Dida",
         added: "2009-07-29"}).
iana(language,gue,tag{
         type: "language",
         subtag: "gue",
         description: "Gurindji",
         added: "2009-07-29"}).
iana(language,guf,tag{
         type: "language",
         subtag: "guf",
         description: "Gupapuyngu",
         added: "2009-07-29"}).
iana(language,gug,tag{
         type: "language",
         subtag: "gug",
         description: "Paraguayan Guaraní",
         added: "2009-07-29",
         macrolanguage: "gn"}).
iana(language,guh,tag{
         type: "language",
         subtag: "guh",
         description: "Guahibo",
         added: "2009-07-29"}).
iana(language,gui,tag{
         type: "language",
         subtag: "gui",
         description: "Eastern Bolivian Guaraní",
         added: "2009-07-29",
         macrolanguage: "gn"}).
iana(language,guk,tag{
         type: "language",
         subtag: "guk",
         description: "Gumuz",
         added: "2009-07-29"}).
iana(language,gul,tag{
         type: "language",
         subtag: "gul",
         description: "Sea Island Creole English",
         added: "2009-07-29"}).
iana(language,gum,tag{
         type: "language",
         subtag: "gum",
         description: "Guambiano",
         added: "2009-07-29"}).
iana(language,gun,tag{
         type: "language",
         subtag: "gun",
         description: "Mbyá Guaraní",
         added: "2009-07-29",
         macrolanguage: "gn"}).
iana(language,guo,tag{
         type: "language",
         subtag: "guo",
         description: "Guayabero",
         added: "2009-07-29"}).
iana(language,gup,tag{
         type: "language",
         subtag: "gup",
         description: "Gunwinggu",
         added: "2009-07-29"}).
iana(language,guq,tag{
         type: "language",
         subtag: "guq",
         description: "Aché",
         added: "2009-07-29"}).
iana(language,gur,tag{
         type: "language",
         subtag: "gur",
         description: "Farefare",
         added: "2009-07-29"}).
iana(language,gus,tag{
         type: "language",
         subtag: "gus",
         description: "Guinean Sign Language",
         added: "2009-07-29"}).
iana(language,gut,tag{
         type: "language",
         subtag: "gut",
         description: "Maléku Jaíka",
         added: "2009-07-29"}).
iana(language,guu,tag{
         type: "language",
         subtag: "guu",
         description: "Yanomamö",
         added: "2009-07-29"}).
iana(language,guv,tag{
         type: "language",
         subtag: "guv",
         description: "Gey",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "duz"}).
iana(language,guw,tag{
         type: "language",
         subtag: "guw",
         description: "Gun",
         added: "2009-07-29"}).
iana(language,gux,tag{
         type: "language",
         subtag: "gux",
         description: "Gourmanchéma",
         added: "2009-07-29"}).
iana(language,guz,tag{
         type: "language",
         subtag: "guz",
         description: ["Gusii","Ekegusii"],
         added: "2009-07-29"}).
iana(language,gva,tag{
         type: "language",
         subtag: "gva",
         description: "Guana (Paraguay)",
         added: "2009-07-29"}).
iana(language,gvc,tag{
         type: "language",
         subtag: "gvc",
         description: "Guanano",
         added: "2009-07-29"}).
iana(language,gve,tag{
         type: "language",
         subtag: "gve",
         description: "Duwet",
         added: "2009-07-29"}).
iana(language,gvf,tag{
         type: "language",
         subtag: "gvf",
         description: "Golin",
         added: "2009-07-29"}).
iana(language,gvj,tag{
         type: "language",
         subtag: "gvj",
         description: "Guajá",
         added: "2009-07-29"}).
iana(language,gvl,tag{
         type: "language",
         subtag: "gvl",
         description: "Gulay",
         added: "2009-07-29"}).
iana(language,gvm,tag{
         type: "language",
         subtag: "gvm",
         description: "Gurmana",
         added: "2009-07-29"}).
iana(language,gvn,tag{
         type: "language",
         subtag: "gvn",
         description: "Kuku-Yalanji",
         added: "2009-07-29"}).
iana(language,gvo,tag{
         type: "language",
         subtag: "gvo",
         description: "Gavião Do Jiparaná",
         added: "2009-07-29"}).
iana(language,gvp,tag{
         type: "language",
         subtag: "gvp",
         description: "Pará Gavião",
         added: "2009-07-29"}).
iana(language,gvr,tag{
         type: "language",
         subtag: "gvr",
         description: "Gurung",
         added: "2009-07-29"}).
iana(language,gvs,tag{
         type: "language",
         subtag: "gvs",
         description: "Gumawana",
         added: "2009-07-29"}).
iana(language,gvy,tag{
         type: "language",
         subtag: "gvy",
         description: "Guyani",
         added: "2009-07-29"}).
iana(language,gwa,tag{
         type: "language",
         subtag: "gwa",
         description: "Mbato",
         added: "2009-07-29"}).
iana(language,gwb,tag{
         type: "language",
         subtag: "gwb",
         description: "Gwa",
         added: "2009-07-29"}).
iana(language,gwc,tag{
         type: "language",
         subtag: "gwc",
         description: ["Gawri","Kalami"],
         added: "2009-07-29"}).
iana(language,gwd,tag{
         type: "language",
         subtag: "gwd",
         description: "Gawwada",
         added: "2009-07-29"}).
iana(language,gwe,tag{
         type: "language",
         subtag: "gwe",
         description: "Gweno",
         added: "2009-07-29"}).
iana(language,gwf,tag{
         type: "language",
         subtag: "gwf",
         description: "Gowro",
         added: "2009-07-29"}).
iana(language,gwg,tag{
         type: "language",
         subtag: "gwg",
         description: "Moo",
         added: "2009-07-29"}).
iana(language,gwi,tag{
         type: "language",
         subtag: "gwi",
         description: "Gwichʼin",
         added: "2005-10-16"}).
iana(language,gwj,tag{
         type: "language",
         subtag: "gwj",
         description: "ǀGwi",
         added: "2009-07-29"}).
iana(language,gwm,tag{
         type: "language",
         subtag: "gwm",
         description: "Awngthim",
         added: "2012-08-12"}).
iana(language,gwn,tag{
         type: "language",
         subtag: "gwn",
         description: "Gwandara",
         added: "2009-07-29"}).
iana(language,gwr,tag{
         type: "language",
         subtag: "gwr",
         description: "Gwere",
         added: "2009-07-29"}).
iana(language,gwt,tag{
         type: "language",
         subtag: "gwt",
         description: "Gawar-Bati",
         added: "2009-07-29"}).
iana(language,gwu,tag{
         type: "language",
         subtag: "gwu",
         description: "Guwamu",
         added: "2009-07-29"}).
iana(language,gww,tag{
         type: "language",
         subtag: "gww",
         description: "Kwini",
         added: "2009-07-29"}).
iana(language,gwx,tag{
         type: "language",
         subtag: "gwx",
         description: "Gua",
         added: "2009-07-29"}).
iana(language,gxx,tag{
         type: "language",
         subtag: "gxx",
         description: "Wè Southern",
         added: "2009-07-29"}).
iana(language,gya,tag{
         type: "language",
         subtag: "gya",
         description: "Northwest Gbaya",
         added: "2009-07-29",
         macrolanguage: "gba"}).
iana(language,gyb,tag{
         type: "language",
         subtag: "gyb",
         description: "Garus",
         added: "2009-07-29"}).
iana(language,gyd,tag{
         type: "language",
         subtag: "gyd",
         description: "Kayardild",
         added: "2009-07-29"}).
iana(language,gye,tag{
         type: "language",
         subtag: "gye",
         description: "Gyem",
         added: "2009-07-29"}).
iana(language,gyf,tag{
         type: "language",
         subtag: "gyf",
         description: "Gungabula",
         added: "2009-07-29"}).
iana(language,gyg,tag{
         type: "language",
         subtag: "gyg",
         description: "Gbayi",
         added: "2009-07-29"}).
iana(language,gyi,tag{
         type: "language",
         subtag: "gyi",
         description: "Gyele",
         added: "2009-07-29"}).
iana(language,gyl,tag{
         type: "language",
         subtag: "gyl",
         description: "Gayil",
         added: "2009-07-29"}).
iana(language,gym,tag{
         type: "language",
         subtag: "gym",
         description: "Ngäbere",
         added: "2009-07-29"}).
iana(language,gyn,tag{
         type: "language",
         subtag: "gyn",
         description: "Guyanese Creole English",
         added: "2009-07-29"}).
iana(language,gyo,tag{
         type: "language",
         subtag: "gyo",
         description: "Gyalsumdo",
         added: "2018-03-08"}).
iana(language,gyr,tag{
         type: "language",
         subtag: "gyr",
         description: "Guarayu",
         added: "2009-07-29"}).
iana(language,gyy,tag{
         type: "language",
         subtag: "gyy",
         description: "Gunya",
         added: "2009-07-29"}).
iana(language,gyz,tag{
         type: "language",
         subtag: "gyz",
         description: ["Geji","Gyaazi"],
         added: "2021-02-20"}).
iana(language,gza,tag{
         type: "language",
         subtag: "gza",
         description: "Ganza",
         added: "2009-07-29"}).
iana(language,gzi,tag{
         type: "language",
         subtag: "gzi",
         description: "Gazi",
         added: "2009-07-29"}).
iana(language,gzn,tag{
         type: "language",
         subtag: "gzn",
         description: "Gane",
         added: "2009-07-29"}).
iana(language,haa,tag{
         type: "language",
         subtag: "haa",
         description: "Han",
         added: "2009-07-29"}).
iana(language,hab,tag{
         type: "language",
         subtag: "hab",
         description: "Hanoi Sign Language",
         added: "2009-07-29"}).
iana(language,hac,tag{
         type: "language",
         subtag: "hac",
         description: "Gurani",
         added: "2009-07-29"}).
iana(language,had,tag{
         type: "language",
         subtag: "had",
         description: "Hatam",
         added: "2009-07-29"}).
iana(language,hae,tag{
         type: "language",
         subtag: "hae",
         description: "Eastern Oromo",
         added: "2009-07-29",
         macrolanguage: "om"}).
iana(language,haf,tag{
         type: "language",
         subtag: "haf",
         description: "Haiphong Sign Language",
         added: "2009-07-29"}).
iana(language,hag,tag{
         type: "language",
         subtag: "hag",
         description: "Hanga",
         added: "2009-07-29"}).
iana(language,hah,tag{
         type: "language",
         subtag: "hah",
         description: "Hahon",
         added: "2009-07-29"}).
iana(language,hai,tag{
         type: "language",
         subtag: "hai",
         description: "Haida",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,haj,tag{
         type: "language",
         subtag: "haj",
         description: "Hajong",
         added: "2009-07-29"}).
iana(language,hak,tag{
         type: "language",
         subtag: "hak",
         description: "Hakka Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,hal,tag{
         type: "language",
         subtag: "hal",
         description: "Halang",
         added: "2009-07-29"}).
iana(language,ham,tag{
         type: "language",
         subtag: "ham",
         description: "Hewa",
         added: "2009-07-29"}).
iana(language,han,tag{
         type: "language",
         subtag: "han",
         description: "Hangaza",
         added: "2009-07-29"}).
iana(language,hao,tag{
         type: "language",
         subtag: "hao",
         description: "Hakö",
         added: "2009-07-29"}).
iana(language,hap,tag{
         type: "language",
         subtag: "hap",
         description: "Hupla",
         added: "2009-07-29"}).
iana(language,haq,tag{
         type: "language",
         subtag: "haq",
         description: "Ha",
         added: "2009-07-29"}).
iana(language,har,tag{
         type: "language",
         subtag: "har",
         description: "Harari",
         added: "2009-07-29"}).
iana(language,has,tag{
         type: "language",
         subtag: "has",
         description: "Haisla",
         added: "2009-07-29"}).
iana(language,hav,tag{
         type: "language",
         subtag: "hav",
         description: "Havu",
         added: "2009-07-29"}).
iana(language,haw,tag{
         type: "language",
         subtag: "haw",
         description: "Hawaiian",
         added: "2005-10-16"}).
iana(language,hax,tag{
         type: "language",
         subtag: "hax",
         description: "Southern Haida",
         added: "2009-07-29",
         macrolanguage: "hai"}).
iana(language,hay,tag{
         type: "language",
         subtag: "hay",
         description: "Haya",
         added: "2009-07-29"}).
iana(language,haz,tag{
         type: "language",
         subtag: "haz",
         description: "Hazaragi",
         added: "2009-07-29"}).
iana(language,hba,tag{
         type: "language",
         subtag: "hba",
         description: "Hamba",
         added: "2009-07-29"}).
iana(language,hbb,tag{
         type: "language",
         subtag: "hbb",
         description: "Huba",
         added: "2009-07-29"}).
iana(language,hbn,tag{
         type: "language",
         subtag: "hbn",
         description: "Heiban",
         added: "2009-07-29"}).
iana(language,hbo,tag{
         type: "language",
         subtag: "hbo",
         description: "Ancient Hebrew",
         added: "2009-07-29"}).
iana(language,hbu,tag{
         type: "language",
         subtag: "hbu",
         description: "Habu",
         added: "2009-07-29"}).
iana(language,hca,tag{
         type: "language",
         subtag: "hca",
         description: "Andaman Creole Hindi",
         added: "2009-07-29"}).
iana(language,hch,tag{
         type: "language",
         subtag: "hch",
         description: "Huichol",
         added: "2009-07-29"}).
iana(language,hdn,tag{
         type: "language",
         subtag: "hdn",
         description: "Northern Haida",
         added: "2009-07-29",
         macrolanguage: "hai"}).
iana(language,hds,tag{
         type: "language",
         subtag: "hds",
         description: "Honduras Sign Language",
         added: "2009-07-29"}).
iana(language,hdy,tag{
         type: "language",
         subtag: "hdy",
         description: "Hadiyya",
         added: "2009-07-29"}).
iana(language,hea,tag{
         type: "language",
         subtag: "hea",
         description: "Northern Qiandong Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hed,tag{
         type: "language",
         subtag: "hed",
         description: "Herdé",
         added: "2009-07-29"}).
iana(language,heg,tag{
         type: "language",
         subtag: "heg",
         description: "Helong",
         added: "2009-07-29"}).
iana(language,heh,tag{
         type: "language",
         subtag: "heh",
         description: "Hehe",
         added: "2009-07-29"}).
iana(language,hei,tag{
         type: "language",
         subtag: "hei",
         description: "Heiltsuk",
         added: "2009-07-29"}).
iana(language,hem,tag{
         type: "language",
         subtag: "hem",
         description: "Hemba",
         added: "2009-07-29"}).
iana(language,hgm,tag{
         type: "language",
         subtag: "hgm",
         description: "Haiǁom",
         added: "2009-07-29"}).
iana(language,hgw,tag{
         type: "language",
         subtag: "hgw",
         description: "Haigwai",
         added: "2009-07-29"}).
iana(language,hhi,tag{
         type: "language",
         subtag: "hhi",
         description: "Hoia Hoia",
         added: "2009-07-29"}).
iana(language,hhr,tag{
         type: "language",
         subtag: "hhr",
         description: "Kerak",
         added: "2009-07-29"}).
iana(language,hhy,tag{
         type: "language",
         subtag: "hhy",
         description: "Hoyahoya",
         added: "2009-07-29"}).
iana(language,hia,tag{
         type: "language",
         subtag: "hia",
         description: "Lamang",
         added: "2009-07-29"}).
iana(language,hib,tag{
         type: "language",
         subtag: "hib",
         description: "Hibito",
         added: "2009-07-29"}).
iana(language,hid,tag{
         type: "language",
         subtag: "hid",
         description: "Hidatsa",
         added: "2009-07-29"}).
iana(language,hif,tag{
         type: "language",
         subtag: "hif",
         description: "Fiji Hindi",
         added: "2009-07-29"}).
iana(language,hig,tag{
         type: "language",
         subtag: "hig",
         description: "Kamwe",
         added: "2009-07-29"}).
iana(language,hih,tag{
         type: "language",
         subtag: "hih",
         description: "Pamosu",
         added: "2009-07-29"}).
iana(language,hii,tag{
         type: "language",
         subtag: "hii",
         description: "Hinduri",
         added: "2009-07-29"}).
iana(language,hij,tag{
         type: "language",
         subtag: "hij",
         description: "Hijuk",
         added: "2009-07-29"}).
iana(language,hik,tag{
         type: "language",
         subtag: "hik",
         description: "Seit-Kaitetu",
         added: "2009-07-29"}).
iana(language,hil,tag{
         type: "language",
         subtag: "hil",
         description: "Hiligaynon",
         added: "2005-10-16"}).
iana(language,him,tag{
         type: "language",
         subtag: "him",
         description: ["Himachali languages","Western Pahari languages"],
         added: "2005-10-16",
         scope: "collection"}).
iana(language,hio,tag{
         type: "language",
         subtag: "hio",
         description: "Tsoa",
         added: "2009-07-29"}).
iana(language,hir,tag{
         type: "language",
         subtag: "hir",
         description: "Himarimã",
         added: "2009-07-29"}).
iana(language,hit,tag{
         type: "language",
         subtag: "hit",
         description: "Hittite",
         added: "2005-10-16"}).
iana(language,hiw,tag{
         type: "language",
         subtag: "hiw",
         description: "Hiw",
         added: "2009-07-29"}).
iana(language,hix,tag{
         type: "language",
         subtag: "hix",
         description: "Hixkaryána",
         added: "2009-07-29"}).
iana(language,hji,tag{
         type: "language",
         subtag: "hji",
         description: "Haji",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,hka,tag{
         type: "language",
         subtag: "hka",
         description: "Kahe",
         added: "2009-07-29"}).
iana(language,hke,tag{
         type: "language",
         subtag: "hke",
         description: "Hunde",
         added: "2009-07-29"}).
iana(language,hkh,tag{
         type: "language",
         subtag: "hkh",
         description: ["Khah","Poguli"],
         added: "2021-02-20"}).
iana(language,hkk,tag{
         type: "language",
         subtag: "hkk",
         description: "Hunjara-Kaina Ke",
         added: "2009-07-29"}).
iana(language,hkn,tag{
         type: "language",
         subtag: "hkn",
         description: "Mel-Khaonh",
         added: "2018-03-08"}).
iana(language,hks,tag{
         type: "language",
         subtag: "hks",
         description: ["Hong Kong Sign Language","Heung Kong Sau Yue"],
         added: "2009-07-29"}).
iana(language,hla,tag{
         type: "language",
         subtag: "hla",
         description: "Halia",
         added: "2009-07-29"}).
iana(language,hlb,tag{
         type: "language",
         subtag: "hlb",
         description: "Halbi",
         added: "2009-07-29"}).
iana(language,hld,tag{
         type: "language",
         subtag: "hld",
         description: "Halang Doan",
         added: "2009-07-29"}).
iana(language,hle,tag{
         type: "language",
         subtag: "hle",
         description: "Hlersu",
         added: "2009-07-29"}).
iana(language,hlt,tag{
         type: "language",
         subtag: "hlt",
         description: "Matu Chin",
         added: "2009-07-29"}).
iana(language,hlu,tag{
         type: "language",
         subtag: "hlu",
         description: "Hieroglyphic Luwian",
         added: "2009-07-29"}).
iana(language,hma,tag{
         type: "language",
         subtag: "hma",
         description: ["Southern Mashan Hmong","Southern Mashan Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmb,tag{
         type: "language",
         subtag: "hmb",
         description: "Humburi Senni Songhay",
         added: "2009-07-29"}).
iana(language,hmc,tag{
         type: "language",
         subtag: "hmc",
         description: ["Central Huishui Hmong","Central Huishui Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmd,tag{
         type: "language",
         subtag: "hmd",
         description: ["Large Flowery Miao","A-hmaos","Da-Hua Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hme,tag{
         type: "language",
         subtag: "hme",
         description: ["Eastern Huishui Hmong","Eastern Huishui Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmf,tag{
         type: "language",
         subtag: "hmf",
         description: "Hmong Don",
         added: "2009-07-29"}).
iana(language,hmg,tag{
         type: "language",
         subtag: "hmg",
         description: "Southwestern Guiyang Hmong",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmh,tag{
         type: "language",
         subtag: "hmh",
         description: ["Southwestern Huishui Hmong","Southwestern Huishui Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmi,tag{
         type: "language",
         subtag: "hmi",
         description: ["Northern Huishui Hmong","Northern Huishui Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmj,tag{
         type: "language",
         subtag: "hmj",
         description: ["Ge","Gejia"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmk,tag{
         type: "language",
         subtag: "hmk",
         description: "Maek",
         added: "2009-07-29"}).
iana(language,hml,tag{
         type: "language",
         subtag: "hml",
         description: ["Luopohe Hmong","Luopohe Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmm,tag{
         type: "language",
         subtag: "hmm",
         description: ["Central Mashan Hmong","Central Mashan Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmn,tag{
         type: "language",
         subtag: "hmn",
         description: ["Hmong","Mong"],
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,hmp,tag{
         type: "language",
         subtag: "hmp",
         description: ["Northern Mashan Hmong","Northern Mashan Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmq,tag{
         type: "language",
         subtag: "hmq",
         description: "Eastern Qiandong Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmr,tag{
         type: "language",
         subtag: "hmr",
         description: "Hmar",
         added: "2009-07-29"}).
iana(language,hms,tag{
         type: "language",
         subtag: "hms",
         description: "Southern Qiandong Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmt,tag{
         type: "language",
         subtag: "hmt",
         description: "Hamtai",
         added: "2009-07-29"}).
iana(language,hmu,tag{
         type: "language",
         subtag: "hmu",
         description: "Hamap",
         added: "2009-07-29"}).
iana(language,hmv,tag{
         type: "language",
         subtag: "hmv",
         description: "Hmong Dô",
         added: "2009-07-29"}).
iana(language,hmw,tag{
         type: "language",
         subtag: "hmw",
         description: ["Western Mashan Hmong","Western Mashan Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmx,tag{
         type: "language",
         subtag: "hmx",
         description: "Hmong-Mien languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,hmy,tag{
         type: "language",
         subtag: "hmy",
         description: ["Southern Guiyang Hmong","Southern Guiyang Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hmz,tag{
         type: "language",
         subtag: "hmz",
         description: ["Hmong Shua","Sinicized Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hna,tag{
         type: "language",
         subtag: "hna",
         description: "Mina (Cameroon)",
         added: "2009-07-29"}).
iana(language,hnd,tag{
         type: "language",
         subtag: "hnd",
         description: "Southern Hindko",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,hne,tag{
         type: "language",
         subtag: "hne",
         description: "Chhattisgarhi",
         added: "2009-07-29"}).
iana(language,hng,tag{
         type: "language",
         subtag: "hng",
         description: "Hungu",
         added: "2020-03-28"}).
iana(language,hnh,tag{
         type: "language",
         subtag: "hnh",
         description: "ǁAni",
         added: "2009-07-29"}).
iana(language,hni,tag{
         type: "language",
         subtag: "hni",
         description: "Hani",
         added: "2009-07-29"}).
iana(language,hnj,tag{
         type: "language",
         subtag: "hnj",
         description: ["Hmong Njua","Mong Leng","Mong Njua"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hnn,tag{
         type: "language",
         subtag: "hnn",
         description: "Hanunoo",
         added: "2009-07-29"}).
iana(language,hno,tag{
         type: "language",
         subtag: "hno",
         description: "Northern Hindko",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,hns,tag{
         type: "language",
         subtag: "hns",
         description: "Caribbean Hindustani",
         added: "2009-07-29"}).
iana(language,hnu,tag{
         type: "language",
         subtag: "hnu",
         description: "Hung",
         added: "2009-07-29"}).
iana(language,hoa,tag{
         type: "language",
         subtag: "hoa",
         description: "Hoava",
         added: "2009-07-29"}).
iana(language,hob,tag{
         type: "language",
         subtag: "hob",
         description: "Mari (Madang Province)",
         added: "2009-07-29"}).
iana(language,hoc,tag{
         type: "language",
         subtag: "hoc",
         description: "Ho",
         added: "2009-07-29"}).
iana(language,hod,tag{
         type: "language",
         subtag: "hod",
         description: "Holma",
         added: "2009-07-29"}).
iana(language,hoe,tag{
         type: "language",
         subtag: "hoe",
         description: "Horom",
         added: "2009-07-29"}).
iana(language,hoh,tag{
         type: "language",
         subtag: "hoh",
         description: "Hobyót",
         added: "2009-07-29"}).
iana(language,hoi,tag{
         type: "language",
         subtag: "hoi",
         description: "Holikachuk",
         added: "2009-07-29"}).
iana(language,hoj,tag{
         type: "language",
         subtag: "hoj",
         description: ["Hadothi","Haroti"],
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,hok,tag{
         type: "language",
         subtag: "hok",
         description: "Hokan languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,hol,tag{
         type: "language",
         subtag: "hol",
         description: "Holu",
         added: "2009-07-29"}).
iana(language,hom,tag{
         type: "language",
         subtag: "hom",
         description: "Homa",
         added: "2009-07-29"}).
iana(language,hoo,tag{
         type: "language",
         subtag: "hoo",
         description: "Holoholo",
         added: "2009-07-29"}).
iana(language,hop,tag{
         type: "language",
         subtag: "hop",
         description: "Hopi",
         added: "2009-07-29"}).
iana(language,hor,tag{
         type: "language",
         subtag: "hor",
         description: "Horo",
         added: "2009-07-29"}).
iana(language,hos,tag{
         type: "language",
         subtag: "hos",
         description: "Ho Chi Minh City Sign Language",
         added: "2009-07-29"}).
iana(language,hot,tag{
         type: "language",
         subtag: "hot",
         description: ["Hote","Malê"],
         added: "2009-07-29"}).
iana(language,hov,tag{
         type: "language",
         subtag: "hov",
         description: "Hovongan",
         added: "2009-07-29"}).
iana(language,how,tag{
         type: "language",
         subtag: "how",
         description: "Honi",
         added: "2009-07-29"}).
iana(language,hoy,tag{
         type: "language",
         subtag: "hoy",
         description: "Holiya",
         added: "2009-07-29"}).
iana(language,hoz,tag{
         type: "language",
         subtag: "hoz",
         description: "Hozo",
         added: "2009-07-29"}).
iana(language,hpo,tag{
         type: "language",
         subtag: "hpo",
         description: "Hpon",
         added: "2009-07-29"}).
iana(language,hps,tag{
         type: "language",
         subtag: "hps",
         description: ["Hawai'i Sign Language (HSL)","Hawai'i Pidgin Sign Language"],
         added: "2009-07-29"}).
iana(language,hra,tag{
         type: "language",
         subtag: "hra",
         description: "Hrangkhol",
         added: "2009-07-29"}).
iana(language,hrc,tag{
         type: "language",
         subtag: "hrc",
         description: "Niwer Mil",
         added: "2013-09-10"}).
iana(language,hre,tag{
         type: "language",
         subtag: "hre",
         description: "Hre",
         added: "2009-07-29"}).
iana(language,hrk,tag{
         type: "language",
         subtag: "hrk",
         description: "Haruku",
         added: "2009-07-29"}).
iana(language,hrm,tag{
         type: "language",
         subtag: "hrm",
         description: "Horned Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,hro,tag{
         type: "language",
         subtag: "hro",
         description: "Haroi",
         added: "2009-07-29"}).
iana(language,hrp,tag{
         type: "language",
         subtag: "hrp",
         description: "Nhirrpi",
         added: "2013-09-10"}).
iana(language,hrr,tag{
         type: "language",
         subtag: "hrr",
         description: "Horuru",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "jal"}).
iana(language,hrt,tag{
         type: "language",
         subtag: "hrt",
         description: "Hértevin",
         added: "2009-07-29"}).
iana(language,hru,tag{
         type: "language",
         subtag: "hru",
         description: "Hruso",
         added: "2009-07-29"}).
iana(language,hrw,tag{
         type: "language",
         subtag: "hrw",
         description: "Warwar Feni",
         added: "2013-09-10"}).
iana(language,hrx,tag{
         type: "language",
         subtag: "hrx",
         description: "Hunsrik",
         added: "2009-07-29"}).
iana(language,hrz,tag{
         type: "language",
         subtag: "hrz",
         description: "Harzani",
         added: "2009-07-29"}).
iana(language,hsb,tag{
         type: "language",
         subtag: "hsb",
         description: "Upper Sorbian",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,hsh,tag{
         type: "language",
         subtag: "hsh",
         description: "Hungarian Sign Language",
         added: "2009-07-29"}).
iana(language,hsl,tag{
         type: "language",
         subtag: "hsl",
         description: "Hausa Sign Language",
         added: "2009-07-29"}).
iana(language,hsn,tag{
         type: "language",
         subtag: "hsn",
         description: "Xiang Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,hss,tag{
         type: "language",
         subtag: "hss",
         description: "Harsusi",
         added: "2009-07-29"}).
iana(language,hti,tag{
         type: "language",
         subtag: "hti",
         description: "Hoti",
         added: "2009-07-29"}).
iana(language,hto,tag{
         type: "language",
         subtag: "hto",
         description: "Minica Huitoto",
         added: "2009-07-29"}).
iana(language,hts,tag{
         type: "language",
         subtag: "hts",
         description: "Hadza",
         added: "2009-07-29"}).
iana(language,htu,tag{
         type: "language",
         subtag: "htu",
         description: "Hitu",
         added: "2009-07-29"}).
iana(language,htx,tag{
         type: "language",
         subtag: "htx",
         description: "Middle Hittite",
         added: "2009-07-29"}).
iana(language,hub,tag{
         type: "language",
         subtag: "hub",
         description: "Huambisa",
         added: "2009-07-29"}).
iana(language,huc,tag{
         type: "language",
         subtag: "huc",
         description: ["ǂHua","ǂʼAmkhoe"],
         added: "2009-07-29"}).
iana(language,hud,tag{
         type: "language",
         subtag: "hud",
         description: "Huaulu",
         added: "2009-07-29"}).
iana(language,hue,tag{
         type: "language",
         subtag: "hue",
         description: "San Francisco Del Mar Huave",
         added: "2009-07-29"}).
iana(language,huf,tag{
         type: "language",
         subtag: "huf",
         description: "Humene",
         added: "2009-07-29"}).
iana(language,hug,tag{
         type: "language",
         subtag: "hug",
         description: "Huachipaeri",
         added: "2009-07-29"}).
iana(language,huh,tag{
         type: "language",
         subtag: "huh",
         description: "Huilliche",
         added: "2009-07-29"}).
iana(language,hui,tag{
         type: "language",
         subtag: "hui",
         description: "Huli",
         added: "2009-07-29"}).
iana(language,huj,tag{
         type: "language",
         subtag: "huj",
         description: ["Northern Guiyang Hmong","Northern Guiyang Miao"],
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,huk,tag{
         type: "language",
         subtag: "huk",
         description: "Hulung",
         added: "2009-07-29"}).
iana(language,hul,tag{
         type: "language",
         subtag: "hul",
         description: "Hula",
         added: "2009-07-29"}).
iana(language,hum,tag{
         type: "language",
         subtag: "hum",
         description: "Hungana",
         added: "2009-07-29"}).
iana(language,huo,tag{
         type: "language",
         subtag: "huo",
         description: "Hu",
         added: "2009-07-29"}).
iana(language,hup,tag{
         type: "language",
         subtag: "hup",
         description: "Hupa",
         added: "2005-10-16"}).
iana(language,huq,tag{
         type: "language",
         subtag: "huq",
         description: "Tsat",
         added: "2009-07-29"}).
iana(language,hur,tag{
         type: "language",
         subtag: "hur",
         description: "Halkomelem",
         added: "2009-07-29"}).
iana(language,hus,tag{
         type: "language",
         subtag: "hus",
         description: "Huastec",
         added: "2009-07-29"}).
iana(language,hut,tag{
         type: "language",
         subtag: "hut",
         description: "Humla",
         added: "2009-07-29"}).
iana(language,huu,tag{
         type: "language",
         subtag: "huu",
         description: "Murui Huitoto",
         added: "2009-07-29"}).
iana(language,huv,tag{
         type: "language",
         subtag: "huv",
         description: "San Mateo Del Mar Huave",
         added: "2009-07-29"}).
iana(language,huw,tag{
         type: "language",
         subtag: "huw",
         description: "Hukumina",
         added: "2009-07-29"}).
iana(language,hux,tag{
         type: "language",
         subtag: "hux",
         description: "Nüpode Huitoto",
         added: "2009-07-29"}).
iana(language,huy,tag{
         type: "language",
         subtag: "huy",
         description: "Hulaulá",
         added: "2009-07-29"}).
iana(language,huz,tag{
         type: "language",
         subtag: "huz",
         description: "Hunzib",
         added: "2009-07-29"}).
iana(language,hvc,tag{
         type: "language",
         subtag: "hvc",
         description: "Haitian Vodoun Culture Language",
         added: "2009-07-29"}).
iana(language,hve,tag{
         type: "language",
         subtag: "hve",
         description: "San Dionisio Del Mar Huave",
         added: "2009-07-29"}).
iana(language,hvk,tag{
         type: "language",
         subtag: "hvk",
         description: "Haveke",
         added: "2009-07-29"}).
iana(language,hvn,tag{
         type: "language",
         subtag: "hvn",
         description: "Sabu",
         added: "2009-07-29"}).
iana(language,hvv,tag{
         type: "language",
         subtag: "hvv",
         description: "Santa María Del Mar Huave",
         added: "2009-07-29"}).
iana(language,hwa,tag{
         type: "language",
         subtag: "hwa",
         description: "Wané",
         added: "2009-07-29"}).
iana(language,hwc,tag{
         type: "language",
         subtag: "hwc",
         description: ["Hawai'i Creole English","Hawai'i Pidgin"],
         added: "2009-07-29"}).
iana(language,hwo,tag{
         type: "language",
         subtag: "hwo",
         description: "Hwana",
         added: "2009-07-29"}).
iana(language,hya,tag{
         type: "language",
         subtag: "hya",
         description: "Hya",
         added: "2009-07-29"}).
iana(language,hyw,tag{
         type: "language",
         subtag: "hyw",
         description: "Western Armenian",
         added: "2018-03-08",
         comments: "see also hy"}).
iana(language,hyx,tag{
         type: "language",
         subtag: "hyx",
         description: "Armenian (family)",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,iai,tag{
         type: "language",
         subtag: "iai",
         description: "Iaai",
         added: "2009-07-29"}).
iana(language,ian,tag{
         type: "language",
         subtag: "ian",
         description: "Iatmul",
         added: "2009-07-29"}).
iana(language,iap,tag{
         type: "language",
         subtag: "iap",
         description: "Iapama",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,iar,tag{
         type: "language",
         subtag: "iar",
         description: "Purari",
         added: "2009-07-29"}).
iana(language,iba,tag{
         type: "language",
         subtag: "iba",
         description: "Iban",
         added: "2005-10-16"}).
iana(language,ibb,tag{
         type: "language",
         subtag: "ibb",
         description: "Ibibio",
         added: "2009-07-29"}).
iana(language,ibd,tag{
         type: "language",
         subtag: "ibd",
         description: "Iwaidja",
         added: "2009-07-29"}).
iana(language,ibe,tag{
         type: "language",
         subtag: "ibe",
         description: "Akpes",
         added: "2009-07-29"}).
iana(language,ibg,tag{
         type: "language",
         subtag: "ibg",
         description: "Ibanag",
         added: "2009-07-29"}).
iana(language,ibh,tag{
         type: "language",
         subtag: "ibh",
         description: "Bih",
         added: "2017-02-23"}).
iana(language,ibi,tag{
         type: "language",
         subtag: "ibi",
         description: "Ibilo",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "opa"}).
iana(language,ibl,tag{
         type: "language",
         subtag: "ibl",
         description: "Ibaloi",
         added: "2009-07-29"}).
iana(language,ibm,tag{
         type: "language",
         subtag: "ibm",
         description: "Agoi",
         added: "2009-07-29"}).
iana(language,ibn,tag{
         type: "language",
         subtag: "ibn",
         description: "Ibino",
         added: "2009-07-29"}).
iana(language,ibr,tag{
         type: "language",
         subtag: "ibr",
         description: "Ibuoro",
         added: "2009-07-29"}).
iana(language,ibu,tag{
         type: "language",
         subtag: "ibu",
         description: "Ibu",
         added: "2009-07-29"}).
iana(language,iby,tag{
         type: "language",
         subtag: "iby",
         description: "Ibani",
         added: "2009-07-29"}).
iana(language,ica,tag{
         type: "language",
         subtag: "ica",
         description: "Ede Ica",
         added: "2009-07-29"}).
iana(language,ich,tag{
         type: "language",
         subtag: "ich",
         description: "Etkywan",
         added: "2009-07-29"}).
iana(language,icl,tag{
         type: "language",
         subtag: "icl",
         description: "Icelandic Sign Language",
         added: "2009-07-29"}).
iana(language,icr,tag{
         type: "language",
         subtag: "icr",
         description: "Islander Creole English",
         added: "2009-07-29"}).
iana(language,ida,tag{
         type: "language",
         subtag: "ida",
         description: ["Idakho-Isukha-Tiriki","Luidakho-Luisukha-Lutirichi"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,idb,tag{
         type: "language",
         subtag: "idb",
         description: "Indo-Portuguese",
         added: "2009-07-29"}).
iana(language,idc,tag{
         type: "language",
         subtag: "idc",
         description: ["Idon","Ajiya"],
         added: "2009-07-29"}).
iana(language,idd,tag{
         type: "language",
         subtag: "idd",
         description: "Ede Idaca",
         added: "2009-07-29"}).
iana(language,ide,tag{
         type: "language",
         subtag: "ide",
         description: "Idere",
         added: "2009-07-29"}).
iana(language,idi,tag{
         type: "language",
         subtag: "idi",
         description: "Idi",
         added: "2009-07-29"}).
iana(language,idr,tag{
         type: "language",
         subtag: "idr",
         description: "Indri",
         added: "2009-07-29"}).
iana(language,ids,tag{
         type: "language",
         subtag: "ids",
         description: "Idesa",
         added: "2009-07-29"}).
iana(language,idt,tag{
         type: "language",
         subtag: "idt",
         description: "Idaté",
         added: "2009-07-29"}).
iana(language,idu,tag{
         type: "language",
         subtag: "idu",
         description: "Idoma",
         added: "2009-07-29"}).
iana(language,ifa,tag{
         type: "language",
         subtag: "ifa",
         description: "Amganad Ifugao",
         added: "2009-07-29"}).
iana(language,ifb,tag{
         type: "language",
         subtag: "ifb",
         description: ["Batad Ifugao","Ayangan Ifugao"],
         added: "2009-07-29"}).
iana(language,ife,tag{
         type: "language",
         subtag: "ife",
         description: "Ifè",
         added: "2009-07-29"}).
iana(language,iff,tag{
         type: "language",
         subtag: "iff",
         description: "Ifo",
         added: "2009-07-29"}).
iana(language,ifk,tag{
         type: "language",
         subtag: "ifk",
         description: "Tuwali Ifugao",
         added: "2009-07-29"}).
iana(language,ifm,tag{
         type: "language",
         subtag: "ifm",
         description: "Teke-Fuumu",
         added: "2009-07-29"}).
iana(language,ifu,tag{
         type: "language",
         subtag: "ifu",
         description: "Mayoyao Ifugao",
         added: "2009-07-29"}).
iana(language,ify,tag{
         type: "language",
         subtag: "ify",
         description: "Keley-I Kallahan",
         added: "2009-07-29"}).
iana(language,igb,tag{
         type: "language",
         subtag: "igb",
         description: "Ebira",
         added: "2009-07-29"}).
iana(language,ige,tag{
         type: "language",
         subtag: "ige",
         description: "Igede",
         added: "2009-07-29"}).
iana(language,igg,tag{
         type: "language",
         subtag: "igg",
         description: "Igana",
         added: "2009-07-29"}).
iana(language,igl,tag{
         type: "language",
         subtag: "igl",
         description: "Igala",
         added: "2009-07-29"}).
iana(language,igm,tag{
         type: "language",
         subtag: "igm",
         description: "Kanggape",
         added: "2009-07-29"}).
iana(language,ign,tag{
         type: "language",
         subtag: "ign",
         description: "Ignaciano",
         added: "2009-07-29"}).
iana(language,igo,tag{
         type: "language",
         subtag: "igo",
         description: "Isebe",
         added: "2009-07-29"}).
iana(language,igs,tag{
         type: "language",
         subtag: "igs",
         description: "Interglossa",
         added: "2009-07-29"}).
iana(language,igw,tag{
         type: "language",
         subtag: "igw",
         description: "Igwe",
         added: "2009-07-29"}).
iana(language,ihb,tag{
         type: "language",
         subtag: "ihb",
         description: "Iha Based Pidgin",
         added: "2009-07-29"}).
iana(language,ihi,tag{
         type: "language",
         subtag: "ihi",
         description: "Ihievbe",
         added: "2009-07-29"}).
iana(language,ihp,tag{
         type: "language",
         subtag: "ihp",
         description: "Iha",
         added: "2009-07-29"}).
iana(language,ihw,tag{
         type: "language",
         subtag: "ihw",
         description: "Bidhawal",
         added: "2012-08-12"}).
iana(language,iin,tag{
         type: "language",
         subtag: "iin",
         description: "Thiin",
         added: "2013-09-10"}).
iana(language,iir,tag{
         type: "language",
         subtag: "iir",
         description: "Indo-Iranian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,ijc,tag{
         type: "language",
         subtag: "ijc",
         description: "Izon",
         added: "2009-07-29"}).
iana(language,ije,tag{
         type: "language",
         subtag: "ije",
         description: "Biseni",
         added: "2009-07-29"}).
iana(language,ijj,tag{
         type: "language",
         subtag: "ijj",
         description: "Ede Ije",
         added: "2009-07-29"}).
iana(language,ijn,tag{
         type: "language",
         subtag: "ijn",
         description: "Kalabari",
         added: "2009-07-29"}).
iana(language,ijo,tag{
         type: "language",
         subtag: "ijo",
         description: "Ijo languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ijs,tag{
         type: "language",
         subtag: "ijs",
         description: "Southeast Ijo",
         added: "2009-07-29"}).
iana(language,ike,tag{
         type: "language",
         subtag: "ike",
         description: "Eastern Canadian Inuktitut",
         added: "2009-07-29",
         macrolanguage: "iu"}).
iana(language,iki,tag{
         type: "language",
         subtag: "iki",
         description: "Iko",
         added: "2009-07-29"}).
iana(language,ikk,tag{
         type: "language",
         subtag: "ikk",
         description: "Ika",
         added: "2009-07-29"}).
iana(language,ikl,tag{
         type: "language",
         subtag: "ikl",
         description: "Ikulu",
         added: "2009-07-29"}).
iana(language,iko,tag{
         type: "language",
         subtag: "iko",
         description: "Olulumo-Ikom",
         added: "2009-07-29"}).
iana(language,ikp,tag{
         type: "language",
         subtag: "ikp",
         description: "Ikpeshi",
         added: "2009-07-29"}).
iana(language,ikr,tag{
         type: "language",
         subtag: "ikr",
         description: "Ikaranggal",
         added: "2013-09-10"}).
iana(language,iks,tag{
         type: "language",
         subtag: "iks",
         description: "Inuit Sign Language",
         added: "2015-02-12"}).
iana(language,ikt,tag{
         type: "language",
         subtag: "ikt",
         description: ["Inuinnaqtun","Western Canadian Inuktitut"],
         added: "2009-07-29",
         macrolanguage: "iu"}).
iana(language,ikv,tag{
         type: "language",
         subtag: "ikv",
         description: "Iku-Gora-Ankwa",
         added: "2009-07-29"}).
iana(language,ikw,tag{
         type: "language",
         subtag: "ikw",
         description: "Ikwere",
         added: "2009-07-29"}).
iana(language,ikx,tag{
         type: "language",
         subtag: "ikx",
         description: "Ik",
         added: "2009-07-29"}).
iana(language,ikz,tag{
         type: "language",
         subtag: "ikz",
         description: "Ikizu",
         added: "2009-07-29"}).
iana(language,ila,tag{
         type: "language",
         subtag: "ila",
         description: "Ile Ape",
         added: "2009-07-29"}).
iana(language,ilb,tag{
         type: "language",
         subtag: "ilb",
         description: "Ila",
         added: "2009-07-29"}).
iana(language,ilg,tag{
         type: "language",
         subtag: "ilg",
         description: "Garig-Ilgar",
         added: "2009-07-29"}).
iana(language,ili,tag{
         type: "language",
         subtag: "ili",
         description: "Ili Turki",
         added: "2009-07-29"}).
iana(language,ilk,tag{
         type: "language",
         subtag: "ilk",
         description: "Ilongot",
         added: "2009-07-29"}).
iana(language,ill,tag{
         type: "language",
         subtag: "ill",
         description: "Iranun",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         comments: "see ilm, ilp"}).
iana(language,ilm,tag{
         type: "language",
         subtag: "ilm",
         description: "Iranun (Malaysia)",
         added: "2016-05-30"}).
iana(language,ilo,tag{
         type: "language",
         subtag: "ilo",
         description: "Iloko",
         added: "2005-10-16"}).
iana(language,ilp,tag{
         type: "language",
         subtag: "ilp",
         description: "Iranun (Philippines)",
         added: "2016-05-30"}).
iana(language,ils,tag{
         type: "language",
         subtag: "ils",
         description: "International Sign",
         added: "2009-07-29"}).
iana(language,ilu,tag{
         type: "language",
         subtag: "ilu",
         description: "Ili'uun",
         added: "2009-07-29"}).
iana(language,ilv,tag{
         type: "language",
         subtag: "ilv",
         description: "Ilue",
         added: "2009-07-29"}).
iana(language,ilw,tag{
         type: "language",
         subtag: "ilw",
         description: "Talur",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "gal"}).
iana(language,ima,tag{
         type: "language",
         subtag: "ima",
         description: "Mala Malasar",
         added: "2009-07-29"}).
iana(language,ime,tag{
         type: "language",
         subtag: "ime",
         description: "Imeraguen",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,imi,tag{
         type: "language",
         subtag: "imi",
         description: "Anamgura",
         added: "2009-07-29"}).
iana(language,iml,tag{
         type: "language",
         subtag: "iml",
         description: "Miluk",
         added: "2009-07-29"}).
iana(language,imn,tag{
         type: "language",
         subtag: "imn",
         description: "Imonda",
         added: "2009-07-29"}).
iana(language,imo,tag{
         type: "language",
         subtag: "imo",
         description: "Imbongu",
         added: "2009-07-29"}).
iana(language,imr,tag{
         type: "language",
         subtag: "imr",
         description: "Imroing",
         added: "2009-07-29"}).
iana(language,ims,tag{
         type: "language",
         subtag: "ims",
         description: "Marsian",
         added: "2009-07-29"}).
iana(language,imt,tag{
         type: "language",
         subtag: "imt",
         description: "Imotong",
         added: "2022-02-25"}).
iana(language,imy,tag{
         type: "language",
         subtag: "imy",
         description: "Milyan",
         added: "2009-07-29"}).
iana(language,inb,tag{
         type: "language",
         subtag: "inb",
         description: "Inga",
         added: "2009-07-29"}).
iana(language,inc,tag{
         type: "language",
         subtag: "inc",
         description: "Indic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ine,tag{
         type: "language",
         subtag: "ine",
         description: "Indo-European languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ing,tag{
         type: "language",
         subtag: "ing",
         description: "Degexit'an",
         added: "2009-07-29"}).
iana(language,inh,tag{
         type: "language",
         subtag: "inh",
         description: "Ingush",
         added: "2005-10-16"}).
iana(language,inj,tag{
         type: "language",
         subtag: "inj",
         description: "Jungle Inga",
         added: "2009-07-29"}).
iana(language,inl,tag{
         type: "language",
         subtag: "inl",
         description: "Indonesian Sign Language",
         added: "2009-07-29"}).
iana(language,inm,tag{
         type: "language",
         subtag: "inm",
         description: "Minaean",
         added: "2009-07-29"}).
iana(language,inn,tag{
         type: "language",
         subtag: "inn",
         description: "Isinai",
         added: "2009-07-29"}).
iana(language,ino,tag{
         type: "language",
         subtag: "ino",
         description: "Inoke-Yate",
         added: "2009-07-29"}).
iana(language,inp,tag{
         type: "language",
         subtag: "inp",
         description: "Iñapari",
         added: "2009-07-29"}).
iana(language,ins,tag{
         type: "language",
         subtag: "ins",
         description: "Indian Sign Language",
         added: "2009-07-29"}).
iana(language,int,tag{
         type: "language",
         subtag: "int",
         description: "Intha",
         added: "2009-07-29"}).
iana(language,inz,tag{
         type: "language",
         subtag: "inz",
         description: "Ineseño",
         added: "2009-07-29"}).
iana(language,ior,tag{
         type: "language",
         subtag: "ior",
         description: "Inor",
         added: "2009-07-29"}).
iana(language,iou,tag{
         type: "language",
         subtag: "iou",
         description: "Tuma-Irumu",
         added: "2009-07-29"}).
iana(language,iow,tag{
         type: "language",
         subtag: "iow",
         description: "Iowa-Oto",
         added: "2009-07-29"}).
iana(language,ipi,tag{
         type: "language",
         subtag: "ipi",
         description: "Ipili",
         added: "2009-07-29"}).
iana(language,ipo,tag{
         type: "language",
         subtag: "ipo",
         description: "Ipiko",
         added: "2009-07-29"}).
iana(language,iqu,tag{
         type: "language",
         subtag: "iqu",
         description: "Iquito",
         added: "2009-07-29"}).
iana(language,iqw,tag{
         type: "language",
         subtag: "iqw",
         description: "Ikwo",
         added: "2013-09-10"}).
iana(language,ira,tag{
         type: "language",
         subtag: "ira",
         description: "Iranian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ire,tag{
         type: "language",
         subtag: "ire",
         description: "Iresim",
         added: "2009-07-29"}).
iana(language,irh,tag{
         type: "language",
         subtag: "irh",
         description: "Irarutu",
         added: "2009-07-29"}).
iana(language,iri,tag{
         type: "language",
         subtag: "iri",
         description: ["Rigwe","Irigwe"],
         added: "2009-07-29"}).
iana(language,irk,tag{
         type: "language",
         subtag: "irk",
         description: "Iraqw",
         added: "2009-07-29"}).
iana(language,irn,tag{
         type: "language",
         subtag: "irn",
         description: "Irántxe",
         added: "2009-07-29"}).
iana(language,iro,tag{
         type: "language",
         subtag: "iro",
         description: "Iroquoian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,irr,tag{
         type: "language",
         subtag: "irr",
         description: "Ir",
         added: "2009-07-29"}).
iana(language,iru,tag{
         type: "language",
         subtag: "iru",
         description: "Irula",
         added: "2009-07-29"}).
iana(language,irx,tag{
         type: "language",
         subtag: "irx",
         description: "Kamberau",
         added: "2009-07-29"}).
iana(language,iry,tag{
         type: "language",
         subtag: "iry",
         description: "Iraya",
         added: "2009-07-29"}).
iana(language,isa,tag{
         type: "language",
         subtag: "isa",
         description: "Isabi",
         added: "2009-07-29"}).
iana(language,isc,tag{
         type: "language",
         subtag: "isc",
         description: "Isconahua",
         added: "2009-07-29"}).
iana(language,isd,tag{
         type: "language",
         subtag: "isd",
         description: "Isnag",
         added: "2009-07-29"}).
iana(language,ise,tag{
         type: "language",
         subtag: "ise",
         description: "Italian Sign Language",
         added: "2009-07-29"}).
iana(language,isg,tag{
         type: "language",
         subtag: "isg",
         description: "Irish Sign Language",
         added: "2009-07-29"}).
iana(language,ish,tag{
         type: "language",
         subtag: "ish",
         description: "Esan",
         added: "2009-07-29"}).
iana(language,isi,tag{
         type: "language",
         subtag: "isi",
         description: "Nkem-Nkum",
         added: "2009-07-29"}).
iana(language,isk,tag{
         type: "language",
         subtag: "isk",
         description: "Ishkashimi",
         added: "2010-03-11"}).
iana(language,ism,tag{
         type: "language",
         subtag: "ism",
         description: "Masimasi",
         added: "2009-07-29"}).
iana(language,isn,tag{
         type: "language",
         subtag: "isn",
         description: "Isanzu",
         added: "2009-07-29"}).
iana(language,iso,tag{
         type: "language",
         subtag: "iso",
         description: "Isoko",
         added: "2009-07-29"}).
iana(language,isr,tag{
         type: "language",
         subtag: "isr",
         description: "Israeli Sign Language",
         added: "2009-07-29"}).
iana(language,ist,tag{
         type: "language",
         subtag: "ist",
         description: "Istriot",
         added: "2009-07-29"}).
iana(language,isu,tag{
         type: "language",
         subtag: "isu",
         description: "Isu (Menchum Division)",
         added: "2009-07-29"}).
iana(language,itb,tag{
         type: "language",
         subtag: "itb",
         description: "Binongan Itneg",
         added: "2009-07-29"}).
iana(language,itc,tag{
         type: "language",
         subtag: "itc",
         description: "Italic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,itd,tag{
         type: "language",
         subtag: "itd",
         description: "Southern Tidung",
         added: "2016-05-30"}).
iana(language,ite,tag{
         type: "language",
         subtag: "ite",
         description: "Itene",
         added: "2009-07-29"}).
iana(language,iti,tag{
         type: "language",
         subtag: "iti",
         description: "Inlaod Itneg",
         added: "2009-07-29"}).
iana(language,itk,tag{
         type: "language",
         subtag: "itk",
         description: "Judeo-Italian",
         added: "2009-07-29"}).
iana(language,itl,tag{
         type: "language",
         subtag: "itl",
         description: "Itelmen",
         added: "2009-07-29"}).
iana(language,itm,tag{
         type: "language",
         subtag: "itm",
         description: "Itu Mbon Uzo",
         added: "2009-07-29"}).
iana(language,ito,tag{
         type: "language",
         subtag: "ito",
         description: "Itonama",
         added: "2009-07-29"}).
iana(language,itr,tag{
         type: "language",
         subtag: "itr",
         description: "Iteri",
         added: "2009-07-29"}).
iana(language,its,tag{
         type: "language",
         subtag: "its",
         description: "Isekiri",
         added: "2009-07-29"}).
iana(language,itt,tag{
         type: "language",
         subtag: "itt",
         description: "Maeng Itneg",
         added: "2009-07-29"}).
iana(language,itv,tag{
         type: "language",
         subtag: "itv",
         description: "Itawit",
         added: "2009-07-29"}).
iana(language,itw,tag{
         type: "language",
         subtag: "itw",
         description: "Ito",
         added: "2009-07-29"}).
iana(language,itx,tag{
         type: "language",
         subtag: "itx",
         description: "Itik",
         added: "2009-07-29"}).
iana(language,ity,tag{
         type: "language",
         subtag: "ity",
         description: "Moyadan Itneg",
         added: "2009-07-29"}).
iana(language,itz,tag{
         type: "language",
         subtag: "itz",
         description: "Itzá",
         added: "2009-07-29"}).
iana(language,ium,tag{
         type: "language",
         subtag: "ium",
         description: "Iu Mien",
         added: "2009-07-29"}).
iana(language,ivb,tag{
         type: "language",
         subtag: "ivb",
         description: "Ibatan",
         added: "2009-07-29"}).
iana(language,ivv,tag{
         type: "language",
         subtag: "ivv",
         description: "Ivatan",
         added: "2009-07-29"}).
iana(language,iwk,tag{
         type: "language",
         subtag: "iwk",
         description: "I-Wak",
         added: "2009-07-29"}).
iana(language,iwm,tag{
         type: "language",
         subtag: "iwm",
         description: "Iwam",
         added: "2009-07-29"}).
iana(language,iwo,tag{
         type: "language",
         subtag: "iwo",
         description: "Iwur",
         added: "2009-07-29"}).
iana(language,iws,tag{
         type: "language",
         subtag: "iws",
         description: "Sepik Iwam",
         added: "2009-07-29"}).
iana(language,ixc,tag{
         type: "language",
         subtag: "ixc",
         description: "Ixcatec",
         added: "2009-07-29"}).
iana(language,ixl,tag{
         type: "language",
         subtag: "ixl",
         description: "Ixil",
         added: "2009-07-29"}).
iana(language,iya,tag{
         type: "language",
         subtag: "iya",
         description: "Iyayu",
         added: "2009-07-29"}).
iana(language,iyo,tag{
         type: "language",
         subtag: "iyo",
         description: "Mesaka",
         added: "2009-07-29"}).
iana(language,iyx,tag{
         type: "language",
         subtag: "iyx",
         description: "Yaka (Congo)",
         added: "2009-07-29"}).
iana(language,izh,tag{
         type: "language",
         subtag: "izh",
         description: "Ingrian",
         added: "2009-07-29"}).
iana(language,izi,tag{
         type: "language",
         subtag: "izi",
         description: "Izi-Ezaa-Ikwo-Mgbo",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see eza, gmz, iqw, izz"}).
iana(language,izr,tag{
         type: "language",
         subtag: "izr",
         description: "Izere",
         added: "2009-07-29"}).
iana(language,izz,tag{
         type: "language",
         subtag: "izz",
         description: "Izii",
         added: "2013-09-10"}).
iana(language,jaa,tag{
         type: "language",
         subtag: "jaa",
         description: "Jamamadí",
         added: "2009-07-29"}).
iana(language,jab,tag{
         type: "language",
         subtag: "jab",
         description: "Hyam",
         added: "2009-07-29"}).
iana(language,jac,tag{
         type: "language",
         subtag: "jac",
         description: ["Popti'","Jakalteko"],
         added: "2009-07-29"}).
iana(language,jad,tag{
         type: "language",
         subtag: "jad",
         description: "Jahanka",
         added: "2009-07-29"}).
iana(language,jae,tag{
         type: "language",
         subtag: "jae",
         description: "Yabem",
         added: "2009-07-29"}).
iana(language,jaf,tag{
         type: "language",
         subtag: "jaf",
         description: "Jara",
         added: "2009-07-29"}).
iana(language,jah,tag{
         type: "language",
         subtag: "jah",
         description: "Jah Hut",
         added: "2009-07-29"}).
iana(language,jaj,tag{
         type: "language",
         subtag: "jaj",
         description: "Zazao",
         added: "2009-07-29"}).
iana(language,jak,tag{
         type: "language",
         subtag: "jak",
         description: "Jakun",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,jal,tag{
         type: "language",
         subtag: "jal",
         description: "Yalahatan",
         added: "2009-07-29"}).
iana(language,jam,tag{
         type: "language",
         subtag: "jam",
         description: "Jamaican Creole English",
         added: "2009-07-29"}).
iana(language,jan,tag{
         type: "language",
         subtag: "jan",
         description: "Jandai",
         added: "2012-08-12"}).
iana(language,jao,tag{
         type: "language",
         subtag: "jao",
         description: "Yanyuwa",
         added: "2009-07-29"}).
iana(language,jaq,tag{
         type: "language",
         subtag: "jaq",
         description: "Yaqay",
         added: "2009-07-29"}).
iana(language,jar,tag{
         type: "language",
         subtag: "jar",
         description: "Jarawa (Nigeria)",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see jgk, jjr"}).
iana(language,jas,tag{
         type: "language",
         subtag: "jas",
         description: "New Caledonian Javanese",
         added: "2009-07-29"}).
iana(language,jat,tag{
         type: "language",
         subtag: "jat",
         description: "Jakati",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,jau,tag{
         type: "language",
         subtag: "jau",
         description: "Yaur",
         added: "2009-07-29"}).
iana(language,jax,tag{
         type: "language",
         subtag: "jax",
         description: "Jambi Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,jay,tag{
         type: "language",
         subtag: "jay",
         description: ["Yan-nhangu","Nhangu"],
         added: "2009-07-29"}).
iana(language,jaz,tag{
         type: "language",
         subtag: "jaz",
         description: "Jawe",
         added: "2009-07-29"}).
iana(language,jbe,tag{
         type: "language",
         subtag: "jbe",
         description: "Judeo-Berber",
         added: "2009-07-29"}).
iana(language,jbi,tag{
         type: "language",
         subtag: "jbi",
         description: "Badjiri",
         added: "2013-09-10"}).
iana(language,jbj,tag{
         type: "language",
         subtag: "jbj",
         description: "Arandai",
         added: "2009-07-29"}).
iana(language,jbk,tag{
         type: "language",
         subtag: "jbk",
         description: "Barikewa",
         added: "2012-08-12"}).
iana(language,jbm,tag{
         type: "language",
         subtag: "jbm",
         description: "Bijim",
         added: "2021-02-20"}).
iana(language,jbn,tag{
         type: "language",
         subtag: "jbn",
         description: "Nafusi",
         added: "2009-07-29"}).
iana(language,jbo,tag{
         type: "language",
         subtag: "jbo",
         description: "Lojban",
         added: "2005-10-16"}).
iana(language,jbr,tag{
         type: "language",
         subtag: "jbr",
         description: "Jofotek-Bromnya",
         added: "2009-07-29"}).
iana(language,jbt,tag{
         type: "language",
         subtag: "jbt",
         description: "Jabutí",
         added: "2009-07-29"}).
iana(language,jbu,tag{
         type: "language",
         subtag: "jbu",
         description: "Jukun Takum",
         added: "2009-07-29"}).
iana(language,jbw,tag{
         type: "language",
         subtag: "jbw",
         description: "Yawijibaya",
         added: "2012-08-12"}).
iana(language,jcs,tag{
         type: "language",
         subtag: "jcs",
         description: "Jamaican Country Sign Language",
         added: "2009-07-29"}).
iana(language,jct,tag{
         type: "language",
         subtag: "jct",
         description: "Krymchak",
         added: "2009-07-29"}).
iana(language,jda,tag{
         type: "language",
         subtag: "jda",
         description: "Jad",
         added: "2009-07-29"}).
iana(language,jdg,tag{
         type: "language",
         subtag: "jdg",
         description: "Jadgali",
         added: "2009-07-29"}).
iana(language,jdt,tag{
         type: "language",
         subtag: "jdt",
         description: "Judeo-Tat",
         added: "2009-07-29"}).
iana(language,jeb,tag{
         type: "language",
         subtag: "jeb",
         description: "Jebero",
         added: "2009-07-29"}).
iana(language,jee,tag{
         type: "language",
         subtag: "jee",
         description: "Jerung",
         added: "2009-07-29"}).
iana(language,jeg,tag{
         type: "language",
         subtag: "jeg",
         description: "Jeng",
         added: "2009-07-29",
         deprecated: "2017-02-23",
         preferred_value: "oyb"}).
iana(language,jeh,tag{
         type: "language",
         subtag: "jeh",
         description: "Jeh",
         added: "2009-07-29"}).
iana(language,jei,tag{
         type: "language",
         subtag: "jei",
         description: "Yei",
         added: "2009-07-29"}).
iana(language,jek,tag{
         type: "language",
         subtag: "jek",
         description: "Jeri Kuo",
         added: "2009-07-29"}).
iana(language,jel,tag{
         type: "language",
         subtag: "jel",
         description: "Yelmek",
         added: "2009-07-29"}).
iana(language,jen,tag{
         type: "language",
         subtag: "jen",
         description: "Dza",
         added: "2009-07-29"}).
iana(language,jer,tag{
         type: "language",
         subtag: "jer",
         description: "Jere",
         added: "2009-07-29"}).
iana(language,jet,tag{
         type: "language",
         subtag: "jet",
         description: "Manem",
         added: "2009-07-29"}).
iana(language,jeu,tag{
         type: "language",
         subtag: "jeu",
         description: "Jonkor Bourmataguil",
         added: "2009-07-29"}).
iana(language,jgb,tag{
         type: "language",
         subtag: "jgb",
         description: "Ngbee",
         added: "2009-07-29"}).
iana(language,jge,tag{
         type: "language",
         subtag: "jge",
         description: "Judeo-Georgian",
         added: "2009-07-29"}).
iana(language,jgk,tag{
         type: "language",
         subtag: "jgk",
         description: "Gwak",
         added: "2012-08-12"}).
iana(language,jgo,tag{
         type: "language",
         subtag: "jgo",
         description: "Ngomba",
         added: "2009-07-29"}).
iana(language,jhi,tag{
         type: "language",
         subtag: "jhi",
         description: "Jehai",
         added: "2009-07-29"}).
iana(language,jhs,tag{
         type: "language",
         subtag: "jhs",
         description: "Jhankot Sign Language",
         added: "2009-07-29"}).
iana(language,jia,tag{
         type: "language",
         subtag: "jia",
         description: "Jina",
         added: "2009-07-29"}).
iana(language,jib,tag{
         type: "language",
         subtag: "jib",
         description: "Jibu",
         added: "2009-07-29"}).
iana(language,jic,tag{
         type: "language",
         subtag: "jic",
         description: "Tol",
         added: "2009-07-29"}).
iana(language,jid,tag{
         type: "language",
         subtag: "jid",
         description: "Bu (Kaduna State)",
         added: "2009-07-29"}).
iana(language,jie,tag{
         type: "language",
         subtag: "jie",
         description: "Jilbe",
         added: "2009-07-29"}).
iana(language,jig,tag{
         type: "language",
         subtag: "jig",
         description: ["Jingulu","Djingili"],
         added: "2009-07-29"}).
iana(language,jih,tag{
         type: "language",
         subtag: "jih",
         description: ["sTodsde","Shangzhai"],
         added: "2009-07-29"}).
iana(language,jii,tag{
         type: "language",
         subtag: "jii",
         description: "Jiiddu",
         added: "2009-07-29"}).
iana(language,jil,tag{
         type: "language",
         subtag: "jil",
         description: "Jilim",
         added: "2009-07-29"}).
iana(language,jim,tag{
         type: "language",
         subtag: "jim",
         description: "Jimi (Cameroon)",
         added: "2009-07-29"}).
iana(language,jio,tag{
         type: "language",
         subtag: "jio",
         description: "Jiamao",
         added: "2009-07-29"}).
iana(language,jiq,tag{
         type: "language",
         subtag: "jiq",
         description: ["Guanyinqiao","Lavrung"],
         added: "2009-07-29"}).
iana(language,jit,tag{
         type: "language",
         subtag: "jit",
         description: "Jita",
         added: "2009-07-29"}).
iana(language,jiu,tag{
         type: "language",
         subtag: "jiu",
         description: "Youle Jinuo",
         added: "2009-07-29"}).
iana(language,jiv,tag{
         type: "language",
         subtag: "jiv",
         description: "Shuar",
         added: "2009-07-29"}).
iana(language,jiy,tag{
         type: "language",
         subtag: "jiy",
         description: "Buyuan Jinuo",
         added: "2009-07-29"}).
iana(language,jje,tag{
         type: "language",
         subtag: "jje",
         description: "Jejueo",
         added: "2015-02-12"}).
iana(language,jjr,tag{
         type: "language",
         subtag: "jjr",
         description: "Bankal",
         added: "2012-08-12"}).
iana(language,jka,tag{
         type: "language",
         subtag: "jka",
         description: "Kaera",
         added: "2016-05-30"}).
iana(language,jkm,tag{
         type: "language",
         subtag: "jkm",
         description: "Mobwa Karen",
         added: "2012-08-12"}).
iana(language,jko,tag{
         type: "language",
         subtag: "jko",
         description: "Kubo",
         added: "2009-07-29"}).
iana(language,jkp,tag{
         type: "language",
         subtag: "jkp",
         description: "Paku Karen",
         added: "2012-08-12"}).
iana(language,jkr,tag{
         type: "language",
         subtag: "jkr",
         description: "Koro (India)",
         added: "2012-08-12"}).
iana(language,jks,tag{
         type: "language",
         subtag: "jks",
         description: "Amami Koniya Sign Language",
         added: "2021-02-20"}).
iana(language,jku,tag{
         type: "language",
         subtag: "jku",
         description: "Labir",
         added: "2009-07-29"}).
iana(language,jle,tag{
         type: "language",
         subtag: "jle",
         description: "Ngile",
         added: "2009-07-29"}).
iana(language,jls,tag{
         type: "language",
         subtag: "jls",
         description: "Jamaican Sign Language",
         added: "2010-03-11"}).
iana(language,jma,tag{
         type: "language",
         subtag: "jma",
         description: "Dima",
         added: "2009-07-29"}).
iana(language,jmb,tag{
         type: "language",
         subtag: "jmb",
         description: "Zumbun",
         added: "2009-07-29"}).
iana(language,jmc,tag{
         type: "language",
         subtag: "jmc",
         description: "Machame",
         added: "2009-07-29"}).
iana(language,jmd,tag{
         type: "language",
         subtag: "jmd",
         description: "Yamdena",
         added: "2009-07-29"}).
iana(language,jmi,tag{
         type: "language",
         subtag: "jmi",
         description: "Jimi (Nigeria)",
         added: "2009-07-29"}).
iana(language,jml,tag{
         type: "language",
         subtag: "jml",
         description: "Jumli",
         added: "2009-07-29"}).
iana(language,jmn,tag{
         type: "language",
         subtag: "jmn",
         description: "Makuri Naga",
         added: "2009-07-29"}).
iana(language,jmr,tag{
         type: "language",
         subtag: "jmr",
         description: "Kamara",
         added: "2009-07-29"}).
iana(language,jms,tag{
         type: "language",
         subtag: "jms",
         description: "Mashi (Nigeria)",
         added: "2009-07-29"}).
iana(language,jmw,tag{
         type: "language",
         subtag: "jmw",
         description: "Mouwase",
         added: "2012-08-12"}).
iana(language,jmx,tag{
         type: "language",
         subtag: "jmx",
         description: "Western Juxtlahuaca Mixtec",
         added: "2009-07-29"}).
iana(language,jna,tag{
         type: "language",
         subtag: "jna",
         description: "Jangshung",
         added: "2009-07-29"}).
iana(language,jnd,tag{
         type: "language",
         subtag: "jnd",
         description: "Jandavra",
         added: "2009-07-29"}).
iana(language,jng,tag{
         type: "language",
         subtag: "jng",
         description: "Yangman",
         added: "2009-07-29"}).
iana(language,jni,tag{
         type: "language",
         subtag: "jni",
         description: "Janji",
         added: "2009-07-29"}).
iana(language,jnj,tag{
         type: "language",
         subtag: "jnj",
         description: "Yemsa",
         added: "2009-07-29"}).
iana(language,jnl,tag{
         type: "language",
         subtag: "jnl",
         description: "Rawat",
         added: "2009-07-29"}).
iana(language,jns,tag{
         type: "language",
         subtag: "jns",
         description: "Jaunsari",
         added: "2009-07-29"}).
iana(language,job,tag{
         type: "language",
         subtag: "job",
         description: "Joba",
         added: "2009-07-29"}).
iana(language,jod,tag{
         type: "language",
         subtag: "jod",
         description: "Wojenaka",
         added: "2009-07-29"}).
iana(language,jog,tag{
         type: "language",
         subtag: "jog",
         description: "Jogi",
         added: "2015-05-27"}).
iana(language,jor,tag{
         type: "language",
         subtag: "jor",
         description: "Jorá",
         added: "2009-07-29"}).
iana(language,jos,tag{
         type: "language",
         subtag: "jos",
         description: "Jordanian Sign Language",
         added: "2009-07-29"}).
iana(language,jow,tag{
         type: "language",
         subtag: "jow",
         description: "Jowulu",
         added: "2009-07-29"}).
iana(language,jpa,tag{
         type: "language",
         subtag: "jpa",
         description: "Jewish Palestinian Aramaic",
         added: "2009-07-29"}).
iana(language,jpr,tag{
         type: "language",
         subtag: "jpr",
         description: "Judeo-Persian",
         added: "2005-10-16"}).
iana(language,jpx,tag{
         type: "language",
         subtag: "jpx",
         description: "Japanese (family)",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,jqr,tag{
         type: "language",
         subtag: "jqr",
         description: "Jaqaru",
         added: "2009-07-29"}).
iana(language,jra,tag{
         type: "language",
         subtag: "jra",
         description: "Jarai",
         added: "2009-07-29"}).
iana(language,jrb,tag{
         type: "language",
         subtag: "jrb",
         description: "Judeo-Arabic",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,jrr,tag{
         type: "language",
         subtag: "jrr",
         description: "Jiru",
         added: "2009-07-29"}).
iana(language,jrt,tag{
         type: "language",
         subtag: "jrt",
         description: "Jakattoe",
         added: "2009-07-29"}).
iana(language,jru,tag{
         type: "language",
         subtag: "jru",
         description: "Japrería",
         added: "2009-07-29"}).
iana(language,jsl,tag{
         type: "language",
         subtag: "jsl",
         description: "Japanese Sign Language",
         added: "2009-07-29"}).
iana(language,jua,tag{
         type: "language",
         subtag: "jua",
         description: "Júma",
         added: "2009-07-29"}).
iana(language,jub,tag{
         type: "language",
         subtag: "jub",
         description: "Wannu",
         added: "2009-07-29"}).
iana(language,juc,tag{
         type: "language",
         subtag: "juc",
         description: "Jurchen",
         added: "2009-07-29"}).
iana(language,jud,tag{
         type: "language",
         subtag: "jud",
         description: "Worodougou",
         added: "2009-07-29"}).
iana(language,juh,tag{
         type: "language",
         subtag: "juh",
         description: "Hõne",
         added: "2009-07-29"}).
iana(language,jui,tag{
         type: "language",
         subtag: "jui",
         description: "Ngadjuri",
         added: "2012-08-12"}).
iana(language,juk,tag{
         type: "language",
         subtag: "juk",
         description: "Wapan",
         added: "2009-07-29"}).
iana(language,jul,tag{
         type: "language",
         subtag: "jul",
         description: "Jirel",
         added: "2009-07-29"}).
iana(language,jum,tag{
         type: "language",
         subtag: "jum",
         description: "Jumjum",
         added: "2009-07-29"}).
iana(language,jun,tag{
         type: "language",
         subtag: "jun",
         description: "Juang",
         added: "2009-07-29"}).
iana(language,juo,tag{
         type: "language",
         subtag: "juo",
         description: "Jiba",
         added: "2009-07-29"}).
iana(language,jup,tag{
         type: "language",
         subtag: "jup",
         description: "Hupdë",
         added: "2009-07-29"}).
iana(language,jur,tag{
         type: "language",
         subtag: "jur",
         description: "Jurúna",
         added: "2009-07-29"}).
iana(language,jus,tag{
         type: "language",
         subtag: "jus",
         description: "Jumla Sign Language",
         added: "2009-07-29"}).
iana(language,jut,tag{
         type: "language",
         subtag: "jut",
         description: "Jutish",
         added: "2009-07-29"}).
iana(language,juu,tag{
         type: "language",
         subtag: "juu",
         description: "Ju",
         added: "2009-07-29"}).
iana(language,juw,tag{
         type: "language",
         subtag: "juw",
         description: "Wãpha",
         added: "2009-07-29"}).
iana(language,juy,tag{
         type: "language",
         subtag: "juy",
         description: "Juray",
         added: "2009-07-29"}).
iana(language,jvd,tag{
         type: "language",
         subtag: "jvd",
         description: "Javindo",
         added: "2009-07-29"}).
iana(language,jvn,tag{
         type: "language",
         subtag: "jvn",
         description: "Caribbean Javanese",
         added: "2009-07-29"}).
iana(language,jwi,tag{
         type: "language",
         subtag: "jwi",
         description: "Jwira-Pepesa",
         added: "2009-07-29"}).
iana(language,jya,tag{
         type: "language",
         subtag: "jya",
         description: "Jiarong",
         added: "2009-07-29"}).
iana(language,jye,tag{
         type: "language",
         subtag: "jye",
         description: "Judeo-Yemeni Arabic",
         added: "2009-07-29",
         macrolanguage: "jrb"}).
iana(language,jyy,tag{
         type: "language",
         subtag: "jyy",
         description: "Jaya",
         added: "2009-07-29"}).
iana(language,kaa,tag{
         type: "language",
         subtag: "kaa",
         description: ["Kara-Kalpak","Karakalpak"],
         added: "2005-10-16"}).
iana(language,kab,tag{
         type: "language",
         subtag: "kab",
         description: "Kabyle",
         added: "2005-10-16"}).
iana(language,kac,tag{
         type: "language",
         subtag: "kac",
         description: ["Kachin","Jingpho"],
         added: "2005-10-16"}).
iana(language,kad,tag{
         type: "language",
         subtag: "kad",
         description: "Adara",
         added: "2009-07-29"}).
iana(language,kae,tag{
         type: "language",
         subtag: "kae",
         description: "Ketangalan",
         added: "2009-07-29"}).
iana(language,kaf,tag{
         type: "language",
         subtag: "kaf",
         description: "Katso",
         added: "2009-07-29"}).
iana(language,kag,tag{
         type: "language",
         subtag: "kag",
         description: "Kajaman",
         added: "2009-07-29"}).
iana(language,kah,tag{
         type: "language",
         subtag: "kah",
         description: "Kara (Central African Republic)",
         added: "2009-07-29"}).
iana(language,kai,tag{
         type: "language",
         subtag: "kai",
         description: "Karekare",
         added: "2009-07-29"}).
iana(language,kaj,tag{
         type: "language",
         subtag: "kaj",
         description: "Jju",
         added: "2009-07-29"}).
iana(language,kak,tag{
         type: "language",
         subtag: "kak",
         description: ["Kalanguya","Kayapa Kallahan"],
         added: "2009-07-29"}).
iana(language,kam,tag{
         type: "language",
         subtag: "kam",
         description: "Kamba (Kenya)",
         added: "2005-10-16"}).
iana(language,kao,tag{
         type: "language",
         subtag: "kao",
         description: "Xaasongaxango",
         added: "2009-07-29"}).
iana(language,kap,tag{
         type: "language",
         subtag: "kap",
         description: "Bezhta",
         added: "2009-07-29"}).
iana(language,kaq,tag{
         type: "language",
         subtag: "kaq",
         description: "Capanahua",
         added: "2009-07-29"}).
iana(language,kar,tag{
         type: "language",
         subtag: "kar",
         description: "Karen languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,kav,tag{
         type: "language",
         subtag: "kav",
         description: "Katukína",
         added: "2009-07-29"}).
iana(language,kaw,tag{
         type: "language",
         subtag: "kaw",
         description: "Kawi",
         added: "2005-10-16"}).
iana(language,kax,tag{
         type: "language",
         subtag: "kax",
         description: "Kao",
         added: "2009-07-29"}).
iana(language,kay,tag{
         type: "language",
         subtag: "kay",
         description: "Kamayurá",
         added: "2009-07-29"}).
iana(language,kba,tag{
         type: "language",
         subtag: "kba",
         description: "Kalarko",
         added: "2009-07-29"}).
iana(language,kbb,tag{
         type: "language",
         subtag: "kbb",
         description: "Kaxuiâna",
         added: "2009-07-29"}).
iana(language,kbc,tag{
         type: "language",
         subtag: "kbc",
         description: "Kadiwéu",
         added: "2009-07-29"}).
iana(language,kbd,tag{
         type: "language",
         subtag: "kbd",
         description: "Kabardian",
         added: "2005-10-16"}).
iana(language,kbe,tag{
         type: "language",
         subtag: "kbe",
         description: "Kanju",
         added: "2009-07-29"}).
iana(language,kbf,tag{
         type: "language",
         subtag: "kbf",
         description: "Kakauhua",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,kbg,tag{
         type: "language",
         subtag: "kbg",
         description: "Khamba",
         added: "2009-07-29"}).
iana(language,kbh,tag{
         type: "language",
         subtag: "kbh",
         description: "Camsá",
         added: "2009-07-29"}).
iana(language,kbi,tag{
         type: "language",
         subtag: "kbi",
         description: "Kaptiau",
         added: "2009-07-29"}).
iana(language,kbj,tag{
         type: "language",
         subtag: "kbj",
         description: "Kari",
         added: "2009-07-29"}).
iana(language,kbk,tag{
         type: "language",
         subtag: "kbk",
         description: "Grass Koiari",
         added: "2009-07-29"}).
iana(language,kbl,tag{
         type: "language",
         subtag: "kbl",
         description: "Kanembu",
         added: "2009-07-29"}).
iana(language,kbm,tag{
         type: "language",
         subtag: "kbm",
         description: "Iwal",
         added: "2009-07-29"}).
iana(language,kbn,tag{
         type: "language",
         subtag: "kbn",
         description: "Kare (Central African Republic)",
         added: "2009-07-29"}).
iana(language,kbo,tag{
         type: "language",
         subtag: "kbo",
         description: "Keliko",
         added: "2009-07-29"}).
iana(language,kbp,tag{
         type: "language",
         subtag: "kbp",
         description: "Kabiyè",
         added: "2009-07-29"}).
iana(language,kbq,tag{
         type: "language",
         subtag: "kbq",
         description: "Kamano",
         added: "2009-07-29"}).
iana(language,kbr,tag{
         type: "language",
         subtag: "kbr",
         description: "Kafa",
         added: "2009-07-29"}).
iana(language,kbs,tag{
         type: "language",
         subtag: "kbs",
         description: "Kande",
         added: "2009-07-29"}).
iana(language,kbt,tag{
         type: "language",
         subtag: "kbt",
         description: "Abadi",
         added: "2009-07-29"}).
iana(language,kbu,tag{
         type: "language",
         subtag: "kbu",
         description: "Kabutra",
         added: "2009-07-29"}).
iana(language,kbv,tag{
         type: "language",
         subtag: "kbv",
         description: "Dera (Indonesia)",
         added: "2009-07-29"}).
iana(language,kbw,tag{
         type: "language",
         subtag: "kbw",
         description: "Kaiep",
         added: "2009-07-29"}).
iana(language,kbx,tag{
         type: "language",
         subtag: "kbx",
         description: "Ap Ma",
         added: "2009-07-29"}).
iana(language,kby,tag{
         type: "language",
         subtag: "kby",
         description: "Manga Kanuri",
         added: "2009-07-29",
         macrolanguage: "kr"}).
iana(language,kbz,tag{
         type: "language",
         subtag: "kbz",
         description: "Duhwa",
         added: "2009-07-29"}).
iana(language,kca,tag{
         type: "language",
         subtag: "kca",
         description: "Khanty",
         added: "2009-07-29"}).
iana(language,kcb,tag{
         type: "language",
         subtag: "kcb",
         description: "Kawacha",
         added: "2009-07-29"}).
iana(language,kcc,tag{
         type: "language",
         subtag: "kcc",
         description: "Lubila",
         added: "2009-07-29"}).
iana(language,kcd,tag{
         type: "language",
         subtag: "kcd",
         description: "Ngkâlmpw Kanum",
         added: "2009-07-29"}).
iana(language,kce,tag{
         type: "language",
         subtag: "kce",
         description: "Kaivi",
         added: "2009-07-29"}).
iana(language,kcf,tag{
         type: "language",
         subtag: "kcf",
         description: "Ukaan",
         added: "2009-07-29"}).
iana(language,kcg,tag{
         type: "language",
         subtag: "kcg",
         description: "Tyap",
         added: "2009-07-29"}).
iana(language,kch,tag{
         type: "language",
         subtag: "kch",
         description: "Vono",
         added: "2009-07-29"}).
iana(language,kci,tag{
         type: "language",
         subtag: "kci",
         description: "Kamantan",
         added: "2009-07-29"}).
iana(language,kcj,tag{
         type: "language",
         subtag: "kcj",
         description: "Kobiana",
         added: "2009-07-29"}).
iana(language,kck,tag{
         type: "language",
         subtag: "kck",
         description: "Kalanga",
         added: "2009-07-29"}).
iana(language,kcl,tag{
         type: "language",
         subtag: "kcl",
         description: ["Kela (Papua New Guinea)","Kala"],
         added: "2009-07-29"}).
iana(language,kcm,tag{
         type: "language",
         subtag: "kcm",
         description: "Gula (Central African Republic)",
         added: "2009-07-29"}).
iana(language,kcn,tag{
         type: "language",
         subtag: "kcn",
         description: "Nubi",
         added: "2009-07-29"}).
iana(language,kco,tag{
         type: "language",
         subtag: "kco",
         description: "Kinalakna",
         added: "2009-07-29"}).
iana(language,kcp,tag{
         type: "language",
         subtag: "kcp",
         description: "Kanga",
         added: "2009-07-29"}).
iana(language,kcq,tag{
         type: "language",
         subtag: "kcq",
         description: "Kamo",
         added: "2009-07-29"}).
iana(language,kcr,tag{
         type: "language",
         subtag: "kcr",
         description: "Katla",
         added: "2009-07-29"}).
iana(language,kcs,tag{
         type: "language",
         subtag: "kcs",
         description: "Koenoem",
         added: "2009-07-29"}).
iana(language,kct,tag{
         type: "language",
         subtag: "kct",
         description: "Kaian",
         added: "2009-07-29"}).
iana(language,kcu,tag{
         type: "language",
         subtag: "kcu",
         description: "Kami (Tanzania)",
         added: "2009-07-29"}).
iana(language,kcv,tag{
         type: "language",
         subtag: "kcv",
         description: "Kete",
         added: "2009-07-29"}).
iana(language,kcw,tag{
         type: "language",
         subtag: "kcw",
         description: "Kabwari",
         added: "2009-07-29"}).
iana(language,kcx,tag{
         type: "language",
         subtag: "kcx",
         description: "Kachama-Ganjule",
         added: "2009-07-29"}).
iana(language,kcy,tag{
         type: "language",
         subtag: "kcy",
         description: "Korandje",
         added: "2009-07-29"}).
iana(language,kcz,tag{
         type: "language",
         subtag: "kcz",
         description: "Konongo",
         added: "2009-07-29"}).
iana(language,kda,tag{
         type: "language",
         subtag: "kda",
         description: "Worimi",
         added: "2009-07-29"}).
iana(language,kdc,tag{
         type: "language",
         subtag: "kdc",
         description: "Kutu",
         added: "2009-07-29"}).
iana(language,kdd,tag{
         type: "language",
         subtag: "kdd",
         description: "Yankunytjatjara",
         added: "2009-07-29"}).
iana(language,kde,tag{
         type: "language",
         subtag: "kde",
         description: "Makonde",
         added: "2009-07-29"}).
iana(language,kdf,tag{
         type: "language",
         subtag: "kdf",
         description: "Mamusi",
         added: "2009-07-29"}).
iana(language,kdg,tag{
         type: "language",
         subtag: "kdg",
         description: "Seba",
         added: "2009-07-29"}).
iana(language,kdh,tag{
         type: "language",
         subtag: "kdh",
         description: "Tem",
         added: "2009-07-29"}).
iana(language,kdi,tag{
         type: "language",
         subtag: "kdi",
         description: "Kumam",
         added: "2009-07-29"}).
iana(language,kdj,tag{
         type: "language",
         subtag: "kdj",
         description: "Karamojong",
         added: "2009-07-29"}).
iana(language,kdk,tag{
         type: "language",
         subtag: "kdk",
         description: ["Numèè","Kwényi"],
         added: "2009-07-29"}).
iana(language,kdl,tag{
         type: "language",
         subtag: "kdl",
         description: "Tsikimba",
         added: "2009-07-29"}).
iana(language,kdm,tag{
         type: "language",
         subtag: "kdm",
         description: "Kagoma",
         added: "2009-07-29"}).
iana(language,kdn,tag{
         type: "language",
         subtag: "kdn",
         description: "Kunda",
         added: "2009-07-29"}).
iana(language,kdo,tag{
         type: "language",
         subtag: "kdo",
         description: "Kordofanian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,kdp,tag{
         type: "language",
         subtag: "kdp",
         description: "Kaningdon-Nindem",
         added: "2009-07-29"}).
iana(language,kdq,tag{
         type: "language",
         subtag: "kdq",
         description: "Koch",
         added: "2009-07-29"}).
iana(language,kdr,tag{
         type: "language",
         subtag: "kdr",
         description: "Karaim",
         added: "2009-07-29"}).
iana(language,kdt,tag{
         type: "language",
         subtag: "kdt",
         description: "Kuy",
         added: "2009-07-29"}).
iana(language,kdu,tag{
         type: "language",
         subtag: "kdu",
         description: "Kadaru",
         added: "2009-07-29"}).
iana(language,kdv,tag{
         type: "language",
         subtag: "kdv",
         description: "Kado",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see zkd, zkn"}).
iana(language,kdw,tag{
         type: "language",
         subtag: "kdw",
         description: "Koneraw",
         added: "2009-07-29"}).
iana(language,kdx,tag{
         type: "language",
         subtag: "kdx",
         description: "Kam",
         added: "2009-07-29"}).
iana(language,kdy,tag{
         type: "language",
         subtag: "kdy",
         description: ["Keder","Keijar"],
         added: "2009-07-29"}).
iana(language,kdz,tag{
         type: "language",
         subtag: "kdz",
         description: "Kwaja",
         added: "2009-07-29"}).
iana(language,kea,tag{
         type: "language",
         subtag: "kea",
         description: "Kabuverdianu",
         added: "2009-07-29"}).
iana(language,keb,tag{
         type: "language",
         subtag: "keb",
         description: "Kélé",
         added: "2009-07-29"}).
iana(language,kec,tag{
         type: "language",
         subtag: "kec",
         description: "Keiga",
         added: "2009-07-29"}).
iana(language,ked,tag{
         type: "language",
         subtag: "ked",
         description: "Kerewe",
         added: "2009-07-29"}).
iana(language,kee,tag{
         type: "language",
         subtag: "kee",
         description: "Eastern Keres",
         added: "2009-07-29"}).
iana(language,kef,tag{
         type: "language",
         subtag: "kef",
         description: "Kpessi",
         added: "2009-07-29"}).
iana(language,keg,tag{
         type: "language",
         subtag: "keg",
         description: "Tese",
         added: "2009-07-29"}).
iana(language,keh,tag{
         type: "language",
         subtag: "keh",
         description: "Keak",
         added: "2009-07-29"}).
iana(language,kei,tag{
         type: "language",
         subtag: "kei",
         description: "Kei",
         added: "2009-07-29"}).
iana(language,kej,tag{
         type: "language",
         subtag: "kej",
         description: "Kadar",
         added: "2009-07-29"}).
iana(language,kek,tag{
         type: "language",
         subtag: "kek",
         description: "Kekchí",
         added: "2009-07-29"}).
iana(language,kel,tag{
         type: "language",
         subtag: "kel",
         description: "Kela (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,kem,tag{
         type: "language",
         subtag: "kem",
         description: "Kemak",
         added: "2009-07-29"}).
iana(language,ken,tag{
         type: "language",
         subtag: "ken",
         description: "Kenyang",
         added: "2009-07-29"}).
iana(language,keo,tag{
         type: "language",
         subtag: "keo",
         description: "Kakwa",
         added: "2009-07-29"}).
iana(language,kep,tag{
         type: "language",
         subtag: "kep",
         description: "Kaikadi",
         added: "2009-07-29"}).
iana(language,keq,tag{
         type: "language",
         subtag: "keq",
         description: "Kamar",
         added: "2009-07-29"}).
iana(language,ker,tag{
         type: "language",
         subtag: "ker",
         description: "Kera",
         added: "2009-07-29"}).
iana(language,kes,tag{
         type: "language",
         subtag: "kes",
         description: "Kugbo",
         added: "2009-07-29"}).
iana(language,ket,tag{
         type: "language",
         subtag: "ket",
         description: "Ket",
         added: "2009-07-29"}).
iana(language,keu,tag{
         type: "language",
         subtag: "keu",
         description: "Akebu",
         added: "2009-07-29"}).
iana(language,kev,tag{
         type: "language",
         subtag: "kev",
         description: "Kanikkaran",
         added: "2009-07-29"}).
iana(language,kew,tag{
         type: "language",
         subtag: "kew",
         description: "West Kewa",
         added: "2009-07-29"}).
iana(language,kex,tag{
         type: "language",
         subtag: "kex",
         description: "Kukna",
         added: "2009-07-29"}).
iana(language,key,tag{
         type: "language",
         subtag: "key",
         description: "Kupia",
         added: "2009-07-29"}).
iana(language,kez,tag{
         type: "language",
         subtag: "kez",
         description: "Kukele",
         added: "2009-07-29"}).
iana(language,kfa,tag{
         type: "language",
         subtag: "kfa",
         description: "Kodava",
         added: "2009-07-29"}).
iana(language,kfb,tag{
         type: "language",
         subtag: "kfb",
         description: "Northwestern Kolami",
         added: "2009-07-29"}).
iana(language,kfc,tag{
         type: "language",
         subtag: "kfc",
         description: "Konda-Dora",
         added: "2009-07-29"}).
iana(language,kfd,tag{
         type: "language",
         subtag: "kfd",
         description: "Korra Koraga",
         added: "2009-07-29"}).
iana(language,kfe,tag{
         type: "language",
         subtag: "kfe",
         description: "Kota (India)",
         added: "2009-07-29"}).
iana(language,kff,tag{
         type: "language",
         subtag: "kff",
         description: "Koya",
         added: "2009-07-29"}).
iana(language,kfg,tag{
         type: "language",
         subtag: "kfg",
         description: "Kudiya",
         added: "2009-07-29"}).
iana(language,kfh,tag{
         type: "language",
         subtag: "kfh",
         description: "Kurichiya",
         added: "2009-07-29"}).
iana(language,kfi,tag{
         type: "language",
         subtag: "kfi",
         description: "Kannada Kurumba",
         added: "2009-07-29"}).
iana(language,kfj,tag{
         type: "language",
         subtag: "kfj",
         description: "Kemiehua",
         added: "2009-07-29"}).
iana(language,kfk,tag{
         type: "language",
         subtag: "kfk",
         description: "Kinnauri",
         added: "2009-07-29"}).
iana(language,kfl,tag{
         type: "language",
         subtag: "kfl",
         description: "Kung",
         added: "2009-07-29"}).
iana(language,kfm,tag{
         type: "language",
         subtag: "kfm",
         description: "Khunsari",
         added: "2009-07-29"}).
iana(language,kfn,tag{
         type: "language",
         subtag: "kfn",
         description: "Kuk",
         added: "2009-07-29"}).
iana(language,kfo,tag{
         type: "language",
         subtag: "kfo",
         description: "Koro (Côte d'Ivoire)",
         added: "2009-07-29"}).
iana(language,kfp,tag{
         type: "language",
         subtag: "kfp",
         description: "Korwa",
         added: "2009-07-29"}).
iana(language,kfq,tag{
         type: "language",
         subtag: "kfq",
         description: "Korku",
         added: "2009-07-29"}).
iana(language,kfr,tag{
         type: "language",
         subtag: "kfr",
         description: ["Kachhi","Kutchi"],
         added: "2009-07-29"}).
iana(language,kfs,tag{
         type: "language",
         subtag: "kfs",
         description: "Bilaspuri",
         added: "2009-07-29"}).
iana(language,kft,tag{
         type: "language",
         subtag: "kft",
         description: "Kanjari",
         added: "2009-07-29"}).
iana(language,kfu,tag{
         type: "language",
         subtag: "kfu",
         description: "Katkari",
         added: "2009-07-29"}).
iana(language,kfv,tag{
         type: "language",
         subtag: "kfv",
         description: "Kurmukar",
         added: "2009-07-29"}).
iana(language,kfw,tag{
         type: "language",
         subtag: "kfw",
         description: "Kharam Naga",
         added: "2009-07-29"}).
iana(language,kfx,tag{
         type: "language",
         subtag: "kfx",
         description: "Kullu Pahari",
         added: "2009-07-29"}).
iana(language,kfy,tag{
         type: "language",
         subtag: "kfy",
         description: "Kumaoni",
         added: "2009-07-29"}).
iana(language,kfz,tag{
         type: "language",
         subtag: "kfz",
         description: "Koromfé",
         added: "2009-07-29"}).
iana(language,kga,tag{
         type: "language",
         subtag: "kga",
         description: "Koyaga",
         added: "2009-07-29"}).
iana(language,kgb,tag{
         type: "language",
         subtag: "kgb",
         description: "Kawe",
         added: "2009-07-29"}).
iana(language,kgc,tag{
         type: "language",
         subtag: "kgc",
         description: "Kasseng",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "tdf"}).
iana(language,kgd,tag{
         type: "language",
         subtag: "kgd",
         description: "Kataang",
         added: "2009-07-29",
         deprecated: "2017-02-23",
         comments: "see ncq, sct"}).
iana(language,kge,tag{
         type: "language",
         subtag: "kge",
         description: "Komering",
         added: "2009-07-29"}).
iana(language,kgf,tag{
         type: "language",
         subtag: "kgf",
         description: "Kube",
         added: "2009-07-29"}).
iana(language,kgg,tag{
         type: "language",
         subtag: "kgg",
         description: "Kusunda",
         added: "2009-07-29"}).
iana(language,kgh,tag{
         type: "language",
         subtag: "kgh",
         description: "Upper Tanudan Kalinga",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "kml"}).
iana(language,kgi,tag{
         type: "language",
         subtag: "kgi",
         description: "Selangor Sign Language",
         added: "2009-07-29"}).
iana(language,kgj,tag{
         type: "language",
         subtag: "kgj",
         description: "Gamale Kham",
         added: "2009-07-29"}).
iana(language,kgk,tag{
         type: "language",
         subtag: "kgk",
         description: "Kaiwá",
         added: "2009-07-29"}).
iana(language,kgl,tag{
         type: "language",
         subtag: "kgl",
         description: "Kunggari",
         added: "2009-07-29"}).
iana(language,kgm,tag{
         type: "language",
         subtag: "kgm",
         description: "Karipúna",
         added: "2009-07-29"}).
iana(language,kgn,tag{
         type: "language",
         subtag: "kgn",
         description: "Karingani",
         added: "2009-07-29"}).
iana(language,kgo,tag{
         type: "language",
         subtag: "kgo",
         description: "Krongo",
         added: "2009-07-29"}).
iana(language,kgp,tag{
         type: "language",
         subtag: "kgp",
         description: "Kaingang",
         added: "2009-07-29"}).
iana(language,kgq,tag{
         type: "language",
         subtag: "kgq",
         description: "Kamoro",
         added: "2009-07-29"}).
iana(language,kgr,tag{
         type: "language",
         subtag: "kgr",
         description: "Abun",
         added: "2009-07-29"}).
iana(language,kgs,tag{
         type: "language",
         subtag: "kgs",
         description: "Kumbainggar",
         added: "2009-07-29"}).
iana(language,kgt,tag{
         type: "language",
         subtag: "kgt",
         description: "Somyev",
         added: "2009-07-29"}).
iana(language,kgu,tag{
         type: "language",
         subtag: "kgu",
         description: "Kobol",
         added: "2009-07-29"}).
iana(language,kgv,tag{
         type: "language",
         subtag: "kgv",
         description: "Karas",
         added: "2009-07-29"}).
iana(language,kgw,tag{
         type: "language",
         subtag: "kgw",
         description: "Karon Dori",
         added: "2009-07-29"}).
iana(language,kgx,tag{
         type: "language",
         subtag: "kgx",
         description: "Kamaru",
         added: "2009-07-29"}).
iana(language,kgy,tag{
         type: "language",
         subtag: "kgy",
         description: "Kyerung",
         added: "2009-07-29"}).
iana(language,kha,tag{
         type: "language",
         subtag: "kha",
         description: "Khasi",
         added: "2005-10-16",
         comments: "as of 2008-04-21 this subtag does not include Lyngngam; see lyg"
     }).
iana(language,khb,tag{
         type: "language",
         subtag: "khb",
         description: "Lü",
         added: "2009-07-29"}).
iana(language,khc,tag{
         type: "language",
         subtag: "khc",
         description: "Tukang Besi North",
         added: "2009-07-29"}).
iana(language,khd,tag{
         type: "language",
         subtag: "khd",
         description: "Bädi Kanum",
         added: "2009-07-29"}).
iana(language,khe,tag{
         type: "language",
         subtag: "khe",
         description: "Korowai",
         added: "2009-07-29"}).
iana(language,khf,tag{
         type: "language",
         subtag: "khf",
         description: "Khuen",
         added: "2009-07-29"}).
iana(language,khg,tag{
         type: "language",
         subtag: "khg",
         description: "Khams Tibetan",
         added: "2009-07-29"}).
iana(language,khh,tag{
         type: "language",
         subtag: "khh",
         description: "Kehu",
         added: "2009-07-29"}).
iana(language,khi,tag{
         type: "language",
         subtag: "khi",
         description: "Khoisan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,khj,tag{
         type: "language",
         subtag: "khj",
         description: "Kuturmi",
         added: "2009-07-29"}).
iana(language,khk,tag{
         type: "language",
         subtag: "khk",
         description: "Halh Mongolian",
         added: "2009-07-29",
         macrolanguage: "mn"}).
iana(language,khl,tag{
         type: "language",
         subtag: "khl",
         description: "Lusi",
         added: "2009-07-29"}).
iana(language,khn,tag{
         type: "language",
         subtag: "khn",
         description: "Khandesi",
         added: "2009-07-29"}).
iana(language,kho,tag{
         type: "language",
         subtag: "kho",
         description: ["Khotanese","Sakan"],
         added: "2005-10-16"}).
iana(language,khp,tag{
         type: "language",
         subtag: "khp",
         description: ["Kapori","Kapauri"],
         added: "2009-07-29"}).
iana(language,khq,tag{
         type: "language",
         subtag: "khq",
         description: "Koyra Chiini Songhay",
         added: "2009-07-29"}).
iana(language,khr,tag{
         type: "language",
         subtag: "khr",
         description: "Kharia",
         added: "2009-07-29"}).
iana(language,khs,tag{
         type: "language",
         subtag: "khs",
         description: "Kasua",
         added: "2009-07-29"}).
iana(language,kht,tag{
         type: "language",
         subtag: "kht",
         description: "Khamti",
         added: "2009-07-29"}).
iana(language,khu,tag{
         type: "language",
         subtag: "khu",
         description: "Nkhumbi",
         added: "2009-07-29"}).
iana(language,khv,tag{
         type: "language",
         subtag: "khv",
         description: "Khvarshi",
         added: "2009-07-29"}).
iana(language,khw,tag{
         type: "language",
         subtag: "khw",
         description: "Khowar",
         added: "2009-07-29"}).
iana(language,khx,tag{
         type: "language",
         subtag: "khx",
         description: "Kanu",
         added: "2009-07-29"}).
iana(language,khy,tag{
         type: "language",
         subtag: "khy",
         description: "Kele (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,khz,tag{
         type: "language",
         subtag: "khz",
         description: "Keapara",
         added: "2009-07-29"}).
iana(language,kia,tag{
         type: "language",
         subtag: "kia",
         description: "Kim",
         added: "2009-07-29"}).
iana(language,kib,tag{
         type: "language",
         subtag: "kib",
         description: "Koalib",
         added: "2009-07-29"}).
iana(language,kic,tag{
         type: "language",
         subtag: "kic",
         description: "Kickapoo",
         added: "2009-07-29"}).
iana(language,kid,tag{
         type: "language",
         subtag: "kid",
         description: "Koshin",
         added: "2009-07-29"}).
iana(language,kie,tag{
         type: "language",
         subtag: "kie",
         description: "Kibet",
         added: "2009-07-29"}).
iana(language,kif,tag{
         type: "language",
         subtag: "kif",
         description: "Eastern Parbate Kham",
         added: "2009-07-29"}).
iana(language,kig,tag{
         type: "language",
         subtag: "kig",
         description: ["Kimaama","Kimaghima"],
         added: "2009-07-29"}).
iana(language,kih,tag{
         type: "language",
         subtag: "kih",
         description: "Kilmeri",
         added: "2009-07-29"}).
iana(language,kii,tag{
         type: "language",
         subtag: "kii",
         description: "Kitsai",
         added: "2009-07-29"}).
iana(language,kij,tag{
         type: "language",
         subtag: "kij",
         description: "Kilivila",
         added: "2009-07-29"}).
iana(language,kil,tag{
         type: "language",
         subtag: "kil",
         description: "Kariya",
         added: "2009-07-29"}).
iana(language,kim,tag{
         type: "language",
         subtag: "kim",
         description: "Karagas",
         added: "2009-07-29"}).
iana(language,kio,tag{
         type: "language",
         subtag: "kio",
         description: "Kiowa",
         added: "2009-07-29"}).
iana(language,kip,tag{
         type: "language",
         subtag: "kip",
         description: "Sheshi Kham",
         added: "2009-07-29"}).
iana(language,kiq,tag{
         type: "language",
         subtag: "kiq",
         description: ["Kosadle","Kosare"],
         added: "2009-07-29"}).
iana(language,kis,tag{
         type: "language",
         subtag: "kis",
         description: "Kis",
         added: "2009-07-29"}).
iana(language,kit,tag{
         type: "language",
         subtag: "kit",
         description: "Agob",
         added: "2009-07-29"}).
iana(language,kiu,tag{
         type: "language",
         subtag: "kiu",
         description: "Kirmanjki (individual language)",
         added: "2009-07-29",
         macrolanguage: "zza"}).
iana(language,kiv,tag{
         type: "language",
         subtag: "kiv",
         description: "Kimbu",
         added: "2009-07-29"}).
iana(language,kiw,tag{
         type: "language",
         subtag: "kiw",
         description: "Northeast Kiwai",
         added: "2009-07-29"}).
iana(language,kix,tag{
         type: "language",
         subtag: "kix",
         description: "Khiamniungan Naga",
         added: "2009-07-29"}).
iana(language,kiy,tag{
         type: "language",
         subtag: "kiy",
         description: "Kirikiri",
         added: "2009-07-29"}).
iana(language,kiz,tag{
         type: "language",
         subtag: "kiz",
         description: "Kisi",
         added: "2009-07-29"}).
iana(language,kja,tag{
         type: "language",
         subtag: "kja",
         description: "Mlap",
         added: "2009-07-29"}).
iana(language,kjb,tag{
         type: "language",
         subtag: "kjb",
         description: ["Q'anjob'al","Kanjobal"],
         added: "2009-07-29"}).
iana(language,kjc,tag{
         type: "language",
         subtag: "kjc",
         description: "Coastal Konjo",
         added: "2009-07-29"}).
iana(language,kjd,tag{
         type: "language",
         subtag: "kjd",
         description: "Southern Kiwai",
         added: "2009-07-29"}).
iana(language,kje,tag{
         type: "language",
         subtag: "kje",
         description: "Kisar",
         added: "2009-07-29"}).
iana(language,kjf,tag{
         type: "language",
         subtag: "kjf",
         description: "Khalaj [Indo-Iranian]",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,kjg,tag{
         type: "language",
         subtag: "kjg",
         description: "Khmu",
         added: "2009-07-29"}).
iana(language,kjh,tag{
         type: "language",
         subtag: "kjh",
         description: "Khakas",
         added: "2009-07-29"}).
iana(language,kji,tag{
         type: "language",
         subtag: "kji",
         description: "Zabana",
         added: "2009-07-29"}).
iana(language,kjj,tag{
         type: "language",
         subtag: "kjj",
         description: "Khinalugh",
         added: "2009-07-29"}).
iana(language,kjk,tag{
         type: "language",
         subtag: "kjk",
         description: "Highland Konjo",
         added: "2009-07-29"}).
iana(language,kjl,tag{
         type: "language",
         subtag: "kjl",
         description: "Western Parbate Kham",
         added: "2009-07-29"}).
iana(language,kjm,tag{
         type: "language",
         subtag: "kjm",
         description: "Kháng",
         added: "2009-07-29"}).
iana(language,kjn,tag{
         type: "language",
         subtag: "kjn",
         description: "Kunjen",
         added: "2009-07-29"}).
iana(language,kjo,tag{
         type: "language",
         subtag: "kjo",
         description: "Harijan Kinnauri",
         added: "2009-07-29"}).
iana(language,kjp,tag{
         type: "language",
         subtag: "kjp",
         description: "Pwo Eastern Karen",
         added: "2009-07-29"}).
iana(language,kjq,tag{
         type: "language",
         subtag: "kjq",
         description: "Western Keres",
         added: "2009-07-29"}).
iana(language,kjr,tag{
         type: "language",
         subtag: "kjr",
         description: "Kurudu",
         added: "2009-07-29"}).
iana(language,kjs,tag{
         type: "language",
         subtag: "kjs",
         description: "East Kewa",
         added: "2009-07-29"}).
iana(language,kjt,tag{
         type: "language",
         subtag: "kjt",
         description: "Phrae Pwo Karen",
         added: "2009-07-29"}).
iana(language,kju,tag{
         type: "language",
         subtag: "kju",
         description: "Kashaya",
         added: "2009-07-29"}).
iana(language,kjv,tag{
         type: "language",
         subtag: "kjv",
         description: "Kaikavian Literary Language",
         added: "2015-02-12"}).
iana(language,kjx,tag{
         type: "language",
         subtag: "kjx",
         description: "Ramopa",
         added: "2009-07-29"}).
iana(language,kjy,tag{
         type: "language",
         subtag: "kjy",
         description: "Erave",
         added: "2009-07-29"}).
iana(language,kjz,tag{
         type: "language",
         subtag: "kjz",
         description: "Bumthangkha",
         added: "2009-07-29"}).
iana(language,kka,tag{
         type: "language",
         subtag: "kka",
         description: "Kakanda",
         added: "2009-07-29"}).
iana(language,kkb,tag{
         type: "language",
         subtag: "kkb",
         description: "Kwerisa",
         added: "2009-07-29"}).
iana(language,kkc,tag{
         type: "language",
         subtag: "kkc",
         description: "Odoodee",
         added: "2009-07-29"}).
iana(language,kkd,tag{
         type: "language",
         subtag: "kkd",
         description: "Kinuku",
         added: "2009-07-29"}).
iana(language,kke,tag{
         type: "language",
         subtag: "kke",
         description: "Kakabe",
         added: "2009-07-29"}).
iana(language,kkf,tag{
         type: "language",
         subtag: "kkf",
         description: "Kalaktang Monpa",
         added: "2009-07-29"}).
iana(language,kkg,tag{
         type: "language",
         subtag: "kkg",
         description: "Mabaka Valley Kalinga",
         added: "2009-07-29"}).
iana(language,kkh,tag{
         type: "language",
         subtag: "kkh",
         description: "Khün",
         added: "2009-07-29"}).
iana(language,kki,tag{
         type: "language",
         subtag: "kki",
         description: "Kagulu",
         added: "2009-07-29"}).
iana(language,kkj,tag{
         type: "language",
         subtag: "kkj",
         description: "Kako",
         added: "2009-07-29"}).
iana(language,kkk,tag{
         type: "language",
         subtag: "kkk",
         description: "Kokota",
         added: "2009-07-29"}).
iana(language,kkl,tag{
         type: "language",
         subtag: "kkl",
         description: "Kosarek Yale",
         added: "2009-07-29"}).
iana(language,kkm,tag{
         type: "language",
         subtag: "kkm",
         description: "Kiong",
         added: "2009-07-29"}).
iana(language,kkn,tag{
         type: "language",
         subtag: "kkn",
         description: "Kon Keu",
         added: "2009-07-29"}).
iana(language,kko,tag{
         type: "language",
         subtag: "kko",
         description: "Karko",
         added: "2009-07-29"}).
iana(language,kkp,tag{
         type: "language",
         subtag: "kkp",
         description: ["Gugubera","Koko-Bera"],
         added: "2009-07-29"}).
iana(language,kkq,tag{
         type: "language",
         subtag: "kkq",
         description: "Kaeku",
         added: "2009-07-29"}).
iana(language,kkr,tag{
         type: "language",
         subtag: "kkr",
         description: "Kir-Balar",
         added: "2009-07-29"}).
iana(language,kks,tag{
         type: "language",
         subtag: "kks",
         description: "Giiwo",
         added: "2009-07-29"}).
iana(language,kkt,tag{
         type: "language",
         subtag: "kkt",
         description: "Koi",
         added: "2009-07-29"}).
iana(language,kku,tag{
         type: "language",
         subtag: "kku",
         description: "Tumi",
         added: "2009-07-29"}).
iana(language,kkv,tag{
         type: "language",
         subtag: "kkv",
         description: "Kangean",
         added: "2009-07-29"}).
iana(language,kkw,tag{
         type: "language",
         subtag: "kkw",
         description: "Teke-Kukuya",
         added: "2009-07-29"}).
iana(language,kkx,tag{
         type: "language",
         subtag: "kkx",
         description: "Kohin",
         added: "2009-07-29"}).
iana(language,kky,tag{
         type: "language",
         subtag: "kky",
         description: ["Guugu Yimidhirr","Guguyimidjir"],
         added: "2009-07-29"}).
iana(language,kkz,tag{
         type: "language",
         subtag: "kkz",
         description: "Kaska",
         added: "2009-07-29"}).
iana(language,kla,tag{
         type: "language",
         subtag: "kla",
         description: "Klamath-Modoc",
         added: "2009-07-29"}).
iana(language,klb,tag{
         type: "language",
         subtag: "klb",
         description: "Kiliwa",
         added: "2009-07-29"}).
iana(language,klc,tag{
         type: "language",
         subtag: "klc",
         description: "Kolbila",
         added: "2009-07-29"}).
iana(language,kld,tag{
         type: "language",
         subtag: "kld",
         description: "Gamilaraay",
         added: "2009-07-29"}).
iana(language,kle,tag{
         type: "language",
         subtag: "kle",
         description: "Kulung (Nepal)",
         added: "2009-07-29"}).
iana(language,klf,tag{
         type: "language",
         subtag: "klf",
         description: "Kendeje",
         added: "2009-07-29"}).
iana(language,klg,tag{
         type: "language",
         subtag: "klg",
         description: "Tagakaulo",
         added: "2009-07-29"}).
iana(language,klh,tag{
         type: "language",
         subtag: "klh",
         description: "Weliki",
         added: "2009-07-29"}).
iana(language,kli,tag{
         type: "language",
         subtag: "kli",
         description: "Kalumpang",
         added: "2009-07-29"}).
iana(language,klj,tag{
         type: "language",
         subtag: "klj",
         description: "Khalaj",
         added: "2009-07-29"}).
iana(language,klk,tag{
         type: "language",
         subtag: "klk",
         description: "Kono (Nigeria)",
         added: "2009-07-29"}).
iana(language,kll,tag{
         type: "language",
         subtag: "kll",
         description: "Kagan Kalagan",
         added: "2009-07-29"}).
iana(language,klm,tag{
         type: "language",
         subtag: "klm",
         description: "Migum",
         added: "2009-07-29"}).
iana(language,kln,tag{
         type: "language",
         subtag: "kln",
         description: "Kalenjin",
         added: "2009-07-29",
         scope: "macrolanguage"}).
iana(language,klo,tag{
         type: "language",
         subtag: "klo",
         description: "Kapya",
         added: "2009-07-29"}).
iana(language,klp,tag{
         type: "language",
         subtag: "klp",
         description: "Kamasa",
         added: "2009-07-29"}).
iana(language,klq,tag{
         type: "language",
         subtag: "klq",
         description: "Rumu",
         added: "2009-07-29"}).
iana(language,klr,tag{
         type: "language",
         subtag: "klr",
         description: "Khaling",
         added: "2009-07-29"}).
iana(language,kls,tag{
         type: "language",
         subtag: "kls",
         description: "Kalasha",
         added: "2009-07-29"}).
iana(language,klt,tag{
         type: "language",
         subtag: "klt",
         description: "Nukna",
         added: "2009-07-29"}).
iana(language,klu,tag{
         type: "language",
         subtag: "klu",
         description: "Klao",
         added: "2009-07-29"}).
iana(language,klv,tag{
         type: "language",
         subtag: "klv",
         description: "Maskelynes",
         added: "2009-07-29"}).
iana(language,klw,tag{
         type: "language",
         subtag: "klw",
         description: ["Tado","Lindu"],
         added: "2009-07-29"}).
iana(language,klx,tag{
         type: "language",
         subtag: "klx",
         description: "Koluwawa",
         added: "2009-07-29"}).
iana(language,kly,tag{
         type: "language",
         subtag: "kly",
         description: "Kalao",
         added: "2009-07-29"}).
iana(language,klz,tag{
         type: "language",
         subtag: "klz",
         description: "Kabola",
         added: "2009-07-29"}).
iana(language,kma,tag{
         type: "language",
         subtag: "kma",
         description: "Konni",
         added: "2009-07-29"}).
iana(language,kmb,tag{
         type: "language",
         subtag: "kmb",
         description: "Kimbundu",
         added: "2005-10-16"}).
iana(language,kmc,tag{
         type: "language",
         subtag: "kmc",
         description: "Southern Dong",
         added: "2009-07-29"}).
iana(language,kmd,tag{
         type: "language",
         subtag: "kmd",
         description: "Majukayang Kalinga",
         added: "2009-07-29"}).
iana(language,kme,tag{
         type: "language",
         subtag: "kme",
         description: "Bakole",
         added: "2009-07-29"}).
iana(language,kmf,tag{
         type: "language",
         subtag: "kmf",
         description: "Kare (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,kmg,tag{
         type: "language",
         subtag: "kmg",
         description: "Kâte",
         added: "2009-07-29"}).
iana(language,kmh,tag{
         type: "language",
         subtag: "kmh",
         description: "Kalam",
         added: "2009-07-29"}).
iana(language,kmi,tag{
         type: "language",
         subtag: "kmi",
         description: "Kami (Nigeria)",
         added: "2009-07-29"}).
iana(language,kmj,tag{
         type: "language",
         subtag: "kmj",
         description: "Kumarbhag Paharia",
         added: "2009-07-29"}).
iana(language,kmk,tag{
         type: "language",
         subtag: "kmk",
         description: "Limos Kalinga",
         added: "2009-07-29"}).
iana(language,kml,tag{
         type: "language",
         subtag: "kml",
         description: "Tanudan Kalinga",
         added: "2009-07-29"}).
iana(language,kmm,tag{
         type: "language",
         subtag: "kmm",
         description: "Kom (India)",
         added: "2009-07-29"}).
iana(language,kmn,tag{
         type: "language",
         subtag: "kmn",
         description: "Awtuw",
         added: "2009-07-29"}).
iana(language,kmo,tag{
         type: "language",
         subtag: "kmo",
         description: "Kwoma",
         added: "2009-07-29"}).
iana(language,kmp,tag{
         type: "language",
         subtag: "kmp",
         description: "Gimme",
         added: "2009-07-29"}).
iana(language,kmq,tag{
         type: "language",
         subtag: "kmq",
         description: "Kwama",
         added: "2009-07-29"}).
iana(language,kmr,tag{
         type: "language",
         subtag: "kmr",
         description: "Northern Kurdish",
         added: "2009-07-29",
         macrolanguage: "ku"}).
iana(language,kms,tag{
         type: "language",
         subtag: "kms",
         description: "Kamasau",
         added: "2009-07-29"}).
iana(language,kmt,tag{
         type: "language",
         subtag: "kmt",
         description: "Kemtuik",
         added: "2009-07-29"}).
iana(language,kmu,tag{
         type: "language",
         subtag: "kmu",
         description: "Kanite",
         added: "2009-07-29"}).
iana(language,kmv,tag{
         type: "language",
         subtag: "kmv",
         description: "Karipúna Creole French",
         added: "2009-07-29"}).
iana(language,kmw,tag{
         type: "language",
         subtag: "kmw",
         description: "Komo (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,kmx,tag{
         type: "language",
         subtag: "kmx",
         description: "Waboda",
         added: "2009-07-29"}).
iana(language,kmy,tag{
         type: "language",
         subtag: "kmy",
         description: "Koma",
         added: "2009-07-29"}).
iana(language,kmz,tag{
         type: "language",
         subtag: "kmz",
         description: "Khorasani Turkish",
         added: "2009-07-29"}).
iana(language,kna,tag{
         type: "language",
         subtag: "kna",
         description: "Dera (Nigeria)",
         added: "2009-07-29"}).
iana(language,knb,tag{
         type: "language",
         subtag: "knb",
         description: "Lubuagan Kalinga",
         added: "2009-07-29"}).
iana(language,knc,tag{
         type: "language",
         subtag: "knc",
         description: "Central Kanuri",
         added: "2009-07-29",
         macrolanguage: "kr"}).
iana(language,knd,tag{
         type: "language",
         subtag: "knd",
         description: "Konda",
         added: "2009-07-29"}).
iana(language,kne,tag{
         type: "language",
         subtag: "kne",
         description: "Kankanaey",
         added: "2009-07-29"}).
iana(language,knf,tag{
         type: "language",
         subtag: "knf",
         description: "Mankanya",
         added: "2009-07-29"}).
iana(language,kng,tag{
         type: "language",
         subtag: "kng",
         description: "Koongo",
         added: "2009-07-29",
         macrolanguage: "kg"}).
iana(language,kni,tag{
         type: "language",
         subtag: "kni",
         description: "Kanufi",
         added: "2009-07-29"}).
iana(language,knj,tag{
         type: "language",
         subtag: "knj",
         description: "Western Kanjobal",
         added: "2009-07-29"}).
iana(language,knk,tag{
         type: "language",
         subtag: "knk",
         description: "Kuranko",
         added: "2009-07-29"}).
iana(language,knl,tag{
         type: "language",
         subtag: "knl",
         description: "Keninjal",
         added: "2009-07-29"}).
iana(language,knm,tag{
         type: "language",
         subtag: "knm",
         description: "Kanamarí",
         added: "2009-07-29"}).
iana(language,knn,tag{
         type: "language",
         subtag: "knn",
         description: "Konkani (individual language)",
         added: "2009-07-29",
         macrolanguage: "kok"}).
iana(language,kno,tag{
         type: "language",
         subtag: "kno",
         description: "Kono (Sierra Leone)",
         added: "2009-07-29"}).
iana(language,knp,tag{
         type: "language",
         subtag: "knp",
         description: "Kwanja",
         added: "2009-07-29"}).
iana(language,knq,tag{
         type: "language",
         subtag: "knq",
         description: "Kintaq",
         added: "2009-07-29"}).
iana(language,knr,tag{
         type: "language",
         subtag: "knr",
         description: "Kaningra",
         added: "2009-07-29"}).
iana(language,kns,tag{
         type: "language",
         subtag: "kns",
         description: "Kensiu",
         added: "2009-07-29"}).
iana(language,knt,tag{
         type: "language",
         subtag: "knt",
         description: "Panoan Katukína",
         added: "2009-07-29"}).
iana(language,knu,tag{
         type: "language",
         subtag: "knu",
         description: "Kono (Guinea)",
         added: "2009-07-29"}).
iana(language,knv,tag{
         type: "language",
         subtag: "knv",
         description: "Tabo",
         added: "2009-07-29"}).
iana(language,knw,tag{
         type: "language",
         subtag: "knw",
         description: "Kung-Ekoka",
         added: "2009-07-29"}).
iana(language,knx,tag{
         type: "language",
         subtag: "knx",
         description: ["Kendayan","Salako"],
         added: "2009-07-29"}).
iana(language,kny,tag{
         type: "language",
         subtag: "kny",
         description: "Kanyok",
         added: "2009-07-29"}).
iana(language,knz,tag{
         type: "language",
         subtag: "knz",
         description: "Kalamsé",
         added: "2009-07-29"}).
iana(language,koa,tag{
         type: "language",
         subtag: "koa",
         description: "Konomala",
         added: "2009-07-29"}).
iana(language,koc,tag{
         type: "language",
         subtag: "koc",
         description: "Kpati",
         added: "2009-07-29"}).
iana(language,kod,tag{
         type: "language",
         subtag: "kod",
         description: "Kodi",
         added: "2009-07-29"}).
iana(language,koe,tag{
         type: "language",
         subtag: "koe",
         description: "Kacipo-Bale Suri",
         added: "2009-07-29"}).
iana(language,kof,tag{
         type: "language",
         subtag: "kof",
         description: "Kubi",
         added: "2009-07-29"}).
iana(language,kog,tag{
         type: "language",
         subtag: "kog",
         description: ["Cogui","Kogi"],
         added: "2009-07-29"}).
iana(language,koh,tag{
         type: "language",
         subtag: "koh",
         description: "Koyo",
         added: "2009-07-29"}).
iana(language,koi,tag{
         type: "language",
         subtag: "koi",
         description: "Komi-Permyak",
         added: "2009-07-29",
         macrolanguage: "kv"}).
iana(language,koj,tag{
         type: "language",
         subtag: "koj",
         description: "Sara Dunjo",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "kwv"}).
iana(language,kok,tag{
         type: "language",
         subtag: "kok",
         description: "Konkani (macrolanguage)",
         added: "2005-10-16",
         suppress_script: "Deva",
         scope: "macrolanguage"}).
iana(language,kol,tag{
         type: "language",
         subtag: "kol",
         description: "Kol (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,koo,tag{
         type: "language",
         subtag: "koo",
         description: "Konzo",
         added: "2009-07-29"}).
iana(language,kop,tag{
         type: "language",
         subtag: "kop",
         description: "Waube",
         added: "2009-07-29"}).
iana(language,koq,tag{
         type: "language",
         subtag: "koq",
         description: "Kota (Gabon)",
         added: "2009-07-29"}).
iana(language,kos,tag{
         type: "language",
         subtag: "kos",
         description: "Kosraean",
         added: "2005-10-16"}).
iana(language,kot,tag{
         type: "language",
         subtag: "kot",
         description: "Lagwan",
         added: "2009-07-29"}).
iana(language,kou,tag{
         type: "language",
         subtag: "kou",
         description: "Koke",
         added: "2009-07-29"}).
iana(language,kov,tag{
         type: "language",
         subtag: "kov",
         description: "Kudu-Camo",
         added: "2009-07-29"}).
iana(language,kow,tag{
         type: "language",
         subtag: "kow",
         description: "Kugama",
         added: "2009-07-29"}).
iana(language,kox,tag{
         type: "language",
         subtag: "kox",
         description: "Coxima",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,koy,tag{
         type: "language",
         subtag: "koy",
         description: "Koyukon",
         added: "2009-07-29"}).
iana(language,koz,tag{
         type: "language",
         subtag: "koz",
         description: "Korak",
         added: "2009-07-29"}).
iana(language,kpa,tag{
         type: "language",
         subtag: "kpa",
         description: "Kutto",
         added: "2009-07-29"}).
iana(language,kpb,tag{
         type: "language",
         subtag: "kpb",
         description: "Mullu Kurumba",
         added: "2009-07-29"}).
iana(language,kpc,tag{
         type: "language",
         subtag: "kpc",
         description: "Curripaco",
         added: "2009-07-29"}).
iana(language,kpd,tag{
         type: "language",
         subtag: "kpd",
         description: "Koba",
         added: "2009-07-29"}).
iana(language,kpe,tag{
         type: "language",
         subtag: "kpe",
         description: "Kpelle",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,kpf,tag{
         type: "language",
         subtag: "kpf",
         description: "Komba",
         added: "2009-07-29"}).
iana(language,kpg,tag{
         type: "language",
         subtag: "kpg",
         description: "Kapingamarangi",
         added: "2009-07-29"}).
iana(language,kph,tag{
         type: "language",
         subtag: "kph",
         description: "Kplang",
         added: "2009-07-29"}).
iana(language,kpi,tag{
         type: "language",
         subtag: "kpi",
         description: "Kofei",
         added: "2009-07-29"}).
iana(language,kpj,tag{
         type: "language",
         subtag: "kpj",
         description: "Karajá",
         added: "2009-07-29"}).
iana(language,kpk,tag{
         type: "language",
         subtag: "kpk",
         description: "Kpan",
         added: "2009-07-29"}).
iana(language,kpl,tag{
         type: "language",
         subtag: "kpl",
         description: "Kpala",
         added: "2009-07-29"}).
iana(language,kpm,tag{
         type: "language",
         subtag: "kpm",
         description: "Koho",
         added: "2009-07-29"}).
iana(language,kpn,tag{
         type: "language",
         subtag: "kpn",
         description: "Kepkiriwát",
         added: "2009-07-29"}).
iana(language,kpo,tag{
         type: "language",
         subtag: "kpo",
         description: "Ikposo",
         added: "2009-07-29"}).
iana(language,kpp,tag{
         type: "language",
         subtag: "kpp",
         description: "Paku Karen",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see jkm, jkp"}).
iana(language,kpq,tag{
         type: "language",
         subtag: "kpq",
         description: "Korupun-Sela",
         added: "2009-07-29"}).
iana(language,kpr,tag{
         type: "language",
         subtag: "kpr",
         description: "Korafe-Yegha",
         added: "2009-07-29"}).
iana(language,kps,tag{
         type: "language",
         subtag: "kps",
         description: "Tehit",
         added: "2009-07-29"}).
iana(language,kpt,tag{
         type: "language",
         subtag: "kpt",
         description: "Karata",
         added: "2009-07-29"}).
iana(language,kpu,tag{
         type: "language",
         subtag: "kpu",
         description: "Kafoa",
         added: "2009-07-29"}).
iana(language,kpv,tag{
         type: "language",
         subtag: "kpv",
         description: "Komi-Zyrian",
         added: "2009-07-29",
         macrolanguage: "kv"}).
iana(language,kpw,tag{
         type: "language",
         subtag: "kpw",
         description: "Kobon",
         added: "2009-07-29"}).
iana(language,kpx,tag{
         type: "language",
         subtag: "kpx",
         description: "Mountain Koiali",
         added: "2009-07-29"}).
iana(language,kpy,tag{
         type: "language",
         subtag: "kpy",
         description: "Koryak",
         added: "2009-07-29"}).
iana(language,kpz,tag{
         type: "language",
         subtag: "kpz",
         description: "Kupsabiny",
         added: "2009-07-29"}).
iana(language,kqa,tag{
         type: "language",
         subtag: "kqa",
         description: "Mum",
         added: "2009-07-29"}).
iana(language,kqb,tag{
         type: "language",
         subtag: "kqb",
         description: "Kovai",
         added: "2009-07-29"}).
iana(language,kqc,tag{
         type: "language",
         subtag: "kqc",
         description: "Doromu-Koki",
         added: "2009-07-29"}).
iana(language,kqd,tag{
         type: "language",
         subtag: "kqd",
         description: "Koy Sanjaq Surat",
         added: "2009-07-29"}).
iana(language,kqe,tag{
         type: "language",
         subtag: "kqe",
         description: "Kalagan",
         added: "2009-07-29"}).
iana(language,kqf,tag{
         type: "language",
         subtag: "kqf",
         description: "Kakabai",
         added: "2009-07-29"}).
iana(language,kqg,tag{
         type: "language",
         subtag: "kqg",
         description: "Khe",
         added: "2009-07-29"}).
iana(language,kqh,tag{
         type: "language",
         subtag: "kqh",
         description: "Kisankasa",
         added: "2009-07-29"}).
iana(language,kqi,tag{
         type: "language",
         subtag: "kqi",
         description: "Koitabu",
         added: "2009-07-29"}).
iana(language,kqj,tag{
         type: "language",
         subtag: "kqj",
         description: "Koromira",
         added: "2009-07-29"}).
iana(language,kqk,tag{
         type: "language",
         subtag: "kqk",
         description: "Kotafon Gbe",
         added: "2009-07-29"}).
iana(language,kql,tag{
         type: "language",
         subtag: "kql",
         description: "Kyenele",
         added: "2009-07-29"}).
iana(language,kqm,tag{
         type: "language",
         subtag: "kqm",
         description: "Khisa",
         added: "2009-07-29"}).
iana(language,kqn,tag{
         type: "language",
         subtag: "kqn",
         description: "Kaonde",
         added: "2009-07-29"}).
iana(language,kqo,tag{
         type: "language",
         subtag: "kqo",
         description: "Eastern Krahn",
         added: "2009-07-29"}).
iana(language,kqp,tag{
         type: "language",
         subtag: "kqp",
         description: "Kimré",
         added: "2009-07-29"}).
iana(language,kqq,tag{
         type: "language",
         subtag: "kqq",
         description: "Krenak",
         added: "2009-07-29"}).
iana(language,kqr,tag{
         type: "language",
         subtag: "kqr",
         description: "Kimaragang",
         added: "2009-07-29"}).
iana(language,kqs,tag{
         type: "language",
         subtag: "kqs",
         description: "Northern Kissi",
         added: "2009-07-29"}).
iana(language,kqt,tag{
         type: "language",
         subtag: "kqt",
         description: "Klias River Kadazan",
         added: "2009-07-29"}).
iana(language,kqu,tag{
         type: "language",
         subtag: "kqu",
         description: "Seroa",
         added: "2009-07-29"}).
iana(language,kqv,tag{
         type: "language",
         subtag: "kqv",
         description: "Okolod",
         added: "2009-07-29"}).
iana(language,kqw,tag{
         type: "language",
         subtag: "kqw",
         description: "Kandas",
         added: "2009-07-29"}).
iana(language,kqx,tag{
         type: "language",
         subtag: "kqx",
         description: "Mser",
         added: "2009-07-29"}).
iana(language,kqy,tag{
         type: "language",
         subtag: "kqy",
         description: "Koorete",
         added: "2009-07-29"}).
iana(language,kqz,tag{
         type: "language",
         subtag: "kqz",
         description: "Korana",
         added: "2009-07-29"}).
iana(language,kra,tag{
         type: "language",
         subtag: "kra",
         description: "Kumhali",
         added: "2009-07-29"}).
iana(language,krb,tag{
         type: "language",
         subtag: "krb",
         description: "Karkin",
         added: "2009-07-29"}).
iana(language,krc,tag{
         type: "language",
         subtag: "krc",
         description: "Karachay-Balkar",
         added: "2005-10-16"}).
iana(language,krd,tag{
         type: "language",
         subtag: "krd",
         description: "Kairui-Midiki",
         added: "2009-07-29"}).
iana(language,kre,tag{
         type: "language",
         subtag: "kre",
         description: "Panará",
         added: "2009-07-29"}).
iana(language,krf,tag{
         type: "language",
         subtag: "krf",
         description: "Koro (Vanuatu)",
         added: "2009-07-29"}).
iana(language,krh,tag{
         type: "language",
         subtag: "krh",
         description: "Kurama",
         added: "2009-07-29"}).
iana(language,kri,tag{
         type: "language",
         subtag: "kri",
         description: "Krio",
         added: "2009-07-29"}).
iana(language,krj,tag{
         type: "language",
         subtag: "krj",
         description: "Kinaray-A",
         added: "2009-07-29"}).
iana(language,krk,tag{
         type: "language",
         subtag: "krk",
         description: "Kerek",
         added: "2009-07-29"}).
iana(language,krl,tag{
         type: "language",
         subtag: "krl",
         description: "Karelian",
         added: "2006-03-08"}).
iana(language,krm,tag{
         type: "language",
         subtag: "krm",
         description: "Krim",
         added: "2009-07-29",
         deprecated: "2017-02-23",
         preferred_value: "bmf"}).
iana(language,krn,tag{
         type: "language",
         subtag: "krn",
         description: "Sapo",
         added: "2009-07-29"}).
iana(language,kro,tag{
         type: "language",
         subtag: "kro",
         description: "Kru languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,krp,tag{
         type: "language",
         subtag: "krp",
         description: "Korop",
         added: "2009-07-29"}).
iana(language,krr,tag{
         type: "language",
         subtag: "krr",
         description: "Krung",
         added: "2009-07-29"}).
iana(language,krs,tag{
         type: "language",
         subtag: "krs",
         description: "Gbaya (Sudan)",
         added: "2009-07-29"}).
iana(language,krt,tag{
         type: "language",
         subtag: "krt",
         description: "Tumari Kanuri",
         added: "2009-07-29",
         macrolanguage: "kr"}).
iana(language,kru,tag{
         type: "language",
         subtag: "kru",
         description: "Kurukh",
         added: "2005-10-16"}).
iana(language,krv,tag{
         type: "language",
         subtag: "krv",
         description: "Kavet",
         added: "2009-07-29"}).
iana(language,krw,tag{
         type: "language",
         subtag: "krw",
         description: "Western Krahn",
         added: "2009-07-29"}).
iana(language,krx,tag{
         type: "language",
         subtag: "krx",
         description: "Karon",
         added: "2009-07-29"}).
iana(language,kry,tag{
         type: "language",
         subtag: "kry",
         description: "Kryts",
         added: "2009-07-29"}).
iana(language,krz,tag{
         type: "language",
         subtag: "krz",
         description: "Sota Kanum",
         added: "2009-07-29"}).
iana(language,ksa,tag{
         type: "language",
         subtag: "ksa",
         description: "Shuwa-Zamani",
         added: "2009-07-29"}).
iana(language,ksb,tag{
         type: "language",
         subtag: "ksb",
         description: "Shambala",
         added: "2009-07-29"}).
iana(language,ksc,tag{
         type: "language",
         subtag: "ksc",
         description: "Southern Kalinga",
         added: "2009-07-29"}).
iana(language,ksd,tag{
         type: "language",
         subtag: "ksd",
         description: "Kuanua",
         added: "2009-07-29"}).
iana(language,kse,tag{
         type: "language",
         subtag: "kse",
         description: "Kuni",
         added: "2009-07-29"}).
iana(language,ksf,tag{
         type: "language",
         subtag: "ksf",
         description: "Bafia",
         added: "2009-07-29"}).
iana(language,ksg,tag{
         type: "language",
         subtag: "ksg",
         description: "Kusaghe",
         added: "2009-07-29"}).
iana(language,ksh,tag{
         type: "language",
         subtag: "ksh",
         description: "Kölsch",
         added: "2009-07-29"}).
iana(language,ksi,tag{
         type: "language",
         subtag: "ksi",
         description: ["Krisa","I'saka"],
         added: "2009-07-29"}).
iana(language,ksj,tag{
         type: "language",
         subtag: "ksj",
         description: "Uare",
         added: "2009-07-29"}).
iana(language,ksk,tag{
         type: "language",
         subtag: "ksk",
         description: "Kansa",
         added: "2009-07-29"}).
iana(language,ksl,tag{
         type: "language",
         subtag: "ksl",
         description: "Kumalu",
         added: "2009-07-29"}).
iana(language,ksm,tag{
         type: "language",
         subtag: "ksm",
         description: "Kumba",
         added: "2009-07-29"}).
iana(language,ksn,tag{
         type: "language",
         subtag: "ksn",
         description: "Kasiguranin",
         added: "2009-07-29"}).
iana(language,kso,tag{
         type: "language",
         subtag: "kso",
         description: "Kofa",
         added: "2009-07-29"}).
iana(language,ksp,tag{
         type: "language",
         subtag: "ksp",
         description: "Kaba",
         added: "2009-07-29"}).
iana(language,ksq,tag{
         type: "language",
         subtag: "ksq",
         description: "Kwaami",
         added: "2009-07-29"}).
iana(language,ksr,tag{
         type: "language",
         subtag: "ksr",
         description: "Borong",
         added: "2009-07-29"}).
iana(language,kss,tag{
         type: "language",
         subtag: "kss",
         description: "Southern Kisi",
         added: "2009-07-29"}).
iana(language,kst,tag{
         type: "language",
         subtag: "kst",
         description: "Winyé",
         added: "2009-07-29"}).
iana(language,ksu,tag{
         type: "language",
         subtag: "ksu",
         description: "Khamyang",
         added: "2009-07-29"}).
iana(language,ksv,tag{
         type: "language",
         subtag: "ksv",
         description: "Kusu",
         added: "2009-07-29"}).
iana(language,ksw,tag{
         type: "language",
         subtag: "ksw",
         description: "S'gaw Karen",
         added: "2009-07-29"}).
iana(language,ksx,tag{
         type: "language",
         subtag: "ksx",
         description: "Kedang",
         added: "2009-07-29"}).
iana(language,ksy,tag{
         type: "language",
         subtag: "ksy",
         description: "Kharia Thar",
         added: "2009-07-29"}).
iana(language,ksz,tag{
         type: "language",
         subtag: "ksz",
         description: "Kodaku",
         added: "2009-07-29"}).
iana(language,kta,tag{
         type: "language",
         subtag: "kta",
         description: "Katua",
         added: "2009-07-29"}).
iana(language,ktb,tag{
         type: "language",
         subtag: "ktb",
         description: "Kambaata",
         added: "2009-07-29"}).
iana(language,ktc,tag{
         type: "language",
         subtag: "ktc",
         description: "Kholok",
         added: "2009-07-29"}).
iana(language,ktd,tag{
         type: "language",
         subtag: "ktd",
         description: ["Kokata","Kukatha"],
         added: "2009-07-29"}).
iana(language,kte,tag{
         type: "language",
         subtag: "kte",
         description: "Nubri",
         added: "2009-07-29"}).
iana(language,ktf,tag{
         type: "language",
         subtag: "ktf",
         description: "Kwami",
         added: "2009-07-29"}).
iana(language,ktg,tag{
         type: "language",
         subtag: "ktg",
         description: "Kalkutung",
         added: "2009-07-29"}).
iana(language,kth,tag{
         type: "language",
         subtag: "kth",
         description: "Karanga",
         added: "2009-07-29"}).
iana(language,kti,tag{
         type: "language",
         subtag: "kti",
         description: "North Muyu",
         added: "2009-07-29"}).
iana(language,ktj,tag{
         type: "language",
         subtag: "ktj",
         description: "Plapo Krumen",
         added: "2009-07-29"}).
iana(language,ktk,tag{
         type: "language",
         subtag: "ktk",
         description: "Kaniet",
         added: "2009-07-29"}).
iana(language,ktl,tag{
         type: "language",
         subtag: "ktl",
         description: "Koroshi",
         added: "2009-07-29"}).
iana(language,ktm,tag{
         type: "language",
         subtag: "ktm",
         description: "Kurti",
         added: "2009-07-29"}).
iana(language,ktn,tag{
         type: "language",
         subtag: "ktn",
         description: "Karitiâna",
         added: "2009-07-29"}).
iana(language,kto,tag{
         type: "language",
         subtag: "kto",
         description: "Kuot",
         added: "2009-07-29"}).
iana(language,ktp,tag{
         type: "language",
         subtag: "ktp",
         description: "Kaduo",
         added: "2009-07-29"}).
iana(language,ktq,tag{
         type: "language",
         subtag: "ktq",
         description: "Katabaga",
         added: "2009-07-29"}).
iana(language,ktr,tag{
         type: "language",
         subtag: "ktr",
         description: "Kota Marudu Tinagas",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "dtp"}).
iana(language,kts,tag{
         type: "language",
         subtag: "kts",
         description: "South Muyu",
         added: "2009-07-29"}).
iana(language,ktt,tag{
         type: "language",
         subtag: "ktt",
         description: "Ketum",
         added: "2009-07-29"}).
iana(language,ktu,tag{
         type: "language",
         subtag: "ktu",
         description: "Kituba (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,ktv,tag{
         type: "language",
         subtag: "ktv",
         description: "Eastern Katu",
         added: "2009-07-29"}).
iana(language,ktw,tag{
         type: "language",
         subtag: "ktw",
         description: "Kato",
         added: "2009-07-29"}).
iana(language,ktx,tag{
         type: "language",
         subtag: "ktx",
         description: "Kaxararí",
         added: "2009-07-29"}).
iana(language,kty,tag{
         type: "language",
         subtag: "kty",
         description: "Kango (Bas-Uélé District)",
         added: "2009-07-29"}).
iana(language,ktz,tag{
         type: "language",
         subtag: "ktz",
         description: ["Juǀʼhoan","Juǀʼhoansi"],
         added: "2009-07-29"}).
iana(language,kub,tag{
         type: "language",
         subtag: "kub",
         description: "Kutep",
         added: "2009-07-29"}).
iana(language,kuc,tag{
         type: "language",
         subtag: "kuc",
         description: "Kwinsu",
         added: "2009-07-29"}).
iana(language,kud,tag{
         type: "language",
         subtag: "kud",
         description: "'Auhelawa",
         added: "2009-07-29"}).
iana(language,kue,tag{
         type: "language",
         subtag: "kue",
         description: "Kuman (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,kuf,tag{
         type: "language",
         subtag: "kuf",
         description: "Western Katu",
         added: "2009-07-29"}).
iana(language,kug,tag{
         type: "language",
         subtag: "kug",
         description: "Kupa",
         added: "2009-07-29"}).
iana(language,kuh,tag{
         type: "language",
         subtag: "kuh",
         description: "Kushi",
         added: "2009-07-29"}).
iana(language,kui,tag{
         type: "language",
         subtag: "kui",
         description: ["Kuikúro-Kalapálo","Kalapalo"],
         added: "2009-07-29"}).
iana(language,kuj,tag{
         type: "language",
         subtag: "kuj",
         description: "Kuria",
         added: "2009-07-29"}).
iana(language,kuk,tag{
         type: "language",
         subtag: "kuk",
         description: "Kepo'",
         added: "2009-07-29"}).
iana(language,kul,tag{
         type: "language",
         subtag: "kul",
         description: "Kulere",
         added: "2009-07-29"}).
iana(language,kum,tag{
         type: "language",
         subtag: "kum",
         description: "Kumyk",
         added: "2005-10-16"}).
iana(language,kun,tag{
         type: "language",
         subtag: "kun",
         description: "Kunama",
         added: "2009-07-29"}).
iana(language,kuo,tag{
         type: "language",
         subtag: "kuo",
         description: "Kumukio",
         added: "2009-07-29"}).
iana(language,kup,tag{
         type: "language",
         subtag: "kup",
         description: "Kunimaipa",
         added: "2009-07-29"}).
iana(language,kuq,tag{
         type: "language",
         subtag: "kuq",
         description: "Karipuna",
         added: "2009-07-29"}).
iana(language,kus,tag{
         type: "language",
         subtag: "kus",
         description: "Kusaal",
         added: "2009-07-29"}).
iana(language,kut,tag{
         type: "language",
         subtag: "kut",
         description: "Kutenai",
         added: "2005-10-16"}).
iana(language,kuu,tag{
         type: "language",
         subtag: "kuu",
         description: "Upper Kuskokwim",
         added: "2009-07-29"}).
iana(language,kuv,tag{
         type: "language",
         subtag: "kuv",
         description: "Kur",
         added: "2009-07-29"}).
iana(language,kuw,tag{
         type: "language",
         subtag: "kuw",
         description: "Kpagua",
         added: "2009-07-29"}).
iana(language,kux,tag{
         type: "language",
         subtag: "kux",
         description: "Kukatja",
         added: "2009-07-29"}).
iana(language,kuy,tag{
         type: "language",
         subtag: "kuy",
         description: "Kuuku-Ya'u",
         added: "2009-07-29"}).
iana(language,kuz,tag{
         type: "language",
         subtag: "kuz",
         description: "Kunza",
         added: "2009-07-29"}).
iana(language,kva,tag{
         type: "language",
         subtag: "kva",
         description: "Bagvalal",
         added: "2009-07-29"}).
iana(language,kvb,tag{
         type: "language",
         subtag: "kvb",
         description: "Kubu",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,kvc,tag{
         type: "language",
         subtag: "kvc",
         description: "Kove",
         added: "2009-07-29"}).
iana(language,kvd,tag{
         type: "language",
         subtag: "kvd",
         description: "Kui (Indonesia)",
         added: "2009-07-29"}).
iana(language,kve,tag{
         type: "language",
         subtag: "kve",
         description: "Kalabakan",
         added: "2009-07-29"}).
iana(language,kvf,tag{
         type: "language",
         subtag: "kvf",
         description: "Kabalai",
         added: "2009-07-29"}).
iana(language,kvg,tag{
         type: "language",
         subtag: "kvg",
         description: "Kuni-Boazi",
         added: "2009-07-29"}).
iana(language,kvh,tag{
         type: "language",
         subtag: "kvh",
         description: "Komodo",
         added: "2009-07-29"}).
iana(language,kvi,tag{
         type: "language",
         subtag: "kvi",
         description: "Kwang",
         added: "2009-07-29"}).
iana(language,kvj,tag{
         type: "language",
         subtag: "kvj",
         description: "Psikye",
         added: "2009-07-29"}).
iana(language,kvk,tag{
         type: "language",
         subtag: "kvk",
         description: "Korean Sign Language",
         added: "2009-07-29"}).
iana(language,kvl,tag{
         type: "language",
         subtag: "kvl",
         description: "Kayaw",
         added: "2009-07-29"}).
iana(language,kvm,tag{
         type: "language",
         subtag: "kvm",
         description: "Kendem",
         added: "2009-07-29"}).
iana(language,kvn,tag{
         type: "language",
         subtag: "kvn",
         description: "Border Kuna",
         added: "2009-07-29"}).
iana(language,kvo,tag{
         type: "language",
         subtag: "kvo",
         description: "Dobel",
         added: "2009-07-29"}).
iana(language,kvp,tag{
         type: "language",
         subtag: "kvp",
         description: "Kompane",
         added: "2009-07-29"}).
iana(language,kvq,tag{
         type: "language",
         subtag: "kvq",
         description: "Geba Karen",
         added: "2009-07-29"}).
iana(language,kvr,tag{
         type: "language",
         subtag: "kvr",
         description: "Kerinci",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,kvs,tag{
         type: "language",
         subtag: "kvs",
         description: "Kunggara",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "gdj"}).
iana(language,kvt,tag{
         type: "language",
         subtag: "kvt",
         description: ["Lahta Karen","Lahta"],
         added: "2009-07-29"}).
iana(language,kvu,tag{
         type: "language",
         subtag: "kvu",
         description: "Yinbaw Karen",
         added: "2009-07-29"}).
iana(language,kvv,tag{
         type: "language",
         subtag: "kvv",
         description: "Kola",
         added: "2009-07-29"}).
iana(language,kvw,tag{
         type: "language",
         subtag: "kvw",
         description: "Wersing",
         added: "2009-07-29"}).
iana(language,kvx,tag{
         type: "language",
         subtag: "kvx",
         description: "Parkari Koli",
         added: "2009-07-29"}).
iana(language,kvy,tag{
         type: "language",
         subtag: "kvy",
         description: ["Yintale Karen","Yintale"],
         added: "2009-07-29"}).
iana(language,kvz,tag{
         type: "language",
         subtag: "kvz",
         description: ["Tsakwambo","Tsaukambo"],
         added: "2009-07-29"}).
iana(language,kwa,tag{
         type: "language",
         subtag: "kwa",
         description: "Dâw",
         added: "2009-07-29"}).
iana(language,kwb,tag{
         type: "language",
         subtag: "kwb",
         description: "Kwa",
         added: "2009-07-29"}).
iana(language,kwc,tag{
         type: "language",
         subtag: "kwc",
         description: "Likwala",
         added: "2009-07-29"}).
iana(language,kwd,tag{
         type: "language",
         subtag: "kwd",
         description: "Kwaio",
         added: "2009-07-29"}).
iana(language,kwe,tag{
         type: "language",
         subtag: "kwe",
         description: "Kwerba",
         added: "2009-07-29"}).
iana(language,kwf,tag{
         type: "language",
         subtag: "kwf",
         description: "Kwara'ae",
         added: "2009-07-29"}).
iana(language,kwg,tag{
         type: "language",
         subtag: "kwg",
         description: "Sara Kaba Deme",
         added: "2009-07-29"}).
iana(language,kwh,tag{
         type: "language",
         subtag: "kwh",
         description: "Kowiai",
         added: "2009-07-29"}).
iana(language,kwi,tag{
         type: "language",
         subtag: "kwi",
         description: "Awa-Cuaiquer",
         added: "2009-07-29"}).
iana(language,kwj,tag{
         type: "language",
         subtag: "kwj",
         description: "Kwanga",
         added: "2009-07-29"}).
iana(language,kwk,tag{
         type: "language",
         subtag: "kwk",
         description: "Kwakiutl",
         added: "2009-07-29"}).
iana(language,kwl,tag{
         type: "language",
         subtag: "kwl",
         description: "Kofyar",
         added: "2009-07-29"}).
iana(language,kwm,tag{
         type: "language",
         subtag: "kwm",
         description: "Kwambi",
         added: "2009-07-29"}).
iana(language,kwn,tag{
         type: "language",
         subtag: "kwn",
         description: "Kwangali",
         added: "2009-07-29"}).
iana(language,kwo,tag{
         type: "language",
         subtag: "kwo",
         description: "Kwomtari",
         added: "2009-07-29"}).
iana(language,kwp,tag{
         type: "language",
         subtag: "kwp",
         description: "Kodia",
         added: "2009-07-29"}).
iana(language,kwq,tag{
         type: "language",
         subtag: "kwq",
         description: "Kwak",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "yam"}).
iana(language,kwr,tag{
         type: "language",
         subtag: "kwr",
         description: "Kwer",
         added: "2009-07-29"}).
iana(language,kws,tag{
         type: "language",
         subtag: "kws",
         description: "Kwese",
         added: "2009-07-29"}).
iana(language,kwt,tag{
         type: "language",
         subtag: "kwt",
         description: "Kwesten",
         added: "2009-07-29"}).
iana(language,kwu,tag{
         type: "language",
         subtag: "kwu",
         description: "Kwakum",
         added: "2009-07-29"}).
iana(language,kwv,tag{
         type: "language",
         subtag: "kwv",
         description: "Sara Kaba Náà",
         added: "2009-07-29"}).
iana(language,kww,tag{
         type: "language",
         subtag: "kww",
         description: "Kwinti",
         added: "2009-07-29"}).
iana(language,kwx,tag{
         type: "language",
         subtag: "kwx",
         description: "Khirwar",
         added: "2009-07-29"}).
iana(language,kwy,tag{
         type: "language",
         subtag: "kwy",
         description: "San Salvador Kongo",
         added: "2009-07-29",
         macrolanguage: "kg"}).
iana(language,kwz,tag{
         type: "language",
         subtag: "kwz",
         description: "Kwadi",
         added: "2009-07-29"}).
iana(language,kxa,tag{
         type: "language",
         subtag: "kxa",
         description: "Kairiru",
         added: "2009-07-29"}).
iana(language,kxb,tag{
         type: "language",
         subtag: "kxb",
         description: "Krobu",
         added: "2009-07-29"}).
iana(language,kxc,tag{
         type: "language",
         subtag: "kxc",
         description: ["Konso","Khonso"],
         added: "2009-07-29"}).
iana(language,kxd,tag{
         type: "language",
         subtag: "kxd",
         description: "Brunei",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,kxe,tag{
         type: "language",
         subtag: "kxe",
         description: "Kakihum",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "tvd"}).
iana(language,kxf,tag{
         type: "language",
         subtag: "kxf",
         description: ["Manumanaw Karen","Manumanaw"],
         added: "2009-07-29"}).
iana(language,kxh,tag{
         type: "language",
         subtag: "kxh",
         description: "Karo (Ethiopia)",
         added: "2009-07-29"}).
iana(language,kxi,tag{
         type: "language",
         subtag: "kxi",
         description: "Keningau Murut",
         added: "2009-07-29"}).
iana(language,kxj,tag{
         type: "language",
         subtag: "kxj",
         description: "Kulfa",
         added: "2009-07-29"}).
iana(language,kxk,tag{
         type: "language",
         subtag: "kxk",
         description: "Zayein Karen",
         added: "2009-07-29"}).
iana(language,kxl,tag{
         type: "language",
         subtag: "kxl",
         description: "Nepali Kurux",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "kru"}).
iana(language,kxm,tag{
         type: "language",
         subtag: "kxm",
         description: "Northern Khmer",
         added: "2009-07-29"}).
iana(language,kxn,tag{
         type: "language",
         subtag: "kxn",
         description: "Kanowit-Tanjong Melanau",
         added: "2009-07-29"}).
iana(language,kxo,tag{
         type: "language",
         subtag: "kxo",
         description: "Kanoé",
         added: "2009-07-29"}).
iana(language,kxp,tag{
         type: "language",
         subtag: "kxp",
         description: "Wadiyara Koli",
         added: "2009-07-29"}).
iana(language,kxq,tag{
         type: "language",
         subtag: "kxq",
         description: "Smärky Kanum",
         added: "2009-07-29"}).
iana(language,kxr,tag{
         type: "language",
         subtag: "kxr",
         description: "Koro (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,kxs,tag{
         type: "language",
         subtag: "kxs",
         description: "Kangjia",
         added: "2009-07-29"}).
iana(language,kxt,tag{
         type: "language",
         subtag: "kxt",
         description: "Koiwat",
         added: "2009-07-29"}).
iana(language,kxu,tag{
         type: "language",
         subtag: "kxu",
         description: "Kui (India)",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         comments: "see dwk, uki"}).
iana(language,kxv,tag{
         type: "language",
         subtag: "kxv",
         description: "Kuvi",
         added: "2009-07-29"}).
iana(language,kxw,tag{
         type: "language",
         subtag: "kxw",
         description: "Konai",
         added: "2009-07-29"}).
iana(language,kxx,tag{
         type: "language",
         subtag: "kxx",
         description: "Likuba",
         added: "2009-07-29"}).
iana(language,kxy,tag{
         type: "language",
         subtag: "kxy",
         description: "Kayong",
         added: "2009-07-29"}).
iana(language,kxz,tag{
         type: "language",
         subtag: "kxz",
         description: "Kerewo",
         added: "2009-07-29"}).
iana(language,kya,tag{
         type: "language",
         subtag: "kya",
         description: "Kwaya",
         added: "2009-07-29"}).
iana(language,kyb,tag{
         type: "language",
         subtag: "kyb",
         description: "Butbut Kalinga",
         added: "2009-07-29"}).
iana(language,kyc,tag{
         type: "language",
         subtag: "kyc",
         description: "Kyaka",
         added: "2009-07-29"}).
iana(language,kyd,tag{
         type: "language",
         subtag: "kyd",
         description: "Karey",
         added: "2009-07-29"}).
iana(language,kye,tag{
         type: "language",
         subtag: "kye",
         description: "Krache",
         added: "2009-07-29"}).
iana(language,kyf,tag{
         type: "language",
         subtag: "kyf",
         description: "Kouya",
         added: "2009-07-29"}).
iana(language,kyg,tag{
         type: "language",
         subtag: "kyg",
         description: "Keyagana",
         added: "2009-07-29"}).
iana(language,kyh,tag{
         type: "language",
         subtag: "kyh",
         description: "Karok",
         added: "2009-07-29"}).
iana(language,kyi,tag{
         type: "language",
         subtag: "kyi",
         description: "Kiput",
         added: "2009-07-29"}).
iana(language,kyj,tag{
         type: "language",
         subtag: "kyj",
         description: "Karao",
         added: "2009-07-29"}).
iana(language,kyk,tag{
         type: "language",
         subtag: "kyk",
         description: "Kamayo",
         added: "2009-07-29"}).
iana(language,kyl,tag{
         type: "language",
         subtag: "kyl",
         description: "Kalapuya",
         added: "2009-07-29"}).
iana(language,kym,tag{
         type: "language",
         subtag: "kym",
         description: "Kpatili",
         added: "2009-07-29"}).
iana(language,kyn,tag{
         type: "language",
         subtag: "kyn",
         description: "Northern Binukidnon",
         added: "2009-07-29"}).
iana(language,kyo,tag{
         type: "language",
         subtag: "kyo",
         description: "Kelon",
         added: "2009-07-29"}).
iana(language,kyp,tag{
         type: "language",
         subtag: "kyp",
         description: "Kang",
         added: "2009-07-29"}).
iana(language,kyq,tag{
         type: "language",
         subtag: "kyq",
         description: "Kenga",
         added: "2009-07-29"}).
iana(language,kyr,tag{
         type: "language",
         subtag: "kyr",
         description: "Kuruáya",
         added: "2009-07-29"}).
iana(language,kys,tag{
         type: "language",
         subtag: "kys",
         description: "Baram Kayan",
         added: "2009-07-29"}).
iana(language,kyt,tag{
         type: "language",
         subtag: "kyt",
         description: "Kayagar",
         added: "2009-07-29"}).
iana(language,kyu,tag{
         type: "language",
         subtag: "kyu",
         description: "Western Kayah",
         added: "2009-07-29"}).
iana(language,kyv,tag{
         type: "language",
         subtag: "kyv",
         description: "Kayort",
         added: "2009-07-29"}).
iana(language,kyw,tag{
         type: "language",
         subtag: "kyw",
         description: "Kudmali",
         added: "2009-07-29"}).
iana(language,kyx,tag{
         type: "language",
         subtag: "kyx",
         description: "Rapoisi",
         added: "2009-07-29"}).
iana(language,kyy,tag{
         type: "language",
         subtag: "kyy",
         description: "Kambaira",
         added: "2009-07-29"}).
iana(language,kyz,tag{
         type: "language",
         subtag: "kyz",
         description: "Kayabí",
         added: "2009-07-29"}).
iana(language,kza,tag{
         type: "language",
         subtag: "kza",
         description: "Western Karaboro",
         added: "2009-07-29"}).
iana(language,kzb,tag{
         type: "language",
         subtag: "kzb",
         description: "Kaibobo",
         added: "2009-07-29"}).
iana(language,kzc,tag{
         type: "language",
         subtag: "kzc",
         description: "Bondoukou Kulango",
         added: "2009-07-29"}).
iana(language,kzd,tag{
         type: "language",
         subtag: "kzd",
         description: "Kadai",
         added: "2009-07-29"}).
iana(language,kze,tag{
         type: "language",
         subtag: "kze",
         description: "Kosena",
         added: "2009-07-29"}).
iana(language,kzf,tag{
         type: "language",
         subtag: "kzf",
         description: "Da'a Kaili",
         added: "2009-07-29"}).
iana(language,kzg,tag{
         type: "language",
         subtag: "kzg",
         description: "Kikai",
         added: "2009-07-29"}).
iana(language,kzh,tag{
         type: "language",
         subtag: "kzh",
         description: "Kenuzi-Dongola",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see dgl, xnz"}).
iana(language,kzi,tag{
         type: "language",
         subtag: "kzi",
         description: "Kelabit",
         added: "2009-07-29"}).
iana(language,kzj,tag{
         type: "language",
         subtag: "kzj",
         description: "Coastal Kadazan",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "dtp"}).
iana(language,kzk,tag{
         type: "language",
         subtag: "kzk",
         description: "Kazukuru",
         added: "2009-07-29"}).
iana(language,kzl,tag{
         type: "language",
         subtag: "kzl",
         description: "Kayeli",
         added: "2009-07-29"}).
iana(language,kzm,tag{
         type: "language",
         subtag: "kzm",
         description: "Kais",
         added: "2009-07-29"}).
iana(language,kzn,tag{
         type: "language",
         subtag: "kzn",
         description: "Kokola",
         added: "2009-07-29"}).
iana(language,kzo,tag{
         type: "language",
         subtag: "kzo",
         description: "Kaningi",
         added: "2009-07-29"}).
iana(language,kzp,tag{
         type: "language",
         subtag: "kzp",
         description: "Kaidipang",
         added: "2009-07-29"}).
iana(language,kzq,tag{
         type: "language",
         subtag: "kzq",
         description: "Kaike",
         added: "2009-07-29"}).
iana(language,kzr,tag{
         type: "language",
         subtag: "kzr",
         description: "Karang",
         added: "2009-07-29"}).
iana(language,kzs,tag{
         type: "language",
         subtag: "kzs",
         description: "Sugut Dusun",
         added: "2009-07-29"}).
iana(language,kzt,tag{
         type: "language",
         subtag: "kzt",
         description: "Tambunan Dusun",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "dtp"}).
iana(language,kzu,tag{
         type: "language",
         subtag: "kzu",
         description: "Kayupulau",
         added: "2009-07-29"}).
iana(language,kzv,tag{
         type: "language",
         subtag: "kzv",
         description: "Komyandaret",
         added: "2009-07-29"}).
iana(language,kzw,tag{
         type: "language",
         subtag: "kzw",
         description: "Karirí-Xocó",
         added: "2009-07-29"}).
iana(language,kzx,tag{
         type: "language",
         subtag: "kzx",
         description: "Kamarian",
         added: "2009-07-29"}).
iana(language,kzy,tag{
         type: "language",
         subtag: "kzy",
         description: "Kango (Tshopo District)",
         added: "2009-07-29"}).
iana(language,kzz,tag{
         type: "language",
         subtag: "kzz",
         description: "Kalabra",
         added: "2009-07-29"}).
iana(language,laa,tag{
         type: "language",
         subtag: "laa",
         description: "Southern Subanen",
         added: "2009-07-29"}).
iana(language,lab,tag{
         type: "language",
         subtag: "lab",
         description: "Linear A",
         added: "2009-07-29"}).
iana(language,lac,tag{
         type: "language",
         subtag: "lac",
         description: "Lacandon",
         added: "2009-07-29"}).
iana(language,lad,tag{
         type: "language",
         subtag: "lad",
         description: "Ladino",
         added: "2005-10-16"}).
iana(language,lae,tag{
         type: "language",
         subtag: "lae",
         description: "Pattani",
         added: "2009-07-29"}).
iana(language,laf,tag{
         type: "language",
         subtag: "laf",
         description: "Lafofa",
         added: "2009-07-29"}).
iana(language,lag,tag{
         type: "language",
         subtag: "lag",
         description: "Langi",
         added: "2009-07-29"}).
iana(language,lah,tag{
         type: "language",
         subtag: "lah",
         description: "Lahnda",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,lai,tag{
         type: "language",
         subtag: "lai",
         description: "Lambya",
         added: "2009-07-29"}).
iana(language,laj,tag{
         type: "language",
         subtag: "laj",
         description: "Lango (Uganda)",
         added: "2009-07-29"}).
iana(language,lak,tag{
         type: "language",
         subtag: "lak",
         description: "Laka (Nigeria)",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         preferred_value: "ksp"}).
iana(language,lal,tag{
         type: "language",
         subtag: "lal",
         description: "Lalia",
         added: "2009-07-29"}).
iana(language,lam,tag{
         type: "language",
         subtag: "lam",
         description: "Lamba",
         added: "2005-10-16"}).
iana(language,lan,tag{
         type: "language",
         subtag: "lan",
         description: "Laru",
         added: "2009-07-29"}).
iana(language,lap,tag{
         type: "language",
         subtag: "lap",
         description: "Laka (Chad)",
         added: "2009-07-29"}).
iana(language,laq,tag{
         type: "language",
         subtag: "laq",
         description: "Qabiao",
         added: "2009-07-29"}).
iana(language,lar,tag{
         type: "language",
         subtag: "lar",
         description: "Larteh",
         added: "2009-07-29"}).
iana(language,las,tag{
         type: "language",
         subtag: "las",
         description: "Lama (Togo)",
         added: "2009-07-29"}).
iana(language,lau,tag{
         type: "language",
         subtag: "lau",
         description: "Laba",
         added: "2009-07-29"}).
iana(language,law,tag{
         type: "language",
         subtag: "law",
         description: "Lauje",
         added: "2009-07-29"}).
iana(language,lax,tag{
         type: "language",
         subtag: "lax",
         description: "Tiwa",
         added: "2009-07-29"}).
iana(language,lay,tag{
         type: "language",
         subtag: "lay",
         description: "Lama Bai",
         added: "2009-07-29"}).
iana(language,laz,tag{
         type: "language",
         subtag: "laz",
         description: "Aribwatsa",
         added: "2009-07-29"}).
iana(language,lba,tag{
         type: "language",
         subtag: "lba",
         description: "Lui",
         added: "2009-07-29",
         deprecated: "2019-04-16"}).
iana(language,lbb,tag{
         type: "language",
         subtag: "lbb",
         description: "Label",
         added: "2009-07-29"}).
iana(language,lbc,tag{
         type: "language",
         subtag: "lbc",
         description: "Lakkia",
         added: "2009-07-29"}).
iana(language,lbe,tag{
         type: "language",
         subtag: "lbe",
         description: "Lak",
         added: "2009-07-29"}).
iana(language,lbf,tag{
         type: "language",
         subtag: "lbf",
         description: "Tinani",
         added: "2009-07-29"}).
iana(language,lbg,tag{
         type: "language",
         subtag: "lbg",
         description: "Laopang",
         added: "2009-07-29"}).
iana(language,lbi,tag{
         type: "language",
         subtag: "lbi",
         description: "La'bi",
         added: "2009-07-29"}).
iana(language,lbj,tag{
         type: "language",
         subtag: "lbj",
         description: "Ladakhi",
         added: "2009-07-29"}).
iana(language,lbk,tag{
         type: "language",
         subtag: "lbk",
         description: "Central Bontok",
         added: "2010-03-11",
         macrolanguage: "bnc"}).
iana(language,lbl,tag{
         type: "language",
         subtag: "lbl",
         description: "Libon Bikol",
         added: "2010-03-11",
         macrolanguage: "bik"}).
iana(language,lbm,tag{
         type: "language",
         subtag: "lbm",
         description: "Lodhi",
         added: "2009-07-29"}).
iana(language,lbn,tag{
         type: "language",
         subtag: "lbn",
         description: "Rmeet",
         added: "2009-07-29"}).
iana(language,lbo,tag{
         type: "language",
         subtag: "lbo",
         description: "Laven",
         added: "2009-07-29"}).
iana(language,lbq,tag{
         type: "language",
         subtag: "lbq",
         description: "Wampar",
         added: "2009-07-29"}).
iana(language,lbr,tag{
         type: "language",
         subtag: "lbr",
         description: "Lohorung",
         added: "2009-07-29"}).
iana(language,lbs,tag{
         type: "language",
         subtag: "lbs",
         description: "Libyan Sign Language",
         added: "2009-07-29"}).
iana(language,lbt,tag{
         type: "language",
         subtag: "lbt",
         description: "Lachi",
         added: "2009-07-29"}).
iana(language,lbu,tag{
         type: "language",
         subtag: "lbu",
         description: "Labu",
         added: "2009-07-29"}).
iana(language,lbv,tag{
         type: "language",
         subtag: "lbv",
         description: "Lavatbura-Lamusong",
         added: "2009-07-29"}).
iana(language,lbw,tag{
         type: "language",
         subtag: "lbw",
         description: "Tolaki",
         added: "2009-07-29"}).
iana(language,lbx,tag{
         type: "language",
         subtag: "lbx",
         description: "Lawangan",
         added: "2009-07-29"}).
iana(language,lby,tag{
         type: "language",
         subtag: "lby",
         description: ["Lamalama","Lamu-Lamu"],
         added: "2009-07-29"}).
iana(language,lbz,tag{
         type: "language",
         subtag: "lbz",
         description: "Lardil",
         added: "2009-07-29"}).
iana(language,lcc,tag{
         type: "language",
         subtag: "lcc",
         description: "Legenyem",
         added: "2009-07-29"}).
iana(language,lcd,tag{
         type: "language",
         subtag: "lcd",
         description: "Lola",
         added: "2009-07-29"}).
iana(language,lce,tag{
         type: "language",
         subtag: "lce",
         description: ["Loncong","Sekak"],
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,lcf,tag{
         type: "language",
         subtag: "lcf",
         description: "Lubu",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,lch,tag{
         type: "language",
         subtag: "lch",
         description: "Luchazi",
         added: "2009-07-29"}).
iana(language,lcl,tag{
         type: "language",
         subtag: "lcl",
         description: "Lisela",
         added: "2009-07-29"}).
iana(language,lcm,tag{
         type: "language",
         subtag: "lcm",
         description: "Tungag",
         added: "2009-07-29"}).
iana(language,lcp,tag{
         type: "language",
         subtag: "lcp",
         description: "Western Lawa",
         added: "2009-07-29"}).
iana(language,lcq,tag{
         type: "language",
         subtag: "lcq",
         description: "Luhu",
         added: "2009-07-29"}).
iana(language,lcs,tag{
         type: "language",
         subtag: "lcs",
         description: "Lisabata-Nuniali",
         added: "2009-07-29"}).
iana(language,lda,tag{
         type: "language",
         subtag: "lda",
         description: "Kla-Dan",
         added: "2013-09-10"}).
iana(language,ldb,tag{
         type: "language",
         subtag: "ldb",
         description: "Dũya",
         added: "2009-07-29"}).
iana(language,ldd,tag{
         type: "language",
         subtag: "ldd",
         description: "Luri",
         added: "2009-07-29"}).
iana(language,ldg,tag{
         type: "language",
         subtag: "ldg",
         description: "Lenyima",
         added: "2009-07-29"}).
iana(language,ldh,tag{
         type: "language",
         subtag: "ldh",
         description: "Lamja-Dengsa-Tola",
         added: "2009-07-29"}).
iana(language,ldi,tag{
         type: "language",
         subtag: "ldi",
         description: "Laari",
         added: "2009-07-29",
         macrolanguage: "kg"}).
iana(language,ldj,tag{
         type: "language",
         subtag: "ldj",
         description: "Lemoro",
         added: "2009-07-29"}).
iana(language,ldk,tag{
         type: "language",
         subtag: "ldk",
         description: "Leelau",
         added: "2009-07-29"}).
iana(language,ldl,tag{
         type: "language",
         subtag: "ldl",
         description: "Kaan",
         added: "2009-07-29"}).
iana(language,ldm,tag{
         type: "language",
         subtag: "ldm",
         description: "Landoma",
         added: "2009-07-29"}).
iana(language,ldn,tag{
         type: "language",
         subtag: "ldn",
         description: "Láadan",
         added: "2009-07-29"}).
iana(language,ldo,tag{
         type: "language",
         subtag: "ldo",
         description: "Loo",
         added: "2009-07-29"}).
iana(language,ldp,tag{
         type: "language",
         subtag: "ldp",
         description: "Tso",
         added: "2009-07-29"}).
iana(language,ldq,tag{
         type: "language",
         subtag: "ldq",
         description: "Lufu",
         added: "2009-07-29"}).
iana(language,lea,tag{
         type: "language",
         subtag: "lea",
         description: "Lega-Shabunda",
         added: "2009-07-29"}).
iana(language,leb,tag{
         type: "language",
         subtag: "leb",
         description: "Lala-Bisa",
         added: "2009-07-29"}).
iana(language,lec,tag{
         type: "language",
         subtag: "lec",
         description: "Leco",
         added: "2009-07-29"}).
iana(language,led,tag{
         type: "language",
         subtag: "led",
         description: "Lendu",
         added: "2009-07-29"}).
iana(language,lee,tag{
         type: "language",
         subtag: "lee",
         description: "Lyélé",
         added: "2009-07-29"}).
iana(language,lef,tag{
         type: "language",
         subtag: "lef",
         description: "Lelemi",
         added: "2009-07-29"}).
iana(language,leg,tag{
         type: "language",
         subtag: "leg",
         description: "Lengua",
         added: "2009-07-29",
         deprecated: "2014-02-28",
         comments: "see enl, enx"}).
iana(language,leh,tag{
         type: "language",
         subtag: "leh",
         description: "Lenje",
         added: "2009-07-29"}).
iana(language,lei,tag{
         type: "language",
         subtag: "lei",
         description: "Lemio",
         added: "2009-07-29"}).
iana(language,lej,tag{
         type: "language",
         subtag: "lej",
         description: "Lengola",
         added: "2009-07-29"}).
iana(language,lek,tag{
         type: "language",
         subtag: "lek",
         description: "Leipon",
         added: "2009-07-29"}).
iana(language,lel,tag{
         type: "language",
         subtag: "lel",
         description: "Lele (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,lem,tag{
         type: "language",
         subtag: "lem",
         description: "Nomaande",
         added: "2009-07-29"}).
iana(language,len,tag{
         type: "language",
         subtag: "len",
         description: "Lenca",
         added: "2009-07-29"}).
iana(language,leo,tag{
         type: "language",
         subtag: "leo",
         description: "Leti (Cameroon)",
         added: "2009-07-29"}).
iana(language,lep,tag{
         type: "language",
         subtag: "lep",
         description: "Lepcha",
         added: "2009-07-29"}).
iana(language,leq,tag{
         type: "language",
         subtag: "leq",
         description: "Lembena",
         added: "2009-07-29"}).
iana(language,ler,tag{
         type: "language",
         subtag: "ler",
         description: "Lenkau",
         added: "2009-07-29"}).
iana(language,les,tag{
         type: "language",
         subtag: "les",
         description: "Lese",
         added: "2009-07-29"}).
iana(language,let,tag{
         type: "language",
         subtag: "let",
         description: ["Lesing-Gelimi","Amio-Gelimi"],
         added: "2009-07-29"}).
iana(language,leu,tag{
         type: "language",
         subtag: "leu",
         description: "Kara (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,lev,tag{
         type: "language",
         subtag: "lev",
         description: "Lamma",
         added: "2009-07-29"}).
iana(language,lew,tag{
         type: "language",
         subtag: "lew",
         description: "Ledo Kaili",
         added: "2009-07-29"}).
iana(language,lex,tag{
         type: "language",
         subtag: "lex",
         description: "Luang",
         added: "2009-07-29"}).
iana(language,ley,tag{
         type: "language",
         subtag: "ley",
         description: "Lemolang",
         added: "2009-07-29"}).
iana(language,lez,tag{
         type: "language",
         subtag: "lez",
         description: "Lezghian",
         added: "2005-10-16"}).
iana(language,lfa,tag{
         type: "language",
         subtag: "lfa",
         description: "Lefa",
         added: "2009-07-29"}).
iana(language,lfn,tag{
         type: "language",
         subtag: "lfn",
         description: "Lingua Franca Nova",
         added: "2009-07-29"}).
iana(language,lga,tag{
         type: "language",
         subtag: "lga",
         description: "Lungga",
         added: "2009-07-29"}).
iana(language,lgb,tag{
         type: "language",
         subtag: "lgb",
         description: "Laghu",
         added: "2009-07-29"}).
iana(language,lgg,tag{
         type: "language",
         subtag: "lgg",
         description: "Lugbara",
         added: "2009-07-29"}).
iana(language,lgh,tag{
         type: "language",
         subtag: "lgh",
         description: "Laghuu",
         added: "2009-07-29"}).
iana(language,lgi,tag{
         type: "language",
         subtag: "lgi",
         description: "Lengilu",
         added: "2009-07-29"}).
iana(language,lgk,tag{
         type: "language",
         subtag: "lgk",
         description: ["Lingarak","Neverver"],
         added: "2009-07-29"}).
iana(language,lgl,tag{
         type: "language",
         subtag: "lgl",
         description: "Wala",
         added: "2009-07-29"}).
iana(language,lgm,tag{
         type: "language",
         subtag: "lgm",
         description: "Lega-Mwenga",
         added: "2009-07-29"}).
iana(language,lgn,tag{
         type: "language",
         subtag: "lgn",
         description: ["T'apo","Opuuo"],
         added: "2009-07-29"}).
iana(language,lgo,tag{
         type: "language",
         subtag: "lgo",
         description: "Lango (South Sudan)",
         added: "2022-02-25"}).
iana(language,lgq,tag{
         type: "language",
         subtag: "lgq",
         description: "Logba",
         added: "2009-07-29"}).
iana(language,lgr,tag{
         type: "language",
         subtag: "lgr",
         description: "Lengo",
         added: "2009-07-29"}).
iana(language,lgt,tag{
         type: "language",
         subtag: "lgt",
         description: "Pahi",
         added: "2009-07-29"}).
iana(language,lgu,tag{
         type: "language",
         subtag: "lgu",
         description: "Longgu",
         added: "2009-07-29"}).
iana(language,lgz,tag{
         type: "language",
         subtag: "lgz",
         description: "Ligenza",
         added: "2009-07-29"}).
iana(language,lha,tag{
         type: "language",
         subtag: "lha",
         description: "Laha (Viet Nam)",
         added: "2009-07-29"}).
iana(language,lhh,tag{
         type: "language",
         subtag: "lhh",
         description: "Laha (Indonesia)",
         added: "2009-07-29"}).
iana(language,lhi,tag{
         type: "language",
         subtag: "lhi",
         description: "Lahu Shi",
         added: "2009-07-29"}).
iana(language,lhl,tag{
         type: "language",
         subtag: "lhl",
         description: "Lahul Lohar",
         added: "2009-07-29"}).
iana(language,lhm,tag{
         type: "language",
         subtag: "lhm",
         description: "Lhomi",
         added: "2009-07-29"}).
iana(language,lhn,tag{
         type: "language",
         subtag: "lhn",
         description: "Lahanan",
         added: "2009-07-29"}).
iana(language,lhp,tag{
         type: "language",
         subtag: "lhp",
         description: "Lhokpu",
         added: "2009-07-29"}).
iana(language,lhs,tag{
         type: "language",
         subtag: "lhs",
         description: "Mlahsö",
         added: "2009-07-29"}).
iana(language,lht,tag{
         type: "language",
         subtag: "lht",
         description: "Lo-Toga",
         added: "2009-07-29"}).
iana(language,lhu,tag{
         type: "language",
         subtag: "lhu",
         description: "Lahu",
         added: "2009-07-29"}).
iana(language,lia,tag{
         type: "language",
         subtag: "lia",
         description: "West-Central Limba",
         added: "2009-07-29"}).
iana(language,lib,tag{
         type: "language",
         subtag: "lib",
         description: "Likum",
         added: "2009-07-29"}).
iana(language,lic,tag{
         type: "language",
         subtag: "lic",
         description: "Hlai",
         added: "2009-07-29"}).
iana(language,lid,tag{
         type: "language",
         subtag: "lid",
         description: "Nyindrou",
         added: "2009-07-29"}).
iana(language,lie,tag{
         type: "language",
         subtag: "lie",
         description: "Likila",
         added: "2009-07-29"}).
iana(language,lif,tag{
         type: "language",
         subtag: "lif",
         description: "Limbu",
         added: "2009-07-29"}).
iana(language,lig,tag{
         type: "language",
         subtag: "lig",
         description: "Ligbi",
         added: "2009-07-29"}).
iana(language,lih,tag{
         type: "language",
         subtag: "lih",
         description: "Lihir",
         added: "2009-07-29"}).
iana(language,lii,tag{
         type: "language",
         subtag: "lii",
         description: "Lingkhim",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "raq"}).
iana(language,lij,tag{
         type: "language",
         subtag: "lij",
         description: "Ligurian",
         added: "2009-07-29"}).
iana(language,lik,tag{
         type: "language",
         subtag: "lik",
         description: "Lika",
         added: "2009-07-29"}).
iana(language,lil,tag{
         type: "language",
         subtag: "lil",
         description: "Lillooet",
         added: "2009-07-29"}).
iana(language,lio,tag{
         type: "language",
         subtag: "lio",
         description: "Liki",
         added: "2009-07-29"}).
iana(language,lip,tag{
         type: "language",
         subtag: "lip",
         description: "Sekpele",
         added: "2009-07-29"}).
iana(language,liq,tag{
         type: "language",
         subtag: "liq",
         description: "Libido",
         added: "2009-07-29"}).
iana(language,lir,tag{
         type: "language",
         subtag: "lir",
         description: "Liberian English",
         added: "2009-07-29"}).
iana(language,lis,tag{
         type: "language",
         subtag: "lis",
         description: "Lisu",
         added: "2009-07-29"}).
iana(language,liu,tag{
         type: "language",
         subtag: "liu",
         description: "Logorik",
         added: "2009-07-29"}).
iana(language,liv,tag{
         type: "language",
         subtag: "liv",
         description: "Liv",
         added: "2009-07-29"}).
iana(language,liw,tag{
         type: "language",
         subtag: "liw",
         description: "Col",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,lix,tag{
         type: "language",
         subtag: "lix",
         description: "Liabuku",
         added: "2009-07-29"}).
iana(language,liy,tag{
         type: "language",
         subtag: "liy",
         description: "Banda-Bambari",
         added: "2009-07-29"}).
iana(language,liz,tag{
         type: "language",
         subtag: "liz",
         description: "Libinza",
         added: "2009-07-29"}).
iana(language,lja,tag{
         type: "language",
         subtag: "lja",
         description: "Golpa",
         added: "2013-09-10"}).
iana(language,lje,tag{
         type: "language",
         subtag: "lje",
         description: "Rampi",
         added: "2009-07-29"}).
iana(language,lji,tag{
         type: "language",
         subtag: "lji",
         description: "Laiyolo",
         added: "2009-07-29"}).
iana(language,ljl,tag{
         type: "language",
         subtag: "ljl",
         description: "Li'o",
         added: "2009-07-29"}).
iana(language,ljp,tag{
         type: "language",
         subtag: "ljp",
         description: "Lampung Api",
         added: "2009-07-29"}).
iana(language,ljw,tag{
         type: "language",
         subtag: "ljw",
         description: "Yirandali",
         added: "2013-09-10"}).
iana(language,ljx,tag{
         type: "language",
         subtag: "ljx",
         description: "Yuru",
         added: "2013-09-10"}).
iana(language,lka,tag{
         type: "language",
         subtag: "lka",
         description: "Lakalei",
         added: "2009-07-29"}).
iana(language,lkb,tag{
         type: "language",
         subtag: "lkb",
         description: ["Kabras","Lukabaras"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lkc,tag{
         type: "language",
         subtag: "lkc",
         description: "Kucong",
         added: "2009-07-29"}).
iana(language,lkd,tag{
         type: "language",
         subtag: "lkd",
         description: "Lakondê",
         added: "2009-07-29"}).
iana(language,lke,tag{
         type: "language",
         subtag: "lke",
         description: "Kenyi",
         added: "2009-07-29"}).
iana(language,lkh,tag{
         type: "language",
         subtag: "lkh",
         description: "Lakha",
         added: "2009-07-29"}).
iana(language,lki,tag{
         type: "language",
         subtag: "lki",
         description: "Laki",
         added: "2009-07-29"}).
iana(language,lkj,tag{
         type: "language",
         subtag: "lkj",
         description: "Remun",
         added: "2009-07-29"}).
iana(language,lkl,tag{
         type: "language",
         subtag: "lkl",
         description: "Laeko-Libuat",
         added: "2009-07-29"}).
iana(language,lkm,tag{
         type: "language",
         subtag: "lkm",
         description: "Kalaamaya",
         added: "2013-09-10"}).
iana(language,lkn,tag{
         type: "language",
         subtag: "lkn",
         description: ["Lakon","Vure"],
         added: "2009-07-29"}).
iana(language,lko,tag{
         type: "language",
         subtag: "lko",
         description: ["Khayo","Olukhayo"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lkr,tag{
         type: "language",
         subtag: "lkr",
         description: "Päri",
         added: "2009-07-29"}).
iana(language,lks,tag{
         type: "language",
         subtag: "lks",
         description: ["Kisa","Olushisa"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lkt,tag{
         type: "language",
         subtag: "lkt",
         description: "Lakota",
         added: "2009-07-29"}).
iana(language,lku,tag{
         type: "language",
         subtag: "lku",
         description: "Kungkari",
         added: "2013-09-10"}).
iana(language,lky,tag{
         type: "language",
         subtag: "lky",
         description: "Lokoya",
         added: "2009-07-29"}).
iana(language,lla,tag{
         type: "language",
         subtag: "lla",
         description: "Lala-Roba",
         added: "2009-07-29"}).
iana(language,llb,tag{
         type: "language",
         subtag: "llb",
         description: "Lolo",
         added: "2009-07-29"}).
iana(language,llc,tag{
         type: "language",
         subtag: "llc",
         description: "Lele (Guinea)",
         added: "2009-07-29"}).
iana(language,lld,tag{
         type: "language",
         subtag: "lld",
         description: "Ladin",
         added: "2009-07-29"}).
iana(language,lle,tag{
         type: "language",
         subtag: "lle",
         description: "Lele (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,llf,tag{
         type: "language",
         subtag: "llf",
         description: "Hermit",
         added: "2009-07-29"}).
iana(language,llg,tag{
         type: "language",
         subtag: "llg",
         description: "Lole",
         added: "2009-07-29"}).
iana(language,llh,tag{
         type: "language",
         subtag: "llh",
         description: "Lamu",
         added: "2009-07-29"}).
iana(language,lli,tag{
         type: "language",
         subtag: "lli",
         description: "Teke-Laali",
         added: "2009-07-29"}).
iana(language,llj,tag{
         type: "language",
         subtag: "llj",
         description: "Ladji Ladji",
         added: "2012-08-12"}).
iana(language,llk,tag{
         type: "language",
         subtag: "llk",
         description: "Lelak",
         added: "2009-07-29"}).
iana(language,lll,tag{
         type: "language",
         subtag: "lll",
         description: "Lilau",
         added: "2009-07-29"}).
iana(language,llm,tag{
         type: "language",
         subtag: "llm",
         description: "Lasalimu",
         added: "2009-07-29"}).
iana(language,lln,tag{
         type: "language",
         subtag: "lln",
         description: "Lele (Chad)",
         added: "2009-07-29"}).
iana(language,llo,tag{
         type: "language",
         subtag: "llo",
         description: "Khlor",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         preferred_value: "ngt"}).
iana(language,llp,tag{
         type: "language",
         subtag: "llp",
         description: "North Efate",
         added: "2009-07-29"}).
iana(language,llq,tag{
         type: "language",
         subtag: "llq",
         description: "Lolak",
         added: "2009-07-29"}).
iana(language,lls,tag{
         type: "language",
         subtag: "lls",
         description: "Lithuanian Sign Language",
         added: "2009-07-29"}).
iana(language,llu,tag{
         type: "language",
         subtag: "llu",
         description: "Lau",
         added: "2009-07-29"}).
iana(language,llx,tag{
         type: "language",
         subtag: "llx",
         description: "Lauan",
         added: "2009-07-29"}).
iana(language,lma,tag{
         type: "language",
         subtag: "lma",
         description: "East Limba",
         added: "2009-07-29"}).
iana(language,lmb,tag{
         type: "language",
         subtag: "lmb",
         description: "Merei",
         added: "2009-07-29"}).
iana(language,lmc,tag{
         type: "language",
         subtag: "lmc",
         description: "Limilngan",
         added: "2009-07-29"}).
iana(language,lmd,tag{
         type: "language",
         subtag: "lmd",
         description: "Lumun",
         added: "2009-07-29"}).
iana(language,lme,tag{
         type: "language",
         subtag: "lme",
         description: "Pévé",
         added: "2009-07-29"}).
iana(language,lmf,tag{
         type: "language",
         subtag: "lmf",
         description: "South Lembata",
         added: "2009-07-29"}).
iana(language,lmg,tag{
         type: "language",
         subtag: "lmg",
         description: "Lamogai",
         added: "2009-07-29"}).
iana(language,lmh,tag{
         type: "language",
         subtag: "lmh",
         description: "Lambichhong",
         added: "2009-07-29"}).
iana(language,lmi,tag{
         type: "language",
         subtag: "lmi",
         description: "Lombi",
         added: "2009-07-29"}).
iana(language,lmj,tag{
         type: "language",
         subtag: "lmj",
         description: "West Lembata",
         added: "2009-07-29"}).
iana(language,lmk,tag{
         type: "language",
         subtag: "lmk",
         description: "Lamkang",
         added: "2009-07-29"}).
iana(language,lml,tag{
         type: "language",
         subtag: "lml",
         description: "Hano",
         added: "2009-07-29"}).
iana(language,lmm,tag{
         type: "language",
         subtag: "lmm",
         description: "Lamam",
         added: "2009-07-29",
         deprecated: "2014-02-28",
         preferred_value: "rmx"}).
iana(language,lmn,tag{
         type: "language",
         subtag: "lmn",
         description: "Lambadi",
         added: "2009-07-29"}).
iana(language,lmo,tag{
         type: "language",
         subtag: "lmo",
         description: "Lombard",
         added: "2009-07-29"}).
iana(language,lmp,tag{
         type: "language",
         subtag: "lmp",
         description: "Limbum",
         added: "2009-07-29"}).
iana(language,lmq,tag{
         type: "language",
         subtag: "lmq",
         description: "Lamatuka",
         added: "2009-07-29"}).
iana(language,lmr,tag{
         type: "language",
         subtag: "lmr",
         description: "Lamalera",
         added: "2009-07-29"}).
iana(language,lmu,tag{
         type: "language",
         subtag: "lmu",
         description: "Lamenu",
         added: "2009-07-29"}).
iana(language,lmv,tag{
         type: "language",
         subtag: "lmv",
         description: "Lomaiviti",
         added: "2009-07-29"}).
iana(language,lmw,tag{
         type: "language",
         subtag: "lmw",
         description: "Lake Miwok",
         added: "2009-07-29"}).
iana(language,lmx,tag{
         type: "language",
         subtag: "lmx",
         description: "Laimbue",
         added: "2009-07-29"}).
iana(language,lmy,tag{
         type: "language",
         subtag: "lmy",
         description: "Lamboya",
         added: "2009-07-29"}).
iana(language,lmz,tag{
         type: "language",
         subtag: "lmz",
         description: "Lumbee",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,lna,tag{
         type: "language",
         subtag: "lna",
         description: "Langbashe",
         added: "2009-07-29"}).
iana(language,lnb,tag{
         type: "language",
         subtag: "lnb",
         description: "Mbalanhu",
         added: "2009-07-29"}).
iana(language,lnd,tag{
         type: "language",
         subtag: "lnd",
         description: ["Lundayeh","Lun Bawang"],
         added: "2009-07-29"}).
iana(language,lng,tag{
         type: "language",
         subtag: "lng",
         description: "Langobardic",
         added: "2009-07-29"}).
iana(language,lnh,tag{
         type: "language",
         subtag: "lnh",
         description: "Lanoh",
         added: "2009-07-29"}).
iana(language,lni,tag{
         type: "language",
         subtag: "lni",
         description: "Daantanai'",
         added: "2009-07-29"}).
iana(language,lnj,tag{
         type: "language",
         subtag: "lnj",
         description: "Leningitij",
         added: "2009-07-29"}).
iana(language,lnl,tag{
         type: "language",
         subtag: "lnl",
         description: "South Central Banda",
         added: "2009-07-29"}).
iana(language,lnm,tag{
         type: "language",
         subtag: "lnm",
         description: "Langam",
         added: "2009-07-29"}).
iana(language,lnn,tag{
         type: "language",
         subtag: "lnn",
         description: "Lorediakarkar",
         added: "2009-07-29"}).
iana(language,lno,tag{
         type: "language",
         subtag: "lno",
         description: "Lango (South Sudan)",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         comments: "see imt, lgo, lqr, oie"}).
iana(language,lns,tag{
         type: "language",
         subtag: "lns",
         description: "Lamnso'",
         added: "2009-07-29"}).
iana(language,lnu,tag{
         type: "language",
         subtag: "lnu",
         description: "Longuda",
         added: "2009-07-29"}).
iana(language,lnw,tag{
         type: "language",
         subtag: "lnw",
         description: "Lanima",
         added: "2013-09-10"}).
iana(language,lnz,tag{
         type: "language",
         subtag: "lnz",
         description: "Lonzo",
         added: "2009-07-29"}).
iana(language,loa,tag{
         type: "language",
         subtag: "loa",
         description: "Loloda",
         added: "2009-07-29"}).
iana(language,lob,tag{
         type: "language",
         subtag: "lob",
         description: "Lobi",
         added: "2009-07-29"}).
iana(language,loc,tag{
         type: "language",
         subtag: "loc",
         description: "Inonhan",
         added: "2009-07-29"}).
iana(language,loe,tag{
         type: "language",
         subtag: "loe",
         description: "Saluan",
         added: "2009-07-29"}).
iana(language,lof,tag{
         type: "language",
         subtag: "lof",
         description: "Logol",
         added: "2009-07-29"}).
iana(language,log,tag{
         type: "language",
         subtag: "log",
         description: "Logo",
         added: "2009-07-29"}).
iana(language,loh,tag{
         type: "language",
         subtag: "loh",
         description: "Narim",
         added: "2009-07-29"}).
iana(language,loi,tag{
         type: "language",
         subtag: "loi",
         description: "Loma (Côte d'Ivoire)",
         added: "2009-07-29"}).
iana(language,loj,tag{
         type: "language",
         subtag: "loj",
         description: "Lou",
         added: "2009-07-29"}).
iana(language,lok,tag{
         type: "language",
         subtag: "lok",
         description: "Loko",
         added: "2009-07-29"}).
iana(language,lol,tag{
         type: "language",
         subtag: "lol",
         description: "Mongo",
         added: "2005-10-16"}).
iana(language,lom,tag{
         type: "language",
         subtag: "lom",
         description: "Loma (Liberia)",
         added: "2009-07-29"}).
iana(language,lon,tag{
         type: "language",
         subtag: "lon",
         description: "Malawi Lomwe",
         added: "2009-07-29"}).
iana(language,loo,tag{
         type: "language",
         subtag: "loo",
         description: "Lombo",
         added: "2009-07-29"}).
iana(language,lop,tag{
         type: "language",
         subtag: "lop",
         description: "Lopa",
         added: "2009-07-29"}).
iana(language,loq,tag{
         type: "language",
         subtag: "loq",
         description: "Lobala",
         added: "2009-07-29"}).
iana(language,lor,tag{
         type: "language",
         subtag: "lor",
         description: "Téén",
         added: "2009-07-29"}).
iana(language,los,tag{
         type: "language",
         subtag: "los",
         description: "Loniu",
         added: "2009-07-29"}).
iana(language,lot,tag{
         type: "language",
         subtag: "lot",
         description: "Otuho",
         added: "2009-07-29"}).
iana(language,lou,tag{
         type: "language",
         subtag: "lou",
         description: "Louisiana Creole",
         added: "2009-07-29"}).
iana(language,lov,tag{
         type: "language",
         subtag: "lov",
         description: "Lopi",
         added: "2009-07-29"}).
iana(language,low,tag{
         type: "language",
         subtag: "low",
         description: "Tampias Lobu",
         added: "2009-07-29"}).
iana(language,lox,tag{
         type: "language",
         subtag: "lox",
         description: "Loun",
         added: "2009-07-29"}).
iana(language,loy,tag{
         type: "language",
         subtag: "loy",
         description: "Loke",
         added: "2009-07-29"}).
iana(language,loz,tag{
         type: "language",
         subtag: "loz",
         description: "Lozi",
         added: "2005-10-16"}).
iana(language,lpa,tag{
         type: "language",
         subtag: "lpa",
         description: "Lelepa",
         added: "2009-07-29"}).
iana(language,lpe,tag{
         type: "language",
         subtag: "lpe",
         description: "Lepki",
         added: "2009-07-29"}).
iana(language,lpn,tag{
         type: "language",
         subtag: "lpn",
         description: "Long Phuri Naga",
         added: "2009-07-29"}).
iana(language,lpo,tag{
         type: "language",
         subtag: "lpo",
         description: "Lipo",
         added: "2009-07-29"}).
iana(language,lpx,tag{
         type: "language",
         subtag: "lpx",
         description: "Lopit",
         added: "2009-07-29"}).
iana(language,lqr,tag{
         type: "language",
         subtag: "lqr",
         description: "Logir",
         added: "2022-02-25"}).
iana(language,lra,tag{
         type: "language",
         subtag: "lra",
         description: "Rara Bakati'",
         added: "2009-07-29"}).
iana(language,lrc,tag{
         type: "language",
         subtag: "lrc",
         description: "Northern Luri",
         added: "2009-07-29"}).
iana(language,lre,tag{
         type: "language",
         subtag: "lre",
         description: "Laurentian",
         added: "2009-07-29"}).
iana(language,lrg,tag{
         type: "language",
         subtag: "lrg",
         description: "Laragia",
         added: "2009-07-29"}).
iana(language,lri,tag{
         type: "language",
         subtag: "lri",
         description: ["Marachi","Olumarachi"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lrk,tag{
         type: "language",
         subtag: "lrk",
         description: "Loarki",
         added: "2009-07-29"}).
iana(language,lrl,tag{
         type: "language",
         subtag: "lrl",
         description: "Lari",
         added: "2009-07-29"}).
iana(language,lrm,tag{
         type: "language",
         subtag: "lrm",
         description: ["Marama","Olumarama"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lrn,tag{
         type: "language",
         subtag: "lrn",
         description: "Lorang",
         added: "2009-07-29"}).
iana(language,lro,tag{
         type: "language",
         subtag: "lro",
         description: "Laro",
         added: "2009-07-29"}).
iana(language,lrr,tag{
         type: "language",
         subtag: "lrr",
         description: "Southern Yamphu",
         added: "2009-07-29"}).
iana(language,lrt,tag{
         type: "language",
         subtag: "lrt",
         description: "Larantuka Malay",
         added: "2009-07-29"}).
iana(language,lrv,tag{
         type: "language",
         subtag: "lrv",
         description: "Larevat",
         added: "2009-07-29"}).
iana(language,lrz,tag{
         type: "language",
         subtag: "lrz",
         description: "Lemerig",
         added: "2009-07-29"}).
iana(language,lsa,tag{
         type: "language",
         subtag: "lsa",
         description: "Lasgerdi",
         added: "2009-07-29"}).
iana(language,lsb,tag{
         type: "language",
         subtag: "lsb",
         description: ["Burundian Sign Language","Langue des Signes Burundaise"],
         added: "2021-02-20"}).
iana(language,lsc,tag{
         type: "language",
         subtag: "lsc",
         description: ["Albarradas Sign Language","Lengua de señas Albarradas"],
         added: "2022-02-25"}).
iana(language,lsd,tag{
         type: "language",
         subtag: "lsd",
         description: "Lishana Deni",
         added: "2009-07-29"}).
iana(language,lse,tag{
         type: "language",
         subtag: "lse",
         description: "Lusengo",
         added: "2009-07-29"}).
iana(language,lsg,tag{
         type: "language",
         subtag: "lsg",
         description: "Lyons Sign Language",
         added: "2009-07-29",
         deprecated: "2018-03-08"}).
iana(language,lsh,tag{
         type: "language",
         subtag: "lsh",
         description: "Lish",
         added: "2009-07-29"}).
iana(language,lsi,tag{
         type: "language",
         subtag: "lsi",
         description: "Lashi",
         added: "2009-07-29"}).
iana(language,lsl,tag{
         type: "language",
         subtag: "lsl",
         description: "Latvian Sign Language",
         added: "2009-07-29"}).
iana(language,lsm,tag{
         type: "language",
         subtag: "lsm",
         description: ["Saamia","Olusamia"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lsn,tag{
         type: "language",
         subtag: "lsn",
         description: "Tibetan Sign Language",
         added: "2019-04-16"}).
iana(language,lso,tag{
         type: "language",
         subtag: "lso",
         description: "Laos Sign Language",
         added: "2009-07-29"}).
iana(language,lsp,tag{
         type: "language",
         subtag: "lsp",
         description: ["Panamanian Sign Language","Lengua de Señas Panameñas"],
         added: "2009-07-29"}).
iana(language,lsr,tag{
         type: "language",
         subtag: "lsr",
         description: "Aruop",
         added: "2009-07-29"}).
iana(language,lss,tag{
         type: "language",
         subtag: "lss",
         description: "Lasi",
         added: "2009-07-29"}).
iana(language,lst,tag{
         type: "language",
         subtag: "lst",
         description: "Trinidad and Tobago Sign Language",
         added: "2009-07-29"}).
iana(language,lsv,tag{
         type: "language",
         subtag: "lsv",
         description: "Sivia Sign Language",
         added: "2019-04-16"}).
iana(language,lsw,tag{
         type: "language",
         subtag: "lsw",
         description: ["Seychelles Sign Language","Lalang Siny Seselwa","Langue des Signes Seychelloise"],
         added: "2022-02-25"}).
iana(language,lsy,tag{
         type: "language",
         subtag: "lsy",
         description: "Mauritian Sign Language",
         added: "2010-03-11"}).
iana(language,ltc,tag{
         type: "language",
         subtag: "ltc",
         description: "Late Middle Chinese",
         added: "2009-07-29"}).
iana(language,ltg,tag{
         type: "language",
         subtag: "ltg",
         description: "Latgalian",
         added: "2010-03-11",
         macrolanguage: "lv"}).
iana(language,lth,tag{
         type: "language",
         subtag: "lth",
         description: "Thur",
         added: "2017-02-23"}).
iana(language,lti,tag{
         type: "language",
         subtag: "lti",
         description: "Leti (Indonesia)",
         added: "2009-07-29"}).
iana(language,ltn,tag{
         type: "language",
         subtag: "ltn",
         description: "Latundê",
         added: "2009-07-29"}).
iana(language,lto,tag{
         type: "language",
         subtag: "lto",
         description: ["Tsotso","Olutsotso"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lts,tag{
         type: "language",
         subtag: "lts",
         description: ["Tachoni","Lutachoni"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,ltu,tag{
         type: "language",
         subtag: "ltu",
         description: "Latu",
         added: "2009-07-29"}).
iana(language,lua,tag{
         type: "language",
         subtag: "lua",
         description: "Luba-Lulua",
         added: "2005-10-16"}).
iana(language,luc,tag{
         type: "language",
         subtag: "luc",
         description: "Aringa",
         added: "2009-07-29"}).
iana(language,lud,tag{
         type: "language",
         subtag: "lud",
         description: "Ludian",
         added: "2009-07-29"}).
iana(language,lue,tag{
         type: "language",
         subtag: "lue",
         description: "Luvale",
         added: "2009-07-29"}).
iana(language,luf,tag{
         type: "language",
         subtag: "luf",
         description: "Laua",
         added: "2009-07-29"}).
iana(language,lui,tag{
         type: "language",
         subtag: "lui",
         description: "Luiseno",
         added: "2005-10-16"}).
iana(language,luj,tag{
         type: "language",
         subtag: "luj",
         description: "Luna",
         added: "2009-07-29"}).
iana(language,luk,tag{
         type: "language",
         subtag: "luk",
         description: "Lunanakha",
         added: "2009-07-29"}).
iana(language,lul,tag{
         type: "language",
         subtag: "lul",
         description: "Olu'bo",
         added: "2009-07-29"}).
iana(language,lum,tag{
         type: "language",
         subtag: "lum",
         description: "Luimbi",
         added: "2009-07-29"}).
iana(language,lun,tag{
         type: "language",
         subtag: "lun",
         description: "Lunda",
         added: "2005-10-16"}).
iana(language,luo,tag{
         type: "language",
         subtag: "luo",
         description: ["Luo (Kenya and Tanzania)","Dholuo"],
         added: "2005-10-16"}).
iana(language,lup,tag{
         type: "language",
         subtag: "lup",
         description: "Lumbu",
         added: "2009-07-29"}).
iana(language,luq,tag{
         type: "language",
         subtag: "luq",
         description: "Lucumi",
         added: "2009-07-29"}).
iana(language,lur,tag{
         type: "language",
         subtag: "lur",
         description: "Laura",
         added: "2009-07-29"}).
iana(language,lus,tag{
         type: "language",
         subtag: "lus",
         description: "Lushai",
         added: "2005-10-16"}).
iana(language,lut,tag{
         type: "language",
         subtag: "lut",
         description: "Lushootseed",
         added: "2009-07-29"}).
iana(language,luu,tag{
         type: "language",
         subtag: "luu",
         description: "Lumba-Yakkha",
         added: "2009-07-29"}).
iana(language,luv,tag{
         type: "language",
         subtag: "luv",
         description: "Luwati",
         added: "2009-07-29"}).
iana(language,luw,tag{
         type: "language",
         subtag: "luw",
         description: "Luo (Cameroon)",
         added: "2009-07-29"}).
iana(language,luy,tag{
         type: "language",
         subtag: "luy",
         description: ["Luyia","Oluluyia"],
         added: "2009-07-29",
         scope: "macrolanguage"}).
iana(language,luz,tag{
         type: "language",
         subtag: "luz",
         description: "Southern Luri",
         added: "2009-07-29"}).
iana(language,lva,tag{
         type: "language",
         subtag: "lva",
         description: "Maku'a",
         added: "2009-07-29"}).
iana(language,lvi,tag{
         type: "language",
         subtag: "lvi",
         description: "Lavi",
         added: "2019-04-16"}).
iana(language,lvk,tag{
         type: "language",
         subtag: "lvk",
         description: "Lavukaleve",
         added: "2009-07-29"}).
iana(language,lvs,tag{
         type: "language",
         subtag: "lvs",
         description: "Standard Latvian",
         added: "2010-03-11",
         macrolanguage: "lv"}).
iana(language,lvu,tag{
         type: "language",
         subtag: "lvu",
         description: "Levuka",
         added: "2009-07-29"}).
iana(language,lwa,tag{
         type: "language",
         subtag: "lwa",
         description: "Lwalu",
         added: "2009-07-29"}).
iana(language,lwe,tag{
         type: "language",
         subtag: "lwe",
         description: "Lewo Eleng",
         added: "2009-07-29"}).
iana(language,lwg,tag{
         type: "language",
         subtag: "lwg",
         description: ["Wanga","Oluwanga"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,lwh,tag{
         type: "language",
         subtag: "lwh",
         description: "White Lachi",
         added: "2009-07-29"}).
iana(language,lwl,tag{
         type: "language",
         subtag: "lwl",
         description: "Eastern Lawa",
         added: "2009-07-29"}).
iana(language,lwm,tag{
         type: "language",
         subtag: "lwm",
         description: "Laomian",
         added: "2009-07-29"}).
iana(language,lwo,tag{
         type: "language",
         subtag: "lwo",
         description: "Luwo",
         added: "2009-07-29"}).
iana(language,lws,tag{
         type: "language",
         subtag: "lws",
         description: "Malawian Sign Language",
         added: "2018-03-08"}).
iana(language,lwt,tag{
         type: "language",
         subtag: "lwt",
         description: "Lewotobi",
         added: "2009-07-29"}).
iana(language,lwu,tag{
         type: "language",
         subtag: "lwu",
         description: "Lawu",
         added: "2013-09-10"}).
iana(language,lww,tag{
         type: "language",
         subtag: "lww",
         description: "Lewo",
         added: "2009-07-29"}).
iana(language,lxm,tag{
         type: "language",
         subtag: "lxm",
         description: "Lakurumau",
         added: "2021-02-20"}).
iana(language,lya,tag{
         type: "language",
         subtag: "lya",
         description: "Layakha",
         added: "2009-07-29"}).
iana(language,lyg,tag{
         type: "language",
         subtag: "lyg",
         description: "Lyngngam",
         added: "2009-07-29"}).
iana(language,lyn,tag{
         type: "language",
         subtag: "lyn",
         description: "Luyana",
         added: "2009-07-29"}).
iana(language,lzh,tag{
         type: "language",
         subtag: "lzh",
         description: "Literary Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,lzl,tag{
         type: "language",
         subtag: "lzl",
         description: "Litzlitz",
         added: "2009-07-29"}).
iana(language,lzn,tag{
         type: "language",
         subtag: "lzn",
         description: "Leinong Naga",
         added: "2009-07-29"}).
iana(language,lzz,tag{
         type: "language",
         subtag: "lzz",
         description: "Laz",
         added: "2009-07-29"}).
iana(language,maa,tag{
         type: "language",
         subtag: "maa",
         description: "San Jerónimo Tecóatl Mazatec",
         added: "2009-07-29",
         comments: "see also pbm"}).
iana(language,mab,tag{
         type: "language",
         subtag: "mab",
         description: "Yutanduchi Mixtec",
         added: "2009-07-29"}).
iana(language,mad,tag{
         type: "language",
         subtag: "mad",
         description: "Madurese",
         added: "2005-10-16"}).
iana(language,mae,tag{
         type: "language",
         subtag: "mae",
         description: "Bo-Rukul",
         added: "2009-07-29"}).
iana(language,maf,tag{
         type: "language",
         subtag: "maf",
         description: "Mafa",
         added: "2009-07-29"}).
iana(language,mag,tag{
         type: "language",
         subtag: "mag",
         description: "Magahi",
         added: "2005-10-16"}).
iana(language,mai,tag{
         type: "language",
         subtag: "mai",
         description: "Maithili",
         added: "2005-10-16",
         suppress_script: "Deva"}).
iana(language,maj,tag{
         type: "language",
         subtag: "maj",
         description: "Jalapa De Díaz Mazatec",
         added: "2009-07-29"}).
iana(language,mak,tag{
         type: "language",
         subtag: "mak",
         description: "Makasar",
         added: "2005-10-16"}).
iana(language,mam,tag{
         type: "language",
         subtag: "mam",
         description: "Mam",
         added: "2009-07-29"}).
iana(language,man,tag{
         type: "language",
         subtag: "man",
         description: ["Mandingo","Manding"],
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,map,tag{
         type: "language",
         subtag: "map",
         description: "Austronesian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,maq,tag{
         type: "language",
         subtag: "maq",
         description: "Chiquihuitlán Mazatec",
         added: "2009-07-29"}).
iana(language,mas,tag{
         type: "language",
         subtag: "mas",
         description: "Masai",
         added: "2005-10-16"}).
iana(language,mat,tag{
         type: "language",
         subtag: "mat",
         description: "San Francisco Matlatzinca",
         added: "2009-07-29"}).
iana(language,mau,tag{
         type: "language",
         subtag: "mau",
         description: "Huautla Mazatec",
         added: "2009-07-29"}).
iana(language,mav,tag{
         type: "language",
         subtag: "mav",
         description: "Sateré-Mawé",
         added: "2009-07-29"}).
iana(language,maw,tag{
         type: "language",
         subtag: "maw",
         description: "Mampruli",
         added: "2009-07-29"}).
iana(language,max,tag{
         type: "language",
         subtag: "max",
         description: "North Moluccan Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,maz,tag{
         type: "language",
         subtag: "maz",
         description: "Central Mazahua",
         added: "2009-07-29"}).
iana(language,mba,tag{
         type: "language",
         subtag: "mba",
         description: "Higaonon",
         added: "2009-07-29"}).
iana(language,mbb,tag{
         type: "language",
         subtag: "mbb",
         description: "Western Bukidnon Manobo",
         added: "2009-07-29"}).
iana(language,mbc,tag{
         type: "language",
         subtag: "mbc",
         description: "Macushi",
         added: "2009-07-29"}).
iana(language,mbd,tag{
         type: "language",
         subtag: "mbd",
         description: "Dibabawon Manobo",
         added: "2009-07-29"}).
iana(language,mbe,tag{
         type: "language",
         subtag: "mbe",
         description: "Molale",
         added: "2009-07-29"}).
iana(language,mbf,tag{
         type: "language",
         subtag: "mbf",
         description: "Baba Malay",
         added: "2009-07-29"}).
iana(language,mbh,tag{
         type: "language",
         subtag: "mbh",
         description: "Mangseng",
         added: "2009-07-29"}).
iana(language,mbi,tag{
         type: "language",
         subtag: "mbi",
         description: "Ilianen Manobo",
         added: "2009-07-29"}).
iana(language,mbj,tag{
         type: "language",
         subtag: "mbj",
         description: "Nadëb",
         added: "2009-07-29"}).
iana(language,mbk,tag{
         type: "language",
         subtag: "mbk",
         description: "Malol",
         added: "2009-07-29"}).
iana(language,mbl,tag{
         type: "language",
         subtag: "mbl",
         description: "Maxakalí",
         added: "2009-07-29"}).
iana(language,mbm,tag{
         type: "language",
         subtag: "mbm",
         description: "Ombamba",
         added: "2009-07-29"}).
iana(language,mbn,tag{
         type: "language",
         subtag: "mbn",
         description: "Macaguán",
         added: "2009-07-29"}).
iana(language,mbo,tag{
         type: "language",
         subtag: "mbo",
         description: "Mbo (Cameroon)",
         added: "2009-07-29"}).
iana(language,mbp,tag{
         type: "language",
         subtag: "mbp",
         description: "Malayo",
         added: "2009-07-29"}).
iana(language,mbq,tag{
         type: "language",
         subtag: "mbq",
         description: "Maisin",
         added: "2009-07-29"}).
iana(language,mbr,tag{
         type: "language",
         subtag: "mbr",
         description: "Nukak Makú",
         added: "2009-07-29"}).
iana(language,mbs,tag{
         type: "language",
         subtag: "mbs",
         description: "Sarangani Manobo",
         added: "2009-07-29"}).
iana(language,mbt,tag{
         type: "language",
         subtag: "mbt",
         description: "Matigsalug Manobo",
         added: "2009-07-29"}).
iana(language,mbu,tag{
         type: "language",
         subtag: "mbu",
         description: "Mbula-Bwazza",
         added: "2009-07-29"}).
iana(language,mbv,tag{
         type: "language",
         subtag: "mbv",
         description: "Mbulungish",
         added: "2009-07-29"}).
iana(language,mbw,tag{
         type: "language",
         subtag: "mbw",
         description: "Maring",
         added: "2009-07-29"}).
iana(language,mbx,tag{
         type: "language",
         subtag: "mbx",
         description: "Mari (East Sepik Province)",
         added: "2009-07-29"}).
iana(language,mby,tag{
         type: "language",
         subtag: "mby",
         description: "Memoni",
         added: "2009-07-29"}).
iana(language,mbz,tag{
         type: "language",
         subtag: "mbz",
         description: "Amoltepec Mixtec",
         added: "2009-07-29"}).
iana(language,mca,tag{
         type: "language",
         subtag: "mca",
         description: "Maca",
         added: "2009-07-29"}).
iana(language,mcb,tag{
         type: "language",
         subtag: "mcb",
         description: "Machiguenga",
         added: "2009-07-29"}).
iana(language,mcc,tag{
         type: "language",
         subtag: "mcc",
         description: "Bitur",
         added: "2009-07-29"}).
iana(language,mcd,tag{
         type: "language",
         subtag: "mcd",
         description: "Sharanahua",
         added: "2009-07-29"}).
iana(language,mce,tag{
         type: "language",
         subtag: "mce",
         description: "Itundujia Mixtec",
         added: "2009-07-29"}).
iana(language,mcf,tag{
         type: "language",
         subtag: "mcf",
         description: "Matsés",
         added: "2009-07-29"}).
iana(language,mcg,tag{
         type: "language",
         subtag: "mcg",
         description: "Mapoyo",
         added: "2009-07-29"}).
iana(language,mch,tag{
         type: "language",
         subtag: "mch",
         description: "Maquiritari",
         added: "2009-07-29"}).
iana(language,mci,tag{
         type: "language",
         subtag: "mci",
         description: "Mese",
         added: "2009-07-29"}).
iana(language,mcj,tag{
         type: "language",
         subtag: "mcj",
         description: "Mvanip",
         added: "2009-07-29"}).
iana(language,mck,tag{
         type: "language",
         subtag: "mck",
         description: "Mbunda",
         added: "2009-07-29"}).
iana(language,mcl,tag{
         type: "language",
         subtag: "mcl",
         description: "Macaguaje",
         added: "2009-07-29"}).
iana(language,mcm,tag{
         type: "language",
         subtag: "mcm",
         description: "Malaccan Creole Portuguese",
         added: "2009-07-29"}).
iana(language,mcn,tag{
         type: "language",
         subtag: "mcn",
         description: "Masana",
         added: "2009-07-29"}).
iana(language,mco,tag{
         type: "language",
         subtag: "mco",
         description: "Coatlán Mixe",
         added: "2009-07-29"}).
iana(language,mcp,tag{
         type: "language",
         subtag: "mcp",
         description: "Makaa",
         added: "2009-07-29"}).
iana(language,mcq,tag{
         type: "language",
         subtag: "mcq",
         description: "Ese",
         added: "2009-07-29"}).
iana(language,mcr,tag{
         type: "language",
         subtag: "mcr",
         description: "Menya",
         added: "2009-07-29"}).
iana(language,mcs,tag{
         type: "language",
         subtag: "mcs",
         description: "Mambai",
         added: "2009-07-29"}).
iana(language,mct,tag{
         type: "language",
         subtag: "mct",
         description: "Mengisa",
         added: "2009-07-29"}).
iana(language,mcu,tag{
         type: "language",
         subtag: "mcu",
         description: "Cameroon Mambila",
         added: "2009-07-29"}).
iana(language,mcv,tag{
         type: "language",
         subtag: "mcv",
         description: "Minanibai",
         added: "2009-07-29"}).
iana(language,mcw,tag{
         type: "language",
         subtag: "mcw",
         description: "Mawa (Chad)",
         added: "2009-07-29"}).
iana(language,mcx,tag{
         type: "language",
         subtag: "mcx",
         description: "Mpiemo",
         added: "2009-07-29"}).
iana(language,mcy,tag{
         type: "language",
         subtag: "mcy",
         description: "South Watut",
         added: "2009-07-29"}).
iana(language,mcz,tag{
         type: "language",
         subtag: "mcz",
         description: "Mawan",
         added: "2009-07-29"}).
iana(language,mda,tag{
         type: "language",
         subtag: "mda",
         description: "Mada (Nigeria)",
         added: "2009-07-29"}).
iana(language,mdb,tag{
         type: "language",
         subtag: "mdb",
         description: "Morigi",
         added: "2009-07-29"}).
iana(language,mdc,tag{
         type: "language",
         subtag: "mdc",
         description: "Male (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,mdd,tag{
         type: "language",
         subtag: "mdd",
         description: "Mbum",
         added: "2009-07-29"}).
iana(language,mde,tag{
         type: "language",
         subtag: "mde",
         description: "Maba (Chad)",
         added: "2009-07-29"}).
iana(language,mdf,tag{
         type: "language",
         subtag: "mdf",
         description: "Moksha",
         added: "2005-10-16"}).
iana(language,mdg,tag{
         type: "language",
         subtag: "mdg",
         description: "Massalat",
         added: "2009-07-29"}).
iana(language,mdh,tag{
         type: "language",
         subtag: "mdh",
         description: "Maguindanaon",
         added: "2009-07-29"}).
iana(language,mdi,tag{
         type: "language",
         subtag: "mdi",
         description: "Mamvu",
         added: "2009-07-29"}).
iana(language,mdj,tag{
         type: "language",
         subtag: "mdj",
         description: "Mangbetu",
         added: "2009-07-29"}).
iana(language,mdk,tag{
         type: "language",
         subtag: "mdk",
         description: "Mangbutu",
         added: "2009-07-29"}).
iana(language,mdl,tag{
         type: "language",
         subtag: "mdl",
         description: "Maltese Sign Language",
         added: "2009-07-29"}).
iana(language,mdm,tag{
         type: "language",
         subtag: "mdm",
         description: "Mayogo",
         added: "2009-07-29"}).
iana(language,mdn,tag{
         type: "language",
         subtag: "mdn",
         description: "Mbati",
         added: "2009-07-29"}).
iana(language,mdp,tag{
         type: "language",
         subtag: "mdp",
         description: "Mbala",
         added: "2009-07-29"}).
iana(language,mdq,tag{
         type: "language",
         subtag: "mdq",
         description: "Mbole",
         added: "2009-07-29"}).
iana(language,mdr,tag{
         type: "language",
         subtag: "mdr",
         description: "Mandar",
         added: "2005-10-16"}).
iana(language,mds,tag{
         type: "language",
         subtag: "mds",
         description: "Maria (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,mdt,tag{
         type: "language",
         subtag: "mdt",
         description: "Mbere",
         added: "2009-07-29"}).
iana(language,mdu,tag{
         type: "language",
         subtag: "mdu",
         description: "Mboko",
         added: "2009-07-29"}).
iana(language,mdv,tag{
         type: "language",
         subtag: "mdv",
         description: "Santa Lucía Monteverde Mixtec",
         added: "2009-07-29"}).
iana(language,mdw,tag{
         type: "language",
         subtag: "mdw",
         description: "Mbosi",
         added: "2009-07-29"}).
iana(language,mdx,tag{
         type: "language",
         subtag: "mdx",
         description: "Dizin",
         added: "2009-07-29"}).
iana(language,mdy,tag{
         type: "language",
         subtag: "mdy",
         description: "Male (Ethiopia)",
         added: "2009-07-29"}).
iana(language,mdz,tag{
         type: "language",
         subtag: "mdz",
         description: "Suruí Do Pará",
         added: "2009-07-29"}).
iana(language,mea,tag{
         type: "language",
         subtag: "mea",
         description: "Menka",
         added: "2009-07-29"}).
iana(language,meb,tag{
         type: "language",
         subtag: "meb",
         description: "Ikobi",
         added: "2009-07-29"}).
iana(language,mec,tag{
         type: "language",
         subtag: "mec",
         description: "Marra",
         added: "2009-07-29"}).
iana(language,med,tag{
         type: "language",
         subtag: "med",
         description: "Melpa",
         added: "2009-07-29"}).
iana(language,mee,tag{
         type: "language",
         subtag: "mee",
         description: "Mengen",
         added: "2009-07-29"}).
iana(language,mef,tag{
         type: "language",
         subtag: "mef",
         description: "Megam",
         added: "2009-07-29"}).
iana(language,meg,tag{
         type: "language",
         subtag: "meg",
         description: "Mea",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "cir"}).
iana(language,meh,tag{
         type: "language",
         subtag: "meh",
         description: "Southwestern Tlaxiaco Mixtec",
         added: "2009-07-29"}).
iana(language,mei,tag{
         type: "language",
         subtag: "mei",
         description: "Midob",
         added: "2009-07-29"}).
iana(language,mej,tag{
         type: "language",
         subtag: "mej",
         description: "Meyah",
         added: "2009-07-29"}).
iana(language,mek,tag{
         type: "language",
         subtag: "mek",
         description: "Mekeo",
         added: "2009-07-29"}).
iana(language,mel,tag{
         type: "language",
         subtag: "mel",
         description: "Central Melanau",
         added: "2009-07-29"}).
iana(language,mem,tag{
         type: "language",
         subtag: "mem",
         description: "Mangala",
         added: "2009-07-29"}).
iana(language,men,tag{
         type: "language",
         subtag: "men",
         description: "Mende (Sierra Leone)",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,meo,tag{
         type: "language",
         subtag: "meo",
         description: "Kedah Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,mep,tag{
         type: "language",
         subtag: "mep",
         description: "Miriwoong",
         added: "2009-07-29"}).
iana(language,meq,tag{
         type: "language",
         subtag: "meq",
         description: "Merey",
         added: "2009-07-29"}).
iana(language,mer,tag{
         type: "language",
         subtag: "mer",
         description: "Meru",
         added: "2009-07-29"}).
iana(language,mes,tag{
         type: "language",
         subtag: "mes",
         description: "Masmaje",
         added: "2009-07-29"}).
iana(language,met,tag{
         type: "language",
         subtag: "met",
         description: "Mato",
         added: "2009-07-29"}).
iana(language,meu,tag{
         type: "language",
         subtag: "meu",
         description: "Motu",
         added: "2009-07-29"}).
iana(language,mev,tag{
         type: "language",
         subtag: "mev",
         description: "Mano",
         added: "2009-07-29"}).
iana(language,mew,tag{
         type: "language",
         subtag: "mew",
         description: "Maaka",
         added: "2009-07-29"}).
iana(language,mey,tag{
         type: "language",
         subtag: "mey",
         description: "Hassaniyya",
         added: "2009-07-29"}).
iana(language,mez,tag{
         type: "language",
         subtag: "mez",
         description: "Menominee",
         added: "2009-07-29"}).
iana(language,mfa,tag{
         type: "language",
         subtag: "mfa",
         description: "Pattani Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,mfb,tag{
         type: "language",
         subtag: "mfb",
         description: "Bangka",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,mfc,tag{
         type: "language",
         subtag: "mfc",
         description: "Mba",
         added: "2009-07-29"}).
iana(language,mfd,tag{
         type: "language",
         subtag: "mfd",
         description: "Mendankwe-Nkwen",
         added: "2009-07-29"}).
iana(language,mfe,tag{
         type: "language",
         subtag: "mfe",
         description: "Morisyen",
         added: "2009-07-29"}).
iana(language,mff,tag{
         type: "language",
         subtag: "mff",
         description: "Naki",
         added: "2009-07-29"}).
iana(language,mfg,tag{
         type: "language",
         subtag: "mfg",
         description: "Mogofin",
         added: "2009-07-29"}).
iana(language,mfh,tag{
         type: "language",
         subtag: "mfh",
         description: "Matal",
         added: "2009-07-29"}).
iana(language,mfi,tag{
         type: "language",
         subtag: "mfi",
         description: "Wandala",
         added: "2009-07-29"}).
iana(language,mfj,tag{
         type: "language",
         subtag: "mfj",
         description: "Mefele",
         added: "2009-07-29"}).
iana(language,mfk,tag{
         type: "language",
         subtag: "mfk",
         description: "North Mofu",
         added: "2009-07-29"}).
iana(language,mfl,tag{
         type: "language",
         subtag: "mfl",
         description: "Putai",
         added: "2009-07-29"}).
iana(language,mfm,tag{
         type: "language",
         subtag: "mfm",
         description: "Marghi South",
         added: "2009-07-29"}).
iana(language,mfn,tag{
         type: "language",
         subtag: "mfn",
         description: "Cross River Mbembe",
         added: "2009-07-29"}).
iana(language,mfo,tag{
         type: "language",
         subtag: "mfo",
         description: "Mbe",
         added: "2009-07-29"}).
iana(language,mfp,tag{
         type: "language",
         subtag: "mfp",
         description: "Makassar Malay",
         added: "2009-07-29"}).
iana(language,mfq,tag{
         type: "language",
         subtag: "mfq",
         description: "Moba",
         added: "2009-07-29"}).
iana(language,mfr,tag{
         type: "language",
         subtag: "mfr",
         description: "Marrithiyel",
         added: "2009-07-29"}).
iana(language,mfs,tag{
         type: "language",
         subtag: "mfs",
         description: "Mexican Sign Language",
         added: "2009-07-29"}).
iana(language,mft,tag{
         type: "language",
         subtag: "mft",
         description: "Mokerang",
         added: "2009-07-29"}).
iana(language,mfu,tag{
         type: "language",
         subtag: "mfu",
         description: "Mbwela",
         added: "2009-07-29"}).
iana(language,mfv,tag{
         type: "language",
         subtag: "mfv",
         description: "Mandjak",
         added: "2009-07-29"}).
iana(language,mfw,tag{
         type: "language",
         subtag: "mfw",
         description: "Mulaha",
         added: "2009-07-29"}).
iana(language,mfx,tag{
         type: "language",
         subtag: "mfx",
         description: "Melo",
         added: "2009-07-29"}).
iana(language,mfy,tag{
         type: "language",
         subtag: "mfy",
         description: "Mayo",
         added: "2009-07-29"}).
iana(language,mfz,tag{
         type: "language",
         subtag: "mfz",
         description: "Mabaan",
         added: "2009-07-29"}).
iana(language,mga,tag{
         type: "language",
         subtag: "mga",
         description: "Middle Irish (900-1200)",
         added: "2005-10-16"}).
iana(language,mgb,tag{
         type: "language",
         subtag: "mgb",
         description: "Mararit",
         added: "2009-07-29"}).
iana(language,mgc,tag{
         type: "language",
         subtag: "mgc",
         description: "Morokodo",
         added: "2009-07-29"}).
iana(language,mgd,tag{
         type: "language",
         subtag: "mgd",
         description: "Moru",
         added: "2009-07-29"}).
iana(language,mge,tag{
         type: "language",
         subtag: "mge",
         description: "Mango",
         added: "2009-07-29"}).
iana(language,mgf,tag{
         type: "language",
         subtag: "mgf",
         description: "Maklew",
         added: "2009-07-29"}).
iana(language,mgg,tag{
         type: "language",
         subtag: "mgg",
         description: "Mpumpong",
         added: "2009-07-29"}).
iana(language,mgh,tag{
         type: "language",
         subtag: "mgh",
         description: "Makhuwa-Meetto",
         added: "2009-07-29"}).
iana(language,mgi,tag{
         type: "language",
         subtag: "mgi",
         description: "Lijili",
         added: "2009-07-29"}).
iana(language,mgj,tag{
         type: "language",
         subtag: "mgj",
         description: "Abureni",
         added: "2009-07-29"}).
iana(language,mgk,tag{
         type: "language",
         subtag: "mgk",
         description: "Mawes",
         added: "2009-07-29"}).
iana(language,mgl,tag{
         type: "language",
         subtag: "mgl",
         description: "Maleu-Kilenge",
         added: "2009-07-29"}).
iana(language,mgm,tag{
         type: "language",
         subtag: "mgm",
         description: "Mambae",
         added: "2009-07-29"}).
iana(language,mgn,tag{
         type: "language",
         subtag: "mgn",
         description: "Mbangi",
         added: "2009-07-29"}).
iana(language,mgo,tag{
         type: "language",
         subtag: "mgo",
         description: "Meta'",
         added: "2009-07-29"}).
iana(language,mgp,tag{
         type: "language",
         subtag: "mgp",
         description: "Eastern Magar",
         added: "2009-07-29"}).
iana(language,mgq,tag{
         type: "language",
         subtag: "mgq",
         description: "Malila",
         added: "2009-07-29"}).
iana(language,mgr,tag{
         type: "language",
         subtag: "mgr",
         description: "Mambwe-Lungu",
         added: "2009-07-29"}).
iana(language,mgs,tag{
         type: "language",
         subtag: "mgs",
         description: "Manda (Tanzania)",
         added: "2009-07-29"}).
iana(language,mgt,tag{
         type: "language",
         subtag: "mgt",
         description: "Mongol",
         added: "2009-07-29"}).
iana(language,mgu,tag{
         type: "language",
         subtag: "mgu",
         description: "Mailu",
         added: "2009-07-29"}).
iana(language,mgv,tag{
         type: "language",
         subtag: "mgv",
         description: "Matengo",
         added: "2009-07-29"}).
iana(language,mgw,tag{
         type: "language",
         subtag: "mgw",
         description: "Matumbi",
         added: "2009-07-29"}).
iana(language,mgx,tag{
         type: "language",
         subtag: "mgx",
         description: "Omati",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see jbk, jmw"}).
iana(language,mgy,tag{
         type: "language",
         subtag: "mgy",
         description: "Mbunga",
         added: "2009-07-29"}).
iana(language,mgz,tag{
         type: "language",
         subtag: "mgz",
         description: "Mbugwe",
         added: "2009-07-29"}).
iana(language,mha,tag{
         type: "language",
         subtag: "mha",
         description: "Manda (India)",
         added: "2009-07-29"}).
iana(language,mhb,tag{
         type: "language",
         subtag: "mhb",
         description: "Mahongwe",
         added: "2009-07-29"}).
iana(language,mhc,tag{
         type: "language",
         subtag: "mhc",
         description: "Mocho",
         added: "2009-07-29"}).
iana(language,mhd,tag{
         type: "language",
         subtag: "mhd",
         description: "Mbugu",
         added: "2009-07-29"}).
iana(language,mhe,tag{
         type: "language",
         subtag: "mhe",
         description: ["Besisi","Mah Meri"],
         added: "2009-07-29"}).
iana(language,mhf,tag{
         type: "language",
         subtag: "mhf",
         description: "Mamaa",
         added: "2009-07-29"}).
iana(language,mhg,tag{
         type: "language",
         subtag: "mhg",
         description: "Margu",
         added: "2009-07-29"}).
iana(language,mhh,tag{
         type: "language",
         subtag: "mhh",
         description: "Maskoy Pidgin",
         added: "2009-07-29",
         deprecated: "2014-02-28"}).
iana(language,mhi,tag{
         type: "language",
         subtag: "mhi",
         description: "Ma'di",
         added: "2009-07-29"}).
iana(language,mhj,tag{
         type: "language",
         subtag: "mhj",
         description: "Mogholi",
         added: "2009-07-29"}).
iana(language,mhk,tag{
         type: "language",
         subtag: "mhk",
         description: "Mungaka",
         added: "2009-07-29"}).
iana(language,mhl,tag{
         type: "language",
         subtag: "mhl",
         description: "Mauwake",
         added: "2009-07-29"}).
iana(language,mhm,tag{
         type: "language",
         subtag: "mhm",
         description: "Makhuwa-Moniga",
         added: "2009-07-29"}).
iana(language,mhn,tag{
         type: "language",
         subtag: "mhn",
         description: "Mócheno",
         added: "2009-07-29"}).
iana(language,mho,tag{
         type: "language",
         subtag: "mho",
         description: "Mashi (Zambia)",
         added: "2009-07-29"}).
iana(language,mhp,tag{
         type: "language",
         subtag: "mhp",
         description: "Balinese Malay",
         added: "2009-07-29"}).
iana(language,mhq,tag{
         type: "language",
         subtag: "mhq",
         description: "Mandan",
         added: "2009-07-29"}).
iana(language,mhr,tag{
         type: "language",
         subtag: "mhr",
         description: "Eastern Mari",
         added: "2009-07-29",
         macrolanguage: "chm"}).
iana(language,mhs,tag{
         type: "language",
         subtag: "mhs",
         description: "Buru (Indonesia)",
         added: "2009-07-29"}).
iana(language,mht,tag{
         type: "language",
         subtag: "mht",
         description: "Mandahuaca",
         added: "2009-07-29"}).
iana(language,mhu,tag{
         type: "language",
         subtag: "mhu",
         description: ["Digaro-Mishmi","Darang Deng"],
         added: "2009-07-29"}).
iana(language,mhw,tag{
         type: "language",
         subtag: "mhw",
         description: "Mbukushu",
         added: "2009-07-29"}).
iana(language,mhx,tag{
         type: "language",
         subtag: "mhx",
         description: ["Maru","Lhaovo"],
         added: "2009-07-29"}).
iana(language,mhy,tag{
         type: "language",
         subtag: "mhy",
         description: "Ma'anyan",
         added: "2009-07-29"}).
iana(language,mhz,tag{
         type: "language",
         subtag: "mhz",
         description: "Mor (Mor Islands)",
         added: "2009-07-29"}).
iana(language,mia,tag{
         type: "language",
         subtag: "mia",
         description: "Miami",
         added: "2009-07-29"}).
iana(language,mib,tag{
         type: "language",
         subtag: "mib",
         description: "Atatláhuca Mixtec",
         added: "2009-07-29"}).
iana(language,mic,tag{
         type: "language",
         subtag: "mic",
         description: ["Mi'kmaq","Micmac"],
         added: "2005-10-16"}).
iana(language,mid,tag{
         type: "language",
         subtag: "mid",
         description: "Mandaic",
         added: "2009-07-29"}).
iana(language,mie,tag{
         type: "language",
         subtag: "mie",
         description: "Ocotepec Mixtec",
         added: "2009-07-29"}).
iana(language,mif,tag{
         type: "language",
         subtag: "mif",
         description: "Mofu-Gudur",
         added: "2009-07-29"}).
iana(language,mig,tag{
         type: "language",
         subtag: "mig",
         description: "San Miguel El Grande Mixtec",
         added: "2009-07-29"}).
iana(language,mih,tag{
         type: "language",
         subtag: "mih",
         description: "Chayuco Mixtec",
         added: "2009-07-29"}).
iana(language,mii,tag{
         type: "language",
         subtag: "mii",
         description: "Chigmecatitlán Mixtec",
         added: "2009-07-29"}).
iana(language,mij,tag{
         type: "language",
         subtag: "mij",
         description: ["Abar","Mungbam"],
         added: "2009-07-29"}).
iana(language,mik,tag{
         type: "language",
         subtag: "mik",
         description: "Mikasuki",
         added: "2009-07-29"}).
iana(language,mil,tag{
         type: "language",
         subtag: "mil",
         description: "Peñoles Mixtec",
         added: "2009-07-29"}).
iana(language,mim,tag{
         type: "language",
         subtag: "mim",
         description: "Alacatlatzala Mixtec",
         added: "2009-07-29"}).
iana(language,min,tag{
         type: "language",
         subtag: "min",
         description: "Minangkabau",
         added: "2005-10-16",
         macrolanguage: "ms"}).
iana(language,mio,tag{
         type: "language",
         subtag: "mio",
         description: "Pinotepa Nacional Mixtec",
         added: "2009-07-29"}).
iana(language,mip,tag{
         type: "language",
         subtag: "mip",
         description: "Apasco-Apoala Mixtec",
         added: "2009-07-29"}).
iana(language,miq,tag{
         type: "language",
         subtag: "miq",
         description: "Mískito",
         added: "2009-07-29"}).
iana(language,mir,tag{
         type: "language",
         subtag: "mir",
         description: "Isthmus Mixe",
         added: "2009-07-29"}).
iana(language,mis,tag{
         type: "language",
         subtag: "mis",
         description: "Uncoded languages",
         added: "2005-10-16",
         scope: "special"}).
iana(language,mit,tag{
         type: "language",
         subtag: "mit",
         description: "Southern Puebla Mixtec",
         added: "2009-07-29"}).
iana(language,miu,tag{
         type: "language",
         subtag: "miu",
         description: "Cacaloxtepec Mixtec",
         added: "2009-07-29"}).
iana(language,miw,tag{
         type: "language",
         subtag: "miw",
         description: "Akoye",
         added: "2009-07-29"}).
iana(language,mix,tag{
         type: "language",
         subtag: "mix",
         description: "Mixtepec Mixtec",
         added: "2009-07-29"}).
iana(language,miy,tag{
         type: "language",
         subtag: "miy",
         description: "Ayutla Mixtec",
         added: "2009-07-29"}).
iana(language,miz,tag{
         type: "language",
         subtag: "miz",
         description: "Coatzospan Mixtec",
         added: "2009-07-29"}).
iana(language,mja,tag{
         type: "language",
         subtag: "mja",
         description: "Mahei",
         added: "2009-07-29",
         deprecated: "2011-08-16"}).
iana(language,mjb,tag{
         type: "language",
         subtag: "mjb",
         description: "Makalero",
         added: "2016-05-30"}).
iana(language,mjc,tag{
         type: "language",
         subtag: "mjc",
         description: "San Juan Colorado Mixtec",
         added: "2009-07-29"}).
iana(language,mjd,tag{
         type: "language",
         subtag: "mjd",
         description: "Northwest Maidu",
         added: "2009-07-29"}).
iana(language,mje,tag{
         type: "language",
         subtag: "mje",
         description: "Muskum",
         added: "2009-07-29"}).
iana(language,mjg,tag{
         type: "language",
         subtag: "mjg",
         description: "Tu",
         added: "2009-07-29"}).
iana(language,mjh,tag{
         type: "language",
         subtag: "mjh",
         description: "Mwera (Nyasa)",
         added: "2009-07-29"}).
iana(language,mji,tag{
         type: "language",
         subtag: "mji",
         description: "Kim Mun",
         added: "2009-07-29"}).
iana(language,mjj,tag{
         type: "language",
         subtag: "mjj",
         description: "Mawak",
         added: "2009-07-29"}).
iana(language,mjk,tag{
         type: "language",
         subtag: "mjk",
         description: "Matukar",
         added: "2009-07-29"}).
iana(language,mjl,tag{
         type: "language",
         subtag: "mjl",
         description: "Mandeali",
         added: "2009-07-29"}).
iana(language,mjm,tag{
         type: "language",
         subtag: "mjm",
         description: "Medebur",
         added: "2009-07-29"}).
iana(language,mjn,tag{
         type: "language",
         subtag: "mjn",
         description: "Ma (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,mjo,tag{
         type: "language",
         subtag: "mjo",
         description: "Malankuravan",
         added: "2009-07-29"}).
iana(language,mjp,tag{
         type: "language",
         subtag: "mjp",
         description: "Malapandaram",
         added: "2009-07-29"}).
iana(language,mjq,tag{
         type: "language",
         subtag: "mjq",
         description: "Malaryan",
         added: "2009-07-29"}).
iana(language,mjr,tag{
         type: "language",
         subtag: "mjr",
         description: "Malavedan",
         added: "2009-07-29"}).
iana(language,mjs,tag{
         type: "language",
         subtag: "mjs",
         description: "Miship",
         added: "2009-07-29"}).
iana(language,mjt,tag{
         type: "language",
         subtag: "mjt",
         description: "Sauria Paharia",
         added: "2009-07-29"}).
iana(language,mju,tag{
         type: "language",
         subtag: "mju",
         description: "Manna-Dora",
         added: "2009-07-29"}).
iana(language,mjv,tag{
         type: "language",
         subtag: "mjv",
         description: "Mannan",
         added: "2009-07-29"}).
iana(language,mjw,tag{
         type: "language",
         subtag: "mjw",
         description: "Karbi",
         added: "2009-07-29"}).
iana(language,mjx,tag{
         type: "language",
         subtag: "mjx",
         description: "Mahali",
         added: "2009-07-29"}).
iana(language,mjy,tag{
         type: "language",
         subtag: "mjy",
         description: "Mahican",
         added: "2009-07-29"}).
iana(language,mjz,tag{
         type: "language",
         subtag: "mjz",
         description: "Majhi",
         added: "2009-07-29"}).
iana(language,mka,tag{
         type: "language",
         subtag: "mka",
         description: "Mbre",
         added: "2009-07-29"}).
iana(language,mkb,tag{
         type: "language",
         subtag: "mkb",
         description: "Mal Paharia",
         added: "2009-07-29"}).
iana(language,mkc,tag{
         type: "language",
         subtag: "mkc",
         description: "Siliput",
         added: "2009-07-29"}).
iana(language,mke,tag{
         type: "language",
         subtag: "mke",
         description: "Mawchi",
         added: "2009-07-29"}).
iana(language,mkf,tag{
         type: "language",
         subtag: "mkf",
         description: "Miya",
         added: "2009-07-29"}).
iana(language,mkg,tag{
         type: "language",
         subtag: "mkg",
         description: "Mak (China)",
         added: "2009-07-29"}).
iana(language,mkh,tag{
         type: "language",
         subtag: "mkh",
         description: "Mon-Khmer languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,mki,tag{
         type: "language",
         subtag: "mki",
         description: "Dhatki",
         added: "2009-07-29"}).
iana(language,mkj,tag{
         type: "language",
         subtag: "mkj",
         description: "Mokilese",
         added: "2009-07-29"}).
iana(language,mkk,tag{
         type: "language",
         subtag: "mkk",
         description: "Byep",
         added: "2009-07-29"}).
iana(language,mkl,tag{
         type: "language",
         subtag: "mkl",
         description: "Mokole",
         added: "2009-07-29"}).
iana(language,mkm,tag{
         type: "language",
         subtag: "mkm",
         description: "Moklen",
         added: "2009-07-29"}).
iana(language,mkn,tag{
         type: "language",
         subtag: "mkn",
         description: "Kupang Malay",
         added: "2009-07-29"}).
iana(language,mko,tag{
         type: "language",
         subtag: "mko",
         description: "Mingang Doso",
         added: "2009-07-29"}).
iana(language,mkp,tag{
         type: "language",
         subtag: "mkp",
         description: "Moikodi",
         added: "2009-07-29"}).
iana(language,mkq,tag{
         type: "language",
         subtag: "mkq",
         description: "Bay Miwok",
         added: "2009-07-29"}).
iana(language,mkr,tag{
         type: "language",
         subtag: "mkr",
         description: "Malas",
         added: "2009-07-29"}).
iana(language,mks,tag{
         type: "language",
         subtag: "mks",
         description: "Silacayoapan Mixtec",
         added: "2009-07-29"}).
iana(language,mkt,tag{
         type: "language",
         subtag: "mkt",
         description: "Vamale",
         added: "2009-07-29"}).
iana(language,mku,tag{
         type: "language",
         subtag: "mku",
         description: "Konyanka Maninka",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,mkv,tag{
         type: "language",
         subtag: "mkv",
         description: "Mafea",
         added: "2009-07-29"}).
iana(language,mkw,tag{
         type: "language",
         subtag: "mkw",
         description: "Kituba (Congo)",
         added: "2009-07-29"}).
iana(language,mkx,tag{
         type: "language",
         subtag: "mkx",
         description: "Kinamiging Manobo",
         added: "2009-07-29"}).
iana(language,mky,tag{
         type: "language",
         subtag: "mky",
         description: "East Makian",
         added: "2009-07-29"}).
iana(language,mkz,tag{
         type: "language",
         subtag: "mkz",
         description: "Makasae",
         added: "2009-07-29"}).
iana(language,mla,tag{
         type: "language",
         subtag: "mla",
         description: "Malo",
         added: "2009-07-29"}).
iana(language,mlb,tag{
         type: "language",
         subtag: "mlb",
         description: "Mbule",
         added: "2009-07-29"}).
iana(language,mlc,tag{
         type: "language",
         subtag: "mlc",
         description: "Cao Lan",
         added: "2009-07-29"}).
iana(language,mld,tag{
         type: "language",
         subtag: "mld",
         description: "Malakhel",
         added: "2009-07-29",
         deprecated: "2013-09-10"}).
iana(language,mle,tag{
         type: "language",
         subtag: "mle",
         description: "Manambu",
         added: "2009-07-29"}).
iana(language,mlf,tag{
         type: "language",
         subtag: "mlf",
         description: "Mal",
         added: "2009-07-29"}).
iana(language,mlh,tag{
         type: "language",
         subtag: "mlh",
         description: "Mape",
         added: "2009-07-29"}).
iana(language,mli,tag{
         type: "language",
         subtag: "mli",
         description: "Malimpung",
         added: "2009-07-29"}).
iana(language,mlj,tag{
         type: "language",
         subtag: "mlj",
         description: "Miltu",
         added: "2009-07-29"}).
iana(language,mlk,tag{
         type: "language",
         subtag: "mlk",
         description: ["Ilwana","Kiwilwana"],
         added: "2009-07-29"}).
iana(language,mll,tag{
         type: "language",
         subtag: "mll",
         description: "Malua Bay",
         added: "2009-07-29"}).
iana(language,mlm,tag{
         type: "language",
         subtag: "mlm",
         description: "Mulam",
         added: "2009-07-29"}).
iana(language,mln,tag{
         type: "language",
         subtag: "mln",
         description: "Malango",
         added: "2009-07-29"}).
iana(language,mlo,tag{
         type: "language",
         subtag: "mlo",
         description: "Mlomp",
         added: "2009-07-29"}).
iana(language,mlp,tag{
         type: "language",
         subtag: "mlp",
         description: "Bargam",
         added: "2009-07-29"}).
iana(language,mlq,tag{
         type: "language",
         subtag: "mlq",
         description: "Western Maninkakan",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,mlr,tag{
         type: "language",
         subtag: "mlr",
         description: "Vame",
         added: "2009-07-29"}).
iana(language,mls,tag{
         type: "language",
         subtag: "mls",
         description: "Masalit",
         added: "2009-07-29"}).
iana(language,mlu,tag{
         type: "language",
         subtag: "mlu",
         description: "To'abaita",
         added: "2009-07-29"}).
iana(language,mlv,tag{
         type: "language",
         subtag: "mlv",
         description: ["Motlav","Mwotlap"],
         added: "2009-07-29"}).
iana(language,mlw,tag{
         type: "language",
         subtag: "mlw",
         description: "Moloko",
         added: "2009-07-29"}).
iana(language,mlx,tag{
         type: "language",
         subtag: "mlx",
         description: ["Malfaxal","Naha'ai"],
         added: "2009-07-29"}).
iana(language,mlz,tag{
         type: "language",
         subtag: "mlz",
         description: "Malaynon",
         added: "2009-07-29"}).
iana(language,mma,tag{
         type: "language",
         subtag: "mma",
         description: "Mama",
         added: "2009-07-29"}).
iana(language,mmb,tag{
         type: "language",
         subtag: "mmb",
         description: "Momina",
         added: "2009-07-29"}).
iana(language,mmc,tag{
         type: "language",
         subtag: "mmc",
         description: "Michoacán Mazahua",
         added: "2009-07-29"}).
iana(language,mmd,tag{
         type: "language",
         subtag: "mmd",
         description: "Maonan",
         added: "2009-07-29"}).
iana(language,mme,tag{
         type: "language",
         subtag: "mme",
         description: "Mae",
         added: "2009-07-29"}).
iana(language,mmf,tag{
         type: "language",
         subtag: "mmf",
         description: "Mundat",
         added: "2009-07-29"}).
iana(language,mmg,tag{
         type: "language",
         subtag: "mmg",
         description: "North Ambrym",
         added: "2009-07-29"}).
iana(language,mmh,tag{
         type: "language",
         subtag: "mmh",
         description: "Mehináku",
         added: "2009-07-29"}).
iana(language,mmi,tag{
         type: "language",
         subtag: "mmi",
         description: "Musar",
         added: "2009-07-29"}).
iana(language,mmj,tag{
         type: "language",
         subtag: "mmj",
         description: "Majhwar",
         added: "2009-07-29"}).
iana(language,mmk,tag{
         type: "language",
         subtag: "mmk",
         description: "Mukha-Dora",
         added: "2009-07-29"}).
iana(language,mml,tag{
         type: "language",
         subtag: "mml",
         description: "Man Met",
         added: "2009-07-29"}).
iana(language,mmm,tag{
         type: "language",
         subtag: "mmm",
         description: "Maii",
         added: "2009-07-29"}).
iana(language,mmn,tag{
         type: "language",
         subtag: "mmn",
         description: "Mamanwa",
         added: "2009-07-29"}).
iana(language,mmo,tag{
         type: "language",
         subtag: "mmo",
         description: "Mangga Buang",
         added: "2009-07-29"}).
iana(language,mmp,tag{
         type: "language",
         subtag: "mmp",
         description: "Siawi",
         added: "2009-07-29"}).
iana(language,mmq,tag{
         type: "language",
         subtag: "mmq",
         description: "Musak",
         added: "2009-07-29"}).
iana(language,mmr,tag{
         type: "language",
         subtag: "mmr",
         description: "Western Xiangxi Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,mmt,tag{
         type: "language",
         subtag: "mmt",
         description: "Malalamai",
         added: "2009-07-29"}).
iana(language,mmu,tag{
         type: "language",
         subtag: "mmu",
         description: "Mmaala",
         added: "2009-07-29"}).
iana(language,mmv,tag{
         type: "language",
         subtag: "mmv",
         description: "Miriti",
         added: "2009-07-29"}).
iana(language,mmw,tag{
         type: "language",
         subtag: "mmw",
         description: "Emae",
         added: "2009-07-29"}).
iana(language,mmx,tag{
         type: "language",
         subtag: "mmx",
         description: "Madak",
         added: "2009-07-29"}).
iana(language,mmy,tag{
         type: "language",
         subtag: "mmy",
         description: "Migaama",
         added: "2009-07-29"}).
iana(language,mmz,tag{
         type: "language",
         subtag: "mmz",
         description: "Mabaale",
         added: "2009-07-29"}).
iana(language,mna,tag{
         type: "language",
         subtag: "mna",
         description: "Mbula",
         added: "2009-07-29"}).
iana(language,mnb,tag{
         type: "language",
         subtag: "mnb",
         description: "Muna",
         added: "2009-07-29"}).
iana(language,mnc,tag{
         type: "language",
         subtag: "mnc",
         description: "Manchu",
         added: "2005-10-16"}).
iana(language,mnd,tag{
         type: "language",
         subtag: "mnd",
         description: "Mondé",
         added: "2009-07-29"}).
iana(language,mne,tag{
         type: "language",
         subtag: "mne",
         description: "Naba",
         added: "2009-07-29"}).
iana(language,mnf,tag{
         type: "language",
         subtag: "mnf",
         description: "Mundani",
         added: "2009-07-29"}).
iana(language,mng,tag{
         type: "language",
         subtag: "mng",
         description: "Eastern Mnong",
         added: "2009-07-29"}).
iana(language,mnh,tag{
         type: "language",
         subtag: "mnh",
         description: "Mono (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,mni,tag{
         type: "language",
         subtag: "mni",
         description: "Manipuri",
         added: "2005-10-16"}).
iana(language,mnj,tag{
         type: "language",
         subtag: "mnj",
         description: "Munji",
         added: "2009-07-29"}).
iana(language,mnk,tag{
         type: "language",
         subtag: "mnk",
         description: "Mandinka",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,mnl,tag{
         type: "language",
         subtag: "mnl",
         description: "Tiale",
         added: "2009-07-29"}).
iana(language,mnm,tag{
         type: "language",
         subtag: "mnm",
         description: "Mapena",
         added: "2009-07-29"}).
iana(language,mnn,tag{
         type: "language",
         subtag: "mnn",
         description: "Southern Mnong",
         added: "2009-07-29"}).
iana(language,mno,tag{
         type: "language",
         subtag: "mno",
         description: "Manobo languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,mnp,tag{
         type: "language",
         subtag: "mnp",
         description: "Min Bei Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,mnq,tag{
         type: "language",
         subtag: "mnq",
         description: "Minriq",
         added: "2009-07-29"}).
iana(language,mnr,tag{
         type: "language",
         subtag: "mnr",
         description: "Mono (USA)",
         added: "2009-07-29"}).
iana(language,mns,tag{
         type: "language",
         subtag: "mns",
         description: "Mansi",
         added: "2009-07-29"}).
iana(language,mnt,tag{
         type: "language",
         subtag: "mnt",
         description: "Maykulan",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see wnn, xyj, xyk, xyt"}).
iana(language,mnu,tag{
         type: "language",
         subtag: "mnu",
         description: "Mer",
         added: "2009-07-29"}).
iana(language,mnv,tag{
         type: "language",
         subtag: "mnv",
         description: "Rennell-Bellona",
         added: "2009-07-29"}).
iana(language,mnw,tag{
         type: "language",
         subtag: "mnw",
         description: "Mon",
         added: "2009-07-29"}).
iana(language,mnx,tag{
         type: "language",
         subtag: "mnx",
         description: "Manikion",
         added: "2009-07-29"}).
iana(language,mny,tag{
         type: "language",
         subtag: "mny",
         description: "Manyawa",
         added: "2009-07-29"}).
iana(language,mnz,tag{
         type: "language",
         subtag: "mnz",
         description: "Moni",
         added: "2009-07-29"}).
iana(language,moa,tag{
         type: "language",
         subtag: "moa",
         description: "Mwan",
         added: "2009-07-29"}).
iana(language,moc,tag{
         type: "language",
         subtag: "moc",
         description: "Mocoví",
         added: "2009-07-29"}).
iana(language,mod,tag{
         type: "language",
         subtag: "mod",
         description: "Mobilian",
         added: "2009-07-29"}).
iana(language,moe,tag{
         type: "language",
         subtag: "moe",
         description: ["Innu","Montagnais"],
         added: "2009-07-29"}).
iana(language,mof,tag{
         type: "language",
         subtag: "mof",
         description: "Mohegan-Montauk-Narragansett",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see xnt, xpq"}).
iana(language,mog,tag{
         type: "language",
         subtag: "mog",
         description: "Mongondow",
         added: "2009-07-29"}).
iana(language,moh,tag{
         type: "language",
         subtag: "moh",
         description: "Mohawk",
         added: "2005-10-16"}).
iana(language,moi,tag{
         type: "language",
         subtag: "moi",
         description: "Mboi",
         added: "2009-07-29"}).
iana(language,moj,tag{
         type: "language",
         subtag: "moj",
         description: "Monzombo",
         added: "2009-07-29"}).
iana(language,mok,tag{
         type: "language",
         subtag: "mok",
         description: "Morori",
         added: "2009-07-29"}).
iana(language,mom,tag{
         type: "language",
         subtag: "mom",
         description: "Mangue",
         added: "2009-07-29"}).
iana(language,moo,tag{
         type: "language",
         subtag: "moo",
         description: "Monom",
         added: "2009-07-29"}).
iana(language,mop,tag{
         type: "language",
         subtag: "mop",
         description: "Mopán Maya",
         added: "2009-07-29"}).
iana(language,moq,tag{
         type: "language",
         subtag: "moq",
         description: "Mor (Bomberai Peninsula)",
         added: "2009-07-29"}).
iana(language,mor,tag{
         type: "language",
         subtag: "mor",
         description: "Moro",
         added: "2009-07-29"}).
iana(language,mos,tag{
         type: "language",
         subtag: "mos",
         description: "Mossi",
         added: "2005-10-16"}).
iana(language,mot,tag{
         type: "language",
         subtag: "mot",
         description: "Barí",
         added: "2009-07-29"}).
iana(language,mou,tag{
         type: "language",
         subtag: "mou",
         description: "Mogum",
         added: "2009-07-29"}).
iana(language,mov,tag{
         type: "language",
         subtag: "mov",
         description: "Mohave",
         added: "2009-07-29"}).
iana(language,mow,tag{
         type: "language",
         subtag: "mow",
         description: "Moi (Congo)",
         added: "2009-07-29"}).
iana(language,mox,tag{
         type: "language",
         subtag: "mox",
         description: "Molima",
         added: "2009-07-29"}).
iana(language,moy,tag{
         type: "language",
         subtag: "moy",
         description: "Shekkacho",
         added: "2009-07-29"}).
iana(language,moz,tag{
         type: "language",
         subtag: "moz",
         description: ["Mukulu","Gergiko"],
         added: "2009-07-29"}).
iana(language,mpa,tag{
         type: "language",
         subtag: "mpa",
         description: "Mpoto",
         added: "2009-07-29"}).
iana(language,mpb,tag{
         type: "language",
         subtag: "mpb",
         description: ["Malak Malak","Mullukmulluk"],
         added: "2009-07-29"}).
iana(language,mpc,tag{
         type: "language",
         subtag: "mpc",
         description: "Mangarrayi",
         added: "2009-07-29"}).
iana(language,mpd,tag{
         type: "language",
         subtag: "mpd",
         description: "Machinere",
         added: "2009-07-29"}).
iana(language,mpe,tag{
         type: "language",
         subtag: "mpe",
         description: "Majang",
         added: "2009-07-29"}).
iana(language,mpg,tag{
         type: "language",
         subtag: "mpg",
         description: "Marba",
         added: "2009-07-29"}).
iana(language,mph,tag{
         type: "language",
         subtag: "mph",
         description: "Maung",
         added: "2009-07-29"}).
iana(language,mpi,tag{
         type: "language",
         subtag: "mpi",
         description: "Mpade",
         added: "2009-07-29"}).
iana(language,mpj,tag{
         type: "language",
         subtag: "mpj",
         description: ["Martu Wangka","Wangkajunga"],
         added: "2009-07-29"}).
iana(language,mpk,tag{
         type: "language",
         subtag: "mpk",
         description: "Mbara (Chad)",
         added: "2009-07-29"}).
iana(language,mpl,tag{
         type: "language",
         subtag: "mpl",
         description: "Middle Watut",
         added: "2009-07-29"}).
iana(language,mpm,tag{
         type: "language",
         subtag: "mpm",
         description: "Yosondúa Mixtec",
         added: "2009-07-29"}).
iana(language,mpn,tag{
         type: "language",
         subtag: "mpn",
         description: "Mindiri",
         added: "2009-07-29"}).
iana(language,mpo,tag{
         type: "language",
         subtag: "mpo",
         description: "Miu",
         added: "2009-07-29"}).
iana(language,mpp,tag{
         type: "language",
         subtag: "mpp",
         description: "Migabac",
         added: "2009-07-29"}).
iana(language,mpq,tag{
         type: "language",
         subtag: "mpq",
         description: "Matís",
         added: "2009-07-29"}).
iana(language,mpr,tag{
         type: "language",
         subtag: "mpr",
         description: "Vangunu",
         added: "2009-07-29"}).
iana(language,mps,tag{
         type: "language",
         subtag: "mps",
         description: "Dadibi",
         added: "2009-07-29"}).
iana(language,mpt,tag{
         type: "language",
         subtag: "mpt",
         description: "Mian",
         added: "2009-07-29"}).
iana(language,mpu,tag{
         type: "language",
         subtag: "mpu",
         description: "Makuráp",
         added: "2009-07-29"}).
iana(language,mpv,tag{
         type: "language",
         subtag: "mpv",
         description: "Mungkip",
         added: "2009-07-29"}).
iana(language,mpw,tag{
         type: "language",
         subtag: "mpw",
         description: "Mapidian",
         added: "2009-07-29"}).
iana(language,mpx,tag{
         type: "language",
         subtag: "mpx",
         description: "Misima-Panaeati",
         added: "2009-07-29"}).
iana(language,mpy,tag{
         type: "language",
         subtag: "mpy",
         description: "Mapia",
         added: "2009-07-29"}).
iana(language,mpz,tag{
         type: "language",
         subtag: "mpz",
         description: "Mpi",
         added: "2009-07-29"}).
iana(language,mqa,tag{
         type: "language",
         subtag: "mqa",
         description: "Maba (Indonesia)",
         added: "2009-07-29"}).
iana(language,mqb,tag{
         type: "language",
         subtag: "mqb",
         description: "Mbuko",
         added: "2009-07-29"}).
iana(language,mqc,tag{
         type: "language",
         subtag: "mqc",
         description: "Mangole",
         added: "2009-07-29"}).
iana(language,mqe,tag{
         type: "language",
         subtag: "mqe",
         description: "Matepi",
         added: "2009-07-29"}).
iana(language,mqf,tag{
         type: "language",
         subtag: "mqf",
         description: "Momuna",
         added: "2009-07-29"}).
iana(language,mqg,tag{
         type: "language",
         subtag: "mqg",
         description: "Kota Bangun Kutai Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,mqh,tag{
         type: "language",
         subtag: "mqh",
         description: "Tlazoyaltepec Mixtec",
         added: "2009-07-29"}).
iana(language,mqi,tag{
         type: "language",
         subtag: "mqi",
         description: "Mariri",
         added: "2009-07-29"}).
iana(language,mqj,tag{
         type: "language",
         subtag: "mqj",
         description: "Mamasa",
         added: "2009-07-29"}).
iana(language,mqk,tag{
         type: "language",
         subtag: "mqk",
         description: "Rajah Kabunsuwan Manobo",
         added: "2009-07-29"}).
iana(language,mql,tag{
         type: "language",
         subtag: "mql",
         description: "Mbelime",
         added: "2009-07-29"}).
iana(language,mqm,tag{
         type: "language",
         subtag: "mqm",
         description: "South Marquesan",
         added: "2009-07-29"}).
iana(language,mqn,tag{
         type: "language",
         subtag: "mqn",
         description: "Moronene",
         added: "2009-07-29"}).
iana(language,mqo,tag{
         type: "language",
         subtag: "mqo",
         description: "Modole",
         added: "2009-07-29"}).
iana(language,mqp,tag{
         type: "language",
         subtag: "mqp",
         description: "Manipa",
         added: "2009-07-29"}).
iana(language,mqq,tag{
         type: "language",
         subtag: "mqq",
         description: "Minokok",
         added: "2009-07-29"}).
iana(language,mqr,tag{
         type: "language",
         subtag: "mqr",
         description: "Mander",
         added: "2009-07-29"}).
iana(language,mqs,tag{
         type: "language",
         subtag: "mqs",
         description: "West Makian",
         added: "2009-07-29"}).
iana(language,mqt,tag{
         type: "language",
         subtag: "mqt",
         description: "Mok",
         added: "2009-07-29"}).
iana(language,mqu,tag{
         type: "language",
         subtag: "mqu",
         description: "Mandari",
         added: "2009-07-29"}).
iana(language,mqv,tag{
         type: "language",
         subtag: "mqv",
         description: "Mosimo",
         added: "2009-07-29"}).
iana(language,mqw,tag{
         type: "language",
         subtag: "mqw",
         description: "Murupi",
         added: "2009-07-29"}).
iana(language,mqx,tag{
         type: "language",
         subtag: "mqx",
         description: "Mamuju",
         added: "2009-07-29"}).
iana(language,mqy,tag{
         type: "language",
         subtag: "mqy",
         description: "Manggarai",
         added: "2009-07-29"}).
iana(language,mqz,tag{
         type: "language",
         subtag: "mqz",
         description: "Pano",
         added: "2009-07-29"}).
iana(language,mra,tag{
         type: "language",
         subtag: "mra",
         description: "Mlabri",
         added: "2009-07-29"}).
iana(language,mrb,tag{
         type: "language",
         subtag: "mrb",
         description: "Marino",
         added: "2009-07-29"}).
iana(language,mrc,tag{
         type: "language",
         subtag: "mrc",
         description: "Maricopa",
         added: "2009-07-29"}).
iana(language,mrd,tag{
         type: "language",
         subtag: "mrd",
         description: "Western Magar",
         added: "2009-07-29"}).
iana(language,mre,tag{
         type: "language",
         subtag: "mre",
         description: "Martha's Vineyard Sign Language",
         added: "2009-07-29"}).
iana(language,mrf,tag{
         type: "language",
         subtag: "mrf",
         description: "Elseng",
         added: "2009-07-29"}).
iana(language,mrg,tag{
         type: "language",
         subtag: "mrg",
         description: "Mising",
         added: "2009-07-29"}).
iana(language,mrh,tag{
         type: "language",
         subtag: "mrh",
         description: "Mara Chin",
         added: "2009-07-29"}).
iana(language,mrj,tag{
         type: "language",
         subtag: "mrj",
         description: "Western Mari",
         added: "2009-07-29",
         macrolanguage: "chm"}).
iana(language,mrk,tag{
         type: "language",
         subtag: "mrk",
         description: "Hmwaveke",
         added: "2009-07-29"}).
iana(language,mrl,tag{
         type: "language",
         subtag: "mrl",
         description: "Mortlockese",
         added: "2009-07-29"}).
iana(language,mrm,tag{
         type: "language",
         subtag: "mrm",
         description: ["Merlav","Mwerlap"],
         added: "2009-07-29"}).
iana(language,mrn,tag{
         type: "language",
         subtag: "mrn",
         description: "Cheke Holo",
         added: "2009-07-29"}).
iana(language,mro,tag{
         type: "language",
         subtag: "mro",
         description: "Mru",
         added: "2009-07-29"}).
iana(language,mrp,tag{
         type: "language",
         subtag: "mrp",
         description: "Morouas",
         added: "2009-07-29"}).
iana(language,mrq,tag{
         type: "language",
         subtag: "mrq",
         description: "North Marquesan",
         added: "2009-07-29"}).
iana(language,mrr,tag{
         type: "language",
         subtag: "mrr",
         description: "Maria (India)",
         added: "2009-07-29"}).
iana(language,mrs,tag{
         type: "language",
         subtag: "mrs",
         description: "Maragus",
         added: "2009-07-29"}).
iana(language,mrt,tag{
         type: "language",
         subtag: "mrt",
         description: "Marghi Central",
         added: "2009-07-29"}).
iana(language,mru,tag{
         type: "language",
         subtag: "mru",
         description: "Mono (Cameroon)",
         added: "2009-07-29"}).
iana(language,mrv,tag{
         type: "language",
         subtag: "mrv",
         description: "Mangareva",
         added: "2009-07-29"}).
iana(language,mrw,tag{
         type: "language",
         subtag: "mrw",
         description: "Maranao",
         added: "2009-07-29"}).
iana(language,mrx,tag{
         type: "language",
         subtag: "mrx",
         description: ["Maremgi","Dineor"],
         added: "2009-07-29"}).
iana(language,mry,tag{
         type: "language",
         subtag: "mry",
         description: "Mandaya",
         added: "2009-07-29"}).
iana(language,mrz,tag{
         type: "language",
         subtag: "mrz",
         description: "Marind",
         added: "2009-07-29"}).
iana(language,msb,tag{
         type: "language",
         subtag: "msb",
         description: "Masbatenyo",
         added: "2009-07-29"}).
iana(language,msc,tag{
         type: "language",
         subtag: "msc",
         description: "Sankaran Maninka",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,msd,tag{
         type: "language",
         subtag: "msd",
         description: "Yucatec Maya Sign Language",
         added: "2009-07-29"}).
iana(language,mse,tag{
         type: "language",
         subtag: "mse",
         description: "Musey",
         added: "2009-07-29"}).
iana(language,msf,tag{
         type: "language",
         subtag: "msf",
         description: "Mekwei",
         added: "2009-07-29"}).
iana(language,msg,tag{
         type: "language",
         subtag: "msg",
         description: "Moraid",
         added: "2009-07-29"}).
iana(language,msh,tag{
         type: "language",
         subtag: "msh",
         description: "Masikoro Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,msi,tag{
         type: "language",
         subtag: "msi",
         description: "Sabah Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,msj,tag{
         type: "language",
         subtag: "msj",
         description: "Ma (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,msk,tag{
         type: "language",
         subtag: "msk",
         description: "Mansaka",
         added: "2009-07-29"}).
iana(language,msl,tag{
         type: "language",
         subtag: "msl",
         description: ["Molof","Poule"],
         added: "2009-07-29"}).
iana(language,msm,tag{
         type: "language",
         subtag: "msm",
         description: "Agusan Manobo",
         added: "2009-07-29"}).
iana(language,msn,tag{
         type: "language",
         subtag: "msn",
         description: "Vurës",
         added: "2009-07-29"}).
iana(language,mso,tag{
         type: "language",
         subtag: "mso",
         description: "Mombum",
         added: "2009-07-29"}).
iana(language,msp,tag{
         type: "language",
         subtag: "msp",
         description: "Maritsauá",
         added: "2009-07-29"}).
iana(language,msq,tag{
         type: "language",
         subtag: "msq",
         description: "Caac",
         added: "2009-07-29"}).
iana(language,msr,tag{
         type: "language",
         subtag: "msr",
         description: "Mongolian Sign Language",
         added: "2009-07-29"}).
iana(language,mss,tag{
         type: "language",
         subtag: "mss",
         description: "West Masela",
         added: "2009-07-29"}).
iana(language,mst,tag{
         type: "language",
         subtag: "mst",
         description: "Cataelano Mandaya",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "mry"}).
iana(language,msu,tag{
         type: "language",
         subtag: "msu",
         description: "Musom",
         added: "2009-07-29"}).
iana(language,msv,tag{
         type: "language",
         subtag: "msv",
         description: "Maslam",
         added: "2009-07-29"}).
iana(language,msw,tag{
         type: "language",
         subtag: "msw",
         description: "Mansoanka",
         added: "2009-07-29"}).
iana(language,msx,tag{
         type: "language",
         subtag: "msx",
         description: "Moresada",
         added: "2009-07-29"}).
iana(language,msy,tag{
         type: "language",
         subtag: "msy",
         description: "Aruamu",
         added: "2009-07-29"}).
iana(language,msz,tag{
         type: "language",
         subtag: "msz",
         description: "Momare",
         added: "2009-07-29"}).
iana(language,mta,tag{
         type: "language",
         subtag: "mta",
         description: "Cotabato Manobo",
         added: "2009-07-29"}).
iana(language,mtb,tag{
         type: "language",
         subtag: "mtb",
         description: "Anyin Morofo",
         added: "2009-07-29"}).
iana(language,mtc,tag{
         type: "language",
         subtag: "mtc",
         description: "Munit",
         added: "2009-07-29"}).
iana(language,mtd,tag{
         type: "language",
         subtag: "mtd",
         description: "Mualang",
         added: "2009-07-29"}).
iana(language,mte,tag{
         type: "language",
         subtag: "mte",
         description: "Mono (Solomon Islands)",
         added: "2009-07-29"}).
iana(language,mtf,tag{
         type: "language",
         subtag: "mtf",
         description: "Murik (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,mtg,tag{
         type: "language",
         subtag: "mtg",
         description: "Una",
         added: "2009-07-29"}).
iana(language,mth,tag{
         type: "language",
         subtag: "mth",
         description: "Munggui",
         added: "2009-07-29"}).
iana(language,mti,tag{
         type: "language",
         subtag: "mti",
         description: "Maiwa (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,mtj,tag{
         type: "language",
         subtag: "mtj",
         description: "Moskona",
         added: "2009-07-29"}).
iana(language,mtk,tag{
         type: "language",
         subtag: "mtk",
         description: "Mbe'",
         added: "2009-07-29"}).
iana(language,mtl,tag{
         type: "language",
         subtag: "mtl",
         description: "Montol",
         added: "2009-07-29"}).
iana(language,mtm,tag{
         type: "language",
         subtag: "mtm",
         description: "Mator",
         added: "2009-07-29"}).
iana(language,mtn,tag{
         type: "language",
         subtag: "mtn",
         description: "Matagalpa",
         added: "2009-07-29"}).
iana(language,mto,tag{
         type: "language",
         subtag: "mto",
         description: "Totontepec Mixe",
         added: "2009-07-29"}).
iana(language,mtp,tag{
         type: "language",
         subtag: "mtp",
         description: "Wichí Lhamtés Nocten",
         added: "2009-07-29"}).
iana(language,mtq,tag{
         type: "language",
         subtag: "mtq",
         description: "Muong",
         added: "2009-07-29"}).
iana(language,mtr,tag{
         type: "language",
         subtag: "mtr",
         description: "Mewari",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,mts,tag{
         type: "language",
         subtag: "mts",
         description: "Yora",
         added: "2009-07-29"}).
iana(language,mtt,tag{
         type: "language",
         subtag: "mtt",
         description: "Mota",
         added: "2009-07-29"}).
iana(language,mtu,tag{
         type: "language",
         subtag: "mtu",
         description: "Tututepec Mixtec",
         added: "2009-07-29"}).
iana(language,mtv,tag{
         type: "language",
         subtag: "mtv",
         description: "Asaro'o",
         added: "2009-07-29"}).
iana(language,mtw,tag{
         type: "language",
         subtag: "mtw",
         description: "Southern Binukidnon",
         added: "2009-07-29"}).
iana(language,mtx,tag{
         type: "language",
         subtag: "mtx",
         description: "Tidaá Mixtec",
         added: "2009-07-29"}).
iana(language,mty,tag{
         type: "language",
         subtag: "mty",
         description: "Nabi",
         added: "2009-07-29"}).
iana(language,mua,tag{
         type: "language",
         subtag: "mua",
         description: "Mundang",
         added: "2009-07-29"}).
iana(language,mub,tag{
         type: "language",
         subtag: "mub",
         description: "Mubi",
         added: "2009-07-29"}).
iana(language,muc,tag{
         type: "language",
         subtag: "muc",
         description: "Ajumbu",
         added: "2009-07-29"}).
iana(language,mud,tag{
         type: "language",
         subtag: "mud",
         description: "Mednyj Aleut",
         added: "2009-07-29"}).
iana(language,mue,tag{
         type: "language",
         subtag: "mue",
         description: "Media Lengua",
         added: "2009-07-29"}).
iana(language,mug,tag{
         type: "language",
         subtag: "mug",
         description: "Musgu",
         added: "2009-07-29"}).
iana(language,muh,tag{
         type: "language",
         subtag: "muh",
         description: "Mündü",
         added: "2009-07-29"}).
iana(language,mui,tag{
         type: "language",
         subtag: "mui",
         description: "Musi",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,muj,tag{
         type: "language",
         subtag: "muj",
         description: "Mabire",
         added: "2009-07-29"}).
iana(language,muk,tag{
         type: "language",
         subtag: "muk",
         description: "Mugom",
         added: "2009-07-29"}).
iana(language,mul,tag{
         type: "language",
         subtag: "mul",
         description: "Multiple languages",
         added: "2005-10-16",
         scope: "special"}).
iana(language,mum,tag{
         type: "language",
         subtag: "mum",
         description: "Maiwala",
         added: "2009-07-29"}).
iana(language,mun,tag{
         type: "language",
         subtag: "mun",
         description: "Munda languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,muo,tag{
         type: "language",
         subtag: "muo",
         description: "Nyong",
         added: "2009-07-29"}).
iana(language,mup,tag{
         type: "language",
         subtag: "mup",
         description: "Malvi",
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,muq,tag{
         type: "language",
         subtag: "muq",
         description: "Eastern Xiangxi Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,mur,tag{
         type: "language",
         subtag: "mur",
         description: "Murle",
         added: "2009-07-29"}).
iana(language,mus,tag{
         type: "language",
         subtag: "mus",
         description: "Creek",
         added: "2005-10-16"}).
iana(language,mut,tag{
         type: "language",
         subtag: "mut",
         description: "Western Muria",
         added: "2009-07-29"}).
iana(language,muu,tag{
         type: "language",
         subtag: "muu",
         description: "Yaaku",
         added: "2009-07-29"}).
iana(language,muv,tag{
         type: "language",
         subtag: "muv",
         description: "Muthuvan",
         added: "2009-07-29"}).
iana(language,mux,tag{
         type: "language",
         subtag: "mux",
         description: "Bo-Ung",
         added: "2009-07-29"}).
iana(language,muy,tag{
         type: "language",
         subtag: "muy",
         description: "Muyang",
         added: "2009-07-29"}).
iana(language,muz,tag{
         type: "language",
         subtag: "muz",
         description: "Mursi",
         added: "2009-07-29"}).
iana(language,mva,tag{
         type: "language",
         subtag: "mva",
         description: "Manam",
         added: "2009-07-29"}).
iana(language,mvb,tag{
         type: "language",
         subtag: "mvb",
         description: "Mattole",
         added: "2009-07-29"}).
iana(language,mvd,tag{
         type: "language",
         subtag: "mvd",
         description: "Mamboru",
         added: "2009-07-29"}).
iana(language,mve,tag{
         type: "language",
         subtag: "mve",
         description: "Marwari (Pakistan)",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,mvf,tag{
         type: "language",
         subtag: "mvf",
         description: "Peripheral Mongolian",
         added: "2009-07-29",
         macrolanguage: "mn"}).
iana(language,mvg,tag{
         type: "language",
         subtag: "mvg",
         description: "Yucuañe Mixtec",
         added: "2009-07-29"}).
iana(language,mvh,tag{
         type: "language",
         subtag: "mvh",
         description: "Mulgi",
         added: "2009-07-29"}).
iana(language,mvi,tag{
         type: "language",
         subtag: "mvi",
         description: "Miyako",
         added: "2009-07-29"}).
iana(language,mvk,tag{
         type: "language",
         subtag: "mvk",
         description: "Mekmek",
         added: "2009-07-29"}).
iana(language,mvl,tag{
         type: "language",
         subtag: "mvl",
         description: "Mbara (Australia)",
         added: "2009-07-29"}).
iana(language,mvm,tag{
         type: "language",
         subtag: "mvm",
         description: "Muya",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see emq, wmg"}).
iana(language,mvn,tag{
         type: "language",
         subtag: "mvn",
         description: "Minaveha",
         added: "2009-07-29"}).
iana(language,mvo,tag{
         type: "language",
         subtag: "mvo",
         description: "Marovo",
         added: "2009-07-29"}).
iana(language,mvp,tag{
         type: "language",
         subtag: "mvp",
         description: "Duri",
         added: "2009-07-29"}).
iana(language,mvq,tag{
         type: "language",
         subtag: "mvq",
         description: "Moere",
         added: "2009-07-29"}).
iana(language,mvr,tag{
         type: "language",
         subtag: "mvr",
         description: "Marau",
         added: "2009-07-29"}).
iana(language,mvs,tag{
         type: "language",
         subtag: "mvs",
         description: "Massep",
         added: "2009-07-29"}).
iana(language,mvt,tag{
         type: "language",
         subtag: "mvt",
         description: "Mpotovoro",
         added: "2009-07-29"}).
iana(language,mvu,tag{
         type: "language",
         subtag: "mvu",
         description: "Marfa",
         added: "2009-07-29"}).
iana(language,mvv,tag{
         type: "language",
         subtag: "mvv",
         description: "Tagal Murut",
         added: "2009-07-29"}).
iana(language,mvw,tag{
         type: "language",
         subtag: "mvw",
         description: "Machinga",
         added: "2009-07-29"}).
iana(language,mvx,tag{
         type: "language",
         subtag: "mvx",
         description: "Meoswar",
         added: "2009-07-29"}).
iana(language,mvy,tag{
         type: "language",
         subtag: "mvy",
         description: "Indus Kohistani",
         added: "2009-07-29"}).
iana(language,mvz,tag{
         type: "language",
         subtag: "mvz",
         description: "Mesqan",
         added: "2009-07-29"}).
iana(language,mwa,tag{
         type: "language",
         subtag: "mwa",
         description: "Mwatebu",
         added: "2009-07-29"}).
iana(language,mwb,tag{
         type: "language",
         subtag: "mwb",
         description: "Juwal",
         added: "2009-07-29"}).
iana(language,mwc,tag{
         type: "language",
         subtag: "mwc",
         description: "Are",
         added: "2009-07-29"}).
iana(language,mwd,tag{
         type: "language",
         subtag: "mwd",
         description: "Mudbura",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see dmw, xrq"}).
iana(language,mwe,tag{
         type: "language",
         subtag: "mwe",
         description: "Mwera (Chimwera)",
         added: "2009-07-29"}).
iana(language,mwf,tag{
         type: "language",
         subtag: "mwf",
         description: "Murrinh-Patha",
         added: "2009-07-29"}).
iana(language,mwg,tag{
         type: "language",
         subtag: "mwg",
         description: "Aiklep",
         added: "2009-07-29"}).
iana(language,mwh,tag{
         type: "language",
         subtag: "mwh",
         description: "Mouk-Aria",
         added: "2009-07-29"}).
iana(language,mwi,tag{
         type: "language",
         subtag: "mwi",
         description: ["Labo","Ninde"],
         added: "2009-07-29"}).
iana(language,mwj,tag{
         type: "language",
         subtag: "mwj",
         description: "Maligo",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "vaj"}).
iana(language,mwk,tag{
         type: "language",
         subtag: "mwk",
         description: "Kita Maninkakan",
         added: "2009-07-29",
         macrolanguage: "man"}).
iana(language,mwl,tag{
         type: "language",
         subtag: "mwl",
         description: "Mirandese",
         added: "2005-10-16"}).
iana(language,mwm,tag{
         type: "language",
         subtag: "mwm",
         description: "Sar",
         added: "2009-07-29"}).
iana(language,mwn,tag{
         type: "language",
         subtag: "mwn",
         description: "Nyamwanga",
         added: "2009-07-29"}).
iana(language,mwo,tag{
         type: "language",
         subtag: "mwo",
         description: "Central Maewo",
         added: "2009-07-29"}).
iana(language,mwp,tag{
         type: "language",
         subtag: "mwp",
         description: "Kala Lagaw Ya",
         added: "2009-07-29"}).
iana(language,mwq,tag{
         type: "language",
         subtag: "mwq",
         description: "Mün Chin",
         added: "2009-07-29"}).
iana(language,mwr,tag{
         type: "language",
         subtag: "mwr",
         description: "Marwari",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,mws,tag{
         type: "language",
         subtag: "mws",
         description: "Mwimbi-Muthambi",
         added: "2009-07-29"}).
iana(language,mwt,tag{
         type: "language",
         subtag: "mwt",
         description: "Moken",
         added: "2009-07-29"}).
iana(language,mwu,tag{
         type: "language",
         subtag: "mwu",
         description: "Mittu",
         added: "2009-07-29"}).
iana(language,mwv,tag{
         type: "language",
         subtag: "mwv",
         description: "Mentawai",
         added: "2009-07-29"}).
iana(language,mww,tag{
         type: "language",
         subtag: "mww",
         description: "Hmong Daw",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,mwx,tag{
         type: "language",
         subtag: "mwx",
         description: "Mediak",
         added: "2009-07-29",
         deprecated: "2018-03-08"}).
iana(language,mwy,tag{
         type: "language",
         subtag: "mwy",
         description: "Mosiro",
         added: "2009-07-29",
         deprecated: "2018-03-08"}).
iana(language,mwz,tag{
         type: "language",
         subtag: "mwz",
         description: "Moingi",
         added: "2009-07-29"}).
iana(language,mxa,tag{
         type: "language",
         subtag: "mxa",
         description: "Northwest Oaxaca Mixtec",
         added: "2009-07-29"}).
iana(language,mxb,tag{
         type: "language",
         subtag: "mxb",
         description: "Tezoatlán Mixtec",
         added: "2009-07-29"}).
iana(language,mxc,tag{
         type: "language",
         subtag: "mxc",
         description: "Manyika",
         added: "2009-07-29"}).
iana(language,mxd,tag{
         type: "language",
         subtag: "mxd",
         description: "Modang",
         added: "2009-07-29"}).
iana(language,mxe,tag{
         type: "language",
         subtag: "mxe",
         description: "Mele-Fila",
         added: "2009-07-29"}).
iana(language,mxf,tag{
         type: "language",
         subtag: "mxf",
         description: "Malgbe",
         added: "2009-07-29"}).
iana(language,mxg,tag{
         type: "language",
         subtag: "mxg",
         description: "Mbangala",
         added: "2009-07-29"}).
iana(language,mxh,tag{
         type: "language",
         subtag: "mxh",
         description: "Mvuba",
         added: "2009-07-29"}).
iana(language,mxi,tag{
         type: "language",
         subtag: "mxi",
         description: "Mozarabic",
         added: "2009-07-29"}).
iana(language,mxj,tag{
         type: "language",
         subtag: "mxj",
         description: ["Miju-Mishmi","Geman Deng"],
         added: "2009-07-29"}).
iana(language,mxk,tag{
         type: "language",
         subtag: "mxk",
         description: "Monumbo",
         added: "2009-07-29"}).
iana(language,mxl,tag{
         type: "language",
         subtag: "mxl",
         description: "Maxi Gbe",
         added: "2009-07-29"}).
iana(language,mxm,tag{
         type: "language",
         subtag: "mxm",
         description: "Meramera",
         added: "2009-07-29"}).
iana(language,mxn,tag{
         type: "language",
         subtag: "mxn",
         description: "Moi (Indonesia)",
         added: "2009-07-29"}).
iana(language,mxo,tag{
         type: "language",
         subtag: "mxo",
         description: "Mbowe",
         added: "2009-07-29"}).
iana(language,mxp,tag{
         type: "language",
         subtag: "mxp",
         description: "Tlahuitoltepec Mixe",
         added: "2009-07-29"}).
iana(language,mxq,tag{
         type: "language",
         subtag: "mxq",
         description: "Juquila Mixe",
         added: "2009-07-29"}).
iana(language,mxr,tag{
         type: "language",
         subtag: "mxr",
         description: "Murik (Malaysia)",
         added: "2009-07-29"}).
iana(language,mxs,tag{
         type: "language",
         subtag: "mxs",
         description: "Huitepec Mixtec",
         added: "2009-07-29"}).
iana(language,mxt,tag{
         type: "language",
         subtag: "mxt",
         description: "Jamiltepec Mixtec",
         added: "2009-07-29"}).
iana(language,mxu,tag{
         type: "language",
         subtag: "mxu",
         description: "Mada (Cameroon)",
         added: "2009-07-29"}).
iana(language,mxv,tag{
         type: "language",
         subtag: "mxv",
         description: "Metlatónoc Mixtec",
         added: "2009-07-29"}).
iana(language,mxw,tag{
         type: "language",
         subtag: "mxw",
         description: "Namo",
         added: "2009-07-29"}).
iana(language,mxx,tag{
         type: "language",
         subtag: "mxx",
         description: ["Mahou","Mawukakan"],
         added: "2009-07-29"}).
iana(language,mxy,tag{
         type: "language",
         subtag: "mxy",
         description: "Southeastern Nochixtlán Mixtec",
         added: "2009-07-29"}).
iana(language,mxz,tag{
         type: "language",
         subtag: "mxz",
         description: "Central Masela",
         added: "2009-07-29"}).
iana(language,myb,tag{
         type: "language",
         subtag: "myb",
         description: "Mbay",
         added: "2009-07-29"}).
iana(language,myc,tag{
         type: "language",
         subtag: "myc",
         description: "Mayeka",
         added: "2009-07-29"}).
iana(language,myd,tag{
         type: "language",
         subtag: "myd",
         description: "Maramba",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         preferred_value: "aog"}).
iana(language,mye,tag{
         type: "language",
         subtag: "mye",
         description: "Myene",
         added: "2009-07-29"}).
iana(language,myf,tag{
         type: "language",
         subtag: "myf",
         description: "Bambassi",
         added: "2009-07-29"}).
iana(language,myg,tag{
         type: "language",
         subtag: "myg",
         description: "Manta",
         added: "2009-07-29"}).
iana(language,myh,tag{
         type: "language",
         subtag: "myh",
         description: "Makah",
         added: "2009-07-29"}).
iana(language,myi,tag{
         type: "language",
         subtag: "myi",
         description: "Mina (India)",
         added: "2009-07-29",
         deprecated: "2019-04-16"}).
iana(language,myj,tag{
         type: "language",
         subtag: "myj",
         description: "Mangayat",
         added: "2009-07-29"}).
iana(language,myk,tag{
         type: "language",
         subtag: "myk",
         description: "Mamara Senoufo",
         added: "2009-07-29"}).
iana(language,myl,tag{
         type: "language",
         subtag: "myl",
         description: "Moma",
         added: "2009-07-29"}).
iana(language,mym,tag{
         type: "language",
         subtag: "mym",
         description: "Me'en",
         added: "2009-07-29"}).
iana(language,myn,tag{
         type: "language",
         subtag: "myn",
         description: "Mayan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,myo,tag{
         type: "language",
         subtag: "myo",
         description: "Anfillo",
         added: "2009-07-29"}).
iana(language,myp,tag{
         type: "language",
         subtag: "myp",
         description: "Pirahã",
         added: "2009-07-29"}).
iana(language,myq,tag{
         type: "language",
         subtag: "myq",
         description: "Forest Maninka",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         macrolanguage: "man"}).
iana(language,myr,tag{
         type: "language",
         subtag: "myr",
         description: "Muniche",
         added: "2009-07-29"}).
iana(language,mys,tag{
         type: "language",
         subtag: "mys",
         description: "Mesmes",
         added: "2009-07-29"}).
iana(language,myt,tag{
         type: "language",
         subtag: "myt",
         description: "Sangab Mandaya",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "mry"}).
iana(language,myu,tag{
         type: "language",
         subtag: "myu",
         description: "Mundurukú",
         added: "2009-07-29"}).
iana(language,myv,tag{
         type: "language",
         subtag: "myv",
         description: "Erzya",
         added: "2005-10-16"}).
iana(language,myw,tag{
         type: "language",
         subtag: "myw",
         description: "Muyuw",
         added: "2009-07-29"}).
iana(language,myx,tag{
         type: "language",
         subtag: "myx",
         description: "Masaaba",
         added: "2009-07-29"}).
iana(language,myy,tag{
         type: "language",
         subtag: "myy",
         description: "Macuna",
         added: "2009-07-29"}).
iana(language,myz,tag{
         type: "language",
         subtag: "myz",
         description: "Classical Mandaic",
         added: "2009-07-29"}).
iana(language,mza,tag{
         type: "language",
         subtag: "mza",
         description: "Santa María Zacatepec Mixtec",
         added: "2009-07-29"}).
iana(language,mzb,tag{
         type: "language",
         subtag: "mzb",
         description: "Tumzabt",
         added: "2009-07-29"}).
iana(language,mzc,tag{
         type: "language",
         subtag: "mzc",
         description: "Madagascar Sign Language",
         added: "2009-07-29"}).
iana(language,mzd,tag{
         type: "language",
         subtag: "mzd",
         description: "Malimba",
         added: "2009-07-29"}).
iana(language,mze,tag{
         type: "language",
         subtag: "mze",
         description: "Morawa",
         added: "2009-07-29"}).
iana(language,mzg,tag{
         type: "language",
         subtag: "mzg",
         description: "Monastic Sign Language",
         added: "2009-07-29"}).
iana(language,mzh,tag{
         type: "language",
         subtag: "mzh",
         description: "Wichí Lhamtés Güisnay",
         added: "2009-07-29"}).
iana(language,mzi,tag{
         type: "language",
         subtag: "mzi",
         description: "Ixcatlán Mazatec",
         added: "2009-07-29"}).
iana(language,mzj,tag{
         type: "language",
         subtag: "mzj",
         description: "Manya",
         added: "2009-07-29"}).
iana(language,mzk,tag{
         type: "language",
         subtag: "mzk",
         description: "Nigeria Mambila",
         added: "2009-07-29"}).
iana(language,mzl,tag{
         type: "language",
         subtag: "mzl",
         description: "Mazatlán Mixe",
         added: "2009-07-29"}).
iana(language,mzm,tag{
         type: "language",
         subtag: "mzm",
         description: "Mumuye",
         added: "2009-07-29"}).
iana(language,mzn,tag{
         type: "language",
         subtag: "mzn",
         description: "Mazanderani",
         added: "2009-07-29"}).
iana(language,mzo,tag{
         type: "language",
         subtag: "mzo",
         description: "Matipuhy",
         added: "2009-07-29"}).
iana(language,mzp,tag{
         type: "language",
         subtag: "mzp",
         description: "Movima",
         added: "2009-07-29"}).
iana(language,mzq,tag{
         type: "language",
         subtag: "mzq",
         description: "Mori Atas",
         added: "2009-07-29"}).
iana(language,mzr,tag{
         type: "language",
         subtag: "mzr",
         description: "Marúbo",
         added: "2009-07-29"}).
iana(language,mzs,tag{
         type: "language",
         subtag: "mzs",
         description: "Macanese",
         added: "2009-07-29"}).
iana(language,mzt,tag{
         type: "language",
         subtag: "mzt",
         description: "Mintil",
         added: "2009-07-29"}).
iana(language,mzu,tag{
         type: "language",
         subtag: "mzu",
         description: "Inapang",
         added: "2009-07-29"}).
iana(language,mzv,tag{
         type: "language",
         subtag: "mzv",
         description: "Manza",
         added: "2009-07-29"}).
iana(language,mzw,tag{
         type: "language",
         subtag: "mzw",
         description: "Deg",
         added: "2009-07-29"}).
iana(language,mzx,tag{
         type: "language",
         subtag: "mzx",
         description: "Mawayana",
         added: "2009-07-29"}).
iana(language,mzy,tag{
         type: "language",
         subtag: "mzy",
         description: "Mozambican Sign Language",
         added: "2009-07-29"}).
iana(language,mzz,tag{
         type: "language",
         subtag: "mzz",
         description: "Maiadomu",
         added: "2009-07-29"}).
iana(language,naa,tag{
         type: "language",
         subtag: "naa",
         description: "Namla",
         added: "2009-07-29"}).
iana(language,nab,tag{
         type: "language",
         subtag: "nab",
         description: "Southern Nambikuára",
         added: "2009-07-29"}).
iana(language,nac,tag{
         type: "language",
         subtag: "nac",
         description: "Narak",
         added: "2009-07-29"}).
iana(language,nad,tag{
         type: "language",
         subtag: "nad",
         description: "Nijadali",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "xny"}).
iana(language,nae,tag{
         type: "language",
         subtag: "nae",
         description: "Naka'ela",
         added: "2009-07-29"}).
iana(language,naf,tag{
         type: "language",
         subtag: "naf",
         description: "Nabak",
         added: "2009-07-29"}).
iana(language,nag,tag{
         type: "language",
         subtag: "nag",
         description: "Naga Pidgin",
         added: "2009-07-29"}).
iana(language,nah,tag{
         type: "language",
         subtag: "nah",
         description: "Nahuatl languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,nai,tag{
         type: "language",
         subtag: "nai",
         description: "North American Indian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,naj,tag{
         type: "language",
         subtag: "naj",
         description: "Nalu",
         added: "2009-07-29"}).
iana(language,nak,tag{
         type: "language",
         subtag: "nak",
         description: "Nakanai",
         added: "2009-07-29"}).
iana(language,nal,tag{
         type: "language",
         subtag: "nal",
         description: "Nalik",
         added: "2009-07-29"}).
iana(language,nam,tag{
         type: "language",
         subtag: "nam",
         description: "Ngan'gityemerri",
         added: "2009-07-29"}).
iana(language,nan,tag{
         type: "language",
         subtag: "nan",
         description: "Min Nan Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,nao,tag{
         type: "language",
         subtag: "nao",
         description: "Naaba",
         added: "2009-07-29"}).
iana(language,nap,tag{
         type: "language",
         subtag: "nap",
         description: "Neapolitan",
         added: "2005-10-16"}).
iana(language,naq,tag{
         type: "language",
         subtag: "naq",
         description: ["Khoekhoe","Nama (Namibia)"],
         added: "2009-07-29"}).
iana(language,nar,tag{
         type: "language",
         subtag: "nar",
         description: "Iguta",
         added: "2009-07-29"}).
iana(language,nas,tag{
         type: "language",
         subtag: "nas",
         description: "Naasioi",
         added: "2009-07-29"}).
iana(language,nat,tag{
         type: "language",
         subtag: "nat",
         description: ["Ca̱hungwa̱rya̱","Hungworo"],
         added: "2009-07-29"}).
iana(language,naw,tag{
         type: "language",
         subtag: "naw",
         description: "Nawuri",
         added: "2009-07-29"}).
iana(language,nax,tag{
         type: "language",
         subtag: "nax",
         description: "Nakwi",
         added: "2009-07-29"}).
iana(language,nay,tag{
         type: "language",
         subtag: "nay",
         description: "Ngarrindjeri",
         added: "2009-07-29"}).
iana(language,naz,tag{
         type: "language",
         subtag: "naz",
         description: "Coatepec Nahuatl",
         added: "2009-07-29"}).
iana(language,nba,tag{
         type: "language",
         subtag: "nba",
         description: "Nyemba",
         added: "2009-07-29"}).
iana(language,nbb,tag{
         type: "language",
         subtag: "nbb",
         description: "Ndoe",
         added: "2009-07-29"}).
iana(language,nbc,tag{
         type: "language",
         subtag: "nbc",
         description: "Chang Naga",
         added: "2009-07-29"}).
iana(language,nbd,tag{
         type: "language",
         subtag: "nbd",
         description: "Ngbinda",
         added: "2009-07-29"}).
iana(language,nbe,tag{
         type: "language",
         subtag: "nbe",
         description: "Konyak Naga",
         added: "2009-07-29"}).
iana(language,nbf,tag{
         type: "language",
         subtag: "nbf",
         description: "Naxi",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         comments: "see nru, nxq"}).
iana(language,nbg,tag{
         type: "language",
         subtag: "nbg",
         description: "Nagarchal",
         added: "2009-07-29"}).
iana(language,nbh,tag{
         type: "language",
         subtag: "nbh",
         description: "Ngamo",
         added: "2009-07-29"}).
iana(language,nbi,tag{
         type: "language",
         subtag: "nbi",
         description: "Mao Naga",
         added: "2009-07-29"}).
iana(language,nbj,tag{
         type: "language",
         subtag: "nbj",
         description: "Ngarinyman",
         added: "2009-07-29"}).
iana(language,nbk,tag{
         type: "language",
         subtag: "nbk",
         description: "Nake",
         added: "2009-07-29"}).
iana(language,nbm,tag{
         type: "language",
         subtag: "nbm",
         description: "Ngbaka Ma'bo",
         added: "2009-07-29"}).
iana(language,nbn,tag{
         type: "language",
         subtag: "nbn",
         description: "Kuri",
         added: "2009-07-29"}).
iana(language,nbo,tag{
         type: "language",
         subtag: "nbo",
         description: "Nkukoli",
         added: "2009-07-29"}).
iana(language,nbp,tag{
         type: "language",
         subtag: "nbp",
         description: "Nnam",
         added: "2009-07-29"}).
iana(language,nbq,tag{
         type: "language",
         subtag: "nbq",
         description: "Nggem",
         added: "2009-07-29"}).
iana(language,nbr,tag{
         type: "language",
         subtag: "nbr",
         description: "Numana",
         added: "2009-07-29"}).
iana(language,nbs,tag{
         type: "language",
         subtag: "nbs",
         description: "Namibian Sign Language",
         added: "2009-07-29"}).
iana(language,nbt,tag{
         type: "language",
         subtag: "nbt",
         description: "Na",
         added: "2009-07-29"}).
iana(language,nbu,tag{
         type: "language",
         subtag: "nbu",
         description: "Rongmei Naga",
         added: "2009-07-29"}).
iana(language,nbv,tag{
         type: "language",
         subtag: "nbv",
         description: "Ngamambo",
         added: "2009-07-29"}).
iana(language,nbw,tag{
         type: "language",
         subtag: "nbw",
         description: "Southern Ngbandi",
         added: "2009-07-29"}).
iana(language,nbx,tag{
         type: "language",
         subtag: "nbx",
         description: "Ngura",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see ekc, gll, jbi, xpt, xwk"}).
iana(language,nby,tag{
         type: "language",
         subtag: "nby",
         description: "Ningera",
         added: "2009-07-29"}).
iana(language,nca,tag{
         type: "language",
         subtag: "nca",
         description: "Iyo",
         added: "2009-07-29"}).
iana(language,ncb,tag{
         type: "language",
         subtag: "ncb",
         description: "Central Nicobarese",
         added: "2009-07-29"}).
iana(language,ncc,tag{
         type: "language",
         subtag: "ncc",
         description: "Ponam",
         added: "2009-07-29"}).
iana(language,ncd,tag{
         type: "language",
         subtag: "ncd",
         description: "Nachering",
         added: "2009-07-29"}).
iana(language,nce,tag{
         type: "language",
         subtag: "nce",
         description: "Yale",
         added: "2009-07-29"}).
iana(language,ncf,tag{
         type: "language",
         subtag: "ncf",
         description: "Notsi",
         added: "2009-07-29"}).
iana(language,ncg,tag{
         type: "language",
         subtag: "ncg",
         description: "Nisga'a",
         added: "2009-07-29"}).
iana(language,nch,tag{
         type: "language",
         subtag: "nch",
         description: "Central Huasteca Nahuatl",
         added: "2009-07-29"}).
iana(language,nci,tag{
         type: "language",
         subtag: "nci",
         description: "Classical Nahuatl",
         added: "2009-07-29"}).
iana(language,ncj,tag{
         type: "language",
         subtag: "ncj",
         description: "Northern Puebla Nahuatl",
         added: "2009-07-29"}).
iana(language,nck,tag{
         type: "language",
         subtag: "nck",
         description: "Na-kara",
         added: "2009-07-29"}).
iana(language,ncl,tag{
         type: "language",
         subtag: "ncl",
         description: "Michoacán Nahuatl",
         added: "2009-07-29"}).
iana(language,ncm,tag{
         type: "language",
         subtag: "ncm",
         description: "Nambo",
         added: "2009-07-29"}).
iana(language,ncn,tag{
         type: "language",
         subtag: "ncn",
         description: "Nauna",
         added: "2009-07-29"}).
iana(language,nco,tag{
         type: "language",
         subtag: "nco",
         description: "Sibe",
         added: "2009-07-29"}).
iana(language,ncp,tag{
         type: "language",
         subtag: "ncp",
         description: "Ndaktup",
         added: "2009-07-29",
         deprecated: "2018-03-08",
         preferred_value: "kdz"}).
iana(language,ncq,tag{
         type: "language",
         subtag: "ncq",
         description: "Northern Katang",
         added: "2017-02-23"}).
iana(language,ncr,tag{
         type: "language",
         subtag: "ncr",
         description: "Ncane",
         added: "2009-07-29"}).
iana(language,ncs,tag{
         type: "language",
         subtag: "ncs",
         description: "Nicaraguan Sign Language",
         added: "2009-07-29"}).
iana(language,nct,tag{
         type: "language",
         subtag: "nct",
         description: "Chothe Naga",
         added: "2009-07-29"}).
iana(language,ncu,tag{
         type: "language",
         subtag: "ncu",
         description: "Chumburung",
         added: "2009-07-29"}).
iana(language,ncx,tag{
         type: "language",
         subtag: "ncx",
         description: "Central Puebla Nahuatl",
         added: "2009-07-29"}).
iana(language,ncz,tag{
         type: "language",
         subtag: "ncz",
         description: "Natchez",
         added: "2009-07-29"}).
iana(language,nda,tag{
         type: "language",
         subtag: "nda",
         description: "Ndasa",
         added: "2009-07-29"}).
iana(language,ndb,tag{
         type: "language",
         subtag: "ndb",
         description: "Kenswei Nsei",
         added: "2009-07-29"}).
iana(language,ndc,tag{
         type: "language",
         subtag: "ndc",
         description: "Ndau",
         added: "2009-07-29"}).
iana(language,ndd,tag{
         type: "language",
         subtag: "ndd",
         description: "Nde-Nsele-Nta",
         added: "2009-07-29"}).
iana(language,ndf,tag{
         type: "language",
         subtag: "ndf",
         description: "Nadruvian",
         added: "2009-07-29"}).
iana(language,ndg,tag{
         type: "language",
         subtag: "ndg",
         description: "Ndengereko",
         added: "2009-07-29"}).
iana(language,ndh,tag{
         type: "language",
         subtag: "ndh",
         description: "Ndali",
         added: "2009-07-29"}).
iana(language,ndi,tag{
         type: "language",
         subtag: "ndi",
         description: "Samba Leko",
         added: "2009-07-29"}).
iana(language,ndj,tag{
         type: "language",
         subtag: "ndj",
         description: "Ndamba",
         added: "2009-07-29"}).
iana(language,ndk,tag{
         type: "language",
         subtag: "ndk",
         description: "Ndaka",
         added: "2009-07-29"}).
iana(language,ndl,tag{
         type: "language",
         subtag: "ndl",
         description: "Ndolo",
         added: "2009-07-29"}).
iana(language,ndm,tag{
         type: "language",
         subtag: "ndm",
         description: "Ndam",
         added: "2009-07-29"}).
iana(language,ndn,tag{
         type: "language",
         subtag: "ndn",
         description: "Ngundi",
         added: "2009-07-29"}).
iana(language,ndp,tag{
         type: "language",
         subtag: "ndp",
         description: "Ndo",
         added: "2009-07-29"}).
iana(language,ndq,tag{
         type: "language",
         subtag: "ndq",
         description: "Ndombe",
         added: "2009-07-29"}).
iana(language,ndr,tag{
         type: "language",
         subtag: "ndr",
         description: "Ndoola",
         added: "2009-07-29"}).
iana(language,nds,tag{
         type: "language",
         subtag: "nds",
         description: ["Low German","Low Saxon"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ndt,tag{
         type: "language",
         subtag: "ndt",
         description: "Ndunga",
         added: "2009-07-29"}).
iana(language,ndu,tag{
         type: "language",
         subtag: "ndu",
         description: "Dugun",
         added: "2009-07-29"}).
iana(language,ndv,tag{
         type: "language",
         subtag: "ndv",
         description: "Ndut",
         added: "2009-07-29"}).
iana(language,ndw,tag{
         type: "language",
         subtag: "ndw",
         description: "Ndobo",
         added: "2009-07-29"}).
iana(language,ndx,tag{
         type: "language",
         subtag: "ndx",
         description: "Nduga",
         added: "2009-07-29"}).
iana(language,ndy,tag{
         type: "language",
         subtag: "ndy",
         description: "Lutos",
         added: "2009-07-29"}).
iana(language,ndz,tag{
         type: "language",
         subtag: "ndz",
         description: "Ndogo",
         added: "2009-07-29"}).
iana(language,nea,tag{
         type: "language",
         subtag: "nea",
         description: "Eastern Ngad'a",
         added: "2009-07-29"}).
iana(language,neb,tag{
         type: "language",
         subtag: "neb",
         description: "Toura (Côte d'Ivoire)",
         added: "2009-07-29"}).
iana(language,nec,tag{
         type: "language",
         subtag: "nec",
         description: "Nedebang",
         added: "2009-07-29"}).
iana(language,ned,tag{
         type: "language",
         subtag: "ned",
         description: "Nde-Gbite",
         added: "2009-07-29"}).
iana(language,nee,tag{
         type: "language",
         subtag: "nee",
         description: "Nêlêmwa-Nixumwak",
         added: "2009-07-29"}).
iana(language,nef,tag{
         type: "language",
         subtag: "nef",
         description: "Nefamese",
         added: "2009-07-29"}).
iana(language,neg,tag{
         type: "language",
         subtag: "neg",
         description: "Negidal",
         added: "2009-07-29"}).
iana(language,neh,tag{
         type: "language",
         subtag: "neh",
         description: "Nyenkha",
         added: "2009-07-29"}).
iana(language,nei,tag{
         type: "language",
         subtag: "nei",
         description: "Neo-Hittite",
         added: "2009-07-29"}).
iana(language,nej,tag{
         type: "language",
         subtag: "nej",
         description: "Neko",
         added: "2009-07-29"}).
iana(language,nek,tag{
         type: "language",
         subtag: "nek",
         description: "Neku",
         added: "2009-07-29"}).
iana(language,nem,tag{
         type: "language",
         subtag: "nem",
         description: "Nemi",
         added: "2009-07-29"}).
iana(language,nen,tag{
         type: "language",
         subtag: "nen",
         description: "Nengone",
         added: "2009-07-29"}).
iana(language,neo,tag{
         type: "language",
         subtag: "neo",
         description: "Ná-Meo",
         added: "2009-07-29"}).
iana(language,neq,tag{
         type: "language",
         subtag: "neq",
         description: "North Central Mixe",
         added: "2009-07-29"}).
iana(language,ner,tag{
         type: "language",
         subtag: "ner",
         description: "Yahadian",
         added: "2009-07-29"}).
iana(language,nes,tag{
         type: "language",
         subtag: "nes",
         description: "Bhoti Kinnauri",
         added: "2009-07-29"}).
iana(language,net,tag{
         type: "language",
         subtag: "net",
         description: "Nete",
         added: "2009-07-29"}).
iana(language,neu,tag{
         type: "language",
         subtag: "neu",
         description: "Neo",
         added: "2012-08-12"}).
iana(language,nev,tag{
         type: "language",
         subtag: "nev",
         description: "Nyaheun",
         added: "2009-07-29"}).
iana(language,new,tag{
         type: "language",
         subtag: "new",
         description: ["Newari","Nepal Bhasa"],
         added: "2005-10-16"}).
iana(language,nex,tag{
         type: "language",
         subtag: "nex",
         description: "Neme",
         added: "2009-07-29"}).
iana(language,ney,tag{
         type: "language",
         subtag: "ney",
         description: "Neyo",
         added: "2009-07-29"}).
iana(language,nez,tag{
         type: "language",
         subtag: "nez",
         description: "Nez Perce",
         added: "2009-07-29"}).
iana(language,nfa,tag{
         type: "language",
         subtag: "nfa",
         description: "Dhao",
         added: "2009-07-29"}).
iana(language,nfd,tag{
         type: "language",
         subtag: "nfd",
         description: "Ahwai",
         added: "2009-07-29"}).
iana(language,nfl,tag{
         type: "language",
         subtag: "nfl",
         description: ["Ayiwo","Äiwoo"],
         added: "2009-07-29"}).
iana(language,nfr,tag{
         type: "language",
         subtag: "nfr",
         description: "Nafaanra",
         added: "2009-07-29"}).
iana(language,nfu,tag{
         type: "language",
         subtag: "nfu",
         description: "Mfumte",
         added: "2009-07-29"}).
iana(language,nga,tag{
         type: "language",
         subtag: "nga",
         description: "Ngbaka",
         added: "2009-07-29"}).
iana(language,ngb,tag{
         type: "language",
         subtag: "ngb",
         description: "Northern Ngbandi",
         added: "2009-07-29"}).
iana(language,ngc,tag{
         type: "language",
         subtag: "ngc",
         description: "Ngombe (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,ngd,tag{
         type: "language",
         subtag: "ngd",
         description: "Ngando (Central African Republic)",
         added: "2009-07-29"}).
iana(language,nge,tag{
         type: "language",
         subtag: "nge",
         description: "Ngemba",
         added: "2009-07-29"}).
iana(language,ngf,tag{
         type: "language",
         subtag: "ngf",
         description: "Trans-New Guinea languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,ngg,tag{
         type: "language",
         subtag: "ngg",
         description: "Ngbaka Manza",
         added: "2009-07-29"}).
iana(language,ngh,tag{
         type: "language",
         subtag: "ngh",
         description: "Nǁng",
         added: "2009-07-29"}).
iana(language,ngi,tag{
         type: "language",
         subtag: "ngi",
         description: "Ngizim",
         added: "2009-07-29"}).
iana(language,ngj,tag{
         type: "language",
         subtag: "ngj",
         description: "Ngie",
         added: "2009-07-29"}).
iana(language,ngk,tag{
         type: "language",
         subtag: "ngk",
         description: "Dalabon",
         added: "2009-07-29"}).
iana(language,ngl,tag{
         type: "language",
         subtag: "ngl",
         description: "Lomwe",
         added: "2009-07-29"}).
iana(language,ngm,tag{
         type: "language",
         subtag: "ngm",
         description: "Ngatik Men's Creole",
         added: "2009-07-29"}).
iana(language,ngn,tag{
         type: "language",
         subtag: "ngn",
         description: "Ngwo",
         added: "2009-07-29"}).
iana(language,ngo,tag{
         type: "language",
         subtag: "ngo",
         description: "Ngoni",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see xnj, xnq"}).
iana(language,ngp,tag{
         type: "language",
         subtag: "ngp",
         description: "Ngulu",
         added: "2009-07-29"}).
iana(language,ngq,tag{
         type: "language",
         subtag: "ngq",
         description: ["Ngurimi","Ngoreme"],
         added: "2009-07-29"}).
iana(language,ngr,tag{
         type: "language",
         subtag: "ngr",
         description: "Engdewu",
         added: "2009-07-29"}).
iana(language,ngs,tag{
         type: "language",
         subtag: "ngs",
         description: "Gvoko",
         added: "2009-07-29"}).
iana(language,ngt,tag{
         type: "language",
         subtag: "ngt",
         description: ["Kriang","Ngeq"],
         added: "2009-07-29"}).
iana(language,ngu,tag{
         type: "language",
         subtag: "ngu",
         description: "Guerrero Nahuatl",
         added: "2009-07-29"}).
iana(language,ngv,tag{
         type: "language",
         subtag: "ngv",
         description: "Nagumi",
         added: "2009-07-29"}).
iana(language,ngw,tag{
         type: "language",
         subtag: "ngw",
         description: "Ngwaba",
         added: "2009-07-29"}).
iana(language,ngx,tag{
         type: "language",
         subtag: "ngx",
         description: "Nggwahyi",
         added: "2009-07-29"}).
iana(language,ngy,tag{
         type: "language",
         subtag: "ngy",
         description: "Tibea",
         added: "2009-07-29"}).
iana(language,ngz,tag{
         type: "language",
         subtag: "ngz",
         description: "Ngungwel",
         added: "2009-07-29"}).
iana(language,nha,tag{
         type: "language",
         subtag: "nha",
         description: "Nhanda",
         added: "2009-07-29"}).
iana(language,nhb,tag{
         type: "language",
         subtag: "nhb",
         description: "Beng",
         added: "2009-07-29"}).
iana(language,nhc,tag{
         type: "language",
         subtag: "nhc",
         description: "Tabasco Nahuatl",
         added: "2009-07-29"}).
iana(language,nhd,tag{
         type: "language",
         subtag: "nhd",
         description: ["Chiripá","Ava Guaraní"],
         added: "2009-07-29",
         macrolanguage: "gn"}).
iana(language,nhe,tag{
         type: "language",
         subtag: "nhe",
         description: "Eastern Huasteca Nahuatl",
         added: "2009-07-29"}).
iana(language,nhf,tag{
         type: "language",
         subtag: "nhf",
         description: "Nhuwala",
         added: "2009-07-29"}).
iana(language,nhg,tag{
         type: "language",
         subtag: "nhg",
         description: "Tetelcingo Nahuatl",
         added: "2009-07-29"}).
iana(language,nhh,tag{
         type: "language",
         subtag: "nhh",
         description: "Nahari",
         added: "2009-07-29"}).
iana(language,nhi,tag{
         type: "language",
         subtag: "nhi",
         description: "Zacatlán-Ahuacatlán-Tepetzintla Nahuatl",
         added: "2009-07-29"}).
iana(language,nhk,tag{
         type: "language",
         subtag: "nhk",
         description: "Isthmus-Cosoleacaque Nahuatl",
         added: "2009-07-29"}).
iana(language,nhm,tag{
         type: "language",
         subtag: "nhm",
         description: "Morelos Nahuatl",
         added: "2009-07-29"}).
iana(language,nhn,tag{
         type: "language",
         subtag: "nhn",
         description: "Central Nahuatl",
         added: "2009-07-29"}).
iana(language,nho,tag{
         type: "language",
         subtag: "nho",
         description: "Takuu",
         added: "2009-07-29"}).
iana(language,nhp,tag{
         type: "language",
         subtag: "nhp",
         description: "Isthmus-Pajapan Nahuatl",
         added: "2009-07-29"}).
iana(language,nhq,tag{
         type: "language",
         subtag: "nhq",
         description: "Huaxcaleca Nahuatl",
         added: "2009-07-29"}).
iana(language,nhr,tag{
         type: "language",
         subtag: "nhr",
         description: "Naro",
         added: "2009-07-29"}).
iana(language,nht,tag{
         type: "language",
         subtag: "nht",
         description: "Ometepec Nahuatl",
         added: "2009-07-29"}).
iana(language,nhu,tag{
         type: "language",
         subtag: "nhu",
         description: "Noone",
         added: "2009-07-29"}).
iana(language,nhv,tag{
         type: "language",
         subtag: "nhv",
         description: "Temascaltepec Nahuatl",
         added: "2009-07-29"}).
iana(language,nhw,tag{
         type: "language",
         subtag: "nhw",
         description: "Western Huasteca Nahuatl",
         added: "2009-07-29"}).
iana(language,nhx,tag{
         type: "language",
         subtag: "nhx",
         description: "Isthmus-Mecayapan Nahuatl",
         added: "2009-07-29"}).
iana(language,nhy,tag{
         type: "language",
         subtag: "nhy",
         description: "Northern Oaxaca Nahuatl",
         added: "2009-07-29"}).
iana(language,nhz,tag{
         type: "language",
         subtag: "nhz",
         description: "Santa María La Alta Nahuatl",
         added: "2009-07-29"}).
iana(language,nia,tag{
         type: "language",
         subtag: "nia",
         description: "Nias",
         added: "2005-10-16"}).
iana(language,nib,tag{
         type: "language",
         subtag: "nib",
         description: "Nakame",
         added: "2009-07-29"}).
iana(language,nic,tag{
         type: "language",
         subtag: "nic",
         description: "Niger-Kordofanian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,nid,tag{
         type: "language",
         subtag: "nid",
         description: "Ngandi",
         added: "2009-07-29"}).
iana(language,nie,tag{
         type: "language",
         subtag: "nie",
         description: "Niellim",
         added: "2009-07-29"}).
iana(language,nif,tag{
         type: "language",
         subtag: "nif",
         description: "Nek",
         added: "2009-07-29"}).
iana(language,nig,tag{
         type: "language",
         subtag: "nig",
         description: "Ngalakgan",
         added: "2009-07-29"}).
iana(language,nih,tag{
         type: "language",
         subtag: "nih",
         description: "Nyiha (Tanzania)",
         added: "2009-07-29"}).
iana(language,nii,tag{
         type: "language",
         subtag: "nii",
         description: "Nii",
         added: "2009-07-29"}).
iana(language,nij,tag{
         type: "language",
         subtag: "nij",
         description: "Ngaju",
         added: "2009-07-29"}).
iana(language,nik,tag{
         type: "language",
         subtag: "nik",
         description: "Southern Nicobarese",
         added: "2009-07-29"}).
iana(language,nil,tag{
         type: "language",
         subtag: "nil",
         description: "Nila",
         added: "2009-07-29"}).
iana(language,nim,tag{
         type: "language",
         subtag: "nim",
         description: "Nilamba",
         added: "2009-07-29"}).
iana(language,nin,tag{
         type: "language",
         subtag: "nin",
         description: "Ninzo",
         added: "2009-07-29"}).
iana(language,nio,tag{
         type: "language",
         subtag: "nio",
         description: "Nganasan",
         added: "2009-07-29"}).
iana(language,niq,tag{
         type: "language",
         subtag: "niq",
         description: "Nandi",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,nir,tag{
         type: "language",
         subtag: "nir",
         description: "Nimboran",
         added: "2009-07-29"}).
iana(language,nis,tag{
         type: "language",
         subtag: "nis",
         description: "Nimi",
         added: "2009-07-29"}).
iana(language,nit,tag{
         type: "language",
         subtag: "nit",
         description: "Southeastern Kolami",
         added: "2009-07-29"}).
iana(language,niu,tag{
         type: "language",
         subtag: "niu",
         description: "Niuean",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,niv,tag{
         type: "language",
         subtag: "niv",
         description: "Gilyak",
         added: "2009-07-29"}).
iana(language,niw,tag{
         type: "language",
         subtag: "niw",
         description: "Nimo",
         added: "2009-07-29"}).
iana(language,nix,tag{
         type: "language",
         subtag: "nix",
         description: "Hema",
         added: "2009-07-29"}).
iana(language,niy,tag{
         type: "language",
         subtag: "niy",
         description: "Ngiti",
         added: "2009-07-29"}).
iana(language,niz,tag{
         type: "language",
         subtag: "niz",
         description: "Ningil",
         added: "2009-07-29"}).
iana(language,nja,tag{
         type: "language",
         subtag: "nja",
         description: "Nzanyi",
         added: "2009-07-29"}).
iana(language,njb,tag{
         type: "language",
         subtag: "njb",
         description: "Nocte Naga",
         added: "2009-07-29"}).
iana(language,njd,tag{
         type: "language",
         subtag: "njd",
         description: "Ndonde Hamba",
         added: "2009-07-29"}).
iana(language,njh,tag{
         type: "language",
         subtag: "njh",
         description: "Lotha Naga",
         added: "2009-07-29"}).
iana(language,nji,tag{
         type: "language",
         subtag: "nji",
         description: "Gudanji",
         added: "2009-07-29"}).
iana(language,njj,tag{
         type: "language",
         subtag: "njj",
         description: "Njen",
         added: "2009-07-29"}).
iana(language,njl,tag{
         type: "language",
         subtag: "njl",
         description: "Njalgulgule",
         added: "2009-07-29"}).
iana(language,njm,tag{
         type: "language",
         subtag: "njm",
         description: "Angami Naga",
         added: "2009-07-29"}).
iana(language,njn,tag{
         type: "language",
         subtag: "njn",
         description: "Liangmai Naga",
         added: "2009-07-29"}).
iana(language,njo,tag{
         type: "language",
         subtag: "njo",
         description: "Ao Naga",
         added: "2009-07-29"}).
iana(language,njr,tag{
         type: "language",
         subtag: "njr",
         description: "Njerep",
         added: "2009-07-29"}).
iana(language,njs,tag{
         type: "language",
         subtag: "njs",
         description: "Nisa",
         added: "2009-07-29"}).
iana(language,njt,tag{
         type: "language",
         subtag: "njt",
         description: "Ndyuka-Trio Pidgin",
         added: "2009-07-29"}).
iana(language,nju,tag{
         type: "language",
         subtag: "nju",
         description: "Ngadjunmaya",
         added: "2009-07-29"}).
iana(language,njx,tag{
         type: "language",
         subtag: "njx",
         description: "Kunyi",
         added: "2009-07-29"}).
iana(language,njy,tag{
         type: "language",
         subtag: "njy",
         description: "Njyem",
         added: "2009-07-29"}).
iana(language,njz,tag{
         type: "language",
         subtag: "njz",
         description: "Nyishi",
         added: "2012-08-12"}).
iana(language,nka,tag{
         type: "language",
         subtag: "nka",
         description: "Nkoya",
         added: "2009-07-29"}).
iana(language,nkb,tag{
         type: "language",
         subtag: "nkb",
         description: "Khoibu Naga",
         added: "2009-07-29"}).
iana(language,nkc,tag{
         type: "language",
         subtag: "nkc",
         description: "Nkongho",
         added: "2009-07-29"}).
iana(language,nkd,tag{
         type: "language",
         subtag: "nkd",
         description: "Koireng",
         added: "2009-07-29"}).
iana(language,nke,tag{
         type: "language",
         subtag: "nke",
         description: "Duke",
         added: "2009-07-29"}).
iana(language,nkf,tag{
         type: "language",
         subtag: "nkf",
         description: "Inpui Naga",
         added: "2009-07-29"}).
iana(language,nkg,tag{
         type: "language",
         subtag: "nkg",
         description: "Nekgini",
         added: "2009-07-29"}).
iana(language,nkh,tag{
         type: "language",
         subtag: "nkh",
         description: "Khezha Naga",
         added: "2009-07-29"}).
iana(language,nki,tag{
         type: "language",
         subtag: "nki",
         description: "Thangal Naga",
         added: "2009-07-29"}).
iana(language,nkj,tag{
         type: "language",
         subtag: "nkj",
         description: "Nakai",
         added: "2009-07-29"}).
iana(language,nkk,tag{
         type: "language",
         subtag: "nkk",
         description: "Nokuku",
         added: "2009-07-29"}).
iana(language,nkm,tag{
         type: "language",
         subtag: "nkm",
         description: "Namat",
         added: "2009-07-29"}).
iana(language,nkn,tag{
         type: "language",
         subtag: "nkn",
         description: "Nkangala",
         added: "2009-07-29"}).
iana(language,nko,tag{
         type: "language",
         subtag: "nko",
         description: "Nkonya",
         added: "2009-07-29"}).
iana(language,nkp,tag{
         type: "language",
         subtag: "nkp",
         description: "Niuatoputapu",
         added: "2009-07-29"}).
iana(language,nkq,tag{
         type: "language",
         subtag: "nkq",
         description: "Nkami",
         added: "2010-04-16"}).
iana(language,nkr,tag{
         type: "language",
         subtag: "nkr",
         description: "Nukuoro",
         added: "2009-07-29"}).
iana(language,nks,tag{
         type: "language",
         subtag: "nks",
         description: "North Asmat",
         added: "2009-07-29"}).
iana(language,nkt,tag{
         type: "language",
         subtag: "nkt",
         description: "Nyika (Tanzania)",
         added: "2009-07-29"}).
iana(language,nku,tag{
         type: "language",
         subtag: "nku",
         description: "Bouna Kulango",
         added: "2009-07-29"}).
iana(language,nkv,tag{
         type: "language",
         subtag: "nkv",
         description: "Nyika (Malawi and Zambia)",
         added: "2009-07-29"}).
iana(language,nkw,tag{
         type: "language",
         subtag: "nkw",
         description: "Nkutu",
         added: "2009-07-29"}).
iana(language,nkx,tag{
         type: "language",
         subtag: "nkx",
         description: "Nkoroo",
         added: "2009-07-29"}).
iana(language,nkz,tag{
         type: "language",
         subtag: "nkz",
         description: "Nkari",
         added: "2009-07-29"}).
iana(language,nla,tag{
         type: "language",
         subtag: "nla",
         description: "Ngombale",
         added: "2009-07-29"}).
iana(language,nlc,tag{
         type: "language",
         subtag: "nlc",
         description: "Nalca",
         added: "2009-07-29"}).
iana(language,nle,tag{
         type: "language",
         subtag: "nle",
         description: "East Nyala",
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,nlg,tag{
         type: "language",
         subtag: "nlg",
         description: "Gela",
         added: "2009-07-29"}).
iana(language,nli,tag{
         type: "language",
         subtag: "nli",
         description: "Grangali",
         added: "2009-07-29"}).
iana(language,nlj,tag{
         type: "language",
         subtag: "nlj",
         description: "Nyali",
         added: "2009-07-29"}).
iana(language,nlk,tag{
         type: "language",
         subtag: "nlk",
         description: "Ninia Yali",
         added: "2009-07-29"}).
iana(language,nll,tag{
         type: "language",
         subtag: "nll",
         description: "Nihali",
         added: "2009-07-29"}).
iana(language,nlm,tag{
         type: "language",
         subtag: "nlm",
         description: "Mankiyali",
         added: "2018-03-08"}).
iana(language,nln,tag{
         type: "language",
         subtag: "nln",
         description: "Durango Nahuatl",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see azd, azn"}).
iana(language,nlo,tag{
         type: "language",
         subtag: "nlo",
         description: "Ngul",
         added: "2009-07-29"}).
iana(language,nlq,tag{
         type: "language",
         subtag: "nlq",
         description: "Lao Naga",
         added: "2013-09-10"}).
iana(language,nlr,tag{
         type: "language",
         subtag: "nlr",
         description: "Ngarla",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see nrk, ywg"}).
iana(language,nlu,tag{
         type: "language",
         subtag: "nlu",
         description: "Nchumbulu",
         added: "2009-07-29"}).
iana(language,nlv,tag{
         type: "language",
         subtag: "nlv",
         description: "Orizaba Nahuatl",
         added: "2009-07-29"}).
iana(language,nlw,tag{
         type: "language",
         subtag: "nlw",
         description: "Walangama",
         added: "2013-09-10"}).
iana(language,nlx,tag{
         type: "language",
         subtag: "nlx",
         description: "Nahali",
         added: "2009-07-29"}).
iana(language,nly,tag{
         type: "language",
         subtag: "nly",
         description: "Nyamal",
         added: "2009-07-29"}).
iana(language,nlz,tag{
         type: "language",
         subtag: "nlz",
         description: "Nalögo",
         added: "2009-07-29"}).
iana(language,nma,tag{
         type: "language",
         subtag: "nma",
         description: "Maram Naga",
         added: "2009-07-29"}).
iana(language,nmb,tag{
         type: "language",
         subtag: "nmb",
         description: ["Big Nambas","V'ënen Taut"],
         added: "2009-07-29"}).
iana(language,nmc,tag{
         type: "language",
         subtag: "nmc",
         description: "Ngam",
         added: "2009-07-29"}).
iana(language,nmd,tag{
         type: "language",
         subtag: "nmd",
         description: "Ndumu",
         added: "2009-07-29"}).
iana(language,nme,tag{
         type: "language",
         subtag: "nme",
         description: "Mzieme Naga",
         added: "2009-07-29"}).
iana(language,nmf,tag{
         type: "language",
         subtag: "nmf",
         description: "Tangkhul Naga (India)",
         added: "2009-07-29",
         comments: "see ntx"}).
iana(language,nmg,tag{
         type: "language",
         subtag: "nmg",
         description: "Kwasio",
         added: "2009-07-29"}).
iana(language,nmh,tag{
         type: "language",
         subtag: "nmh",
         description: "Monsang Naga",
         added: "2009-07-29"}).
iana(language,nmi,tag{
         type: "language",
         subtag: "nmi",
         description: "Nyam",
         added: "2009-07-29"}).
iana(language,nmj,tag{
         type: "language",
         subtag: "nmj",
         description: "Ngombe (Central African Republic)",
         added: "2009-07-29"}).
iana(language,nmk,tag{
         type: "language",
         subtag: "nmk",
         description: "Namakura",
         added: "2009-07-29"}).
iana(language,nml,tag{
         type: "language",
         subtag: "nml",
         description: "Ndemli",
         added: "2009-07-29"}).
iana(language,nmm,tag{
         type: "language",
         subtag: "nmm",
         description: "Manangba",
         added: "2009-07-29"}).
iana(language,nmn,tag{
         type: "language",
         subtag: "nmn",
         description: "ǃXóõ",
         added: "2009-07-29"}).
iana(language,nmo,tag{
         type: "language",
         subtag: "nmo",
         description: "Moyon Naga",
         added: "2009-07-29"}).
iana(language,nmp,tag{
         type: "language",
         subtag: "nmp",
         description: "Nimanbur",
         added: "2009-07-29"}).
iana(language,nmq,tag{
         type: "language",
         subtag: "nmq",
         description: "Nambya",
         added: "2009-07-29"}).
iana(language,nmr,tag{
         type: "language",
         subtag: "nmr",
         description: "Nimbari",
         added: "2009-07-29"}).
iana(language,nms,tag{
         type: "language",
         subtag: "nms",
         description: "Letemboi",
         added: "2009-07-29"}).
iana(language,nmt,tag{
         type: "language",
         subtag: "nmt",
         description: "Namonuito",
         added: "2009-07-29"}).
iana(language,nmu,tag{
         type: "language",
         subtag: "nmu",
         description: "Northeast Maidu",
         added: "2009-07-29"}).
iana(language,nmv,tag{
         type: "language",
         subtag: "nmv",
         description: "Ngamini",
         added: "2009-07-29"}).
iana(language,nmw,tag{
         type: "language",
         subtag: "nmw",
         description: ["Nimoa","Rifao"],
         added: "2009-07-29"}).
iana(language,nmx,tag{
         type: "language",
         subtag: "nmx",
         description: "Nama (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,nmy,tag{
         type: "language",
         subtag: "nmy",
         description: "Namuyi",
         added: "2009-07-29"}).
iana(language,nmz,tag{
         type: "language",
         subtag: "nmz",
         description: "Nawdm",
         added: "2009-07-29"}).
iana(language,nna,tag{
         type: "language",
         subtag: "nna",
         description: "Nyangumarta",
         added: "2009-07-29"}).
iana(language,nnb,tag{
         type: "language",
         subtag: "nnb",
         description: "Nande",
         added: "2009-07-29"}).
iana(language,nnc,tag{
         type: "language",
         subtag: "nnc",
         description: "Nancere",
         added: "2009-07-29"}).
iana(language,nnd,tag{
         type: "language",
         subtag: "nnd",
         description: "West Ambae",
         added: "2009-07-29"}).
iana(language,nne,tag{
         type: "language",
         subtag: "nne",
         description: "Ngandyera",
         added: "2009-07-29"}).
iana(language,nnf,tag{
         type: "language",
         subtag: "nnf",
         description: "Ngaing",
         added: "2009-07-29"}).
iana(language,nng,tag{
         type: "language",
         subtag: "nng",
         description: "Maring Naga",
         added: "2009-07-29"}).
iana(language,nnh,tag{
         type: "language",
         subtag: "nnh",
         description: "Ngiemboon",
         added: "2009-07-29"}).
iana(language,nni,tag{
         type: "language",
         subtag: "nni",
         description: "North Nuaulu",
         added: "2009-07-29"}).
iana(language,nnj,tag{
         type: "language",
         subtag: "nnj",
         description: "Nyangatom",
         added: "2009-07-29"}).
iana(language,nnk,tag{
         type: "language",
         subtag: "nnk",
         description: "Nankina",
         added: "2009-07-29"}).
iana(language,nnl,tag{
         type: "language",
         subtag: "nnl",
         description: "Northern Rengma Naga",
         added: "2009-07-29"}).
iana(language,nnm,tag{
         type: "language",
         subtag: "nnm",
         description: "Namia",
         added: "2009-07-29"}).
iana(language,nnn,tag{
         type: "language",
         subtag: "nnn",
         description: "Ngete",
         added: "2009-07-29"}).
iana(language,nnp,tag{
         type: "language",
         subtag: "nnp",
         description: "Wancho Naga",
         added: "2009-07-29"}).
iana(language,nnq,tag{
         type: "language",
         subtag: "nnq",
         description: "Ngindo",
         added: "2009-07-29"}).
iana(language,nnr,tag{
         type: "language",
         subtag: "nnr",
         description: "Narungga",
         added: "2009-07-29"}).
iana(language,nns,tag{
         type: "language",
         subtag: "nns",
         description: "Ningye",
         added: "2009-07-29",
         deprecated: "2019-04-16",
         preferred_value: "nbr"}).
iana(language,nnt,tag{
         type: "language",
         subtag: "nnt",
         description: "Nanticoke",
         added: "2009-07-29"}).
iana(language,nnu,tag{
         type: "language",
         subtag: "nnu",
         description: "Dwang",
         added: "2009-07-29"}).
iana(language,nnv,tag{
         type: "language",
         subtag: "nnv",
         description: "Nugunu (Australia)",
         added: "2009-07-29"}).
iana(language,nnw,tag{
         type: "language",
         subtag: "nnw",
         description: "Southern Nuni",
         added: "2009-07-29"}).
iana(language,nnx,tag{
         type: "language",
         subtag: "nnx",
         description: "Ngong",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "ngv"}).
iana(language,nny,tag{
         type: "language",
         subtag: "nny",
         description: "Nyangga",
         added: "2009-07-29"}).
iana(language,nnz,tag{
         type: "language",
         subtag: "nnz",
         description: "Nda'nda'",
         added: "2009-07-29"}).
iana(language,noa,tag{
         type: "language",
         subtag: "noa",
         description: "Woun Meu",
         added: "2009-07-29"}).
iana(language,noc,tag{
         type: "language",
         subtag: "noc",
         description: "Nuk",
         added: "2009-07-29"}).
iana(language,nod,tag{
         type: "language",
         subtag: "nod",
         description: "Northern Thai",
         added: "2009-07-29"}).
iana(language,noe,tag{
         type: "language",
         subtag: "noe",
         description: "Nimadi",
         added: "2009-07-29"}).
iana(language,nof,tag{
         type: "language",
         subtag: "nof",
         description: "Nomane",
         added: "2009-07-29"}).
iana(language,nog,tag{
         type: "language",
         subtag: "nog",
         description: "Nogai",
         added: "2005-10-16"}).
iana(language,noh,tag{
         type: "language",
         subtag: "noh",
         description: "Nomu",
         added: "2009-07-29"}).
iana(language,noi,tag{
         type: "language",
         subtag: "noi",
         description: "Noiri",
         added: "2009-07-29"}).
iana(language,noj,tag{
         type: "language",
         subtag: "noj",
         description: "Nonuya",
         added: "2010-03-11"}).
iana(language,nok,tag{
         type: "language",
         subtag: "nok",
         description: "Nooksack",
         added: "2009-07-29"}).
iana(language,nol,tag{
         type: "language",
         subtag: "nol",
         description: "Nomlaki",
         added: "2013-09-10"}).
iana(language,nom,tag{
         type: "language",
         subtag: "nom",
         description: "Nocamán",
         added: "2009-07-29"}).
iana(language,non,tag{
         type: "language",
         subtag: "non",
         description: "Old Norse",
         added: "2005-10-16"}).
iana(language,noo,tag{
         type: "language",
         subtag: "noo",
         description: "Nootka",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         comments: "see dtd, nuk"}).
iana(language,nop,tag{
         type: "language",
         subtag: "nop",
         description: "Numanggang",
         added: "2009-07-29"}).
iana(language,noq,tag{
         type: "language",
         subtag: "noq",
         description: "Ngongo",
         added: "2009-07-29"}).
iana(language,nos,tag{
         type: "language",
         subtag: "nos",
         description: "Eastern Nisu",
         added: "2009-07-29"}).
iana(language,not,tag{
         type: "language",
         subtag: "not",
         description: "Nomatsiguenga",
         added: "2009-07-29"}).
iana(language,nou,tag{
         type: "language",
         subtag: "nou",
         description: "Ewage-Notu",
         added: "2009-07-29"}).
iana(language,nov,tag{
         type: "language",
         subtag: "nov",
         description: "Novial",
         added: "2009-07-29"}).
iana(language,now,tag{
         type: "language",
         subtag: "now",
         description: "Nyambo",
         added: "2009-07-29"}).
iana(language,noy,tag{
         type: "language",
         subtag: "noy",
         description: "Noy",
         added: "2009-07-29"}).
iana(language,noz,tag{
         type: "language",
         subtag: "noz",
         description: "Nayi",
         added: "2009-07-29"}).
iana(language,npa,tag{
         type: "language",
         subtag: "npa",
         description: "Nar Phu",
         added: "2009-07-29"}).
iana(language,npb,tag{
         type: "language",
         subtag: "npb",
         description: "Nupbikha",
         added: "2009-07-29"}).
iana(language,npg,tag{
         type: "language",
         subtag: "npg",
         description: "Ponyo-Gongwang Naga",
         added: "2012-08-12"}).
iana(language,nph,tag{
         type: "language",
         subtag: "nph",
         description: "Phom Naga",
         added: "2009-07-29"}).
iana(language,npi,tag{
         type: "language",
         subtag: "npi",
         description: "Nepali (individual language)",
         added: "2012-08-12",
         macrolanguage: "ne"}).
iana(language,npl,tag{
         type: "language",
         subtag: "npl",
         description: "Southeastern Puebla Nahuatl",
         added: "2009-07-29"}).
iana(language,npn,tag{
         type: "language",
         subtag: "npn",
         description: "Mondropolon",
         added: "2009-07-29"}).
iana(language,npo,tag{
         type: "language",
         subtag: "npo",
         description: "Pochuri Naga",
         added: "2009-07-29"}).
iana(language,nps,tag{
         type: "language",
         subtag: "nps",
         description: "Nipsan",
         added: "2009-07-29"}).
iana(language,npu,tag{
         type: "language",
         subtag: "npu",
         description: "Puimei Naga",
         added: "2009-07-29"}).
iana(language,npx,tag{
         type: "language",
         subtag: "npx",
         description: "Noipx",
         added: "2017-02-23"}).
iana(language,npy,tag{
         type: "language",
         subtag: "npy",
         description: "Napu",
         added: "2009-07-29"}).
iana(language,nqg,tag{
         type: "language",
         subtag: "nqg",
         description: "Southern Nago",
         added: "2009-07-29"}).
iana(language,nqk,tag{
         type: "language",
         subtag: "nqk",
         description: "Kura Ede Nago",
         added: "2009-07-29"}).
iana(language,nql,tag{
         type: "language",
         subtag: "nql",
         description: "Ngendelengo",
         added: "2017-02-23"}).
iana(language,nqm,tag{
         type: "language",
         subtag: "nqm",
         description: "Ndom",
         added: "2009-07-29"}).
iana(language,nqn,tag{
         type: "language",
         subtag: "nqn",
         description: "Nen",
         added: "2009-07-29"}).
iana(language,nqo,tag{
         type: "language",
         subtag: "nqo",
         description: ["N'Ko","N’Ko"],
         added: "2006-06-05",
         suppress_script: "Nkoo"}).
iana(language,nqq,tag{
         type: "language",
         subtag: "nqq",
         description: "Kyan-Karyaw Naga",
         added: "2013-09-10"}).
iana(language,nqt,tag{
         type: "language",
         subtag: "nqt",
         description: "Nteng",
         added: "2021-02-20"}).
iana(language,nqy,tag{
         type: "language",
         subtag: "nqy",
         description: "Akyaung Ari Naga",
         added: "2012-08-12"}).
iana(language,nra,tag{
         type: "language",
         subtag: "nra",
         description: "Ngom",
         added: "2009-07-29"}).
iana(language,nrb,tag{
         type: "language",
         subtag: "nrb",
         description: "Nara",
         added: "2009-07-29"}).
iana(language,nrc,tag{
         type: "language",
         subtag: "nrc",
         description: "Noric",
         added: "2009-07-29"}).
iana(language,nre,tag{
         type: "language",
         subtag: "nre",
         description: "Southern Rengma Naga",
         added: "2009-07-29"}).
iana(language,nrf,tag{
         type: "language",
         subtag: "nrf",
         description: ["Jèrriais","Guernésiais"],
         added: "2015-02-12"}).
iana(language,nrg,tag{
         type: "language",
         subtag: "nrg",
         description: "Narango",
         added: "2009-07-29"}).
iana(language,nri,tag{
         type: "language",
         subtag: "nri",
         description: "Chokri Naga",
         added: "2009-07-29"}).
iana(language,nrk,tag{
         type: "language",
         subtag: "nrk",
         description: "Ngarla",
         added: "2013-09-10"}).
iana(language,nrl,tag{
         type: "language",
         subtag: "nrl",
         description: "Ngarluma",
         added: "2009-07-29"}).
iana(language,nrm,tag{
         type: "language",
         subtag: "nrm",
         description: "Narom",
         added: "2009-07-29"}).
iana(language,nrn,tag{
         type: "language",
         subtag: "nrn",
         description: "Norn",
         added: "2009-07-29"}).
iana(language,nrp,tag{
         type: "language",
         subtag: "nrp",
         description: "North Picene",
         added: "2009-07-29"}).
iana(language,nrr,tag{
         type: "language",
         subtag: "nrr",
         description: ["Norra","Nora"],
         added: "2009-07-29"}).
iana(language,nrt,tag{
         type: "language",
         subtag: "nrt",
         description: "Northern Kalapuya",
         added: "2009-07-29"}).
iana(language,nru,tag{
         type: "language",
         subtag: "nru",
         description: "Narua",
         added: "2011-08-16"}).
iana(language,nrx,tag{
         type: "language",
         subtag: "nrx",
         description: "Ngurmbur",
         added: "2009-07-29"}).
iana(language,nrz,tag{
         type: "language",
         subtag: "nrz",
         description: "Lala",
         added: "2009-07-29"}).
iana(language,nsa,tag{
         type: "language",
         subtag: "nsa",
         description: "Sangtam Naga",
         added: "2009-07-29"}).
iana(language,nsb,tag{
         type: "language",
         subtag: "nsb",
         description: "Lower Nossob",
         added: "2020-03-28"}).
iana(language,nsc,tag{
         type: "language",
         subtag: "nsc",
         description: "Nshi",
         added: "2009-07-29"}).
iana(language,nsd,tag{
         type: "language",
         subtag: "nsd",
         description: "Southern Nisu",
         added: "2009-07-29"}).
iana(language,nse,tag{
         type: "language",
         subtag: "nse",
         description: "Nsenga",
         added: "2009-07-29"}).
iana(language,nsf,tag{
         type: "language",
         subtag: "nsf",
         description: "Northwestern Nisu",
         added: "2012-08-12"}).
iana(language,nsg,tag{
         type: "language",
         subtag: "nsg",
         description: "Ngasa",
         added: "2009-07-29"}).
iana(language,nsh,tag{
         type: "language",
         subtag: "nsh",
         description: "Ngoshie",
         added: "2009-07-29"}).
iana(language,nsi,tag{
         type: "language",
         subtag: "nsi",
         description: "Nigerian Sign Language",
         added: "2009-07-29"}).
iana(language,nsk,tag{
         type: "language",
         subtag: "nsk",
         description: "Naskapi",
         added: "2009-07-29"}).
iana(language,nsl,tag{
         type: "language",
         subtag: "nsl",
         description: "Norwegian Sign Language",
         added: "2009-07-29"}).
iana(language,nsm,tag{
         type: "language",
         subtag: "nsm",
         description: "Sumi Naga",
         added: "2009-07-29"}).
iana(language,nsn,tag{
         type: "language",
         subtag: "nsn",
         description: "Nehan",
         added: "2009-07-29"}).
iana(language,nso,tag{
         type: "language",
         subtag: "nso",
         description: ["Pedi","Northern Sotho","Sepedi"],
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,nsp,tag{
         type: "language",
         subtag: "nsp",
         description: "Nepalese Sign Language",
         added: "2009-07-29"}).
iana(language,nsq,tag{
         type: "language",
         subtag: "nsq",
         description: "Northern Sierra Miwok",
         added: "2009-07-29"}).
iana(language,nsr,tag{
         type: "language",
         subtag: "nsr",
         description: "Maritime Sign Language",
         added: "2009-07-29"}).
iana(language,nss,tag{
         type: "language",
         subtag: "nss",
         description: "Nali",
         added: "2009-07-29"}).
iana(language,nst,tag{
         type: "language",
         subtag: "nst",
         description: "Tase Naga",
         added: "2009-07-29"}).
iana(language,nsu,tag{
         type: "language",
         subtag: "nsu",
         description: "Sierra Negra Nahuatl",
         added: "2009-07-29"}).
iana(language,nsv,tag{
         type: "language",
         subtag: "nsv",
         description: "Southwestern Nisu",
         added: "2009-07-29"}).
iana(language,nsw,tag{
         type: "language",
         subtag: "nsw",
         description: "Navut",
         added: "2009-07-29"}).
iana(language,nsx,tag{
         type: "language",
         subtag: "nsx",
         description: "Nsongo",
         added: "2009-07-29"}).
iana(language,nsy,tag{
         type: "language",
         subtag: "nsy",
         description: "Nasal",
         added: "2009-07-29"}).
iana(language,nsz,tag{
         type: "language",
         subtag: "nsz",
         description: "Nisenan",
         added: "2009-07-29"}).
iana(language,ntd,tag{
         type: "language",
         subtag: "ntd",
         description: "Northern Tidung",
         added: "2016-05-30"}).
iana(language,nte,tag{
         type: "language",
         subtag: "nte",
         description: "Nathembo",
         added: "2009-07-29"}).
iana(language,ntg,tag{
         type: "language",
         subtag: "ntg",
         description: "Ngantangarra",
         added: "2013-09-10"}).
iana(language,nti,tag{
         type: "language",
         subtag: "nti",
         description: "Natioro",
         added: "2009-07-29"}).
iana(language,ntj,tag{
         type: "language",
         subtag: "ntj",
         description: "Ngaanyatjarra",
         added: "2009-07-29"}).
iana(language,ntk,tag{
         type: "language",
         subtag: "ntk",
         description: "Ikoma-Nata-Isenye",
         added: "2009-07-29"}).
iana(language,ntm,tag{
         type: "language",
         subtag: "ntm",
         description: "Nateni",
         added: "2009-07-29"}).
iana(language,nto,tag{
         type: "language",
         subtag: "nto",
         description: "Ntomba",
         added: "2009-07-29"}).
iana(language,ntp,tag{
         type: "language",
         subtag: "ntp",
         description: "Northern Tepehuan",
         added: "2009-07-29"}).
iana(language,ntr,tag{
         type: "language",
         subtag: "ntr",
         description: "Delo",
         added: "2009-07-29"}).
iana(language,nts,tag{
         type: "language",
         subtag: "nts",
         description: "Natagaimas",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "pij"}).
iana(language,ntu,tag{
         type: "language",
         subtag: "ntu",
         description: "Natügu",
         added: "2009-07-29"}).
iana(language,ntw,tag{
         type: "language",
         subtag: "ntw",
         description: "Nottoway",
         added: "2009-07-29"}).
iana(language,ntx,tag{
         type: "language",
         subtag: "ntx",
         description: "Tangkhul Naga (Myanmar)",
         added: "2012-08-12",
         comments: "see nmf"}).
iana(language,nty,tag{
         type: "language",
         subtag: "nty",
         description: "Mantsi",
         added: "2009-07-29"}).
iana(language,ntz,tag{
         type: "language",
         subtag: "ntz",
         description: "Natanzi",
         added: "2009-07-29"}).
iana(language,nua,tag{
         type: "language",
         subtag: "nua",
         description: "Yuanga",
         added: "2009-07-29"}).
iana(language,nub,tag{
         type: "language",
         subtag: "nub",
         description: "Nubian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,nuc,tag{
         type: "language",
         subtag: "nuc",
         description: "Nukuini",
         added: "2009-07-29"}).
iana(language,nud,tag{
         type: "language",
         subtag: "nud",
         description: "Ngala",
         added: "2009-07-29"}).
iana(language,nue,tag{
         type: "language",
         subtag: "nue",
         description: "Ngundu",
         added: "2009-07-29"}).
iana(language,nuf,tag{
         type: "language",
         subtag: "nuf",
         description: "Nusu",
         added: "2009-07-29"}).
iana(language,nug,tag{
         type: "language",
         subtag: "nug",
         description: "Nungali",
         added: "2009-07-29"}).
iana(language,nuh,tag{
         type: "language",
         subtag: "nuh",
         description: "Ndunda",
         added: "2009-07-29"}).
iana(language,nui,tag{
         type: "language",
         subtag: "nui",
         description: "Ngumbi",
         added: "2009-07-29"}).
iana(language,nuj,tag{
         type: "language",
         subtag: "nuj",
         description: "Nyole",
         added: "2009-07-29"}).
iana(language,nuk,tag{
         type: "language",
         subtag: "nuk",
         description: ["Nuu-chah-nulth","Nuuchahnulth"],
         added: "2011-08-16"}).
iana(language,nul,tag{
         type: "language",
         subtag: "nul",
         description: "Nusa Laut",
         added: "2009-07-29"}).
iana(language,num,tag{
         type: "language",
         subtag: "num",
         description: "Niuafo'ou",
         added: "2009-07-29"}).
iana(language,nun,tag{
         type: "language",
         subtag: "nun",
         description: "Anong",
         added: "2009-07-29"}).
iana(language,nuo,tag{
         type: "language",
         subtag: "nuo",
         description: "Nguôn",
         added: "2009-07-29"}).
iana(language,nup,tag{
         type: "language",
         subtag: "nup",
         description: "Nupe-Nupe-Tako",
         added: "2009-07-29"}).
iana(language,nuq,tag{
         type: "language",
         subtag: "nuq",
         description: "Nukumanu",
         added: "2009-07-29"}).
iana(language,nur,tag{
         type: "language",
         subtag: "nur",
         description: "Nukuria",
         added: "2009-07-29"}).
iana(language,nus,tag{
         type: "language",
         subtag: "nus",
         description: "Nuer",
         added: "2009-07-29"}).
iana(language,nut,tag{
         type: "language",
         subtag: "nut",
         description: "Nung (Viet Nam)",
         added: "2009-07-29"}).
iana(language,nuu,tag{
         type: "language",
         subtag: "nuu",
         description: "Ngbundu",
         added: "2009-07-29"}).
iana(language,nuv,tag{
         type: "language",
         subtag: "nuv",
         description: "Northern Nuni",
         added: "2009-07-29"}).
iana(language,nuw,tag{
         type: "language",
         subtag: "nuw",
         description: "Nguluwan",
         added: "2009-07-29"}).
iana(language,nux,tag{
         type: "language",
         subtag: "nux",
         description: "Mehek",
         added: "2009-07-29"}).
iana(language,nuy,tag{
         type: "language",
         subtag: "nuy",
         description: "Nunggubuyu",
         added: "2009-07-29"}).
iana(language,nuz,tag{
         type: "language",
         subtag: "nuz",
         description: "Tlamacazapa Nahuatl",
         added: "2009-07-29"}).
iana(language,nvh,tag{
         type: "language",
         subtag: "nvh",
         description: "Nasarian",
         added: "2009-07-29"}).
iana(language,nvm,tag{
         type: "language",
         subtag: "nvm",
         description: "Namiae",
         added: "2009-07-29"}).
iana(language,nvo,tag{
         type: "language",
         subtag: "nvo",
         description: "Nyokon",
         added: "2012-08-12"}).
iana(language,nwa,tag{
         type: "language",
         subtag: "nwa",
         description: "Nawathinehena",
         added: "2009-07-29"}).
iana(language,nwb,tag{
         type: "language",
         subtag: "nwb",
         description: "Nyabwa",
         added: "2009-07-29"}).
iana(language,nwc,tag{
         type: "language",
         subtag: "nwc",
         description: ["Classical Newari","Classical Nepal Bhasa","Old Newari"],
         added: "2005-10-16"}).
iana(language,nwe,tag{
         type: "language",
         subtag: "nwe",
         description: "Ngwe",
         added: "2009-07-29"}).
iana(language,nwg,tag{
         type: "language",
         subtag: "nwg",
         description: "Ngayawung",
         added: "2012-08-12"}).
iana(language,nwi,tag{
         type: "language",
         subtag: "nwi",
         description: "Southwest Tanna",
         added: "2009-07-29"}).
iana(language,nwm,tag{
         type: "language",
         subtag: "nwm",
         description: "Nyamusa-Molo",
         added: "2009-07-29"}).
iana(language,nwo,tag{
         type: "language",
         subtag: "nwo",
         description: "Nauo",
         added: "2012-08-12"}).
iana(language,nwr,tag{
         type: "language",
         subtag: "nwr",
         description: "Nawaru",
         added: "2009-07-29"}).
iana(language,nww,tag{
         type: "language",
         subtag: "nww",
         description: "Ndwewe",
         added: "2022-02-25"}).
iana(language,nwx,tag{
         type: "language",
         subtag: "nwx",
         description: "Middle Newar",
         added: "2009-07-29"}).
iana(language,nwy,tag{
         type: "language",
         subtag: "nwy",
         description: "Nottoway-Meherrin",
         added: "2009-07-29"}).
iana(language,nxa,tag{
         type: "language",
         subtag: "nxa",
         description: "Nauete",
         added: "2009-07-29"}).
iana(language,nxd,tag{
         type: "language",
         subtag: "nxd",
         description: "Ngando (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,nxe,tag{
         type: "language",
         subtag: "nxe",
         description: "Nage",
         added: "2009-07-29"}).
iana(language,nxg,tag{
         type: "language",
         subtag: "nxg",
         description: "Ngad'a",
         added: "2009-07-29"}).
iana(language,nxi,tag{
         type: "language",
         subtag: "nxi",
         description: "Nindi",
         added: "2009-07-29"}).
iana(language,nxk,tag{
         type: "language",
         subtag: "nxk",
         description: "Koki Naga",
         added: "2012-08-12"}).
iana(language,nxl,tag{
         type: "language",
         subtag: "nxl",
         description: "South Nuaulu",
         added: "2009-07-29"}).
iana(language,nxm,tag{
         type: "language",
         subtag: "nxm",
         description: "Numidian",
         added: "2009-07-29"}).
iana(language,nxn,tag{
         type: "language",
         subtag: "nxn",
         description: "Ngawun",
         added: "2009-07-29"}).
iana(language,nxo,tag{
         type: "language",
         subtag: "nxo",
         description: "Ndambomo",
         added: "2015-02-12"}).
iana(language,nxq,tag{
         type: "language",
         subtag: "nxq",
         description: "Naxi",
         added: "2011-08-16"}).
iana(language,nxr,tag{
         type: "language",
         subtag: "nxr",
         description: "Ninggerum",
         added: "2009-07-29"}).
iana(language,nxu,tag{
         type: "language",
         subtag: "nxu",
         description: "Narau",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "bpp"}).
iana(language,nxx,tag{
         type: "language",
         subtag: "nxx",
         description: "Nafri",
         added: "2009-07-29"}).
iana(language,nyb,tag{
         type: "language",
         subtag: "nyb",
         description: "Nyangbo",
         added: "2009-07-29"}).
iana(language,nyc,tag{
         type: "language",
         subtag: "nyc",
         description: "Nyanga-li",
         added: "2009-07-29"}).
iana(language,nyd,tag{
         type: "language",
         subtag: "nyd",
         description: ["Nyore","Olunyole"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,nye,tag{
         type: "language",
         subtag: "nye",
         description: "Nyengo",
         added: "2009-07-29"}).
iana(language,nyf,tag{
         type: "language",
         subtag: "nyf",
         description: ["Giryama","Kigiryama"],
         added: "2009-07-29"}).
iana(language,nyg,tag{
         type: "language",
         subtag: "nyg",
         description: "Nyindu",
         added: "2009-07-29"}).
iana(language,nyh,tag{
         type: "language",
         subtag: "nyh",
         description: "Nyikina",
         added: "2009-07-29"}).
iana(language,nyi,tag{
         type: "language",
         subtag: "nyi",
         description: "Ama (Sudan)",
         added: "2009-07-29"}).
iana(language,nyj,tag{
         type: "language",
         subtag: "nyj",
         description: "Nyanga",
         added: "2009-07-29"}).
iana(language,nyk,tag{
         type: "language",
         subtag: "nyk",
         description: "Nyaneka",
         added: "2009-07-29"}).
iana(language,nyl,tag{
         type: "language",
         subtag: "nyl",
         description: "Nyeu",
         added: "2009-07-29"}).
iana(language,nym,tag{
         type: "language",
         subtag: "nym",
         description: "Nyamwezi",
         added: "2005-10-16"}).
iana(language,nyn,tag{
         type: "language",
         subtag: "nyn",
         description: "Nyankole",
         added: "2005-10-16"}).
iana(language,nyo,tag{
         type: "language",
         subtag: "nyo",
         description: "Nyoro",
         added: "2005-10-16"}).
iana(language,nyp,tag{
         type: "language",
         subtag: "nyp",
         description: "Nyang'i",
         added: "2009-07-29"}).
iana(language,nyq,tag{
         type: "language",
         subtag: "nyq",
         description: "Nayini",
         added: "2009-07-29"}).
iana(language,nyr,tag{
         type: "language",
         subtag: "nyr",
         description: "Nyiha (Malawi)",
         added: "2009-07-29"}).
iana(language,nys,tag{
         type: "language",
         subtag: "nys",
         description: "Nyungar",
         added: "2009-07-29"}).
iana(language,nyt,tag{
         type: "language",
         subtag: "nyt",
         description: "Nyawaygi",
         added: "2009-07-29"}).
iana(language,nyu,tag{
         type: "language",
         subtag: "nyu",
         description: "Nyungwe",
         added: "2009-07-29"}).
iana(language,nyv,tag{
         type: "language",
         subtag: "nyv",
         description: "Nyulnyul",
         added: "2009-07-29"}).
iana(language,nyw,tag{
         type: "language",
         subtag: "nyw",
         description: "Nyaw",
         added: "2009-07-29"}).
iana(language,nyx,tag{
         type: "language",
         subtag: "nyx",
         description: "Nganyaywana",
         added: "2009-07-29"}).
iana(language,nyy,tag{
         type: "language",
         subtag: "nyy",
         description: "Nyakyusa-Ngonde",
         added: "2009-07-29"}).
iana(language,nza,tag{
         type: "language",
         subtag: "nza",
         description: "Tigon Mbembe",
         added: "2009-07-29"}).
iana(language,nzb,tag{
         type: "language",
         subtag: "nzb",
         description: "Njebi",
         added: "2009-07-29"}).
iana(language,nzd,tag{
         type: "language",
         subtag: "nzd",
         description: "Nzadi",
         added: "2018-03-08"}).
iana(language,nzi,tag{
         type: "language",
         subtag: "nzi",
         description: "Nzima",
         added: "2005-10-16"}).
iana(language,nzk,tag{
         type: "language",
         subtag: "nzk",
         description: "Nzakara",
         added: "2009-07-29"}).
iana(language,nzm,tag{
         type: "language",
         subtag: "nzm",
         description: "Zeme Naga",
         added: "2009-07-29"}).
iana(language,nzs,tag{
         type: "language",
         subtag: "nzs",
         description: "New Zealand Sign Language",
         added: "2009-07-29"}).
iana(language,nzu,tag{
         type: "language",
         subtag: "nzu",
         description: "Teke-Nzikou",
         added: "2009-07-29"}).
iana(language,nzy,tag{
         type: "language",
         subtag: "nzy",
         description: "Nzakambay",
         added: "2009-07-29"}).
iana(language,nzz,tag{
         type: "language",
         subtag: "nzz",
         description: "Nanga Dama Dogon",
         added: "2010-03-11"}).
iana(language,oaa,tag{
         type: "language",
         subtag: "oaa",
         description: "Orok",
         added: "2009-07-29"}).
iana(language,oac,tag{
         type: "language",
         subtag: "oac",
         description: "Oroch",
         added: "2009-07-29"}).
iana(language,oar,tag{
         type: "language",
         subtag: "oar",
         description: ["Old Aramaic (up to 700 BCE)","Ancient Aramaic (up to 700 BCE)"],
         added: "2009-07-29"}).
iana(language,oav,tag{
         type: "language",
         subtag: "oav",
         description: "Old Avar",
         added: "2009-07-29"}).
iana(language,obi,tag{
         type: "language",
         subtag: "obi",
         description: "Obispeño",
         added: "2009-07-29"}).
iana(language,obk,tag{
         type: "language",
         subtag: "obk",
         description: "Southern Bontok",
         added: "2010-03-11",
         macrolanguage: "bnc"}).
iana(language,obl,tag{
         type: "language",
         subtag: "obl",
         description: "Oblo",
         added: "2009-07-29"}).
iana(language,obm,tag{
         type: "language",
         subtag: "obm",
         description: "Moabite",
         added: "2009-07-29"}).
iana(language,obo,tag{
         type: "language",
         subtag: "obo",
         description: "Obo Manobo",
         added: "2009-07-29"}).
iana(language,obr,tag{
         type: "language",
         subtag: "obr",
         description: "Old Burmese",
         added: "2009-07-29"}).
iana(language,obt,tag{
         type: "language",
         subtag: "obt",
         description: "Old Breton",
         added: "2009-07-29"}).
iana(language,obu,tag{
         type: "language",
         subtag: "obu",
         description: "Obulom",
         added: "2009-07-29"}).
iana(language,oca,tag{
         type: "language",
         subtag: "oca",
         description: "Ocaina",
         added: "2009-07-29"}).
iana(language,och,tag{
         type: "language",
         subtag: "och",
         description: "Old Chinese",
         added: "2009-07-29"}).
iana(language,ocm,tag{
         type: "language",
         subtag: "ocm",
         description: "Old Cham",
         added: "2021-02-20"}).
iana(language,oco,tag{
         type: "language",
         subtag: "oco",
         description: "Old Cornish",
         added: "2009-07-29"}).
iana(language,ocu,tag{
         type: "language",
         subtag: "ocu",
         description: "Atzingo Matlatzinca",
         added: "2009-07-29"}).
iana(language,oda,tag{
         type: "language",
         subtag: "oda",
         description: "Odut",
         added: "2009-07-29"}).
iana(language,odk,tag{
         type: "language",
         subtag: "odk",
         description: "Od",
         added: "2009-07-29"}).
iana(language,odt,tag{
         type: "language",
         subtag: "odt",
         description: "Old Dutch",
         added: "2009-07-29"}).
iana(language,odu,tag{
         type: "language",
         subtag: "odu",
         description: "Odual",
         added: "2009-07-29"}).
iana(language,ofo,tag{
         type: "language",
         subtag: "ofo",
         description: "Ofo",
         added: "2009-07-29"}).
iana(language,ofs,tag{
         type: "language",
         subtag: "ofs",
         description: "Old Frisian",
         added: "2009-07-29"}).
iana(language,ofu,tag{
         type: "language",
         subtag: "ofu",
         description: "Efutop",
         added: "2009-07-29"}).
iana(language,ogb,tag{
         type: "language",
         subtag: "ogb",
         description: "Ogbia",
         added: "2009-07-29"}).
iana(language,ogc,tag{
         type: "language",
         subtag: "ogc",
         description: "Ogbah",
         added: "2009-07-29"}).
iana(language,oge,tag{
         type: "language",
         subtag: "oge",
         description: "Old Georgian",
         added: "2009-07-29"}).
iana(language,ogg,tag{
         type: "language",
         subtag: "ogg",
         description: "Ogbogolo",
         added: "2009-07-29"}).
iana(language,ogo,tag{
         type: "language",
         subtag: "ogo",
         description: "Khana",
         added: "2009-07-29"}).
iana(language,ogu,tag{
         type: "language",
         subtag: "ogu",
         description: "Ogbronuagum",
         added: "2009-07-29"}).
iana(language,oht,tag{
         type: "language",
         subtag: "oht",
         description: "Old Hittite",
         added: "2009-07-29"}).
iana(language,ohu,tag{
         type: "language",
         subtag: "ohu",
         description: "Old Hungarian",
         added: "2009-07-29"}).
iana(language,oia,tag{
         type: "language",
         subtag: "oia",
         description: "Oirata",
         added: "2009-07-29"}).
iana(language,oie,tag{
         type: "language",
         subtag: "oie",
         description: "Okolie",
         added: "2022-02-25"}).
iana(language,oin,tag{
         type: "language",
         subtag: "oin",
         description: "Inebu One",
         added: "2009-07-29"}).
iana(language,ojb,tag{
         type: "language",
         subtag: "ojb",
         description: "Northwestern Ojibwa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,ojc,tag{
         type: "language",
         subtag: "ojc",
         description: "Central Ojibwa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,ojg,tag{
         type: "language",
         subtag: "ojg",
         description: "Eastern Ojibwa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,ojp,tag{
         type: "language",
         subtag: "ojp",
         description: "Old Japanese",
         added: "2009-07-29"}).
iana(language,ojs,tag{
         type: "language",
         subtag: "ojs",
         description: "Severn Ojibwa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,ojv,tag{
         type: "language",
         subtag: "ojv",
         description: "Ontong Java",
         added: "2009-07-29"}).
iana(language,ojw,tag{
         type: "language",
         subtag: "ojw",
         description: "Western Ojibwa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,oka,tag{
         type: "language",
         subtag: "oka",
         description: "Okanagan",
         added: "2009-07-29"}).
iana(language,okb,tag{
         type: "language",
         subtag: "okb",
         description: "Okobo",
         added: "2009-07-29"}).
iana(language,okc,tag{
         type: "language",
         subtag: "okc",
         description: "Kobo",
         added: "2021-02-20"}).
iana(language,okd,tag{
         type: "language",
         subtag: "okd",
         description: "Okodia",
         added: "2009-07-29"}).
iana(language,oke,tag{
         type: "language",
         subtag: "oke",
         description: "Okpe (Southwestern Edo)",
         added: "2009-07-29"}).
iana(language,okg,tag{
         type: "language",
         subtag: "okg",
         description: "Koko Babangk",
         added: "2012-08-12"}).
iana(language,okh,tag{
         type: "language",
         subtag: "okh",
         description: "Koresh-e Rostam",
         added: "2009-07-29"}).
iana(language,oki,tag{
         type: "language",
         subtag: "oki",
         description: "Okiek",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,okj,tag{
         type: "language",
         subtag: "okj",
         description: "Oko-Juwoi",
         added: "2009-07-29"}).
iana(language,okk,tag{
         type: "language",
         subtag: "okk",
         description: "Kwamtim One",
         added: "2009-07-29"}).
iana(language,okl,tag{
         type: "language",
         subtag: "okl",
         description: "Old Kentish Sign Language",
         added: "2009-07-29"}).
iana(language,okm,tag{
         type: "language",
         subtag: "okm",
         description: "Middle Korean (10th-16th cent.)",
         added: "2009-07-29"}).
iana(language,okn,tag{
         type: "language",
         subtag: "okn",
         description: "Oki-No-Erabu",
         added: "2009-07-29"}).
iana(language,oko,tag{
         type: "language",
         subtag: "oko",
         description: "Old Korean (3rd-9th cent.)",
         added: "2009-07-29"}).
iana(language,okr,tag{
         type: "language",
         subtag: "okr",
         description: "Kirike",
         added: "2009-07-29"}).
iana(language,oks,tag{
         type: "language",
         subtag: "oks",
         description: "Oko-Eni-Osayen",
         added: "2009-07-29"}).
iana(language,oku,tag{
         type: "language",
         subtag: "oku",
         description: "Oku",
         added: "2009-07-29"}).
iana(language,okv,tag{
         type: "language",
         subtag: "okv",
         description: "Orokaiva",
         added: "2009-07-29"}).
iana(language,okx,tag{
         type: "language",
         subtag: "okx",
         description: "Okpe (Northwestern Edo)",
         added: "2009-07-29"}).
iana(language,okz,tag{
         type: "language",
         subtag: "okz",
         description: "Old Khmer",
         added: "2021-02-20"}).
iana(language,ola,tag{
         type: "language",
         subtag: "ola",
         description: "Walungge",
         added: "2009-07-29"}).
iana(language,old,tag{
         type: "language",
         subtag: "old",
         description: "Mochi",
         added: "2009-07-29"}).
iana(language,ole,tag{
         type: "language",
         subtag: "ole",
         description: "Olekha",
         added: "2009-07-29"}).
iana(language,olk,tag{
         type: "language",
         subtag: "olk",
         description: "Olkol",
         added: "2013-09-10"}).
iana(language,olm,tag{
         type: "language",
         subtag: "olm",
         description: "Oloma",
         added: "2009-07-29"}).
iana(language,olo,tag{
         type: "language",
         subtag: "olo",
         description: "Livvi",
         added: "2009-07-29"}).
iana(language,olr,tag{
         type: "language",
         subtag: "olr",
         description: "Olrat",
         added: "2009-07-29"}).
iana(language,olt,tag{
         type: "language",
         subtag: "olt",
         description: "Old Lithuanian",
         added: "2014-02-28"}).
iana(language,olu,tag{
         type: "language",
         subtag: "olu",
         description: "Kuvale",
         added: "2016-05-30"}).
iana(language,oma,tag{
         type: "language",
         subtag: "oma",
         description: "Omaha-Ponca",
         added: "2009-07-29"}).
iana(language,omb,tag{
         type: "language",
         subtag: "omb",
         description: "East Ambae",
         added: "2009-07-29"}).
iana(language,omc,tag{
         type: "language",
         subtag: "omc",
         description: "Mochica",
         added: "2009-07-29"}).
iana(language,ome,tag{
         type: "language",
         subtag: "ome",
         description: "Omejes",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,omg,tag{
         type: "language",
         subtag: "omg",
         description: "Omagua",
         added: "2009-07-29"}).
iana(language,omi,tag{
         type: "language",
         subtag: "omi",
         description: "Omi",
         added: "2009-07-29"}).
iana(language,omk,tag{
         type: "language",
         subtag: "omk",
         description: "Omok",
         added: "2009-07-29"}).
iana(language,oml,tag{
         type: "language",
         subtag: "oml",
         description: "Ombo",
         added: "2009-07-29"}).
iana(language,omn,tag{
         type: "language",
         subtag: "omn",
         description: "Minoan",
         added: "2009-07-29"}).
iana(language,omo,tag{
         type: "language",
         subtag: "omo",
         description: "Utarmbung",
         added: "2009-07-29"}).
iana(language,omp,tag{
         type: "language",
         subtag: "omp",
         description: "Old Manipuri",
         added: "2009-07-29"}).
iana(language,omq,tag{
         type: "language",
         subtag: "omq",
         description: "Oto-Manguean languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,omr,tag{
         type: "language",
         subtag: "omr",
         description: "Old Marathi",
         added: "2009-07-29"}).
iana(language,omt,tag{
         type: "language",
         subtag: "omt",
         description: "Omotik",
         added: "2009-07-29"}).
iana(language,omu,tag{
         type: "language",
         subtag: "omu",
         description: "Omurano",
         added: "2009-07-29"}).
iana(language,omv,tag{
         type: "language",
         subtag: "omv",
         description: "Omotic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,omw,tag{
         type: "language",
         subtag: "omw",
         description: "South Tairora",
         added: "2009-07-29"}).
iana(language,omx,tag{
         type: "language",
         subtag: "omx",
         description: "Old Mon",
         added: "2009-07-29"}).
iana(language,omy,tag{
         type: "language",
         subtag: "omy",
         description: "Old Malay",
         added: "2021-02-20"}).
iana(language,ona,tag{
         type: "language",
         subtag: "ona",
         description: "Ona",
         added: "2009-07-29"}).
iana(language,onb,tag{
         type: "language",
         subtag: "onb",
         description: "Lingao",
         added: "2009-07-29"}).
iana(language,one,tag{
         type: "language",
         subtag: "one",
         description: "Oneida",
         added: "2009-07-29"}).
iana(language,ong,tag{
         type: "language",
         subtag: "ong",
         description: "Olo",
         added: "2009-07-29"}).
iana(language,oni,tag{
         type: "language",
         subtag: "oni",
         description: "Onin",
         added: "2009-07-29"}).
iana(language,onj,tag{
         type: "language",
         subtag: "onj",
         description: "Onjob",
         added: "2009-07-29"}).
iana(language,onk,tag{
         type: "language",
         subtag: "onk",
         description: "Kabore One",
         added: "2009-07-29"}).
iana(language,onn,tag{
         type: "language",
         subtag: "onn",
         description: "Onobasulu",
         added: "2009-07-29"}).
iana(language,ono,tag{
         type: "language",
         subtag: "ono",
         description: "Onondaga",
         added: "2009-07-29"}).
iana(language,onp,tag{
         type: "language",
         subtag: "onp",
         description: "Sartang",
         added: "2009-07-29"}).
iana(language,onr,tag{
         type: "language",
         subtag: "onr",
         description: "Northern One",
         added: "2009-07-29"}).
iana(language,ons,tag{
         type: "language",
         subtag: "ons",
         description: "Ono",
         added: "2009-07-29"}).
iana(language,ont,tag{
         type: "language",
         subtag: "ont",
         description: "Ontenu",
         added: "2009-07-29"}).
iana(language,onu,tag{
         type: "language",
         subtag: "onu",
         description: "Unua",
         added: "2009-07-29"}).
iana(language,onw,tag{
         type: "language",
         subtag: "onw",
         description: "Old Nubian",
         added: "2009-07-29"}).
iana(language,onx,tag{
         type: "language",
         subtag: "onx",
         description: "Onin Based Pidgin",
         added: "2009-07-29"}).
iana(language,ood,tag{
         type: "language",
         subtag: "ood",
         description: "Tohono O'odham",
         added: "2009-07-29"}).
iana(language,oog,tag{
         type: "language",
         subtag: "oog",
         description: "Ong",
         added: "2009-07-29"}).
iana(language,oon,tag{
         type: "language",
         subtag: "oon",
         description: "Önge",
         added: "2009-07-29"}).
iana(language,oor,tag{
         type: "language",
         subtag: "oor",
         description: "Oorlams",
         added: "2009-07-29"}).
iana(language,oos,tag{
         type: "language",
         subtag: "oos",
         description: "Old Ossetic",
         added: "2009-07-29"}).
iana(language,opa,tag{
         type: "language",
         subtag: "opa",
         description: "Okpamheri",
         added: "2009-07-29"}).
iana(language,opk,tag{
         type: "language",
         subtag: "opk",
         description: "Kopkaka",
         added: "2009-07-29"}).
iana(language,opm,tag{
         type: "language",
         subtag: "opm",
         description: "Oksapmin",
         added: "2009-07-29"}).
iana(language,opo,tag{
         type: "language",
         subtag: "opo",
         description: "Opao",
         added: "2009-07-29"}).
iana(language,opt,tag{
         type: "language",
         subtag: "opt",
         description: "Opata",
         added: "2009-07-29"}).
iana(language,opy,tag{
         type: "language",
         subtag: "opy",
         description: "Ofayé",
         added: "2009-07-29"}).
iana(language,ora,tag{
         type: "language",
         subtag: "ora",
         description: "Oroha",
         added: "2009-07-29"}).
iana(language,orc,tag{
         type: "language",
         subtag: "orc",
         description: "Orma",
         added: "2009-07-29",
         macrolanguage: "om"}).
iana(language,ore,tag{
         type: "language",
         subtag: "ore",
         description: "Orejón",
         added: "2009-07-29"}).
iana(language,org,tag{
         type: "language",
         subtag: "org",
         description: "Oring",
         added: "2009-07-29"}).
iana(language,orh,tag{
         type: "language",
         subtag: "orh",
         description: "Oroqen",
         added: "2009-07-29"}).
iana(language,orn,tag{
         type: "language",
         subtag: "orn",
         description: "Orang Kanaq",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,oro,tag{
         type: "language",
         subtag: "oro",
         description: "Orokolo",
         added: "2009-07-29"}).
iana(language,orr,tag{
         type: "language",
         subtag: "orr",
         description: "Oruma",
         added: "2009-07-29"}).
iana(language,ors,tag{
         type: "language",
         subtag: "ors",
         description: "Orang Seletar",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,ort,tag{
         type: "language",
         subtag: "ort",
         description: "Adivasi Oriya",
         added: "2009-07-29"}).
iana(language,oru,tag{
         type: "language",
         subtag: "oru",
         description: "Ormuri",
         added: "2009-07-29"}).
iana(language,orv,tag{
         type: "language",
         subtag: "orv",
         description: "Old Russian",
         added: "2009-07-29"}).
iana(language,orw,tag{
         type: "language",
         subtag: "orw",
         description: "Oro Win",
         added: "2009-07-29"}).
iana(language,orx,tag{
         type: "language",
         subtag: "orx",
         description: "Oro",
         added: "2009-07-29"}).
iana(language,ory,tag{
         type: "language",
         subtag: "ory",
         description: ["Odia (individual language)","Oriya (individual language)"],
         added: "2012-08-12",
         macrolanguage: "or"}).
iana(language,orz,tag{
         type: "language",
         subtag: "orz",
         description: "Ormu",
         added: "2009-07-29"}).
iana(language,osa,tag{
         type: "language",
         subtag: "osa",
         description: "Osage",
         added: "2005-10-16"}).
iana(language,osc,tag{
         type: "language",
         subtag: "osc",
         description: "Oscan",
         added: "2009-07-29"}).
iana(language,osi,tag{
         type: "language",
         subtag: "osi",
         description: "Osing",
         added: "2009-07-29"}).
iana(language,osn,tag{
         type: "language",
         subtag: "osn",
         description: "Old Sundanese",
         added: "2021-02-20"}).
iana(language,oso,tag{
         type: "language",
         subtag: "oso",
         description: "Ososo",
         added: "2009-07-29"}).
iana(language,osp,tag{
         type: "language",
         subtag: "osp",
         description: "Old Spanish",
         added: "2009-07-29"}).
iana(language,ost,tag{
         type: "language",
         subtag: "ost",
         description: "Osatu",
         added: "2009-07-29"}).
iana(language,osu,tag{
         type: "language",
         subtag: "osu",
         description: "Southern One",
         added: "2009-07-29"}).
iana(language,osx,tag{
         type: "language",
         subtag: "osx",
         description: "Old Saxon",
         added: "2009-07-29"}).
iana(language,ota,tag{
         type: "language",
         subtag: "ota",
         description: "Ottoman Turkish (1500-1928)",
         added: "2005-10-16"}).
iana(language,otb,tag{
         type: "language",
         subtag: "otb",
         description: "Old Tibetan",
         added: "2009-07-29"}).
iana(language,otd,tag{
         type: "language",
         subtag: "otd",
         description: "Ot Danum",
         added: "2009-07-29"}).
iana(language,ote,tag{
         type: "language",
         subtag: "ote",
         description: "Mezquital Otomi",
         added: "2009-07-29"}).
iana(language,oti,tag{
         type: "language",
         subtag: "oti",
         description: "Oti",
         added: "2009-07-29"}).
iana(language,otk,tag{
         type: "language",
         subtag: "otk",
         description: "Old Turkish",
         added: "2009-07-29"}).
iana(language,otl,tag{
         type: "language",
         subtag: "otl",
         description: "Tilapa Otomi",
         added: "2009-07-29"}).
iana(language,otm,tag{
         type: "language",
         subtag: "otm",
         description: "Eastern Highland Otomi",
         added: "2009-07-29"}).
iana(language,otn,tag{
         type: "language",
         subtag: "otn",
         description: "Tenango Otomi",
         added: "2009-07-29"}).
iana(language,oto,tag{
         type: "language",
         subtag: "oto",
         description: "Otomian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,otq,tag{
         type: "language",
         subtag: "otq",
         description: "Querétaro Otomi",
         added: "2009-07-29"}).
iana(language,otr,tag{
         type: "language",
         subtag: "otr",
         description: "Otoro",
         added: "2009-07-29"}).
iana(language,ots,tag{
         type: "language",
         subtag: "ots",
         description: "Estado de México Otomi",
         added: "2009-07-29"}).
iana(language,ott,tag{
         type: "language",
         subtag: "ott",
         description: "Temoaya Otomi",
         added: "2009-07-29"}).
iana(language,otu,tag{
         type: "language",
         subtag: "otu",
         description: "Otuke",
         added: "2009-07-29"}).
iana(language,otw,tag{
         type: "language",
         subtag: "otw",
         description: "Ottawa",
         added: "2009-07-29",
         macrolanguage: "oj"}).
iana(language,otx,tag{
         type: "language",
         subtag: "otx",
         description: "Texcatepec Otomi",
         added: "2009-07-29"}).
iana(language,oty,tag{
         type: "language",
         subtag: "oty",
         description: "Old Tamil",
         added: "2009-07-29"}).
iana(language,otz,tag{
         type: "language",
         subtag: "otz",
         description: "Ixtenco Otomi",
         added: "2009-07-29"}).
iana(language,oua,tag{
         type: "language",
         subtag: "oua",
         description: "Tagargrent",
         added: "2009-07-29"}).
iana(language,oub,tag{
         type: "language",
         subtag: "oub",
         description: "Glio-Oubi",
         added: "2009-07-29"}).
iana(language,oue,tag{
         type: "language",
         subtag: "oue",
         description: "Oune",
         added: "2009-07-29"}).
iana(language,oui,tag{
         type: "language",
         subtag: "oui",
         description: "Old Uighur",
         added: "2009-07-29"}).
iana(language,oum,tag{
         type: "language",
         subtag: "oum",
         description: "Ouma",
         added: "2009-07-29"}).
iana(language,oun,tag{
         type: "language",
         subtag: "oun",
         description: "ǃOǃung",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "vaj"}).
iana(language,ovd,tag{
         type: "language",
         subtag: "ovd",
         description: ["Elfdalian","Övdalian"],
         added: "2016-06-16"}).
iana(language,owi,tag{
         type: "language",
         subtag: "owi",
         description: "Owiniga",
         added: "2009-07-29"}).
iana(language,owl,tag{
         type: "language",
         subtag: "owl",
         description: "Old Welsh",
         added: "2009-07-29"}).
iana(language,oyb,tag{
         type: "language",
         subtag: "oyb",
         description: "Oy",
         added: "2009-07-29"}).
iana(language,oyd,tag{
         type: "language",
         subtag: "oyd",
         description: "Oyda",
         added: "2009-07-29"}).
iana(language,oym,tag{
         type: "language",
         subtag: "oym",
         description: "Wayampi",
         added: "2009-07-29"}).
iana(language,oyy,tag{
         type: "language",
         subtag: "oyy",
         description: "Oya'oya",
         added: "2009-07-29"}).
iana(language,ozm,tag{
         type: "language",
         subtag: "ozm",
         description: "Koonzime",
         added: "2009-07-29"}).
iana(language,paa,tag{
         type: "language",
         subtag: "paa",
         description: "Papuan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,pab,tag{
         type: "language",
         subtag: "pab",
         description: "Parecís",
         added: "2009-07-29"}).
iana(language,pac,tag{
         type: "language",
         subtag: "pac",
         description: "Pacoh",
         added: "2009-07-29"}).
iana(language,pad,tag{
         type: "language",
         subtag: "pad",
         description: "Paumarí",
         added: "2009-07-29"}).
iana(language,pae,tag{
         type: "language",
         subtag: "pae",
         description: "Pagibete",
         added: "2009-07-29"}).
iana(language,paf,tag{
         type: "language",
         subtag: "paf",
         description: "Paranawát",
         added: "2009-07-29"}).
iana(language,pag,tag{
         type: "language",
         subtag: "pag",
         description: "Pangasinan",
         added: "2005-10-16"}).
iana(language,pah,tag{
         type: "language",
         subtag: "pah",
         description: "Tenharim",
         added: "2009-07-29"}).
iana(language,pai,tag{
         type: "language",
         subtag: "pai",
         description: "Pe",
         added: "2009-07-29"}).
iana(language,pak,tag{
         type: "language",
         subtag: "pak",
         description: "Parakanã",
         added: "2009-07-29"}).
iana(language,pal,tag{
         type: "language",
         subtag: "pal",
         description: "Pahlavi",
         added: "2005-10-16"}).
iana(language,pam,tag{
         type: "language",
         subtag: "pam",
         description: ["Pampanga","Kapampangan"],
         added: "2005-10-16"}).
iana(language,pao,tag{
         type: "language",
         subtag: "pao",
         description: "Northern Paiute",
         added: "2009-07-29"}).
iana(language,pap,tag{
         type: "language",
         subtag: "pap",
         description: "Papiamento",
         added: "2005-10-16"}).
iana(language,paq,tag{
         type: "language",
         subtag: "paq",
         description: "Parya",
         added: "2009-07-29"}).
iana(language,par,tag{
         type: "language",
         subtag: "par",
         description: ["Panamint","Timbisha"],
         added: "2009-07-29"}).
iana(language,pas,tag{
         type: "language",
         subtag: "pas",
         description: "Papasena",
         added: "2009-07-29"}).
iana(language,pat,tag{
         type: "language",
         subtag: "pat",
         description: "Papitalai",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         preferred_value: "kxr"}).
iana(language,pau,tag{
         type: "language",
         subtag: "pau",
         description: "Palauan",
         added: "2005-10-16"}).
iana(language,pav,tag{
         type: "language",
         subtag: "pav",
         description: "Pakaásnovos",
         added: "2009-07-29"}).
iana(language,paw,tag{
         type: "language",
         subtag: "paw",
         description: "Pawnee",
         added: "2009-07-29"}).
iana(language,pax,tag{
         type: "language",
         subtag: "pax",
         description: "Pankararé",
         added: "2009-07-29"}).
iana(language,pay,tag{
         type: "language",
         subtag: "pay",
         description: "Pech",
         added: "2009-07-29"}).
iana(language,paz,tag{
         type: "language",
         subtag: "paz",
         description: "Pankararú",
         added: "2009-07-29"}).
iana(language,pbb,tag{
         type: "language",
         subtag: "pbb",
         description: "Páez",
         added: "2009-07-29"}).
iana(language,pbc,tag{
         type: "language",
         subtag: "pbc",
         description: "Patamona",
         added: "2009-07-29"}).
iana(language,pbe,tag{
         type: "language",
         subtag: "pbe",
         description: "Mezontla Popoloca",
         added: "2009-07-29"}).
iana(language,pbf,tag{
         type: "language",
         subtag: "pbf",
         description: "Coyotepec Popoloca",
         added: "2009-07-29"}).
iana(language,pbg,tag{
         type: "language",
         subtag: "pbg",
         description: "Paraujano",
         added: "2009-07-29"}).
iana(language,pbh,tag{
         type: "language",
         subtag: "pbh",
         description: "E'ñapa Woromaipu",
         added: "2009-07-29"}).
iana(language,pbi,tag{
         type: "language",
         subtag: "pbi",
         description: "Parkwa",
         added: "2009-07-29"}).
iana(language,pbl,tag{
         type: "language",
         subtag: "pbl",
         description: "Mak (Nigeria)",
         added: "2009-07-29"}).
iana(language,pbm,tag{
         type: "language",
         subtag: "pbm",
         description: "Puebla Mazatec",
         added: "2018-03-08",
         comments: "see also maa"}).
iana(language,pbn,tag{
         type: "language",
         subtag: "pbn",
         description: "Kpasam",
         added: "2009-07-29"}).
iana(language,pbo,tag{
         type: "language",
         subtag: "pbo",
         description: "Papel",
         added: "2009-07-29"}).
iana(language,pbp,tag{
         type: "language",
         subtag: "pbp",
         description: "Badyara",
         added: "2009-07-29"}).
iana(language,pbr,tag{
         type: "language",
         subtag: "pbr",
         description: "Pangwa",
         added: "2009-07-29"}).
iana(language,pbs,tag{
         type: "language",
         subtag: "pbs",
         description: "Central Pame",
         added: "2009-07-29"}).
iana(language,pbt,tag{
         type: "language",
         subtag: "pbt",
         description: "Southern Pashto",
         added: "2009-07-29",
         macrolanguage: "ps"}).
iana(language,pbu,tag{
         type: "language",
         subtag: "pbu",
         description: "Northern Pashto",
         added: "2009-07-29",
         macrolanguage: "ps"}).
iana(language,pbv,tag{
         type: "language",
         subtag: "pbv",
         description: "Pnar",
         added: "2009-07-29"}).
iana(language,pby,tag{
         type: "language",
         subtag: "pby",
         description: "Pyu (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,pbz,tag{
         type: "language",
         subtag: "pbz",
         description: "Palu",
         added: "2009-07-29",
         deprecated: "2012-08-12"}).
iana(language,pca,tag{
         type: "language",
         subtag: "pca",
         description: "Santa Inés Ahuatempan Popoloca",
         added: "2009-07-29"}).
iana(language,pcb,tag{
         type: "language",
         subtag: "pcb",
         description: "Pear",
         added: "2009-07-29"}).
iana(language,pcc,tag{
         type: "language",
         subtag: "pcc",
         description: "Bouyei",
         added: "2009-07-29"}).
iana(language,pcd,tag{
         type: "language",
         subtag: "pcd",
         description: "Picard",
         added: "2009-07-29"}).
iana(language,pce,tag{
         type: "language",
         subtag: "pce",
         description: "Ruching Palaung",
         added: "2009-07-29"}).
iana(language,pcf,tag{
         type: "language",
         subtag: "pcf",
         description: "Paliyan",
         added: "2009-07-29"}).
iana(language,pcg,tag{
         type: "language",
         subtag: "pcg",
         description: "Paniya",
         added: "2009-07-29"}).
iana(language,pch,tag{
         type: "language",
         subtag: "pch",
         description: "Pardhan",
         added: "2009-07-29"}).
iana(language,pci,tag{
         type: "language",
         subtag: "pci",
         description: "Duruwa",
         added: "2009-07-29"}).
iana(language,pcj,tag{
         type: "language",
         subtag: "pcj",
         description: "Parenga",
         added: "2009-07-29"}).
iana(language,pck,tag{
         type: "language",
         subtag: "pck",
         description: "Paite Chin",
         added: "2009-07-29"}).
iana(language,pcl,tag{
         type: "language",
         subtag: "pcl",
         description: "Pardhi",
         added: "2009-07-29"}).
iana(language,pcm,tag{
         type: "language",
         subtag: "pcm",
         description: "Nigerian Pidgin",
         added: "2009-07-29"}).
iana(language,pcn,tag{
         type: "language",
         subtag: "pcn",
         description: "Piti",
         added: "2009-07-29"}).
iana(language,pcp,tag{
         type: "language",
         subtag: "pcp",
         description: "Pacahuara",
         added: "2009-07-29"}).
iana(language,pcr,tag{
         type: "language",
         subtag: "pcr",
         description: "Panang",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "adx"}).
iana(language,pcw,tag{
         type: "language",
         subtag: "pcw",
         description: "Pyapun",
         added: "2009-07-29"}).
iana(language,pda,tag{
         type: "language",
         subtag: "pda",
         description: "Anam",
         added: "2009-07-29"}).
iana(language,pdc,tag{
         type: "language",
         subtag: "pdc",
         description: "Pennsylvania German",
         added: "2009-07-29"}).
iana(language,pdi,tag{
         type: "language",
         subtag: "pdi",
         description: "Pa Di",
         added: "2009-07-29"}).
iana(language,pdn,tag{
         type: "language",
         subtag: "pdn",
         description: ["Podena","Fedan"],
         added: "2009-07-29"}).
iana(language,pdo,tag{
         type: "language",
         subtag: "pdo",
         description: "Padoe",
         added: "2009-07-29"}).
iana(language,pdt,tag{
         type: "language",
         subtag: "pdt",
         description: "Plautdietsch",
         added: "2009-07-29"}).
iana(language,pdu,tag{
         type: "language",
         subtag: "pdu",
         description: "Kayan",
         added: "2009-07-29"}).
iana(language,pea,tag{
         type: "language",
         subtag: "pea",
         description: "Peranakan Indonesian",
         added: "2009-07-29"}).
iana(language,peb,tag{
         type: "language",
         subtag: "peb",
         description: "Eastern Pomo",
         added: "2009-07-29"}).
iana(language,ped,tag{
         type: "language",
         subtag: "ped",
         description: "Mala (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,pee,tag{
         type: "language",
         subtag: "pee",
         description: "Taje",
         added: "2009-07-29"}).
iana(language,pef,tag{
         type: "language",
         subtag: "pef",
         description: "Northeastern Pomo",
         added: "2009-07-29"}).
iana(language,peg,tag{
         type: "language",
         subtag: "peg",
         description: "Pengo",
         added: "2009-07-29"}).
iana(language,peh,tag{
         type: "language",
         subtag: "peh",
         description: "Bonan",
         added: "2009-07-29"}).
iana(language,pei,tag{
         type: "language",
         subtag: "pei",
         description: "Chichimeca-Jonaz",
         added: "2009-07-29"}).
iana(language,pej,tag{
         type: "language",
         subtag: "pej",
         description: "Northern Pomo",
         added: "2009-07-29"}).
iana(language,pek,tag{
         type: "language",
         subtag: "pek",
         description: "Penchal",
         added: "2009-07-29"}).
iana(language,pel,tag{
         type: "language",
         subtag: "pel",
         description: "Pekal",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,pem,tag{
         type: "language",
         subtag: "pem",
         description: "Phende",
         added: "2009-07-29"}).
iana(language,peo,tag{
         type: "language",
         subtag: "peo",
         description: "Old Persian (ca. 600-400 B.C.)",
         added: "2005-10-16"}).
iana(language,pep,tag{
         type: "language",
         subtag: "pep",
         description: "Kunja",
         added: "2009-07-29"}).
iana(language,peq,tag{
         type: "language",
         subtag: "peq",
         description: "Southern Pomo",
         added: "2009-07-29"}).
iana(language,pes,tag{
         type: "language",
         subtag: "pes",
         description: "Iranian Persian",
         added: "2009-07-29",
         macrolanguage: "fa"}).
iana(language,pev,tag{
         type: "language",
         subtag: "pev",
         description: "Pémono",
         added: "2009-07-29"}).
iana(language,pex,tag{
         type: "language",
         subtag: "pex",
         description: "Petats",
         added: "2009-07-29"}).
iana(language,pey,tag{
         type: "language",
         subtag: "pey",
         description: "Petjo",
         added: "2009-07-29"}).
iana(language,pez,tag{
         type: "language",
         subtag: "pez",
         description: "Eastern Penan",
         added: "2009-07-29"}).
iana(language,pfa,tag{
         type: "language",
         subtag: "pfa",
         description: "Pááfang",
         added: "2009-07-29"}).
iana(language,pfe,tag{
         type: "language",
         subtag: "pfe",
         description: "Pere",
         added: "2009-07-29"}).
iana(language,pfl,tag{
         type: "language",
         subtag: "pfl",
         description: "Pfaelzisch",
         added: "2009-07-29"}).
iana(language,pga,tag{
         type: "language",
         subtag: "pga",
         description: "Sudanese Creole Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,pgd,tag{
         type: "language",
         subtag: "pgd",
         description: "Gāndhārī",
         added: "2015-02-12"}).
iana(language,pgg,tag{
         type: "language",
         subtag: "pgg",
         description: "Pangwali",
         added: "2009-07-29"}).
iana(language,pgi,tag{
         type: "language",
         subtag: "pgi",
         description: "Pagi",
         added: "2009-07-29"}).
iana(language,pgk,tag{
         type: "language",
         subtag: "pgk",
         description: "Rerep",
         added: "2009-07-29"}).
iana(language,pgl,tag{
         type: "language",
         subtag: "pgl",
         description: "Primitive Irish",
         added: "2011-08-16"}).
iana(language,pgn,tag{
         type: "language",
         subtag: "pgn",
         description: "Paelignian",
         added: "2009-07-29"}).
iana(language,pgs,tag{
         type: "language",
         subtag: "pgs",
         description: "Pangseng",
         added: "2009-07-29"}).
iana(language,pgu,tag{
         type: "language",
         subtag: "pgu",
         description: "Pagu",
         added: "2009-07-29"}).
iana(language,pgy,tag{
         type: "language",
         subtag: "pgy",
         description: "Pongyong",
         added: "2009-07-29",
         deprecated: "2012-08-12"}).
iana(language,pgz,tag{
         type: "language",
         subtag: "pgz",
         description: "Papua New Guinean Sign Language",
         added: "2016-05-30"}).
iana(language,pha,tag{
         type: "language",
         subtag: "pha",
         description: "Pa-Hng",
         added: "2009-07-29"}).
iana(language,phd,tag{
         type: "language",
         subtag: "phd",
         description: "Phudagi",
         added: "2009-07-29"}).
iana(language,phg,tag{
         type: "language",
         subtag: "phg",
         description: "Phuong",
         added: "2009-07-29"}).
iana(language,phh,tag{
         type: "language",
         subtag: "phh",
         description: "Phukha",
         added: "2009-07-29"}).
iana(language,phi,tag{
         type: "language",
         subtag: "phi",
         description: "Philippine languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,phj,tag{
         type: "language",
         subtag: "phj",
         description: "Pahari",
         added: "2022-02-25"}).
iana(language,phk,tag{
         type: "language",
         subtag: "phk",
         description: "Phake",
         added: "2009-07-29"}).
iana(language,phl,tag{
         type: "language",
         subtag: "phl",
         description: ["Phalura","Palula"],
         added: "2009-07-29"}).
iana(language,phm,tag{
         type: "language",
         subtag: "phm",
         description: "Phimbi",
         added: "2009-07-29"}).
iana(language,phn,tag{
         type: "language",
         subtag: "phn",
         description: "Phoenician",
         added: "2005-10-16"}).
iana(language,pho,tag{
         type: "language",
         subtag: "pho",
         description: "Phunoi",
         added: "2009-07-29"}).
iana(language,phq,tag{
         type: "language",
         subtag: "phq",
         description: "Phana'",
         added: "2009-07-29"}).
iana(language,phr,tag{
         type: "language",
         subtag: "phr",
         description: "Pahari-Potwari",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,pht,tag{
         type: "language",
         subtag: "pht",
         description: "Phu Thai",
         added: "2009-07-29"}).
iana(language,phu,tag{
         type: "language",
         subtag: "phu",
         description: "Phuan",
         added: "2009-07-29"}).
iana(language,phv,tag{
         type: "language",
         subtag: "phv",
         description: "Pahlavani",
         added: "2009-07-29"}).
iana(language,phw,tag{
         type: "language",
         subtag: "phw",
         description: "Phangduwali",
         added: "2009-07-29"}).
iana(language,pia,tag{
         type: "language",
         subtag: "pia",
         description: "Pima Bajo",
         added: "2009-07-29"}).
iana(language,pib,tag{
         type: "language",
         subtag: "pib",
         description: "Yine",
         added: "2009-07-29"}).
iana(language,pic,tag{
         type: "language",
         subtag: "pic",
         description: "Pinji",
         added: "2009-07-29"}).
iana(language,pid,tag{
         type: "language",
         subtag: "pid",
         description: "Piaroa",
         added: "2009-07-29"}).
iana(language,pie,tag{
         type: "language",
         subtag: "pie",
         description: "Piro",
         added: "2009-07-29"}).
iana(language,pif,tag{
         type: "language",
         subtag: "pif",
         description: "Pingelapese",
         added: "2009-07-29"}).
iana(language,pig,tag{
         type: "language",
         subtag: "pig",
         description: "Pisabo",
         added: "2009-07-29"}).
iana(language,pih,tag{
         type: "language",
         subtag: "pih",
         description: "Pitcairn-Norfolk",
         added: "2009-07-29"}).
iana(language,pii,tag{
         type: "language",
         subtag: "pii",
         description: "Pini",
         added: "2009-07-29",
         deprecated: "2022-02-25"}).
iana(language,pij,tag{
         type: "language",
         subtag: "pij",
         description: "Pijao",
         added: "2009-07-29"}).
iana(language,pil,tag{
         type: "language",
         subtag: "pil",
         description: "Yom",
         added: "2009-07-29"}).
iana(language,pim,tag{
         type: "language",
         subtag: "pim",
         description: "Powhatan",
         added: "2009-07-29"}).
iana(language,pin,tag{
         type: "language",
         subtag: "pin",
         description: "Piame",
         added: "2009-07-29"}).
iana(language,pio,tag{
         type: "language",
         subtag: "pio",
         description: "Piapoco",
         added: "2009-07-29"}).
iana(language,pip,tag{
         type: "language",
         subtag: "pip",
         description: "Pero",
         added: "2009-07-29"}).
iana(language,pir,tag{
         type: "language",
         subtag: "pir",
         description: "Piratapuyo",
         added: "2009-07-29"}).
iana(language,pis,tag{
         type: "language",
         subtag: "pis",
         description: "Pijin",
         added: "2009-07-29"}).
iana(language,pit,tag{
         type: "language",
         subtag: "pit",
         description: "Pitta Pitta",
         added: "2009-07-29"}).
iana(language,piu,tag{
         type: "language",
         subtag: "piu",
         description: "Pintupi-Luritja",
         added: "2009-07-29"}).
iana(language,piv,tag{
         type: "language",
         subtag: "piv",
         description: ["Pileni","Vaeakau-Taumako"],
         added: "2009-07-29"}).
iana(language,piw,tag{
         type: "language",
         subtag: "piw",
         description: "Pimbwe",
         added: "2009-07-29"}).
iana(language,pix,tag{
         type: "language",
         subtag: "pix",
         description: "Piu",
         added: "2009-07-29"}).
iana(language,piy,tag{
         type: "language",
         subtag: "piy",
         description: "Piya-Kwonci",
         added: "2009-07-29"}).
iana(language,piz,tag{
         type: "language",
         subtag: "piz",
         description: "Pije",
         added: "2009-07-29"}).
iana(language,pjt,tag{
         type: "language",
         subtag: "pjt",
         description: "Pitjantjatjara",
         added: "2009-07-29"}).
iana(language,pka,tag{
         type: "language",
         subtag: "pka",
         description: "Ardhamāgadhī Prākrit",
         added: "2009-07-29"}).
iana(language,pkb,tag{
         type: "language",
         subtag: "pkb",
         description: ["Pokomo","Kipfokomo"],
         added: "2009-07-29"}).
iana(language,pkc,tag{
         type: "language",
         subtag: "pkc",
         description: "Paekche",
         added: "2009-07-29"}).
iana(language,pkg,tag{
         type: "language",
         subtag: "pkg",
         description: "Pak-Tong",
         added: "2009-07-29"}).
iana(language,pkh,tag{
         type: "language",
         subtag: "pkh",
         description: "Pankhu",
         added: "2009-07-29"}).
iana(language,pkn,tag{
         type: "language",
         subtag: "pkn",
         description: "Pakanha",
         added: "2009-07-29"}).
iana(language,pko,tag{
         type: "language",
         subtag: "pko",
         description: "Pökoot",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,pkp,tag{
         type: "language",
         subtag: "pkp",
         description: "Pukapuka",
         added: "2009-07-29"}).
iana(language,pkr,tag{
         type: "language",
         subtag: "pkr",
         description: "Attapady Kurumba",
         added: "2009-07-29"}).
iana(language,pks,tag{
         type: "language",
         subtag: "pks",
         description: "Pakistan Sign Language",
         added: "2009-07-29"}).
iana(language,pkt,tag{
         type: "language",
         subtag: "pkt",
         description: "Maleng",
         added: "2009-07-29"}).
iana(language,pku,tag{
         type: "language",
         subtag: "pku",
         description: "Paku",
         added: "2009-07-29"}).
iana(language,pla,tag{
         type: "language",
         subtag: "pla",
         description: "Miani",
         added: "2009-07-29"}).
iana(language,plb,tag{
         type: "language",
         subtag: "plb",
         description: "Polonombauk",
         added: "2009-07-29"}).
iana(language,plc,tag{
         type: "language",
         subtag: "plc",
         description: "Central Palawano",
         added: "2009-07-29"}).
iana(language,pld,tag{
         type: "language",
         subtag: "pld",
         description: "Polari",
         added: "2009-07-29"}).
iana(language,ple,tag{
         type: "language",
         subtag: "ple",
         description: "Palu'e",
         added: "2009-07-29"}).
iana(language,plf,tag{
         type: "language",
         subtag: "plf",
         description: "Central Malayo-Polynesian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,plg,tag{
         type: "language",
         subtag: "plg",
         description: "Pilagá",
         added: "2009-07-29"}).
iana(language,plh,tag{
         type: "language",
         subtag: "plh",
         description: "Paulohi",
         added: "2009-07-29"}).
iana(language,plj,tag{
         type: "language",
         subtag: "plj",
         description: "Polci",
         added: "2009-07-29"}).
iana(language,plk,tag{
         type: "language",
         subtag: "plk",
         description: "Kohistani Shina",
         added: "2009-07-29"}).
iana(language,pll,tag{
         type: "language",
         subtag: "pll",
         description: "Shwe Palaung",
         added: "2009-07-29"}).
iana(language,pln,tag{
         type: "language",
         subtag: "pln",
         description: "Palenquero",
         added: "2009-07-29"}).
iana(language,plo,tag{
         type: "language",
         subtag: "plo",
         description: "Oluta Popoluca",
         added: "2009-07-29"}).
iana(language,plp,tag{
         type: "language",
         subtag: "plp",
         description: "Palpa",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,plq,tag{
         type: "language",
         subtag: "plq",
         description: "Palaic",
         added: "2009-07-29"}).
iana(language,plr,tag{
         type: "language",
         subtag: "plr",
         description: "Palaka Senoufo",
         added: "2009-07-29"}).
iana(language,pls,tag{
         type: "language",
         subtag: "pls",
         description: ["San Marcos Tlacoyalco Popoloca","San Marcos Tlalcoyalco Popoloca"],
         added: "2009-07-29"}).
iana(language,plt,tag{
         type: "language",
         subtag: "plt",
         description: "Plateau Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,plu,tag{
         type: "language",
         subtag: "plu",
         description: "Palikúr",
         added: "2009-07-29"}).
iana(language,plv,tag{
         type: "language",
         subtag: "plv",
         description: "Southwest Palawano",
         added: "2009-07-29"}).
iana(language,plw,tag{
         type: "language",
         subtag: "plw",
         description: "Brooke's Point Palawano",
         added: "2009-07-29"}).
iana(language,ply,tag{
         type: "language",
         subtag: "ply",
         description: "Bolyu",
         added: "2009-07-29"}).
iana(language,plz,tag{
         type: "language",
         subtag: "plz",
         description: "Paluan",
         added: "2009-07-29"}).
iana(language,pma,tag{
         type: "language",
         subtag: "pma",
         description: "Paama",
         added: "2009-07-29"}).
iana(language,pmb,tag{
         type: "language",
         subtag: "pmb",
         description: "Pambia",
         added: "2009-07-29"}).
iana(language,pmc,tag{
         type: "language",
         subtag: "pmc",
         description: "Palumata",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "huw"}).
iana(language,pmd,tag{
         type: "language",
         subtag: "pmd",
         description: "Pallanganmiddang",
         added: "2012-08-12"}).
iana(language,pme,tag{
         type: "language",
         subtag: "pme",
         description: "Pwaamei",
         added: "2009-07-29"}).
iana(language,pmf,tag{
         type: "language",
         subtag: "pmf",
         description: "Pamona",
         added: "2009-07-29"}).
iana(language,pmh,tag{
         type: "language",
         subtag: "pmh",
         description: "Māhārāṣṭri Prākrit",
         added: "2009-07-29"}).
iana(language,pmi,tag{
         type: "language",
         subtag: "pmi",
         description: "Northern Pumi",
         added: "2009-07-29"}).
iana(language,pmj,tag{
         type: "language",
         subtag: "pmj",
         description: "Southern Pumi",
         added: "2009-07-29"}).
iana(language,pmk,tag{
         type: "language",
         subtag: "pmk",
         description: "Pamlico",
         added: "2009-07-29"}).
iana(language,pml,tag{
         type: "language",
         subtag: "pml",
         description: "Lingua Franca",
         added: "2009-07-29"}).
iana(language,pmm,tag{
         type: "language",
         subtag: "pmm",
         description: "Pomo",
         added: "2009-07-29"}).
iana(language,pmn,tag{
         type: "language",
         subtag: "pmn",
         description: "Pam",
         added: "2009-07-29"}).
iana(language,pmo,tag{
         type: "language",
         subtag: "pmo",
         description: "Pom",
         added: "2009-07-29"}).
iana(language,pmq,tag{
         type: "language",
         subtag: "pmq",
         description: "Northern Pame",
         added: "2009-07-29"}).
iana(language,pmr,tag{
         type: "language",
         subtag: "pmr",
         description: "Paynamar",
         added: "2009-07-29"}).
iana(language,pms,tag{
         type: "language",
         subtag: "pms",
         description: "Piemontese",
         added: "2009-07-29"}).
iana(language,pmt,tag{
         type: "language",
         subtag: "pmt",
         description: "Tuamotuan",
         added: "2009-07-29"}).
iana(language,pmu,tag{
         type: "language",
         subtag: "pmu",
         description: "Mirpur Panjabi",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "phr",
         macrolanguage: "lah"}).
iana(language,pmw,tag{
         type: "language",
         subtag: "pmw",
         description: "Plains Miwok",
         added: "2009-07-29"}).
iana(language,pmx,tag{
         type: "language",
         subtag: "pmx",
         description: "Poumei Naga",
         added: "2009-07-29"}).
iana(language,pmy,tag{
         type: "language",
         subtag: "pmy",
         description: "Papuan Malay",
         added: "2009-07-29"}).
iana(language,pmz,tag{
         type: "language",
         subtag: "pmz",
         description: "Southern Pame",
         added: "2009-07-29"}).
iana(language,pna,tag{
         type: "language",
         subtag: "pna",
         description: "Punan Bah-Biau",
         added: "2009-07-29"}).
iana(language,pnb,tag{
         type: "language",
         subtag: "pnb",
         description: "Western Panjabi",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,pnc,tag{
         type: "language",
         subtag: "pnc",
         description: "Pannei",
         added: "2009-07-29"}).
iana(language,pnd,tag{
         type: "language",
         subtag: "pnd",
         description: "Mpinda",
         added: "2019-04-16"}).
iana(language,pne,tag{
         type: "language",
         subtag: "pne",
         description: "Western Penan",
         added: "2009-07-29"}).
iana(language,png,tag{
         type: "language",
         subtag: "png",
         description: ["Pangu","Pongu"],
         added: "2009-07-29"}).
iana(language,pnh,tag{
         type: "language",
         subtag: "pnh",
         description: "Penrhyn",
         added: "2009-07-29"}).
iana(language,pni,tag{
         type: "language",
         subtag: "pni",
         description: "Aoheng",
         added: "2009-07-29"}).
iana(language,pnj,tag{
         type: "language",
         subtag: "pnj",
         description: "Pinjarup",
         added: "2013-09-10"}).
iana(language,pnk,tag{
         type: "language",
         subtag: "pnk",
         description: "Paunaka",
         added: "2012-08-12"}).
iana(language,pnl,tag{
         type: "language",
         subtag: "pnl",
         description: "Paleni",
         added: "2013-09-10",
         comments: "see also wbf"}).
iana(language,pnm,tag{
         type: "language",
         subtag: "pnm",
         description: "Punan Batu 1",
         added: "2009-07-29"}).
iana(language,pnn,tag{
         type: "language",
         subtag: "pnn",
         description: "Pinai-Hagahai",
         added: "2009-07-29"}).
iana(language,pno,tag{
         type: "language",
         subtag: "pno",
         description: "Panobo",
         added: "2009-07-29"}).
iana(language,pnp,tag{
         type: "language",
         subtag: "pnp",
         description: "Pancana",
         added: "2009-07-29"}).
iana(language,pnq,tag{
         type: "language",
         subtag: "pnq",
         description: "Pana (Burkina Faso)",
         added: "2009-07-29"}).
iana(language,pnr,tag{
         type: "language",
         subtag: "pnr",
         description: "Panim",
         added: "2009-07-29"}).
iana(language,pns,tag{
         type: "language",
         subtag: "pns",
         description: "Ponosakan",
         added: "2009-07-29"}).
iana(language,pnt,tag{
         type: "language",
         subtag: "pnt",
         description: "Pontic",
         added: "2009-07-29"}).
iana(language,pnu,tag{
         type: "language",
         subtag: "pnu",
         description: "Jiongnai Bunu",
         added: "2009-07-29"}).
iana(language,pnv,tag{
         type: "language",
         subtag: "pnv",
         description: "Pinigura",
         added: "2009-07-29"}).
iana(language,pnw,tag{
         type: "language",
         subtag: "pnw",
         description: ["Banyjima","Panytyima"],
         added: "2009-07-29"}).
iana(language,pnx,tag{
         type: "language",
         subtag: "pnx",
         description: "Phong-Kniang",
         added: "2009-07-29"}).
iana(language,pny,tag{
         type: "language",
         subtag: "pny",
         description: "Pinyin",
         added: "2009-07-29",
         comments: "a Niger-Congo language spoken in Cameroon; not to be confused with the Pinyin romanization systems used for Chinese and Tibetan"
     }).
iana(language,pnz,tag{
         type: "language",
         subtag: "pnz",
         description: "Pana (Central African Republic)",
         added: "2009-07-29"}).
iana(language,poc,tag{
         type: "language",
         subtag: "poc",
         description: "Poqomam",
         added: "2009-07-29"}).
iana(language,pod,tag{
         type: "language",
         subtag: "pod",
         description: "Ponares",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,poe,tag{
         type: "language",
         subtag: "poe",
         description: "San Juan Atzingo Popoloca",
         added: "2009-07-29"}).
iana(language,pof,tag{
         type: "language",
         subtag: "pof",
         description: "Poke",
         added: "2009-07-29"}).
iana(language,pog,tag{
         type: "language",
         subtag: "pog",
         description: "Potiguára",
         added: "2009-07-29"}).
iana(language,poh,tag{
         type: "language",
         subtag: "poh",
         description: "Poqomchi'",
         added: "2009-07-29"}).
iana(language,poi,tag{
         type: "language",
         subtag: "poi",
         description: "Highland Popoluca",
         added: "2009-07-29"}).
iana(language,pok,tag{
         type: "language",
         subtag: "pok",
         description: "Pokangá",
         added: "2009-07-29"}).
iana(language,pom,tag{
         type: "language",
         subtag: "pom",
         description: "Southeastern Pomo",
         added: "2009-07-29"}).
iana(language,pon,tag{
         type: "language",
         subtag: "pon",
         description: "Pohnpeian",
         added: "2005-10-16"}).
iana(language,poo,tag{
         type: "language",
         subtag: "poo",
         description: "Central Pomo",
         added: "2009-07-29"}).
iana(language,pop,tag{
         type: "language",
         subtag: "pop",
         description: "Pwapwâ",
         added: "2009-07-29"}).
iana(language,poq,tag{
         type: "language",
         subtag: "poq",
         description: "Texistepec Popoluca",
         added: "2009-07-29"}).
iana(language,pos,tag{
         type: "language",
         subtag: "pos",
         description: "Sayula Popoluca",
         added: "2009-07-29"}).
iana(language,pot,tag{
         type: "language",
         subtag: "pot",
         description: "Potawatomi",
         added: "2009-07-29"}).
iana(language,pov,tag{
         type: "language",
         subtag: "pov",
         description: "Upper Guinea Crioulo",
         added: "2009-07-29"}).
iana(language,pow,tag{
         type: "language",
         subtag: "pow",
         description: "San Felipe Otlaltepec Popoloca",
         added: "2009-07-29"}).
iana(language,pox,tag{
         type: "language",
         subtag: "pox",
         description: "Polabian",
         added: "2009-07-29"}).
iana(language,poy,tag{
         type: "language",
         subtag: "poy",
         description: "Pogolo",
         added: "2009-07-29"}).
iana(language,poz,tag{
         type: "language",
         subtag: "poz",
         description: "Malayo-Polynesian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,ppa,tag{
         type: "language",
         subtag: "ppa",
         description: "Pao",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "bfy"}).
iana(language,ppe,tag{
         type: "language",
         subtag: "ppe",
         description: "Papi",
         added: "2009-07-29"}).
iana(language,ppi,tag{
         type: "language",
         subtag: "ppi",
         description: "Paipai",
         added: "2009-07-29"}).
iana(language,ppk,tag{
         type: "language",
         subtag: "ppk",
         description: "Uma",
         added: "2009-07-29"}).
iana(language,ppl,tag{
         type: "language",
         subtag: "ppl",
         description: ["Pipil","Nicarao"],
         added: "2009-07-29"}).
iana(language,ppm,tag{
         type: "language",
         subtag: "ppm",
         description: "Papuma",
         added: "2009-07-29"}).
iana(language,ppn,tag{
         type: "language",
         subtag: "ppn",
         description: "Papapana",
         added: "2009-07-29"}).
iana(language,ppo,tag{
         type: "language",
         subtag: "ppo",
         description: "Folopa",
         added: "2009-07-29"}).
iana(language,ppp,tag{
         type: "language",
         subtag: "ppp",
         description: "Pelende",
         added: "2009-07-29"}).
iana(language,ppq,tag{
         type: "language",
         subtag: "ppq",
         description: "Pei",
         added: "2009-07-29"}).
iana(language,ppr,tag{
         type: "language",
         subtag: "ppr",
         description: "Piru",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "lcq"}).
iana(language,pps,tag{
         type: "language",
         subtag: "pps",
         description: "San Luís Temalacayuca Popoloca",
         added: "2009-07-29"}).
iana(language,ppt,tag{
         type: "language",
         subtag: "ppt",
         description: "Pare",
         added: "2009-07-29"}).
iana(language,ppu,tag{
         type: "language",
         subtag: "ppu",
         description: "Papora",
         added: "2009-07-29"}).
iana(language,pqa,tag{
         type: "language",
         subtag: "pqa",
         description: "Pa'a",
         added: "2009-07-29"}).
iana(language,pqe,tag{
         type: "language",
         subtag: "pqe",
         description: "Eastern Malayo-Polynesian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,pqm,tag{
         type: "language",
         subtag: "pqm",
         description: "Malecite-Passamaquoddy",
         added: "2009-07-29"}).
iana(language,pqw,tag{
         type: "language",
         subtag: "pqw",
         description: "Western Malayo-Polynesian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,pra,tag{
         type: "language",
         subtag: "pra",
         description: "Prakrit languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,prb,tag{
         type: "language",
         subtag: "prb",
         description: "Lua'",
         added: "2009-07-29",
         deprecated: "2017-02-23"}).
iana(language,prc,tag{
         type: "language",
         subtag: "prc",
         description: "Parachi",
         added: "2009-07-29"}).
iana(language,prd,tag{
         type: "language",
         subtag: "prd",
         description: "Parsi-Dari",
         added: "2009-07-29"}).
iana(language,pre,tag{
         type: "language",
         subtag: "pre",
         description: "Principense",
         added: "2009-07-29"}).
iana(language,prf,tag{
         type: "language",
         subtag: "prf",
         description: "Paranan",
         added: "2010-03-11"}).
iana(language,prg,tag{
         type: "language",
         subtag: "prg",
         description: "Prussian",
         added: "2009-07-29"}).
iana(language,prh,tag{
         type: "language",
         subtag: "prh",
         description: "Porohanon",
         added: "2009-07-29"}).
iana(language,pri,tag{
         type: "language",
         subtag: "pri",
         description: "Paicî",
         added: "2009-07-29"}).
iana(language,prk,tag{
         type: "language",
         subtag: "prk",
         description: "Parauk",
         added: "2009-07-29"}).
iana(language,prl,tag{
         type: "language",
         subtag: "prl",
         description: "Peruvian Sign Language",
         added: "2009-07-29"}).
iana(language,prm,tag{
         type: "language",
         subtag: "prm",
         description: "Kibiri",
         added: "2009-07-29"}).
iana(language,prn,tag{
         type: "language",
         subtag: "prn",
         description: "Prasuni",
         added: "2009-07-29"}).
iana(language,pro,tag{
         type: "language",
         subtag: "pro",
         description: ["Old Provençal (to 1500)","Old Occitan (to 1500)"],
         added: "2005-10-16"}).
iana(language,prp,tag{
         type: "language",
         subtag: "prp",
         description: "Parsi",
         added: "2009-07-29"}).
iana(language,prq,tag{
         type: "language",
         subtag: "prq",
         description: "Ashéninka Perené",
         added: "2009-07-29"}).
iana(language,prr,tag{
         type: "language",
         subtag: "prr",
         description: "Puri",
         added: "2009-07-29"}).
iana(language,prs,tag{
         type: "language",
         subtag: "prs",
         description: ["Dari","Afghan Persian"],
         added: "2009-07-29",
         macrolanguage: "fa"}).
iana(language,prt,tag{
         type: "language",
         subtag: "prt",
         description: "Phai",
         added: "2009-07-29"}).
iana(language,pru,tag{
         type: "language",
         subtag: "pru",
         description: "Puragi",
         added: "2009-07-29"}).
iana(language,prw,tag{
         type: "language",
         subtag: "prw",
         description: "Parawen",
         added: "2009-07-29"}).
iana(language,prx,tag{
         type: "language",
         subtag: "prx",
         description: "Purik",
         added: "2009-07-29"}).
iana(language,pry,tag{
         type: "language",
         subtag: "pry",
         description: "Pray 3",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "prt"}).
iana(language,prz,tag{
         type: "language",
         subtag: "prz",
         description: "Providencia Sign Language",
         added: "2009-07-29"}).
iana(language,psa,tag{
         type: "language",
         subtag: "psa",
         description: "Asue Awyu",
         added: "2009-07-29"}).
iana(language,psc,tag{
         type: "language",
         subtag: "psc",
         description: ["Iranian Sign Language","Persian Sign Language"],
         added: "2009-07-29"}).
iana(language,psd,tag{
         type: "language",
         subtag: "psd",
         description: "Plains Indian Sign Language",
         added: "2009-07-29"}).
iana(language,pse,tag{
         type: "language",
         subtag: "pse",
         description: "Central Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,psg,tag{
         type: "language",
         subtag: "psg",
         description: "Penang Sign Language",
         added: "2009-07-29"}).
iana(language,psh,tag{
         type: "language",
         subtag: "psh",
         description: ["Southwest Pashai","Southwest Pashayi"],
         added: "2009-07-29"}).
iana(language,psi,tag{
         type: "language",
         subtag: "psi",
         description: ["Southeast Pashai","Southeast Pashayi"],
         added: "2009-07-29"}).
iana(language,psl,tag{
         type: "language",
         subtag: "psl",
         description: "Puerto Rican Sign Language",
         added: "2009-07-29"}).
iana(language,psm,tag{
         type: "language",
         subtag: "psm",
         description: "Pauserna",
         added: "2009-07-29"}).
iana(language,psn,tag{
         type: "language",
         subtag: "psn",
         description: "Panasuan",
         added: "2009-07-29"}).
iana(language,pso,tag{
         type: "language",
         subtag: "pso",
         description: "Polish Sign Language",
         added: "2009-07-29"}).
iana(language,psp,tag{
         type: "language",
         subtag: "psp",
         description: "Philippine Sign Language",
         added: "2009-07-29"}).
iana(language,psq,tag{
         type: "language",
         subtag: "psq",
         description: "Pasi",
         added: "2009-07-29"}).
iana(language,psr,tag{
         type: "language",
         subtag: "psr",
         description: "Portuguese Sign Language",
         added: "2009-07-29"}).
iana(language,pss,tag{
         type: "language",
         subtag: "pss",
         description: "Kaulong",
         added: "2009-07-29"}).
iana(language,pst,tag{
         type: "language",
         subtag: "pst",
         description: "Central Pashto",
         added: "2009-07-29",
         macrolanguage: "ps"}).
iana(language,psu,tag{
         type: "language",
         subtag: "psu",
         description: "Sauraseni Prākrit",
         added: "2009-07-29"}).
iana(language,psw,tag{
         type: "language",
         subtag: "psw",
         description: "Port Sandwich",
         added: "2009-07-29"}).
iana(language,psy,tag{
         type: "language",
         subtag: "psy",
         description: "Piscataway",
         added: "2009-07-29"}).
iana(language,pta,tag{
         type: "language",
         subtag: "pta",
         description: "Pai Tavytera",
         added: "2009-07-29"}).
iana(language,pth,tag{
         type: "language",
         subtag: "pth",
         description: "Pataxó Hã-Ha-Hãe",
         added: "2009-07-29"}).
iana(language,pti,tag{
         type: "language",
         subtag: "pti",
         description: ["Pindiini","Wangkatha"],
         added: "2009-07-29"}).
iana(language,ptn,tag{
         type: "language",
         subtag: "ptn",
         description: "Patani",
         added: "2009-07-29"}).
iana(language,pto,tag{
         type: "language",
         subtag: "pto",
         description: "Zo'é",
         added: "2009-07-29"}).
iana(language,ptp,tag{
         type: "language",
         subtag: "ptp",
         description: "Patep",
         added: "2009-07-29"}).
iana(language,ptq,tag{
         type: "language",
         subtag: "ptq",
         description: "Pattapu",
         added: "2014-04-06"}).
iana(language,ptr,tag{
         type: "language",
         subtag: "ptr",
         description: "Piamatsina",
         added: "2009-07-29"}).
iana(language,ptt,tag{
         type: "language",
         subtag: "ptt",
         description: "Enrekang",
         added: "2009-07-29"}).
iana(language,ptu,tag{
         type: "language",
         subtag: "ptu",
         description: "Bambam",
         added: "2009-07-29"}).
iana(language,ptv,tag{
         type: "language",
         subtag: "ptv",
         description: "Port Vato",
         added: "2009-07-29"}).
iana(language,ptw,tag{
         type: "language",
         subtag: "ptw",
         description: "Pentlatch",
         added: "2009-07-29"}).
iana(language,pty,tag{
         type: "language",
         subtag: "pty",
         description: "Pathiya",
         added: "2009-07-29"}).
iana(language,pua,tag{
         type: "language",
         subtag: "pua",
         description: "Western Highland Purepecha",
         added: "2009-07-29"}).
iana(language,pub,tag{
         type: "language",
         subtag: "pub",
         description: "Purum",
         added: "2009-07-29"}).
iana(language,puc,tag{
         type: "language",
         subtag: "puc",
         description: "Punan Merap",
         added: "2009-07-29"}).
iana(language,pud,tag{
         type: "language",
         subtag: "pud",
         description: "Punan Aput",
         added: "2009-07-29"}).
iana(language,pue,tag{
         type: "language",
         subtag: "pue",
         description: "Puelche",
         added: "2009-07-29"}).
iana(language,puf,tag{
         type: "language",
         subtag: "puf",
         description: "Punan Merah",
         added: "2009-07-29"}).
iana(language,pug,tag{
         type: "language",
         subtag: "pug",
         description: "Phuie",
         added: "2009-07-29"}).
iana(language,pui,tag{
         type: "language",
         subtag: "pui",
         description: "Puinave",
         added: "2009-07-29"}).
iana(language,puj,tag{
         type: "language",
         subtag: "puj",
         description: "Punan Tubu",
         added: "2009-07-29"}).
iana(language,puk,tag{
         type: "language",
         subtag: "puk",
         description: "Pu Ko",
         added: "2009-07-29",
         deprecated: "2017-02-23"}).
iana(language,pum,tag{
         type: "language",
         subtag: "pum",
         description: "Puma",
         added: "2009-07-29"}).
iana(language,puo,tag{
         type: "language",
         subtag: "puo",
         description: "Puoc",
         added: "2009-07-29"}).
iana(language,pup,tag{
         type: "language",
         subtag: "pup",
         description: "Pulabu",
         added: "2009-07-29"}).
iana(language,puq,tag{
         type: "language",
         subtag: "puq",
         description: "Puquina",
         added: "2009-07-29"}).
iana(language,pur,tag{
         type: "language",
         subtag: "pur",
         description: "Puruborá",
         added: "2009-07-29"}).
iana(language,put,tag{
         type: "language",
         subtag: "put",
         description: "Putoh",
         added: "2009-07-29"}).
iana(language,puu,tag{
         type: "language",
         subtag: "puu",
         description: "Punu",
         added: "2009-07-29"}).
iana(language,puw,tag{
         type: "language",
         subtag: "puw",
         description: "Puluwatese",
         added: "2009-07-29"}).
iana(language,pux,tag{
         type: "language",
         subtag: "pux",
         description: "Puare",
         added: "2009-07-29"}).
iana(language,puy,tag{
         type: "language",
         subtag: "puy",
         description: "Purisimeño",
         added: "2009-07-29"}).
iana(language,puz,tag{
         type: "language",
         subtag: "puz",
         description: "Purum Naga",
         added: "2009-07-29",
         deprecated: "2014-02-28",
         preferred_value: "pub"}).
iana(language,pwa,tag{
         type: "language",
         subtag: "pwa",
         description: "Pawaia",
         added: "2009-07-29"}).
iana(language,pwb,tag{
         type: "language",
         subtag: "pwb",
         description: "Panawa",
         added: "2009-07-29"}).
iana(language,pwg,tag{
         type: "language",
         subtag: "pwg",
         description: "Gapapaiwa",
         added: "2009-07-29"}).
iana(language,pwi,tag{
         type: "language",
         subtag: "pwi",
         description: "Patwin",
         added: "2013-09-10"}).
iana(language,pwm,tag{
         type: "language",
         subtag: "pwm",
         description: "Molbog",
         added: "2009-07-29"}).
iana(language,pwn,tag{
         type: "language",
         subtag: "pwn",
         description: "Paiwan",
         added: "2009-07-29"}).
iana(language,pwo,tag{
         type: "language",
         subtag: "pwo",
         description: "Pwo Western Karen",
         added: "2009-07-29"}).
iana(language,pwr,tag{
         type: "language",
         subtag: "pwr",
         description: "Powari",
         added: "2009-07-29"}).
iana(language,pww,tag{
         type: "language",
         subtag: "pww",
         description: "Pwo Northern Karen",
         added: "2009-07-29"}).
iana(language,pxm,tag{
         type: "language",
         subtag: "pxm",
         description: "Quetzaltepec Mixe",
         added: "2009-07-29"}).
iana(language,pye,tag{
         type: "language",
         subtag: "pye",
         description: "Pye Krumen",
         added: "2009-07-29"}).
iana(language,pym,tag{
         type: "language",
         subtag: "pym",
         description: "Fyam",
         added: "2009-07-29"}).
iana(language,pyn,tag{
         type: "language",
         subtag: "pyn",
         description: "Poyanáwa",
         added: "2009-07-29"}).
iana(language,pys,tag{
         type: "language",
         subtag: "pys",
         description: ["Paraguayan Sign Language","Lengua de Señas del Paraguay"],
         added: "2010-03-11"}).
iana(language,pyu,tag{
         type: "language",
         subtag: "pyu",
         description: "Puyuma",
         added: "2009-07-29"}).
iana(language,pyx,tag{
         type: "language",
         subtag: "pyx",
         description: "Pyu (Myanmar)",
         added: "2009-07-29"}).
iana(language,pyy,tag{
         type: "language",
         subtag: "pyy",
         description: "Pyen",
         added: "2009-07-29"}).
iana(language,pzh,tag{
         type: "language",
         subtag: "pzh",
         description: "Pazeh",
         added: "2022-02-25"}).
iana(language,pzn,tag{
         type: "language",
         subtag: "pzn",
         description: ["Jejara Naga","Para Naga"],
         added: "2009-07-29"}).
iana(language,qaa,tag{
         type: "language",
         subtag: "qaa..qtz",
         description: "Private use",
         added: "2005-10-16",
         scope: "private-use"}).
iana(language,qua,tag{
         type: "language",
         subtag: "qua",
         description: "Quapaw",
         added: "2009-07-29"}).
iana(language,qub,tag{
         type: "language",
         subtag: "qub",
         description: "Huallaga Huánuco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quc,tag{
         type: "language",
         subtag: "quc",
         description: ["K'iche'","Quiché"],
         added: "2009-07-29"}).
iana(language,qud,tag{
         type: "language",
         subtag: "qud",
         description: "Calderón Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quf,tag{
         type: "language",
         subtag: "quf",
         description: "Lambayeque Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qug,tag{
         type: "language",
         subtag: "qug",
         description: "Chimborazo Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quh,tag{
         type: "language",
         subtag: "quh",
         description: "South Bolivian Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qui,tag{
         type: "language",
         subtag: "qui",
         description: "Quileute",
         added: "2009-07-29"}).
iana(language,quk,tag{
         type: "language",
         subtag: "quk",
         description: "Chachapoyas Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qul,tag{
         type: "language",
         subtag: "qul",
         description: "North Bolivian Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qum,tag{
         type: "language",
         subtag: "qum",
         description: "Sipacapense",
         added: "2009-07-29"}).
iana(language,qun,tag{
         type: "language",
         subtag: "qun",
         description: "Quinault",
         added: "2009-07-29"}).
iana(language,qup,tag{
         type: "language",
         subtag: "qup",
         description: "Southern Pastaza Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quq,tag{
         type: "language",
         subtag: "quq",
         description: "Quinqui",
         added: "2009-07-29"}).
iana(language,qur,tag{
         type: "language",
         subtag: "qur",
         description: "Yanahuanca Pasco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qus,tag{
         type: "language",
         subtag: "qus",
         description: "Santiago del Estero Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quv,tag{
         type: "language",
         subtag: "quv",
         description: "Sacapulteco",
         added: "2009-07-29"}).
iana(language,quw,tag{
         type: "language",
         subtag: "quw",
         description: "Tena Lowland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qux,tag{
         type: "language",
         subtag: "qux",
         description: "Yauyos Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quy,tag{
         type: "language",
         subtag: "quy",
         description: "Ayacucho Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,quz,tag{
         type: "language",
         subtag: "quz",
         description: "Cusco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qva,tag{
         type: "language",
         subtag: "qva",
         description: "Ambo-Pasco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvc,tag{
         type: "language",
         subtag: "qvc",
         description: "Cajamarca Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qve,tag{
         type: "language",
         subtag: "qve",
         description: "Eastern Apurímac Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvh,tag{
         type: "language",
         subtag: "qvh",
         description: "Huamalíes-Dos de Mayo Huánuco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvi,tag{
         type: "language",
         subtag: "qvi",
         description: "Imbabura Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvj,tag{
         type: "language",
         subtag: "qvj",
         description: "Loja Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvl,tag{
         type: "language",
         subtag: "qvl",
         description: "Cajatambo North Lima Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvm,tag{
         type: "language",
         subtag: "qvm",
         description: "Margos-Yarowilca-Lauricocha Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvn,tag{
         type: "language",
         subtag: "qvn",
         description: "North Junín Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvo,tag{
         type: "language",
         subtag: "qvo",
         description: "Napo Lowland Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvp,tag{
         type: "language",
         subtag: "qvp",
         description: "Pacaraos Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvs,tag{
         type: "language",
         subtag: "qvs",
         description: "San Martín Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvw,tag{
         type: "language",
         subtag: "qvw",
         description: "Huaylla Wanca Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qvy,tag{
         type: "language",
         subtag: "qvy",
         description: "Queyu",
         added: "2009-07-29"}).
iana(language,qvz,tag{
         type: "language",
         subtag: "qvz",
         description: "Northern Pastaza Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qwa,tag{
         type: "language",
         subtag: "qwa",
         description: "Corongo Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qwc,tag{
         type: "language",
         subtag: "qwc",
         description: "Classical Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qwe,tag{
         type: "language",
         subtag: "qwe",
         description: "Quechuan (family)",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,qwh,tag{
         type: "language",
         subtag: "qwh",
         description: "Huaylas Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qwm,tag{
         type: "language",
         subtag: "qwm",
         description: "Kuman (Russia)",
         added: "2009-07-29"}).
iana(language,qws,tag{
         type: "language",
         subtag: "qws",
         description: "Sihuas Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qwt,tag{
         type: "language",
         subtag: "qwt",
         description: "Kwalhioqua-Tlatskanai",
         added: "2009-07-29"}).
iana(language,qxa,tag{
         type: "language",
         subtag: "qxa",
         description: "Chiquián Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxc,tag{
         type: "language",
         subtag: "qxc",
         description: "Chincha Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxh,tag{
         type: "language",
         subtag: "qxh",
         description: "Panao Huánuco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxl,tag{
         type: "language",
         subtag: "qxl",
         description: "Salasaca Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxn,tag{
         type: "language",
         subtag: "qxn",
         description: "Northern Conchucos Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxo,tag{
         type: "language",
         subtag: "qxo",
         description: "Southern Conchucos Ancash Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxp,tag{
         type: "language",
         subtag: "qxp",
         description: "Puno Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxq,tag{
         type: "language",
         subtag: "qxq",
         description: "Qashqa'i",
         added: "2009-07-29"}).
iana(language,qxr,tag{
         type: "language",
         subtag: "qxr",
         description: "Cañar Highland Quichua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxs,tag{
         type: "language",
         subtag: "qxs",
         description: "Southern Qiang",
         added: "2009-07-29"}).
iana(language,qxt,tag{
         type: "language",
         subtag: "qxt",
         description: "Santa Ana de Tusi Pasco Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxu,tag{
         type: "language",
         subtag: "qxu",
         description: "Arequipa-La Unión Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qxw,tag{
         type: "language",
         subtag: "qxw",
         description: "Jauja Wanca Quechua",
         added: "2009-07-29",
         macrolanguage: "qu"}).
iana(language,qya,tag{
         type: "language",
         subtag: "qya",
         description: "Quenya",
         added: "2009-07-29"}).
iana(language,qyp,tag{
         type: "language",
         subtag: "qyp",
         description: "Quiripi",
         added: "2009-07-29"}).
iana(language,raa,tag{
         type: "language",
         subtag: "raa",
         description: "Dungmali",
         added: "2009-07-29"}).
iana(language,rab,tag{
         type: "language",
         subtag: "rab",
         description: "Camling",
         added: "2009-07-29"}).
iana(language,rac,tag{
         type: "language",
         subtag: "rac",
         description: "Rasawa",
         added: "2009-07-29"}).
iana(language,rad,tag{
         type: "language",
         subtag: "rad",
         description: "Rade",
         added: "2009-07-29"}).
iana(language,raf,tag{
         type: "language",
         subtag: "raf",
         description: "Western Meohang",
         added: "2009-07-29"}).
iana(language,rag,tag{
         type: "language",
         subtag: "rag",
         description: ["Logooli","Lulogooli"],
         added: "2009-07-29",
         macrolanguage: "luy"}).
iana(language,rah,tag{
         type: "language",
         subtag: "rah",
         description: "Rabha",
         added: "2009-07-29"}).
iana(language,rai,tag{
         type: "language",
         subtag: "rai",
         description: "Ramoaaina",
         added: "2009-07-29"}).
iana(language,raj,tag{
         type: "language",
         subtag: "raj",
         description: "Rajasthani",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,rak,tag{
         type: "language",
         subtag: "rak",
         description: "Tulu-Bohuai",
         added: "2009-07-29"}).
iana(language,ral,tag{
         type: "language",
         subtag: "ral",
         description: "Ralte",
         added: "2009-07-29"}).
iana(language,ram,tag{
         type: "language",
         subtag: "ram",
         description: "Canela",
         added: "2009-07-29"}).
iana(language,ran,tag{
         type: "language",
         subtag: "ran",
         description: "Riantana",
         added: "2009-07-29"}).
iana(language,rao,tag{
         type: "language",
         subtag: "rao",
         description: "Rao",
         added: "2009-07-29"}).
iana(language,rap,tag{
         type: "language",
         subtag: "rap",
         description: "Rapanui",
         added: "2005-10-16"}).
iana(language,raq,tag{
         type: "language",
         subtag: "raq",
         description: "Saam",
         added: "2009-07-29"}).
iana(language,rar,tag{
         type: "language",
         subtag: "rar",
         description: ["Rarotongan","Cook Islands Maori"],
         added: "2005-10-16"}).
iana(language,ras,tag{
         type: "language",
         subtag: "ras",
         description: "Tegali",
         added: "2009-07-29"}).
iana(language,rat,tag{
         type: "language",
         subtag: "rat",
         description: "Razajerdi",
         added: "2009-07-29"}).
iana(language,rau,tag{
         type: "language",
         subtag: "rau",
         description: "Raute",
         added: "2009-07-29"}).
iana(language,rav,tag{
         type: "language",
         subtag: "rav",
         description: "Sampang",
         added: "2009-07-29"}).
iana(language,raw,tag{
         type: "language",
         subtag: "raw",
         description: "Rawang",
         added: "2009-07-29"}).
iana(language,rax,tag{
         type: "language",
         subtag: "rax",
         description: "Rang",
         added: "2009-07-29"}).
iana(language,ray,tag{
         type: "language",
         subtag: "ray",
         description: "Rapa",
         added: "2009-07-29"}).
iana(language,raz,tag{
         type: "language",
         subtag: "raz",
         description: "Rahambuu",
         added: "2009-07-29"}).
iana(language,rbb,tag{
         type: "language",
         subtag: "rbb",
         description: "Rumai Palaung",
         added: "2009-07-29"}).
iana(language,rbk,tag{
         type: "language",
         subtag: "rbk",
         description: "Northern Bontok",
         added: "2010-03-11",
         macrolanguage: "bnc"}).
iana(language,rbl,tag{
         type: "language",
         subtag: "rbl",
         description: "Miraya Bikol",
         added: "2010-03-11",
         macrolanguage: "bik"}).
iana(language,rbp,tag{
         type: "language",
         subtag: "rbp",
         description: "Barababaraba",
         added: "2012-08-12"}).
iana(language,rcf,tag{
         type: "language",
         subtag: "rcf",
         description: "Réunion Creole French",
         added: "2009-07-29"}).
iana(language,rdb,tag{
         type: "language",
         subtag: "rdb",
         description: "Rudbari",
         added: "2009-07-29"}).
iana(language,rea,tag{
         type: "language",
         subtag: "rea",
         description: "Rerau",
         added: "2009-07-29"}).
iana(language,reb,tag{
         type: "language",
         subtag: "reb",
         description: "Rembong",
         added: "2009-07-29"}).
iana(language,ree,tag{
         type: "language",
         subtag: "ree",
         description: "Rejang Kayan",
         added: "2009-07-29"}).
iana(language,reg,tag{
         type: "language",
         subtag: "reg",
         description: "Kara (Tanzania)",
         added: "2009-07-29"}).
iana(language,rei,tag{
         type: "language",
         subtag: "rei",
         description: "Reli",
         added: "2009-07-29"}).
iana(language,rej,tag{
         type: "language",
         subtag: "rej",
         description: "Rejang",
         added: "2009-07-29"}).
iana(language,rel,tag{
         type: "language",
         subtag: "rel",
         description: "Rendille",
         added: "2009-07-29"}).
iana(language,rem,tag{
         type: "language",
         subtag: "rem",
         description: "Remo",
         added: "2009-07-29"}).
iana(language,ren,tag{
         type: "language",
         subtag: "ren",
         description: "Rengao",
         added: "2009-07-29"}).
iana(language,rer,tag{
         type: "language",
         subtag: "rer",
         description: "Rer Bare",
         added: "2009-07-29"}).
iana(language,res,tag{
         type: "language",
         subtag: "res",
         description: "Reshe",
         added: "2009-07-29"}).
iana(language,ret,tag{
         type: "language",
         subtag: "ret",
         description: "Retta",
         added: "2009-07-29"}).
iana(language,rey,tag{
         type: "language",
         subtag: "rey",
         description: "Reyesano",
         added: "2009-07-29"}).
iana(language,rga,tag{
         type: "language",
         subtag: "rga",
         description: "Roria",
         added: "2009-07-29"}).
iana(language,rge,tag{
         type: "language",
         subtag: "rge",
         description: "Romano-Greek",
         added: "2009-07-29"}).
iana(language,rgk,tag{
         type: "language",
         subtag: "rgk",
         description: "Rangkas",
         added: "2009-07-29"}).
iana(language,rgn,tag{
         type: "language",
         subtag: "rgn",
         description: "Romagnol",
         added: "2009-07-29"}).
iana(language,rgr,tag{
         type: "language",
         subtag: "rgr",
         description: "Resígaro",
         added: "2009-07-29"}).
iana(language,rgs,tag{
         type: "language",
         subtag: "rgs",
         description: "Southern Roglai",
         added: "2009-07-29"}).
iana(language,rgu,tag{
         type: "language",
         subtag: "rgu",
         description: "Ringgou",
         added: "2009-07-29"}).
iana(language,rhg,tag{
         type: "language",
         subtag: "rhg",
         description: "Rohingya",
         added: "2009-07-29"}).
iana(language,rhp,tag{
         type: "language",
         subtag: "rhp",
         description: "Yahang",
         added: "2009-07-29"}).
iana(language,ria,tag{
         type: "language",
         subtag: "ria",
         description: "Riang (India)",
         added: "2009-07-29"}).
iana(language,rib,tag{
         type: "language",
         subtag: "rib",
         description: "Bribri Sign Language",
         added: "2022-02-25"}).
iana(language,rie,tag{
         type: "language",
         subtag: "rie",
         description: "Rien",
         added: "2009-07-29",
         deprecated: "2017-02-23"}).
iana(language,rif,tag{
         type: "language",
         subtag: "rif",
         description: "Tarifit",
         added: "2009-07-29"}).
iana(language,ril,tag{
         type: "language",
         subtag: "ril",
         description: ["Riang Lang","Riang (Myanmar)"],
         added: "2009-07-29"}).
iana(language,rim,tag{
         type: "language",
         subtag: "rim",
         description: "Nyaturu",
         added: "2009-07-29"}).
iana(language,rin,tag{
         type: "language",
         subtag: "rin",
         description: "Nungu",
         added: "2009-07-29"}).
iana(language,rir,tag{
         type: "language",
         subtag: "rir",
         description: "Ribun",
         added: "2009-07-29"}).
iana(language,rit,tag{
         type: "language",
         subtag: "rit",
         description: "Ritharrngu",
         added: "2009-07-29"}).
iana(language,riu,tag{
         type: "language",
         subtag: "riu",
         description: "Riung",
         added: "2009-07-29"}).
iana(language,rjg,tag{
         type: "language",
         subtag: "rjg",
         description: "Rajong",
         added: "2009-07-29"}).
iana(language,rji,tag{
         type: "language",
         subtag: "rji",
         description: "Raji",
         added: "2009-07-29"}).
iana(language,rjs,tag{
         type: "language",
         subtag: "rjs",
         description: "Rajbanshi",
         added: "2009-07-29"}).
iana(language,rka,tag{
         type: "language",
         subtag: "rka",
         description: "Kraol",
         added: "2009-07-29"}).
iana(language,rkb,tag{
         type: "language",
         subtag: "rkb",
         description: "Rikbaktsa",
         added: "2009-07-29"}).
iana(language,rkh,tag{
         type: "language",
         subtag: "rkh",
         description: "Rakahanga-Manihiki",
         added: "2009-07-29"}).
iana(language,rki,tag{
         type: "language",
         subtag: "rki",
         description: "Rakhine",
         added: "2009-07-29"}).
iana(language,rkm,tag{
         type: "language",
         subtag: "rkm",
         description: "Marka",
         added: "2009-07-29"}).
iana(language,rkt,tag{
         type: "language",
         subtag: "rkt",
         description: ["Rangpuri","Kamta"],
         added: "2009-07-29"}).
iana(language,rkw,tag{
         type: "language",
         subtag: "rkw",
         description: "Arakwal",
         added: "2013-09-10"}).
iana(language,rma,tag{
         type: "language",
         subtag: "rma",
         description: "Rama",
         added: "2009-07-29"}).
iana(language,rmb,tag{
         type: "language",
         subtag: "rmb",
         description: "Rembarrnga",
         added: "2009-07-29"}).
iana(language,rmc,tag{
         type: "language",
         subtag: "rmc",
         description: "Carpathian Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmd,tag{
         type: "language",
         subtag: "rmd",
         description: "Traveller Danish",
         added: "2009-07-29"}).
iana(language,rme,tag{
         type: "language",
         subtag: "rme",
         description: "Angloromani",
         added: "2009-07-29"}).
iana(language,rmf,tag{
         type: "language",
         subtag: "rmf",
         description: "Kalo Finnish Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmg,tag{
         type: "language",
         subtag: "rmg",
         description: "Traveller Norwegian",
         added: "2009-07-29"}).
iana(language,rmh,tag{
         type: "language",
         subtag: "rmh",
         description: "Murkim",
         added: "2009-07-29"}).
iana(language,rmi,tag{
         type: "language",
         subtag: "rmi",
         description: "Lomavren",
         added: "2009-07-29"}).
iana(language,rmk,tag{
         type: "language",
         subtag: "rmk",
         description: "Romkun",
         added: "2009-07-29"}).
iana(language,rml,tag{
         type: "language",
         subtag: "rml",
         description: "Baltic Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmm,tag{
         type: "language",
         subtag: "rmm",
         description: "Roma",
         added: "2009-07-29"}).
iana(language,rmn,tag{
         type: "language",
         subtag: "rmn",
         description: "Balkan Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmo,tag{
         type: "language",
         subtag: "rmo",
         description: "Sinte Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmp,tag{
         type: "language",
         subtag: "rmp",
         description: "Rempi",
         added: "2009-07-29"}).
iana(language,rmq,tag{
         type: "language",
         subtag: "rmq",
         description: "Caló",
         added: "2010-03-11"}).
iana(language,rmr,tag{
         type: "language",
         subtag: "rmr",
         description: "Caló",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see emx, rmq"}).
iana(language,rms,tag{
         type: "language",
         subtag: "rms",
         description: "Romanian Sign Language",
         added: "2009-07-29"}).
iana(language,rmt,tag{
         type: "language",
         subtag: "rmt",
         description: "Domari",
         added: "2009-07-29"}).
iana(language,rmu,tag{
         type: "language",
         subtag: "rmu",
         description: "Tavringer Romani",
         added: "2009-07-29"}).
iana(language,rmv,tag{
         type: "language",
         subtag: "rmv",
         description: "Romanova",
         added: "2009-07-29"}).
iana(language,rmw,tag{
         type: "language",
         subtag: "rmw",
         description: "Welsh Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmx,tag{
         type: "language",
         subtag: "rmx",
         description: "Romam",
         added: "2009-07-29"}).
iana(language,rmy,tag{
         type: "language",
         subtag: "rmy",
         description: "Vlax Romani",
         added: "2009-07-29",
         macrolanguage: "rom"}).
iana(language,rmz,tag{
         type: "language",
         subtag: "rmz",
         description: "Marma",
         added: "2009-07-29"}).
iana(language,rna,tag{
         type: "language",
         subtag: "rna",
         description: "Runa",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,rnb,tag{
         type: "language",
         subtag: "rnb",
         description: "Brunca Sign Language",
         added: "2022-02-25"}).
iana(language,rnd,tag{
         type: "language",
         subtag: "rnd",
         description: "Ruund",
         added: "2009-07-29"}).
iana(language,rng,tag{
         type: "language",
         subtag: "rng",
         description: "Ronga",
         added: "2009-07-29"}).
iana(language,rnl,tag{
         type: "language",
         subtag: "rnl",
         description: "Ranglong",
         added: "2009-07-29"}).
iana(language,rnn,tag{
         type: "language",
         subtag: "rnn",
         description: "Roon",
         added: "2009-07-29"}).
iana(language,rnp,tag{
         type: "language",
         subtag: "rnp",
         description: "Rongpo",
         added: "2009-07-29"}).
iana(language,rnr,tag{
         type: "language",
         subtag: "rnr",
         description: "Nari Nari",
         added: "2012-08-20"}).
iana(language,rnw,tag{
         type: "language",
         subtag: "rnw",
         description: "Rungwa",
         added: "2009-07-29"}).
iana(language,roa,tag{
         type: "language",
         subtag: "roa",
         description: "Romance languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,rob,tag{
         type: "language",
         subtag: "rob",
         description: "Tae'",
         added: "2009-07-29"}).
iana(language,roc,tag{
         type: "language",
         subtag: "roc",
         description: "Cacgia Roglai",
         added: "2009-07-29"}).
iana(language,rod,tag{
         type: "language",
         subtag: "rod",
         description: "Rogo",
         added: "2009-07-29"}).
iana(language,roe,tag{
         type: "language",
         subtag: "roe",
         description: "Ronji",
         added: "2009-07-29"}).
iana(language,rof,tag{
         type: "language",
         subtag: "rof",
         description: "Rombo",
         added: "2009-07-29"}).
iana(language,rog,tag{
         type: "language",
         subtag: "rog",
         description: "Northern Roglai",
         added: "2009-07-29"}).
iana(language,rol,tag{
         type: "language",
         subtag: "rol",
         description: "Romblomanon",
         added: "2009-07-29"}).
iana(language,rom,tag{
         type: "language",
         subtag: "rom",
         description: "Romany",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,roo,tag{
         type: "language",
         subtag: "roo",
         description: "Rotokas",
         added: "2009-07-29"}).
iana(language,rop,tag{
         type: "language",
         subtag: "rop",
         description: "Kriol",
         added: "2009-07-29"}).
iana(language,ror,tag{
         type: "language",
         subtag: "ror",
         description: "Rongga",
         added: "2009-07-29"}).
iana(language,rou,tag{
         type: "language",
         subtag: "rou",
         description: "Runga",
         added: "2009-07-29"}).
iana(language,row,tag{
         type: "language",
         subtag: "row",
         description: "Dela-Oenale",
         added: "2009-07-29"}).
iana(language,rpn,tag{
         type: "language",
         subtag: "rpn",
         description: "Repanbitip",
         added: "2009-07-29"}).
iana(language,rpt,tag{
         type: "language",
         subtag: "rpt",
         description: "Rapting",
         added: "2009-07-29"}).
iana(language,rri,tag{
         type: "language",
         subtag: "rri",
         description: "Ririo",
         added: "2009-07-29"}).
iana(language,rro,tag{
         type: "language",
         subtag: "rro",
         description: "Waima",
         added: "2009-07-29"}).
iana(language,rrt,tag{
         type: "language",
         subtag: "rrt",
         description: "Arritinngithigh",
         added: "2012-08-12"}).
iana(language,rsb,tag{
         type: "language",
         subtag: "rsb",
         description: "Romano-Serbian",
         added: "2009-07-29"}).
iana(language,rsi,tag{
         type: "language",
         subtag: "rsi",
         description: "Rennellese Sign Language",
         added: "2009-07-29",
         deprecated: "2017-02-23"}).
iana(language,rsk,tag{
         type: "language",
         subtag: "rsk",
         description: ["Ruthenian","Rusyn"],
         added: "2022-02-25"}).
iana(language,rsl,tag{
         type: "language",
         subtag: "rsl",
         description: "Russian Sign Language",
         added: "2009-07-29"}).
iana(language,rsm,tag{
         type: "language",
         subtag: "rsm",
         description: "Miriwoong Sign Language",
         added: "2016-05-30"}).
iana(language,rsn,tag{
         type: "language",
         subtag: "rsn",
         description: "Rwandan Sign Language",
         added: "2022-02-25"}).
iana(language,rtc,tag{
         type: "language",
         subtag: "rtc",
         description: "Rungtu Chin",
         added: "2012-08-12"}).
iana(language,rth,tag{
         type: "language",
         subtag: "rth",
         description: "Ratahan",
         added: "2009-07-29"}).
iana(language,rtm,tag{
         type: "language",
         subtag: "rtm",
         description: "Rotuman",
         added: "2009-07-29"}).
iana(language,rts,tag{
         type: "language",
         subtag: "rts",
         description: "Yurats",
         added: "2015-02-12"}).
iana(language,rtw,tag{
         type: "language",
         subtag: "rtw",
         description: "Rathawi",
         added: "2009-07-29"}).
iana(language,rub,tag{
         type: "language",
         subtag: "rub",
         description: "Gungu",
         added: "2009-07-29"}).
iana(language,ruc,tag{
         type: "language",
         subtag: "ruc",
         description: "Ruuli",
         added: "2009-07-29"}).
iana(language,rue,tag{
         type: "language",
         subtag: "rue",
         description: "Rusyn",
         added: "2009-07-29"}).
iana(language,ruf,tag{
         type: "language",
         subtag: "ruf",
         description: "Luguru",
         added: "2009-07-29"}).
iana(language,rug,tag{
         type: "language",
         subtag: "rug",
         description: "Roviana",
         added: "2009-07-29"}).
iana(language,ruh,tag{
         type: "language",
         subtag: "ruh",
         description: "Ruga",
         added: "2009-07-29"}).
iana(language,rui,tag{
         type: "language",
         subtag: "rui",
         description: "Rufiji",
         added: "2009-07-29"}).
iana(language,ruk,tag{
         type: "language",
         subtag: "ruk",
         description: "Che",
         added: "2009-07-29"}).
iana(language,ruo,tag{
         type: "language",
         subtag: "ruo",
         description: "Istro Romanian",
         added: "2009-07-29"}).
iana(language,rup,tag{
         type: "language",
         subtag: "rup",
         description: ["Macedo-Romanian","Aromanian","Arumanian"],
         added: "2005-10-16"}).
iana(language,ruq,tag{
         type: "language",
         subtag: "ruq",
         description: "Megleno Romanian",
         added: "2009-07-29"}).
iana(language,rut,tag{
         type: "language",
         subtag: "rut",
         description: "Rutul",
         added: "2009-07-29"}).
iana(language,ruu,tag{
         type: "language",
         subtag: "ruu",
         description: "Lanas Lobu",
         added: "2009-07-29"}).
iana(language,ruy,tag{
         type: "language",
         subtag: "ruy",
         description: "Mala (Nigeria)",
         added: "2009-07-29"}).
iana(language,ruz,tag{
         type: "language",
         subtag: "ruz",
         description: "Ruma",
         added: "2009-07-29"}).
iana(language,rwa,tag{
         type: "language",
         subtag: "rwa",
         description: "Rawo",
         added: "2009-07-29"}).
iana(language,rwk,tag{
         type: "language",
         subtag: "rwk",
         description: "Rwa",
         added: "2009-07-29"}).
iana(language,rwl,tag{
         type: "language",
         subtag: "rwl",
         description: "Ruwila",
         added: "2021-02-20"}).
iana(language,rwm,tag{
         type: "language",
         subtag: "rwm",
         description: "Amba (Uganda)",
         added: "2009-07-29"}).
iana(language,rwo,tag{
         type: "language",
         subtag: "rwo",
         description: "Rawa",
         added: "2009-07-29"}).
iana(language,rwr,tag{
         type: "language",
         subtag: "rwr",
         description: "Marwari (India)",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,rxd,tag{
         type: "language",
         subtag: "rxd",
         description: "Ngardi",
         added: "2013-09-10"}).
iana(language,rxw,tag{
         type: "language",
         subtag: "rxw",
         description: ["Karuwali","Garuwali"],
         added: "2013-09-10"}).
iana(language,ryn,tag{
         type: "language",
         subtag: "ryn",
         description: "Northern Amami-Oshima",
         added: "2009-07-29"}).
iana(language,rys,tag{
         type: "language",
         subtag: "rys",
         description: "Yaeyama",
         added: "2009-07-29"}).
iana(language,ryu,tag{
         type: "language",
         subtag: "ryu",
         description: "Central Okinawan",
         added: "2009-07-29"}).
iana(language,rzh,tag{
         type: "language",
         subtag: "rzh",
         description: "Rāziḥī",
         added: "2016-05-30"}).
iana(language,saa,tag{
         type: "language",
         subtag: "saa",
         description: "Saba",
         added: "2009-07-29"}).
iana(language,sab,tag{
         type: "language",
         subtag: "sab",
         description: "Buglere",
         added: "2009-07-29"}).
iana(language,sac,tag{
         type: "language",
         subtag: "sac",
         description: "Meskwaki",
         added: "2009-07-29"}).
iana(language,sad,tag{
         type: "language",
         subtag: "sad",
         description: "Sandawe",
         added: "2005-10-16"}).
iana(language,sae,tag{
         type: "language",
         subtag: "sae",
         description: "Sabanê",
         added: "2009-07-29"}).
iana(language,saf,tag{
         type: "language",
         subtag: "saf",
         description: "Safaliba",
         added: "2009-07-29"}).
iana(language,sah,tag{
         type: "language",
         subtag: "sah",
         description: "Yakut",
         added: "2005-10-16"}).
iana(language,sai,tag{
         type: "language",
         subtag: "sai",
         description: "South American Indian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,saj,tag{
         type: "language",
         subtag: "saj",
         description: "Sahu",
         added: "2009-07-29"}).
iana(language,sak,tag{
         type: "language",
         subtag: "sak",
         description: "Sake",
         added: "2009-07-29"}).
iana(language,sal,tag{
         type: "language",
         subtag: "sal",
         description: "Salishan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,sam,tag{
         type: "language",
         subtag: "sam",
         description: "Samaritan Aramaic",
         added: "2005-10-16"}).
iana(language,sao,tag{
         type: "language",
         subtag: "sao",
         description: "Sause",
         added: "2009-07-29"}).
iana(language,sap,tag{
         type: "language",
         subtag: "sap",
         description: "Sanapaná",
         added: "2009-07-29",
         deprecated: "2014-02-28",
         comments: "see aqt, spn"}).
iana(language,saq,tag{
         type: "language",
         subtag: "saq",
         description: "Samburu",
         added: "2009-07-29"}).
iana(language,sar,tag{
         type: "language",
         subtag: "sar",
         description: "Saraveca",
         added: "2009-07-29"}).
iana(language,sas,tag{
         type: "language",
         subtag: "sas",
         description: "Sasak",
         added: "2005-10-16"}).
iana(language,sat,tag{
         type: "language",
         subtag: "sat",
         description: "Santali",
         added: "2005-10-16"}).
iana(language,sau,tag{
         type: "language",
         subtag: "sau",
         description: "Saleman",
         added: "2009-07-29"}).
iana(language,sav,tag{
         type: "language",
         subtag: "sav",
         description: "Saafi-Saafi",
         added: "2009-07-29"}).
iana(language,saw,tag{
         type: "language",
         subtag: "saw",
         description: "Sawi",
         added: "2009-07-29"}).
iana(language,sax,tag{
         type: "language",
         subtag: "sax",
         description: "Sa",
         added: "2009-07-29"}).
iana(language,say,tag{
         type: "language",
         subtag: "say",
         description: "Saya",
         added: "2009-07-29"}).
iana(language,saz,tag{
         type: "language",
         subtag: "saz",
         description: "Saurashtra",
         added: "2009-07-29"}).
iana(language,sba,tag{
         type: "language",
         subtag: "sba",
         description: "Ngambay",
         added: "2009-07-29"}).
iana(language,sbb,tag{
         type: "language",
         subtag: "sbb",
         description: "Simbo",
         added: "2009-07-29"}).
iana(language,sbc,tag{
         type: "language",
         subtag: "sbc",
         description: "Kele (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,sbd,tag{
         type: "language",
         subtag: "sbd",
         description: "Southern Samo",
         added: "2009-07-29"}).
iana(language,sbe,tag{
         type: "language",
         subtag: "sbe",
         description: "Saliba",
         added: "2009-07-29"}).
iana(language,sbf,tag{
         type: "language",
         subtag: "sbf",
         description: ["Chabu","Shabo"],
         added: "2009-07-29"}).
iana(language,sbg,tag{
         type: "language",
         subtag: "sbg",
         description: "Seget",
         added: "2009-07-29"}).
iana(language,sbh,tag{
         type: "language",
         subtag: "sbh",
         description: "Sori-Harengan",
         added: "2009-07-29"}).
iana(language,sbi,tag{
         type: "language",
         subtag: "sbi",
         description: "Seti",
         added: "2009-07-29"}).
iana(language,sbj,tag{
         type: "language",
         subtag: "sbj",
         description: "Surbakhal",
         added: "2009-07-29"}).
iana(language,sbk,tag{
         type: "language",
         subtag: "sbk",
         description: "Safwa",
         added: "2009-07-29"}).
iana(language,sbl,tag{
         type: "language",
         subtag: "sbl",
         description: "Botolan Sambal",
         added: "2009-07-29"}).
iana(language,sbm,tag{
         type: "language",
         subtag: "sbm",
         description: "Sagala",
         added: "2009-07-29"}).
iana(language,sbn,tag{
         type: "language",
         subtag: "sbn",
         description: "Sindhi Bhil",
         added: "2009-07-29"}).
iana(language,sbo,tag{
         type: "language",
         subtag: "sbo",
         description: "Sabüm",
         added: "2009-07-29"}).
iana(language,sbp,tag{
         type: "language",
         subtag: "sbp",
         description: "Sangu (Tanzania)",
         added: "2009-07-29"}).
iana(language,sbq,tag{
         type: "language",
         subtag: "sbq",
         description: "Sileibi",
         added: "2009-07-29"}).
iana(language,sbr,tag{
         type: "language",
         subtag: "sbr",
         description: "Sembakung Murut",
         added: "2009-07-29"}).
iana(language,sbs,tag{
         type: "language",
         subtag: "sbs",
         description: "Subiya",
         added: "2009-07-29"}).
iana(language,sbt,tag{
         type: "language",
         subtag: "sbt",
         description: "Kimki",
         added: "2009-07-29"}).
iana(language,sbu,tag{
         type: "language",
         subtag: "sbu",
         description: "Stod Bhoti",
         added: "2009-07-29"}).
iana(language,sbv,tag{
         type: "language",
         subtag: "sbv",
         description: "Sabine",
         added: "2009-07-29"}).
iana(language,sbw,tag{
         type: "language",
         subtag: "sbw",
         description: "Simba",
         added: "2009-07-29"}).
iana(language,sbx,tag{
         type: "language",
         subtag: "sbx",
         description: "Seberuang",
         added: "2009-07-29"}).
iana(language,sby,tag{
         type: "language",
         subtag: "sby",
         description: "Soli",
         added: "2009-07-29"}).
iana(language,sbz,tag{
         type: "language",
         subtag: "sbz",
         description: "Sara Kaba",
         added: "2009-07-29"}).
iana(language,sca,tag{
         type: "language",
         subtag: "sca",
         description: "Sansu",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "hle"}).
iana(language,scb,tag{
         type: "language",
         subtag: "scb",
         description: "Chut",
         added: "2009-07-29"}).
iana(language,sce,tag{
         type: "language",
         subtag: "sce",
         description: "Dongxiang",
         added: "2009-07-29"}).
iana(language,scf,tag{
         type: "language",
         subtag: "scf",
         description: "San Miguel Creole French",
         added: "2009-07-29"}).
iana(language,scg,tag{
         type: "language",
         subtag: "scg",
         description: "Sanggau",
         added: "2009-07-29"}).
iana(language,sch,tag{
         type: "language",
         subtag: "sch",
         description: "Sakachep",
         added: "2009-07-29"}).
iana(language,sci,tag{
         type: "language",
         subtag: "sci",
         description: "Sri Lankan Creole Malay",
         added: "2009-07-29"}).
iana(language,sck,tag{
         type: "language",
         subtag: "sck",
         description: "Sadri",
         added: "2009-07-29"}).
iana(language,scl,tag{
         type: "language",
         subtag: "scl",
         description: "Shina",
         added: "2009-07-29"}).
iana(language,scn,tag{
         type: "language",
         subtag: "scn",
         description: "Sicilian",
         added: "2005-10-16"}).
iana(language,sco,tag{
         type: "language",
         subtag: "sco",
         description: "Scots",
         added: "2005-10-16"}).
iana(language,scp,tag{
         type: "language",
         subtag: "scp",
         description: ["Hyolmo","Helambu Sherpa"],
         added: "2009-07-29"}).
iana(language,scq,tag{
         type: "language",
         subtag: "scq",
         description: "Sa'och",
         added: "2009-07-29"}).
iana(language,scs,tag{
         type: "language",
         subtag: "scs",
         description: "North Slavey",
         added: "2009-07-29",
         macrolanguage: "den"}).
iana(language,sct,tag{
         type: "language",
         subtag: "sct",
         description: "Southern Katang",
         added: "2017-02-23"}).
iana(language,scu,tag{
         type: "language",
         subtag: "scu",
         description: "Shumcho",
         added: "2009-07-29"}).
iana(language,scv,tag{
         type: "language",
         subtag: "scv",
         description: "Sheni",
         added: "2009-07-29"}).
iana(language,scw,tag{
         type: "language",
         subtag: "scw",
         description: "Sha",
         added: "2009-07-29"}).
iana(language,scx,tag{
         type: "language",
         subtag: "scx",
         description: "Sicel",
         added: "2009-07-29"}).
iana(language,sda,tag{
         type: "language",
         subtag: "sda",
         description: "Toraja-Sa'dan",
         added: "2009-07-29"}).
iana(language,sdb,tag{
         type: "language",
         subtag: "sdb",
         description: "Shabak",
         added: "2009-07-29"}).
iana(language,sdc,tag{
         type: "language",
         subtag: "sdc",
         description: "Sassarese Sardinian",
         added: "2009-07-29",
         macrolanguage: "sc"}).
iana(language,sde,tag{
         type: "language",
         subtag: "sde",
         description: "Surubu",
         added: "2009-07-29"}).
iana(language,sdf,tag{
         type: "language",
         subtag: "sdf",
         description: "Sarli",
         added: "2009-07-29"}).
iana(language,sdg,tag{
         type: "language",
         subtag: "sdg",
         description: "Savi",
         added: "2009-07-29"}).
iana(language,sdh,tag{
         type: "language",
         subtag: "sdh",
         description: "Southern Kurdish",
         added: "2009-07-29",
         macrolanguage: "ku"}).
iana(language,sdj,tag{
         type: "language",
         subtag: "sdj",
         description: "Suundi",
         added: "2009-07-29"}).
iana(language,sdk,tag{
         type: "language",
         subtag: "sdk",
         description: "Sos Kundi",
         added: "2009-07-29"}).
iana(language,sdl,tag{
         type: "language",
         subtag: "sdl",
         description: "Saudi Arabian Sign Language",
         added: "2009-07-29"}).
iana(language,sdm,tag{
         type: "language",
         subtag: "sdm",
         description: "Semandang",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         comments: "see ebc, gef, sdq"}).
iana(language,sdn,tag{
         type: "language",
         subtag: "sdn",
         description: "Gallurese Sardinian",
         added: "2009-07-29",
         macrolanguage: "sc"}).
iana(language,sdo,tag{
         type: "language",
         subtag: "sdo",
         description: "Bukar-Sadung Bidayuh",
         added: "2009-07-29"}).
iana(language,sdp,tag{
         type: "language",
         subtag: "sdp",
         description: "Sherdukpen",
         added: "2009-07-29"}).
iana(language,sdq,tag{
         type: "language",
         subtag: "sdq",
         description: "Semandang",
         added: "2020-03-28"}).
iana(language,sdr,tag{
         type: "language",
         subtag: "sdr",
         description: "Oraon Sadri",
         added: "2009-07-29"}).
iana(language,sds,tag{
         type: "language",
         subtag: "sds",
         description: "Sened",
         added: "2009-07-29"}).
iana(language,sdt,tag{
         type: "language",
         subtag: "sdt",
         description: "Shuadit",
         added: "2009-07-29"}).
iana(language,sdu,tag{
         type: "language",
         subtag: "sdu",
         description: "Sarudu",
         added: "2009-07-29"}).
iana(language,sdv,tag{
         type: "language",
         subtag: "sdv",
         description: "Eastern Sudanic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,sdx,tag{
         type: "language",
         subtag: "sdx",
         description: "Sibu Melanau",
         added: "2009-07-29"}).
iana(language,sdz,tag{
         type: "language",
         subtag: "sdz",
         description: "Sallands",
         added: "2009-07-29"}).
iana(language,sea,tag{
         type: "language",
         subtag: "sea",
         description: "Semai",
         added: "2009-07-29"}).
iana(language,seb,tag{
         type: "language",
         subtag: "seb",
         description: "Shempire Senoufo",
         added: "2009-07-29"}).
iana(language,sec,tag{
         type: "language",
         subtag: "sec",
         description: "Sechelt",
         added: "2009-07-29"}).
iana(language,sed,tag{
         type: "language",
         subtag: "sed",
         description: "Sedang",
         added: "2009-07-29"}).
iana(language,see,tag{
         type: "language",
         subtag: "see",
         description: "Seneca",
         added: "2009-07-29"}).
iana(language,sef,tag{
         type: "language",
         subtag: "sef",
         description: "Cebaara Senoufo",
         added: "2009-07-29"}).
iana(language,seg,tag{
         type: "language",
         subtag: "seg",
         description: "Segeju",
         added: "2009-07-29"}).
iana(language,seh,tag{
         type: "language",
         subtag: "seh",
         description: "Sena",
         added: "2009-07-29"}).
iana(language,sei,tag{
         type: "language",
         subtag: "sei",
         description: "Seri",
         added: "2009-07-29"}).
iana(language,sej,tag{
         type: "language",
         subtag: "sej",
         description: "Sene",
         added: "2009-07-29"}).
iana(language,sek,tag{
         type: "language",
         subtag: "sek",
         description: "Sekani",
         added: "2009-07-29"}).
iana(language,sel,tag{
         type: "language",
         subtag: "sel",
         description: "Selkup",
         added: "2005-10-16"}).
iana(language,sem,tag{
         type: "language",
         subtag: "sem",
         description: "Semitic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,sen,tag{
         type: "language",
         subtag: "sen",
         description: "Nanerigé Sénoufo",
         added: "2009-07-29"}).
iana(language,seo,tag{
         type: "language",
         subtag: "seo",
         description: "Suarmin",
         added: "2009-07-29"}).
iana(language,sep,tag{
         type: "language",
         subtag: "sep",
         description: "Sìcìté Sénoufo",
         added: "2009-07-29"}).
iana(language,seq,tag{
         type: "language",
         subtag: "seq",
         description: "Senara Sénoufo",
         added: "2009-07-29"}).
iana(language,ser,tag{
         type: "language",
         subtag: "ser",
         description: "Serrano",
         added: "2009-07-29"}).
iana(language,ses,tag{
         type: "language",
         subtag: "ses",
         description: "Koyraboro Senni Songhai",
         added: "2009-07-29"}).
iana(language,set,tag{
         type: "language",
         subtag: "set",
         description: "Sentani",
         added: "2009-07-29"}).
iana(language,seu,tag{
         type: "language",
         subtag: "seu",
         description: "Serui-Laut",
         added: "2009-07-29"}).
iana(language,sev,tag{
         type: "language",
         subtag: "sev",
         description: "Nyarafolo Senoufo",
         added: "2009-07-29"}).
iana(language,sew,tag{
         type: "language",
         subtag: "sew",
         description: "Sewa Bay",
         added: "2009-07-29"}).
iana(language,sey,tag{
         type: "language",
         subtag: "sey",
         description: "Secoya",
         added: "2009-07-29"}).
iana(language,sez,tag{
         type: "language",
         subtag: "sez",
         description: "Senthang Chin",
         added: "2009-07-29"}).
iana(language,sfb,tag{
         type: "language",
         subtag: "sfb",
         description: ["Langue des signes de Belgique Francophone","French Belgian Sign Language"],
         added: "2009-07-29"}).
iana(language,sfe,tag{
         type: "language",
         subtag: "sfe",
         description: "Eastern Subanen",
         added: "2012-08-12"}).
iana(language,sfm,tag{
         type: "language",
         subtag: "sfm",
         description: "Small Flowery Miao",
         added: "2009-07-29",
         macrolanguage: "hmn"}).
iana(language,sfs,tag{
         type: "language",
         subtag: "sfs",
         description: "South African Sign Language",
         added: "2009-07-29"}).
iana(language,sfw,tag{
         type: "language",
         subtag: "sfw",
         description: "Sehwi",
         added: "2009-07-29"}).
iana(language,sga,tag{
         type: "language",
         subtag: "sga",
         description: "Old Irish (to 900)",
         added: "2005-10-16"}).
iana(language,sgb,tag{
         type: "language",
         subtag: "sgb",
         description: "Mag-antsi Ayta",
         added: "2009-07-29"}).
iana(language,sgc,tag{
         type: "language",
         subtag: "sgc",
         description: "Kipsigis",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,sgd,tag{
         type: "language",
         subtag: "sgd",
         description: "Surigaonon",
         added: "2010-03-11"}).
iana(language,sge,tag{
         type: "language",
         subtag: "sge",
         description: "Segai",
         added: "2009-07-29"}).
iana(language,sgg,tag{
         type: "language",
         subtag: "sgg",
         description: "Swiss-German Sign Language",
         added: "2009-07-29"}).
iana(language,sgh,tag{
         type: "language",
         subtag: "sgh",
         description: "Shughni",
         added: "2009-07-29"}).
iana(language,sgi,tag{
         type: "language",
         subtag: "sgi",
         description: "Suga",
         added: "2009-07-29"}).
iana(language,sgj,tag{
         type: "language",
         subtag: "sgj",
         description: "Surgujia",
         added: "2012-08-12"}).
iana(language,sgk,tag{
         type: "language",
         subtag: "sgk",
         description: "Sangkong",
         added: "2009-07-29"}).
iana(language,sgl,tag{
         type: "language",
         subtag: "sgl",
         description: "Sanglechi-Ishkashimi",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see isk, sgy"}).
iana(language,sgm,tag{
         type: "language",
         subtag: "sgm",
         description: "Singa",
         added: "2009-07-29"}).
iana(language,sgn,tag{
         type: "language",
         subtag: "sgn",
         description: "Sign languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,sgo,tag{
         type: "language",
         subtag: "sgo",
         description: "Songa",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,sgp,tag{
         type: "language",
         subtag: "sgp",
         description: "Singpho",
         added: "2009-07-29"}).
iana(language,sgr,tag{
         type: "language",
         subtag: "sgr",
         description: "Sangisari",
         added: "2009-07-29"}).
iana(language,sgs,tag{
         type: "language",
         subtag: "sgs",
         description: "Samogitian",
         added: "2010-07-26"}).
iana(language,sgt,tag{
         type: "language",
         subtag: "sgt",
         description: "Brokpake",
         added: "2009-07-29"}).
iana(language,sgu,tag{
         type: "language",
         subtag: "sgu",
         description: "Salas",
         added: "2009-07-29"}).
iana(language,sgw,tag{
         type: "language",
         subtag: "sgw",
         description: "Sebat Bet Gurage",
         added: "2009-07-29"}).
iana(language,sgx,tag{
         type: "language",
         subtag: "sgx",
         description: "Sierra Leone Sign Language",
         added: "2009-07-29"}).
iana(language,sgy,tag{
         type: "language",
         subtag: "sgy",
         description: "Sanglechi",
         added: "2010-03-11"}).
iana(language,sgz,tag{
         type: "language",
         subtag: "sgz",
         description: "Sursurunga",
         added: "2009-07-29"}).
iana(language,sha,tag{
         type: "language",
         subtag: "sha",
         description: "Shall-Zwall",
         added: "2009-07-29"}).
iana(language,shb,tag{
         type: "language",
         subtag: "shb",
         description: "Ninam",
         added: "2009-07-29"}).
iana(language,shc,tag{
         type: "language",
         subtag: "shc",
         description: "Sonde",
         added: "2009-07-29"}).
iana(language,shd,tag{
         type: "language",
         subtag: "shd",
         description: "Kundal Shahi",
         added: "2010-03-11"}).
iana(language,she,tag{
         type: "language",
         subtag: "she",
         description: "Sheko",
         added: "2009-07-29"}).
iana(language,shg,tag{
         type: "language",
         subtag: "shg",
         description: "Shua",
         added: "2009-07-29"}).
iana(language,shh,tag{
         type: "language",
         subtag: "shh",
         description: "Shoshoni",
         added: "2009-07-29"}).
iana(language,shi,tag{
         type: "language",
         subtag: "shi",
         description: "Tachelhit",
         added: "2009-07-29"}).
iana(language,shj,tag{
         type: "language",
         subtag: "shj",
         description: "Shatt",
         added: "2009-07-29"}).
iana(language,shk,tag{
         type: "language",
         subtag: "shk",
         description: "Shilluk",
         added: "2009-07-29"}).
iana(language,shl,tag{
         type: "language",
         subtag: "shl",
         description: "Shendu",
         added: "2009-07-29"}).
iana(language,shm,tag{
         type: "language",
         subtag: "shm",
         description: "Shahrudi",
         added: "2009-07-29"}).
iana(language,shn,tag{
         type: "language",
         subtag: "shn",
         description: "Shan",
         added: "2005-10-16"}).
iana(language,sho,tag{
         type: "language",
         subtag: "sho",
         description: "Shanga",
         added: "2009-07-29"}).
iana(language,shp,tag{
         type: "language",
         subtag: "shp",
         description: "Shipibo-Conibo",
         added: "2009-07-29"}).
iana(language,shq,tag{
         type: "language",
         subtag: "shq",
         description: "Sala",
         added: "2009-07-29"}).
iana(language,shr,tag{
         type: "language",
         subtag: "shr",
         description: "Shi",
         added: "2009-07-29"}).
iana(language,shs,tag{
         type: "language",
         subtag: "shs",
         description: "Shuswap",
         added: "2009-07-29"}).
iana(language,sht,tag{
         type: "language",
         subtag: "sht",
         description: "Shasta",
         added: "2009-07-29"}).
iana(language,shu,tag{
         type: "language",
         subtag: "shu",
         description: "Chadian Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,shv,tag{
         type: "language",
         subtag: "shv",
         description: "Shehri",
         added: "2009-07-29"}).
iana(language,shw,tag{
         type: "language",
         subtag: "shw",
         description: "Shwai",
         added: "2009-07-29"}).
iana(language,shx,tag{
         type: "language",
         subtag: "shx",
         description: "She",
         added: "2009-07-29"}).
iana(language,shy,tag{
         type: "language",
         subtag: "shy",
         description: "Tachawit",
         added: "2009-07-29"}).
iana(language,shz,tag{
         type: "language",
         subtag: "shz",
         description: "Syenara Senoufo",
         added: "2009-07-29"}).
iana(language,sia,tag{
         type: "language",
         subtag: "sia",
         description: "Akkala Sami",
         added: "2009-07-29"}).
iana(language,sib,tag{
         type: "language",
         subtag: "sib",
         description: "Sebop",
         added: "2009-07-29"}).
iana(language,sid,tag{
         type: "language",
         subtag: "sid",
         description: "Sidamo",
         added: "2005-10-16"}).
iana(language,sie,tag{
         type: "language",
         subtag: "sie",
         description: "Simaa",
         added: "2009-07-29"}).
iana(language,sif,tag{
         type: "language",
         subtag: "sif",
         description: "Siamou",
         added: "2009-07-29"}).
iana(language,sig,tag{
         type: "language",
         subtag: "sig",
         description: "Paasaal",
         added: "2009-07-29"}).
iana(language,sih,tag{
         type: "language",
         subtag: "sih",
         description: ["Zire","Sîshëë"],
         added: "2009-07-29"}).
iana(language,sii,tag{
         type: "language",
         subtag: "sii",
         description: "Shom Peng",
         added: "2009-07-29"}).
iana(language,sij,tag{
         type: "language",
         subtag: "sij",
         description: "Numbami",
         added: "2009-07-29"}).
iana(language,sik,tag{
         type: "language",
         subtag: "sik",
         description: "Sikiana",
         added: "2009-07-29"}).
iana(language,sil,tag{
         type: "language",
         subtag: "sil",
         description: "Tumulung Sisaala",
         added: "2009-07-29"}).
iana(language,sim,tag{
         type: "language",
         subtag: "sim",
         description: "Mende (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,sio,tag{
         type: "language",
         subtag: "sio",
         description: "Siouan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,sip,tag{
         type: "language",
         subtag: "sip",
         description: "Sikkimese",
         added: "2009-07-29"}).
iana(language,siq,tag{
         type: "language",
         subtag: "siq",
         description: "Sonia",
         added: "2009-07-29"}).
iana(language,sir,tag{
         type: "language",
         subtag: "sir",
         description: "Siri",
         added: "2009-07-29"}).
iana(language,sis,tag{
         type: "language",
         subtag: "sis",
         description: "Siuslaw",
         added: "2009-07-29"}).
iana(language,sit,tag{
         type: "language",
         subtag: "sit",
         description: "Sino-Tibetan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,siu,tag{
         type: "language",
         subtag: "siu",
         description: "Sinagen",
         added: "2009-07-29"}).
iana(language,siv,tag{
         type: "language",
         subtag: "siv",
         description: "Sumariup",
         added: "2009-07-29"}).
iana(language,siw,tag{
         type: "language",
         subtag: "siw",
         description: "Siwai",
         added: "2009-07-29"}).
iana(language,six,tag{
         type: "language",
         subtag: "six",
         description: "Sumau",
         added: "2009-07-29"}).
iana(language,siy,tag{
         type: "language",
         subtag: "siy",
         description: "Sivandi",
         added: "2009-07-29"}).
iana(language,siz,tag{
         type: "language",
         subtag: "siz",
         description: "Siwi",
         added: "2009-07-29"}).
iana(language,sja,tag{
         type: "language",
         subtag: "sja",
         description: "Epena",
         added: "2009-07-29"}).
iana(language,sjb,tag{
         type: "language",
         subtag: "sjb",
         description: "Sajau Basap",
         added: "2009-07-29"}).
iana(language,sjd,tag{
         type: "language",
         subtag: "sjd",
         description: "Kildin Sami",
         added: "2009-07-29"}).
iana(language,sje,tag{
         type: "language",
         subtag: "sje",
         description: "Pite Sami",
         added: "2009-07-29"}).
iana(language,sjg,tag{
         type: "language",
         subtag: "sjg",
         description: "Assangori",
         added: "2009-07-29"}).
iana(language,sjk,tag{
         type: "language",
         subtag: "sjk",
         description: "Kemi Sami",
         added: "2009-07-29"}).
iana(language,sjl,tag{
         type: "language",
         subtag: "sjl",
         description: ["Sajalong","Miji"],
         added: "2009-07-29"}).
iana(language,sjm,tag{
         type: "language",
         subtag: "sjm",
         description: "Mapun",
         added: "2009-07-29"}).
iana(language,sjn,tag{
         type: "language",
         subtag: "sjn",
         description: "Sindarin",
         added: "2009-07-29"}).
iana(language,sjo,tag{
         type: "language",
         subtag: "sjo",
         description: "Xibe",
         added: "2009-07-29"}).
iana(language,sjp,tag{
         type: "language",
         subtag: "sjp",
         description: "Surjapuri",
         added: "2009-07-29"}).
iana(language,sjr,tag{
         type: "language",
         subtag: "sjr",
         description: "Siar-Lak",
         added: "2009-07-29"}).
iana(language,sjs,tag{
         type: "language",
         subtag: "sjs",
         description: "Senhaja De Srair",
         added: "2009-07-29"}).
iana(language,sjt,tag{
         type: "language",
         subtag: "sjt",
         description: "Ter Sami",
         added: "2009-07-29"}).
iana(language,sju,tag{
         type: "language",
         subtag: "sju",
         description: "Ume Sami",
         added: "2009-07-29"}).
iana(language,sjw,tag{
         type: "language",
         subtag: "sjw",
         description: "Shawnee",
         added: "2009-07-29"}).
iana(language,ska,tag{
         type: "language",
         subtag: "ska",
         description: "Skagit",
         added: "2009-07-29"}).
iana(language,skb,tag{
         type: "language",
         subtag: "skb",
         description: "Saek",
         added: "2009-07-29"}).
iana(language,skc,tag{
         type: "language",
         subtag: "skc",
         description: "Ma Manda",
         added: "2009-07-29"}).
iana(language,skd,tag{
         type: "language",
         subtag: "skd",
         description: "Southern Sierra Miwok",
         added: "2009-07-29"}).
iana(language,ske,tag{
         type: "language",
         subtag: "ske",
         description: "Seke (Vanuatu)",
         added: "2009-07-29"}).
iana(language,skf,tag{
         type: "language",
         subtag: "skf",
         description: "Sakirabiá",
         added: "2009-07-29"}).
iana(language,skg,tag{
         type: "language",
         subtag: "skg",
         description: "Sakalava Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,skh,tag{
         type: "language",
         subtag: "skh",
         description: "Sikule",
         added: "2009-07-29"}).
iana(language,ski,tag{
         type: "language",
         subtag: "ski",
         description: "Sika",
         added: "2009-07-29"}).
iana(language,skj,tag{
         type: "language",
         subtag: "skj",
         description: "Seke (Nepal)",
         added: "2009-07-29"}).
iana(language,skk,tag{
         type: "language",
         subtag: "skk",
         description: "Sok",
         added: "2009-07-29",
         deprecated: "2017-02-23",
         preferred_value: "oyb"}).
iana(language,skm,tag{
         type: "language",
         subtag: "skm",
         description: "Kutong",
         added: "2009-07-29"}).
iana(language,skn,tag{
         type: "language",
         subtag: "skn",
         description: "Kolibugan Subanon",
         added: "2009-07-29"}).
iana(language,sko,tag{
         type: "language",
         subtag: "sko",
         description: "Seko Tengah",
         added: "2009-07-29"}).
iana(language,skp,tag{
         type: "language",
         subtag: "skp",
         description: "Sekapan",
         added: "2009-07-29"}).
iana(language,skq,tag{
         type: "language",
         subtag: "skq",
         description: "Sininkere",
         added: "2009-07-29"}).
iana(language,skr,tag{
         type: "language",
         subtag: "skr",
         description: ["Saraiki","Seraiki"],
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,sks,tag{
         type: "language",
         subtag: "sks",
         description: "Maia",
         added: "2009-07-29"}).
iana(language,skt,tag{
         type: "language",
         subtag: "skt",
         description: "Sakata",
         added: "2009-07-29"}).
iana(language,sku,tag{
         type: "language",
         subtag: "sku",
         description: "Sakao",
         added: "2009-07-29"}).
iana(language,skv,tag{
         type: "language",
         subtag: "skv",
         description: "Skou",
         added: "2009-07-29"}).
iana(language,skw,tag{
         type: "language",
         subtag: "skw",
         description: "Skepi Creole Dutch",
         added: "2009-07-29"}).
iana(language,skx,tag{
         type: "language",
         subtag: "skx",
         description: "Seko Padang",
         added: "2009-07-29"}).
iana(language,sky,tag{
         type: "language",
         subtag: "sky",
         description: "Sikaiana",
         added: "2009-07-29"}).
iana(language,skz,tag{
         type: "language",
         subtag: "skz",
         description: "Sekar",
         added: "2009-07-29"}).
iana(language,sla,tag{
         type: "language",
         subtag: "sla",
         description: "Slavic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,slc,tag{
         type: "language",
         subtag: "slc",
         description: "Sáliba",
         added: "2009-07-29"}).
iana(language,sld,tag{
         type: "language",
         subtag: "sld",
         description: "Sissala",
         added: "2009-07-29"}).
iana(language,sle,tag{
         type: "language",
         subtag: "sle",
         description: "Sholaga",
         added: "2009-07-29"}).
iana(language,slf,tag{
         type: "language",
         subtag: "slf",
         description: "Swiss-Italian Sign Language",
         added: "2009-07-29"}).
iana(language,slg,tag{
         type: "language",
         subtag: "slg",
         description: "Selungai Murut",
         added: "2009-07-29"}).
iana(language,slh,tag{
         type: "language",
         subtag: "slh",
         description: "Southern Puget Sound Salish",
         added: "2009-07-29"}).
iana(language,sli,tag{
         type: "language",
         subtag: "sli",
         description: "Lower Silesian",
         added: "2009-07-29"}).
iana(language,slj,tag{
         type: "language",
         subtag: "slj",
         description: "Salumá",
         added: "2009-07-29"}).
iana(language,sll,tag{
         type: "language",
         subtag: "sll",
         description: "Salt-Yui",
         added: "2009-07-29"}).
iana(language,slm,tag{
         type: "language",
         subtag: "slm",
         description: "Pangutaran Sama",
         added: "2009-07-29"}).
iana(language,sln,tag{
         type: "language",
         subtag: "sln",
         description: "Salinan",
         added: "2009-07-29"}).
iana(language,slp,tag{
         type: "language",
         subtag: "slp",
         description: "Lamaholot",
         added: "2009-07-29"}).
iana(language,slq,tag{
         type: "language",
         subtag: "slq",
         description: "Salchuq",
         added: "2009-07-29"}).
iana(language,slr,tag{
         type: "language",
         subtag: "slr",
         description: "Salar",
         added: "2009-07-29"}).
iana(language,sls,tag{
         type: "language",
         subtag: "sls",
         description: "Singapore Sign Language",
         added: "2009-07-29"}).
iana(language,slt,tag{
         type: "language",
         subtag: "slt",
         description: "Sila",
         added: "2009-07-29"}).
iana(language,slu,tag{
         type: "language",
         subtag: "slu",
         description: "Selaru",
         added: "2009-07-29"}).
iana(language,slw,tag{
         type: "language",
         subtag: "slw",
         description: "Sialum",
         added: "2009-07-29"}).
iana(language,slx,tag{
         type: "language",
         subtag: "slx",
         description: "Salampasu",
         added: "2009-07-29"}).
iana(language,sly,tag{
         type: "language",
         subtag: "sly",
         description: "Selayar",
         added: "2009-07-29"}).
iana(language,slz,tag{
         type: "language",
         subtag: "slz",
         description: "Ma'ya",
         added: "2009-07-29"}).
iana(language,sma,tag{
         type: "language",
         subtag: "sma",
         description: "Southern Sami",
         added: "2005-10-16"}).
iana(language,smb,tag{
         type: "language",
         subtag: "smb",
         description: "Simbari",
         added: "2009-07-29"}).
iana(language,smc,tag{
         type: "language",
         subtag: "smc",
         description: "Som",
         added: "2009-07-29"}).
iana(language,smd,tag{
         type: "language",
         subtag: "smd",
         description: "Sama",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         preferred_value: "kmb"}).
iana(language,smf,tag{
         type: "language",
         subtag: "smf",
         description: "Auwe",
         added: "2009-07-29"}).
iana(language,smg,tag{
         type: "language",
         subtag: "smg",
         description: "Simbali",
         added: "2009-07-29"}).
iana(language,smh,tag{
         type: "language",
         subtag: "smh",
         description: "Samei",
         added: "2009-07-29"}).
iana(language,smi,tag{
         type: "language",
         subtag: "smi",
         description: "Sami languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,smj,tag{
         type: "language",
         subtag: "smj",
         description: "Lule Sami",
         added: "2005-10-16"}).
iana(language,smk,tag{
         type: "language",
         subtag: "smk",
         description: "Bolinao",
         added: "2009-07-29"}).
iana(language,sml,tag{
         type: "language",
         subtag: "sml",
         description: "Central Sama",
         added: "2009-07-29"}).
iana(language,smm,tag{
         type: "language",
         subtag: "smm",
         description: "Musasa",
         added: "2009-07-29"}).
iana(language,smn,tag{
         type: "language",
         subtag: "smn",
         description: "Inari Sami",
         added: "2005-10-16"}).
iana(language,smp,tag{
         type: "language",
         subtag: "smp",
         description: "Samaritan",
         added: "2009-07-29"}).
iana(language,smq,tag{
         type: "language",
         subtag: "smq",
         description: "Samo",
         added: "2009-07-29"}).
iana(language,smr,tag{
         type: "language",
         subtag: "smr",
         description: "Simeulue",
         added: "2009-07-29"}).
iana(language,sms,tag{
         type: "language",
         subtag: "sms",
         description: "Skolt Sami",
         added: "2005-10-16"}).
iana(language,smt,tag{
         type: "language",
         subtag: "smt",
         description: "Simte",
         added: "2009-07-29"}).
iana(language,smu,tag{
         type: "language",
         subtag: "smu",
         description: "Somray",
         added: "2009-07-29"}).
iana(language,smv,tag{
         type: "language",
         subtag: "smv",
         description: "Samvedi",
         added: "2009-07-29"}).
iana(language,smw,tag{
         type: "language",
         subtag: "smw",
         description: "Sumbawa",
         added: "2009-07-29"}).
iana(language,smx,tag{
         type: "language",
         subtag: "smx",
         description: "Samba",
         added: "2009-07-29"}).
iana(language,smy,tag{
         type: "language",
         subtag: "smy",
         description: "Semnani",
         added: "2009-07-29"}).
iana(language,smz,tag{
         type: "language",
         subtag: "smz",
         description: "Simeku",
         added: "2009-07-29"}).
iana(language,snb,tag{
         type: "language",
         subtag: "snb",
         description: "Sebuyau",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         preferred_value: "iba"}).
iana(language,snc,tag{
         type: "language",
         subtag: "snc",
         description: "Sinaugoro",
         added: "2009-07-29"}).
iana(language,sne,tag{
         type: "language",
         subtag: "sne",
         description: "Bau Bidayuh",
         added: "2009-07-29"}).
iana(language,snf,tag{
         type: "language",
         subtag: "snf",
         description: "Noon",
         added: "2009-07-29"}).
iana(language,sng,tag{
         type: "language",
         subtag: "sng",
         description: "Sanga (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,snh,tag{
         type: "language",
         subtag: "snh",
         description: "Shinabo",
         added: "2009-07-29",
         deprecated: "2017-02-23"}).
iana(language,sni,tag{
         type: "language",
         subtag: "sni",
         description: "Sensi",
         added: "2009-07-29"}).
iana(language,snj,tag{
         type: "language",
         subtag: "snj",
         description: "Riverain Sango",
         added: "2009-07-29"}).
iana(language,snk,tag{
         type: "language",
         subtag: "snk",
         description: "Soninke",
         added: "2005-10-16"}).
iana(language,snl,tag{
         type: "language",
         subtag: "snl",
         description: "Sangil",
         added: "2009-07-29"}).
iana(language,snm,tag{
         type: "language",
         subtag: "snm",
         description: "Southern Ma'di",
         added: "2009-07-29"}).
iana(language,snn,tag{
         type: "language",
         subtag: "snn",
         description: "Siona",
         added: "2009-07-29"}).
iana(language,sno,tag{
         type: "language",
         subtag: "sno",
         description: "Snohomish",
         added: "2009-07-29"}).
iana(language,snp,tag{
         type: "language",
         subtag: "snp",
         description: "Siane",
         added: "2009-07-29"}).
iana(language,snq,tag{
         type: "language",
         subtag: "snq",
         description: "Sangu (Gabon)",
         added: "2009-07-29"}).
iana(language,snr,tag{
         type: "language",
         subtag: "snr",
         description: "Sihan",
         added: "2009-07-29"}).
iana(language,sns,tag{
         type: "language",
         subtag: "sns",
         description: ["South West Bay","Nahavaq"],
         added: "2009-07-29"}).
iana(language,snu,tag{
         type: "language",
         subtag: "snu",
         description: ["Senggi","Viid"],
         added: "2009-07-29"}).
iana(language,snv,tag{
         type: "language",
         subtag: "snv",
         description: "Sa'ban",
         added: "2009-07-29"}).
iana(language,snw,tag{
         type: "language",
         subtag: "snw",
         description: "Selee",
         added: "2009-07-29"}).
iana(language,snx,tag{
         type: "language",
         subtag: "snx",
         description: "Sam",
         added: "2009-07-29"}).
iana(language,sny,tag{
         type: "language",
         subtag: "sny",
         description: "Saniyo-Hiyewe",
         added: "2009-07-29"}).
iana(language,snz,tag{
         type: "language",
         subtag: "snz",
         description: "Kou",
         added: "2009-07-29"}).
iana(language,soa,tag{
         type: "language",
         subtag: "soa",
         description: "Thai Song",
         added: "2009-07-29"}).
iana(language,sob,tag{
         type: "language",
         subtag: "sob",
         description: "Sobei",
         added: "2009-07-29"}).
iana(language,soc,tag{
         type: "language",
         subtag: "soc",
         description: "So (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,sod,tag{
         type: "language",
         subtag: "sod",
         description: "Songoora",
         added: "2009-07-29"}).
iana(language,soe,tag{
         type: "language",
         subtag: "soe",
         description: "Songomeno",
         added: "2009-07-29"}).
iana(language,sog,tag{
         type: "language",
         subtag: "sog",
         description: "Sogdian",
         added: "2005-10-16"}).
iana(language,soh,tag{
         type: "language",
         subtag: "soh",
         description: "Aka",
         added: "2009-07-29"}).
iana(language,soi,tag{
         type: "language",
         subtag: "soi",
         description: "Sonha",
         added: "2009-07-29"}).
iana(language,soj,tag{
         type: "language",
         subtag: "soj",
         description: "Soi",
         added: "2009-07-29"}).
iana(language,sok,tag{
         type: "language",
         subtag: "sok",
         description: "Sokoro",
         added: "2009-07-29"}).
iana(language,sol,tag{
         type: "language",
         subtag: "sol",
         description: "Solos",
         added: "2009-07-29"}).
iana(language,son,tag{
         type: "language",
         subtag: "son",
         description: "Songhai languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,soo,tag{
         type: "language",
         subtag: "soo",
         description: "Songo",
         added: "2009-07-29"}).
iana(language,sop,tag{
         type: "language",
         subtag: "sop",
         description: "Songe",
         added: "2009-07-29"}).
iana(language,soq,tag{
         type: "language",
         subtag: "soq",
         description: "Kanasi",
         added: "2009-07-29"}).
iana(language,sor,tag{
         type: "language",
         subtag: "sor",
         description: "Somrai",
         added: "2009-07-29"}).
iana(language,sos,tag{
         type: "language",
         subtag: "sos",
         description: "Seeku",
         added: "2009-07-29"}).
iana(language,sou,tag{
         type: "language",
         subtag: "sou",
         description: "Southern Thai",
         added: "2009-07-29"}).
iana(language,sov,tag{
         type: "language",
         subtag: "sov",
         description: "Sonsorol",
         added: "2009-07-29"}).
iana(language,sow,tag{
         type: "language",
         subtag: "sow",
         description: "Sowanda",
         added: "2009-07-29"}).
iana(language,sox,tag{
         type: "language",
         subtag: "sox",
         description: "Swo",
         added: "2009-07-29"}).
iana(language,soy,tag{
         type: "language",
         subtag: "soy",
         description: "Miyobe",
         added: "2009-07-29"}).
iana(language,soz,tag{
         type: "language",
         subtag: "soz",
         description: "Temi",
         added: "2009-07-29"}).
iana(language,spb,tag{
         type: "language",
         subtag: "spb",
         description: "Sepa (Indonesia)",
         added: "2009-07-29"}).
iana(language,spc,tag{
         type: "language",
         subtag: "spc",
         description: "Sapé",
         added: "2009-07-29"}).
iana(language,spd,tag{
         type: "language",
         subtag: "spd",
         description: "Saep",
         added: "2009-07-29"}).
iana(language,spe,tag{
         type: "language",
         subtag: "spe",
         description: "Sepa (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,spg,tag{
         type: "language",
         subtag: "spg",
         description: "Sian",
         added: "2009-07-29"}).
iana(language,spi,tag{
         type: "language",
         subtag: "spi",
         description: "Saponi",
         added: "2009-07-29"}).
iana(language,spk,tag{
         type: "language",
         subtag: "spk",
         description: "Sengo",
         added: "2009-07-29"}).
iana(language,spl,tag{
         type: "language",
         subtag: "spl",
         description: "Selepet",
         added: "2009-07-29"}).
iana(language,spm,tag{
         type: "language",
         subtag: "spm",
         description: "Akukem",
         added: "2009-07-29"}).
iana(language,spn,tag{
         type: "language",
         subtag: "spn",
         description: "Sanapaná",
         added: "2014-02-28"}).
iana(language,spo,tag{
         type: "language",
         subtag: "spo",
         description: "Spokane",
         added: "2009-07-29"}).
iana(language,spp,tag{
         type: "language",
         subtag: "spp",
         description: "Supyire Senoufo",
         added: "2009-07-29"}).
iana(language,spq,tag{
         type: "language",
         subtag: "spq",
         description: "Loreto-Ucayali Spanish",
         added: "2009-07-29"}).
iana(language,spr,tag{
         type: "language",
         subtag: "spr",
         description: "Saparua",
         added: "2009-07-29"}).
iana(language,sps,tag{
         type: "language",
         subtag: "sps",
         description: "Saposa",
         added: "2009-07-29"}).
iana(language,spt,tag{
         type: "language",
         subtag: "spt",
         description: "Spiti Bhoti",
         added: "2009-07-29"}).
iana(language,spu,tag{
         type: "language",
         subtag: "spu",
         description: "Sapuan",
         added: "2009-07-29"}).
iana(language,spv,tag{
         type: "language",
         subtag: "spv",
         description: ["Sambalpuri","Kosli"],
         added: "2012-08-12",
         macrolanguage: "or"}).
iana(language,spx,tag{
         type: "language",
         subtag: "spx",
         description: "South Picene",
         added: "2009-07-29"}).
iana(language,spy,tag{
         type: "language",
         subtag: "spy",
         description: "Sabaot",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,sqa,tag{
         type: "language",
         subtag: "sqa",
         description: "Shama-Sambuga",
         added: "2009-07-29"}).
iana(language,sqh,tag{
         type: "language",
         subtag: "sqh",
         description: "Shau",
         added: "2009-07-29"}).
iana(language,sqj,tag{
         type: "language",
         subtag: "sqj",
         description: "Albanian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,sqk,tag{
         type: "language",
         subtag: "sqk",
         description: "Albanian Sign Language",
         added: "2012-08-12"}).
iana(language,sqm,tag{
         type: "language",
         subtag: "sqm",
         description: "Suma",
         added: "2009-07-29"}).
iana(language,sqn,tag{
         type: "language",
         subtag: "sqn",
         description: "Susquehannock",
         added: "2009-07-29"}).
iana(language,sqo,tag{
         type: "language",
         subtag: "sqo",
         description: "Sorkhei",
         added: "2009-07-29"}).
iana(language,sqq,tag{
         type: "language",
         subtag: "sqq",
         description: "Sou",
         added: "2009-07-29"}).
iana(language,sqr,tag{
         type: "language",
         subtag: "sqr",
         description: "Siculo Arabic",
         added: "2009-07-29"}).
iana(language,sqs,tag{
         type: "language",
         subtag: "sqs",
         description: "Sri Lankan Sign Language",
         added: "2009-07-29"}).
iana(language,sqt,tag{
         type: "language",
         subtag: "sqt",
         description: "Soqotri",
         added: "2009-07-29"}).
iana(language,squ,tag{
         type: "language",
         subtag: "squ",
         description: "Squamish",
         added: "2009-07-29"}).
iana(language,sqx,tag{
         type: "language",
         subtag: "sqx",
         description: "Kufr Qassem Sign Language (KQSL)",
         added: "2021-02-20"}).
iana(language,sra,tag{
         type: "language",
         subtag: "sra",
         description: "Saruga",
         added: "2009-07-29"}).
iana(language,srb,tag{
         type: "language",
         subtag: "srb",
         description: "Sora",
         added: "2009-07-29"}).
iana(language,src,tag{
         type: "language",
         subtag: "src",
         description: "Logudorese Sardinian",
         added: "2009-07-29",
         macrolanguage: "sc"}).
iana(language,sre,tag{
         type: "language",
         subtag: "sre",
         description: "Sara",
         added: "2009-07-29"}).
iana(language,srf,tag{
         type: "language",
         subtag: "srf",
         description: "Nafi",
         added: "2009-07-29"}).
iana(language,srg,tag{
         type: "language",
         subtag: "srg",
         description: "Sulod",
         added: "2009-07-29"}).
iana(language,srh,tag{
         type: "language",
         subtag: "srh",
         description: "Sarikoli",
         added: "2009-07-29"}).
iana(language,sri,tag{
         type: "language",
         subtag: "sri",
         description: "Siriano",
         added: "2009-07-29"}).
iana(language,srk,tag{
         type: "language",
         subtag: "srk",
         description: "Serudung Murut",
         added: "2009-07-29"}).
iana(language,srl,tag{
         type: "language",
         subtag: "srl",
         description: "Isirawa",
         added: "2009-07-29"}).
iana(language,srm,tag{
         type: "language",
         subtag: "srm",
         description: "Saramaccan",
         added: "2009-07-29"}).
iana(language,srn,tag{
         type: "language",
         subtag: "srn",
         description: "Sranan Tongo",
         added: "2005-10-16"}).
iana(language,sro,tag{
         type: "language",
         subtag: "sro",
         description: "Campidanese Sardinian",
         added: "2009-07-29",
         macrolanguage: "sc"}).
iana(language,srq,tag{
         type: "language",
         subtag: "srq",
         description: "Sirionó",
         added: "2009-07-29"}).
iana(language,srr,tag{
         type: "language",
         subtag: "srr",
         description: "Serer",
         added: "2005-10-16"}).
iana(language,srs,tag{
         type: "language",
         subtag: "srs",
         description: "Sarsi",
         added: "2009-07-29"}).
iana(language,srt,tag{
         type: "language",
         subtag: "srt",
         description: "Sauri",
         added: "2009-07-29"}).
iana(language,sru,tag{
         type: "language",
         subtag: "sru",
         description: "Suruí",
         added: "2009-07-29"}).
iana(language,srv,tag{
         type: "language",
         subtag: "srv",
         description: "Southern Sorsoganon",
         added: "2009-07-29"}).
iana(language,srw,tag{
         type: "language",
         subtag: "srw",
         description: "Serua",
         added: "2009-07-29"}).
iana(language,srx,tag{
         type: "language",
         subtag: "srx",
         description: "Sirmauri",
         added: "2009-07-29"}).
iana(language,sry,tag{
         type: "language",
         subtag: "sry",
         description: "Sera",
         added: "2009-07-29"}).
iana(language,srz,tag{
         type: "language",
         subtag: "srz",
         description: "Shahmirzadi",
         added: "2009-07-29"}).
iana(language,ssa,tag{
         type: "language",
         subtag: "ssa",
         description: "Nilo-Saharan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ssb,tag{
         type: "language",
         subtag: "ssb",
         description: "Southern Sama",
         added: "2009-07-29"}).
iana(language,ssc,tag{
         type: "language",
         subtag: "ssc",
         description: "Suba-Simbiti",
         added: "2009-07-29"}).
iana(language,ssd,tag{
         type: "language",
         subtag: "ssd",
         description: "Siroi",
         added: "2009-07-29"}).
iana(language,sse,tag{
         type: "language",
         subtag: "sse",
         description: ["Balangingi","Bangingih Sama"],
         added: "2009-07-29"}).
iana(language,ssf,tag{
         type: "language",
         subtag: "ssf",
         description: "Thao",
         added: "2009-07-29"}).
iana(language,ssg,tag{
         type: "language",
         subtag: "ssg",
         description: "Seimat",
         added: "2009-07-29"}).
iana(language,ssh,tag{
         type: "language",
         subtag: "ssh",
         description: "Shihhi Arabic",
         added: "2009-07-29",
         macrolanguage: "ar"}).
iana(language,ssi,tag{
         type: "language",
         subtag: "ssi",
         description: "Sansi",
         added: "2009-07-29"}).
iana(language,ssj,tag{
         type: "language",
         subtag: "ssj",
         description: "Sausi",
         added: "2009-07-29"}).
iana(language,ssk,tag{
         type: "language",
         subtag: "ssk",
         description: "Sunam",
         added: "2009-07-29"}).
iana(language,ssl,tag{
         type: "language",
         subtag: "ssl",
         description: "Western Sisaala",
         added: "2009-07-29"}).
iana(language,ssm,tag{
         type: "language",
         subtag: "ssm",
         description: "Semnam",
         added: "2009-07-29"}).
iana(language,ssn,tag{
         type: "language",
         subtag: "ssn",
         description: "Waata",
         added: "2009-07-29"}).
iana(language,sso,tag{
         type: "language",
         subtag: "sso",
         description: "Sissano",
         added: "2009-07-29"}).
iana(language,ssp,tag{
         type: "language",
         subtag: "ssp",
         description: "Spanish Sign Language",
         added: "2009-07-29"}).
iana(language,ssq,tag{
         type: "language",
         subtag: "ssq",
         description: "So'a",
         added: "2009-07-29"}).
iana(language,ssr,tag{
         type: "language",
         subtag: "ssr",
         description: "Swiss-French Sign Language",
         added: "2009-07-29"}).
iana(language,sss,tag{
         type: "language",
         subtag: "sss",
         description: "Sô",
         added: "2009-07-29"}).
iana(language,sst,tag{
         type: "language",
         subtag: "sst",
         description: "Sinasina",
         added: "2009-07-29"}).
iana(language,ssu,tag{
         type: "language",
         subtag: "ssu",
         description: "Susuami",
         added: "2009-07-29"}).
iana(language,ssv,tag{
         type: "language",
         subtag: "ssv",
         description: "Shark Bay",
         added: "2009-07-29"}).
iana(language,ssx,tag{
         type: "language",
         subtag: "ssx",
         description: "Samberigi",
         added: "2009-07-29"}).
iana(language,ssy,tag{
         type: "language",
         subtag: "ssy",
         description: "Saho",
         added: "2009-07-29"}).
iana(language,ssz,tag{
         type: "language",
         subtag: "ssz",
         description: "Sengseng",
         added: "2009-07-29"}).
iana(language,sta,tag{
         type: "language",
         subtag: "sta",
         description: "Settla",
         added: "2009-07-29"}).
iana(language,stb,tag{
         type: "language",
         subtag: "stb",
         description: "Northern Subanen",
         added: "2009-07-29"}).
iana(language,std,tag{
         type: "language",
         subtag: "std",
         description: "Sentinel",
         added: "2009-07-29"}).
iana(language,ste,tag{
         type: "language",
         subtag: "ste",
         description: "Liana-Seti",
         added: "2009-07-29"}).
iana(language,stf,tag{
         type: "language",
         subtag: "stf",
         description: "Seta",
         added: "2009-07-29"}).
iana(language,stg,tag{
         type: "language",
         subtag: "stg",
         description: "Trieng",
         added: "2009-07-29"}).
iana(language,sth,tag{
         type: "language",
         subtag: "sth",
         description: "Shelta",
         added: "2009-07-29"}).
iana(language,sti,tag{
         type: "language",
         subtag: "sti",
         description: "Bulo Stieng",
         added: "2009-07-29"}).
iana(language,stj,tag{
         type: "language",
         subtag: "stj",
         description: "Matya Samo",
         added: "2009-07-29"}).
iana(language,stk,tag{
         type: "language",
         subtag: "stk",
         description: "Arammba",
         added: "2009-07-29"}).
iana(language,stl,tag{
         type: "language",
         subtag: "stl",
         description: "Stellingwerfs",
         added: "2009-07-29"}).
iana(language,stm,tag{
         type: "language",
         subtag: "stm",
         description: "Setaman",
         added: "2009-07-29"}).
iana(language,stn,tag{
         type: "language",
         subtag: "stn",
         description: "Owa",
         added: "2009-07-29"}).
iana(language,sto,tag{
         type: "language",
         subtag: "sto",
         description: "Stoney",
         added: "2009-07-29"}).
iana(language,stp,tag{
         type: "language",
         subtag: "stp",
         description: "Southeastern Tepehuan",
         added: "2009-07-29"}).
iana(language,stq,tag{
         type: "language",
         subtag: "stq",
         description: "Saterfriesisch",
         added: "2009-07-29"}).
iana(language,str,tag{
         type: "language",
         subtag: "str",
         description: "Straits Salish",
         added: "2009-07-29"}).
iana(language,sts,tag{
         type: "language",
         subtag: "sts",
         description: "Shumashti",
         added: "2009-07-29"}).
iana(language,stt,tag{
         type: "language",
         subtag: "stt",
         description: "Budeh Stieng",
         added: "2009-07-29"}).
iana(language,stu,tag{
         type: "language",
         subtag: "stu",
         description: "Samtao",
         added: "2009-07-29"}).
iana(language,stv,tag{
         type: "language",
         subtag: "stv",
         description: "Silt'e",
         added: "2009-07-29"}).
iana(language,stw,tag{
         type: "language",
         subtag: "stw",
         description: "Satawalese",
         added: "2009-07-29"}).
iana(language,sty,tag{
         type: "language",
         subtag: "sty",
         description: "Siberian Tatar",
         added: "2013-09-10"}).
iana(language,sua,tag{
         type: "language",
         subtag: "sua",
         description: "Sulka",
         added: "2009-07-29"}).
iana(language,sub,tag{
         type: "language",
         subtag: "sub",
         description: "Suku",
         added: "2009-07-29"}).
iana(language,suc,tag{
         type: "language",
         subtag: "suc",
         description: "Western Subanon",
         added: "2009-07-29"}).
iana(language,sue,tag{
         type: "language",
         subtag: "sue",
         description: "Suena",
         added: "2009-07-29"}).
iana(language,sug,tag{
         type: "language",
         subtag: "sug",
         description: "Suganga",
         added: "2009-07-29"}).
iana(language,sui,tag{
         type: "language",
         subtag: "sui",
         description: "Suki",
         added: "2009-07-29"}).
iana(language,suj,tag{
         type: "language",
         subtag: "suj",
         description: "Shubi",
         added: "2009-07-29",
         comments: "see also xsj"}).
iana(language,suk,tag{
         type: "language",
         subtag: "suk",
         description: "Sukuma",
         added: "2005-10-16"}).
iana(language,sul,tag{
         type: "language",
         subtag: "sul",
         description: "Surigaonon",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see sgd, tgn"}).
iana(language,sum,tag{
         type: "language",
         subtag: "sum",
         description: "Sumo-Mayangna",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see ulw, yan"}).
iana(language,suo,tag{
         type: "language",
         subtag: "suo",
         description: "Bouni",
         added: "2021-02-20"}).
iana(language,suq,tag{
         type: "language",
         subtag: "suq",
         description: ["Tirmaga-Chai Suri","Suri"],
         added: "2009-07-29"}).
iana(language,sur,tag{
         type: "language",
         subtag: "sur",
         description: "Mwaghavul",
         added: "2009-07-29"}).
iana(language,sus,tag{
         type: "language",
         subtag: "sus",
         description: "Susu",
         added: "2005-10-16"}).
iana(language,sut,tag{
         type: "language",
         subtag: "sut",
         description: "Subtiaba",
         added: "2009-07-29"}).
iana(language,suv,tag{
         type: "language",
         subtag: "suv",
         description: "Puroik",
         added: "2009-07-29"}).
iana(language,suw,tag{
         type: "language",
         subtag: "suw",
         description: "Sumbwa",
         added: "2009-07-29"}).
iana(language,sux,tag{
         type: "language",
         subtag: "sux",
         description: "Sumerian",
         added: "2005-10-16"}).
iana(language,suy,tag{
         type: "language",
         subtag: "suy",
         description: "Suyá",
         added: "2009-07-29"}).
iana(language,suz,tag{
         type: "language",
         subtag: "suz",
         description: "Sunwar",
         added: "2009-07-29"}).
iana(language,sva,tag{
         type: "language",
         subtag: "sva",
         description: "Svan",
         added: "2009-07-29"}).
iana(language,svb,tag{
         type: "language",
         subtag: "svb",
         description: "Ulau-Suain",
         added: "2009-07-29"}).
iana(language,svc,tag{
         type: "language",
         subtag: "svc",
         description: "Vincentian Creole English",
         added: "2009-07-29"}).
iana(language,sve,tag{
         type: "language",
         subtag: "sve",
         description: "Serili",
         added: "2009-07-29"}).
iana(language,svk,tag{
         type: "language",
         subtag: "svk",
         description: "Slovakian Sign Language",
         added: "2009-07-29"}).
iana(language,svm,tag{
         type: "language",
         subtag: "svm",
         description: "Slavomolisano",
         added: "2013-09-10"}).
iana(language,svr,tag{
         type: "language",
         subtag: "svr",
         description: "Savara",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,svs,tag{
         type: "language",
         subtag: "svs",
         description: "Savosavo",
         added: "2009-07-29"}).
iana(language,svx,tag{
         type: "language",
         subtag: "svx",
         description: "Skalvian",
         added: "2009-07-29"}).
iana(language,swb,tag{
         type: "language",
         subtag: "swb",
         description: "Maore Comorian",
         added: "2009-07-29"}).
iana(language,swc,tag{
         type: "language",
         subtag: "swc",
         description: "Congo Swahili",
         added: "2009-07-29",
         macrolanguage: "sw"}).
iana(language,swf,tag{
         type: "language",
         subtag: "swf",
         description: "Sere",
         added: "2009-07-29"}).
iana(language,swg,tag{
         type: "language",
         subtag: "swg",
         description: "Swabian",
         added: "2009-07-29"}).
iana(language,swh,tag{
         type: "language",
         subtag: "swh",
         description: ["Swahili (individual language)","Kiswahili"],
         added: "2009-07-29",
         macrolanguage: "sw"}).
iana(language,swi,tag{
         type: "language",
         subtag: "swi",
         description: "Sui",
         added: "2009-07-29"}).
iana(language,swj,tag{
         type: "language",
         subtag: "swj",
         description: "Sira",
         added: "2009-07-29"}).
iana(language,swk,tag{
         type: "language",
         subtag: "swk",
         description: "Malawi Sena",
         added: "2009-07-29"}).
iana(language,swl,tag{
         type: "language",
         subtag: "swl",
         description: "Swedish Sign Language",
         added: "2009-07-29"}).
iana(language,swm,tag{
         type: "language",
         subtag: "swm",
         description: "Samosa",
         added: "2009-07-29"}).
iana(language,swn,tag{
         type: "language",
         subtag: "swn",
         description: "Sawknah",
         added: "2009-07-29"}).
iana(language,swo,tag{
         type: "language",
         subtag: "swo",
         description: "Shanenawa",
         added: "2009-07-29"}).
iana(language,swp,tag{
         type: "language",
         subtag: "swp",
         description: "Suau",
         added: "2009-07-29"}).
iana(language,swq,tag{
         type: "language",
         subtag: "swq",
         description: "Sharwa",
         added: "2009-07-29"}).
iana(language,swr,tag{
         type: "language",
         subtag: "swr",
         description: "Saweru",
         added: "2009-07-29"}).
iana(language,sws,tag{
         type: "language",
         subtag: "sws",
         description: "Seluwasan",
         added: "2009-07-29"}).
iana(language,swt,tag{
         type: "language",
         subtag: "swt",
         description: "Sawila",
         added: "2009-07-29"}).
iana(language,swu,tag{
         type: "language",
         subtag: "swu",
         description: "Suwawa",
         added: "2009-07-29"}).
iana(language,swv,tag{
         type: "language",
         subtag: "swv",
         description: "Shekhawati",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,sww,tag{
         type: "language",
         subtag: "sww",
         description: "Sowa",
         added: "2009-07-29"}).
iana(language,swx,tag{
         type: "language",
         subtag: "swx",
         description: "Suruahá",
         added: "2009-07-29"}).
iana(language,swy,tag{
         type: "language",
         subtag: "swy",
         description: "Sarua",
         added: "2009-07-29"}).
iana(language,sxb,tag{
         type: "language",
         subtag: "sxb",
         description: "Suba",
         added: "2009-07-29"}).
iana(language,sxc,tag{
         type: "language",
         subtag: "sxc",
         description: "Sicanian",
         added: "2009-07-29"}).
iana(language,sxe,tag{
         type: "language",
         subtag: "sxe",
         description: "Sighu",
         added: "2009-07-29"}).
iana(language,sxg,tag{
         type: "language",
         subtag: "sxg",
         description: ["Shuhi","Shixing"],
         added: "2009-07-29"}).
iana(language,sxk,tag{
         type: "language",
         subtag: "sxk",
         description: "Southern Kalapuya",
         added: "2009-07-29"}).
iana(language,sxl,tag{
         type: "language",
         subtag: "sxl",
         description: "Selian",
         added: "2009-07-29"}).
iana(language,sxm,tag{
         type: "language",
         subtag: "sxm",
         description: "Samre",
         added: "2009-07-29"}).
iana(language,sxn,tag{
         type: "language",
         subtag: "sxn",
         description: "Sangir",
         added: "2009-07-29"}).
iana(language,sxo,tag{
         type: "language",
         subtag: "sxo",
         description: "Sorothaptic",
         added: "2009-07-29"}).
iana(language,sxr,tag{
         type: "language",
         subtag: "sxr",
         description: "Saaroa",
         added: "2009-07-29"}).
iana(language,sxs,tag{
         type: "language",
         subtag: "sxs",
         description: "Sasaru",
         added: "2009-07-29"}).
iana(language,sxu,tag{
         type: "language",
         subtag: "sxu",
         description: "Upper Saxon",
         added: "2009-07-29"}).
iana(language,sxw,tag{
         type: "language",
         subtag: "sxw",
         description: "Saxwe Gbe",
         added: "2009-07-29"}).
iana(language,sya,tag{
         type: "language",
         subtag: "sya",
         description: "Siang",
         added: "2009-07-29"}).
iana(language,syb,tag{
         type: "language",
         subtag: "syb",
         description: "Central Subanen",
         added: "2009-07-29"}).
iana(language,syc,tag{
         type: "language",
         subtag: "syc",
         description: "Classical Syriac",
         added: "2007-04-03"}).
iana(language,syd,tag{
         type: "language",
         subtag: "syd",
         description: "Samoyedic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,syi,tag{
         type: "language",
         subtag: "syi",
         description: "Seki",
         added: "2009-07-29"}).
iana(language,syk,tag{
         type: "language",
         subtag: "syk",
         description: "Sukur",
         added: "2009-07-29"}).
iana(language,syl,tag{
         type: "language",
         subtag: "syl",
         description: "Sylheti",
         added: "2009-07-29"}).
iana(language,sym,tag{
         type: "language",
         subtag: "sym",
         description: "Maya Samo",
         added: "2009-07-29"}).
iana(language,syn,tag{
         type: "language",
         subtag: "syn",
         description: "Senaya",
         added: "2009-07-29"}).
iana(language,syo,tag{
         type: "language",
         subtag: "syo",
         description: "Suoy",
         added: "2009-07-29"}).
iana(language,syr,tag{
         type: "language",
         subtag: "syr",
         description: "Syriac",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,sys,tag{
         type: "language",
         subtag: "sys",
         description: "Sinyar",
         added: "2009-07-29"}).
iana(language,syw,tag{
         type: "language",
         subtag: "syw",
         description: "Kagate",
         added: "2009-07-29"}).
iana(language,syx,tag{
         type: "language",
         subtag: "syx",
         description: "Samay",
         added: "2015-02-12"}).
iana(language,syy,tag{
         type: "language",
         subtag: "syy",
         description: "Al-Sayyid Bedouin Sign Language",
         added: "2009-07-29"}).
iana(language,sza,tag{
         type: "language",
         subtag: "sza",
         description: "Semelai",
         added: "2009-07-29"}).
iana(language,szb,tag{
         type: "language",
         subtag: "szb",
         description: "Ngalum",
         added: "2009-07-29"}).
iana(language,szc,tag{
         type: "language",
         subtag: "szc",
         description: "Semaq Beri",
         added: "2009-07-29"}).
iana(language,szd,tag{
         type: "language",
         subtag: "szd",
         description: "Seru",
         added: "2009-07-29"}).
iana(language,sze,tag{
         type: "language",
         subtag: "sze",
         description: "Seze",
         added: "2009-07-29"}).
iana(language,szg,tag{
         type: "language",
         subtag: "szg",
         description: "Sengele",
         added: "2009-07-29"}).
iana(language,szl,tag{
         type: "language",
         subtag: "szl",
         description: "Silesian",
         added: "2009-07-29"}).
iana(language,szn,tag{
         type: "language",
         subtag: "szn",
         description: "Sula",
         added: "2009-07-29"}).
iana(language,szp,tag{
         type: "language",
         subtag: "szp",
         description: "Suabo",
         added: "2009-07-29"}).
iana(language,szs,tag{
         type: "language",
         subtag: "szs",
         description: "Solomon Islands Sign Language",
         added: "2017-02-23"}).
iana(language,szv,tag{
         type: "language",
         subtag: "szv",
         description: "Isu (Fako Division)",
         added: "2009-07-29"}).
iana(language,szw,tag{
         type: "language",
         subtag: "szw",
         description: "Sawai",
         added: "2009-07-29"}).
iana(language,szy,tag{
         type: "language",
         subtag: "szy",
         description: "Sakizaya",
         added: "2019-04-16"}).
iana(language,taa,tag{
         type: "language",
         subtag: "taa",
         description: "Lower Tanana",
         added: "2009-07-29"}).
iana(language,tab,tag{
         type: "language",
         subtag: "tab",
         description: "Tabassaran",
         added: "2009-07-29"}).
iana(language,tac,tag{
         type: "language",
         subtag: "tac",
         description: "Lowland Tarahumara",
         added: "2009-07-29"}).
iana(language,tad,tag{
         type: "language",
         subtag: "tad",
         description: "Tause",
         added: "2009-07-29"}).
iana(language,tae,tag{
         type: "language",
         subtag: "tae",
         description: "Tariana",
         added: "2009-07-29"}).
iana(language,taf,tag{
         type: "language",
         subtag: "taf",
         description: "Tapirapé",
         added: "2009-07-29"}).
iana(language,tag,tag{
         type: "language",
         subtag: "tag",
         description: "Tagoi",
         added: "2009-07-29"}).
iana(language,tai,tag{
         type: "language",
         subtag: "tai",
         description: "Tai languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,taj,tag{
         type: "language",
         subtag: "taj",
         description: "Eastern Tamang",
         added: "2009-07-29"}).
iana(language,tak,tag{
         type: "language",
         subtag: "tak",
         description: "Tala",
         added: "2009-07-29"}).
iana(language,tal,tag{
         type: "language",
         subtag: "tal",
         description: "Tal",
         added: "2009-07-29"}).
iana(language,tan,tag{
         type: "language",
         subtag: "tan",
         description: "Tangale",
         added: "2009-07-29"}).
iana(language,tao,tag{
         type: "language",
         subtag: "tao",
         description: "Yami",
         added: "2009-07-29"}).
iana(language,tap,tag{
         type: "language",
         subtag: "tap",
         description: "Taabwa",
         added: "2009-07-29"}).
iana(language,taq,tag{
         type: "language",
         subtag: "taq",
         description: "Tamasheq",
         added: "2009-07-29",
         macrolanguage: "tmh"}).
iana(language,tar,tag{
         type: "language",
         subtag: "tar",
         description: "Central Tarahumara",
         added: "2009-07-29"}).
iana(language,tas,tag{
         type: "language",
         subtag: "tas",
         description: "Tay Boi",
         added: "2009-07-29"}).
iana(language,tau,tag{
         type: "language",
         subtag: "tau",
         description: "Upper Tanana",
         added: "2009-07-29"}).
iana(language,tav,tag{
         type: "language",
         subtag: "tav",
         description: "Tatuyo",
         added: "2009-07-29"}).
iana(language,taw,tag{
         type: "language",
         subtag: "taw",
         description: "Tai",
         added: "2009-07-29"}).
iana(language,tax,tag{
         type: "language",
         subtag: "tax",
         description: "Tamki",
         added: "2009-07-29"}).
iana(language,tay,tag{
         type: "language",
         subtag: "tay",
         description: "Atayal",
         added: "2009-07-29"}).
iana(language,taz,tag{
         type: "language",
         subtag: "taz",
         description: "Tocho",
         added: "2009-07-29"}).
iana(language,tba,tag{
         type: "language",
         subtag: "tba",
         description: "Aikanã",
         added: "2009-07-29"}).
iana(language,tbb,tag{
         type: "language",
         subtag: "tbb",
         description: "Tapeba",
         added: "2009-07-29",
         deprecated: "2020-03-28"}).
iana(language,tbc,tag{
         type: "language",
         subtag: "tbc",
         description: "Takia",
         added: "2009-07-29"}).
iana(language,tbd,tag{
         type: "language",
         subtag: "tbd",
         description: "Kaki Ae",
         added: "2009-07-29"}).
iana(language,tbe,tag{
         type: "language",
         subtag: "tbe",
         description: "Tanimbili",
         added: "2009-07-29"}).
iana(language,tbf,tag{
         type: "language",
         subtag: "tbf",
         description: "Mandara",
         added: "2009-07-29"}).
iana(language,tbg,tag{
         type: "language",
         subtag: "tbg",
         description: "North Tairora",
         added: "2009-07-29"}).
iana(language,tbh,tag{
         type: "language",
         subtag: "tbh",
         description: ["Dharawal","Thurawal"],
         added: "2009-07-29"}).
iana(language,tbi,tag{
         type: "language",
         subtag: "tbi",
         description: "Gaam",
         added: "2009-07-29"}).
iana(language,tbj,tag{
         type: "language",
         subtag: "tbj",
         description: "Tiang",
         added: "2009-07-29"}).
iana(language,tbk,tag{
         type: "language",
         subtag: "tbk",
         description: "Calamian Tagbanwa",
         added: "2009-07-29"}).
iana(language,tbl,tag{
         type: "language",
         subtag: "tbl",
         description: "Tboli",
         added: "2009-07-29"}).
iana(language,tbm,tag{
         type: "language",
         subtag: "tbm",
         description: "Tagbu",
         added: "2009-07-29"}).
iana(language,tbn,tag{
         type: "language",
         subtag: "tbn",
         description: "Barro Negro Tunebo",
         added: "2009-07-29"}).
iana(language,tbo,tag{
         type: "language",
         subtag: "tbo",
         description: "Tawala",
         added: "2009-07-29"}).
iana(language,tbp,tag{
         type: "language",
         subtag: "tbp",
         description: ["Taworta","Diebroud"],
         added: "2009-07-29"}).
iana(language,tbq,tag{
         type: "language",
         subtag: "tbq",
         description: "Tibeto-Burman languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,tbr,tag{
         type: "language",
         subtag: "tbr",
         description: "Tumtum",
         added: "2009-07-29"}).
iana(language,tbs,tag{
         type: "language",
         subtag: "tbs",
         description: "Tanguat",
         added: "2009-07-29"}).
iana(language,tbt,tag{
         type: "language",
         subtag: "tbt",
         description: "Tembo (Kitembo)",
         added: "2009-07-29"}).
iana(language,tbu,tag{
         type: "language",
         subtag: "tbu",
         description: "Tubar",
         added: "2009-07-29"}).
iana(language,tbv,tag{
         type: "language",
         subtag: "tbv",
         description: "Tobo",
         added: "2009-07-29"}).
iana(language,tbw,tag{
         type: "language",
         subtag: "tbw",
         description: "Tagbanwa",
         added: "2009-07-29"}).
iana(language,tbx,tag{
         type: "language",
         subtag: "tbx",
         description: "Kapin",
         added: "2009-07-29"}).
iana(language,tby,tag{
         type: "language",
         subtag: "tby",
         description: "Tabaru",
         added: "2009-07-29"}).
iana(language,tbz,tag{
         type: "language",
         subtag: "tbz",
         description: "Ditammari",
         added: "2009-07-29"}).
iana(language,tca,tag{
         type: "language",
         subtag: "tca",
         description: "Ticuna",
         added: "2009-07-29"}).
iana(language,tcb,tag{
         type: "language",
         subtag: "tcb",
         description: "Tanacross",
         added: "2009-07-29"}).
iana(language,tcc,tag{
         type: "language",
         subtag: "tcc",
         description: "Datooga",
         added: "2009-07-29"}).
iana(language,tcd,tag{
         type: "language",
         subtag: "tcd",
         description: "Tafi",
         added: "2009-07-29"}).
iana(language,tce,tag{
         type: "language",
         subtag: "tce",
         description: "Southern Tutchone",
         added: "2009-07-29"}).
iana(language,tcf,tag{
         type: "language",
         subtag: "tcf",
         description: ["Malinaltepec Me'phaa","Malinaltepec Tlapanec"],
         added: "2009-07-29"}).
iana(language,tcg,tag{
         type: "language",
         subtag: "tcg",
         description: "Tamagario",
         added: "2009-07-29"}).
iana(language,tch,tag{
         type: "language",
         subtag: "tch",
         description: "Turks And Caicos Creole English",
         added: "2009-07-29"}).
iana(language,tci,tag{
         type: "language",
         subtag: "tci",
         description: "Wára",
         added: "2009-07-29"}).
iana(language,tck,tag{
         type: "language",
         subtag: "tck",
         description: "Tchitchege",
         added: "2009-07-29"}).
iana(language,tcl,tag{
         type: "language",
         subtag: "tcl",
         description: "Taman (Myanmar)",
         added: "2009-07-29"}).
iana(language,tcm,tag{
         type: "language",
         subtag: "tcm",
         description: "Tanahmerah",
         added: "2009-07-29"}).
iana(language,tcn,tag{
         type: "language",
         subtag: "tcn",
         description: "Tichurong",
         added: "2009-07-29"}).
iana(language,tco,tag{
         type: "language",
         subtag: "tco",
         description: "Taungyo",
         added: "2009-07-29"}).
iana(language,tcp,tag{
         type: "language",
         subtag: "tcp",
         description: "Tawr Chin",
         added: "2009-07-29"}).
iana(language,tcq,tag{
         type: "language",
         subtag: "tcq",
         description: "Kaiy",
         added: "2009-07-29"}).
iana(language,tcs,tag{
         type: "language",
         subtag: "tcs",
         description: ["Torres Strait Creole","Yumplatok"],
         added: "2009-07-29"}).
iana(language,tct,tag{
         type: "language",
         subtag: "tct",
         description: "T'en",
         added: "2009-07-29"}).
iana(language,tcu,tag{
         type: "language",
         subtag: "tcu",
         description: "Southeastern Tarahumara",
         added: "2009-07-29"}).
iana(language,tcw,tag{
         type: "language",
         subtag: "tcw",
         description: "Tecpatlán Totonac",
         added: "2009-07-29"}).
iana(language,tcx,tag{
         type: "language",
         subtag: "tcx",
         description: "Toda",
         added: "2009-07-29"}).
iana(language,tcy,tag{
         type: "language",
         subtag: "tcy",
         description: "Tulu",
         added: "2009-07-29"}).
iana(language,tcz,tag{
         type: "language",
         subtag: "tcz",
         description: "Thado Chin",
         added: "2009-07-29"}).
iana(language,tda,tag{
         type: "language",
         subtag: "tda",
         description: "Tagdal",
         added: "2009-07-29"}).
iana(language,tdb,tag{
         type: "language",
         subtag: "tdb",
         description: "Panchpargania",
         added: "2009-07-29"}).
iana(language,tdc,tag{
         type: "language",
         subtag: "tdc",
         description: "Emberá-Tadó",
         added: "2009-07-29"}).
iana(language,tdd,tag{
         type: "language",
         subtag: "tdd",
         description: "Tai Nüa",
         added: "2009-07-29"}).
iana(language,tde,tag{
         type: "language",
         subtag: "tde",
         description: "Tiranige Diga Dogon",
         added: "2010-03-11"}).
iana(language,tdf,tag{
         type: "language",
         subtag: "tdf",
         description: "Talieng",
         added: "2009-07-29"}).
iana(language,tdg,tag{
         type: "language",
         subtag: "tdg",
         description: "Western Tamang",
         added: "2009-07-29"}).
iana(language,tdh,tag{
         type: "language",
         subtag: "tdh",
         description: "Thulung",
         added: "2009-07-29"}).
iana(language,tdi,tag{
         type: "language",
         subtag: "tdi",
         description: "Tomadino",
         added: "2009-07-29"}).
iana(language,tdj,tag{
         type: "language",
         subtag: "tdj",
         description: "Tajio",
         added: "2009-07-29"}).
iana(language,tdk,tag{
         type: "language",
         subtag: "tdk",
         description: "Tambas",
         added: "2009-07-29"}).
iana(language,tdl,tag{
         type: "language",
         subtag: "tdl",
         description: "Sur",
         added: "2009-07-29"}).
iana(language,tdm,tag{
         type: "language",
         subtag: "tdm",
         description: "Taruma",
         added: "2016-05-30"}).
iana(language,tdn,tag{
         type: "language",
         subtag: "tdn",
         description: "Tondano",
         added: "2009-07-29"}).
iana(language,tdo,tag{
         type: "language",
         subtag: "tdo",
         description: "Teme",
         added: "2009-07-29"}).
iana(language,tdq,tag{
         type: "language",
         subtag: "tdq",
         description: "Tita",
         added: "2009-07-29"}).
iana(language,tdr,tag{
         type: "language",
         subtag: "tdr",
         description: "Todrah",
         added: "2009-07-29"}).
iana(language,tds,tag{
         type: "language",
         subtag: "tds",
         description: "Doutai",
         added: "2009-07-29"}).
iana(language,tdt,tag{
         type: "language",
         subtag: "tdt",
         description: "Tetun Dili",
         added: "2009-07-29"}).
iana(language,tdu,tag{
         type: "language",
         subtag: "tdu",
         description: "Tempasuk Dusun",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "dtp"}).
iana(language,tdv,tag{
         type: "language",
         subtag: "tdv",
         description: "Toro",
         added: "2009-07-29"}).
iana(language,tdx,tag{
         type: "language",
         subtag: "tdx",
         description: "Tandroy-Mahafaly Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,tdy,tag{
         type: "language",
         subtag: "tdy",
         description: "Tadyawan",
         added: "2009-07-29"}).
iana(language,tea,tag{
         type: "language",
         subtag: "tea",
         description: "Temiar",
         added: "2009-07-29"}).
iana(language,teb,tag{
         type: "language",
         subtag: "teb",
         description: "Tetete",
         added: "2009-07-29"}).
iana(language,tec,tag{
         type: "language",
         subtag: "tec",
         description: "Terik",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,ted,tag{
         type: "language",
         subtag: "ted",
         description: "Tepo Krumen",
         added: "2009-07-29"}).
iana(language,tee,tag{
         type: "language",
         subtag: "tee",
         description: "Huehuetla Tepehua",
         added: "2009-07-29"}).
iana(language,tef,tag{
         type: "language",
         subtag: "tef",
         description: "Teressa",
         added: "2009-07-29"}).
iana(language,teg,tag{
         type: "language",
         subtag: "teg",
         description: "Teke-Tege",
         added: "2009-07-29"}).
iana(language,teh,tag{
         type: "language",
         subtag: "teh",
         description: "Tehuelche",
         added: "2009-07-29"}).
iana(language,tei,tag{
         type: "language",
         subtag: "tei",
         description: "Torricelli",
         added: "2009-07-29"}).
iana(language,tek,tag{
         type: "language",
         subtag: "tek",
         description: "Ibali Teke",
         added: "2009-07-29"}).
iana(language,tem,tag{
         type: "language",
         subtag: "tem",
         description: "Timne",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,ten,tag{
         type: "language",
         subtag: "ten",
         description: "Tama (Colombia)",
         added: "2009-07-29"}).
iana(language,teo,tag{
         type: "language",
         subtag: "teo",
         description: "Teso",
         added: "2009-07-29"}).
iana(language,tep,tag{
         type: "language",
         subtag: "tep",
         description: "Tepecano",
         added: "2009-07-29"}).
iana(language,teq,tag{
         type: "language",
         subtag: "teq",
         description: "Temein",
         added: "2009-07-29"}).
iana(language,ter,tag{
         type: "language",
         subtag: "ter",
         description: "Tereno",
         added: "2005-10-16"}).
iana(language,tes,tag{
         type: "language",
         subtag: "tes",
         description: "Tengger",
         added: "2009-07-29"}).
iana(language,tet,tag{
         type: "language",
         subtag: "tet",
         description: "Tetum",
         added: "2005-10-16"}).
iana(language,teu,tag{
         type: "language",
         subtag: "teu",
         description: "Soo",
         added: "2009-07-29"}).
iana(language,tev,tag{
         type: "language",
         subtag: "tev",
         description: "Teor",
         added: "2009-07-29"}).
iana(language,tew,tag{
         type: "language",
         subtag: "tew",
         description: "Tewa (USA)",
         added: "2009-07-29"}).
iana(language,tex,tag{
         type: "language",
         subtag: "tex",
         description: "Tennet",
         added: "2009-07-29"}).
iana(language,tey,tag{
         type: "language",
         subtag: "tey",
         description: "Tulishi",
         added: "2009-07-29"}).
iana(language,tez,tag{
         type: "language",
         subtag: "tez",
         description: "Tetserret",
         added: "2018-03-08"}).
iana(language,tfi,tag{
         type: "language",
         subtag: "tfi",
         description: "Tofin Gbe",
         added: "2009-07-29"}).
iana(language,tfn,tag{
         type: "language",
         subtag: "tfn",
         description: "Tanaina",
         added: "2009-07-29"}).
iana(language,tfo,tag{
         type: "language",
         subtag: "tfo",
         description: "Tefaro",
         added: "2009-07-29"}).
iana(language,tfr,tag{
         type: "language",
         subtag: "tfr",
         description: "Teribe",
         added: "2009-07-29"}).
iana(language,tft,tag{
         type: "language",
         subtag: "tft",
         description: "Ternate",
         added: "2009-07-29"}).
iana(language,tga,tag{
         type: "language",
         subtag: "tga",
         description: "Sagalla",
         added: "2009-07-29"}).
iana(language,tgb,tag{
         type: "language",
         subtag: "tgb",
         description: "Tobilung",
         added: "2009-07-29"}).
iana(language,tgc,tag{
         type: "language",
         subtag: "tgc",
         description: "Tigak",
         added: "2009-07-29"}).
iana(language,tgd,tag{
         type: "language",
         subtag: "tgd",
         description: "Ciwogai",
         added: "2009-07-29"}).
iana(language,tge,tag{
         type: "language",
         subtag: "tge",
         description: "Eastern Gorkha Tamang",
         added: "2009-07-29"}).
iana(language,tgf,tag{
         type: "language",
         subtag: "tgf",
         description: "Chalikha",
         added: "2009-07-29"}).
iana(language,tgg,tag{
         type: "language",
         subtag: "tgg",
         description: "Tangga",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see bjp, hrc, hrw"}).
iana(language,tgh,tag{
         type: "language",
         subtag: "tgh",
         description: "Tobagonian Creole English",
         added: "2009-07-29"}).
iana(language,tgi,tag{
         type: "language",
         subtag: "tgi",
         description: "Lawunuia",
         added: "2009-07-29"}).
iana(language,tgj,tag{
         type: "language",
         subtag: "tgj",
         description: "Tagin",
         added: "2012-08-12"}).
iana(language,tgn,tag{
         type: "language",
         subtag: "tgn",
         description: "Tandaganon",
         added: "2010-03-11"}).
iana(language,tgo,tag{
         type: "language",
         subtag: "tgo",
         description: "Sudest",
         added: "2009-07-29"}).
iana(language,tgp,tag{
         type: "language",
         subtag: "tgp",
         description: "Tangoa",
         added: "2009-07-29"}).
iana(language,tgq,tag{
         type: "language",
         subtag: "tgq",
         description: "Tring",
         added: "2009-07-29"}).
iana(language,tgr,tag{
         type: "language",
         subtag: "tgr",
         description: "Tareng",
         added: "2009-07-29"}).
iana(language,tgs,tag{
         type: "language",
         subtag: "tgs",
         description: "Nume",
         added: "2009-07-29"}).
iana(language,tgt,tag{
         type: "language",
         subtag: "tgt",
         description: "Central Tagbanwa",
         added: "2009-07-29"}).
iana(language,tgu,tag{
         type: "language",
         subtag: "tgu",
         description: "Tanggu",
         added: "2009-07-29"}).
iana(language,tgv,tag{
         type: "language",
         subtag: "tgv",
         description: "Tingui-Boto",
         added: "2009-07-29"}).
iana(language,tgw,tag{
         type: "language",
         subtag: "tgw",
         description: "Tagwana Senoufo",
         added: "2009-07-29"}).
iana(language,tgx,tag{
         type: "language",
         subtag: "tgx",
         description: "Tagish",
         added: "2009-07-29"}).
iana(language,tgy,tag{
         type: "language",
         subtag: "tgy",
         description: "Togoyo",
         added: "2009-07-29"}).
iana(language,tgz,tag{
         type: "language",
         subtag: "tgz",
         description: "Tagalaka",
         added: "2012-08-12"}).
iana(language,thc,tag{
         type: "language",
         subtag: "thc",
         description: "Tai Hang Tong",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "tpo"}).
iana(language,thd,tag{
         type: "language",
         subtag: "thd",
         description: ["Kuuk Thaayorre","Thayore"],
         added: "2009-07-29"}).
iana(language,the,tag{
         type: "language",
         subtag: "the",
         description: "Chitwania Tharu",
         added: "2009-07-29"}).
iana(language,thf,tag{
         type: "language",
         subtag: "thf",
         description: "Thangmi",
         added: "2009-07-29"}).
iana(language,thh,tag{
         type: "language",
         subtag: "thh",
         description: "Northern Tarahumara",
         added: "2009-07-29"}).
iana(language,thi,tag{
         type: "language",
         subtag: "thi",
         description: "Tai Long",
         added: "2009-07-29"}).
iana(language,thk,tag{
         type: "language",
         subtag: "thk",
         description: ["Tharaka","Kitharaka"],
         added: "2009-07-29"}).
iana(language,thl,tag{
         type: "language",
         subtag: "thl",
         description: "Dangaura Tharu",
         added: "2009-07-29"}).
iana(language,thm,tag{
         type: "language",
         subtag: "thm",
         description: "Aheu",
         added: "2009-07-29"}).
iana(language,thn,tag{
         type: "language",
         subtag: "thn",
         description: "Thachanadan",
         added: "2009-07-29"}).
iana(language,thp,tag{
         type: "language",
         subtag: "thp",
         description: "Thompson",
         added: "2009-07-29"}).
iana(language,thq,tag{
         type: "language",
         subtag: "thq",
         description: "Kochila Tharu",
         added: "2009-07-29"}).
iana(language,thr,tag{
         type: "language",
         subtag: "thr",
         description: "Rana Tharu",
         added: "2009-07-29"}).
iana(language,ths,tag{
         type: "language",
         subtag: "ths",
         description: "Thakali",
         added: "2009-07-29"}).
iana(language,tht,tag{
         type: "language",
         subtag: "tht",
         description: "Tahltan",
         added: "2009-07-29"}).
iana(language,thu,tag{
         type: "language",
         subtag: "thu",
         description: "Thuri",
         added: "2009-07-29"}).
iana(language,thv,tag{
         type: "language",
         subtag: "thv",
         description: "Tahaggart Tamahaq",
         added: "2009-07-29",
         macrolanguage: "tmh"}).
iana(language,thw,tag{
         type: "language",
         subtag: "thw",
         description: "Thudam",
         added: "2009-07-29",
         deprecated: "2020-06-08",
         preferred_value: "ola"}).
iana(language,thx,tag{
         type: "language",
         subtag: "thx",
         description: "The",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "oyb"}).
iana(language,thy,tag{
         type: "language",
         subtag: "thy",
         description: "Tha",
         added: "2009-07-29"}).
iana(language,thz,tag{
         type: "language",
         subtag: "thz",
         description: "Tayart Tamajeq",
         added: "2009-07-29",
         macrolanguage: "tmh"}).
iana(language,tia,tag{
         type: "language",
         subtag: "tia",
         description: "Tidikelt Tamazight",
         added: "2009-07-29"}).
iana(language,tic,tag{
         type: "language",
         subtag: "tic",
         description: "Tira",
         added: "2009-07-29"}).
iana(language,tid,tag{
         type: "language",
         subtag: "tid",
         description: "Tidong",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         comments: "see itd, ntd"}).
iana(language,tie,tag{
         type: "language",
         subtag: "tie",
         description: "Tingal",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         preferred_value: "ras"}).
iana(language,tif,tag{
         type: "language",
         subtag: "tif",
         description: "Tifal",
         added: "2009-07-29"}).
iana(language,tig,tag{
         type: "language",
         subtag: "tig",
         description: "Tigre",
         added: "2005-10-16"}).
iana(language,tih,tag{
         type: "language",
         subtag: "tih",
         description: "Timugon Murut",
         added: "2009-07-29"}).
iana(language,tii,tag{
         type: "language",
         subtag: "tii",
         description: "Tiene",
         added: "2009-07-29"}).
iana(language,tij,tag{
         type: "language",
         subtag: "tij",
         description: "Tilung",
         added: "2009-07-29"}).
iana(language,tik,tag{
         type: "language",
         subtag: "tik",
         description: "Tikar",
         added: "2009-07-29"}).
iana(language,til,tag{
         type: "language",
         subtag: "til",
         description: "Tillamook",
         added: "2009-07-29"}).
iana(language,tim,tag{
         type: "language",
         subtag: "tim",
         description: "Timbe",
         added: "2009-07-29"}).
iana(language,tin,tag{
         type: "language",
         subtag: "tin",
         description: "Tindi",
         added: "2009-07-29"}).
iana(language,tio,tag{
         type: "language",
         subtag: "tio",
         description: "Teop",
         added: "2009-07-29"}).
iana(language,tip,tag{
         type: "language",
         subtag: "tip",
         description: "Trimuris",
         added: "2009-07-29"}).
iana(language,tiq,tag{
         type: "language",
         subtag: "tiq",
         description: "Tiéfo",
         added: "2009-07-29"}).
iana(language,tis,tag{
         type: "language",
         subtag: "tis",
         description: "Masadiit Itneg",
         added: "2009-07-29"}).
iana(language,tit,tag{
         type: "language",
         subtag: "tit",
         description: "Tinigua",
         added: "2009-07-29"}).
iana(language,tiu,tag{
         type: "language",
         subtag: "tiu",
         description: "Adasen",
         added: "2009-07-29"}).
iana(language,tiv,tag{
         type: "language",
         subtag: "tiv",
         description: "Tiv",
         added: "2005-10-16"}).
iana(language,tiw,tag{
         type: "language",
         subtag: "tiw",
         description: "Tiwi",
         added: "2009-07-29"}).
iana(language,tix,tag{
         type: "language",
         subtag: "tix",
         description: "Southern Tiwa",
         added: "2009-07-29"}).
iana(language,tiy,tag{
         type: "language",
         subtag: "tiy",
         description: "Tiruray",
         added: "2009-07-29"}).
iana(language,tiz,tag{
         type: "language",
         subtag: "tiz",
         description: "Tai Hongjin",
         added: "2009-07-29"}).
iana(language,tja,tag{
         type: "language",
         subtag: "tja",
         description: "Tajuasohn",
         added: "2009-07-29"}).
iana(language,tjg,tag{
         type: "language",
         subtag: "tjg",
         description: "Tunjung",
         added: "2009-07-29"}).
iana(language,tji,tag{
         type: "language",
         subtag: "tji",
         description: "Northern Tujia",
         added: "2009-07-29"}).
iana(language,tjj,tag{
         type: "language",
         subtag: "tjj",
         description: "Tjungundji",
         added: "2019-04-16"}).
iana(language,tjl,tag{
         type: "language",
         subtag: "tjl",
         description: "Tai Laing",
         added: "2012-08-12"}).
iana(language,tjm,tag{
         type: "language",
         subtag: "tjm",
         description: "Timucua",
         added: "2009-07-29"}).
iana(language,tjn,tag{
         type: "language",
         subtag: "tjn",
         description: "Tonjon",
         added: "2009-07-29"}).
iana(language,tjo,tag{
         type: "language",
         subtag: "tjo",
         description: "Temacine Tamazight",
         added: "2009-07-29"}).
iana(language,tjp,tag{
         type: "language",
         subtag: "tjp",
         description: "Tjupany",
         added: "2019-04-16"}).
iana(language,tjs,tag{
         type: "language",
         subtag: "tjs",
         description: "Southern Tujia",
         added: "2009-07-29"}).
iana(language,tju,tag{
         type: "language",
         subtag: "tju",
         description: "Tjurruru",
         added: "2009-07-29"}).
iana(language,tjw,tag{
         type: "language",
         subtag: "tjw",
         description: "Djabwurrung",
         added: "2013-09-10"}).
iana(language,tka,tag{
         type: "language",
         subtag: "tka",
         description: "Truká",
         added: "2009-07-29"}).
iana(language,tkb,tag{
         type: "language",
         subtag: "tkb",
         description: "Buksa",
         added: "2009-07-29"}).
iana(language,tkd,tag{
         type: "language",
         subtag: "tkd",
         description: "Tukudede",
         added: "2009-07-29"}).
iana(language,tke,tag{
         type: "language",
         subtag: "tke",
         description: "Takwane",
         added: "2009-07-29"}).
iana(language,tkf,tag{
         type: "language",
         subtag: "tkf",
         description: "Tukumanféd",
         added: "2009-07-29"}).
iana(language,tkg,tag{
         type: "language",
         subtag: "tkg",
         description: "Tesaka Malagasy",
         added: "2011-08-16",
         macrolanguage: "mg"}).
iana(language,tkk,tag{
         type: "language",
         subtag: "tkk",
         description: "Takpa",
         added: "2009-07-29",
         deprecated: "2011-08-16",
         preferred_value: "twm"}).
iana(language,tkl,tag{
         type: "language",
         subtag: "tkl",
         description: "Tokelau",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tkm,tag{
         type: "language",
         subtag: "tkm",
         description: "Takelma",
         added: "2009-07-29"}).
iana(language,tkn,tag{
         type: "language",
         subtag: "tkn",
         description: "Toku-No-Shima",
         added: "2009-07-29"}).
iana(language,tkp,tag{
         type: "language",
         subtag: "tkp",
         description: "Tikopia",
         added: "2009-07-29"}).
iana(language,tkq,tag{
         type: "language",
         subtag: "tkq",
         description: "Tee",
         added: "2009-07-29"}).
iana(language,tkr,tag{
         type: "language",
         subtag: "tkr",
         description: "Tsakhur",
         added: "2009-07-29"}).
iana(language,tks,tag{
         type: "language",
         subtag: "tks",
         description: "Takestani",
         added: "2009-07-29"}).
iana(language,tkt,tag{
         type: "language",
         subtag: "tkt",
         description: "Kathoriya Tharu",
         added: "2009-07-29"}).
iana(language,tku,tag{
         type: "language",
         subtag: "tku",
         description: "Upper Necaxa Totonac",
         added: "2009-07-29"}).
iana(language,tkv,tag{
         type: "language",
         subtag: "tkv",
         description: "Mur Pano",
         added: "2014-02-28"}).
iana(language,tkw,tag{
         type: "language",
         subtag: "tkw",
         description: "Teanu",
         added: "2009-07-29"}).
iana(language,tkx,tag{
         type: "language",
         subtag: "tkx",
         description: "Tangko",
         added: "2009-07-29"}).
iana(language,tkz,tag{
         type: "language",
         subtag: "tkz",
         description: "Takua",
         added: "2009-07-29"}).
iana(language,tla,tag{
         type: "language",
         subtag: "tla",
         description: "Southwestern Tepehuan",
         added: "2009-07-29"}).
iana(language,tlb,tag{
         type: "language",
         subtag: "tlb",
         description: "Tobelo",
         added: "2009-07-29"}).
iana(language,tlc,tag{
         type: "language",
         subtag: "tlc",
         description: "Yecuatla Totonac",
         added: "2009-07-29"}).
iana(language,tld,tag{
         type: "language",
         subtag: "tld",
         description: "Talaud",
         added: "2009-07-29"}).
iana(language,tlf,tag{
         type: "language",
         subtag: "tlf",
         description: "Telefol",
         added: "2009-07-29"}).
iana(language,tlg,tag{
         type: "language",
         subtag: "tlg",
         description: "Tofanma",
         added: "2009-07-29"}).
iana(language,tlh,tag{
         type: "language",
         subtag: "tlh",
         description: ["Klingon","tlhIngan Hol"],
         added: "2005-10-16"}).
iana(language,tli,tag{
         type: "language",
         subtag: "tli",
         description: "Tlingit",
         added: "2005-10-16"}).
iana(language,tlj,tag{
         type: "language",
         subtag: "tlj",
         description: "Talinga-Bwisi",
         added: "2009-07-29"}).
iana(language,tlk,tag{
         type: "language",
         subtag: "tlk",
         description: "Taloki",
         added: "2009-07-29"}).
iana(language,tll,tag{
         type: "language",
         subtag: "tll",
         description: "Tetela",
         added: "2009-07-29"}).
iana(language,tlm,tag{
         type: "language",
         subtag: "tlm",
         description: "Tolomako",
         added: "2009-07-29"}).
iana(language,tln,tag{
         type: "language",
         subtag: "tln",
         description: "Talondo'",
         added: "2009-07-29"}).
iana(language,tlo,tag{
         type: "language",
         subtag: "tlo",
         description: "Talodi",
         added: "2009-07-29"}).
iana(language,tlp,tag{
         type: "language",
         subtag: "tlp",
         description: "Filomena Mata-Coahuitlán Totonac",
         added: "2009-07-29"}).
iana(language,tlq,tag{
         type: "language",
         subtag: "tlq",
         description: "Tai Loi",
         added: "2009-07-29"}).
iana(language,tlr,tag{
         type: "language",
         subtag: "tlr",
         description: "Talise",
         added: "2009-07-29"}).
iana(language,tls,tag{
         type: "language",
         subtag: "tls",
         description: "Tambotalo",
         added: "2009-07-29"}).
iana(language,tlt,tag{
         type: "language",
         subtag: "tlt",
         description: ["Sou Nama","Teluti"],
         added: "2009-07-29"}).
iana(language,tlu,tag{
         type: "language",
         subtag: "tlu",
         description: "Tulehu",
         added: "2009-07-29"}).
iana(language,tlv,tag{
         type: "language",
         subtag: "tlv",
         description: "Taliabu",
         added: "2009-07-29"}).
iana(language,tlw,tag{
         type: "language",
         subtag: "tlw",
         description: "South Wemale",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "weo"}).
iana(language,tlx,tag{
         type: "language",
         subtag: "tlx",
         description: "Khehek",
         added: "2009-07-29"}).
iana(language,tly,tag{
         type: "language",
         subtag: "tly",
         description: "Talysh",
         added: "2009-07-29"}).
iana(language,tma,tag{
         type: "language",
         subtag: "tma",
         description: "Tama (Chad)",
         added: "2009-07-29"}).
iana(language,tmb,tag{
         type: "language",
         subtag: "tmb",
         description: ["Katbol","Avava"],
         added: "2009-07-29"}).
iana(language,tmc,tag{
         type: "language",
         subtag: "tmc",
         description: "Tumak",
         added: "2009-07-29"}).
iana(language,tmd,tag{
         type: "language",
         subtag: "tmd",
         description: "Haruai",
         added: "2009-07-29"}).
iana(language,tme,tag{
         type: "language",
         subtag: "tme",
         description: "Tremembé",
         added: "2009-07-29"}).
iana(language,tmf,tag{
         type: "language",
         subtag: "tmf",
         description: "Toba-Maskoy",
         added: "2009-07-29"}).
iana(language,tmg,tag{
         type: "language",
         subtag: "tmg",
         description: "Ternateño",
         added: "2009-07-29"}).
iana(language,tmh,tag{
         type: "language",
         subtag: "tmh",
         description: "Tamashek",
         added: "2005-10-16",
         suppress_script: "Latn",
         scope: "macrolanguage"}).
iana(language,tmi,tag{
         type: "language",
         subtag: "tmi",
         description: "Tutuba",
         added: "2009-07-29"}).
iana(language,tmj,tag{
         type: "language",
         subtag: "tmj",
         description: "Samarokena",
         added: "2009-07-29"}).
iana(language,tmk,tag{
         type: "language",
         subtag: "tmk",
         description: "Northwestern Tamang",
         added: "2009-07-29"}).
iana(language,tml,tag{
         type: "language",
         subtag: "tml",
         description: "Tamnim Citak",
         added: "2009-07-29"}).
iana(language,tmm,tag{
         type: "language",
         subtag: "tmm",
         description: "Tai Thanh",
         added: "2009-07-29"}).
iana(language,tmn,tag{
         type: "language",
         subtag: "tmn",
         description: "Taman (Indonesia)",
         added: "2009-07-29"}).
iana(language,tmo,tag{
         type: "language",
         subtag: "tmo",
         description: "Temoq",
         added: "2009-07-29"}).
iana(language,tmp,tag{
         type: "language",
         subtag: "tmp",
         description: "Tai Mène",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "tyj"}).
iana(language,tmq,tag{
         type: "language",
         subtag: "tmq",
         description: "Tumleo",
         added: "2009-07-29"}).
iana(language,tmr,tag{
         type: "language",
         subtag: "tmr",
         description: "Jewish Babylonian Aramaic (ca. 200-1200 CE)",
         added: "2009-07-29"}).
iana(language,tms,tag{
         type: "language",
         subtag: "tms",
         description: "Tima",
         added: "2009-07-29"}).
iana(language,tmt,tag{
         type: "language",
         subtag: "tmt",
         description: "Tasmate",
         added: "2009-07-29"}).
iana(language,tmu,tag{
         type: "language",
         subtag: "tmu",
         description: "Iau",
         added: "2009-07-29"}).
iana(language,tmv,tag{
         type: "language",
         subtag: "tmv",
         description: "Tembo (Motembo)",
         added: "2009-07-29"}).
iana(language,tmw,tag{
         type: "language",
         subtag: "tmw",
         description: "Temuan",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,tmy,tag{
         type: "language",
         subtag: "tmy",
         description: "Tami",
         added: "2009-07-29"}).
iana(language,tmz,tag{
         type: "language",
         subtag: "tmz",
         description: "Tamanaku",
         added: "2009-07-29"}).
iana(language,tna,tag{
         type: "language",
         subtag: "tna",
         description: "Tacana",
         added: "2009-07-29"}).
iana(language,tnb,tag{
         type: "language",
         subtag: "tnb",
         description: "Western Tunebo",
         added: "2009-07-29"}).
iana(language,tnc,tag{
         type: "language",
         subtag: "tnc",
         description: "Tanimuca-Retuarã",
         added: "2009-07-29"}).
iana(language,tnd,tag{
         type: "language",
         subtag: "tnd",
         description: "Angosturas Tunebo",
         added: "2009-07-29"}).
iana(language,tne,tag{
         type: "language",
         subtag: "tne",
         description: "Tinoc Kallahan",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "kak"}).
iana(language,tnf,tag{
         type: "language",
         subtag: "tnf",
         description: "Tangshewi",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         preferred_value: "prs"}).
iana(language,tng,tag{
         type: "language",
         subtag: "tng",
         description: "Tobanga",
         added: "2009-07-29"}).
iana(language,tnh,tag{
         type: "language",
         subtag: "tnh",
         description: "Maiani",
         added: "2009-07-29"}).
iana(language,tni,tag{
         type: "language",
         subtag: "tni",
         description: "Tandia",
         added: "2009-07-29"}).
iana(language,tnk,tag{
         type: "language",
         subtag: "tnk",
         description: "Kwamera",
         added: "2009-07-29"}).
iana(language,tnl,tag{
         type: "language",
         subtag: "tnl",
         description: "Lenakel",
         added: "2009-07-29"}).
iana(language,tnm,tag{
         type: "language",
         subtag: "tnm",
         description: "Tabla",
         added: "2009-07-29"}).
iana(language,tnn,tag{
         type: "language",
         subtag: "tnn",
         description: "North Tanna",
         added: "2009-07-29"}).
iana(language,tno,tag{
         type: "language",
         subtag: "tno",
         description: "Toromono",
         added: "2009-07-29"}).
iana(language,tnp,tag{
         type: "language",
         subtag: "tnp",
         description: "Whitesands",
         added: "2009-07-29"}).
iana(language,tnq,tag{
         type: "language",
         subtag: "tnq",
         description: "Taino",
         added: "2009-07-29"}).
iana(language,tnr,tag{
         type: "language",
         subtag: "tnr",
         description: "Ménik",
         added: "2009-07-29"}).
iana(language,tns,tag{
         type: "language",
         subtag: "tns",
         description: "Tenis",
         added: "2009-07-29"}).
iana(language,tnt,tag{
         type: "language",
         subtag: "tnt",
         description: "Tontemboan",
         added: "2009-07-29"}).
iana(language,tnu,tag{
         type: "language",
         subtag: "tnu",
         description: "Tay Khang",
         added: "2009-07-29"}).
iana(language,tnv,tag{
         type: "language",
         subtag: "tnv",
         description: "Tangchangya",
         added: "2009-07-29"}).
iana(language,tnw,tag{
         type: "language",
         subtag: "tnw",
         description: "Tonsawang",
         added: "2009-07-29"}).
iana(language,tnx,tag{
         type: "language",
         subtag: "tnx",
         description: "Tanema",
         added: "2009-07-29"}).
iana(language,tny,tag{
         type: "language",
         subtag: "tny",
         description: "Tongwe",
         added: "2009-07-29"}).
iana(language,tnz,tag{
         type: "language",
         subtag: "tnz",
         description: "Ten'edn",
         added: "2009-07-29"}).
iana(language,tob,tag{
         type: "language",
         subtag: "tob",
         description: "Toba",
         added: "2009-07-29"}).
iana(language,toc,tag{
         type: "language",
         subtag: "toc",
         description: "Coyutla Totonac",
         added: "2009-07-29"}).
iana(language,tod,tag{
         type: "language",
         subtag: "tod",
         description: "Toma",
         added: "2009-07-29"}).
iana(language,toe,tag{
         type: "language",
         subtag: "toe",
         description: "Tomedes",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,tof,tag{
         type: "language",
         subtag: "tof",
         description: "Gizrra",
         added: "2009-07-29"}).
iana(language,tog,tag{
         type: "language",
         subtag: "tog",
         description: "Tonga (Nyasa)",
         added: "2005-10-16"}).
iana(language,toh,tag{
         type: "language",
         subtag: "toh",
         description: "Gitonga",
         added: "2009-07-29"}).
iana(language,toi,tag{
         type: "language",
         subtag: "toi",
         description: "Tonga (Zambia)",
         added: "2009-07-29"}).
iana(language,toj,tag{
         type: "language",
         subtag: "toj",
         description: "Tojolabal",
         added: "2009-07-29"}).
iana(language,tok,tag{
         type: "language",
         subtag: "tok",
         description: "Toki Pona",
         added: "2022-02-25"}).
iana(language,tol,tag{
         type: "language",
         subtag: "tol",
         description: "Tolowa",
         added: "2009-07-29"}).
iana(language,tom,tag{
         type: "language",
         subtag: "tom",
         description: "Tombulu",
         added: "2009-07-29"}).
iana(language,too,tag{
         type: "language",
         subtag: "too",
         description: "Xicotepec De Juárez Totonac",
         added: "2009-07-29"}).
iana(language,top,tag{
         type: "language",
         subtag: "top",
         description: "Papantla Totonac",
         added: "2009-07-29"}).
iana(language,toq,tag{
         type: "language",
         subtag: "toq",
         description: "Toposa",
         added: "2009-07-29"}).
iana(language,tor,tag{
         type: "language",
         subtag: "tor",
         description: "Togbo-Vara Banda",
         added: "2009-07-29"}).
iana(language,tos,tag{
         type: "language",
         subtag: "tos",
         description: "Highland Totonac",
         added: "2009-07-29"}).
iana(language,tou,tag{
         type: "language",
         subtag: "tou",
         description: "Tho",
         added: "2009-07-29"}).
iana(language,tov,tag{
         type: "language",
         subtag: "tov",
         description: "Upper Taromi",
         added: "2009-07-29"}).
iana(language,tow,tag{
         type: "language",
         subtag: "tow",
         description: "Jemez",
         added: "2009-07-29"}).
iana(language,tox,tag{
         type: "language",
         subtag: "tox",
         description: "Tobian",
         added: "2009-07-29"}).
iana(language,toy,tag{
         type: "language",
         subtag: "toy",
         description: "Topoiyo",
         added: "2009-07-29"}).
iana(language,toz,tag{
         type: "language",
         subtag: "toz",
         description: "To",
         added: "2009-07-29"}).
iana(language,tpa,tag{
         type: "language",
         subtag: "tpa",
         description: "Taupota",
         added: "2009-07-29"}).
iana(language,tpc,tag{
         type: "language",
         subtag: "tpc",
         description: ["Azoyú Me'phaa","Azoyú Tlapanec"],
         added: "2009-07-29"}).
iana(language,tpe,tag{
         type: "language",
         subtag: "tpe",
         description: "Tippera",
         added: "2009-07-29"}).
iana(language,tpf,tag{
         type: "language",
         subtag: "tpf",
         description: "Tarpia",
         added: "2009-07-29"}).
iana(language,tpg,tag{
         type: "language",
         subtag: "tpg",
         description: "Kula",
         added: "2009-07-29"}).
iana(language,tpi,tag{
         type: "language",
         subtag: "tpi",
         description: "Tok Pisin",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tpj,tag{
         type: "language",
         subtag: "tpj",
         description: "Tapieté",
         added: "2009-07-29"}).
iana(language,tpk,tag{
         type: "language",
         subtag: "tpk",
         description: "Tupinikin",
         added: "2009-07-29"}).
iana(language,tpl,tag{
         type: "language",
         subtag: "tpl",
         description: ["Tlacoapa Me'phaa","Tlacoapa Tlapanec"],
         added: "2009-07-29"}).
iana(language,tpm,tag{
         type: "language",
         subtag: "tpm",
         description: "Tampulma",
         added: "2009-07-29"}).
iana(language,tpn,tag{
         type: "language",
         subtag: "tpn",
         description: "Tupinambá",
         added: "2009-07-29"}).
iana(language,tpo,tag{
         type: "language",
         subtag: "tpo",
         description: "Tai Pao",
         added: "2009-07-29"}).
iana(language,tpp,tag{
         type: "language",
         subtag: "tpp",
         description: "Pisaflores Tepehua",
         added: "2009-07-29"}).
iana(language,tpq,tag{
         type: "language",
         subtag: "tpq",
         description: "Tukpa",
         added: "2009-07-29"}).
iana(language,tpr,tag{
         type: "language",
         subtag: "tpr",
         description: "Tuparí",
         added: "2009-07-29"}).
iana(language,tpt,tag{
         type: "language",
         subtag: "tpt",
         description: "Tlachichilco Tepehua",
         added: "2009-07-29"}).
iana(language,tpu,tag{
         type: "language",
         subtag: "tpu",
         description: "Tampuan",
         added: "2009-07-29"}).
iana(language,tpv,tag{
         type: "language",
         subtag: "tpv",
         description: "Tanapag",
         added: "2009-07-29"}).
iana(language,tpw,tag{
         type: "language",
         subtag: "tpw",
         description: "Tupí",
         added: "2009-07-29"}).
iana(language,tpx,tag{
         type: "language",
         subtag: "tpx",
         description: ["Acatepec Me'phaa","Acatepec Tlapanec"],
         added: "2009-07-29"}).
iana(language,tpy,tag{
         type: "language",
         subtag: "tpy",
         description: "Trumai",
         added: "2009-07-29"}).
iana(language,tpz,tag{
         type: "language",
         subtag: "tpz",
         description: "Tinputz",
         added: "2009-07-29"}).
iana(language,tqb,tag{
         type: "language",
         subtag: "tqb",
         description: "Tembé",
         added: "2009-07-29"}).
iana(language,tql,tag{
         type: "language",
         subtag: "tql",
         description: "Lehali",
         added: "2009-07-29"}).
iana(language,tqm,tag{
         type: "language",
         subtag: "tqm",
         description: "Turumsa",
         added: "2009-07-29"}).
iana(language,tqn,tag{
         type: "language",
         subtag: "tqn",
         description: "Tenino",
         added: "2009-07-29"}).
iana(language,tqo,tag{
         type: "language",
         subtag: "tqo",
         description: "Toaripi",
         added: "2009-07-29"}).
iana(language,tqp,tag{
         type: "language",
         subtag: "tqp",
         description: "Tomoip",
         added: "2009-07-29"}).
iana(language,tqq,tag{
         type: "language",
         subtag: "tqq",
         description: "Tunni",
         added: "2009-07-29"}).
iana(language,tqr,tag{
         type: "language",
         subtag: "tqr",
         description: "Torona",
         added: "2009-07-29"}).
iana(language,tqt,tag{
         type: "language",
         subtag: "tqt",
         description: "Western Totonac",
         added: "2009-07-29"}).
iana(language,tqu,tag{
         type: "language",
         subtag: "tqu",
         description: "Touo",
         added: "2009-07-29"}).
iana(language,tqw,tag{
         type: "language",
         subtag: "tqw",
         description: "Tonkawa",
         added: "2009-07-29"}).
iana(language,tra,tag{
         type: "language",
         subtag: "tra",
         description: "Tirahi",
         added: "2009-07-29"}).
iana(language,trb,tag{
         type: "language",
         subtag: "trb",
         description: "Terebu",
         added: "2009-07-29"}).
iana(language,trc,tag{
         type: "language",
         subtag: "trc",
         description: "Copala Triqui",
         added: "2009-07-29"}).
iana(language,trd,tag{
         type: "language",
         subtag: "trd",
         description: "Turi",
         added: "2009-07-29"}).
iana(language,tre,tag{
         type: "language",
         subtag: "tre",
         description: "East Tarangan",
         added: "2009-07-29"}).
iana(language,trf,tag{
         type: "language",
         subtag: "trf",
         description: "Trinidadian Creole English",
         added: "2009-07-29"}).
iana(language,trg,tag{
         type: "language",
         subtag: "trg",
         description: "Lishán Didán",
         added: "2009-07-29"}).
iana(language,trh,tag{
         type: "language",
         subtag: "trh",
         description: "Turaka",
         added: "2009-07-29"}).
iana(language,tri,tag{
         type: "language",
         subtag: "tri",
         description: "Trió",
         added: "2009-07-29"}).
iana(language,trj,tag{
         type: "language",
         subtag: "trj",
         description: "Toram",
         added: "2009-07-29"}).
iana(language,trk,tag{
         type: "language",
         subtag: "trk",
         description: "Turkic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,trl,tag{
         type: "language",
         subtag: "trl",
         description: "Traveller Scottish",
         added: "2009-07-29"}).
iana(language,trm,tag{
         type: "language",
         subtag: "trm",
         description: "Tregami",
         added: "2009-07-29"}).
iana(language,trn,tag{
         type: "language",
         subtag: "trn",
         description: "Trinitario",
         added: "2009-07-29"}).
iana(language,tro,tag{
         type: "language",
         subtag: "tro",
         description: "Tarao Naga",
         added: "2009-07-29"}).
iana(language,trp,tag{
         type: "language",
         subtag: "trp",
         description: "Kok Borok",
         added: "2009-07-29"}).
iana(language,trq,tag{
         type: "language",
         subtag: "trq",
         description: "San Martín Itunyoso Triqui",
         added: "2009-07-29"}).
iana(language,trr,tag{
         type: "language",
         subtag: "trr",
         description: "Taushiro",
         added: "2009-07-29"}).
iana(language,trs,tag{
         type: "language",
         subtag: "trs",
         description: "Chicahuaxtla Triqui",
         added: "2009-07-29"}).
iana(language,trt,tag{
         type: "language",
         subtag: "trt",
         description: "Tunggare",
         added: "2009-07-29"}).
iana(language,tru,tag{
         type: "language",
         subtag: "tru",
         description: ["Turoyo","Surayt"],
         added: "2009-07-29"}).
iana(language,trv,tag{
         type: "language",
         subtag: "trv",
         description: ["Sediq","Seediq","Taroko"],
         added: "2009-07-29"}).
iana(language,trw,tag{
         type: "language",
         subtag: "trw",
         description: "Torwali",
         added: "2009-07-29"}).
iana(language,trx,tag{
         type: "language",
         subtag: "trx",
         description: "Tringgus-Sembaan Bidayuh",
         added: "2009-07-29"}).
iana(language,try,tag{
         type: "language",
         subtag: "try",
         description: "Turung",
         added: "2009-07-29"}).
iana(language,trz,tag{
         type: "language",
         subtag: "trz",
         description: "Torá",
         added: "2009-07-29"}).
iana(language,tsa,tag{
         type: "language",
         subtag: "tsa",
         description: "Tsaangi",
         added: "2009-07-29"}).
iana(language,tsb,tag{
         type: "language",
         subtag: "tsb",
         description: "Tsamai",
         added: "2009-07-29"}).
iana(language,tsc,tag{
         type: "language",
         subtag: "tsc",
         description: "Tswa",
         added: "2009-07-29"}).
iana(language,tsd,tag{
         type: "language",
         subtag: "tsd",
         description: "Tsakonian",
         added: "2009-07-29"}).
iana(language,tse,tag{
         type: "language",
         subtag: "tse",
         description: "Tunisian Sign Language",
         added: "2009-07-29"}).
iana(language,tsf,tag{
         type: "language",
         subtag: "tsf",
         description: "Southwestern Tamang",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "taj"}).
iana(language,tsg,tag{
         type: "language",
         subtag: "tsg",
         description: "Tausug",
         added: "2009-07-29"}).
iana(language,tsh,tag{
         type: "language",
         subtag: "tsh",
         description: "Tsuvan",
         added: "2009-07-29"}).
iana(language,tsi,tag{
         type: "language",
         subtag: "tsi",
         description: "Tsimshian",
         added: "2005-10-16"}).
iana(language,tsj,tag{
         type: "language",
         subtag: "tsj",
         description: "Tshangla",
         added: "2009-07-29"}).
iana(language,tsk,tag{
         type: "language",
         subtag: "tsk",
         description: "Tseku",
         added: "2009-07-29"}).
iana(language,tsl,tag{
         type: "language",
         subtag: "tsl",
         description: "Ts'ün-Lao",
         added: "2009-07-29"}).
iana(language,tsm,tag{
         type: "language",
         subtag: "tsm",
         description: ["Turkish Sign Language","Türk İşaret Dili"],
         added: "2009-07-29"}).
iana(language,tsp,tag{
         type: "language",
         subtag: "tsp",
         description: "Northern Toussian",
         added: "2009-07-29"}).
iana(language,tsq,tag{
         type: "language",
         subtag: "tsq",
         description: "Thai Sign Language",
         added: "2009-07-29"}).
iana(language,tsr,tag{
         type: "language",
         subtag: "tsr",
         description: "Akei",
         added: "2009-07-29"}).
iana(language,tss,tag{
         type: "language",
         subtag: "tss",
         description: "Taiwan Sign Language",
         added: "2009-07-29"}).
iana(language,tst,tag{
         type: "language",
         subtag: "tst",
         description: "Tondi Songway Kiini",
         added: "2011-08-16"}).
iana(language,tsu,tag{
         type: "language",
         subtag: "tsu",
         description: "Tsou",
         added: "2009-07-29"}).
iana(language,tsv,tag{
         type: "language",
         subtag: "tsv",
         description: "Tsogo",
         added: "2009-07-29"}).
iana(language,tsw,tag{
         type: "language",
         subtag: "tsw",
         description: "Tsishingini",
         added: "2009-07-29"}).
iana(language,tsx,tag{
         type: "language",
         subtag: "tsx",
         description: "Mubami",
         added: "2009-07-29"}).
iana(language,tsy,tag{
         type: "language",
         subtag: "tsy",
         description: "Tebul Sign Language",
         added: "2009-07-29"}).
iana(language,tsz,tag{
         type: "language",
         subtag: "tsz",
         description: "Purepecha",
         added: "2009-07-29"}).
iana(language,tta,tag{
         type: "language",
         subtag: "tta",
         description: "Tutelo",
         added: "2009-07-29"}).
iana(language,ttb,tag{
         type: "language",
         subtag: "ttb",
         description: "Gaa",
         added: "2009-07-29"}).
iana(language,ttc,tag{
         type: "language",
         subtag: "ttc",
         description: "Tektiteko",
         added: "2009-07-29"}).
iana(language,ttd,tag{
         type: "language",
         subtag: "ttd",
         description: "Tauade",
         added: "2009-07-29"}).
iana(language,tte,tag{
         type: "language",
         subtag: "tte",
         description: "Bwanabwana",
         added: "2009-07-29"}).
iana(language,ttf,tag{
         type: "language",
         subtag: "ttf",
         description: "Tuotomb",
         added: "2009-07-29"}).
iana(language,ttg,tag{
         type: "language",
         subtag: "ttg",
         description: "Tutong",
         added: "2009-07-29"}).
iana(language,tth,tag{
         type: "language",
         subtag: "tth",
         description: "Upper Ta'oih",
         added: "2009-07-29"}).
iana(language,tti,tag{
         type: "language",
         subtag: "tti",
         description: "Tobati",
         added: "2009-07-29"}).
iana(language,ttj,tag{
         type: "language",
         subtag: "ttj",
         description: "Tooro",
         added: "2009-07-29"}).
iana(language,ttk,tag{
         type: "language",
         subtag: "ttk",
         description: "Totoro",
         added: "2009-07-29"}).
iana(language,ttl,tag{
         type: "language",
         subtag: "ttl",
         description: "Totela",
         added: "2009-07-29"}).
iana(language,ttm,tag{
         type: "language",
         subtag: "ttm",
         description: "Northern Tutchone",
         added: "2009-07-29"}).
iana(language,ttn,tag{
         type: "language",
         subtag: "ttn",
         description: "Towei",
         added: "2009-07-29"}).
iana(language,tto,tag{
         type: "language",
         subtag: "tto",
         description: "Lower Ta'oih",
         added: "2009-07-29"}).
iana(language,ttp,tag{
         type: "language",
         subtag: "ttp",
         description: "Tombelala",
         added: "2009-07-29"}).
iana(language,ttq,tag{
         type: "language",
         subtag: "ttq",
         description: "Tawallammat Tamajaq",
         added: "2009-07-29",
         macrolanguage: "tmh"}).
iana(language,ttr,tag{
         type: "language",
         subtag: "ttr",
         description: "Tera",
         added: "2009-07-29"}).
iana(language,tts,tag{
         type: "language",
         subtag: "tts",
         description: "Northeastern Thai",
         added: "2009-07-29"}).
iana(language,ttt,tag{
         type: "language",
         subtag: "ttt",
         description: "Muslim Tat",
         added: "2009-07-29"}).
iana(language,ttu,tag{
         type: "language",
         subtag: "ttu",
         description: "Torau",
         added: "2009-07-29"}).
iana(language,ttv,tag{
         type: "language",
         subtag: "ttv",
         description: "Titan",
         added: "2009-07-29"}).
iana(language,ttw,tag{
         type: "language",
         subtag: "ttw",
         description: "Long Wat",
         added: "2009-07-29"}).
iana(language,tty,tag{
         type: "language",
         subtag: "tty",
         description: "Sikaritai",
         added: "2009-07-29"}).
iana(language,ttz,tag{
         type: "language",
         subtag: "ttz",
         description: "Tsum",
         added: "2009-07-29"}).
iana(language,tua,tag{
         type: "language",
         subtag: "tua",
         description: "Wiarumus",
         added: "2009-07-29"}).
iana(language,tub,tag{
         type: "language",
         subtag: "tub",
         description: "Tübatulabal",
         added: "2009-07-29"}).
iana(language,tuc,tag{
         type: "language",
         subtag: "tuc",
         description: "Mutu",
         added: "2009-07-29"}).
iana(language,tud,tag{
         type: "language",
         subtag: "tud",
         description: "Tuxá",
         added: "2009-07-29"}).
iana(language,tue,tag{
         type: "language",
         subtag: "tue",
         description: "Tuyuca",
         added: "2009-07-29"}).
iana(language,tuf,tag{
         type: "language",
         subtag: "tuf",
         description: "Central Tunebo",
         added: "2009-07-29"}).
iana(language,tug,tag{
         type: "language",
         subtag: "tug",
         description: "Tunia",
         added: "2009-07-29"}).
iana(language,tuh,tag{
         type: "language",
         subtag: "tuh",
         description: "Taulil",
         added: "2009-07-29"}).
iana(language,tui,tag{
         type: "language",
         subtag: "tui",
         description: "Tupuri",
         added: "2009-07-29"}).
iana(language,tuj,tag{
         type: "language",
         subtag: "tuj",
         description: "Tugutil",
         added: "2009-07-29"}).
iana(language,tul,tag{
         type: "language",
         subtag: "tul",
         description: "Tula",
         added: "2009-07-29"}).
iana(language,tum,tag{
         type: "language",
         subtag: "tum",
         description: "Tumbuka",
         added: "2005-10-16"}).
iana(language,tun,tag{
         type: "language",
         subtag: "tun",
         description: "Tunica",
         added: "2009-07-29"}).
iana(language,tuo,tag{
         type: "language",
         subtag: "tuo",
         description: "Tucano",
         added: "2009-07-29"}).
iana(language,tup,tag{
         type: "language",
         subtag: "tup",
         description: "Tupi languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,tuq,tag{
         type: "language",
         subtag: "tuq",
         description: "Tedaga",
         added: "2009-07-29"}).
iana(language,tus,tag{
         type: "language",
         subtag: "tus",
         description: "Tuscarora",
         added: "2009-07-29"}).
iana(language,tut,tag{
         type: "language",
         subtag: "tut",
         description: "Altaic languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,tuu,tag{
         type: "language",
         subtag: "tuu",
         description: "Tututni",
         added: "2009-07-29"}).
iana(language,tuv,tag{
         type: "language",
         subtag: "tuv",
         description: "Turkana",
         added: "2009-07-29"}).
iana(language,tuw,tag{
         type: "language",
         subtag: "tuw",
         description: "Tungus languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,tux,tag{
         type: "language",
         subtag: "tux",
         description: "Tuxináwa",
         added: "2009-07-29"}).
iana(language,tuy,tag{
         type: "language",
         subtag: "tuy",
         description: "Tugen",
         added: "2009-07-29",
         macrolanguage: "kln"}).
iana(language,tuz,tag{
         type: "language",
         subtag: "tuz",
         description: "Turka",
         added: "2009-07-29"}).
iana(language,tva,tag{
         type: "language",
         subtag: "tva",
         description: "Vaghua",
         added: "2009-07-29"}).
iana(language,tvd,tag{
         type: "language",
         subtag: "tvd",
         description: "Tsuvadi",
         added: "2009-07-29"}).
iana(language,tve,tag{
         type: "language",
         subtag: "tve",
         description: "Te'un",
         added: "2009-07-29"}).
iana(language,tvk,tag{
         type: "language",
         subtag: "tvk",
         description: "Southeast Ambrym",
         added: "2009-07-29"}).
iana(language,tvl,tag{
         type: "language",
         subtag: "tvl",
         description: "Tuvalu",
         added: "2005-10-16",
         suppress_script: "Latn"}).
iana(language,tvm,tag{
         type: "language",
         subtag: "tvm",
         description: "Tela-Masbuar",
         added: "2009-07-29"}).
iana(language,tvn,tag{
         type: "language",
         subtag: "tvn",
         description: "Tavoyan",
         added: "2009-07-29"}).
iana(language,tvo,tag{
         type: "language",
         subtag: "tvo",
         description: "Tidore",
         added: "2009-07-29"}).
iana(language,tvs,tag{
         type: "language",
         subtag: "tvs",
         description: "Taveta",
         added: "2009-07-29"}).
iana(language,tvt,tag{
         type: "language",
         subtag: "tvt",
         description: "Tutsa Naga",
         added: "2009-07-29"}).
iana(language,tvu,tag{
         type: "language",
         subtag: "tvu",
         description: "Tunen",
         added: "2012-08-12"}).
iana(language,tvw,tag{
         type: "language",
         subtag: "tvw",
         description: "Sedoa",
         added: "2009-07-29"}).
iana(language,tvx,tag{
         type: "language",
         subtag: "tvx",
         description: "Taivoan",
         added: "2019-04-16"}).
iana(language,tvy,tag{
         type: "language",
         subtag: "tvy",
         description: "Timor Pidgin",
         added: "2009-07-29"}).
iana(language,twa,tag{
         type: "language",
         subtag: "twa",
         description: "Twana",
         added: "2009-07-29"}).
iana(language,twb,tag{
         type: "language",
         subtag: "twb",
         description: "Western Tawbuid",
         added: "2009-07-29"}).
iana(language,twc,tag{
         type: "language",
         subtag: "twc",
         description: "Teshenawa",
         added: "2009-07-29"}).
iana(language,twd,tag{
         type: "language",
         subtag: "twd",
         description: "Twents",
         added: "2009-07-29"}).
iana(language,twe,tag{
         type: "language",
         subtag: "twe",
         description: "Tewa (Indonesia)",
         added: "2009-07-29"}).
iana(language,twf,tag{
         type: "language",
         subtag: "twf",
         description: "Northern Tiwa",
         added: "2009-07-29"}).
iana(language,twg,tag{
         type: "language",
         subtag: "twg",
         description: "Tereweng",
         added: "2009-07-29"}).
iana(language,twh,tag{
         type: "language",
         subtag: "twh",
         description: "Tai Dón",
         added: "2009-07-29"}).
iana(language,twl,tag{
         type: "language",
         subtag: "twl",
         description: "Tawara",
         added: "2009-07-29"}).
iana(language,twm,tag{
         type: "language",
         subtag: "twm",
         description: "Tawang Monpa",
         added: "2009-07-29"}).
iana(language,twn,tag{
         type: "language",
         subtag: "twn",
         description: "Twendi",
         added: "2009-07-29"}).
iana(language,two,tag{
         type: "language",
         subtag: "two",
         description: "Tswapong",
         added: "2009-07-29"}).
iana(language,twp,tag{
         type: "language",
         subtag: "twp",
         description: "Ere",
         added: "2009-07-29"}).
iana(language,twq,tag{
         type: "language",
         subtag: "twq",
         description: "Tasawaq",
         added: "2009-07-29"}).
iana(language,twr,tag{
         type: "language",
         subtag: "twr",
         description: "Southwestern Tarahumara",
         added: "2009-07-29"}).
iana(language,twt,tag{
         type: "language",
         subtag: "twt",
         description: "Turiwára",
         added: "2009-07-29"}).
iana(language,twu,tag{
         type: "language",
         subtag: "twu",
         description: "Termanu",
         added: "2009-07-29"}).
iana(language,tww,tag{
         type: "language",
         subtag: "tww",
         description: "Tuwari",
         added: "2009-07-29"}).
iana(language,twx,tag{
         type: "language",
         subtag: "twx",
         description: "Tewe",
         added: "2009-07-29"}).
iana(language,twy,tag{
         type: "language",
         subtag: "twy",
         description: "Tawoyan",
         added: "2009-07-29"}).
iana(language,txa,tag{
         type: "language",
         subtag: "txa",
         description: "Tombonuo",
         added: "2009-07-29"}).
iana(language,txb,tag{
         type: "language",
         subtag: "txb",
         description: "Tokharian B",
         added: "2009-07-29"}).
iana(language,txc,tag{
         type: "language",
         subtag: "txc",
         description: "Tsetsaut",
         added: "2009-07-29"}).
iana(language,txe,tag{
         type: "language",
         subtag: "txe",
         description: "Totoli",
         added: "2009-07-29"}).
iana(language,txg,tag{
         type: "language",
         subtag: "txg",
         description: "Tangut",
         added: "2009-07-29"}).
iana(language,txh,tag{
         type: "language",
         subtag: "txh",
         description: "Thracian",
         added: "2009-07-29"}).
iana(language,txi,tag{
         type: "language",
         subtag: "txi",
         description: "Ikpeng",
         added: "2009-07-29"}).
iana(language,txj,tag{
         type: "language",
         subtag: "txj",
         description: "Tarjumo",
         added: "2015-02-12"}).
iana(language,txm,tag{
         type: "language",
         subtag: "txm",
         description: "Tomini",
         added: "2009-07-29"}).
iana(language,txn,tag{
         type: "language",
         subtag: "txn",
         description: "West Tarangan",
         added: "2009-07-29"}).
iana(language,txo,tag{
         type: "language",
         subtag: "txo",
         description: "Toto",
         added: "2009-07-29"}).
iana(language,txq,tag{
         type: "language",
         subtag: "txq",
         description: "Tii",
         added: "2009-07-29"}).
iana(language,txr,tag{
         type: "language",
         subtag: "txr",
         description: "Tartessian",
         added: "2009-07-29"}).
iana(language,txs,tag{
         type: "language",
         subtag: "txs",
         description: "Tonsea",
         added: "2009-07-29"}).
iana(language,txt,tag{
         type: "language",
         subtag: "txt",
         description: "Citak",
         added: "2009-07-29"}).
iana(language,txu,tag{
         type: "language",
         subtag: "txu",
         description: "Kayapó",
         added: "2009-07-29"}).
iana(language,txx,tag{
         type: "language",
         subtag: "txx",
         description: "Tatana",
         added: "2009-07-29"}).
iana(language,txy,tag{
         type: "language",
         subtag: "txy",
         description: "Tanosy Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,tya,tag{
         type: "language",
         subtag: "tya",
         description: "Tauya",
         added: "2009-07-29"}).
iana(language,tye,tag{
         type: "language",
         subtag: "tye",
         description: "Kyanga",
         added: "2009-07-29"}).
iana(language,tyh,tag{
         type: "language",
         subtag: "tyh",
         description: "O'du",
         added: "2009-07-29"}).
iana(language,tyi,tag{
         type: "language",
         subtag: "tyi",
         description: "Teke-Tsaayi",
         added: "2009-07-29"}).
iana(language,tyj,tag{
         type: "language",
         subtag: "tyj",
         description: ["Tai Do","Tai Yo"],
         added: "2009-07-29"}).
iana(language,tyl,tag{
         type: "language",
         subtag: "tyl",
         description: "Thu Lao",
         added: "2009-07-29"}).
iana(language,tyn,tag{
         type: "language",
         subtag: "tyn",
         description: "Kombai",
         added: "2009-07-29"}).
iana(language,typ,tag{
         type: "language",
         subtag: "typ",
         description: "Thaypan",
         added: "2009-07-29"}).
iana(language,tyr,tag{
         type: "language",
         subtag: "tyr",
         description: "Tai Daeng",
         added: "2009-07-29"}).
iana(language,tys,tag{
         type: "language",
         subtag: "tys",
         description: "Tày Sa Pa",
         added: "2009-07-29"}).
iana(language,tyt,tag{
         type: "language",
         subtag: "tyt",
         description: "Tày Tac",
         added: "2009-07-29"}).
iana(language,tyu,tag{
         type: "language",
         subtag: "tyu",
         description: "Kua",
         added: "2009-07-29"}).
iana(language,tyv,tag{
         type: "language",
         subtag: "tyv",
         description: "Tuvinian",
         added: "2005-10-16"}).
iana(language,tyx,tag{
         type: "language",
         subtag: "tyx",
         description: "Teke-Tyee",
         added: "2009-07-29"}).
iana(language,tyy,tag{
         type: "language",
         subtag: "tyy",
         description: "Tiyaa",
         added: "2021-02-20"}).
iana(language,tyz,tag{
         type: "language",
         subtag: "tyz",
         description: "Tày",
         added: "2009-07-29"}).
iana(language,tza,tag{
         type: "language",
         subtag: "tza",
         description: "Tanzanian Sign Language",
         added: "2009-07-29"}).
iana(language,tzh,tag{
         type: "language",
         subtag: "tzh",
         description: "Tzeltal",
         added: "2009-07-29"}).
iana(language,tzj,tag{
         type: "language",
         subtag: "tzj",
         description: "Tz'utujil",
         added: "2009-07-29"}).
iana(language,tzl,tag{
         type: "language",
         subtag: "tzl",
         description: "Talossan",
         added: "2013-09-10"}).
iana(language,tzm,tag{
         type: "language",
         subtag: "tzm",
         description: "Central Atlas Tamazight",
         added: "2009-07-29"}).
iana(language,tzn,tag{
         type: "language",
         subtag: "tzn",
         description: "Tugun",
         added: "2009-07-29"}).
iana(language,tzo,tag{
         type: "language",
         subtag: "tzo",
         description: "Tzotzil",
         added: "2009-07-29"}).
iana(language,tzx,tag{
         type: "language",
         subtag: "tzx",
         description: "Tabriak",
         added: "2009-07-29"}).
iana(language,uam,tag{
         type: "language",
         subtag: "uam",
         description: "Uamué",
         added: "2009-07-29"}).
iana(language,uan,tag{
         type: "language",
         subtag: "uan",
         description: "Kuan",
         added: "2009-07-29"}).
iana(language,uar,tag{
         type: "language",
         subtag: "uar",
         description: "Tairuma",
         added: "2009-07-29"}).
iana(language,uba,tag{
         type: "language",
         subtag: "uba",
         description: "Ubang",
         added: "2009-07-29"}).
iana(language,ubi,tag{
         type: "language",
         subtag: "ubi",
         description: "Ubi",
         added: "2009-07-29"}).
iana(language,ubl,tag{
         type: "language",
         subtag: "ubl",
         description: "Buhi'non Bikol",
         added: "2010-03-11",
         macrolanguage: "bik"}).
iana(language,ubr,tag{
         type: "language",
         subtag: "ubr",
         description: "Ubir",
         added: "2009-07-29"}).
iana(language,ubu,tag{
         type: "language",
         subtag: "ubu",
         description: "Umbu-Ungu",
         added: "2009-07-29"}).
iana(language,uby,tag{
         type: "language",
         subtag: "uby",
         description: "Ubykh",
         added: "2009-07-29"}).
iana(language,uda,tag{
         type: "language",
         subtag: "uda",
         description: "Uda",
         added: "2009-07-29"}).
iana(language,ude,tag{
         type: "language",
         subtag: "ude",
         description: "Udihe",
         added: "2009-07-29"}).
iana(language,udg,tag{
         type: "language",
         subtag: "udg",
         description: "Muduga",
         added: "2009-07-29"}).
iana(language,udi,tag{
         type: "language",
         subtag: "udi",
         description: "Udi",
         added: "2009-07-29"}).
iana(language,udj,tag{
         type: "language",
         subtag: "udj",
         description: "Ujir",
         added: "2009-07-29"}).
iana(language,udl,tag{
         type: "language",
         subtag: "udl",
         description: "Wuzlam",
         added: "2009-07-29"}).
iana(language,udm,tag{
         type: "language",
         subtag: "udm",
         description: "Udmurt",
         added: "2005-10-16"}).
iana(language,udu,tag{
         type: "language",
         subtag: "udu",
         description: "Uduk",
         added: "2009-07-29"}).
iana(language,ues,tag{
         type: "language",
         subtag: "ues",
         description: "Kioko",
         added: "2009-07-29"}).
iana(language,ufi,tag{
         type: "language",
         subtag: "ufi",
         description: "Ufim",
         added: "2009-07-29"}).
iana(language,uga,tag{
         type: "language",
         subtag: "uga",
         description: "Ugaritic",
         added: "2005-10-16"}).
iana(language,ugb,tag{
         type: "language",
         subtag: "ugb",
         description: "Kuku-Ugbanh",
         added: "2009-07-29"}).
iana(language,uge,tag{
         type: "language",
         subtag: "uge",
         description: "Ughele",
         added: "2009-07-29"}).
iana(language,ugh,tag{
         type: "language",
         subtag: "ugh",
         description: "Kubachi",
         added: "2022-02-25"}).
iana(language,ugn,tag{
         type: "language",
         subtag: "ugn",
         description: "Ugandan Sign Language",
         added: "2009-07-29"}).
iana(language,ugo,tag{
         type: "language",
         subtag: "ugo",
         description: "Ugong",
         added: "2009-07-29"}).
iana(language,ugy,tag{
         type: "language",
         subtag: "ugy",
         description: "Uruguayan Sign Language",
         added: "2009-07-29"}).
iana(language,uha,tag{
         type: "language",
         subtag: "uha",
         description: "Uhami",
         added: "2009-07-29"}).
iana(language,uhn,tag{
         type: "language",
         subtag: "uhn",
         description: "Damal",
         added: "2009-07-29"}).
iana(language,uis,tag{
         type: "language",
         subtag: "uis",
         description: "Uisai",
         added: "2009-07-29"}).
iana(language,uiv,tag{
         type: "language",
         subtag: "uiv",
         description: "Iyive",
         added: "2009-07-29"}).
iana(language,uji,tag{
         type: "language",
         subtag: "uji",
         description: "Tanjijili",
         added: "2009-07-29"}).
iana(language,uka,tag{
         type: "language",
         subtag: "uka",
         description: "Kaburi",
         added: "2009-07-29"}).
iana(language,ukg,tag{
         type: "language",
         subtag: "ukg",
         description: "Ukuriguma",
         added: "2009-07-29"}).
iana(language,ukh,tag{
         type: "language",
         subtag: "ukh",
         description: "Ukhwejo",
         added: "2009-07-29"}).
iana(language,uki,tag{
         type: "language",
         subtag: "uki",
         description: "Kui (India)",
         added: "2020-03-28"}).
iana(language,ukk,tag{
         type: "language",
         subtag: "ukk",
         description: "Muak Sa-aak",
         added: "2017-02-23"}).
iana(language,ukl,tag{
         type: "language",
         subtag: "ukl",
         description: "Ukrainian Sign Language",
         added: "2009-07-29"}).
iana(language,ukp,tag{
         type: "language",
         subtag: "ukp",
         description: "Ukpe-Bayobiri",
         added: "2009-07-29"}).
iana(language,ukq,tag{
         type: "language",
         subtag: "ukq",
         description: "Ukwa",
         added: "2009-07-29"}).
iana(language,uks,tag{
         type: "language",
         subtag: "uks",
         description: ["Urubú-Kaapor Sign Language","Kaapor Sign Language"],
         added: "2009-07-29"}).
iana(language,uku,tag{
         type: "language",
         subtag: "uku",
         description: "Ukue",
         added: "2009-07-29"}).
iana(language,ukv,tag{
         type: "language",
         subtag: "ukv",
         description: "Kuku",
         added: "2020-03-28"}).
iana(language,ukw,tag{
         type: "language",
         subtag: "ukw",
         description: "Ukwuani-Aboh-Ndoni",
         added: "2009-07-29"}).
iana(language,uky,tag{
         type: "language",
         subtag: "uky",
         description: "Kuuk-Yak",
         added: "2012-08-12"}).
iana(language,ula,tag{
         type: "language",
         subtag: "ula",
         description: "Fungwa",
         added: "2009-07-29"}).
iana(language,ulb,tag{
         type: "language",
         subtag: "ulb",
         description: "Ulukwumi",
         added: "2009-07-29"}).
iana(language,ulc,tag{
         type: "language",
         subtag: "ulc",
         description: "Ulch",
         added: "2009-07-29"}).
iana(language,ule,tag{
         type: "language",
         subtag: "ule",
         description: "Lule",
         added: "2013-09-10"}).
iana(language,ulf,tag{
         type: "language",
         subtag: "ulf",
         description: ["Usku","Afra"],
         added: "2009-07-29"}).
iana(language,uli,tag{
         type: "language",
         subtag: "uli",
         description: "Ulithian",
         added: "2009-07-29"}).
iana(language,ulk,tag{
         type: "language",
         subtag: "ulk",
         description: "Meriam Mir",
         added: "2009-07-29"}).
iana(language,ull,tag{
         type: "language",
         subtag: "ull",
         description: "Ullatan",
         added: "2009-07-29"}).
iana(language,ulm,tag{
         type: "language",
         subtag: "ulm",
         description: "Ulumanda'",
         added: "2009-07-29"}).
iana(language,uln,tag{
         type: "language",
         subtag: "uln",
         description: "Unserdeutsch",
         added: "2009-07-29"}).
iana(language,ulu,tag{
         type: "language",
         subtag: "ulu",
         description: "Uma' Lung",
         added: "2009-07-29"}).
iana(language,ulw,tag{
         type: "language",
         subtag: "ulw",
         description: "Ulwa",
         added: "2010-03-11"}).
iana(language,uma,tag{
         type: "language",
         subtag: "uma",
         description: "Umatilla",
         added: "2009-07-29"}).
iana(language,umb,tag{
         type: "language",
         subtag: "umb",
         description: "Umbundu",
         added: "2005-10-16"}).
iana(language,umc,tag{
         type: "language",
         subtag: "umc",
         description: "Marrucinian",
         added: "2009-07-29"}).
iana(language,umd,tag{
         type: "language",
         subtag: "umd",
         description: "Umbindhamu",
         added: "2009-07-29"}).
iana(language,umg,tag{
         type: "language",
         subtag: "umg",
         description: ["Morrobalama","Umbuygamu"],
         added: "2009-07-29"}).
iana(language,umi,tag{
         type: "language",
         subtag: "umi",
         description: "Ukit",
         added: "2009-07-29"}).
iana(language,umm,tag{
         type: "language",
         subtag: "umm",
         description: "Umon",
         added: "2009-07-29"}).
iana(language,umn,tag{
         type: "language",
         subtag: "umn",
         description: "Makyan Naga",
         added: "2009-07-29"}).
iana(language,umo,tag{
         type: "language",
         subtag: "umo",
         description: "Umotína",
         added: "2009-07-29"}).
iana(language,ump,tag{
         type: "language",
         subtag: "ump",
         description: "Umpila",
         added: "2009-07-29"}).
iana(language,umr,tag{
         type: "language",
         subtag: "umr",
         description: "Umbugarla",
         added: "2009-07-29"}).
iana(language,ums,tag{
         type: "language",
         subtag: "ums",
         description: "Pendau",
         added: "2009-07-29"}).
iana(language,umu,tag{
         type: "language",
         subtag: "umu",
         description: "Munsee",
         added: "2009-07-29",
         macrolanguage: "del"}).
iana(language,una,tag{
         type: "language",
         subtag: "una",
         description: "North Watut",
         added: "2009-07-29"}).
iana(language,und,tag{
         type: "language",
         subtag: "und",
         description: "Undetermined",
         added: "2005-10-16",
         scope: "special"}).
iana(language,une,tag{
         type: "language",
         subtag: "une",
         description: "Uneme",
         added: "2009-07-29"}).
iana(language,ung,tag{
         type: "language",
         subtag: "ung",
         description: "Ngarinyin",
         added: "2009-07-29"}).
iana(language,uni,tag{
         type: "language",
         subtag: "uni",
         description: "Uni",
         added: "2021-02-20"}).
iana(language,unk,tag{
         type: "language",
         subtag: "unk",
         description: "Enawené-Nawé",
         added: "2009-07-29"}).
iana(language,unm,tag{
         type: "language",
         subtag: "unm",
         description: "Unami",
         added: "2009-07-29",
         macrolanguage: "del"}).
iana(language,unn,tag{
         type: "language",
         subtag: "unn",
         description: "Kurnai",
         added: "2013-09-10"}).
iana(language,unp,tag{
         type: "language",
         subtag: "unp",
         description: "Worora",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see wro, xgu"}).
iana(language,unr,tag{
         type: "language",
         subtag: "unr",
         description: "Mundari",
         added: "2009-07-29"}).
iana(language,unu,tag{
         type: "language",
         subtag: "unu",
         description: "Unubahe",
         added: "2013-09-10"}).
iana(language,unx,tag{
         type: "language",
         subtag: "unx",
         description: "Munda",
         added: "2009-07-29"}).
iana(language,unz,tag{
         type: "language",
         subtag: "unz",
         description: "Unde Kaili",
         added: "2009-07-29"}).
iana(language,uok,tag{
         type: "language",
         subtag: "uok",
         description: "Uokha",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "ema"}).
iana(language,uon,tag{
         type: "language",
         subtag: "uon",
         description: "Kulon",
         added: "2022-02-25"}).
iana(language,upi,tag{
         type: "language",
         subtag: "upi",
         description: "Umeda",
         added: "2009-07-29"}).
iana(language,upv,tag{
         type: "language",
         subtag: "upv",
         description: "Uripiv-Wala-Rano-Atchin",
         added: "2009-07-29"}).
iana(language,ura,tag{
         type: "language",
         subtag: "ura",
         description: "Urarina",
         added: "2009-07-29"}).
iana(language,urb,tag{
         type: "language",
         subtag: "urb",
         description: ["Urubú-Kaapor","Kaapor"],
         added: "2009-07-29"}).
iana(language,urc,tag{
         type: "language",
         subtag: "urc",
         description: "Urningangg",
         added: "2009-07-29"}).
iana(language,ure,tag{
         type: "language",
         subtag: "ure",
         description: "Uru",
         added: "2009-07-29"}).
iana(language,urf,tag{
         type: "language",
         subtag: "urf",
         description: "Uradhi",
         added: "2009-07-29"}).
iana(language,urg,tag{
         type: "language",
         subtag: "urg",
         description: "Urigina",
         added: "2009-07-29"}).
iana(language,urh,tag{
         type: "language",
         subtag: "urh",
         description: "Urhobo",
         added: "2009-07-29"}).
iana(language,uri,tag{
         type: "language",
         subtag: "uri",
         description: "Urim",
         added: "2009-07-29"}).
iana(language,urj,tag{
         type: "language",
         subtag: "urj",
         description: "Uralic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,urk,tag{
         type: "language",
         subtag: "urk",
         description: "Urak Lawoi'",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,url,tag{
         type: "language",
         subtag: "url",
         description: "Urali",
         added: "2009-07-29"}).
iana(language,urm,tag{
         type: "language",
         subtag: "urm",
         description: "Urapmin",
         added: "2009-07-29"}).
iana(language,urn,tag{
         type: "language",
         subtag: "urn",
         description: "Uruangnirin",
         added: "2009-07-29"}).
iana(language,uro,tag{
         type: "language",
         subtag: "uro",
         description: "Ura (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,urp,tag{
         type: "language",
         subtag: "urp",
         description: "Uru-Pa-In",
         added: "2009-07-29"}).
iana(language,urr,tag{
         type: "language",
         subtag: "urr",
         description: ["Lehalurup","Löyöp"],
         added: "2009-07-29"}).
iana(language,urt,tag{
         type: "language",
         subtag: "urt",
         description: "Urat",
         added: "2009-07-29"}).
iana(language,uru,tag{
         type: "language",
         subtag: "uru",
         description: "Urumi",
         added: "2009-07-29"}).
iana(language,urv,tag{
         type: "language",
         subtag: "urv",
         description: "Uruava",
         added: "2009-07-29"}).
iana(language,urw,tag{
         type: "language",
         subtag: "urw",
         description: "Sop",
         added: "2009-07-29"}).
iana(language,urx,tag{
         type: "language",
         subtag: "urx",
         description: "Urimo",
         added: "2009-07-29"}).
iana(language,ury,tag{
         type: "language",
         subtag: "ury",
         description: "Orya",
         added: "2009-07-29"}).
iana(language,urz,tag{
         type: "language",
         subtag: "urz",
         description: "Uru-Eu-Wau-Wau",
         added: "2009-07-29"}).
iana(language,usa,tag{
         type: "language",
         subtag: "usa",
         description: "Usarufa",
         added: "2009-07-29"}).
iana(language,ush,tag{
         type: "language",
         subtag: "ush",
         description: "Ushojo",
         added: "2009-07-29"}).
iana(language,usi,tag{
         type: "language",
         subtag: "usi",
         description: "Usui",
         added: "2009-07-29"}).
iana(language,usk,tag{
         type: "language",
         subtag: "usk",
         description: "Usaghade",
         added: "2009-07-29"}).
iana(language,usp,tag{
         type: "language",
         subtag: "usp",
         description: "Uspanteco",
         added: "2009-07-29"}).
iana(language,uss,tag{
         type: "language",
         subtag: "uss",
         description: "us-Saare",
         added: "2019-04-16"}).
iana(language,usu,tag{
         type: "language",
         subtag: "usu",
         description: "Uya",
         added: "2009-07-29"}).
iana(language,uta,tag{
         type: "language",
         subtag: "uta",
         description: "Otank",
         added: "2009-07-29"}).
iana(language,ute,tag{
         type: "language",
         subtag: "ute",
         description: "Ute-Southern Paiute",
         added: "2009-07-29"}).
iana(language,uth,tag{
         type: "language",
         subtag: "uth",
         description: "ut-Hun",
         added: "2019-04-16"}).
iana(language,utp,tag{
         type: "language",
         subtag: "utp",
         description: "Amba (Solomon Islands)",
         added: "2009-07-29"}).
iana(language,utr,tag{
         type: "language",
         subtag: "utr",
         description: "Etulo",
         added: "2009-07-29"}).
iana(language,utu,tag{
         type: "language",
         subtag: "utu",
         description: "Utu",
         added: "2009-07-29"}).
iana(language,uum,tag{
         type: "language",
         subtag: "uum",
         description: "Urum",
         added: "2009-07-29"}).
iana(language,uun,tag{
         type: "language",
         subtag: "uun",
         description: "Kulon-Pazeh",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         comments: "see pzh, uon"}).
iana(language,uur,tag{
         type: "language",
         subtag: "uur",
         description: "Ura (Vanuatu)",
         added: "2009-07-29"}).
iana(language,uuu,tag{
         type: "language",
         subtag: "uuu",
         description: "U",
         added: "2009-07-29"}).
iana(language,uve,tag{
         type: "language",
         subtag: "uve",
         description: ["West Uvean","Fagauvea"],
         added: "2009-07-29"}).
iana(language,uvh,tag{
         type: "language",
         subtag: "uvh",
         description: "Uri",
         added: "2009-07-29"}).
iana(language,uvl,tag{
         type: "language",
         subtag: "uvl",
         description: "Lote",
         added: "2009-07-29"}).
iana(language,uwa,tag{
         type: "language",
         subtag: "uwa",
         description: "Kuku-Uwanh",
         added: "2009-07-29"}).
iana(language,uya,tag{
         type: "language",
         subtag: "uya",
         description: "Doko-Uyanga",
         added: "2009-07-29"}).
iana(language,uzn,tag{
         type: "language",
         subtag: "uzn",
         description: "Northern Uzbek",
         added: "2009-07-29",
         macrolanguage: "uz"}).
iana(language,uzs,tag{
         type: "language",
         subtag: "uzs",
         description: "Southern Uzbek",
         added: "2009-07-29",
         macrolanguage: "uz"}).
iana(language,vaa,tag{
         type: "language",
         subtag: "vaa",
         description: "Vaagri Booli",
         added: "2009-07-29"}).
iana(language,vae,tag{
         type: "language",
         subtag: "vae",
         description: "Vale",
         added: "2009-07-29"}).
iana(language,vaf,tag{
         type: "language",
         subtag: "vaf",
         description: "Vafsi",
         added: "2009-07-29"}).
iana(language,vag,tag{
         type: "language",
         subtag: "vag",
         description: "Vagla",
         added: "2009-07-29"}).
iana(language,vah,tag{
         type: "language",
         subtag: "vah",
         description: "Varhadi-Nagpuri",
         added: "2009-07-29"}).
iana(language,vai,tag{
         type: "language",
         subtag: "vai",
         description: "Vai",
         added: "2005-10-16"}).
iana(language,vaj,tag{
         type: "language",
         subtag: "vaj",
         description: ["Sekele","Northwestern ǃKung","Vasekele"],
         added: "2009-07-29"}).
iana(language,val,tag{
         type: "language",
         subtag: "val",
         description: "Vehes",
         added: "2009-07-29"}).
iana(language,vam,tag{
         type: "language",
         subtag: "vam",
         description: "Vanimo",
         added: "2009-07-29"}).
iana(language,van,tag{
         type: "language",
         subtag: "van",
         description: "Valman",
         added: "2009-07-29"}).
iana(language,vao,tag{
         type: "language",
         subtag: "vao",
         description: "Vao",
         added: "2009-07-29"}).
iana(language,vap,tag{
         type: "language",
         subtag: "vap",
         description: "Vaiphei",
         added: "2009-07-29"}).
iana(language,var,tag{
         type: "language",
         subtag: "var",
         description: "Huarijio",
         added: "2009-07-29"}).
iana(language,vas,tag{
         type: "language",
         subtag: "vas",
         description: "Vasavi",
         added: "2009-07-29"}).
iana(language,vau,tag{
         type: "language",
         subtag: "vau",
         description: "Vanuma",
         added: "2009-07-29"}).
iana(language,vav,tag{
         type: "language",
         subtag: "vav",
         description: "Varli",
         added: "2009-07-29"}).
iana(language,vay,tag{
         type: "language",
         subtag: "vay",
         description: "Wayu",
         added: "2009-07-29"}).
iana(language,vbb,tag{
         type: "language",
         subtag: "vbb",
         description: "Southeast Babar",
         added: "2009-07-29"}).
iana(language,vbk,tag{
         type: "language",
         subtag: "vbk",
         description: "Southwestern Bontok",
         added: "2010-03-11",
         macrolanguage: "bnc"}).
iana(language,vec,tag{
         type: "language",
         subtag: "vec",
         description: "Venetian",
         added: "2009-07-29"}).
iana(language,ved,tag{
         type: "language",
         subtag: "ved",
         description: "Veddah",
         added: "2009-07-29"}).
iana(language,vel,tag{
         type: "language",
         subtag: "vel",
         description: "Veluws",
         added: "2009-07-29"}).
iana(language,vem,tag{
         type: "language",
         subtag: "vem",
         description: "Vemgo-Mabas",
         added: "2009-07-29"}).
iana(language,veo,tag{
         type: "language",
         subtag: "veo",
         description: "Ventureño",
         added: "2009-07-29"}).
iana(language,vep,tag{
         type: "language",
         subtag: "vep",
         description: "Veps",
         added: "2009-07-29"}).
iana(language,ver,tag{
         type: "language",
         subtag: "ver",
         description: "Mom Jango",
         added: "2009-07-29"}).
iana(language,vgr,tag{
         type: "language",
         subtag: "vgr",
         description: "Vaghri",
         added: "2009-07-29"}).
iana(language,vgt,tag{
         type: "language",
         subtag: "vgt",
         description: ["Vlaamse Gebarentaal","Flemish Sign Language"],
         added: "2009-07-29"}).
iana(language,vic,tag{
         type: "language",
         subtag: "vic",
         description: "Virgin Islands Creole English",
         added: "2009-07-29"}).
iana(language,vid,tag{
         type: "language",
         subtag: "vid",
         description: "Vidunda",
         added: "2009-07-29"}).
iana(language,vif,tag{
         type: "language",
         subtag: "vif",
         description: "Vili",
         added: "2009-07-29"}).
iana(language,vig,tag{
         type: "language",
         subtag: "vig",
         description: "Viemo",
         added: "2009-07-29"}).
iana(language,vil,tag{
         type: "language",
         subtag: "vil",
         description: "Vilela",
         added: "2009-07-29"}).
iana(language,vin,tag{
         type: "language",
         subtag: "vin",
         description: "Vinza",
         added: "2009-07-29"}).
iana(language,vis,tag{
         type: "language",
         subtag: "vis",
         description: "Vishavan",
         added: "2009-07-29"}).
iana(language,vit,tag{
         type: "language",
         subtag: "vit",
         description: "Viti",
         added: "2009-07-29"}).
iana(language,viv,tag{
         type: "language",
         subtag: "viv",
         description: "Iduna",
         added: "2009-07-29"}).
iana(language,vka,tag{
         type: "language",
         subtag: "vka",
         description: "Kariyarra",
         added: "2009-07-29"}).
iana(language,vki,tag{
         type: "language",
         subtag: "vki",
         description: "Ija-Zuba",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see vkn, vkz"}).
iana(language,vkj,tag{
         type: "language",
         subtag: "vkj",
         description: "Kujarge",
         added: "2009-07-29"}).
iana(language,vkk,tag{
         type: "language",
         subtag: "vkk",
         description: "Kaur",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,vkl,tag{
         type: "language",
         subtag: "vkl",
         description: "Kulisusu",
         added: "2009-07-29"}).
iana(language,vkm,tag{
         type: "language",
         subtag: "vkm",
         description: "Kamakan",
         added: "2009-07-29"}).
iana(language,vkn,tag{
         type: "language",
         subtag: "vkn",
         description: "Koro Nulu",
         added: "2021-02-20"}).
iana(language,vko,tag{
         type: "language",
         subtag: "vko",
         description: "Kodeoha",
         added: "2009-07-29"}).
iana(language,vkp,tag{
         type: "language",
         subtag: "vkp",
         description: "Korlai Creole Portuguese",
         added: "2009-07-29"}).
iana(language,vkt,tag{
         type: "language",
         subtag: "vkt",
         description: "Tenggarong Kutai Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,vku,tag{
         type: "language",
         subtag: "vku",
         description: "Kurrama",
         added: "2009-07-29"}).
iana(language,vkz,tag{
         type: "language",
         subtag: "vkz",
         description: "Koro Zuba",
         added: "2021-02-20"}).
iana(language,vlp,tag{
         type: "language",
         subtag: "vlp",
         description: "Valpei",
         added: "2009-07-29"}).
iana(language,vls,tag{
         type: "language",
         subtag: "vls",
         description: "Vlaams",
         added: "2009-07-29"}).
iana(language,vma,tag{
         type: "language",
         subtag: "vma",
         description: "Martuyhunira",
         added: "2009-07-29"}).
iana(language,vmb,tag{
         type: "language",
         subtag: "vmb",
         description: "Barbaram",
         added: "2009-07-29"}).
iana(language,vmc,tag{
         type: "language",
         subtag: "vmc",
         description: "Juxtlahuaca Mixtec",
         added: "2009-07-29"}).
iana(language,vmd,tag{
         type: "language",
         subtag: "vmd",
         description: "Mudu Koraga",
         added: "2009-07-29"}).
iana(language,vme,tag{
         type: "language",
         subtag: "vme",
         description: "East Masela",
         added: "2009-07-29"}).
iana(language,vmf,tag{
         type: "language",
         subtag: "vmf",
         description: "Mainfränkisch",
         added: "2009-07-29"}).
iana(language,vmg,tag{
         type: "language",
         subtag: "vmg",
         description: "Lungalunga",
         added: "2009-07-29",
         comments: "see also bxf"}).
iana(language,vmh,tag{
         type: "language",
         subtag: "vmh",
         description: "Maraghei",
         added: "2009-07-29"}).
iana(language,vmi,tag{
         type: "language",
         subtag: "vmi",
         description: "Miwa",
         added: "2009-07-29"}).
iana(language,vmj,tag{
         type: "language",
         subtag: "vmj",
         description: "Ixtayutla Mixtec",
         added: "2009-07-29"}).
iana(language,vmk,tag{
         type: "language",
         subtag: "vmk",
         description: "Makhuwa-Shirima",
         added: "2009-07-29"}).
iana(language,vml,tag{
         type: "language",
         subtag: "vml",
         description: "Malgana",
         added: "2009-07-29"}).
iana(language,vmm,tag{
         type: "language",
         subtag: "vmm",
         description: "Mitlatongo Mixtec",
         added: "2009-07-29"}).
iana(language,vmp,tag{
         type: "language",
         subtag: "vmp",
         description: "Soyaltepec Mazatec",
         added: "2009-07-29"}).
iana(language,vmq,tag{
         type: "language",
         subtag: "vmq",
         description: "Soyaltepec Mixtec",
         added: "2009-07-29"}).
iana(language,vmr,tag{
         type: "language",
         subtag: "vmr",
         description: "Marenje",
         added: "2009-07-29"}).
iana(language,vms,tag{
         type: "language",
         subtag: "vms",
         description: "Moksela",
         added: "2009-07-29"}).
iana(language,vmu,tag{
         type: "language",
         subtag: "vmu",
         description: "Muluridyi",
         added: "2009-07-29"}).
iana(language,vmv,tag{
         type: "language",
         subtag: "vmv",
         description: "Valley Maidu",
         added: "2009-07-29"}).
iana(language,vmw,tag{
         type: "language",
         subtag: "vmw",
         description: "Makhuwa",
         added: "2009-07-29"}).
iana(language,vmx,tag{
         type: "language",
         subtag: "vmx",
         description: "Tamazola Mixtec",
         added: "2009-07-29"}).
iana(language,vmy,tag{
         type: "language",
         subtag: "vmy",
         description: "Ayautla Mazatec",
         added: "2009-07-29"}).
iana(language,vmz,tag{
         type: "language",
         subtag: "vmz",
         description: "Mazatlán Mazatec",
         added: "2009-07-29"}).
iana(language,vnk,tag{
         type: "language",
         subtag: "vnk",
         description: ["Vano","Lovono"],
         added: "2009-07-29"}).
iana(language,vnm,tag{
         type: "language",
         subtag: "vnm",
         description: ["Vinmavis","Neve'ei"],
         added: "2009-07-29"}).
iana(language,vnp,tag{
         type: "language",
         subtag: "vnp",
         description: "Vunapu",
         added: "2009-07-29"}).
iana(language,vor,tag{
         type: "language",
         subtag: "vor",
         description: "Voro",
         added: "2009-07-29"}).
iana(language,vot,tag{
         type: "language",
         subtag: "vot",
         description: "Votic",
         added: "2005-10-16"}).
iana(language,vra,tag{
         type: "language",
         subtag: "vra",
         description: "Vera'a",
         added: "2009-07-29"}).
iana(language,vro,tag{
         type: "language",
         subtag: "vro",
         description: "Võro",
         added: "2009-07-29",
         macrolanguage: "et"}).
iana(language,vrs,tag{
         type: "language",
         subtag: "vrs",
         description: "Varisi",
         added: "2009-07-29"}).
iana(language,vrt,tag{
         type: "language",
         subtag: "vrt",
         description: ["Burmbar","Banam Bay"],
         added: "2009-07-29"}).
iana(language,vsi,tag{
         type: "language",
         subtag: "vsi",
         description: "Moldova Sign Language",
         added: "2009-07-29"}).
iana(language,vsl,tag{
         type: "language",
         subtag: "vsl",
         description: "Venezuelan Sign Language",
         added: "2009-07-29"}).
iana(language,vsv,tag{
         type: "language",
         subtag: "vsv",
         description: ["Valencian Sign Language","Llengua de signes valenciana"],
         added: "2009-07-29"}).
iana(language,vto,tag{
         type: "language",
         subtag: "vto",
         description: "Vitou",
         added: "2009-07-29"}).
iana(language,vum,tag{
         type: "language",
         subtag: "vum",
         description: "Vumbu",
         added: "2009-07-29"}).
iana(language,vun,tag{
         type: "language",
         subtag: "vun",
         description: "Vunjo",
         added: "2009-07-29"}).
iana(language,vut,tag{
         type: "language",
         subtag: "vut",
         description: "Vute",
         added: "2009-07-29"}).
iana(language,vwa,tag{
         type: "language",
         subtag: "vwa",
         description: "Awa (China)",
         added: "2009-07-29"}).
iana(language,waa,tag{
         type: "language",
         subtag: "waa",
         description: "Walla Walla",
         added: "2009-07-29"}).
iana(language,wab,tag{
         type: "language",
         subtag: "wab",
         description: "Wab",
         added: "2009-07-29"}).
iana(language,wac,tag{
         type: "language",
         subtag: "wac",
         description: "Wasco-Wishram",
         added: "2009-07-29"}).
iana(language,wad,tag{
         type: "language",
         subtag: "wad",
         description: ["Wamesa","Wondama"],
         added: "2009-07-29"}).
iana(language,wae,tag{
         type: "language",
         subtag: "wae",
         description: "Walser",
         added: "2009-07-29"}).
iana(language,waf,tag{
         type: "language",
         subtag: "waf",
         description: "Wakoná",
         added: "2009-07-29"}).
iana(language,wag,tag{
         type: "language",
         subtag: "wag",
         description: "Wa'ema",
         added: "2009-07-29"}).
iana(language,wah,tag{
         type: "language",
         subtag: "wah",
         description: "Watubela",
         added: "2009-07-29"}).
iana(language,wai,tag{
         type: "language",
         subtag: "wai",
         description: "Wares",
         added: "2009-07-29"}).
iana(language,waj,tag{
         type: "language",
         subtag: "waj",
         description: "Waffa",
         added: "2009-07-29"}).
iana(language,wak,tag{
         type: "language",
         subtag: "wak",
         description: "Wakashan languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,wal,tag{
         type: "language",
         subtag: "wal",
         description: ["Wolaytta","Wolaitta"],
         added: "2005-10-16"}).
iana(language,wam,tag{
         type: "language",
         subtag: "wam",
         description: "Wampanoag",
         added: "2009-07-29"}).
iana(language,wan,tag{
         type: "language",
         subtag: "wan",
         description: "Wan",
         added: "2009-07-29"}).
iana(language,wao,tag{
         type: "language",
         subtag: "wao",
         description: "Wappo",
         added: "2009-07-29"}).
iana(language,wap,tag{
         type: "language",
         subtag: "wap",
         description: "Wapishana",
         added: "2009-07-29"}).
iana(language,waq,tag{
         type: "language",
         subtag: "waq",
         description: "Wagiman",
         added: "2009-07-29"}).
iana(language,war,tag{
         type: "language",
         subtag: "war",
         description: "Waray (Philippines)",
         added: "2005-10-16"}).
iana(language,was,tag{
         type: "language",
         subtag: "was",
         description: "Washo",
         added: "2005-10-16"}).
iana(language,wat,tag{
         type: "language",
         subtag: "wat",
         description: "Kaninuwa",
         added: "2009-07-29"}).
iana(language,wau,tag{
         type: "language",
         subtag: "wau",
         description: "Waurá",
         added: "2009-07-29"}).
iana(language,wav,tag{
         type: "language",
         subtag: "wav",
         description: "Waka",
         added: "2009-07-29"}).
iana(language,waw,tag{
         type: "language",
         subtag: "waw",
         description: "Waiwai",
         added: "2009-07-29"}).
iana(language,wax,tag{
         type: "language",
         subtag: "wax",
         description: ["Watam","Marangis"],
         added: "2009-07-29"}).
iana(language,way,tag{
         type: "language",
         subtag: "way",
         description: "Wayana",
         added: "2009-07-29"}).
iana(language,waz,tag{
         type: "language",
         subtag: "waz",
         description: "Wampur",
         added: "2009-07-29"}).
iana(language,wba,tag{
         type: "language",
         subtag: "wba",
         description: "Warao",
         added: "2009-07-29"}).
iana(language,wbb,tag{
         type: "language",
         subtag: "wbb",
         description: "Wabo",
         added: "2009-07-29"}).
iana(language,wbe,tag{
         type: "language",
         subtag: "wbe",
         description: "Waritai",
         added: "2009-07-29"}).
iana(language,wbf,tag{
         type: "language",
         subtag: "wbf",
         description: "Wara",
         added: "2009-07-29",
         comments: "see also pnl"}).
iana(language,wbh,tag{
         type: "language",
         subtag: "wbh",
         description: "Wanda",
         added: "2009-07-29"}).
iana(language,wbi,tag{
         type: "language",
         subtag: "wbi",
         description: "Vwanji",
         added: "2009-07-29"}).
iana(language,wbj,tag{
         type: "language",
         subtag: "wbj",
         description: "Alagwa",
         added: "2009-07-29"}).
iana(language,wbk,tag{
         type: "language",
         subtag: "wbk",
         description: "Waigali",
         added: "2009-07-29"}).
iana(language,wbl,tag{
         type: "language",
         subtag: "wbl",
         description: "Wakhi",
         added: "2009-07-29"}).
iana(language,wbm,tag{
         type: "language",
         subtag: "wbm",
         description: "Wa",
         added: "2009-07-29"}).
iana(language,wbp,tag{
         type: "language",
         subtag: "wbp",
         description: "Warlpiri",
         added: "2009-07-29"}).
iana(language,wbq,tag{
         type: "language",
         subtag: "wbq",
         description: "Waddar",
         added: "2009-07-29"}).
iana(language,wbr,tag{
         type: "language",
         subtag: "wbr",
         description: "Wagdi",
         added: "2009-07-29",
         macrolanguage: "raj"}).
iana(language,wbs,tag{
         type: "language",
         subtag: "wbs",
         description: "West Bengal Sign Language",
         added: "2017-02-23"}).
iana(language,wbt,tag{
         type: "language",
         subtag: "wbt",
         description: "Warnman",
         added: "2009-07-29"}).
iana(language,wbv,tag{
         type: "language",
         subtag: "wbv",
         description: "Wajarri",
         added: "2009-07-29"}).
iana(language,wbw,tag{
         type: "language",
         subtag: "wbw",
         description: "Woi",
         added: "2009-07-29"}).
iana(language,wca,tag{
         type: "language",
         subtag: "wca",
         description: "Yanomámi",
         added: "2009-07-29"}).
iana(language,wci,tag{
         type: "language",
         subtag: "wci",
         description: "Waci Gbe",
         added: "2009-07-29"}).
iana(language,wdd,tag{
         type: "language",
         subtag: "wdd",
         description: "Wandji",
         added: "2009-07-29"}).
iana(language,wdg,tag{
         type: "language",
         subtag: "wdg",
         description: "Wadaginam",
         added: "2009-07-29"}).
iana(language,wdj,tag{
         type: "language",
         subtag: "wdj",
         description: "Wadjiginy",
         added: "2009-07-29"}).
iana(language,wdk,tag{
         type: "language",
         subtag: "wdk",
         description: "Wadikali",
         added: "2013-09-10"}).
iana(language,wdt,tag{
         type: "language",
         subtag: "wdt",
         description: "Wendat",
         added: "2022-02-25"}).
iana(language,wdu,tag{
         type: "language",
         subtag: "wdu",
         description: "Wadjigu",
         added: "2009-07-29"}).
iana(language,wdy,tag{
         type: "language",
         subtag: "wdy",
         description: "Wadjabangayi",
         added: "2013-09-10"}).
iana(language,wea,tag{
         type: "language",
         subtag: "wea",
         description: "Wewaw",
         added: "2009-07-29"}).
iana(language,wec,tag{
         type: "language",
         subtag: "wec",
         description: "Wè Western",
         added: "2009-07-29"}).
iana(language,wed,tag{
         type: "language",
         subtag: "wed",
         description: "Wedau",
         added: "2009-07-29"}).
iana(language,weg,tag{
         type: "language",
         subtag: "weg",
         description: "Wergaia",
         added: "2013-09-10"}).
iana(language,weh,tag{
         type: "language",
         subtag: "weh",
         description: "Weh",
         added: "2009-07-29"}).
iana(language,wei,tag{
         type: "language",
         subtag: "wei",
         description: "Kiunum",
         added: "2009-07-29"}).
iana(language,wem,tag{
         type: "language",
         subtag: "wem",
         description: "Weme Gbe",
         added: "2009-07-29"}).
iana(language,wen,tag{
         type: "language",
         subtag: "wen",
         description: "Sorbian languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,weo,tag{
         type: "language",
         subtag: "weo",
         description: "Wemale",
         added: "2009-07-29"}).
iana(language,wep,tag{
         type: "language",
         subtag: "wep",
         description: "Westphalien",
         added: "2009-07-29"}).
iana(language,wer,tag{
         type: "language",
         subtag: "wer",
         description: "Weri",
         added: "2009-07-29"}).
iana(language,wes,tag{
         type: "language",
         subtag: "wes",
         description: "Cameroon Pidgin",
         added: "2009-07-29"}).
iana(language,wet,tag{
         type: "language",
         subtag: "wet",
         description: "Perai",
         added: "2009-07-29"}).
iana(language,weu,tag{
         type: "language",
         subtag: "weu",
         description: "Rawngtu Chin",
         added: "2009-07-29"}).
iana(language,wew,tag{
         type: "language",
         subtag: "wew",
         description: "Wejewa",
         added: "2009-07-29"}).
iana(language,wfg,tag{
         type: "language",
         subtag: "wfg",
         description: ["Yafi","Zorop"],
         added: "2009-07-29"}).
iana(language,wga,tag{
         type: "language",
         subtag: "wga",
         description: "Wagaya",
         added: "2009-07-29"}).
iana(language,wgb,tag{
         type: "language",
         subtag: "wgb",
         description: "Wagawaga",
         added: "2010-03-11"}).
iana(language,wgg,tag{
         type: "language",
         subtag: "wgg",
         description: ["Wangkangurru","Wangganguru"],
         added: "2009-07-29"}).
iana(language,wgi,tag{
         type: "language",
         subtag: "wgi",
         description: "Wahgi",
         added: "2009-07-29"}).
iana(language,wgo,tag{
         type: "language",
         subtag: "wgo",
         description: "Waigeo",
         added: "2009-07-29"}).
iana(language,wgu,tag{
         type: "language",
         subtag: "wgu",
         description: "Wirangu",
         added: "2012-08-12"}).
iana(language,wgw,tag{
         type: "language",
         subtag: "wgw",
         description: "Wagawaga",
         added: "2009-07-29",
         deprecated: "2010-03-11",
         comments: "see wgb, ylb"}).
iana(language,wgy,tag{
         type: "language",
         subtag: "wgy",
         description: "Warrgamay",
         added: "2009-07-29"}).
iana(language,wha,tag{
         type: "language",
         subtag: "wha",
         description: ["Sou Upaa","Manusela"],
         added: "2009-07-29"}).
iana(language,whg,tag{
         type: "language",
         subtag: "whg",
         description: "North Wahgi",
         added: "2009-07-29"}).
iana(language,whk,tag{
         type: "language",
         subtag: "whk",
         description: "Wahau Kenyah",
         added: "2009-07-29"}).
iana(language,whu,tag{
         type: "language",
         subtag: "whu",
         description: "Wahau Kayan",
         added: "2009-07-29"}).
iana(language,wib,tag{
         type: "language",
         subtag: "wib",
         description: "Southern Toussian",
         added: "2009-07-29"}).
iana(language,wic,tag{
         type: "language",
         subtag: "wic",
         description: "Wichita",
         added: "2009-07-29"}).
iana(language,wie,tag{
         type: "language",
         subtag: "wie",
         description: "Wik-Epa",
         added: "2009-07-29"}).
iana(language,wif,tag{
         type: "language",
         subtag: "wif",
         description: "Wik-Keyangan",
         added: "2009-07-29"}).
iana(language,wig,tag{
         type: "language",
         subtag: "wig",
         description: "Wik Ngathan",
         added: "2009-07-29"}).
iana(language,wih,tag{
         type: "language",
         subtag: "wih",
         description: "Wik-Me'anha",
         added: "2009-07-29"}).
iana(language,wii,tag{
         type: "language",
         subtag: "wii",
         description: "Minidien",
         added: "2009-07-29"}).
iana(language,wij,tag{
         type: "language",
         subtag: "wij",
         description: "Wik-Iiyanh",
         added: "2009-07-29"}).
iana(language,wik,tag{
         type: "language",
         subtag: "wik",
         description: "Wikalkan",
         added: "2009-07-29"}).
iana(language,wil,tag{
         type: "language",
         subtag: "wil",
         description: "Wilawila",
         added: "2009-07-29"}).
iana(language,wim,tag{
         type: "language",
         subtag: "wim",
         description: "Wik-Mungkan",
         added: "2009-07-29"}).
iana(language,win,tag{
         type: "language",
         subtag: "win",
         description: "Ho-Chunk",
         added: "2009-07-29"}).
iana(language,wir,tag{
         type: "language",
         subtag: "wir",
         description: "Wiraféd",
         added: "2009-07-29"}).
iana(language,wit,tag{
         type: "language",
         subtag: "wit",
         description: "Wintu",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see nol, pwi, wnw"}).
iana(language,wiu,tag{
         type: "language",
         subtag: "wiu",
         description: "Wiru",
         added: "2009-07-29"}).
iana(language,wiv,tag{
         type: "language",
         subtag: "wiv",
         description: "Vitu",
         added: "2009-07-29"}).
iana(language,wiw,tag{
         type: "language",
         subtag: "wiw",
         description: "Wirangu",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see nwo, wgu"}).
iana(language,wiy,tag{
         type: "language",
         subtag: "wiy",
         description: "Wiyot",
         added: "2009-07-29"}).
iana(language,wja,tag{
         type: "language",
         subtag: "wja",
         description: "Waja",
         added: "2009-07-29"}).
iana(language,wji,tag{
         type: "language",
         subtag: "wji",
         description: "Warji",
         added: "2009-07-29"}).
iana(language,wka,tag{
         type: "language",
         subtag: "wka",
         description: "Kw'adza",
         added: "2009-07-29"}).
iana(language,wkb,tag{
         type: "language",
         subtag: "wkb",
         description: "Kumbaran",
         added: "2009-07-29"}).
iana(language,wkd,tag{
         type: "language",
         subtag: "wkd",
         description: ["Wakde","Mo"],
         added: "2009-07-29"}).
iana(language,wkl,tag{
         type: "language",
         subtag: "wkl",
         description: "Kalanadi",
         added: "2009-07-29"}).
iana(language,wkr,tag{
         type: "language",
         subtag: "wkr",
         description: "Keerray-Woorroong",
         added: "2019-04-16"}).
iana(language,wku,tag{
         type: "language",
         subtag: "wku",
         description: "Kunduvadi",
         added: "2009-07-29"}).
iana(language,wkw,tag{
         type: "language",
         subtag: "wkw",
         description: "Wakawaka",
         added: "2009-07-29"}).
iana(language,wky,tag{
         type: "language",
         subtag: "wky",
         description: "Wangkayutyuru",
         added: "2013-09-10"}).
iana(language,wla,tag{
         type: "language",
         subtag: "wla",
         description: "Walio",
         added: "2009-07-29"}).
iana(language,wlc,tag{
         type: "language",
         subtag: "wlc",
         description: "Mwali Comorian",
         added: "2009-07-29"}).
iana(language,wle,tag{
         type: "language",
         subtag: "wle",
         description: "Wolane",
         added: "2009-07-29"}).
iana(language,wlg,tag{
         type: "language",
         subtag: "wlg",
         description: "Kunbarlang",
         added: "2009-07-29"}).
iana(language,wlh,tag{
         type: "language",
         subtag: "wlh",
         description: "Welaun",
         added: "2020-03-28"}).
iana(language,wli,tag{
         type: "language",
         subtag: "wli",
         description: "Waioli",
         added: "2009-07-29"}).
iana(language,wlk,tag{
         type: "language",
         subtag: "wlk",
         description: "Wailaki",
         added: "2009-07-29"}).
iana(language,wll,tag{
         type: "language",
         subtag: "wll",
         description: "Wali (Sudan)",
         added: "2009-07-29"}).
iana(language,wlm,tag{
         type: "language",
         subtag: "wlm",
         description: "Middle Welsh",
         added: "2009-07-29"}).
iana(language,wlo,tag{
         type: "language",
         subtag: "wlo",
         description: "Wolio",
         added: "2009-07-29"}).
iana(language,wlr,tag{
         type: "language",
         subtag: "wlr",
         description: "Wailapa",
         added: "2009-07-29"}).
iana(language,wls,tag{
         type: "language",
         subtag: "wls",
         description: "Wallisian",
         added: "2009-07-29"}).
iana(language,wlu,tag{
         type: "language",
         subtag: "wlu",
         description: "Wuliwuli",
         added: "2009-07-29"}).
iana(language,wlv,tag{
         type: "language",
         subtag: "wlv",
         description: "Wichí Lhamtés Vejoz",
         added: "2009-07-29"}).
iana(language,wlw,tag{
         type: "language",
         subtag: "wlw",
         description: "Walak",
         added: "2009-07-29"}).
iana(language,wlx,tag{
         type: "language",
         subtag: "wlx",
         description: "Wali (Ghana)",
         added: "2009-07-29"}).
iana(language,wly,tag{
         type: "language",
         subtag: "wly",
         description: "Waling",
         added: "2009-07-29"}).
iana(language,wma,tag{
         type: "language",
         subtag: "wma",
         description: "Mawa (Nigeria)",
         added: "2009-07-29"}).
iana(language,wmb,tag{
         type: "language",
         subtag: "wmb",
         description: "Wambaya",
         added: "2009-07-29"}).
iana(language,wmc,tag{
         type: "language",
         subtag: "wmc",
         description: "Wamas",
         added: "2009-07-29"}).
iana(language,wmd,tag{
         type: "language",
         subtag: "wmd",
         description: "Mamaindé",
         added: "2009-07-29"}).
iana(language,wme,tag{
         type: "language",
         subtag: "wme",
         description: "Wambule",
         added: "2009-07-29"}).
iana(language,wmg,tag{
         type: "language",
         subtag: "wmg",
         description: "Western Minyag",
         added: "2021-02-20"}).
iana(language,wmh,tag{
         type: "language",
         subtag: "wmh",
         description: "Waima'a",
         added: "2009-07-29"}).
iana(language,wmi,tag{
         type: "language",
         subtag: "wmi",
         description: "Wamin",
         added: "2009-07-29"}).
iana(language,wmm,tag{
         type: "language",
         subtag: "wmm",
         description: "Maiwa (Indonesia)",
         added: "2009-07-29"}).
iana(language,wmn,tag{
         type: "language",
         subtag: "wmn",
         description: "Waamwang",
         added: "2009-07-29"}).
iana(language,wmo,tag{
         type: "language",
         subtag: "wmo",
         description: "Wom (Papua New Guinea)",
         added: "2009-07-29"}).
iana(language,wms,tag{
         type: "language",
         subtag: "wms",
         description: "Wambon",
         added: "2009-07-29"}).
iana(language,wmt,tag{
         type: "language",
         subtag: "wmt",
         description: "Walmajarri",
         added: "2009-07-29"}).
iana(language,wmw,tag{
         type: "language",
         subtag: "wmw",
         description: "Mwani",
         added: "2009-07-29"}).
iana(language,wmx,tag{
         type: "language",
         subtag: "wmx",
         description: "Womo",
         added: "2009-07-29"}).
iana(language,wnb,tag{
         type: "language",
         subtag: "wnb",
         description: "Wanambre",
         added: "2009-07-29"}).
iana(language,wnc,tag{
         type: "language",
         subtag: "wnc",
         description: "Wantoat",
         added: "2009-07-29"}).
iana(language,wnd,tag{
         type: "language",
         subtag: "wnd",
         description: "Wandarang",
         added: "2009-07-29"}).
iana(language,wne,tag{
         type: "language",
         subtag: "wne",
         description: "Waneci",
         added: "2009-07-29"}).
iana(language,wng,tag{
         type: "language",
         subtag: "wng",
         description: "Wanggom",
         added: "2009-07-29"}).
iana(language,wni,tag{
         type: "language",
         subtag: "wni",
         description: "Ndzwani Comorian",
         added: "2009-07-29"}).
iana(language,wnk,tag{
         type: "language",
         subtag: "wnk",
         description: "Wanukaka",
         added: "2009-07-29"}).
iana(language,wnm,tag{
         type: "language",
         subtag: "wnm",
         description: "Wanggamala",
         added: "2009-07-29"}).
iana(language,wnn,tag{
         type: "language",
         subtag: "wnn",
         description: "Wunumara",
         added: "2013-09-10"}).
iana(language,wno,tag{
         type: "language",
         subtag: "wno",
         description: "Wano",
         added: "2009-07-29"}).
iana(language,wnp,tag{
         type: "language",
         subtag: "wnp",
         description: "Wanap",
         added: "2009-07-29"}).
iana(language,wnu,tag{
         type: "language",
         subtag: "wnu",
         description: "Usan",
         added: "2009-07-29"}).
iana(language,wnw,tag{
         type: "language",
         subtag: "wnw",
         description: "Wintu",
         added: "2013-09-10"}).
iana(language,wny,tag{
         type: "language",
         subtag: "wny",
         description: ["Wanyi","Waanyi"],
         added: "2012-08-12"}).
iana(language,woa,tag{
         type: "language",
         subtag: "woa",
         description: ["Kuwema","Tyaraity"],
         added: "2009-07-29"}).
iana(language,wob,tag{
         type: "language",
         subtag: "wob",
         description: "Wè Northern",
         added: "2009-07-29"}).
iana(language,woc,tag{
         type: "language",
         subtag: "woc",
         description: "Wogeo",
         added: "2009-07-29"}).
iana(language,wod,tag{
         type: "language",
         subtag: "wod",
         description: "Wolani",
         added: "2009-07-29"}).
iana(language,woe,tag{
         type: "language",
         subtag: "woe",
         description: "Woleaian",
         added: "2009-07-29"}).
iana(language,wof,tag{
         type: "language",
         subtag: "wof",
         description: "Gambian Wolof",
         added: "2009-07-29"}).
iana(language,wog,tag{
         type: "language",
         subtag: "wog",
         description: "Wogamusin",
         added: "2009-07-29"}).
iana(language,woi,tag{
         type: "language",
         subtag: "woi",
         description: "Kamang",
         added: "2009-07-29"}).
iana(language,wok,tag{
         type: "language",
         subtag: "wok",
         description: "Longto",
         added: "2009-07-29"}).
iana(language,wom,tag{
         type: "language",
         subtag: "wom",
         description: "Wom (Nigeria)",
         added: "2009-07-29"}).
iana(language,won,tag{
         type: "language",
         subtag: "won",
         description: "Wongo",
         added: "2009-07-29"}).
iana(language,woo,tag{
         type: "language",
         subtag: "woo",
         description: "Manombai",
         added: "2009-07-29"}).
iana(language,wor,tag{
         type: "language",
         subtag: "wor",
         description: "Woria",
         added: "2009-07-29"}).
iana(language,wos,tag{
         type: "language",
         subtag: "wos",
         description: "Hanga Hundi",
         added: "2009-07-29"}).
iana(language,wow,tag{
         type: "language",
         subtag: "wow",
         description: "Wawonii",
         added: "2009-07-29"}).
iana(language,woy,tag{
         type: "language",
         subtag: "woy",
         description: "Weyto",
         added: "2009-07-29"}).
iana(language,wpc,tag{
         type: "language",
         subtag: "wpc",
         description: "Maco",
         added: "2009-07-29"}).
iana(language,wra,tag{
         type: "language",
         subtag: "wra",
         description: "Warapu",
         added: "2009-07-29",
         deprecated: "2021-02-20",
         comments: "see bpe, suo, uni"}).
iana(language,wrb,tag{
         type: "language",
         subtag: "wrb",
         description: ["Waluwarra","Warluwara"],
         added: "2009-07-29"}).
iana(language,wrd,tag{
         type: "language",
         subtag: "wrd",
         description: "Warduji",
         added: "2009-07-29",
         deprecated: "2022-02-25"}).
iana(language,wrg,tag{
         type: "language",
         subtag: "wrg",
         description: ["Warungu","Gudjal"],
         added: "2009-07-29"}).
iana(language,wrh,tag{
         type: "language",
         subtag: "wrh",
         description: "Wiradjuri",
         added: "2009-07-29"}).
iana(language,wri,tag{
         type: "language",
         subtag: "wri",
         description: "Wariyangga",
         added: "2009-07-29"}).
iana(language,wrk,tag{
         type: "language",
         subtag: "wrk",
         description: "Garrwa",
         added: "2012-08-12"}).
iana(language,wrl,tag{
         type: "language",
         subtag: "wrl",
         description: "Warlmanpa",
         added: "2009-07-29"}).
iana(language,wrm,tag{
         type: "language",
         subtag: "wrm",
         description: "Warumungu",
         added: "2009-07-29"}).
iana(language,wrn,tag{
         type: "language",
         subtag: "wrn",
         description: "Warnang",
         added: "2009-07-29"}).
iana(language,wro,tag{
         type: "language",
         subtag: "wro",
         description: "Worrorra",
         added: "2012-08-12"}).
iana(language,wrp,tag{
         type: "language",
         subtag: "wrp",
         description: "Waropen",
         added: "2009-07-29"}).
iana(language,wrr,tag{
         type: "language",
         subtag: "wrr",
         description: "Wardaman",
         added: "2009-07-29"}).
iana(language,wrs,tag{
         type: "language",
         subtag: "wrs",
         description: "Waris",
         added: "2009-07-29"}).
iana(language,wru,tag{
         type: "language",
         subtag: "wru",
         description: "Waru",
         added: "2009-07-29"}).
iana(language,wrv,tag{
         type: "language",
         subtag: "wrv",
         description: "Waruna",
         added: "2009-07-29"}).
iana(language,wrw,tag{
         type: "language",
         subtag: "wrw",
         description: "Gugu Warra",
         added: "2009-07-29"}).
iana(language,wrx,tag{
         type: "language",
         subtag: "wrx",
         description: "Wae Rana",
         added: "2009-07-29"}).
iana(language,wry,tag{
         type: "language",
         subtag: "wry",
         description: "Merwari",
         added: "2009-07-29",
         macrolanguage: "mwr"}).
iana(language,wrz,tag{
         type: "language",
         subtag: "wrz",
         description: "Waray (Australia)",
         added: "2009-07-29"}).
iana(language,wsa,tag{
         type: "language",
         subtag: "wsa",
         description: "Warembori",
         added: "2009-07-29"}).
iana(language,wsg,tag{
         type: "language",
         subtag: "wsg",
         description: "Adilabad Gondi",
         added: "2016-05-30",
         macrolanguage: "gon"}).
iana(language,wsi,tag{
         type: "language",
         subtag: "wsi",
         description: "Wusi",
         added: "2009-07-29"}).
iana(language,wsk,tag{
         type: "language",
         subtag: "wsk",
         description: "Waskia",
         added: "2009-07-29"}).
iana(language,wsr,tag{
         type: "language",
         subtag: "wsr",
         description: "Owenia",
         added: "2009-07-29"}).
iana(language,wss,tag{
         type: "language",
         subtag: "wss",
         description: "Wasa",
         added: "2009-07-29"}).
iana(language,wsu,tag{
         type: "language",
         subtag: "wsu",
         description: "Wasu",
         added: "2009-07-29"}).
iana(language,wsv,tag{
         type: "language",
         subtag: "wsv",
         description: "Wotapuri-Katarqalai",
         added: "2009-07-29"}).
iana(language,wtf,tag{
         type: "language",
         subtag: "wtf",
         description: "Watiwa",
         added: "2009-07-29"}).
iana(language,wth,tag{
         type: "language",
         subtag: "wth",
         description: "Wathawurrung",
         added: "2013-09-10"}).
iana(language,wti,tag{
         type: "language",
         subtag: "wti",
         description: "Berta",
         added: "2009-07-29"}).
iana(language,wtk,tag{
         type: "language",
         subtag: "wtk",
         description: "Watakataui",
         added: "2009-07-29"}).
iana(language,wtm,tag{
         type: "language",
         subtag: "wtm",
         description: "Mewati",
         added: "2009-07-29"}).
iana(language,wtw,tag{
         type: "language",
         subtag: "wtw",
         description: "Wotu",
         added: "2009-07-29"}).
iana(language,wua,tag{
         type: "language",
         subtag: "wua",
         description: "Wikngenchera",
         added: "2009-07-29"}).
iana(language,wub,tag{
         type: "language",
         subtag: "wub",
         description: "Wunambal",
         added: "2009-07-29"}).
iana(language,wud,tag{
         type: "language",
         subtag: "wud",
         description: "Wudu",
         added: "2009-07-29"}).
iana(language,wuh,tag{
         type: "language",
         subtag: "wuh",
         description: "Wutunhua",
         added: "2009-07-29"}).
iana(language,wul,tag{
         type: "language",
         subtag: "wul",
         description: "Silimo",
         added: "2009-07-29"}).
iana(language,wum,tag{
         type: "language",
         subtag: "wum",
         description: "Wumbvu",
         added: "2009-07-29"}).
iana(language,wun,tag{
         type: "language",
         subtag: "wun",
         description: "Bungu",
         added: "2009-07-29"}).
iana(language,wur,tag{
         type: "language",
         subtag: "wur",
         description: "Wurrugu",
         added: "2009-07-29"}).
iana(language,wut,tag{
         type: "language",
         subtag: "wut",
         description: "Wutung",
         added: "2009-07-29"}).
iana(language,wuu,tag{
         type: "language",
         subtag: "wuu",
         description: "Wu Chinese",
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,wuv,tag{
         type: "language",
         subtag: "wuv",
         description: "Wuvulu-Aua",
         added: "2009-07-29"}).
iana(language,wux,tag{
         type: "language",
         subtag: "wux",
         description: "Wulna",
         added: "2009-07-29"}).
iana(language,wuy,tag{
         type: "language",
         subtag: "wuy",
         description: "Wauyai",
         added: "2009-07-29"}).
iana(language,wwa,tag{
         type: "language",
         subtag: "wwa",
         description: "Waama",
         added: "2009-07-29"}).
iana(language,wwb,tag{
         type: "language",
         subtag: "wwb",
         description: "Wakabunga",
         added: "2012-08-12"}).
iana(language,wwo,tag{
         type: "language",
         subtag: "wwo",
         description: ["Wetamut","Dorig"],
         added: "2009-07-29"}).
iana(language,wwr,tag{
         type: "language",
         subtag: "wwr",
         description: "Warrwa",
         added: "2009-07-29"}).
iana(language,www,tag{
         type: "language",
         subtag: "www",
         description: "Wawa",
         added: "2009-07-29"}).
iana(language,wxa,tag{
         type: "language",
         subtag: "wxa",
         description: "Waxianghua",
         added: "2009-07-29"}).
iana(language,wxw,tag{
         type: "language",
         subtag: "wxw",
         description: "Wardandi",
         added: "2013-09-10"}).
iana(language,wya,tag{
         type: "language",
         subtag: "wya",
         description: "Wyandot",
         added: "2009-07-29",
         deprecated: "2022-02-25",
         comments: "see wdt, wyn"}).
iana(language,wyb,tag{
         type: "language",
         subtag: "wyb",
         description: "Wangaaybuwan-Ngiyambaa",
         added: "2009-07-29"}).
iana(language,wyi,tag{
         type: "language",
         subtag: "wyi",
         description: "Woiwurrung",
         added: "2013-09-10"}).
iana(language,wym,tag{
         type: "language",
         subtag: "wym",
         description: "Wymysorys",
         added: "2009-07-29"}).
iana(language,wyn,tag{
         type: "language",
         subtag: "wyn",
         description: "Wyandot",
         added: "2022-02-25"}).
iana(language,wyr,tag{
         type: "language",
         subtag: "wyr",
         description: "Wayoró",
         added: "2009-07-29"}).
iana(language,wyy,tag{
         type: "language",
         subtag: "wyy",
         description: "Western Fijian",
         added: "2009-07-29"}).
iana(language,xaa,tag{
         type: "language",
         subtag: "xaa",
         description: "Andalusian Arabic",
         added: "2009-07-29"}).
iana(language,xab,tag{
         type: "language",
         subtag: "xab",
         description: "Sambe",
         added: "2009-07-29"}).
iana(language,xac,tag{
         type: "language",
         subtag: "xac",
         description: "Kachari",
         added: "2009-07-29"}).
iana(language,xad,tag{
         type: "language",
         subtag: "xad",
         description: "Adai",
         added: "2009-07-29"}).
iana(language,xae,tag{
         type: "language",
         subtag: "xae",
         description: "Aequian",
         added: "2009-07-29"}).
iana(language,xag,tag{
         type: "language",
         subtag: "xag",
         description: "Aghwan",
         added: "2009-07-29"}).
iana(language,xai,tag{
         type: "language",
         subtag: "xai",
         description: "Kaimbé",
         added: "2009-07-29"}).
iana(language,xaj,tag{
         type: "language",
         subtag: "xaj",
         description: "Ararandewára",
         added: "2014-02-28"}).
iana(language,xak,tag{
         type: "language",
         subtag: "xak",
         description: "Máku",
         added: "2016-05-30"}).
iana(language,xal,tag{
         type: "language",
         subtag: "xal",
         description: ["Kalmyk","Oirat"],
         added: "2005-10-16"}).
iana(language,xam,tag{
         type: "language",
         subtag: "xam",
         description: "ǀXam",
         added: "2009-07-29"}).
iana(language,xan,tag{
         type: "language",
         subtag: "xan",
         description: "Xamtanga",
         added: "2009-07-29"}).
iana(language,xao,tag{
         type: "language",
         subtag: "xao",
         description: "Khao",
         added: "2009-07-29"}).
iana(language,xap,tag{
         type: "language",
         subtag: "xap",
         description: "Apalachee",
         added: "2009-07-29"}).
iana(language,xaq,tag{
         type: "language",
         subtag: "xaq",
         description: "Aquitanian",
         added: "2009-07-29"}).
iana(language,xar,tag{
         type: "language",
         subtag: "xar",
         description: "Karami",
         added: "2009-07-29"}).
iana(language,xas,tag{
         type: "language",
         subtag: "xas",
         description: "Kamas",
         added: "2009-07-29"}).
iana(language,xat,tag{
         type: "language",
         subtag: "xat",
         description: "Katawixi",
         added: "2009-07-29"}).
iana(language,xau,tag{
         type: "language",
         subtag: "xau",
         description: "Kauwera",
         added: "2009-07-29"}).
iana(language,xav,tag{
         type: "language",
         subtag: "xav",
         description: "Xavánte",
         added: "2009-07-29"}).
iana(language,xaw,tag{
         type: "language",
         subtag: "xaw",
         description: "Kawaiisu",
         added: "2009-07-29"}).
iana(language,xay,tag{
         type: "language",
         subtag: "xay",
         description: "Kayan Mahakam",
         added: "2009-07-29"}).
iana(language,xba,tag{
         type: "language",
         subtag: "xba",
         description: "Kamba (Brazil)",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "cax"}).
iana(language,xbb,tag{
         type: "language",
         subtag: "xbb",
         description: "Lower Burdekin",
         added: "2010-03-11"}).
iana(language,xbc,tag{
         type: "language",
         subtag: "xbc",
         description: "Bactrian",
         added: "2009-07-29"}).
iana(language,xbd,tag{
         type: "language",
         subtag: "xbd",
         description: "Bindal",
         added: "2013-09-10"}).
iana(language,xbe,tag{
         type: "language",
         subtag: "xbe",
         description: "Bigambal",
         added: "2013-09-10"}).
iana(language,xbg,tag{
         type: "language",
         subtag: "xbg",
         description: "Bunganditj",
         added: "2013-09-10"}).
iana(language,xbi,tag{
         type: "language",
         subtag: "xbi",
         description: "Kombio",
         added: "2009-07-29"}).
iana(language,xbj,tag{
         type: "language",
         subtag: "xbj",
         description: "Birrpayi",
         added: "2013-09-10"}).
iana(language,xbm,tag{
         type: "language",
         subtag: "xbm",
         description: "Middle Breton",
         added: "2009-07-29"}).
iana(language,xbn,tag{
         type: "language",
         subtag: "xbn",
         description: "Kenaboi",
         added: "2010-03-11"}).
iana(language,xbo,tag{
         type: "language",
         subtag: "xbo",
         description: "Bolgarian",
         added: "2009-07-29"}).
iana(language,xbp,tag{
         type: "language",
         subtag: "xbp",
         description: "Bibbulman",
         added: "2013-09-10"}).
iana(language,xbr,tag{
         type: "language",
         subtag: "xbr",
         description: "Kambera",
         added: "2009-07-29"}).
iana(language,xbw,tag{
         type: "language",
         subtag: "xbw",
         description: "Kambiwá",
         added: "2009-07-29"}).
iana(language,xbx,tag{
         type: "language",
         subtag: "xbx",
         description: "Kabixí",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,xby,tag{
         type: "language",
         subtag: "xby",
         description: ["Batjala","Batyala"],
         added: "2013-09-10"}).
iana(language,xcb,tag{
         type: "language",
         subtag: "xcb",
         description: "Cumbric",
         added: "2009-07-29"}).
iana(language,xcc,tag{
         type: "language",
         subtag: "xcc",
         description: "Camunic",
         added: "2009-07-29"}).
iana(language,xce,tag{
         type: "language",
         subtag: "xce",
         description: "Celtiberian",
         added: "2009-07-29"}).
iana(language,xcg,tag{
         type: "language",
         subtag: "xcg",
         description: "Cisalpine Gaulish",
         added: "2009-07-29"}).
iana(language,xch,tag{
         type: "language",
         subtag: "xch",
         description: ["Chemakum","Chimakum"],
         added: "2009-07-29"}).
iana(language,xcl,tag{
         type: "language",
         subtag: "xcl",
         description: "Classical Armenian",
         added: "2009-07-29"}).
iana(language,xcm,tag{
         type: "language",
         subtag: "xcm",
         description: "Comecrudo",
         added: "2009-07-29"}).
iana(language,xcn,tag{
         type: "language",
         subtag: "xcn",
         description: "Cotoname",
         added: "2009-07-29"}).
iana(language,xco,tag{
         type: "language",
         subtag: "xco",
         description: "Chorasmian",
         added: "2009-07-29"}).
iana(language,xcr,tag{
         type: "language",
         subtag: "xcr",
         description: "Carian",
         added: "2009-07-29"}).
iana(language,xct,tag{
         type: "language",
         subtag: "xct",
         description: "Classical Tibetan",
         added: "2009-07-29"}).
iana(language,xcu,tag{
         type: "language",
         subtag: "xcu",
         description: "Curonian",
         added: "2009-07-29"}).
iana(language,xcv,tag{
         type: "language",
         subtag: "xcv",
         description: "Chuvantsy",
         added: "2009-07-29"}).
iana(language,xcw,tag{
         type: "language",
         subtag: "xcw",
         description: "Coahuilteco",
         added: "2009-07-29"}).
iana(language,xcy,tag{
         type: "language",
         subtag: "xcy",
         description: "Cayuse",
         added: "2009-07-29"}).
iana(language,xda,tag{
         type: "language",
         subtag: "xda",
         description: "Darkinyung",
         added: "2013-09-10"}).
iana(language,xdc,tag{
         type: "language",
         subtag: "xdc",
         description: "Dacian",
         added: "2009-07-29"}).
iana(language,xdk,tag{
         type: "language",
         subtag: "xdk",
         description: "Dharuk",
         added: "2013-09-10"}).
iana(language,xdm,tag{
         type: "language",
         subtag: "xdm",
         description: "Edomite",
         added: "2009-07-29"}).
iana(language,xdo,tag{
         type: "language",
         subtag: "xdo",
         description: "Kwandu",
         added: "2017-02-23"}).
iana(language,xdq,tag{
         type: "language",
         subtag: "xdq",
         description: "Kaitag",
         added: "2022-02-25"}).
iana(language,xdy,tag{
         type: "language",
         subtag: "xdy",
         description: "Malayic Dayak",
         added: "2009-07-29"}).
iana(language,xeb,tag{
         type: "language",
         subtag: "xeb",
         description: "Eblan",
         added: "2009-07-29"}).
iana(language,xed,tag{
         type: "language",
         subtag: "xed",
         description: "Hdi",
         added: "2009-07-29"}).
iana(language,xeg,tag{
         type: "language",
         subtag: "xeg",
         description: "ǁXegwi",
         added: "2009-07-29"}).
iana(language,xel,tag{
         type: "language",
         subtag: "xel",
         description: "Kelo",
         added: "2009-07-29"}).
iana(language,xem,tag{
         type: "language",
         subtag: "xem",
         description: "Kembayan",
         added: "2009-07-29"}).
iana(language,xep,tag{
         type: "language",
         subtag: "xep",
         description: "Epi-Olmec",
         added: "2009-07-29"}).
iana(language,xer,tag{
         type: "language",
         subtag: "xer",
         description: "Xerénte",
         added: "2009-07-29"}).
iana(language,xes,tag{
         type: "language",
         subtag: "xes",
         description: "Kesawai",
         added: "2009-07-29"}).
iana(language,xet,tag{
         type: "language",
         subtag: "xet",
         description: "Xetá",
         added: "2009-07-29"}).
iana(language,xeu,tag{
         type: "language",
         subtag: "xeu",
         description: "Keoru-Ahia",
         added: "2009-07-29"}).
iana(language,xfa,tag{
         type: "language",
         subtag: "xfa",
         description: "Faliscan",
         added: "2009-07-29"}).
iana(language,xga,tag{
         type: "language",
         subtag: "xga",
         description: "Galatian",
         added: "2009-07-29"}).
iana(language,xgb,tag{
         type: "language",
         subtag: "xgb",
         description: "Gbin",
         added: "2012-08-12"}).
iana(language,xgd,tag{
         type: "language",
         subtag: "xgd",
         description: "Gudang",
         added: "2013-09-10"}).
iana(language,xgf,tag{
         type: "language",
         subtag: "xgf",
         description: "Gabrielino-Fernandeño",
         added: "2009-07-29"}).
iana(language,xgg,tag{
         type: "language",
         subtag: "xgg",
         description: "Goreng",
         added: "2013-09-10"}).
iana(language,xgi,tag{
         type: "language",
         subtag: "xgi",
         description: "Garingbal",
         added: "2013-09-10"}).
iana(language,xgl,tag{
         type: "language",
         subtag: "xgl",
         description: "Galindan",
         added: "2009-07-29"}).
iana(language,xgm,tag{
         type: "language",
         subtag: "xgm",
         description: ["Dharumbal","Guwinmal"],
         added: "2013-09-10"}).
iana(language,xgn,tag{
         type: "language",
         subtag: "xgn",
         description: "Mongolian languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,xgr,tag{
         type: "language",
         subtag: "xgr",
         description: "Garza",
         added: "2009-07-29"}).
iana(language,xgu,tag{
         type: "language",
         subtag: "xgu",
         description: "Unggumi",
         added: "2012-08-12"}).
iana(language,xgw,tag{
         type: "language",
         subtag: "xgw",
         description: "Guwa",
         added: "2013-09-10"}).
iana(language,xha,tag{
         type: "language",
         subtag: "xha",
         description: "Harami",
         added: "2009-07-29"}).
iana(language,xhc,tag{
         type: "language",
         subtag: "xhc",
         description: "Hunnic",
         added: "2009-07-29"}).
iana(language,xhd,tag{
         type: "language",
         subtag: "xhd",
         description: "Hadrami",
         added: "2009-07-29"}).
iana(language,xhe,tag{
         type: "language",
         subtag: "xhe",
         description: "Khetrani",
         added: "2009-07-29",
         macrolanguage: "lah"}).
iana(language,xhm,tag{
         type: "language",
         subtag: "xhm",
         description: "Middle Khmer (1400 to 1850 CE)",
         added: "2022-02-25"}).
iana(language,xhr,tag{
         type: "language",
         subtag: "xhr",
         description: "Hernican",
         added: "2009-07-29"}).
iana(language,xht,tag{
         type: "language",
         subtag: "xht",
         description: "Hattic",
         added: "2009-07-29"}).
iana(language,xhu,tag{
         type: "language",
         subtag: "xhu",
         description: "Hurrian",
         added: "2009-07-29"}).
iana(language,xhv,tag{
         type: "language",
         subtag: "xhv",
         description: "Khua",
         added: "2009-07-29"}).
iana(language,xia,tag{
         type: "language",
         subtag: "xia",
         description: "Xiandao",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "acn"}).
iana(language,xib,tag{
         type: "language",
         subtag: "xib",
         description: "Iberian",
         added: "2009-07-29"}).
iana(language,xii,tag{
         type: "language",
         subtag: "xii",
         description: "Xiri",
         added: "2009-07-29"}).
iana(language,xil,tag{
         type: "language",
         subtag: "xil",
         description: "Illyrian",
         added: "2009-07-29"}).
iana(language,xin,tag{
         type: "language",
         subtag: "xin",
         description: "Xinca",
         added: "2009-07-29"}).
iana(language,xip,tag{
         type: "language",
         subtag: "xip",
         description: "Xipináwa",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,xir,tag{
         type: "language",
         subtag: "xir",
         description: "Xiriâna",
         added: "2009-07-29"}).
iana(language,xis,tag{
         type: "language",
         subtag: "xis",
         description: "Kisan",
         added: "2014-02-28"}).
iana(language,xiv,tag{
         type: "language",
         subtag: "xiv",
         description: "Indus Valley Language",
         added: "2009-07-29"}).
iana(language,xiy,tag{
         type: "language",
         subtag: "xiy",
         description: "Xipaya",
         added: "2009-07-29"}).
iana(language,xjb,tag{
         type: "language",
         subtag: "xjb",
         description: "Minjungbal",
         added: "2013-09-10"}).
iana(language,xjt,tag{
         type: "language",
         subtag: "xjt",
         description: "Jaitmatang",
         added: "2013-09-10"}).
iana(language,xka,tag{
         type: "language",
         subtag: "xka",
         description: "Kalkoti",
         added: "2009-07-29"}).
iana(language,xkb,tag{
         type: "language",
         subtag: "xkb",
         description: "Northern Nago",
         added: "2009-07-29"}).
iana(language,xkc,tag{
         type: "language",
         subtag: "xkc",
         description: "Kho'ini",
         added: "2009-07-29"}).
iana(language,xkd,tag{
         type: "language",
         subtag: "xkd",
         description: "Mendalam Kayan",
         added: "2009-07-29"}).
iana(language,xke,tag{
         type: "language",
         subtag: "xke",
         description: "Kereho",
         added: "2009-07-29"}).
iana(language,xkf,tag{
         type: "language",
         subtag: "xkf",
         description: "Khengkha",
         added: "2009-07-29"}).
iana(language,xkg,tag{
         type: "language",
         subtag: "xkg",
         description: "Kagoro",
         added: "2009-07-29"}).
iana(language,xkh,tag{
         type: "language",
         subtag: "xkh",
         description: "Karahawyana",
         added: "2009-07-29",
         deprecated: "2016-05-30",
         preferred_value: "waw"}).
iana(language,xki,tag{
         type: "language",
         subtag: "xki",
         description: "Kenyan Sign Language",
         added: "2009-07-29"}).
iana(language,xkj,tag{
         type: "language",
         subtag: "xkj",
         description: "Kajali",
         added: "2009-07-29"}).
iana(language,xkk,tag{
         type: "language",
         subtag: "xkk",
         description: ["Kachok","Kaco'"],
         added: "2009-07-29"}).
iana(language,xkl,tag{
         type: "language",
         subtag: "xkl",
         description: "Mainstream Kenyah",
         added: "2009-07-29"}).
iana(language,xkn,tag{
         type: "language",
         subtag: "xkn",
         description: "Kayan River Kayan",
         added: "2009-07-29"}).
iana(language,xko,tag{
         type: "language",
         subtag: "xko",
         description: "Kiorr",
         added: "2009-07-29"}).
iana(language,xkp,tag{
         type: "language",
         subtag: "xkp",
         description: "Kabatei",
         added: "2009-07-29"}).
iana(language,xkq,tag{
         type: "language",
         subtag: "xkq",
         description: "Koroni",
         added: "2009-07-29"}).
iana(language,xkr,tag{
         type: "language",
         subtag: "xkr",
         description: "Xakriabá",
         added: "2009-07-29"}).
iana(language,xks,tag{
         type: "language",
         subtag: "xks",
         description: "Kumbewaha",
         added: "2009-07-29"}).
iana(language,xkt,tag{
         type: "language",
         subtag: "xkt",
         description: "Kantosi",
         added: "2009-07-29"}).
iana(language,xku,tag{
         type: "language",
         subtag: "xku",
         description: "Kaamba",
         added: "2009-07-29"}).
iana(language,xkv,tag{
         type: "language",
         subtag: "xkv",
         description: "Kgalagadi",
         added: "2009-07-29"}).
iana(language,xkw,tag{
         type: "language",
         subtag: "xkw",
         description: "Kembra",
         added: "2009-07-29"}).
iana(language,xkx,tag{
         type: "language",
         subtag: "xkx",
         description: "Karore",
         added: "2009-07-29"}).
iana(language,xky,tag{
         type: "language",
         subtag: "xky",
         description: "Uma' Lasan",
         added: "2009-07-29"}).
iana(language,xkz,tag{
         type: "language",
         subtag: "xkz",
         description: "Kurtokha",
         added: "2009-07-29"}).
iana(language,xla,tag{
         type: "language",
         subtag: "xla",
         description: "Kamula",
         added: "2009-07-29"}).
iana(language,xlb,tag{
         type: "language",
         subtag: "xlb",
         description: "Loup B",
         added: "2009-07-29"}).
iana(language,xlc,tag{
         type: "language",
         subtag: "xlc",
         description: "Lycian",
         added: "2009-07-29"}).
iana(language,xld,tag{
         type: "language",
         subtag: "xld",
         description: "Lydian",
         added: "2009-07-29"}).
iana(language,xle,tag{
         type: "language",
         subtag: "xle",
         description: "Lemnian",
         added: "2009-07-29"}).
iana(language,xlg,tag{
         type: "language",
         subtag: "xlg",
         description: "Ligurian (Ancient)",
         added: "2009-07-29"}).
iana(language,xli,tag{
         type: "language",
         subtag: "xli",
         description: "Liburnian",
         added: "2009-07-29"}).
iana(language,xln,tag{
         type: "language",
         subtag: "xln",
         description: "Alanic",
         added: "2009-07-29"}).
iana(language,xlo,tag{
         type: "language",
         subtag: "xlo",
         description: "Loup A",
         added: "2009-07-29"}).
iana(language,xlp,tag{
         type: "language",
         subtag: "xlp",
         description: "Lepontic",
         added: "2009-07-29"}).
iana(language,xls,tag{
         type: "language",
         subtag: "xls",
         description: "Lusitanian",
         added: "2009-07-29"}).
iana(language,xlu,tag{
         type: "language",
         subtag: "xlu",
         description: "Cuneiform Luwian",
         added: "2009-07-29"}).
iana(language,xly,tag{
         type: "language",
         subtag: "xly",
         description: "Elymian",
         added: "2009-07-29"}).
iana(language,xma,tag{
         type: "language",
         subtag: "xma",
         description: "Mushungulu",
         added: "2009-07-29"}).
iana(language,xmb,tag{
         type: "language",
         subtag: "xmb",
         description: "Mbonga",
         added: "2009-07-29"}).
iana(language,xmc,tag{
         type: "language",
         subtag: "xmc",
         description: "Makhuwa-Marrevone",
         added: "2009-07-29"}).
iana(language,xmd,tag{
         type: "language",
         subtag: "xmd",
         description: "Mbudum",
         added: "2009-07-29"}).
iana(language,xme,tag{
         type: "language",
         subtag: "xme",
         description: "Median",
         added: "2009-07-29"}).
iana(language,xmf,tag{
         type: "language",
         subtag: "xmf",
         description: "Mingrelian",
         added: "2009-07-29"}).
iana(language,xmg,tag{
         type: "language",
         subtag: "xmg",
         description: "Mengaka",
         added: "2009-07-29"}).
iana(language,xmh,tag{
         type: "language",
         subtag: "xmh",
         description: "Kugu-Muminh",
         added: "2009-07-29"}).
iana(language,xmj,tag{
         type: "language",
         subtag: "xmj",
         description: "Majera",
         added: "2009-07-29"}).
iana(language,xmk,tag{
         type: "language",
         subtag: "xmk",
         description: "Ancient Macedonian",
         added: "2009-07-29"}).
iana(language,xml,tag{
         type: "language",
         subtag: "xml",
         description: "Malaysian Sign Language",
         added: "2009-07-29"}).
iana(language,xmm,tag{
         type: "language",
         subtag: "xmm",
         description: "Manado Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,xmn,tag{
         type: "language",
         subtag: "xmn",
         description: "Manichaean Middle Persian",
         added: "2009-07-29"}).
iana(language,xmo,tag{
         type: "language",
         subtag: "xmo",
         description: "Morerebi",
         added: "2009-07-29"}).
iana(language,xmp,tag{
         type: "language",
         subtag: "xmp",
         description: "Kuku-Mu'inh",
         added: "2009-07-29"}).
iana(language,xmq,tag{
         type: "language",
         subtag: "xmq",
         description: "Kuku-Mangk",
         added: "2009-07-29"}).
iana(language,xmr,tag{
         type: "language",
         subtag: "xmr",
         description: "Meroitic",
         added: "2009-07-29"}).
iana(language,xms,tag{
         type: "language",
         subtag: "xms",
         description: "Moroccan Sign Language",
         added: "2009-07-29"}).
iana(language,xmt,tag{
         type: "language",
         subtag: "xmt",
         description: "Matbat",
         added: "2009-07-29"}).
iana(language,xmu,tag{
         type: "language",
         subtag: "xmu",
         description: "Kamu",
         added: "2009-07-29"}).
iana(language,xmv,tag{
         type: "language",
         subtag: "xmv",
         description: ["Antankarana Malagasy","Tankarana Malagasy"],
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,xmw,tag{
         type: "language",
         subtag: "xmw",
         description: "Tsimihety Malagasy",
         added: "2009-07-29",
         macrolanguage: "mg"}).
iana(language,xmx,tag{
         type: "language",
         subtag: "xmx",
         description: ["Salawati","Maden"],
         added: "2009-07-29"}).
iana(language,xmy,tag{
         type: "language",
         subtag: "xmy",
         description: "Mayaguduna",
         added: "2009-07-29"}).
iana(language,xmz,tag{
         type: "language",
         subtag: "xmz",
         description: "Mori Bawah",
         added: "2009-07-29"}).
iana(language,xna,tag{
         type: "language",
         subtag: "xna",
         description: "Ancient North Arabian",
         added: "2009-07-29"}).
iana(language,xnb,tag{
         type: "language",
         subtag: "xnb",
         description: "Kanakanabu",
         added: "2009-07-29"}).
iana(language,xnd,tag{
         type: "language",
         subtag: "xnd",
         description: "Na-Dene languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,xng,tag{
         type: "language",
         subtag: "xng",
         description: "Middle Mongolian",
         added: "2009-07-29"}).
iana(language,xnh,tag{
         type: "language",
         subtag: "xnh",
         description: "Kuanhua",
         added: "2009-07-29"}).
iana(language,xni,tag{
         type: "language",
         subtag: "xni",
         description: "Ngarigu",
         added: "2013-09-10"}).
iana(language,xnj,tag{
         type: "language",
         subtag: "xnj",
         description: "Ngoni (Tanzania)",
         added: "2021-02-20"}).
iana(language,xnk,tag{
         type: "language",
         subtag: "xnk",
         description: "Nganakarti",
         added: "2013-09-10"}).
iana(language,xnm,tag{
         type: "language",
         subtag: "xnm",
         description: "Ngumbarl",
         added: "2020-03-28"}).
iana(language,xnn,tag{
         type: "language",
         subtag: "xnn",
         description: "Northern Kankanay",
         added: "2009-07-29"}).
iana(language,xno,tag{
         type: "language",
         subtag: "xno",
         description: "Anglo-Norman",
         added: "2009-07-29"}).
iana(language,xnq,tag{
         type: "language",
         subtag: "xnq",
         description: "Ngoni (Mozambique)",
         added: "2021-02-20"}).
iana(language,xnr,tag{
         type: "language",
         subtag: "xnr",
         description: "Kangri",
         added: "2009-07-29",
         macrolanguage: "doi"}).
iana(language,xns,tag{
         type: "language",
         subtag: "xns",
         description: "Kanashi",
         added: "2009-07-29"}).
iana(language,xnt,tag{
         type: "language",
         subtag: "xnt",
         description: "Narragansett",
         added: "2010-03-11"}).
iana(language,xnu,tag{
         type: "language",
         subtag: "xnu",
         description: "Nukunul",
         added: "2013-09-10"}).
iana(language,xny,tag{
         type: "language",
         subtag: "xny",
         description: "Nyiyaparli",
         added: "2013-09-10"}).
iana(language,xnz,tag{
         type: "language",
         subtag: "xnz",
         description: ["Kenzi","Mattoki"],
         added: "2012-08-12"}).
iana(language,xoc,tag{
         type: "language",
         subtag: "xoc",
         description: "O'chi'chi'",
         added: "2009-07-29"}).
iana(language,xod,tag{
         type: "language",
         subtag: "xod",
         description: "Kokoda",
         added: "2009-07-29"}).
iana(language,xog,tag{
         type: "language",
         subtag: "xog",
         description: "Soga",
         added: "2009-07-29"}).
iana(language,xoi,tag{
         type: "language",
         subtag: "xoi",
         description: "Kominimung",
         added: "2009-07-29"}).
iana(language,xok,tag{
         type: "language",
         subtag: "xok",
         description: "Xokleng",
         added: "2009-07-29"}).
iana(language,xom,tag{
         type: "language",
         subtag: "xom",
         description: "Komo (Sudan)",
         added: "2009-07-29"}).
iana(language,xon,tag{
         type: "language",
         subtag: "xon",
         description: "Konkomba",
         added: "2009-07-29"}).
iana(language,xoo,tag{
         type: "language",
         subtag: "xoo",
         description: "Xukurú",
         added: "2009-07-29"}).
iana(language,xop,tag{
         type: "language",
         subtag: "xop",
         description: "Kopar",
         added: "2009-07-29"}).
iana(language,xor,tag{
         type: "language",
         subtag: "xor",
         description: "Korubo",
         added: "2009-07-29"}).
iana(language,xow,tag{
         type: "language",
         subtag: "xow",
         description: "Kowaki",
         added: "2009-07-29"}).
iana(language,xpa,tag{
         type: "language",
         subtag: "xpa",
         description: "Pirriya",
         added: "2013-09-10"}).
iana(language,xpb,tag{
         type: "language",
         subtag: "xpb",
         description: ["Northeastern Tasmanian","Pyemmairrener"],
         added: "2020-03-28"}).
iana(language,xpc,tag{
         type: "language",
         subtag: "xpc",
         description: "Pecheneg",
         added: "2009-07-29"}).
iana(language,xpd,tag{
         type: "language",
         subtag: "xpd",
         description: "Oyster Bay Tasmanian",
         added: "2020-03-28"}).
iana(language,xpe,tag{
         type: "language",
         subtag: "xpe",
         description: "Liberia Kpelle",
         added: "2009-07-29",
         macrolanguage: "kpe"}).
iana(language,xpf,tag{
         type: "language",
         subtag: "xpf",
         description: ["Southeast Tasmanian","Nuenonne"],
         added: "2020-03-28"}).
iana(language,xpg,tag{
         type: "language",
         subtag: "xpg",
         description: "Phrygian",
         added: "2009-07-29"}).
iana(language,xph,tag{
         type: "language",
         subtag: "xph",
         description: ["North Midlands Tasmanian","Tyerrenoterpanner"],
         added: "2020-03-28"}).
iana(language,xpi,tag{
         type: "language",
         subtag: "xpi",
         description: "Pictish",
         added: "2009-07-29"}).
iana(language,xpj,tag{
         type: "language",
         subtag: "xpj",
         description: "Mpalitjanh",
         added: "2012-08-12"}).
iana(language,xpk,tag{
         type: "language",
         subtag: "xpk",
         description: "Kulina Pano",
         added: "2009-07-29"}).
iana(language,xpl,tag{
         type: "language",
         subtag: "xpl",
         description: "Port Sorell Tasmanian",
         added: "2020-03-28"}).
iana(language,xpm,tag{
         type: "language",
         subtag: "xpm",
         description: "Pumpokol",
         added: "2009-07-29"}).
iana(language,xpn,tag{
         type: "language",
         subtag: "xpn",
         description: "Kapinawá",
         added: "2009-07-29"}).
iana(language,xpo,tag{
         type: "language",
         subtag: "xpo",
         description: "Pochutec",
         added: "2009-07-29"}).
iana(language,xpp,tag{
         type: "language",
         subtag: "xpp",
         description: "Puyo-Paekche",
         added: "2009-07-29"}).
iana(language,xpq,tag{
         type: "language",
         subtag: "xpq",
         description: "Mohegan-Pequot",
         added: "2010-03-11"}).
iana(language,xpr,tag{
         type: "language",
         subtag: "xpr",
         description: "Parthian",
         added: "2009-07-29"}).
iana(language,xps,tag{
         type: "language",
         subtag: "xps",
         description: "Pisidian",
         added: "2009-07-29"}).
iana(language,xpt,tag{
         type: "language",
         subtag: "xpt",
         description: "Punthamara",
         added: "2013-09-10"}).
iana(language,xpu,tag{
         type: "language",
         subtag: "xpu",
         description: "Punic",
         added: "2009-07-29"}).
iana(language,xpv,tag{
         type: "language",
         subtag: "xpv",
         description: ["Northern Tasmanian","Tommeginne"],
         added: "2020-03-28"}).
iana(language,xpw,tag{
         type: "language",
         subtag: "xpw",
         description: ["Northwestern Tasmanian","Peerapper"],
         added: "2020-03-28"}).
iana(language,xpx,tag{
         type: "language",
         subtag: "xpx",
         description: ["Southwestern Tasmanian","Toogee"],
         added: "2020-03-28"}).
iana(language,xpy,tag{
         type: "language",
         subtag: "xpy",
         description: "Puyo",
         added: "2009-07-29"}).
iana(language,xpz,tag{
         type: "language",
         subtag: "xpz",
         description: "Bruny Island Tasmanian",
         added: "2020-03-28"}).
iana(language,xqa,tag{
         type: "language",
         subtag: "xqa",
         description: "Karakhanid",
         added: "2009-07-29"}).
iana(language,xqt,tag{
         type: "language",
         subtag: "xqt",
         description: "Qatabanian",
         added: "2009-07-29"}).
iana(language,xra,tag{
         type: "language",
         subtag: "xra",
         description: "Krahô",
         added: "2009-07-29"}).
iana(language,xrb,tag{
         type: "language",
         subtag: "xrb",
         description: "Eastern Karaboro",
         added: "2009-07-29"}).
iana(language,xrd,tag{
         type: "language",
         subtag: "xrd",
         description: "Gundungurra",
         added: "2013-09-10"}).
iana(language,xre,tag{
         type: "language",
         subtag: "xre",
         description: "Kreye",
         added: "2009-07-29"}).
iana(language,xrg,tag{
         type: "language",
         subtag: "xrg",
         description: "Minang",
         added: "2013-09-10"}).
iana(language,xri,tag{
         type: "language",
         subtag: "xri",
         description: "Krikati-Timbira",
         added: "2009-07-29"}).
iana(language,xrm,tag{
         type: "language",
         subtag: "xrm",
         description: "Armazic",
         added: "2009-07-29"}).
iana(language,xrn,tag{
         type: "language",
         subtag: "xrn",
         description: "Arin",
         added: "2009-07-29"}).
iana(language,xrq,tag{
         type: "language",
         subtag: "xrq",
         description: "Karranga",
         added: "2013-09-10",
         deprecated: "2020-03-28",
         preferred_value: "dmw"}).
iana(language,xrr,tag{
         type: "language",
         subtag: "xrr",
         description: "Raetic",
         added: "2009-07-29"}).
iana(language,xrt,tag{
         type: "language",
         subtag: "xrt",
         description: "Aranama-Tamique",
         added: "2009-07-29"}).
iana(language,xru,tag{
         type: "language",
         subtag: "xru",
         description: "Marriammu",
         added: "2009-07-29"}).
iana(language,xrw,tag{
         type: "language",
         subtag: "xrw",
         description: "Karawa",
         added: "2009-07-29"}).
iana(language,xsa,tag{
         type: "language",
         subtag: "xsa",
         description: "Sabaean",
         added: "2009-07-29"}).
iana(language,xsb,tag{
         type: "language",
         subtag: "xsb",
         description: "Sambal",
         added: "2009-07-29"}).
iana(language,xsc,tag{
         type: "language",
         subtag: "xsc",
         description: "Scythian",
         added: "2009-07-29"}).
iana(language,xsd,tag{
         type: "language",
         subtag: "xsd",
         description: "Sidetic",
         added: "2009-07-29"}).
iana(language,xse,tag{
         type: "language",
         subtag: "xse",
         description: "Sempan",
         added: "2009-07-29"}).
iana(language,xsh,tag{
         type: "language",
         subtag: "xsh",
         description: "Shamang",
         added: "2009-07-29"}).
iana(language,xsi,tag{
         type: "language",
         subtag: "xsi",
         description: "Sio",
         added: "2009-07-29"}).
iana(language,xsj,tag{
         type: "language",
         subtag: "xsj",
         description: "Subi",
         added: "2009-07-29",
         comments: "see also suj"}).
iana(language,xsl,tag{
         type: "language",
         subtag: "xsl",
         description: "South Slavey",
         added: "2009-07-29",
         macrolanguage: "den"}).
iana(language,xsm,tag{
         type: "language",
         subtag: "xsm",
         description: "Kasem",
         added: "2009-07-29"}).
iana(language,xsn,tag{
         type: "language",
         subtag: "xsn",
         description: "Sanga (Nigeria)",
         added: "2009-07-29"}).
iana(language,xso,tag{
         type: "language",
         subtag: "xso",
         description: "Solano",
         added: "2009-07-29"}).
iana(language,xsp,tag{
         type: "language",
         subtag: "xsp",
         description: "Silopi",
         added: "2009-07-29"}).
iana(language,xsq,tag{
         type: "language",
         subtag: "xsq",
         description: "Makhuwa-Saka",
         added: "2009-07-29"}).
iana(language,xsr,tag{
         type: "language",
         subtag: "xsr",
         description: "Sherpa",
         added: "2009-07-29"}).
iana(language,xss,tag{
         type: "language",
         subtag: "xss",
         description: "Assan",
         added: "2009-07-29"}).
iana(language,xsu,tag{
         type: "language",
         subtag: "xsu",
         description: "Sanumá",
         added: "2009-07-29"}).
iana(language,xsv,tag{
         type: "language",
         subtag: "xsv",
         description: "Sudovian",
         added: "2009-07-29"}).
iana(language,xsy,tag{
         type: "language",
         subtag: "xsy",
         description: "Saisiyat",
         added: "2009-07-29"}).
iana(language,xta,tag{
         type: "language",
         subtag: "xta",
         description: "Alcozauca Mixtec",
         added: "2009-07-29"}).
iana(language,xtb,tag{
         type: "language",
         subtag: "xtb",
         description: "Chazumba Mixtec",
         added: "2009-07-29"}).
iana(language,xtc,tag{
         type: "language",
         subtag: "xtc",
         description: "Katcha-Kadugli-Miri",
         added: "2009-07-29"}).
iana(language,xtd,tag{
         type: "language",
         subtag: "xtd",
         description: "Diuxi-Tilantongo Mixtec",
         added: "2009-07-29"}).
iana(language,xte,tag{
         type: "language",
         subtag: "xte",
         description: "Ketengban",
         added: "2009-07-29"}).
iana(language,xtg,tag{
         type: "language",
         subtag: "xtg",
         description: "Transalpine Gaulish",
         added: "2009-07-29"}).
iana(language,xth,tag{
         type: "language",
         subtag: "xth",
         description: "Yitha Yitha",
         added: "2013-09-10"}).
iana(language,xti,tag{
         type: "language",
         subtag: "xti",
         description: "Sinicahua Mixtec",
         added: "2009-07-29"}).
iana(language,xtj,tag{
         type: "language",
         subtag: "xtj",
         description: "San Juan Teita Mixtec",
         added: "2009-07-29"}).
iana(language,xtl,tag{
         type: "language",
         subtag: "xtl",
         description: "Tijaltepec Mixtec",
         added: "2009-07-29"}).
iana(language,xtm,tag{
         type: "language",
         subtag: "xtm",
         description: "Magdalena Peñasco Mixtec",
         added: "2009-07-29"}).
iana(language,xtn,tag{
         type: "language",
         subtag: "xtn",
         description: "Northern Tlaxiaco Mixtec",
         added: "2009-07-29"}).
iana(language,xto,tag{
         type: "language",
         subtag: "xto",
         description: "Tokharian A",
         added: "2009-07-29"}).
iana(language,xtp,tag{
         type: "language",
         subtag: "xtp",
         description: "San Miguel Piedras Mixtec",
         added: "2009-07-29"}).
iana(language,xtq,tag{
         type: "language",
         subtag: "xtq",
         description: "Tumshuqese",
         added: "2009-07-29"}).
iana(language,xtr,tag{
         type: "language",
         subtag: "xtr",
         description: "Early Tripuri",
         added: "2009-07-29"}).
iana(language,xts,tag{
         type: "language",
         subtag: "xts",
         description: "Sindihui Mixtec",
         added: "2009-07-29"}).
iana(language,xtt,tag{
         type: "language",
         subtag: "xtt",
         description: "Tacahua Mixtec",
         added: "2009-07-29"}).
iana(language,xtu,tag{
         type: "language",
         subtag: "xtu",
         description: "Cuyamecalco Mixtec",
         added: "2009-07-29"}).
iana(language,xtv,tag{
         type: "language",
         subtag: "xtv",
         description: "Thawa",
         added: "2013-09-10"}).
iana(language,xtw,tag{
         type: "language",
         subtag: "xtw",
         description: "Tawandê",
         added: "2009-07-29"}).
iana(language,xty,tag{
         type: "language",
         subtag: "xty",
         description: "Yoloxochitl Mixtec",
         added: "2009-07-29"}).
iana(language,xtz,tag{
         type: "language",
         subtag: "xtz",
         description: "Tasmanian",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         comments: "see xpb, xpd, xpf, xph, xpl, xpv, xpw, xpx, xpz"}).
iana(language,xua,tag{
         type: "language",
         subtag: "xua",
         description: "Alu Kurumba",
         added: "2009-07-29"}).
iana(language,xub,tag{
         type: "language",
         subtag: "xub",
         description: "Betta Kurumba",
         added: "2009-07-29"}).
iana(language,xud,tag{
         type: "language",
         subtag: "xud",
         description: "Umiida",
         added: "2012-08-12"}).
iana(language,xug,tag{
         type: "language",
         subtag: "xug",
         description: "Kunigami",
         added: "2009-07-29"}).
iana(language,xuj,tag{
         type: "language",
         subtag: "xuj",
         description: "Jennu Kurumba",
         added: "2009-07-29"}).
iana(language,xul,tag{
         type: "language",
         subtag: "xul",
         description: ["Ngunawal","Nunukul"],
         added: "2013-09-10"}).
iana(language,xum,tag{
         type: "language",
         subtag: "xum",
         description: "Umbrian",
         added: "2009-07-29"}).
iana(language,xun,tag{
         type: "language",
         subtag: "xun",
         description: "Unggaranggu",
         added: "2012-08-12"}).
iana(language,xuo,tag{
         type: "language",
         subtag: "xuo",
         description: "Kuo",
         added: "2009-07-29"}).
iana(language,xup,tag{
         type: "language",
         subtag: "xup",
         description: "Upper Umpqua",
         added: "2009-07-29"}).
iana(language,xur,tag{
         type: "language",
         subtag: "xur",
         description: "Urartian",
         added: "2009-07-29"}).
iana(language,xut,tag{
         type: "language",
         subtag: "xut",
         description: "Kuthant",
         added: "2009-07-29"}).
iana(language,xuu,tag{
         type: "language",
         subtag: "xuu",
         description: ["Kxoe","Khwedam"],
         added: "2009-07-29"}).
iana(language,xve,tag{
         type: "language",
         subtag: "xve",
         description: "Venetic",
         added: "2009-07-29"}).
iana(language,xvi,tag{
         type: "language",
         subtag: "xvi",
         description: "Kamviri",
         added: "2009-07-29"}).
iana(language,xvn,tag{
         type: "language",
         subtag: "xvn",
         description: "Vandalic",
         added: "2009-07-29"}).
iana(language,xvo,tag{
         type: "language",
         subtag: "xvo",
         description: "Volscian",
         added: "2009-07-29"}).
iana(language,xvs,tag{
         type: "language",
         subtag: "xvs",
         description: "Vestinian",
         added: "2009-07-29"}).
iana(language,xwa,tag{
         type: "language",
         subtag: "xwa",
         description: "Kwaza",
         added: "2009-07-29"}).
iana(language,xwc,tag{
         type: "language",
         subtag: "xwc",
         description: "Woccon",
         added: "2009-07-29"}).
iana(language,xwd,tag{
         type: "language",
         subtag: "xwd",
         description: "Wadi Wadi",
         added: "2013-09-10"}).
iana(language,xwe,tag{
         type: "language",
         subtag: "xwe",
         description: "Xwela Gbe",
         added: "2009-07-29"}).
iana(language,xwg,tag{
         type: "language",
         subtag: "xwg",
         description: "Kwegu",
         added: "2009-07-29"}).
iana(language,xwj,tag{
         type: "language",
         subtag: "xwj",
         description: "Wajuk",
         added: "2013-09-10"}).
iana(language,xwk,tag{
         type: "language",
         subtag: "xwk",
         description: "Wangkumara",
         added: "2013-09-10"}).
iana(language,xwl,tag{
         type: "language",
         subtag: "xwl",
         description: "Western Xwla Gbe",
         added: "2009-07-29"}).
iana(language,xwo,tag{
         type: "language",
         subtag: "xwo",
         description: "Written Oirat",
         added: "2009-07-29"}).
iana(language,xwr,tag{
         type: "language",
         subtag: "xwr",
         description: "Kwerba Mamberamo",
         added: "2009-07-29"}).
iana(language,xwt,tag{
         type: "language",
         subtag: "xwt",
         description: "Wotjobaluk",
         added: "2013-09-10"}).
iana(language,xww,tag{
         type: "language",
         subtag: "xww",
         description: "Wemba Wemba",
         added: "2013-09-10"}).
iana(language,xxb,tag{
         type: "language",
         subtag: "xxb",
         description: "Boro (Ghana)",
         added: "2009-07-29"}).
iana(language,xxk,tag{
         type: "language",
         subtag: "xxk",
         description: "Ke'o",
         added: "2009-07-29"}).
iana(language,xxm,tag{
         type: "language",
         subtag: "xxm",
         description: "Minkin",
         added: "2013-09-10"}).
iana(language,xxr,tag{
         type: "language",
         subtag: "xxr",
         description: "Koropó",
         added: "2009-07-29"}).
iana(language,xxt,tag{
         type: "language",
         subtag: "xxt",
         description: "Tambora",
         added: "2009-07-29"}).
iana(language,xya,tag{
         type: "language",
         subtag: "xya",
         description: "Yaygir",
         added: "2013-09-10"}).
iana(language,xyb,tag{
         type: "language",
         subtag: "xyb",
         description: "Yandjibara",
         added: "2013-09-10"}).
iana(language,xyj,tag{
         type: "language",
         subtag: "xyj",
         description: "Mayi-Yapi",
         added: "2013-09-10"}).
iana(language,xyk,tag{
         type: "language",
         subtag: "xyk",
         description: "Mayi-Kulan",
         added: "2013-09-10"}).
iana(language,xyl,tag{
         type: "language",
         subtag: "xyl",
         description: "Yalakalore",
         added: "2009-07-29"}).
iana(language,xyt,tag{
         type: "language",
         subtag: "xyt",
         description: "Mayi-Thakurti",
         added: "2013-09-10"}).
iana(language,xyy,tag{
         type: "language",
         subtag: "xyy",
         description: "Yorta Yorta",
         added: "2012-08-12"}).
iana(language,xzh,tag{
         type: "language",
         subtag: "xzh",
         description: "Zhang-Zhung",
         added: "2009-07-29"}).
iana(language,xzm,tag{
         type: "language",
         subtag: "xzm",
         description: "Zemgalian",
         added: "2009-07-29"}).
iana(language,xzp,tag{
         type: "language",
         subtag: "xzp",
         description: "Ancient Zapotec",
         added: "2009-07-29"}).
iana(language,yaa,tag{
         type: "language",
         subtag: "yaa",
         description: "Yaminahua",
         added: "2009-07-29"}).
iana(language,yab,tag{
         type: "language",
         subtag: "yab",
         description: "Yuhup",
         added: "2009-07-29"}).
iana(language,yac,tag{
         type: "language",
         subtag: "yac",
         description: "Pass Valley Yali",
         added: "2009-07-29"}).
iana(language,yad,tag{
         type: "language",
         subtag: "yad",
         description: "Yagua",
         added: "2009-07-29"}).
iana(language,yae,tag{
         type: "language",
         subtag: "yae",
         description: "Pumé",
         added: "2009-07-29"}).
iana(language,yaf,tag{
         type: "language",
         subtag: "yaf",
         description: "Yaka (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,yag,tag{
         type: "language",
         subtag: "yag",
         description: "Yámana",
         added: "2009-07-29"}).
iana(language,yah,tag{
         type: "language",
         subtag: "yah",
         description: "Yazgulyam",
         added: "2009-07-29"}).
iana(language,yai,tag{
         type: "language",
         subtag: "yai",
         description: "Yagnobi",
         added: "2009-07-29"}).
iana(language,yaj,tag{
         type: "language",
         subtag: "yaj",
         description: "Banda-Yangere",
         added: "2009-07-29"}).
iana(language,yak,tag{
         type: "language",
         subtag: "yak",
         description: "Yakama",
         added: "2009-07-29"}).
iana(language,yal,tag{
         type: "language",
         subtag: "yal",
         description: "Yalunka",
         added: "2009-07-29"}).
iana(language,yam,tag{
         type: "language",
         subtag: "yam",
         description: "Yamba",
         added: "2009-07-29"}).
iana(language,yan,tag{
         type: "language",
         subtag: "yan",
         description: "Mayangna",
         added: "2010-03-11"}).
iana(language,yao,tag{
         type: "language",
         subtag: "yao",
         description: "Yao",
         added: "2005-10-16"}).
iana(language,yap,tag{
         type: "language",
         subtag: "yap",
         description: "Yapese",
         added: "2005-10-16"}).
iana(language,yaq,tag{
         type: "language",
         subtag: "yaq",
         description: "Yaqui",
         added: "2009-07-29"}).
iana(language,yar,tag{
         type: "language",
         subtag: "yar",
         description: "Yabarana",
         added: "2009-07-29"}).
iana(language,yas,tag{
         type: "language",
         subtag: "yas",
         description: "Nugunu (Cameroon)",
         added: "2009-07-29"}).
iana(language,yat,tag{
         type: "language",
         subtag: "yat",
         description: "Yambeta",
         added: "2009-07-29"}).
iana(language,yau,tag{
         type: "language",
         subtag: "yau",
         description: "Yuwana",
         added: "2009-07-29"}).
iana(language,yav,tag{
         type: "language",
         subtag: "yav",
         description: "Yangben",
         added: "2009-07-29"}).
iana(language,yaw,tag{
         type: "language",
         subtag: "yaw",
         description: "Yawalapití",
         added: "2009-07-29"}).
iana(language,yax,tag{
         type: "language",
         subtag: "yax",
         description: "Yauma",
         added: "2009-07-29"}).
iana(language,yay,tag{
         type: "language",
         subtag: "yay",
         description: "Agwagwune",
         added: "2009-07-29"}).
iana(language,yaz,tag{
         type: "language",
         subtag: "yaz",
         description: "Lokaa",
         added: "2009-07-29"}).
iana(language,yba,tag{
         type: "language",
         subtag: "yba",
         description: "Yala",
         added: "2009-07-29"}).
iana(language,ybb,tag{
         type: "language",
         subtag: "ybb",
         description: "Yemba",
         added: "2009-07-29"}).
iana(language,ybd,tag{
         type: "language",
         subtag: "ybd",
         description: "Yangbye",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "rki"}).
iana(language,ybe,tag{
         type: "language",
         subtag: "ybe",
         description: "West Yugur",
         added: "2009-07-29"}).
iana(language,ybh,tag{
         type: "language",
         subtag: "ybh",
         description: "Yakha",
         added: "2009-07-29"}).
iana(language,ybi,tag{
         type: "language",
         subtag: "ybi",
         description: "Yamphu",
         added: "2009-07-29"}).
iana(language,ybj,tag{
         type: "language",
         subtag: "ybj",
         description: "Hasha",
         added: "2009-07-29"}).
iana(language,ybk,tag{
         type: "language",
         subtag: "ybk",
         description: "Bokha",
         added: "2009-07-29"}).
iana(language,ybl,tag{
         type: "language",
         subtag: "ybl",
         description: "Yukuben",
         added: "2009-07-29"}).
iana(language,ybm,tag{
         type: "language",
         subtag: "ybm",
         description: "Yaben",
         added: "2009-07-29"}).
iana(language,ybn,tag{
         type: "language",
         subtag: "ybn",
         description: "Yabaâna",
         added: "2009-07-29"}).
iana(language,ybo,tag{
         type: "language",
         subtag: "ybo",
         description: "Yabong",
         added: "2009-07-29"}).
iana(language,ybx,tag{
         type: "language",
         subtag: "ybx",
         description: "Yawiyo",
         added: "2009-07-29"}).
iana(language,yby,tag{
         type: "language",
         subtag: "yby",
         description: "Yaweyuha",
         added: "2009-07-29"}).
iana(language,ych,tag{
         type: "language",
         subtag: "ych",
         description: "Chesu",
         added: "2009-07-29"}).
iana(language,ycl,tag{
         type: "language",
         subtag: "ycl",
         description: "Lolopo",
         added: "2009-07-29"}).
iana(language,ycn,tag{
         type: "language",
         subtag: "ycn",
         description: "Yucuna",
         added: "2009-07-29"}).
iana(language,ycp,tag{
         type: "language",
         subtag: "ycp",
         description: "Chepya",
         added: "2009-07-29"}).
iana(language,yda,tag{
         type: "language",
         subtag: "yda",
         description: "Yanda",
         added: "2013-09-10"}).
iana(language,ydd,tag{
         type: "language",
         subtag: "ydd",
         description: "Eastern Yiddish",
         added: "2009-07-29",
         macrolanguage: "yi"}).
iana(language,yde,tag{
         type: "language",
         subtag: "yde",
         description: "Yangum Dey",
         added: "2009-07-29"}).
iana(language,ydg,tag{
         type: "language",
         subtag: "ydg",
         description: "Yidgha",
         added: "2009-07-29"}).
iana(language,ydk,tag{
         type: "language",
         subtag: "ydk",
         description: "Yoidik",
         added: "2009-07-29"}).
iana(language,yds,tag{
         type: "language",
         subtag: "yds",
         description: "Yiddish Sign Language",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,yea,tag{
         type: "language",
         subtag: "yea",
         description: "Ravula",
         added: "2009-07-29"}).
iana(language,yec,tag{
         type: "language",
         subtag: "yec",
         description: "Yeniche",
         added: "2009-07-29"}).
iana(language,yee,tag{
         type: "language",
         subtag: "yee",
         description: "Yimas",
         added: "2009-07-29"}).
iana(language,yei,tag{
         type: "language",
         subtag: "yei",
         description: "Yeni",
         added: "2009-07-29"}).
iana(language,yej,tag{
         type: "language",
         subtag: "yej",
         description: "Yevanic",
         added: "2009-07-29"}).
iana(language,yel,tag{
         type: "language",
         subtag: "yel",
         description: "Yela",
         added: "2009-07-29"}).
iana(language,yen,tag{
         type: "language",
         subtag: "yen",
         description: "Yendang",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         comments: "see ynq, yot"}).
iana(language,yer,tag{
         type: "language",
         subtag: "yer",
         description: "Tarok",
         added: "2009-07-29"}).
iana(language,yes,tag{
         type: "language",
         subtag: "yes",
         description: "Nyankpa",
         added: "2009-07-29"}).
iana(language,yet,tag{
         type: "language",
         subtag: "yet",
         description: "Yetfa",
         added: "2009-07-29"}).
iana(language,yeu,tag{
         type: "language",
         subtag: "yeu",
         description: "Yerukula",
         added: "2009-07-29"}).
iana(language,yev,tag{
         type: "language",
         subtag: "yev",
         description: "Yapunda",
         added: "2009-07-29"}).
iana(language,yey,tag{
         type: "language",
         subtag: "yey",
         description: "Yeyi",
         added: "2009-07-29"}).
iana(language,yga,tag{
         type: "language",
         subtag: "yga",
         description: "Malyangapa",
         added: "2012-08-12"}).
iana(language,ygi,tag{
         type: "language",
         subtag: "ygi",
         description: "Yiningayi",
         added: "2013-09-10"}).
iana(language,ygl,tag{
         type: "language",
         subtag: "ygl",
         description: "Yangum Gel",
         added: "2009-07-29"}).
iana(language,ygm,tag{
         type: "language",
         subtag: "ygm",
         description: "Yagomi",
         added: "2009-07-29"}).
iana(language,ygp,tag{
         type: "language",
         subtag: "ygp",
         description: "Gepo",
         added: "2009-07-29"}).
iana(language,ygr,tag{
         type: "language",
         subtag: "ygr",
         description: "Yagaria",
         added: "2009-07-29"}).
iana(language,ygs,tag{
         type: "language",
         subtag: "ygs",
         description: "Yolŋu Sign Language",
         added: "2014-02-28"}).
iana(language,ygu,tag{
         type: "language",
         subtag: "ygu",
         description: "Yugul",
         added: "2013-09-10"}).
iana(language,ygw,tag{
         type: "language",
         subtag: "ygw",
         description: "Yagwoia",
         added: "2009-07-29"}).
iana(language,yha,tag{
         type: "language",
         subtag: "yha",
         description: "Baha Buyang",
         added: "2009-07-29"}).
iana(language,yhd,tag{
         type: "language",
         subtag: "yhd",
         description: "Judeo-Iraqi Arabic",
         added: "2009-07-29",
         macrolanguage: "jrb"}).
iana(language,yhl,tag{
         type: "language",
         subtag: "yhl",
         description: "Hlepho Phowa",
         added: "2009-07-29"}).
iana(language,yhs,tag{
         type: "language",
         subtag: "yhs",
         description: "Yan-nhaŋu Sign Language",
         added: "2015-04-17"}).
iana(language,yia,tag{
         type: "language",
         subtag: "yia",
         description: "Yinggarda",
         added: "2009-07-29"}).
iana(language,yif,tag{
         type: "language",
         subtag: "yif",
         description: "Ache",
         added: "2009-07-29"}).
iana(language,yig,tag{
         type: "language",
         subtag: "yig",
         description: "Wusa Nasu",
         added: "2009-07-29"}).
iana(language,yih,tag{
         type: "language",
         subtag: "yih",
         description: "Western Yiddish",
         added: "2009-07-29",
         macrolanguage: "yi"}).
iana(language,yii,tag{
         type: "language",
         subtag: "yii",
         description: "Yidiny",
         added: "2009-07-29"}).
iana(language,yij,tag{
         type: "language",
         subtag: "yij",
         description: "Yindjibarndi",
         added: "2009-07-29"}).
iana(language,yik,tag{
         type: "language",
         subtag: "yik",
         description: "Dongshanba Lalo",
         added: "2009-07-29"}).
iana(language,yil,tag{
         type: "language",
         subtag: "yil",
         description: "Yindjilandji",
         added: "2009-07-29"}).
iana(language,yim,tag{
         type: "language",
         subtag: "yim",
         description: "Yimchungru Naga",
         added: "2009-07-29"}).
iana(language,yin,tag{
         type: "language",
         subtag: "yin",
         description: ["Riang Lai","Yinchia"],
         added: "2009-07-29"}).
iana(language,yip,tag{
         type: "language",
         subtag: "yip",
         description: "Pholo",
         added: "2009-07-29"}).
iana(language,yiq,tag{
         type: "language",
         subtag: "yiq",
         description: "Miqie",
         added: "2009-07-29"}).
iana(language,yir,tag{
         type: "language",
         subtag: "yir",
         description: "North Awyu",
         added: "2009-07-29"}).
iana(language,yis,tag{
         type: "language",
         subtag: "yis",
         description: "Yis",
         added: "2009-07-29"}).
iana(language,yit,tag{
         type: "language",
         subtag: "yit",
         description: "Eastern Lalu",
         added: "2009-07-29"}).
iana(language,yiu,tag{
         type: "language",
         subtag: "yiu",
         description: "Awu",
         added: "2009-07-29"}).
iana(language,yiv,tag{
         type: "language",
         subtag: "yiv",
         description: "Northern Nisu",
         added: "2009-07-29"}).
iana(language,yix,tag{
         type: "language",
         subtag: "yix",
         description: "Axi Yi",
         added: "2009-07-29"}).
iana(language,yiy,tag{
         type: "language",
         subtag: "yiy",
         description: "Yir Yoront",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         comments: "see yrm, yyr"}).
iana(language,yiz,tag{
         type: "language",
         subtag: "yiz",
         description: "Azhe",
         added: "2009-07-29"}).
iana(language,yka,tag{
         type: "language",
         subtag: "yka",
         description: "Yakan",
         added: "2009-07-29"}).
iana(language,ykg,tag{
         type: "language",
         subtag: "ykg",
         description: "Northern Yukaghir",
         added: "2009-07-29"}).
iana(language,yki,tag{
         type: "language",
         subtag: "yki",
         description: "Yoke",
         added: "2009-07-29"}).
iana(language,ykk,tag{
         type: "language",
         subtag: "ykk",
         description: "Yakaikeke",
         added: "2009-07-29"}).
iana(language,ykl,tag{
         type: "language",
         subtag: "ykl",
         description: "Khlula",
         added: "2009-07-29"}).
iana(language,ykm,tag{
         type: "language",
         subtag: "ykm",
         description: "Kap",
         added: "2009-07-29"}).
iana(language,ykn,tag{
         type: "language",
         subtag: "ykn",
         description: "Kua-nsi",
         added: "2012-08-12"}).
iana(language,yko,tag{
         type: "language",
         subtag: "yko",
         description: "Yasa",
         added: "2009-07-29"}).
iana(language,ykr,tag{
         type: "language",
         subtag: "ykr",
         description: "Yekora",
         added: "2009-07-29"}).
iana(language,ykt,tag{
         type: "language",
         subtag: "ykt",
         description: "Kathu",
         added: "2009-07-29"}).
iana(language,yku,tag{
         type: "language",
         subtag: "yku",
         description: "Kuamasi",
         added: "2012-08-12"}).
iana(language,yky,tag{
         type: "language",
         subtag: "yky",
         description: "Yakoma",
         added: "2009-07-29"}).
iana(language,yla,tag{
         type: "language",
         subtag: "yla",
         description: "Yaul",
         added: "2009-07-29"}).
iana(language,ylb,tag{
         type: "language",
         subtag: "ylb",
         description: "Yaleba",
         added: "2010-03-11"}).
iana(language,yle,tag{
         type: "language",
         subtag: "yle",
         description: "Yele",
         added: "2009-07-29"}).
iana(language,ylg,tag{
         type: "language",
         subtag: "ylg",
         description: "Yelogu",
         added: "2009-07-29"}).
iana(language,yli,tag{
         type: "language",
         subtag: "yli",
         description: "Angguruk Yali",
         added: "2009-07-29"}).
iana(language,yll,tag{
         type: "language",
         subtag: "yll",
         description: "Yil",
         added: "2009-07-29"}).
iana(language,ylm,tag{
         type: "language",
         subtag: "ylm",
         description: "Limi",
         added: "2009-07-29"}).
iana(language,yln,tag{
         type: "language",
         subtag: "yln",
         description: "Langnian Buyang",
         added: "2009-07-29"}).
iana(language,ylo,tag{
         type: "language",
         subtag: "ylo",
         description: "Naluo Yi",
         added: "2009-07-29"}).
iana(language,ylr,tag{
         type: "language",
         subtag: "ylr",
         description: "Yalarnnga",
         added: "2009-07-29"}).
iana(language,ylu,tag{
         type: "language",
         subtag: "ylu",
         description: "Aribwaung",
         added: "2009-07-29"}).
iana(language,yly,tag{
         type: "language",
         subtag: "yly",
         description: ["Nyâlayu","Nyelâyu"],
         added: "2009-07-29"}).
iana(language,yma,tag{
         type: "language",
         subtag: "yma",
         description: "Yamphe",
         added: "2009-07-29",
         deprecated: "2012-08-12",
         preferred_value: "lrr"}).
iana(language,ymb,tag{
         type: "language",
         subtag: "ymb",
         description: "Yambes",
         added: "2009-07-29"}).
iana(language,ymc,tag{
         type: "language",
         subtag: "ymc",
         description: "Southern Muji",
         added: "2009-07-29"}).
iana(language,ymd,tag{
         type: "language",
         subtag: "ymd",
         description: "Muda",
         added: "2009-07-29"}).
iana(language,yme,tag{
         type: "language",
         subtag: "yme",
         description: "Yameo",
         added: "2009-07-29"}).
iana(language,ymg,tag{
         type: "language",
         subtag: "ymg",
         description: "Yamongeri",
         added: "2009-07-29"}).
iana(language,ymh,tag{
         type: "language",
         subtag: "ymh",
         description: "Mili",
         added: "2009-07-29"}).
iana(language,ymi,tag{
         type: "language",
         subtag: "ymi",
         description: "Moji",
         added: "2009-07-29"}).
iana(language,ymk,tag{
         type: "language",
         subtag: "ymk",
         description: "Makwe",
         added: "2009-07-29"}).
iana(language,yml,tag{
         type: "language",
         subtag: "yml",
         description: "Iamalele",
         added: "2009-07-29"}).
iana(language,ymm,tag{
         type: "language",
         subtag: "ymm",
         description: "Maay",
         added: "2009-07-29"}).
iana(language,ymn,tag{
         type: "language",
         subtag: "ymn",
         description: ["Yamna","Sunum"],
         added: "2009-07-29"}).
iana(language,ymo,tag{
         type: "language",
         subtag: "ymo",
         description: "Yangum Mon",
         added: "2009-07-29"}).
iana(language,ymp,tag{
         type: "language",
         subtag: "ymp",
         description: "Yamap",
         added: "2009-07-29"}).
iana(language,ymq,tag{
         type: "language",
         subtag: "ymq",
         description: "Qila Muji",
         added: "2009-07-29"}).
iana(language,ymr,tag{
         type: "language",
         subtag: "ymr",
         description: "Malasar",
         added: "2009-07-29"}).
iana(language,yms,tag{
         type: "language",
         subtag: "yms",
         description: "Mysian",
         added: "2009-07-29"}).
iana(language,ymt,tag{
         type: "language",
         subtag: "ymt",
         description: "Mator-Taygi-Karagas",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "mtm"}).
iana(language,ymx,tag{
         type: "language",
         subtag: "ymx",
         description: "Northern Muji",
         added: "2009-07-29"}).
iana(language,ymz,tag{
         type: "language",
         subtag: "ymz",
         description: "Muzi",
         added: "2009-07-29"}).
iana(language,yna,tag{
         type: "language",
         subtag: "yna",
         description: "Aluo",
         added: "2009-07-29"}).
iana(language,ynd,tag{
         type: "language",
         subtag: "ynd",
         description: "Yandruwandha",
         added: "2009-07-29"}).
iana(language,yne,tag{
         type: "language",
         subtag: "yne",
         description: "Lang'e",
         added: "2009-07-29"}).
iana(language,yng,tag{
         type: "language",
         subtag: "yng",
         description: "Yango",
         added: "2009-07-29"}).
iana(language,ynh,tag{
         type: "language",
         subtag: "ynh",
         description: "Yangho",
         added: "2009-07-29",
         deprecated: "2015-02-12"}).
iana(language,ynk,tag{
         type: "language",
         subtag: "ynk",
         description: "Naukan Yupik",
         added: "2009-07-29"}).
iana(language,ynl,tag{
         type: "language",
         subtag: "ynl",
         description: "Yangulam",
         added: "2009-07-29"}).
iana(language,ynn,tag{
         type: "language",
         subtag: "ynn",
         description: "Yana",
         added: "2009-07-29"}).
iana(language,yno,tag{
         type: "language",
         subtag: "yno",
         description: "Yong",
         added: "2009-07-29"}).
iana(language,ynq,tag{
         type: "language",
         subtag: "ynq",
         description: "Yendang",
         added: "2012-08-12"}).
iana(language,yns,tag{
         type: "language",
         subtag: "yns",
         description: "Yansi",
         added: "2009-07-29"}).
iana(language,ynu,tag{
         type: "language",
         subtag: "ynu",
         description: "Yahuna",
         added: "2009-07-29"}).
iana(language,yob,tag{
         type: "language",
         subtag: "yob",
         description: "Yoba",
         added: "2009-07-29"}).
iana(language,yog,tag{
         type: "language",
         subtag: "yog",
         description: "Yogad",
         added: "2009-07-29"}).
iana(language,yoi,tag{
         type: "language",
         subtag: "yoi",
         description: "Yonaguni",
         added: "2009-07-29"}).
iana(language,yok,tag{
         type: "language",
         subtag: "yok",
         description: "Yokuts",
         added: "2009-07-29"}).
iana(language,yol,tag{
         type: "language",
         subtag: "yol",
         description: "Yola",
         added: "2009-07-29"}).
iana(language,yom,tag{
         type: "language",
         subtag: "yom",
         description: "Yombe",
         added: "2009-07-29"}).
iana(language,yon,tag{
         type: "language",
         subtag: "yon",
         description: "Yongkom",
         added: "2009-07-29"}).
iana(language,yos,tag{
         type: "language",
         subtag: "yos",
         description: "Yos",
         added: "2009-07-29",
         deprecated: "2013-09-10",
         preferred_value: "zom"}).
iana(language,yot,tag{
         type: "language",
         subtag: "yot",
         description: "Yotti",
         added: "2012-08-12"}).
iana(language,yox,tag{
         type: "language",
         subtag: "yox",
         description: "Yoron",
         added: "2009-07-29"}).
iana(language,yoy,tag{
         type: "language",
         subtag: "yoy",
         description: "Yoy",
         added: "2009-07-29"}).
iana(language,ypa,tag{
         type: "language",
         subtag: "ypa",
         description: "Phala",
         added: "2009-07-29"}).
iana(language,ypb,tag{
         type: "language",
         subtag: "ypb",
         description: "Labo Phowa",
         added: "2009-07-29"}).
iana(language,ypg,tag{
         type: "language",
         subtag: "ypg",
         description: "Phola",
         added: "2009-07-29"}).
iana(language,yph,tag{
         type: "language",
         subtag: "yph",
         description: "Phupha",
         added: "2009-07-29"}).
iana(language,ypk,tag{
         type: "language",
         subtag: "ypk",
         description: "Yupik languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,ypm,tag{
         type: "language",
         subtag: "ypm",
         description: "Phuma",
         added: "2009-07-29"}).
iana(language,ypn,tag{
         type: "language",
         subtag: "ypn",
         description: "Ani Phowa",
         added: "2009-07-29"}).
iana(language,ypo,tag{
         type: "language",
         subtag: "ypo",
         description: "Alo Phola",
         added: "2009-07-29"}).
iana(language,ypp,tag{
         type: "language",
         subtag: "ypp",
         description: "Phupa",
         added: "2009-07-29"}).
iana(language,ypz,tag{
         type: "language",
         subtag: "ypz",
         description: "Phuza",
         added: "2009-07-29"}).
iana(language,yra,tag{
         type: "language",
         subtag: "yra",
         description: "Yerakai",
         added: "2009-07-29"}).
iana(language,yrb,tag{
         type: "language",
         subtag: "yrb",
         description: "Yareba",
         added: "2009-07-29"}).
iana(language,yre,tag{
         type: "language",
         subtag: "yre",
         description: "Yaouré",
         added: "2009-07-29"}).
iana(language,yri,tag{
         type: "language",
         subtag: "yri",
         description: "Yarí",
         added: "2009-07-29",
         deprecated: "2016-05-30"}).
iana(language,yrk,tag{
         type: "language",
         subtag: "yrk",
         description: "Nenets",
         added: "2009-07-29"}).
iana(language,yrl,tag{
         type: "language",
         subtag: "yrl",
         description: "Nhengatu",
         added: "2009-07-29"}).
iana(language,yrm,tag{
         type: "language",
         subtag: "yrm",
         description: "Yirrk-Mel",
         added: "2013-09-10"}).
iana(language,yrn,tag{
         type: "language",
         subtag: "yrn",
         description: "Yerong",
         added: "2009-07-29"}).
iana(language,yro,tag{
         type: "language",
         subtag: "yro",
         description: "Yaroamë",
         added: "2016-05-30"}).
iana(language,yrs,tag{
         type: "language",
         subtag: "yrs",
         description: "Yarsun",
         added: "2009-07-29"}).
iana(language,yrw,tag{
         type: "language",
         subtag: "yrw",
         description: "Yarawata",
         added: "2009-07-29"}).
iana(language,yry,tag{
         type: "language",
         subtag: "yry",
         description: "Yarluyandi",
         added: "2013-09-10"}).
iana(language,ysc,tag{
         type: "language",
         subtag: "ysc",
         description: "Yassic",
         added: "2009-07-29"}).
iana(language,ysd,tag{
         type: "language",
         subtag: "ysd",
         description: "Samatao",
         added: "2009-07-29"}).
iana(language,ysg,tag{
         type: "language",
         subtag: "ysg",
         description: "Sonaga",
         added: "2012-08-12"}).
iana(language,ysl,tag{
         type: "language",
         subtag: "ysl",
         description: "Yugoslavian Sign Language",
         added: "2009-07-29"}).
iana(language,ysm,tag{
         type: "language",
         subtag: "ysm",
         description: "Myanmar Sign Language",
         added: "2021-02-20"}).
iana(language,ysn,tag{
         type: "language",
         subtag: "ysn",
         description: "Sani",
         added: "2009-07-29"}).
iana(language,yso,tag{
         type: "language",
         subtag: "yso",
         description: "Nisi (China)",
         added: "2009-07-29"}).
iana(language,ysp,tag{
         type: "language",
         subtag: "ysp",
         description: "Southern Lolopo",
         added: "2009-07-29"}).
iana(language,ysr,tag{
         type: "language",
         subtag: "ysr",
         description: "Sirenik Yupik",
         added: "2009-07-29"}).
iana(language,yss,tag{
         type: "language",
         subtag: "yss",
         description: "Yessan-Mayo",
         added: "2009-07-29"}).
iana(language,ysy,tag{
         type: "language",
         subtag: "ysy",
         description: "Sanie",
         added: "2009-07-29"}).
iana(language,yta,tag{
         type: "language",
         subtag: "yta",
         description: "Talu",
         added: "2009-07-29"}).
iana(language,ytl,tag{
         type: "language",
         subtag: "ytl",
         description: "Tanglang",
         added: "2009-07-29"}).
iana(language,ytp,tag{
         type: "language",
         subtag: "ytp",
         description: "Thopho",
         added: "2009-07-29"}).
iana(language,ytw,tag{
         type: "language",
         subtag: "ytw",
         description: "Yout Wam",
         added: "2010-03-11"}).
iana(language,yty,tag{
         type: "language",
         subtag: "yty",
         description: "Yatay",
         added: "2013-09-10"}).
iana(language,yua,tag{
         type: "language",
         subtag: "yua",
         description: ["Yucateco","Yucatec Maya"],
         added: "2009-07-29"}).
iana(language,yub,tag{
         type: "language",
         subtag: "yub",
         description: "Yugambal",
         added: "2009-07-29"}).
iana(language,yuc,tag{
         type: "language",
         subtag: "yuc",
         description: "Yuchi",
         added: "2009-07-29"}).
iana(language,yud,tag{
         type: "language",
         subtag: "yud",
         description: "Judeo-Tripolitanian Arabic",
         added: "2009-07-29",
         macrolanguage: "jrb"}).
iana(language,yue,tag{
         type: "language",
         subtag: "yue",
         description: ["Yue Chinese","Cantonese"],
         added: "2009-07-29",
         macrolanguage: "zh"}).
iana(language,yuf,tag{
         type: "language",
         subtag: "yuf",
         description: "Havasupai-Walapai-Yavapai",
         added: "2009-07-29"}).
iana(language,yug,tag{
         type: "language",
         subtag: "yug",
         description: "Yug",
         added: "2009-07-29"}).
iana(language,yui,tag{
         type: "language",
         subtag: "yui",
         description: "Yurutí",
         added: "2009-07-29"}).
iana(language,yuj,tag{
         type: "language",
         subtag: "yuj",
         description: "Karkar-Yuri",
         added: "2009-07-29"}).
iana(language,yuk,tag{
         type: "language",
         subtag: "yuk",
         description: "Yuki",
         added: "2009-07-29"}).
iana(language,yul,tag{
         type: "language",
         subtag: "yul",
         description: "Yulu",
         added: "2009-07-29"}).
iana(language,yum,tag{
         type: "language",
         subtag: "yum",
         description: "Quechan",
         added: "2009-07-29"}).
iana(language,yun,tag{
         type: "language",
         subtag: "yun",
         description: "Bena (Nigeria)",
         added: "2009-07-29"}).
iana(language,yup,tag{
         type: "language",
         subtag: "yup",
         description: "Yukpa",
         added: "2009-07-29"}).
iana(language,yuq,tag{
         type: "language",
         subtag: "yuq",
         description: "Yuqui",
         added: "2009-07-29"}).
iana(language,yur,tag{
         type: "language",
         subtag: "yur",
         description: "Yurok",
         added: "2009-07-29"}).
iana(language,yut,tag{
         type: "language",
         subtag: "yut",
         description: "Yopno",
         added: "2009-07-29"}).
iana(language,yuu,tag{
         type: "language",
         subtag: "yuu",
         description: "Yugh",
         added: "2009-07-29",
         deprecated: "2014-02-28",
         preferred_value: "yug"}).
iana(language,yuw,tag{
         type: "language",
         subtag: "yuw",
         description: "Yau (Morobe Province)",
         added: "2009-07-29"}).
iana(language,yux,tag{
         type: "language",
         subtag: "yux",
         description: "Southern Yukaghir",
         added: "2009-07-29"}).
iana(language,yuy,tag{
         type: "language",
         subtag: "yuy",
         description: "East Yugur",
         added: "2009-07-29"}).
iana(language,yuz,tag{
         type: "language",
         subtag: "yuz",
         description: "Yuracare",
         added: "2009-07-29"}).
iana(language,yva,tag{
         type: "language",
         subtag: "yva",
         description: "Yawa",
         added: "2009-07-29"}).
iana(language,yvt,tag{
         type: "language",
         subtag: "yvt",
         description: "Yavitero",
         added: "2009-07-29"}).
iana(language,ywa,tag{
         type: "language",
         subtag: "ywa",
         description: "Kalou",
         added: "2009-07-29"}).
iana(language,ywg,tag{
         type: "language",
         subtag: "ywg",
         description: "Yinhawangka",
         added: "2013-09-10"}).
iana(language,ywl,tag{
         type: "language",
         subtag: "ywl",
         description: "Western Lalu",
         added: "2009-07-29"}).
iana(language,ywn,tag{
         type: "language",
         subtag: "ywn",
         description: "Yawanawa",
         added: "2009-07-29"}).
iana(language,ywq,tag{
         type: "language",
         subtag: "ywq",
         description: "Wuding-Luquan Yi",
         added: "2009-07-29"}).
iana(language,ywr,tag{
         type: "language",
         subtag: "ywr",
         description: "Yawuru",
         added: "2009-07-29"}).
iana(language,ywt,tag{
         type: "language",
         subtag: "ywt",
         description: ["Xishanba Lalo","Central Lalo"],
         added: "2009-07-29"}).
iana(language,ywu,tag{
         type: "language",
         subtag: "ywu",
         description: "Wumeng Nasu",
         added: "2009-07-29"}).
iana(language,yww,tag{
         type: "language",
         subtag: "yww",
         description: "Yawarawarga",
         added: "2009-07-29"}).
iana(language,yxa,tag{
         type: "language",
         subtag: "yxa",
         description: "Mayawali",
         added: "2013-09-10"}).
iana(language,yxg,tag{
         type: "language",
         subtag: "yxg",
         description: "Yagara",
         added: "2012-08-12"}).
iana(language,yxl,tag{
         type: "language",
         subtag: "yxl",
         description: "Yardliyawarra",
         added: "2013-09-10"}).
iana(language,yxm,tag{
         type: "language",
         subtag: "yxm",
         description: "Yinwum",
         added: "2013-09-10"}).
iana(language,yxu,tag{
         type: "language",
         subtag: "yxu",
         description: "Yuyu",
         added: "2013-09-10"}).
iana(language,yxy,tag{
         type: "language",
         subtag: "yxy",
         description: "Yabula Yabula",
         added: "2012-08-12"}).
iana(language,yyr,tag{
         type: "language",
         subtag: "yyr",
         description: "Yir Yoront",
         added: "2013-09-03"}).
iana(language,yyu,tag{
         type: "language",
         subtag: "yyu",
         description: "Yau (Sandaun Province)",
         added: "2009-07-29"}).
iana(language,yyz,tag{
         type: "language",
         subtag: "yyz",
         description: "Ayizi",
         added: "2009-07-29"}).
iana(language,yzg,tag{
         type: "language",
         subtag: "yzg",
         description: "E'ma Buyang",
         added: "2009-07-29"}).
iana(language,yzk,tag{
         type: "language",
         subtag: "yzk",
         description: "Zokhuo",
         added: "2009-07-29"}).
iana(language,zaa,tag{
         type: "language",
         subtag: "zaa",
         description: "Sierra de Juárez Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zab,tag{
         type: "language",
         subtag: "zab",
         description: ["Western Tlacolula Valley Zapotec","San Juan Guelavía Zapotec"],
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zac,tag{
         type: "language",
         subtag: "zac",
         description: "Ocotlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zad,tag{
         type: "language",
         subtag: "zad",
         description: "Cajonos Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zae,tag{
         type: "language",
         subtag: "zae",
         description: "Yareni Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zaf,tag{
         type: "language",
         subtag: "zaf",
         description: "Ayoquesco Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zag,tag{
         type: "language",
         subtag: "zag",
         description: "Zaghawa",
         added: "2009-07-29"}).
iana(language,zah,tag{
         type: "language",
         subtag: "zah",
         description: "Zangwal",
         added: "2009-07-29"}).
iana(language,zai,tag{
         type: "language",
         subtag: "zai",
         description: "Isthmus Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zaj,tag{
         type: "language",
         subtag: "zaj",
         description: "Zaramo",
         added: "2009-07-29"}).
iana(language,zak,tag{
         type: "language",
         subtag: "zak",
         description: "Zanaki",
         added: "2009-07-29"}).
iana(language,zal,tag{
         type: "language",
         subtag: "zal",
         description: "Zauzou",
         added: "2009-07-29"}).
iana(language,zam,tag{
         type: "language",
         subtag: "zam",
         description: "Miahuatlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zao,tag{
         type: "language",
         subtag: "zao",
         description: "Ozolotepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zap,tag{
         type: "language",
         subtag: "zap",
         description: "Zapotec",
         added: "2005-10-16",
         scope: "macrolanguage"}).
iana(language,zaq,tag{
         type: "language",
         subtag: "zaq",
         description: "Aloápam Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zar,tag{
         type: "language",
         subtag: "zar",
         description: "Rincón Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zas,tag{
         type: "language",
         subtag: "zas",
         description: "Santo Domingo Albarradas Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zat,tag{
         type: "language",
         subtag: "zat",
         description: "Tabaa Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zau,tag{
         type: "language",
         subtag: "zau",
         description: "Zangskari",
         added: "2009-07-29"}).
iana(language,zav,tag{
         type: "language",
         subtag: "zav",
         description: "Yatzachi Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zaw,tag{
         type: "language",
         subtag: "zaw",
         description: "Mitla Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zax,tag{
         type: "language",
         subtag: "zax",
         description: "Xadani Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zay,tag{
         type: "language",
         subtag: "zay",
         description: ["Zayse-Zergulla","Zaysete"],
         added: "2009-07-29"}).
iana(language,zaz,tag{
         type: "language",
         subtag: "zaz",
         description: "Zari",
         added: "2009-07-29"}).
iana(language,zba,tag{
         type: "language",
         subtag: "zba",
         description: "Balaibalan",
         added: "2020-03-28"}).
iana(language,zbc,tag{
         type: "language",
         subtag: "zbc",
         description: "Central Berawan",
         added: "2009-07-29"}).
iana(language,zbe,tag{
         type: "language",
         subtag: "zbe",
         description: "East Berawan",
         added: "2009-07-29"}).
iana(language,zbl,tag{
         type: "language",
         subtag: "zbl",
         description: ["Blissymbols","Bliss","Blissymbolics"],
         added: "2007-08-21",
         suppress_script: "Blis"}).
iana(language,zbt,tag{
         type: "language",
         subtag: "zbt",
         description: "Batui",
         added: "2009-07-29"}).
iana(language,zbu,tag{
         type: "language",
         subtag: "zbu",
         description: "Bu (Bauchi State)",
         added: "2021-02-20"}).
iana(language,zbw,tag{
         type: "language",
         subtag: "zbw",
         description: "West Berawan",
         added: "2009-07-29"}).
iana(language,zca,tag{
         type: "language",
         subtag: "zca",
         description: "Coatecas Altas Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zcd,tag{
         type: "language",
         subtag: "zcd",
         description: "Las Delicias Zapotec",
         added: "2022-02-25",
         macrolanguage: "zap"}).
iana(language,zch,tag{
         type: "language",
         subtag: "zch",
         description: "Central Hongshuihe Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zdj,tag{
         type: "language",
         subtag: "zdj",
         description: "Ngazidja Comorian",
         added: "2009-07-29"}).
iana(language,zea,tag{
         type: "language",
         subtag: "zea",
         description: "Zeeuws",
         added: "2009-07-29"}).
iana(language,zeg,tag{
         type: "language",
         subtag: "zeg",
         description: "Zenag",
         added: "2009-07-29"}).
iana(language,zeh,tag{
         type: "language",
         subtag: "zeh",
         description: "Eastern Hongshuihe Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zen,tag{
         type: "language",
         subtag: "zen",
         description: "Zenaga",
         added: "2005-10-16"}).
iana(language,zga,tag{
         type: "language",
         subtag: "zga",
         description: "Kinga",
         added: "2009-07-29"}).
iana(language,zgb,tag{
         type: "language",
         subtag: "zgb",
         description: "Guibei Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zgh,tag{
         type: "language",
         subtag: "zgh",
         description: "Standard Moroccan Tamazight",
         added: "2013-01-25"}).
iana(language,zgm,tag{
         type: "language",
         subtag: "zgm",
         description: "Minz Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zgn,tag{
         type: "language",
         subtag: "zgn",
         description: "Guibian Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zgr,tag{
         type: "language",
         subtag: "zgr",
         description: "Magori",
         added: "2009-07-29"}).
iana(language,zhb,tag{
         type: "language",
         subtag: "zhb",
         description: "Zhaba",
         added: "2009-07-29"}).
iana(language,zhd,tag{
         type: "language",
         subtag: "zhd",
         description: "Dai Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zhi,tag{
         type: "language",
         subtag: "zhi",
         description: "Zhire",
         added: "2009-07-29"}).
iana(language,zhn,tag{
         type: "language",
         subtag: "zhn",
         description: "Nong Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zhw,tag{
         type: "language",
         subtag: "zhw",
         description: "Zhoa",
         added: "2009-07-29"}).
iana(language,zhx,tag{
         type: "language",
         subtag: "zhx",
         description: "Chinese (family)",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,zia,tag{
         type: "language",
         subtag: "zia",
         description: "Zia",
         added: "2009-07-29"}).
iana(language,zib,tag{
         type: "language",
         subtag: "zib",
         description: "Zimbabwe Sign Language",
         added: "2009-07-29"}).
iana(language,zik,tag{
         type: "language",
         subtag: "zik",
         description: "Zimakani",
         added: "2009-07-29"}).
iana(language,zil,tag{
         type: "language",
         subtag: "zil",
         description: "Zialo",
         added: "2011-08-16"}).
iana(language,zim,tag{
         type: "language",
         subtag: "zim",
         description: "Mesme",
         added: "2009-07-29"}).
iana(language,zin,tag{
         type: "language",
         subtag: "zin",
         description: "Zinza",
         added: "2009-07-29"}).
iana(language,zir,tag{
         type: "language",
         subtag: "zir",
         description: "Ziriya",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "scv"}).
iana(language,ziw,tag{
         type: "language",
         subtag: "ziw",
         description: "Zigula",
         added: "2009-07-29"}).
iana(language,ziz,tag{
         type: "language",
         subtag: "ziz",
         description: "Zizilivakan",
         added: "2009-07-29"}).
iana(language,zka,tag{
         type: "language",
         subtag: "zka",
         description: "Kaimbulawa",
         added: "2009-07-29"}).
iana(language,zkb,tag{
         type: "language",
         subtag: "zkb",
         description: "Koibal",
         added: "2009-07-29"}).
iana(language,zkd,tag{
         type: "language",
         subtag: "zkd",
         description: "Kadu",
         added: "2012-08-12"}).
iana(language,zkg,tag{
         type: "language",
         subtag: "zkg",
         description: "Koguryo",
         added: "2009-07-29"}).
iana(language,zkh,tag{
         type: "language",
         subtag: "zkh",
         description: "Khorezmian",
         added: "2009-07-29"}).
iana(language,zkk,tag{
         type: "language",
         subtag: "zkk",
         description: "Karankawa",
         added: "2009-07-29"}).
iana(language,zkn,tag{
         type: "language",
         subtag: "zkn",
         description: "Kanan",
         added: "2012-08-12"}).
iana(language,zko,tag{
         type: "language",
         subtag: "zko",
         description: "Kott",
         added: "2009-07-29"}).
iana(language,zkp,tag{
         type: "language",
         subtag: "zkp",
         description: "São Paulo Kaingáng",
         added: "2009-07-29"}).
iana(language,zkr,tag{
         type: "language",
         subtag: "zkr",
         description: "Zakhring",
         added: "2009-07-29"}).
iana(language,zkt,tag{
         type: "language",
         subtag: "zkt",
         description: "Kitan",
         added: "2009-07-29"}).
iana(language,zku,tag{
         type: "language",
         subtag: "zku",
         description: "Kaurna",
         added: "2009-07-29"}).
iana(language,zkv,tag{
         type: "language",
         subtag: "zkv",
         description: "Krevinian",
         added: "2009-07-29"}).
iana(language,zkz,tag{
         type: "language",
         subtag: "zkz",
         description: "Khazar",
         added: "2009-07-29"}).
iana(language,zla,tag{
         type: "language",
         subtag: "zla",
         description: "Zula",
         added: "2021-02-20"}).
iana(language,zle,tag{
         type: "language",
         subtag: "zle",
         description: "East Slavic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,zlj,tag{
         type: "language",
         subtag: "zlj",
         description: "Liujiang Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zlm,tag{
         type: "language",
         subtag: "zlm",
         description: "Malay (individual language)",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,zln,tag{
         type: "language",
         subtag: "zln",
         description: "Lianshan Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zlq,tag{
         type: "language",
         subtag: "zlq",
         description: "Liuqian Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zls,tag{
         type: "language",
         subtag: "zls",
         description: "South Slavic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,zlw,tag{
         type: "language",
         subtag: "zlw",
         description: "West Slavic languages",
         added: "2009-07-29",
         scope: "collection"}).
iana(language,zma,tag{
         type: "language",
         subtag: "zma",
         description: "Manda (Australia)",
         added: "2009-07-29"}).
iana(language,zmb,tag{
         type: "language",
         subtag: "zmb",
         description: "Zimba",
         added: "2009-07-29"}).
iana(language,zmc,tag{
         type: "language",
         subtag: "zmc",
         description: "Margany",
         added: "2009-07-29"}).
iana(language,zmd,tag{
         type: "language",
         subtag: "zmd",
         description: "Maridan",
         added: "2009-07-29"}).
iana(language,zme,tag{
         type: "language",
         subtag: "zme",
         description: "Mangerr",
         added: "2009-07-29"}).
iana(language,zmf,tag{
         type: "language",
         subtag: "zmf",
         description: "Mfinu",
         added: "2009-07-29"}).
iana(language,zmg,tag{
         type: "language",
         subtag: "zmg",
         description: "Marti Ke",
         added: "2009-07-29"}).
iana(language,zmh,tag{
         type: "language",
         subtag: "zmh",
         description: "Makolkol",
         added: "2009-07-29"}).
iana(language,zmi,tag{
         type: "language",
         subtag: "zmi",
         description: "Negeri Sembilan Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,zmj,tag{
         type: "language",
         subtag: "zmj",
         description: "Maridjabin",
         added: "2009-07-29"}).
iana(language,zmk,tag{
         type: "language",
         subtag: "zmk",
         description: "Mandandanyi",
         added: "2009-07-29"}).
iana(language,zml,tag{
         type: "language",
         subtag: "zml",
         description: "Matngala",
         added: "2009-07-29"}).
iana(language,zmm,tag{
         type: "language",
         subtag: "zmm",
         description: ["Marimanindji","Marramaninyshi"],
         added: "2009-07-29"}).
iana(language,zmn,tag{
         type: "language",
         subtag: "zmn",
         description: "Mbangwe",
         added: "2009-07-29"}).
iana(language,zmo,tag{
         type: "language",
         subtag: "zmo",
         description: "Molo",
         added: "2009-07-29"}).
iana(language,zmp,tag{
         type: "language",
         subtag: "zmp",
         description: "Mpuono",
         added: "2009-07-29"}).
iana(language,zmq,tag{
         type: "language",
         subtag: "zmq",
         description: "Mituku",
         added: "2009-07-29"}).
iana(language,zmr,tag{
         type: "language",
         subtag: "zmr",
         description: "Maranunggu",
         added: "2009-07-29"}).
iana(language,zms,tag{
         type: "language",
         subtag: "zms",
         description: "Mbesa",
         added: "2009-07-29"}).
iana(language,zmt,tag{
         type: "language",
         subtag: "zmt",
         description: "Maringarr",
         added: "2009-07-29"}).
iana(language,zmu,tag{
         type: "language",
         subtag: "zmu",
         description: "Muruwari",
         added: "2009-07-29"}).
iana(language,zmv,tag{
         type: "language",
         subtag: "zmv",
         description: "Mbariman-Gudhinma",
         added: "2009-07-29"}).
iana(language,zmw,tag{
         type: "language",
         subtag: "zmw",
         description: "Mbo (Democratic Republic of Congo)",
         added: "2009-07-29"}).
iana(language,zmx,tag{
         type: "language",
         subtag: "zmx",
         description: "Bomitaba",
         added: "2009-07-29"}).
iana(language,zmy,tag{
         type: "language",
         subtag: "zmy",
         description: "Mariyedi",
         added: "2009-07-29"}).
iana(language,zmz,tag{
         type: "language",
         subtag: "zmz",
         description: "Mbandja",
         added: "2009-07-29"}).
iana(language,zna,tag{
         type: "language",
         subtag: "zna",
         description: "Zan Gula",
         added: "2009-07-29"}).
iana(language,znd,tag{
         type: "language",
         subtag: "znd",
         description: "Zande languages",
         added: "2005-10-16",
         scope: "collection"}).
iana(language,zne,tag{
         type: "language",
         subtag: "zne",
         description: "Zande (individual language)",
         added: "2009-07-29"}).
iana(language,zng,tag{
         type: "language",
         subtag: "zng",
         description: "Mang",
         added: "2009-07-29"}).
iana(language,znk,tag{
         type: "language",
         subtag: "znk",
         description: "Manangkari",
         added: "2009-07-29"}).
iana(language,zns,tag{
         type: "language",
         subtag: "zns",
         description: "Mangas",
         added: "2009-07-29"}).
iana(language,zoc,tag{
         type: "language",
         subtag: "zoc",
         description: "Copainalá Zoque",
         added: "2009-07-29"}).
iana(language,zoh,tag{
         type: "language",
         subtag: "zoh",
         description: "Chimalapa Zoque",
         added: "2009-07-29"}).
iana(language,zom,tag{
         type: "language",
         subtag: "zom",
         description: "Zou",
         added: "2009-07-29"}).
iana(language,zoo,tag{
         type: "language",
         subtag: "zoo",
         description: "Asunción Mixtepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zoq,tag{
         type: "language",
         subtag: "zoq",
         description: "Tabasco Zoque",
         added: "2009-07-29"}).
iana(language,zor,tag{
         type: "language",
         subtag: "zor",
         description: "Rayón Zoque",
         added: "2009-07-29"}).
iana(language,zos,tag{
         type: "language",
         subtag: "zos",
         description: "Francisco León Zoque",
         added: "2009-07-29"}).
iana(language,zpa,tag{
         type: "language",
         subtag: "zpa",
         description: "Lachiguiri Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpb,tag{
         type: "language",
         subtag: "zpb",
         description: "Yautepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpc,tag{
         type: "language",
         subtag: "zpc",
         description: "Choapan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpd,tag{
         type: "language",
         subtag: "zpd",
         description: "Southeastern Ixtlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpe,tag{
         type: "language",
         subtag: "zpe",
         description: "Petapa Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpf,tag{
         type: "language",
         subtag: "zpf",
         description: "San Pedro Quiatoni Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpg,tag{
         type: "language",
         subtag: "zpg",
         description: "Guevea De Humboldt Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zph,tag{
         type: "language",
         subtag: "zph",
         description: "Totomachapan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpi,tag{
         type: "language",
         subtag: "zpi",
         description: "Santa María Quiegolani Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpj,tag{
         type: "language",
         subtag: "zpj",
         description: "Quiavicuzas Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpk,tag{
         type: "language",
         subtag: "zpk",
         description: "Tlacolulita Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpl,tag{
         type: "language",
         subtag: "zpl",
         description: "Lachixío Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpm,tag{
         type: "language",
         subtag: "zpm",
         description: "Mixtepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpn,tag{
         type: "language",
         subtag: "zpn",
         description: "Santa Inés Yatzechi Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpo,tag{
         type: "language",
         subtag: "zpo",
         description: "Amatlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpp,tag{
         type: "language",
         subtag: "zpp",
         description: "El Alto Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpq,tag{
         type: "language",
         subtag: "zpq",
         description: "Zoogocho Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpr,tag{
         type: "language",
         subtag: "zpr",
         description: "Santiago Xanica Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zps,tag{
         type: "language",
         subtag: "zps",
         description: "Coatlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpt,tag{
         type: "language",
         subtag: "zpt",
         description: "San Vicente Coatlán Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpu,tag{
         type: "language",
         subtag: "zpu",
         description: "Yalálag Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpv,tag{
         type: "language",
         subtag: "zpv",
         description: "Chichicapan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpw,tag{
         type: "language",
         subtag: "zpw",
         description: "Zaniza Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpx,tag{
         type: "language",
         subtag: "zpx",
         description: "San Baltazar Loxicha Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpy,tag{
         type: "language",
         subtag: "zpy",
         description: "Mazaltepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zpz,tag{
         type: "language",
         subtag: "zpz",
         description: "Texmelucan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zqe,tag{
         type: "language",
         subtag: "zqe",
         description: "Qiubei Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zra,tag{
         type: "language",
         subtag: "zra",
         description: "Kara (Korea)",
         added: "2009-07-29"}).
iana(language,zrg,tag{
         type: "language",
         subtag: "zrg",
         description: "Mirgan",
         added: "2009-07-29"}).
iana(language,zrn,tag{
         type: "language",
         subtag: "zrn",
         description: "Zerenkel",
         added: "2009-07-29"}).
iana(language,zro,tag{
         type: "language",
         subtag: "zro",
         description: "Záparo",
         added: "2009-07-29"}).
iana(language,zrp,tag{
         type: "language",
         subtag: "zrp",
         description: "Zarphatic",
         added: "2009-07-29"}).
iana(language,zrs,tag{
         type: "language",
         subtag: "zrs",
         description: "Mairasi",
         added: "2009-07-29"}).
iana(language,zsa,tag{
         type: "language",
         subtag: "zsa",
         description: "Sarasira",
         added: "2009-07-29"}).
iana(language,zsk,tag{
         type: "language",
         subtag: "zsk",
         description: "Kaskean",
         added: "2009-07-29"}).
iana(language,zsl,tag{
         type: "language",
         subtag: "zsl",
         description: "Zambian Sign Language",
         added: "2009-07-29"}).
iana(language,zsm,tag{
         type: "language",
         subtag: "zsm",
         description: "Standard Malay",
         added: "2009-07-29",
         macrolanguage: "ms"}).
iana(language,zsr,tag{
         type: "language",
         subtag: "zsr",
         description: "Southern Rincon Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zsu,tag{
         type: "language",
         subtag: "zsu",
         description: "Sukurum",
         added: "2009-07-29"}).
iana(language,zte,tag{
         type: "language",
         subtag: "zte",
         description: "Elotepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztg,tag{
         type: "language",
         subtag: "ztg",
         description: "Xanaguía Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztl,tag{
         type: "language",
         subtag: "ztl",
         description: "Lapaguía-Guivini Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztm,tag{
         type: "language",
         subtag: "ztm",
         description: "San Agustín Mixtepec Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztn,tag{
         type: "language",
         subtag: "ztn",
         description: "Santa Catarina Albarradas Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztp,tag{
         type: "language",
         subtag: "ztp",
         description: "Loxicha Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztq,tag{
         type: "language",
         subtag: "ztq",
         description: "Quioquitani-Quierí Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zts,tag{
         type: "language",
         subtag: "zts",
         description: "Tilquiapan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztt,tag{
         type: "language",
         subtag: "ztt",
         description: "Tejalapan Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztu,tag{
         type: "language",
         subtag: "ztu",
         description: "Güilá Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,ztx,tag{
         type: "language",
         subtag: "ztx",
         description: "Zaachila Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zty,tag{
         type: "language",
         subtag: "zty",
         description: "Yatee Zapotec",
         added: "2009-07-29",
         macrolanguage: "zap"}).
iana(language,zua,tag{
         type: "language",
         subtag: "zua",
         description: "Zeem",
         added: "2009-07-29"}).
iana(language,zuh,tag{
         type: "language",
         subtag: "zuh",
         description: "Tokano",
         added: "2009-07-29"}).
iana(language,zum,tag{
         type: "language",
         subtag: "zum",
         description: "Kumzari",
         added: "2009-07-29"}).
iana(language,zun,tag{
         type: "language",
         subtag: "zun",
         description: "Zuni",
         added: "2005-10-16"}).
iana(language,zuy,tag{
         type: "language",
         subtag: "zuy",
         description: "Zumaya",
         added: "2009-07-29"}).
iana(language,zwa,tag{
         type: "language",
         subtag: "zwa",
         description: "Zay",
         added: "2009-07-29"}).
iana(language,zxx,tag{
         type: "language",
         subtag: "zxx",
         description: ["No linguistic content","Not applicable"],
         added: "2006-03-08",
         scope: "special"}).
iana(language,zyb,tag{
         type: "language",
         subtag: "zyb",
         description: "Yongbei Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zyg,tag{
         type: "language",
         subtag: "zyg",
         description: "Yang Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zyj,tag{
         type: "language",
         subtag: "zyj",
         description: "Youjiang Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zyn,tag{
         type: "language",
         subtag: "zyn",
         description: "Yongnan Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(language,zyp,tag{
         type: "language",
         subtag: "zyp",
         description: "Zyphe Chin",
         added: "2009-07-29"}).
iana(language,zza,tag{
         type: "language",
         subtag: "zza",
         description: ["Zaza","Dimili","Dimli (macrolanguage)","Kirdki","Kirmanjki (macrolanguage)","Zazaki"],
         added: "2006-08-24",
         scope: "macrolanguage"}).
iana(language,zzj,tag{
         type: "language",
         subtag: "zzj",
         description: "Zuojiang Zhuang",
         added: "2009-07-29",
         macrolanguage: "za"}).
iana(extlang,aao,tag{
         type: "extlang",
         subtag: "aao",
         description: "Algerian Saharan Arabic",
         added: "2009-07-29",
         preferred_value: "aao",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,abh,tag{
         type: "extlang",
         subtag: "abh",
         description: "Tajiki Arabic",
         added: "2009-07-29",
         preferred_value: "abh",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,abv,tag{
         type: "extlang",
         subtag: "abv",
         description: "Baharna Arabic",
         added: "2009-07-29",
         preferred_value: "abv",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,acm,tag{
         type: "extlang",
         subtag: "acm",
         description: "Mesopotamian Arabic",
         added: "2009-07-29",
         preferred_value: "acm",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,acq,tag{
         type: "extlang",
         subtag: "acq",
         description: "Ta'izzi-Adeni Arabic",
         added: "2009-07-29",
         preferred_value: "acq",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,acw,tag{
         type: "extlang",
         subtag: "acw",
         description: "Hijazi Arabic",
         added: "2009-07-29",
         preferred_value: "acw",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,acx,tag{
         type: "extlang",
         subtag: "acx",
         description: "Omani Arabic",
         added: "2009-07-29",
         preferred_value: "acx",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,acy,tag{
         type: "extlang",
         subtag: "acy",
         description: "Cypriot Arabic",
         added: "2009-07-29",
         preferred_value: "acy",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,adf,tag{
         type: "extlang",
         subtag: "adf",
         description: "Dhofari Arabic",
         added: "2009-07-29",
         preferred_value: "adf",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ads,tag{
         type: "extlang",
         subtag: "ads",
         description: "Adamorobe Sign Language",
         added: "2009-07-29",
         preferred_value: "ads",
         prefix: "sgn"}).
iana(extlang,aeb,tag{
         type: "extlang",
         subtag: "aeb",
         description: "Tunisian Arabic",
         added: "2009-07-29",
         preferred_value: "aeb",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,aec,tag{
         type: "extlang",
         subtag: "aec",
         description: "Saidi Arabic",
         added: "2009-07-29",
         preferred_value: "aec",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,aed,tag{
         type: "extlang",
         subtag: "aed",
         description: "Argentine Sign Language",
         added: "2009-07-29",
         preferred_value: "aed",
         prefix: "sgn"}).
iana(extlang,aen,tag{
         type: "extlang",
         subtag: "aen",
         description: "Armenian Sign Language",
         added: "2009-07-29",
         preferred_value: "aen",
         prefix: "sgn"}).
iana(extlang,afb,tag{
         type: "extlang",
         subtag: "afb",
         description: "Gulf Arabic",
         added: "2009-07-29",
         preferred_value: "afb",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,afg,tag{
         type: "extlang",
         subtag: "afg",
         description: "Afghan Sign Language",
         added: "2009-07-29",
         preferred_value: "afg",
         prefix: "sgn"}).
iana(extlang,ajp,tag{
         type: "extlang",
         subtag: "ajp",
         description: "South Levantine Arabic",
         added: "2009-07-29",
         preferred_value: "ajp",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ajs,tag{
         type: "extlang",
         subtag: "ajs",
         description: "Algerian Jewish Sign Language",
         added: "2022-02-25",
         preferred_value: "ajs",
         prefix: "sgn"}).
iana(extlang,apc,tag{
         type: "extlang",
         subtag: "apc",
         description: "North Levantine Arabic",
         added: "2009-07-29",
         preferred_value: "apc",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,apd,tag{
         type: "extlang",
         subtag: "apd",
         description: "Sudanese Arabic",
         added: "2009-07-29",
         preferred_value: "apd",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,arb,tag{
         type: "extlang",
         subtag: "arb",
         description: "Standard Arabic",
         added: "2009-07-29",
         preferred_value: "arb",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,arq,tag{
         type: "extlang",
         subtag: "arq",
         description: "Algerian Arabic",
         added: "2009-07-29",
         preferred_value: "arq",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ars,tag{
         type: "extlang",
         subtag: "ars",
         description: "Najdi Arabic",
         added: "2009-07-29",
         preferred_value: "ars",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ary,tag{
         type: "extlang",
         subtag: "ary",
         description: "Moroccan Arabic",
         added: "2009-07-29",
         preferred_value: "ary",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,arz,tag{
         type: "extlang",
         subtag: "arz",
         description: "Egyptian Arabic",
         added: "2009-07-29",
         preferred_value: "arz",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ase,tag{
         type: "extlang",
         subtag: "ase",
         description: "American Sign Language",
         added: "2009-07-29",
         preferred_value: "ase",
         prefix: "sgn"}).
iana(extlang,asf,tag{
         type: "extlang",
         subtag: "asf",
         description: ["Auslan","Australian Sign Language"],
         added: "2009-07-29",
         preferred_value: "asf",
         prefix: "sgn"}).
iana(extlang,asp,tag{
         type: "extlang",
         subtag: "asp",
         description: "Algerian Sign Language",
         added: "2009-07-29",
         preferred_value: "asp",
         prefix: "sgn"}).
iana(extlang,asq,tag{
         type: "extlang",
         subtag: "asq",
         description: "Austrian Sign Language",
         added: "2009-07-29",
         preferred_value: "asq",
         prefix: "sgn"}).
iana(extlang,asw,tag{
         type: "extlang",
         subtag: "asw",
         description: "Australian Aborigines Sign Language",
         added: "2009-07-29",
         preferred_value: "asw",
         prefix: "sgn"}).
iana(extlang,auz,tag{
         type: "extlang",
         subtag: "auz",
         description: "Uzbeki Arabic",
         added: "2009-07-29",
         preferred_value: "auz",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,avl,tag{
         type: "extlang",
         subtag: "avl",
         description: "Eastern Egyptian Bedawi Arabic",
         added: "2009-07-29",
         preferred_value: "avl",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ayh,tag{
         type: "extlang",
         subtag: "ayh",
         description: "Hadrami Arabic",
         added: "2009-07-29",
         preferred_value: "ayh",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ayl,tag{
         type: "extlang",
         subtag: "ayl",
         description: "Libyan Arabic",
         added: "2009-07-29",
         preferred_value: "ayl",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ayn,tag{
         type: "extlang",
         subtag: "ayn",
         description: "Sanaani Arabic",
         added: "2009-07-29",
         preferred_value: "ayn",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ayp,tag{
         type: "extlang",
         subtag: "ayp",
         description: "North Mesopotamian Arabic",
         added: "2009-07-29",
         preferred_value: "ayp",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,bbz,tag{
         type: "extlang",
         subtag: "bbz",
         description: "Babalia Creole Arabic",
         added: "2009-07-29",
         deprecated: "2020-03-28",
         preferred_value: "bbz",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,bfi,tag{
         type: "extlang",
         subtag: "bfi",
         description: "British Sign Language",
         added: "2009-07-29",
         preferred_value: "bfi",
         prefix: "sgn"}).
iana(extlang,bfk,tag{
         type: "extlang",
         subtag: "bfk",
         description: "Ban Khor Sign Language",
         added: "2009-07-29",
         preferred_value: "bfk",
         prefix: "sgn"}).
iana(extlang,bjn,tag{
         type: "extlang",
         subtag: "bjn",
         description: "Banjar",
         added: "2009-07-29",
         preferred_value: "bjn",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,bog,tag{
         type: "extlang",
         subtag: "bog",
         description: "Bamako Sign Language",
         added: "2009-07-29",
         preferred_value: "bog",
         prefix: "sgn"}).
iana(extlang,bqn,tag{
         type: "extlang",
         subtag: "bqn",
         description: "Bulgarian Sign Language",
         added: "2009-07-29",
         preferred_value: "bqn",
         prefix: "sgn"}).
iana(extlang,bqy,tag{
         type: "extlang",
         subtag: "bqy",
         description: "Bengkala Sign Language",
         added: "2009-07-29",
         preferred_value: "bqy",
         prefix: "sgn"}).
iana(extlang,btj,tag{
         type: "extlang",
         subtag: "btj",
         description: "Bacanese Malay",
         added: "2009-07-29",
         preferred_value: "btj",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,bve,tag{
         type: "extlang",
         subtag: "bve",
         description: "Berau Malay",
         added: "2009-07-29",
         preferred_value: "bve",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,bvl,tag{
         type: "extlang",
         subtag: "bvl",
         description: "Bolivian Sign Language",
         added: "2009-07-29",
         preferred_value: "bvl",
         prefix: "sgn"}).
iana(extlang,bvu,tag{
         type: "extlang",
         subtag: "bvu",
         description: "Bukit Malay",
         added: "2009-07-29",
         preferred_value: "bvu",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,bzs,tag{
         type: "extlang",
         subtag: "bzs",
         description: "Brazilian Sign Language",
         added: "2009-07-29",
         preferred_value: "bzs",
         prefix: "sgn"}).
iana(extlang,cdo,tag{
         type: "extlang",
         subtag: "cdo",
         description: "Min Dong Chinese",
         added: "2009-07-29",
         preferred_value: "cdo",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,cds,tag{
         type: "extlang",
         subtag: "cds",
         description: "Chadian Sign Language",
         added: "2009-07-29",
         preferred_value: "cds",
         prefix: "sgn"}).
iana(extlang,cjy,tag{
         type: "extlang",
         subtag: "cjy",
         description: "Jinyu Chinese",
         added: "2009-07-29",
         preferred_value: "cjy",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,cmn,tag{
         type: "extlang",
         subtag: "cmn",
         description: "Mandarin Chinese",
         added: "2009-07-29",
         preferred_value: "cmn",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,cnp,tag{
         type: "extlang",
         subtag: "cnp",
         description: ["Northern Ping Chinese","Northern Pinghua"],
         added: "2020-03-28",
         preferred_value: "cnp",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,coa,tag{
         type: "extlang",
         subtag: "coa",
         description: "Cocos Islands Malay",
         added: "2009-07-29",
         preferred_value: "coa",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,cpx,tag{
         type: "extlang",
         subtag: "cpx",
         description: "Pu-Xian Chinese",
         added: "2009-07-29",
         preferred_value: "cpx",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,csc,tag{
         type: "extlang",
         subtag: "csc",
         description: ["Catalan Sign Language","Lengua de señas catalana","Llengua de Signes Catalana"],
         added: "2009-07-29",
         preferred_value: "csc",
         prefix: "sgn"}).
iana(extlang,csd,tag{
         type: "extlang",
         subtag: "csd",
         description: "Chiangmai Sign Language",
         added: "2009-07-29",
         preferred_value: "csd",
         prefix: "sgn"}).
iana(extlang,cse,tag{
         type: "extlang",
         subtag: "cse",
         description: "Czech Sign Language",
         added: "2009-07-29",
         preferred_value: "cse",
         prefix: "sgn"}).
iana(extlang,csf,tag{
         type: "extlang",
         subtag: "csf",
         description: "Cuba Sign Language",
         added: "2009-07-29",
         preferred_value: "csf",
         prefix: "sgn"}).
iana(extlang,csg,tag{
         type: "extlang",
         subtag: "csg",
         description: "Chilean Sign Language",
         added: "2009-07-29",
         preferred_value: "csg",
         prefix: "sgn"}).
iana(extlang,csl,tag{
         type: "extlang",
         subtag: "csl",
         description: "Chinese Sign Language",
         added: "2009-07-29",
         preferred_value: "csl",
         prefix: "sgn"}).
iana(extlang,csn,tag{
         type: "extlang",
         subtag: "csn",
         description: "Colombian Sign Language",
         added: "2009-07-29",
         preferred_value: "csn",
         prefix: "sgn"}).
iana(extlang,csp,tag{
         type: "extlang",
         subtag: "csp",
         description: ["Southern Ping Chinese","Southern Pinghua"],
         added: "2020-03-28",
         preferred_value: "csp",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,csq,tag{
         type: "extlang",
         subtag: "csq",
         description: "Croatia Sign Language",
         added: "2009-07-29",
         preferred_value: "csq",
         prefix: "sgn"}).
iana(extlang,csr,tag{
         type: "extlang",
         subtag: "csr",
         description: "Costa Rican Sign Language",
         added: "2009-07-29",
         preferred_value: "csr",
         prefix: "sgn"}).
iana(extlang,csx,tag{
         type: "extlang",
         subtag: "csx",
         description: "Cambodian Sign Language",
         added: "2021-02-20",
         preferred_value: "csx",
         prefix: "sgn"}).
iana(extlang,czh,tag{
         type: "extlang",
         subtag: "czh",
         description: "Huizhou Chinese",
         added: "2009-07-29",
         preferred_value: "czh",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,czo,tag{
         type: "extlang",
         subtag: "czo",
         description: "Min Zhong Chinese",
         added: "2009-07-29",
         preferred_value: "czo",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,doq,tag{
         type: "extlang",
         subtag: "doq",
         description: "Dominican Sign Language",
         added: "2009-07-29",
         preferred_value: "doq",
         prefix: "sgn"}).
iana(extlang,dse,tag{
         type: "extlang",
         subtag: "dse",
         description: "Dutch Sign Language",
         added: "2009-07-29",
         preferred_value: "dse",
         prefix: "sgn"}).
iana(extlang,dsl,tag{
         type: "extlang",
         subtag: "dsl",
         description: "Danish Sign Language",
         added: "2009-07-29",
         preferred_value: "dsl",
         prefix: "sgn"}).
iana(extlang,dsz,tag{
         type: "extlang",
         subtag: "dsz",
         description: "Mardin Sign Language",
         added: "2022-02-25",
         preferred_value: "dsz",
         prefix: "sgn"}).
iana(extlang,dup,tag{
         type: "extlang",
         subtag: "dup",
         description: "Duano",
         added: "2009-07-29",
         preferred_value: "dup",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,ecs,tag{
         type: "extlang",
         subtag: "ecs",
         description: "Ecuadorian Sign Language",
         added: "2009-07-29",
         preferred_value: "ecs",
         prefix: "sgn"}).
iana(extlang,ehs,tag{
         type: "extlang",
         subtag: "ehs",
         description: "Miyakubo Sign Language",
         added: "2021-02-20",
         preferred_value: "ehs",
         prefix: "sgn"}).
iana(extlang,esl,tag{
         type: "extlang",
         subtag: "esl",
         description: "Egypt Sign Language",
         added: "2009-07-29",
         preferred_value: "esl",
         prefix: "sgn"}).
iana(extlang,esn,tag{
         type: "extlang",
         subtag: "esn",
         description: "Salvadoran Sign Language",
         added: "2009-07-29",
         preferred_value: "esn",
         prefix: "sgn"}).
iana(extlang,eso,tag{
         type: "extlang",
         subtag: "eso",
         description: "Estonian Sign Language",
         added: "2009-07-29",
         preferred_value: "eso",
         prefix: "sgn"}).
iana(extlang,eth,tag{
         type: "extlang",
         subtag: "eth",
         description: "Ethiopian Sign Language",
         added: "2009-07-29",
         preferred_value: "eth",
         prefix: "sgn"}).
iana(extlang,fcs,tag{
         type: "extlang",
         subtag: "fcs",
         description: "Quebec Sign Language",
         added: "2009-07-29",
         preferred_value: "fcs",
         prefix: "sgn"}).
iana(extlang,fse,tag{
         type: "extlang",
         subtag: "fse",
         description: "Finnish Sign Language",
         added: "2009-07-29",
         preferred_value: "fse",
         prefix: "sgn"}).
iana(extlang,fsl,tag{
         type: "extlang",
         subtag: "fsl",
         description: "French Sign Language",
         added: "2009-07-29",
         preferred_value: "fsl",
         prefix: "sgn"}).
iana(extlang,fss,tag{
         type: "extlang",
         subtag: "fss",
         description: ["Finland-Swedish Sign Language","finlandssvenskt teckenspråk","suomenruotsalainen viittomakieli"],
         added: "2009-07-29",
         preferred_value: "fss",
         prefix: "sgn"}).
iana(extlang,gan,tag{
         type: "extlang",
         subtag: "gan",
         description: "Gan Chinese",
         added: "2009-07-29",
         preferred_value: "gan",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,gds,tag{
         type: "extlang",
         subtag: "gds",
         description: "Ghandruk Sign Language",
         added: "2012-08-12",
         preferred_value: "gds",
         prefix: "sgn"}).
iana(extlang,gom,tag{
         type: "extlang",
         subtag: "gom",
         description: "Goan Konkani",
         added: "2009-07-29",
         preferred_value: "gom",
         prefix: "kok",
         macrolanguage: "kok"}).
iana(extlang,gse,tag{
         type: "extlang",
         subtag: "gse",
         description: "Ghanaian Sign Language",
         added: "2009-07-29",
         preferred_value: "gse",
         prefix: "sgn"}).
iana(extlang,gsg,tag{
         type: "extlang",
         subtag: "gsg",
         description: "German Sign Language",
         added: "2009-07-29",
         preferred_value: "gsg",
         prefix: "sgn"}).
iana(extlang,gsm,tag{
         type: "extlang",
         subtag: "gsm",
         description: "Guatemalan Sign Language",
         added: "2009-07-29",
         preferred_value: "gsm",
         prefix: "sgn"}).
iana(extlang,gss,tag{
         type: "extlang",
         subtag: "gss",
         description: "Greek Sign Language",
         added: "2009-07-29",
         preferred_value: "gss",
         prefix: "sgn"}).
iana(extlang,gus,tag{
         type: "extlang",
         subtag: "gus",
         description: "Guinean Sign Language",
         added: "2009-07-29",
         preferred_value: "gus",
         prefix: "sgn"}).
iana(extlang,hab,tag{
         type: "extlang",
         subtag: "hab",
         description: "Hanoi Sign Language",
         added: "2009-07-29",
         preferred_value: "hab",
         prefix: "sgn"}).
iana(extlang,haf,tag{
         type: "extlang",
         subtag: "haf",
         description: "Haiphong Sign Language",
         added: "2009-07-29",
         preferred_value: "haf",
         prefix: "sgn"}).
iana(extlang,hak,tag{
         type: "extlang",
         subtag: "hak",
         description: "Hakka Chinese",
         added: "2009-07-29",
         preferred_value: "hak",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,hds,tag{
         type: "extlang",
         subtag: "hds",
         description: "Honduras Sign Language",
         added: "2009-07-29",
         preferred_value: "hds",
         prefix: "sgn"}).
iana(extlang,hji,tag{
         type: "extlang",
         subtag: "hji",
         description: "Haji",
         added: "2009-07-29",
         preferred_value: "hji",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,hks,tag{
         type: "extlang",
         subtag: "hks",
         description: ["Hong Kong Sign Language","Heung Kong Sau Yue"],
         added: "2009-07-29",
         preferred_value: "hks",
         prefix: "sgn"}).
iana(extlang,hos,tag{
         type: "extlang",
         subtag: "hos",
         description: "Ho Chi Minh City Sign Language",
         added: "2009-07-29",
         preferred_value: "hos",
         prefix: "sgn"}).
iana(extlang,hps,tag{
         type: "extlang",
         subtag: "hps",
         description: ["Hawai'i Sign Language (HSL)","Hawai'i Pidgin Sign Language"],
         added: "2009-07-29",
         preferred_value: "hps",
         prefix: "sgn"}).
iana(extlang,hsh,tag{
         type: "extlang",
         subtag: "hsh",
         description: "Hungarian Sign Language",
         added: "2009-07-29",
         preferred_value: "hsh",
         prefix: "sgn"}).
iana(extlang,hsl,tag{
         type: "extlang",
         subtag: "hsl",
         description: "Hausa Sign Language",
         added: "2009-07-29",
         preferred_value: "hsl",
         prefix: "sgn"}).
iana(extlang,hsn,tag{
         type: "extlang",
         subtag: "hsn",
         description: "Xiang Chinese",
         added: "2009-07-29",
         preferred_value: "hsn",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,icl,tag{
         type: "extlang",
         subtag: "icl",
         description: "Icelandic Sign Language",
         added: "2009-07-29",
         preferred_value: "icl",
         prefix: "sgn"}).
iana(extlang,iks,tag{
         type: "extlang",
         subtag: "iks",
         description: "Inuit Sign Language",
         added: "2015-02-12",
         preferred_value: "iks",
         prefix: "sgn"}).
iana(extlang,ils,tag{
         type: "extlang",
         subtag: "ils",
         description: "International Sign",
         added: "2009-07-29",
         preferred_value: "ils",
         prefix: "sgn"}).
iana(extlang,inl,tag{
         type: "extlang",
         subtag: "inl",
         description: "Indonesian Sign Language",
         added: "2009-07-29",
         preferred_value: "inl",
         prefix: "sgn"}).
iana(extlang,ins,tag{
         type: "extlang",
         subtag: "ins",
         description: "Indian Sign Language",
         added: "2009-07-29",
         preferred_value: "ins",
         prefix: "sgn"}).
iana(extlang,ise,tag{
         type: "extlang",
         subtag: "ise",
         description: "Italian Sign Language",
         added: "2009-07-29",
         preferred_value: "ise",
         prefix: "sgn"}).
iana(extlang,isg,tag{
         type: "extlang",
         subtag: "isg",
         description: "Irish Sign Language",
         added: "2009-07-29",
         preferred_value: "isg",
         prefix: "sgn"}).
iana(extlang,isr,tag{
         type: "extlang",
         subtag: "isr",
         description: "Israeli Sign Language",
         added: "2009-07-29",
         preferred_value: "isr",
         prefix: "sgn"}).
iana(extlang,jak,tag{
         type: "extlang",
         subtag: "jak",
         description: "Jakun",
         added: "2009-07-29",
         preferred_value: "jak",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,jax,tag{
         type: "extlang",
         subtag: "jax",
         description: "Jambi Malay",
         added: "2009-07-29",
         preferred_value: "jax",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,jcs,tag{
         type: "extlang",
         subtag: "jcs",
         description: "Jamaican Country Sign Language",
         added: "2009-07-29",
         preferred_value: "jcs",
         prefix: "sgn"}).
iana(extlang,jhs,tag{
         type: "extlang",
         subtag: "jhs",
         description: "Jhankot Sign Language",
         added: "2009-07-29",
         preferred_value: "jhs",
         prefix: "sgn"}).
iana(extlang,jks,tag{
         type: "extlang",
         subtag: "jks",
         description: "Amami Koniya Sign Language",
         added: "2021-02-20",
         preferred_value: "jks",
         prefix: "sgn"}).
iana(extlang,jls,tag{
         type: "extlang",
         subtag: "jls",
         description: "Jamaican Sign Language",
         added: "2010-03-11",
         preferred_value: "jls",
         prefix: "sgn"}).
iana(extlang,jos,tag{
         type: "extlang",
         subtag: "jos",
         description: "Jordanian Sign Language",
         added: "2009-07-29",
         preferred_value: "jos",
         prefix: "sgn"}).
iana(extlang,jsl,tag{
         type: "extlang",
         subtag: "jsl",
         description: "Japanese Sign Language",
         added: "2009-07-29",
         preferred_value: "jsl",
         prefix: "sgn"}).
iana(extlang,jus,tag{
         type: "extlang",
         subtag: "jus",
         description: "Jumla Sign Language",
         added: "2009-07-29",
         preferred_value: "jus",
         prefix: "sgn"}).
iana(extlang,kgi,tag{
         type: "extlang",
         subtag: "kgi",
         description: "Selangor Sign Language",
         added: "2009-07-29",
         preferred_value: "kgi",
         prefix: "sgn"}).
iana(extlang,knn,tag{
         type: "extlang",
         subtag: "knn",
         description: "Konkani (individual language)",
         added: "2009-07-29",
         preferred_value: "knn",
         prefix: "kok",
         macrolanguage: "kok"}).
iana(extlang,kvb,tag{
         type: "extlang",
         subtag: "kvb",
         description: "Kubu",
         added: "2009-07-29",
         preferred_value: "kvb",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,kvk,tag{
         type: "extlang",
         subtag: "kvk",
         description: "Korean Sign Language",
         added: "2009-07-29",
         preferred_value: "kvk",
         prefix: "sgn"}).
iana(extlang,kvr,tag{
         type: "extlang",
         subtag: "kvr",
         description: "Kerinci",
         added: "2009-07-29",
         preferred_value: "kvr",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,kxd,tag{
         type: "extlang",
         subtag: "kxd",
         description: "Brunei",
         added: "2009-07-29",
         preferred_value: "kxd",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,lbs,tag{
         type: "extlang",
         subtag: "lbs",
         description: "Libyan Sign Language",
         added: "2009-07-29",
         preferred_value: "lbs",
         prefix: "sgn"}).
iana(extlang,lce,tag{
         type: "extlang",
         subtag: "lce",
         description: ["Loncong","Sekak"],
         added: "2009-07-29",
         preferred_value: "lce",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,lcf,tag{
         type: "extlang",
         subtag: "lcf",
         description: "Lubu",
         added: "2009-07-29",
         preferred_value: "lcf",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,liw,tag{
         type: "extlang",
         subtag: "liw",
         description: "Col",
         added: "2009-07-29",
         preferred_value: "liw",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,lls,tag{
         type: "extlang",
         subtag: "lls",
         description: "Lithuanian Sign Language",
         added: "2009-07-29",
         preferred_value: "lls",
         prefix: "sgn"}).
iana(extlang,lsb,tag{
         type: "extlang",
         subtag: "lsb",
         description: ["Burundian Sign Language","Langue des Signes Burundaise"],
         added: "2021-02-20",
         preferred_value: "lsb",
         prefix: "sgn"}).
iana(extlang,lsc,tag{
         type: "extlang",
         subtag: "lsc",
         description: ["Albarradas Sign Language","Lengua de señas Albarradas"],
         added: "2022-02-25",
         preferred_value: "lsc",
         prefix: "sgn"}).
iana(extlang,lsg,tag{
         type: "extlang",
         subtag: "lsg",
         description: "Lyons Sign Language",
         added: "2009-07-29",
         deprecated: "2018-03-08",
         preferred_value: "lsg",
         prefix: "sgn"}).
iana(extlang,lsl,tag{
         type: "extlang",
         subtag: "lsl",
         description: "Latvian Sign Language",
         added: "2009-07-29",
         preferred_value: "lsl",
         prefix: "sgn"}).
iana(extlang,lsn,tag{
         type: "extlang",
         subtag: "lsn",
         description: "Tibetan Sign Language",
         added: "2019-04-16",
         preferred_value: "lsn",
         prefix: "sgn"}).
iana(extlang,lso,tag{
         type: "extlang",
         subtag: "lso",
         description: "Laos Sign Language",
         added: "2009-07-29",
         preferred_value: "lso",
         prefix: "sgn"}).
iana(extlang,lsp,tag{
         type: "extlang",
         subtag: "lsp",
         description: ["Panamanian Sign Language","Lengua de Señas Panameñas"],
         added: "2009-07-29",
         preferred_value: "lsp",
         prefix: "sgn"}).
iana(extlang,lst,tag{
         type: "extlang",
         subtag: "lst",
         description: "Trinidad and Tobago Sign Language",
         added: "2009-07-29",
         preferred_value: "lst",
         prefix: "sgn"}).
iana(extlang,lsv,tag{
         type: "extlang",
         subtag: "lsv",
         description: "Sivia Sign Language",
         added: "2019-04-16",
         preferred_value: "lsv",
         prefix: "sgn"}).
iana(extlang,lsw,tag{
         type: "extlang",
         subtag: "lsw",
         description: ["Seychelles Sign Language","Lalang Siny Seselwa","Langue des Signes Seychelloise"],
         added: "2022-02-25",
         preferred_value: "lsw",
         prefix: "sgn"}).
iana(extlang,lsy,tag{
         type: "extlang",
         subtag: "lsy",
         description: "Mauritian Sign Language",
         added: "2010-03-11",
         preferred_value: "lsy",
         prefix: "sgn"}).
iana(extlang,ltg,tag{
         type: "extlang",
         subtag: "ltg",
         description: "Latgalian",
         added: "2010-03-11",
         preferred_value: "ltg",
         prefix: "lv",
         macrolanguage: "lv"}).
iana(extlang,lvs,tag{
         type: "extlang",
         subtag: "lvs",
         description: "Standard Latvian",
         added: "2010-03-11",
         preferred_value: "lvs",
         prefix: "lv",
         macrolanguage: "lv"}).
iana(extlang,lws,tag{
         type: "extlang",
         subtag: "lws",
         description: "Malawian Sign Language",
         added: "2018-03-08",
         preferred_value: "lws",
         prefix: "sgn"}).
iana(extlang,lzh,tag{
         type: "extlang",
         subtag: "lzh",
         description: "Literary Chinese",
         added: "2009-07-29",
         preferred_value: "lzh",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,max,tag{
         type: "extlang",
         subtag: "max",
         description: "North Moluccan Malay",
         added: "2009-07-29",
         preferred_value: "max",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mdl,tag{
         type: "extlang",
         subtag: "mdl",
         description: "Maltese Sign Language",
         added: "2009-07-29",
         preferred_value: "mdl",
         prefix: "sgn"}).
iana(extlang,meo,tag{
         type: "extlang",
         subtag: "meo",
         description: "Kedah Malay",
         added: "2009-07-29",
         preferred_value: "meo",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mfa,tag{
         type: "extlang",
         subtag: "mfa",
         description: "Pattani Malay",
         added: "2009-07-29",
         preferred_value: "mfa",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mfb,tag{
         type: "extlang",
         subtag: "mfb",
         description: "Bangka",
         added: "2009-07-29",
         preferred_value: "mfb",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mfs,tag{
         type: "extlang",
         subtag: "mfs",
         description: "Mexican Sign Language",
         added: "2009-07-29",
         preferred_value: "mfs",
         prefix: "sgn"}).
iana(extlang,min,tag{
         type: "extlang",
         subtag: "min",
         description: "Minangkabau",
         added: "2009-07-29",
         preferred_value: "min",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mnp,tag{
         type: "extlang",
         subtag: "mnp",
         description: "Min Bei Chinese",
         added: "2009-07-29",
         preferred_value: "mnp",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,mqg,tag{
         type: "extlang",
         subtag: "mqg",
         description: "Kota Bangun Kutai Malay",
         added: "2009-07-29",
         preferred_value: "mqg",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mre,tag{
         type: "extlang",
         subtag: "mre",
         description: "Martha's Vineyard Sign Language",
         added: "2009-07-29",
         preferred_value: "mre",
         prefix: "sgn"}).
iana(extlang,msd,tag{
         type: "extlang",
         subtag: "msd",
         description: "Yucatec Maya Sign Language",
         added: "2009-07-29",
         preferred_value: "msd",
         prefix: "sgn"}).
iana(extlang,msi,tag{
         type: "extlang",
         subtag: "msi",
         description: "Sabah Malay",
         added: "2009-07-29",
         preferred_value: "msi",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,msr,tag{
         type: "extlang",
         subtag: "msr",
         description: "Mongolian Sign Language",
         added: "2009-07-29",
         preferred_value: "msr",
         prefix: "sgn"}).
iana(extlang,mui,tag{
         type: "extlang",
         subtag: "mui",
         description: "Musi",
         added: "2009-07-29",
         preferred_value: "mui",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,mzc,tag{
         type: "extlang",
         subtag: "mzc",
         description: "Madagascar Sign Language",
         added: "2009-07-29",
         preferred_value: "mzc",
         prefix: "sgn"}).
iana(extlang,mzg,tag{
         type: "extlang",
         subtag: "mzg",
         description: "Monastic Sign Language",
         added: "2009-07-29",
         preferred_value: "mzg",
         prefix: "sgn"}).
iana(extlang,mzy,tag{
         type: "extlang",
         subtag: "mzy",
         description: "Mozambican Sign Language",
         added: "2009-07-29",
         preferred_value: "mzy",
         prefix: "sgn"}).
iana(extlang,nan,tag{
         type: "extlang",
         subtag: "nan",
         description: "Min Nan Chinese",
         added: "2009-07-29",
         preferred_value: "nan",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,nbs,tag{
         type: "extlang",
         subtag: "nbs",
         description: "Namibian Sign Language",
         added: "2009-07-29",
         preferred_value: "nbs",
         prefix: "sgn"}).
iana(extlang,ncs,tag{
         type: "extlang",
         subtag: "ncs",
         description: "Nicaraguan Sign Language",
         added: "2009-07-29",
         preferred_value: "ncs",
         prefix: "sgn"}).
iana(extlang,nsi,tag{
         type: "extlang",
         subtag: "nsi",
         description: "Nigerian Sign Language",
         added: "2009-07-29",
         preferred_value: "nsi",
         prefix: "sgn"}).
iana(extlang,nsl,tag{
         type: "extlang",
         subtag: "nsl",
         description: "Norwegian Sign Language",
         added: "2009-07-29",
         preferred_value: "nsl",
         prefix: "sgn"}).
iana(extlang,nsp,tag{
         type: "extlang",
         subtag: "nsp",
         description: "Nepalese Sign Language",
         added: "2009-07-29",
         preferred_value: "nsp",
         prefix: "sgn"}).
iana(extlang,nsr,tag{
         type: "extlang",
         subtag: "nsr",
         description: "Maritime Sign Language",
         added: "2009-07-29",
         preferred_value: "nsr",
         prefix: "sgn"}).
iana(extlang,nzs,tag{
         type: "extlang",
         subtag: "nzs",
         description: "New Zealand Sign Language",
         added: "2009-07-29",
         preferred_value: "nzs",
         prefix: "sgn"}).
iana(extlang,okl,tag{
         type: "extlang",
         subtag: "okl",
         description: "Old Kentish Sign Language",
         added: "2009-07-29",
         preferred_value: "okl",
         prefix: "sgn"}).
iana(extlang,orn,tag{
         type: "extlang",
         subtag: "orn",
         description: "Orang Kanaq",
         added: "2009-07-29",
         preferred_value: "orn",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,ors,tag{
         type: "extlang",
         subtag: "ors",
         description: "Orang Seletar",
         added: "2009-07-29",
         preferred_value: "ors",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,pel,tag{
         type: "extlang",
         subtag: "pel",
         description: "Pekal",
         added: "2009-07-29",
         preferred_value: "pel",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,pga,tag{
         type: "extlang",
         subtag: "pga",
         description: "Sudanese Creole Arabic",
         added: "2009-07-29",
         preferred_value: "pga",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,pgz,tag{
         type: "extlang",
         subtag: "pgz",
         description: "Papua New Guinean Sign Language",
         added: "2016-05-30",
         preferred_value: "pgz",
         prefix: "sgn"}).
iana(extlang,pks,tag{
         type: "extlang",
         subtag: "pks",
         description: "Pakistan Sign Language",
         added: "2009-07-29",
         preferred_value: "pks",
         prefix: "sgn"}).
iana(extlang,prl,tag{
         type: "extlang",
         subtag: "prl",
         description: "Peruvian Sign Language",
         added: "2009-07-29",
         preferred_value: "prl",
         prefix: "sgn"}).
iana(extlang,prz,tag{
         type: "extlang",
         subtag: "prz",
         description: "Providencia Sign Language",
         added: "2009-07-29",
         preferred_value: "prz",
         prefix: "sgn"}).
iana(extlang,psc,tag{
         type: "extlang",
         subtag: "psc",
         description: ["Iranian Sign Language","Persian Sign Language"],
         added: "2009-07-29",
         preferred_value: "psc",
         prefix: "sgn"}).
iana(extlang,psd,tag{
         type: "extlang",
         subtag: "psd",
         description: "Plains Indian Sign Language",
         added: "2009-07-29",
         preferred_value: "psd",
         prefix: "sgn"}).
iana(extlang,pse,tag{
         type: "extlang",
         subtag: "pse",
         description: "Central Malay",
         added: "2009-07-29",
         preferred_value: "pse",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,psg,tag{
         type: "extlang",
         subtag: "psg",
         description: "Penang Sign Language",
         added: "2009-07-29",
         preferred_value: "psg",
         prefix: "sgn"}).
iana(extlang,psl,tag{
         type: "extlang",
         subtag: "psl",
         description: "Puerto Rican Sign Language",
         added: "2009-07-29",
         preferred_value: "psl",
         prefix: "sgn"}).
iana(extlang,pso,tag{
         type: "extlang",
         subtag: "pso",
         description: "Polish Sign Language",
         added: "2009-07-29",
         preferred_value: "pso",
         prefix: "sgn"}).
iana(extlang,psp,tag{
         type: "extlang",
         subtag: "psp",
         description: "Philippine Sign Language",
         added: "2009-07-29",
         preferred_value: "psp",
         prefix: "sgn"}).
iana(extlang,psr,tag{
         type: "extlang",
         subtag: "psr",
         description: "Portuguese Sign Language",
         added: "2009-07-29",
         preferred_value: "psr",
         prefix: "sgn"}).
iana(extlang,pys,tag{
         type: "extlang",
         subtag: "pys",
         description: ["Paraguayan Sign Language","Lengua de Señas del Paraguay"],
         added: "2010-03-11",
         preferred_value: "pys",
         prefix: "sgn"}).
iana(extlang,rib,tag{
         type: "extlang",
         subtag: "rib",
         description: "Bribri Sign Language",
         added: "2022-02-25",
         preferred_value: "rib",
         prefix: "sgn"}).
iana(extlang,rms,tag{
         type: "extlang",
         subtag: "rms",
         description: "Romanian Sign Language",
         added: "2009-07-29",
         preferred_value: "rms",
         prefix: "sgn"}).
iana(extlang,rnb,tag{
         type: "extlang",
         subtag: "rnb",
         description: "Brunca Sign Language",
         added: "2022-02-25",
         preferred_value: "rnb",
         prefix: "sgn"}).
iana(extlang,rsi,tag{
         type: "extlang",
         subtag: "rsi",
         description: "Rennellese Sign Language",
         added: "2009-07-29",
         deprecated: "2017-02-23",
         preferred_value: "rsi",
         prefix: "sgn"}).
iana(extlang,rsl,tag{
         type: "extlang",
         subtag: "rsl",
         description: "Russian Sign Language",
         added: "2009-07-29",
         preferred_value: "rsl",
         prefix: "sgn"}).
iana(extlang,rsm,tag{
         type: "extlang",
         subtag: "rsm",
         description: "Miriwoong Sign Language",
         added: "2016-05-30",
         preferred_value: "rsm",
         prefix: "sgn"}).
iana(extlang,rsn,tag{
         type: "extlang",
         subtag: "rsn",
         description: "Rwandan Sign Language",
         added: "2022-02-25",
         preferred_value: "rsn",
         prefix: "sgn"}).
iana(extlang,sdl,tag{
         type: "extlang",
         subtag: "sdl",
         description: "Saudi Arabian Sign Language",
         added: "2009-07-29",
         preferred_value: "sdl",
         prefix: "sgn"}).
iana(extlang,sfb,tag{
         type: "extlang",
         subtag: "sfb",
         description: ["Langue des signes de Belgique Francophone","French Belgian Sign Language"],
         added: "2009-07-29",
         preferred_value: "sfb",
         prefix: "sgn"}).
iana(extlang,sfs,tag{
         type: "extlang",
         subtag: "sfs",
         description: "South African Sign Language",
         added: "2009-07-29",
         preferred_value: "sfs",
         prefix: "sgn"}).
iana(extlang,sgg,tag{
         type: "extlang",
         subtag: "sgg",
         description: "Swiss-German Sign Language",
         added: "2009-07-29",
         preferred_value: "sgg",
         prefix: "sgn"}).
iana(extlang,sgx,tag{
         type: "extlang",
         subtag: "sgx",
         description: "Sierra Leone Sign Language",
         added: "2009-07-29",
         preferred_value: "sgx",
         prefix: "sgn"}).
iana(extlang,shu,tag{
         type: "extlang",
         subtag: "shu",
         description: "Chadian Arabic",
         added: "2009-07-29",
         preferred_value: "shu",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,slf,tag{
         type: "extlang",
         subtag: "slf",
         description: "Swiss-Italian Sign Language",
         added: "2009-07-29",
         preferred_value: "slf",
         prefix: "sgn"}).
iana(extlang,sls,tag{
         type: "extlang",
         subtag: "sls",
         description: "Singapore Sign Language",
         added: "2009-07-29",
         preferred_value: "sls",
         prefix: "sgn"}).
iana(extlang,sqk,tag{
         type: "extlang",
         subtag: "sqk",
         description: "Albanian Sign Language",
         added: "2012-08-12",
         preferred_value: "sqk",
         prefix: "sgn"}).
iana(extlang,sqs,tag{
         type: "extlang",
         subtag: "sqs",
         description: "Sri Lankan Sign Language",
         added: "2009-07-29",
         preferred_value: "sqs",
         prefix: "sgn"}).
iana(extlang,sqx,tag{
         type: "extlang",
         subtag: "sqx",
         description: "Kufr Qassem Sign Language (KQSL)",
         added: "2021-02-20",
         preferred_value: "sqx",
         prefix: "sgn"}).
iana(extlang,ssh,tag{
         type: "extlang",
         subtag: "ssh",
         description: "Shihhi Arabic",
         added: "2009-07-29",
         preferred_value: "ssh",
         prefix: "ar",
         macrolanguage: "ar"}).
iana(extlang,ssp,tag{
         type: "extlang",
         subtag: "ssp",
         description: "Spanish Sign Language",
         added: "2009-07-29",
         preferred_value: "ssp",
         prefix: "sgn"}).
iana(extlang,ssr,tag{
         type: "extlang",
         subtag: "ssr",
         description: "Swiss-French Sign Language",
         added: "2009-07-29",
         preferred_value: "ssr",
         prefix: "sgn"}).
iana(extlang,svk,tag{
         type: "extlang",
         subtag: "svk",
         description: "Slovakian Sign Language",
         added: "2009-07-29",
         preferred_value: "svk",
         prefix: "sgn"}).
iana(extlang,swc,tag{
         type: "extlang",
         subtag: "swc",
         description: "Congo Swahili",
         added: "2009-07-29",
         preferred_value: "swc",
         prefix: "sw",
         macrolanguage: "sw"}).
iana(extlang,swh,tag{
         type: "extlang",
         subtag: "swh",
         description: ["Swahili (individual language)","Kiswahili"],
         added: "2009-07-29",
         preferred_value: "swh",
         prefix: "sw",
         macrolanguage: "sw"}).
iana(extlang,swl,tag{
         type: "extlang",
         subtag: "swl",
         description: "Swedish Sign Language",
         added: "2009-07-29",
         preferred_value: "swl",
         prefix: "sgn"}).
iana(extlang,syy,tag{
         type: "extlang",
         subtag: "syy",
         description: "Al-Sayyid Bedouin Sign Language",
         added: "2009-07-29",
         preferred_value: "syy",
         prefix: "sgn"}).
iana(extlang,szs,tag{
         type: "extlang",
         subtag: "szs",
         description: "Solomon Islands Sign Language",
         added: "2017-02-23",
         preferred_value: "szs",
         prefix: "sgn"}).
iana(extlang,tmw,tag{
         type: "extlang",
         subtag: "tmw",
         description: "Temuan",
         added: "2009-07-29",
         preferred_value: "tmw",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,tse,tag{
         type: "extlang",
         subtag: "tse",
         description: "Tunisian Sign Language",
         added: "2009-07-29",
         preferred_value: "tse",
         prefix: "sgn"}).
iana(extlang,tsm,tag{
         type: "extlang",
         subtag: "tsm",
         description: ["Turkish Sign Language","Türk İşaret Dili"],
         added: "2009-07-29",
         preferred_value: "tsm",
         prefix: "sgn"}).
iana(extlang,tsq,tag{
         type: "extlang",
         subtag: "tsq",
         description: "Thai Sign Language",
         added: "2009-07-29",
         preferred_value: "tsq",
         prefix: "sgn"}).
iana(extlang,tss,tag{
         type: "extlang",
         subtag: "tss",
         description: "Taiwan Sign Language",
         added: "2009-07-29",
         preferred_value: "tss",
         prefix: "sgn"}).
iana(extlang,tsy,tag{
         type: "extlang",
         subtag: "tsy",
         description: "Tebul Sign Language",
         added: "2009-07-29",
         preferred_value: "tsy",
         prefix: "sgn"}).
iana(extlang,tza,tag{
         type: "extlang",
         subtag: "tza",
         description: "Tanzanian Sign Language",
         added: "2009-07-29",
         preferred_value: "tza",
         prefix: "sgn"}).
iana(extlang,ugn,tag{
         type: "extlang",
         subtag: "ugn",
         description: "Ugandan Sign Language",
         added: "2009-07-29",
         preferred_value: "ugn",
         prefix: "sgn"}).
iana(extlang,ugy,tag{
         type: "extlang",
         subtag: "ugy",
         description: "Uruguayan Sign Language",
         added: "2009-07-29",
         preferred_value: "ugy",
         prefix: "sgn"}).
iana(extlang,ukl,tag{
         type: "extlang",
         subtag: "ukl",
         description: "Ukrainian Sign Language",
         added: "2009-07-29",
         preferred_value: "ukl",
         prefix: "sgn"}).
iana(extlang,uks,tag{
         type: "extlang",
         subtag: "uks",
         description: ["Urubú-Kaapor Sign Language","Kaapor Sign Language"],
         added: "2009-07-29",
         preferred_value: "uks",
         prefix: "sgn"}).
iana(extlang,urk,tag{
         type: "extlang",
         subtag: "urk",
         description: "Urak Lawoi'",
         added: "2009-07-29",
         preferred_value: "urk",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,uzn,tag{
         type: "extlang",
         subtag: "uzn",
         description: "Northern Uzbek",
         added: "2009-07-29",
         preferred_value: "uzn",
         prefix: "uz",
         macrolanguage: "uz"}).
iana(extlang,uzs,tag{
         type: "extlang",
         subtag: "uzs",
         description: "Southern Uzbek",
         added: "2009-07-29",
         preferred_value: "uzs",
         prefix: "uz",
         macrolanguage: "uz"}).
iana(extlang,vgt,tag{
         type: "extlang",
         subtag: "vgt",
         description: ["Vlaamse Gebarentaal","Flemish Sign Language"],
         added: "2009-07-29",
         preferred_value: "vgt",
         prefix: "sgn"}).
iana(extlang,vkk,tag{
         type: "extlang",
         subtag: "vkk",
         description: "Kaur",
         added: "2009-07-29",
         preferred_value: "vkk",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,vkt,tag{
         type: "extlang",
         subtag: "vkt",
         description: "Tenggarong Kutai Malay",
         added: "2009-07-29",
         preferred_value: "vkt",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,vsi,tag{
         type: "extlang",
         subtag: "vsi",
         description: "Moldova Sign Language",
         added: "2009-07-29",
         preferred_value: "vsi",
         prefix: "sgn"}).
iana(extlang,vsl,tag{
         type: "extlang",
         subtag: "vsl",
         description: "Venezuelan Sign Language",
         added: "2009-07-29",
         preferred_value: "vsl",
         prefix: "sgn"}).
iana(extlang,vsv,tag{
         type: "extlang",
         subtag: "vsv",
         description: ["Valencian Sign Language","Llengua de signes valenciana"],
         added: "2009-07-29",
         preferred_value: "vsv",
         prefix: "sgn"}).
iana(extlang,wbs,tag{
         type: "extlang",
         subtag: "wbs",
         description: "West Bengal Sign Language",
         added: "2017-02-23",
         preferred_value: "wbs",
         prefix: "sgn"}).
iana(extlang,wuu,tag{
         type: "extlang",
         subtag: "wuu",
         description: "Wu Chinese",
         added: "2009-07-29",
         preferred_value: "wuu",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,xki,tag{
         type: "extlang",
         subtag: "xki",
         description: "Kenyan Sign Language",
         added: "2009-07-29",
         preferred_value: "xki",
         prefix: "sgn"}).
iana(extlang,xml,tag{
         type: "extlang",
         subtag: "xml",
         description: "Malaysian Sign Language",
         added: "2009-07-29",
         preferred_value: "xml",
         prefix: "sgn"}).
iana(extlang,xmm,tag{
         type: "extlang",
         subtag: "xmm",
         description: "Manado Malay",
         added: "2009-07-29",
         preferred_value: "xmm",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,xms,tag{
         type: "extlang",
         subtag: "xms",
         description: "Moroccan Sign Language",
         added: "2009-07-29",
         preferred_value: "xms",
         prefix: "sgn"}).
iana(extlang,yds,tag{
         type: "extlang",
         subtag: "yds",
         description: "Yiddish Sign Language",
         added: "2009-07-29",
         deprecated: "2015-02-12",
         preferred_value: "yds",
         prefix: "sgn"}).
iana(extlang,ygs,tag{
         type: "extlang",
         subtag: "ygs",
         description: "Yolŋu Sign Language",
         added: "2014-02-28",
         preferred_value: "ygs",
         prefix: "sgn"}).
iana(extlang,yhs,tag{
         type: "extlang",
         subtag: "yhs",
         description: "Yan-nhaŋu Sign Language",
         added: "2015-04-17",
         preferred_value: "yhs",
         prefix: "sgn"}).
iana(extlang,ysl,tag{
         type: "extlang",
         subtag: "ysl",
         description: "Yugoslavian Sign Language",
         added: "2009-07-29",
         preferred_value: "ysl",
         prefix: "sgn"}).
iana(extlang,ysm,tag{
         type: "extlang",
         subtag: "ysm",
         description: "Myanmar Sign Language",
         added: "2021-02-20",
         preferred_value: "ysm",
         prefix: "sgn"}).
iana(extlang,yue,tag{
         type: "extlang",
         subtag: "yue",
         description: ["Yue Chinese","Cantonese"],
         added: "2009-07-29",
         preferred_value: "yue",
         prefix: "zh",
         macrolanguage: "zh"}).
iana(extlang,zib,tag{
         type: "extlang",
         subtag: "zib",
         description: "Zimbabwe Sign Language",
         added: "2009-07-29",
         preferred_value: "zib",
         prefix: "sgn"}).
iana(extlang,zlm,tag{
         type: "extlang",
         subtag: "zlm",
         description: "Malay (individual language)",
         added: "2009-07-29",
         preferred_value: "zlm",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,zmi,tag{
         type: "extlang",
         subtag: "zmi",
         description: "Negeri Sembilan Malay",
         added: "2009-07-29",
         preferred_value: "zmi",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(extlang,zsl,tag{
         type: "extlang",
         subtag: "zsl",
         description: "Zambian Sign Language",
         added: "2009-07-29",
         preferred_value: "zsl",
         prefix: "sgn"}).
iana(extlang,zsm,tag{
         type: "extlang",
         subtag: "zsm",
         description: "Standard Malay",
         added: "2009-07-29",
         preferred_value: "zsm",
         prefix: "ms",
         macrolanguage: "ms"}).
iana(script,'Adlm',tag{
         type: "script",
         subtag: "Adlm",
         description: "Adlam",
         added: "2014-12-11"}).
iana(script,'Afak',tag{
         type: "script",
         subtag: "Afak",
         description: "Afaka",
         added: "2011-01-07"}).
iana(script,'Aghb',tag{
         type: "script",
         subtag: "Aghb",
         description: "Caucasian Albanian",
         added: "2012-11-01"}).
iana(script,'Ahom',tag{
         type: "script",
         subtag: "Ahom",
         description: ["Ahom","Tai Ahom"],
         added: "2013-12-02"}).
iana(script,'Arab',tag{
         type: "script",
         subtag: "Arab",
         description: "Arabic",
         added: "2005-10-16"}).
iana(script,'Aran',tag{
         type: "script",
         subtag: "Aran",
         description: "Arabic (Nastaliq variant)",
         added: "2014-12-11"}).
iana(script,'Armi',tag{
         type: "script",
         subtag: "Armi",
         description: "Imperial Aramaic",
         added: "2007-12-05"}).
iana(script,'Armn',tag{
         type: "script",
         subtag: "Armn",
         description: "Armenian",
         added: "2005-10-16"}).
iana(script,'Avst',tag{
         type: "script",
         subtag: "Avst",
         description: "Avestan",
         added: "2007-07-28"}).
iana(script,'Bali',tag{
         type: "script",
         subtag: "Bali",
         description: "Balinese",
         added: "2005-10-16"}).
iana(script,'Bamu',tag{
         type: "script",
         subtag: "Bamu",
         description: "Bamum",
         added: "2009-07-30"}).
iana(script,'Bass',tag{
         type: "script",
         subtag: "Bass",
         description: "Bassa Vah",
         added: "2010-04-10"}).
iana(script,'Batk',tag{
         type: "script",
         subtag: "Batk",
         description: "Batak",
         added: "2005-10-16"}).
iana(script,'Beng',tag{
         type: "script",
         subtag: "Beng",
         description: ["Bengali","Bangla"],
         added: "2005-10-16"}).
iana(script,'Bhks',tag{
         type: "script",
         subtag: "Bhks",
         description: "Bhaiksuki",
         added: "2015-07-24"}).
iana(script,'Blis',tag{
         type: "script",
         subtag: "Blis",
         description: "Blissymbols",
         added: "2005-10-16"}).
iana(script,'Bopo',tag{
         type: "script",
         subtag: "Bopo",
         description: "Bopomofo",
         added: "2005-10-16"}).
iana(script,'Brah',tag{
         type: "script",
         subtag: "Brah",
         description: "Brahmi",
         added: "2005-10-16"}).
iana(script,'Brai',tag{
         type: "script",
         subtag: "Brai",
         description: "Braille",
         added: "2005-10-16"}).
iana(script,'Bugi',tag{
         type: "script",
         subtag: "Bugi",
         description: "Buginese",
         added: "2005-10-16"}).
iana(script,'Buhd',tag{
         type: "script",
         subtag: "Buhd",
         description: "Buhid",
         added: "2005-10-16"}).
iana(script,'Cakm',tag{
         type: "script",
         subtag: "Cakm",
         description: "Chakma",
         added: "2007-12-05"}).
iana(script,'Cans',tag{
         type: "script",
         subtag: "Cans",
         description: "Unified Canadian Aboriginal Syllabics",
         added: "2005-10-16"}).
iana(script,'Cari',tag{
         type: "script",
         subtag: "Cari",
         description: "Carian",
         added: "2006-07-21"}).
iana(script,'Cham',tag{
         type: "script",
         subtag: "Cham",
         description: "Cham",
         added: "2005-10-16"}).
iana(script,'Cher',tag{
         type: "script",
         subtag: "Cher",
         description: "Cherokee",
         added: "2005-10-16"}).
iana(script,'Chrs',tag{
         type: "script",
         subtag: "Chrs",
         description: "Chorasmian",
         added: "2019-09-11"}).
iana(script,'Cirt',tag{
         type: "script",
         subtag: "Cirt",
         description: "Cirth",
         added: "2005-10-16"}).
iana(script,'Copt',tag{
         type: "script",
         subtag: "Copt",
         description: "Coptic",
         added: "2005-10-16"}).
iana(script,'Cpmn',tag{
         type: "script",
         subtag: "Cpmn",
         description: "Cypro-Minoan",
         added: "2017-08-13"}).
iana(script,'Cprt',tag{
         type: "script",
         subtag: "Cprt",
         description: "Cypriot syllabary",
         added: "2005-10-16"}).
iana(script,'Cyrl',tag{
         type: "script",
         subtag: "Cyrl",
         description: "Cyrillic",
         added: "2005-10-16"}).
iana(script,'Cyrs',tag{
         type: "script",
         subtag: "Cyrs",
         description: "Cyrillic (Old Church Slavonic variant)",
         added: "2005-10-16"}).
iana(script,'Deva',tag{
         type: "script",
         subtag: "Deva",
         description: ["Devanagari","Nagari"],
         added: "2005-10-16"}).
iana(script,'Diak',tag{
         type: "script",
         subtag: "Diak",
         description: "Dives Akuru",
         added: "2019-09-11"}).
iana(script,'Dogr',tag{
         type: "script",
         subtag: "Dogr",
         description: "Dogra",
         added: "2017-01-13"}).
iana(script,'Dsrt',tag{
         type: "script",
         subtag: "Dsrt",
         description: ["Deseret","Mormon"],
         added: "2005-10-16"}).
iana(script,'Dupl',tag{
         type: "script",
         subtag: "Dupl",
         description: ["Duployan shorthand","Duployan stenography"],
         added: "2010-08-16"}).
iana(script,'Egyd',tag{
         type: "script",
         subtag: "Egyd",
         description: "Egyptian demotic",
         added: "2005-10-16"}).
iana(script,'Egyh',tag{
         type: "script",
         subtag: "Egyh",
         description: "Egyptian hieratic",
         added: "2005-10-16"}).
iana(script,'Egyp',tag{
         type: "script",
         subtag: "Egyp",
         description: "Egyptian hieroglyphs",
         added: "2005-10-16"}).
iana(script,'Elba',tag{
         type: "script",
         subtag: "Elba",
         description: "Elbasan",
         added: "2010-08-16"}).
iana(script,'Elym',tag{
         type: "script",
         subtag: "Elym",
         description: "Elymaic",
         added: "2018-10-28"}).
iana(script,'Ethi',tag{
         type: "script",
         subtag: "Ethi",
         description: ["Ethiopic","Geʻez","Ge'ez"],
         added: "2005-10-16"}).
iana(script,'Geok',tag{
         type: "script",
         subtag: "Geok",
         description: "Khutsuri (Asomtavruli and Nuskhuri)",
         added: "2005-10-16"}).
iana(script,'Geor',tag{
         type: "script",
         subtag: "Geor",
         description: "Georgian (Mkhedruli and Mtavruli)",
         added: "2005-10-16"}).
iana(script,'Glag',tag{
         type: "script",
         subtag: "Glag",
         description: "Glagolitic",
         added: "2005-10-16"}).
iana(script,'Gong',tag{
         type: "script",
         subtag: "Gong",
         description: "Gunjala Gondi",
         added: "2017-01-13"}).
iana(script,'Gonm',tag{
         type: "script",
         subtag: "Gonm",
         description: "Masaram Gondi",
         added: "2017-01-13"}).
iana(script,'Goth',tag{
         type: "script",
         subtag: "Goth",
         description: "Gothic",
         added: "2005-10-16"}).
iana(script,'Gran',tag{
         type: "script",
         subtag: "Gran",
         description: "Grantha",
         added: "2009-12-09"}).
iana(script,'Grek',tag{
         type: "script",
         subtag: "Grek",
         description: "Greek",
         added: "2005-10-16"}).
iana(script,'Gujr',tag{
         type: "script",
         subtag: "Gujr",
         description: "Gujarati",
         added: "2005-10-16"}).
iana(script,'Guru',tag{
         type: "script",
         subtag: "Guru",
         description: "Gurmukhi",
         added: "2005-10-16"}).
iana(script,'Hanb',tag{
         type: "script",
         subtag: "Hanb",
         description: "Han with Bopomofo (alias for Han + Bopomofo)",
         added: "2016-02-08"}).
iana(script,'Hang',tag{
         type: "script",
         subtag: "Hang",
         description: ["Hangul","Hangŭl","Hangeul"],
         added: "2005-10-16"}).
iana(script,'Hani',tag{
         type: "script",
         subtag: "Hani",
         description: ["Han","Hanzi","Kanji","Hanja"],
         added: "2005-10-16"}).
iana(script,'Hano',tag{
         type: "script",
         subtag: "Hano",
         description: ["Hanunoo","Hanunóo"],
         added: "2005-10-16"}).
iana(script,'Hans',tag{
         type: "script",
         subtag: "Hans",
         description: "Han (Simplified variant)",
         added: "2005-10-16"}).
iana(script,'Hant',tag{
         type: "script",
         subtag: "Hant",
         description: "Han (Traditional variant)",
         added: "2005-10-16"}).
iana(script,'Hatr',tag{
         type: "script",
         subtag: "Hatr",
         description: "Hatran",
         added: "2013-12-02"}).
iana(script,'Hebr',tag{
         type: "script",
         subtag: "Hebr",
         description: "Hebrew",
         added: "2005-10-16"}).
iana(script,'Hira',tag{
         type: "script",
         subtag: "Hira",
         description: "Hiragana",
         added: "2005-10-16"}).
iana(script,'Hluw',tag{
         type: "script",
         subtag: "Hluw",
         description: ["Anatolian Hieroglyphs","Luwian Hieroglyphs","Hittite Hieroglyphs"],
         added: "2011-12-28"}).
iana(script,'Hmng',tag{
         type: "script",
         subtag: "Hmng",
         description: "Pahawh Hmong",
         added: "2005-10-16"}).
iana(script,'Hmnp',tag{
         type: "script",
         subtag: "Hmnp",
         description: "Nyiakeng Puachue Hmong",
         added: "2017-08-13"}).
iana(script,'Hrkt',tag{
         type: "script",
         subtag: "Hrkt",
         description: "Japanese syllabaries (alias for Hiragana + Katakana)",
         added: "2005-10-16"}).
iana(script,'Hung',tag{
         type: "script",
         subtag: "Hung",
         description: ["Old Hungarian","Hungarian Runic"],
         added: "2005-10-16"}).
iana(script,'Inds',tag{
         type: "script",
         subtag: "Inds",
         description: ["Indus","Harappan"],
         added: "2005-10-16"}).
iana(script,'Ital',tag{
         type: "script",
         subtag: "Ital",
         description: "Old Italic (Etruscan, Oscan, etc.)",
         added: "2005-10-16"}).
iana(script,'Jamo',tag{
         type: "script",
         subtag: "Jamo",
         description: "Jamo (alias for Jamo subset of Hangul)",
         added: "2016-02-08"}).
iana(script,'Java',tag{
         type: "script",
         subtag: "Java",
         description: "Javanese",
         added: "2005-10-16"}).
iana(script,'Jpan',tag{
         type: "script",
         subtag: "Jpan",
         description: "Japanese (alias for Han + Hiragana + Katakana)",
         added: "2006-07-21"}).
iana(script,'Jurc',tag{
         type: "script",
         subtag: "Jurc",
         description: "Jurchen",
         added: "2011-01-07"}).
iana(script,'Kali',tag{
         type: "script",
         subtag: "Kali",
         description: "Kayah Li",
         added: "2005-10-16"}).
iana(script,'Kana',tag{
         type: "script",
         subtag: "Kana",
         description: "Katakana",
         added: "2005-10-16"}).
iana(script,'Kawi',tag{
         type: "script",
         subtag: "Kawi",
         description: "Kawi",
         added: "2021-12-24"}).
iana(script,'Khar',tag{
         type: "script",
         subtag: "Khar",
         description: "Kharoshthi",
         added: "2005-10-16"}).
iana(script,'Khmr',tag{
         type: "script",
         subtag: "Khmr",
         description: "Khmer",
         added: "2005-10-16"}).
iana(script,'Khoj',tag{
         type: "script",
         subtag: "Khoj",
         description: "Khojki",
         added: "2011-08-16"}).
iana(script,'Kitl',tag{
         type: "script",
         subtag: "Kitl",
         description: "Khitan large script",
         added: "2014-12-11"}).
iana(script,'Kits',tag{
         type: "script",
         subtag: "Kits",
         description: "Khitan small script",
         added: "2014-12-11"}).
iana(script,'Knda',tag{
         type: "script",
         subtag: "Knda",
         description: "Kannada",
         added: "2005-10-16"}).
iana(script,'Kore',tag{
         type: "script",
         subtag: "Kore",
         description: "Korean (alias for Hangul + Han)",
         added: "2007-07-05"}).
iana(script,'Kpel',tag{
         type: "script",
         subtag: "Kpel",
         description: "Kpelle",
         added: "2010-04-10"}).
iana(script,'Kthi',tag{
         type: "script",
         subtag: "Kthi",
         description: "Kaithi",
         added: "2007-12-05"}).
iana(script,'Lana',tag{
         type: "script",
         subtag: "Lana",
         description: ["Tai Tham","Lanna"],
         added: "2006-07-21"}).
iana(script,'Laoo',tag{
         type: "script",
         subtag: "Laoo",
         description: "Lao",
         added: "2005-10-16"}).
iana(script,'Latf',tag{
         type: "script",
         subtag: "Latf",
         description: "Latin (Fraktur variant)",
         added: "2005-10-16"}).
iana(script,'Latg',tag{
         type: "script",
         subtag: "Latg",
         description: "Latin (Gaelic variant)",
         added: "2005-10-16"}).
iana(script,'Latn',tag{
         type: "script",
         subtag: "Latn",
         description: "Latin",
         added: "2005-10-16"}).
iana(script,'Leke',tag{
         type: "script",
         subtag: "Leke",
         description: "Leke",
         added: "2015-07-24"}).
iana(script,'Lepc',tag{
         type: "script",
         subtag: "Lepc",
         description: ["Lepcha","Róng"],
         added: "2005-10-16"}).
iana(script,'Limb',tag{
         type: "script",
         subtag: "Limb",
         description: "Limbu",
         added: "2005-10-16"}).
iana(script,'Lina',tag{
         type: "script",
         subtag: "Lina",
         description: "Linear A",
         added: "2005-10-16"}).
iana(script,'Linb',tag{
         type: "script",
         subtag: "Linb",
         description: "Linear B",
         added: "2005-10-16"}).
iana(script,'Lisu',tag{
         type: "script",
         subtag: "Lisu",
         description: ["Lisu","Fraser"],
         added: "2009-03-13"}).
iana(script,'Loma',tag{
         type: "script",
         subtag: "Loma",
         description: "Loma",
         added: "2010-04-10"}).
iana(script,'Lyci',tag{
         type: "script",
         subtag: "Lyci",
         description: "Lycian",
         added: "2006-07-21"}).
iana(script,'Lydi',tag{
         type: "script",
         subtag: "Lydi",
         description: "Lydian",
         added: "2006-07-21"}).
iana(script,'Mahj',tag{
         type: "script",
         subtag: "Mahj",
         description: "Mahajani",
         added: "2012-11-01"}).
iana(script,'Maka',tag{
         type: "script",
         subtag: "Maka",
         description: "Makasar",
         added: "2017-01-13"}).
iana(script,'Mand',tag{
         type: "script",
         subtag: "Mand",
         description: ["Mandaic","Mandaean"],
         added: "2005-10-16"}).
iana(script,'Mani',tag{
         type: "script",
         subtag: "Mani",
         description: "Manichaean",
         added: "2007-07-28"}).
iana(script,'Marc',tag{
         type: "script",
         subtag: "Marc",
         description: "Marchen",
         added: "2014-12-11"}).
iana(script,'Maya',tag{
         type: "script",
         subtag: "Maya",
         description: "Mayan hieroglyphs",
         added: "2005-10-16"}).
iana(script,'Medf',tag{
         type: "script",
         subtag: "Medf",
         description: ["Medefaidrin","Oberi Okaime","Oberi Ɔkaimɛ"],
         added: "2017-01-13"}).
iana(script,'Mend',tag{
         type: "script",
         subtag: "Mend",
         description: "Mende Kikakui",
         added: "2010-04-10"}).
iana(script,'Merc',tag{
         type: "script",
         subtag: "Merc",
         description: "Meroitic Cursive",
         added: "2009-12-09"}).
iana(script,'Mero',tag{
         type: "script",
         subtag: "Mero",
         description: "Meroitic Hieroglyphs",
         added: "2005-10-16"}).
iana(script,'Mlym',tag{
         type: "script",
         subtag: "Mlym",
         description: "Malayalam",
         added: "2005-10-16"}).
iana(script,'Modi',tag{
         type: "script",
         subtag: "Modi",
         description: ["Modi","Moḍī"],
         added: "2013-12-02"}).
iana(script,'Mong',tag{
         type: "script",
         subtag: "Mong",
         description: "Mongolian",
         added: "2005-10-16"}).
iana(script,'Moon',tag{
         type: "script",
         subtag: "Moon",
         description: ["Moon","Moon code","Moon script","Moon type"],
         added: "2007-01-26"}).
iana(script,'Mroo',tag{
         type: "script",
         subtag: "Mroo",
         description: ["Mro","Mru"],
         added: "2011-01-07"}).
iana(script,'Mtei',tag{
         type: "script",
         subtag: "Mtei",
         description: ["Meitei Mayek","Meithei","Meetei"],
         added: "2007-01-26"}).
iana(script,'Mult',tag{
         type: "script",
         subtag: "Mult",
         description: "Multani",
         added: "2013-12-02"}).
iana(script,'Mymr',tag{
         type: "script",
         subtag: "Mymr",
         description: ["Myanmar","Burmese"],
         added: "2005-10-16"}).
iana(script,'Nagm',tag{
         type: "script",
         subtag: "Nagm",
         description: "Nag Mundari",
         added: "2021-12-24"}).
iana(script,'Nand',tag{
         type: "script",
         subtag: "Nand",
         description: "Nandinagari",
         added: "2018-10-28"}).
iana(script,'Narb',tag{
         type: "script",
         subtag: "Narb",
         description: ["Old North Arabian","Ancient North Arabian"],
         added: "2010-04-10"}).
iana(script,'Nbat',tag{
         type: "script",
         subtag: "Nbat",
         description: "Nabataean",
         added: "2010-04-10"}).
iana(script,'Newa',tag{
         type: "script",
         subtag: "Newa",
         description: ["Newa","Newar","Newari","Nepāla lipi"],
         added: "2016-01-04"}).
iana(script,'Nkdb',tag{
         type: "script",
         subtag: "Nkdb",
         description: ["Naxi Dongba","na²¹ɕi³³ to³³ba²¹","Nakhi Tomba"],
         added: "2017-08-13"}).
iana(script,'Nkgb',tag{
         type: "script",
         subtag: "Nkgb",
         description: ["Naxi Geba","na²¹ɕi³³ gʌ²¹ba²¹","'Na-'Khi ²Ggŏ-¹baw","Nakhi Geba"],
         added: "2009-03-13"}).
iana(script,'Nkoo',tag{
         type: "script",
         subtag: "Nkoo",
         description: ["N’Ko","N'Ko"],
         added: "2005-10-16"}).
iana(script,'Nshu',tag{
         type: "script",
         subtag: "Nshu",
         description: "Nüshu",
         added: "2011-01-07"}).
iana(script,'Ogam',tag{
         type: "script",
         subtag: "Ogam",
         description: "Ogham",
         added: "2005-10-16"}).
iana(script,'Olck',tag{
         type: "script",
         subtag: "Olck",
         description: ["Ol Chiki","Ol Cemet'","Ol","Santali"],
         added: "2006-07-21"}).
iana(script,'Orkh',tag{
         type: "script",
         subtag: "Orkh",
         description: ["Old Turkic","Orkhon Runic"],
         added: "2009-07-30"}).
iana(script,'Orya',tag{
         type: "script",
         subtag: "Orya",
         description: ["Oriya","Odia"],
         added: "2005-10-16"}).
iana(script,'Osge',tag{
         type: "script",
         subtag: "Osge",
         description: "Osage",
         added: "2014-12-11"}).
iana(script,'Osma',tag{
         type: "script",
         subtag: "Osma",
         description: "Osmanya",
         added: "2005-10-16"}).
iana(script,'Ougr',tag{
         type: "script",
         subtag: "Ougr",
         description: "Old Uyghur",
         added: "2021-02-12"}).
iana(script,'Palm',tag{
         type: "script",
         subtag: "Palm",
         description: "Palmyrene",
         added: "2010-04-10"}).
iana(script,'Pauc',tag{
         type: "script",
         subtag: "Pauc",
         description: "Pau Cin Hau",
         added: "2013-12-02"}).
iana(script,'Pcun',tag{
         type: "script",
         subtag: "Pcun",
         description: "Proto-Cuneiform",
         added: "2021-02-12"}).
iana(script,'Pelm',tag{
         type: "script",
         subtag: "Pelm",
         description: "Proto-Elamite",
         added: "2021-02-12"}).
iana(script,'Perm',tag{
         type: "script",
         subtag: "Perm",
         description: "Old Permic",
         added: "2005-10-16"}).
iana(script,'Phag',tag{
         type: "script",
         subtag: "Phag",
         description: "Phags-pa",
         added: "2005-10-16"}).
iana(script,'Phli',tag{
         type: "script",
         subtag: "Phli",
         description: "Inscriptional Pahlavi",
         added: "2007-12-05"}).
iana(script,'Phlp',tag{
         type: "script",
         subtag: "Phlp",
         description: "Psalter Pahlavi",
         added: "2007-12-05"}).
iana(script,'Phlv',tag{
         type: "script",
         subtag: "Phlv",
         description: "Book Pahlavi",
         added: "2007-07-28"}).
iana(script,'Phnx',tag{
         type: "script",
         subtag: "Phnx",
         description: "Phoenician",
         added: "2005-10-16"}).
iana(script,'Piqd',tag{
         type: "script",
         subtag: "Piqd",
         description: "Klingon (KLI pIqaD)",
         added: "2016-01-04"}).
iana(script,'Plrd',tag{
         type: "script",
         subtag: "Plrd",
         description: ["Miao","Pollard"],
         added: "2005-10-16"}).
iana(script,'Prti',tag{
         type: "script",
         subtag: "Prti",
         description: "Inscriptional Parthian",
         added: "2007-12-05"}).
iana(script,'Psin',tag{
         type: "script",
         subtag: "Psin",
         description: "Proto-Sinaitic",
         added: "2021-02-12"}).
iana(script,'Qaaa',tag{
         type: "script",
         subtag: "Qaaa..Qabx",
         description: "Private use",
         added: "2005-10-16"}).
iana(script,'Ranj',tag{
         type: "script",
         subtag: "Ranj",
         description: "Ranjana",
         added: "2021-02-12"}).
iana(script,'Rjng',tag{
         type: "script",
         subtag: "Rjng",
         description: ["Rejang","Redjang","Kaganga"],
         added: "2006-10-17"}).
iana(script,'Rohg',tag{
         type: "script",
         subtag: "Rohg",
         description: "Hanifi Rohingya",
         added: "2017-12-13"}).
iana(script,'Roro',tag{
         type: "script",
         subtag: "Roro",
         description: "Rongorongo",
         added: "2005-10-16"}).
iana(script,'Runr',tag{
         type: "script",
         subtag: "Runr",
         description: "Runic",
         added: "2005-10-16"}).
iana(script,'Samr',tag{
         type: "script",
         subtag: "Samr",
         description: "Samaritan",
         added: "2007-07-28"}).
iana(script,'Sara',tag{
         type: "script",
         subtag: "Sara",
         description: "Sarati",
         added: "2005-10-16"}).
iana(script,'Sarb',tag{
         type: "script",
         subtag: "Sarb",
         description: "Old South Arabian",
         added: "2009-07-30"}).
iana(script,'Saur',tag{
         type: "script",
         subtag: "Saur",
         description: "Saurashtra",
         added: "2006-07-21"}).
iana(script,'Sgnw',tag{
         type: "script",
         subtag: "Sgnw",
         description: "SignWriting",
         added: "2006-10-17"}).
iana(script,'Shaw',tag{
         type: "script",
         subtag: "Shaw",
         description: ["Shavian","Shaw"],
         added: "2005-10-16"}).
iana(script,'Shrd',tag{
         type: "script",
         subtag: "Shrd",
         description: ["Sharada","Śāradā"],
         added: "2011-01-07"}).
iana(script,'Shui',tag{
         type: "script",
         subtag: "Shui",
         description: "Shuishu",
         added: "2017-08-13"}).
iana(script,'Sidd',tag{
         type: "script",
         subtag: "Sidd",
         description: ["Siddham","Siddhaṃ","Siddhamātṛkā"],
         added: "2013-12-02"}).
iana(script,'Sind',tag{
         type: "script",
         subtag: "Sind",
         description: ["Khudawadi","Sindhi"],
         added: "2010-08-16"}).
iana(script,'Sinh',tag{
         type: "script",
         subtag: "Sinh",
         description: "Sinhala",
         added: "2005-10-16"}).
iana(script,'Sogd',tag{
         type: "script",
         subtag: "Sogd",
         description: "Sogdian",
         added: "2017-12-13"}).
iana(script,'Sogo',tag{
         type: "script",
         subtag: "Sogo",
         description: "Old Sogdian",
         added: "2017-12-13"}).
iana(script,'Sora',tag{
         type: "script",
         subtag: "Sora",
         description: "Sora Sompeng",
         added: "2011-01-07"}).
iana(script,'Soyo',tag{
         type: "script",
         subtag: "Soyo",
         description: "Soyombo",
         added: "2017-01-13"}).
iana(script,'Sund',tag{
         type: "script",
         subtag: "Sund",
         description: "Sundanese",
         added: "2006-07-21"}).
iana(script,'Sunu',tag{
         type: "script",
         subtag: "Sunu",
         description: "Sunuwar",
         added: "2021-12-24"}).
iana(script,'Sylo',tag{
         type: "script",
         subtag: "Sylo",
         description: "Syloti Nagri",
         added: "2005-10-16"}).
iana(script,'Syrc',tag{
         type: "script",
         subtag: "Syrc",
         description: "Syriac",
         added: "2005-10-16"}).
iana(script,'Syre',tag{
         type: "script",
         subtag: "Syre",
         description: "Syriac (Estrangelo variant)",
         added: "2005-10-16"}).
iana(script,'Syrj',tag{
         type: "script",
         subtag: "Syrj",
         description: "Syriac (Western variant)",
         added: "2005-10-16"}).
iana(script,'Syrn',tag{
         type: "script",
         subtag: "Syrn",
         description: "Syriac (Eastern variant)",
         added: "2005-10-16"}).
iana(script,'Tagb',tag{
         type: "script",
         subtag: "Tagb",
         description: "Tagbanwa",
         added: "2005-10-16"}).
iana(script,'Takr',tag{
         type: "script",
         subtag: "Takr",
         description: ["Takri","Ṭākrī","Ṭāṅkrī"],
         added: "2011-01-07"}).
iana(script,'Tale',tag{
         type: "script",
         subtag: "Tale",
         description: "Tai Le",
         added: "2005-10-16"}).
iana(script,'Talu',tag{
         type: "script",
         subtag: "Talu",
         description: "New Tai Lue",
         added: "2005-10-16"}).
iana(script,'Taml',tag{
         type: "script",
         subtag: "Taml",
         description: "Tamil",
         added: "2005-10-16"}).
iana(script,'Tang',tag{
         type: "script",
         subtag: "Tang",
         description: "Tangut",
         added: "2011-01-07"}).
iana(script,'Tavt',tag{
         type: "script",
         subtag: "Tavt",
         description: "Tai Viet",
         added: "2007-12-05"}).
iana(script,'Telu',tag{
         type: "script",
         subtag: "Telu",
         description: "Telugu",
         added: "2005-10-16"}).
iana(script,'Teng',tag{
         type: "script",
         subtag: "Teng",
         description: "Tengwar",
         added: "2005-10-16"}).
iana(script,'Tfng',tag{
         type: "script",
         subtag: "Tfng",
         description: ["Tifinagh","Berber"],
         added: "2005-10-16"}).
iana(script,'Tglg',tag{
         type: "script",
         subtag: "Tglg",
         description: ["Tagalog","Baybayin","Alibata"],
         added: "2005-10-16"}).
iana(script,'Thaa',tag{
         type: "script",
         subtag: "Thaa",
         description: "Thaana",
         added: "2005-10-16"}).
iana(script,'Thai',tag{
         type: "script",
         subtag: "Thai",
         description: "Thai",
         added: "2005-10-16"}).
iana(script,'Tibt',tag{
         type: "script",
         subtag: "Tibt",
         description: "Tibetan",
         added: "2005-10-16"}).
iana(script,'Tirh',tag{
         type: "script",
         subtag: "Tirh",
         description: "Tirhuta",
         added: "2011-08-16"}).
iana(script,'Tnsa',tag{
         type: "script",
         subtag: "Tnsa",
         description: "Tangsa",
         added: "2021-03-05"}).
iana(script,'Toto',tag{
         type: "script",
         subtag: "Toto",
         description: "Toto",
         added: "2020-05-12"}).
iana(script,'Ugar',tag{
         type: "script",
         subtag: "Ugar",
         description: "Ugaritic",
         added: "2005-10-16"}).
iana(script,'Vaii',tag{
         type: "script",
         subtag: "Vaii",
         description: "Vai",
         added: "2005-10-16"}).
iana(script,'Visp',tag{
         type: "script",
         subtag: "Visp",
         description: "Visible Speech",
         added: "2005-10-16"}).
iana(script,'Vith',tag{
         type: "script",
         subtag: "Vith",
         description: "Vithkuqi",
         added: "2021-03-05"}).
iana(script,'Wara',tag{
         type: "script",
         subtag: "Wara",
         description: ["Warang Citi","Varang Kshiti"],
         added: "2009-12-09"}).
iana(script,'Wcho',tag{
         type: "script",
         subtag: "Wcho",
         description: "Wancho",
         added: "2017-08-13"}).
iana(script,'Wole',tag{
         type: "script",
         subtag: "Wole",
         description: "Woleai",
         added: "2011-01-07"}).
iana(script,'Xpeo',tag{
         type: "script",
         subtag: "Xpeo",
         description: "Old Persian",
         added: "2005-10-16"}).
iana(script,'Xsux',tag{
         type: "script",
         subtag: "Xsux",
         description: "Sumero-Akkadian cuneiform",
         added: "2005-10-16"}).
iana(script,'Yezi',tag{
         type: "script",
         subtag: "Yezi",
         description: "Yezidi",
         added: "2019-09-11"}).
iana(script,'Yiii',tag{
         type: "script",
         subtag: "Yiii",
         description: "Yi",
         added: "2005-10-16"}).
iana(script,'Zanb',tag{
         type: "script",
         subtag: "Zanb",
         description: ["Zanabazar Square","Zanabazarin Dörböljin Useg","Xewtee Dörböljin Bicig","Horizontal Square Script"],
         added: "2017-01-13"}).
iana(script,'Zinh',tag{
         type: "script",
         subtag: "Zinh",
         description: "Code for inherited script",
         added: "2009-04-03",
         comments: "Not intended for use as a language subtag"}).
iana(script,'Zmth',tag{
         type: "script",
         subtag: "Zmth",
         description: "Mathematical notation",
         added: "2007-12-05"}).
iana(script,'Zsye',tag{
         type: "script",
         subtag: "Zsye",
         description: "Symbols (Emoji variant)",
         added: "2016-01-04"}).
iana(script,'Zsym',tag{
         type: "script",
         subtag: "Zsym",
         description: "Symbols",
         added: "2007-12-05"}).
iana(script,'Zxxx',tag{
         type: "script",
         subtag: "Zxxx",
         description: "Code for unwritten documents",
         added: "2005-10-16"}).
iana(script,'Zyyy',tag{
         type: "script",
         subtag: "Zyyy",
         description: "Code for undetermined script",
         added: "2005-10-16"}).
iana(script,'Zzzz',tag{
         type: "script",
         subtag: "Zzzz",
         description: "Code for uncoded script",
         added: "2005-10-16"}).
iana(region,'AA',tag{
         type: "region",
         subtag: "AA",
         description: "Private use",
         added: "2005-10-16"}).
iana(region,'AC',tag{
         type: "region",
         subtag: "AC",
         description: "Ascension Island",
         added: "2009-07-29"}).
iana(region,'AD',tag{
         type: "region",
         subtag: "AD",
         description: "Andorra",
         added: "2005-10-16"}).
iana(region,'AE',tag{
         type: "region",
         subtag: "AE",
         description: "United Arab Emirates",
         added: "2005-10-16"}).
iana(region,'AF',tag{
         type: "region",
         subtag: "AF",
         description: "Afghanistan",
         added: "2005-10-16"}).
iana(region,'AG',tag{
         type: "region",
         subtag: "AG",
         description: "Antigua and Barbuda",
         added: "2005-10-16"}).
iana(region,'AI',tag{
         type: "region",
         subtag: "AI",
         description: "Anguilla",
         added: "2005-10-16"}).
iana(region,'AL',tag{
         type: "region",
         subtag: "AL",
         description: "Albania",
         added: "2005-10-16"}).
iana(region,'AM',tag{
         type: "region",
         subtag: "AM",
         description: "Armenia",
         added: "2005-10-16"}).
iana(region,'AN',tag{
         type: "region",
         subtag: "AN",
         description: "Netherlands Antilles",
         added: "2005-10-16",
         deprecated: "2011-01-07",
         comments: "see BQ, CW, and SX"}).
iana(region,'AO',tag{
         type: "region",
         subtag: "AO",
         description: "Angola",
         added: "2005-10-16"}).
iana(region,'AQ',tag{
         type: "region",
         subtag: "AQ",
         description: "Antarctica",
         added: "2005-10-16"}).
iana(region,'AR',tag{
         type: "region",
         subtag: "AR",
         description: "Argentina",
         added: "2005-10-16"}).
iana(region,'AS',tag{
         type: "region",
         subtag: "AS",
         description: "American Samoa",
         added: "2005-10-16"}).
iana(region,'AT',tag{
         type: "region",
         subtag: "AT",
         description: "Austria",
         added: "2005-10-16"}).
iana(region,'AU',tag{
         type: "region",
         subtag: "AU",
         description: "Australia",
         added: "2005-10-16"}).
iana(region,'AW',tag{
         type: "region",
         subtag: "AW",
         description: "Aruba",
         added: "2005-10-16"}).
iana(region,'AX',tag{
         type: "region",
         subtag: "AX",
         description: "Åland Islands",
         added: "2005-10-16"}).
iana(region,'AZ',tag{
         type: "region",
         subtag: "AZ",
         description: "Azerbaijan",
         added: "2005-10-16"}).
iana(region,'BA',tag{
         type: "region",
         subtag: "BA",
         description: "Bosnia and Herzegovina",
         added: "2005-10-16"}).
iana(region,'BB',tag{
         type: "region",
         subtag: "BB",
         description: "Barbados",
         added: "2005-10-16"}).
iana(region,'BD',tag{
         type: "region",
         subtag: "BD",
         description: "Bangladesh",
         added: "2005-10-16"}).
iana(region,'BE',tag{
         type: "region",
         subtag: "BE",
         description: "Belgium",
         added: "2005-10-16"}).
iana(region,'BF',tag{
         type: "region",
         subtag: "BF",
         description: "Burkina Faso",
         added: "2005-10-16"}).
iana(region,'BG',tag{
         type: "region",
         subtag: "BG",
         description: "Bulgaria",
         added: "2005-10-16"}).
iana(region,'BH',tag{
         type: "region",
         subtag: "BH",
         description: "Bahrain",
         added: "2005-10-16"}).
iana(region,'BI',tag{
         type: "region",
         subtag: "BI",
         description: "Burundi",
         added: "2005-10-16"}).
iana(region,'BJ',tag{
         type: "region",
         subtag: "BJ",
         description: "Benin",
         added: "2005-10-16"}).
iana(region,'BL',tag{
         type: "region",
         subtag: "BL",
         description: "Saint Barthélemy",
         added: "2007-11-02"}).
iana(region,'BM',tag{
         type: "region",
         subtag: "BM",
         description: "Bermuda",
         added: "2005-10-16"}).
iana(region,'BN',tag{
         type: "region",
         subtag: "BN",
         description: "Brunei Darussalam",
         added: "2005-10-16"}).
iana(region,'BO',tag{
         type: "region",
         subtag: "BO",
         description: "Bolivia",
         added: "2005-10-16"}).
iana(region,'BQ',tag{
         type: "region",
         subtag: "BQ",
         description: "Bonaire, Sint Eustatius and Saba",
         added: "2011-01-07"}).
iana(region,'BR',tag{
         type: "region",
         subtag: "BR",
         description: "Brazil",
         added: "2005-10-16"}).
iana(region,'BS',tag{
         type: "region",
         subtag: "BS",
         description: "Bahamas",
         added: "2005-10-16"}).
iana(region,'BT',tag{
         type: "region",
         subtag: "BT",
         description: "Bhutan",
         added: "2005-10-16"}).
iana(region,'BU',tag{
         type: "region",
         subtag: "BU",
         description: "Burma",
         added: "2005-10-16",
         deprecated: "1989-12-05",
         preferred_value: "MM"}).
iana(region,'BV',tag{
         type: "region",
         subtag: "BV",
         description: "Bouvet Island",
         added: "2005-10-16"}).
iana(region,'BW',tag{
         type: "region",
         subtag: "BW",
         description: "Botswana",
         added: "2005-10-16"}).
iana(region,'BY',tag{
         type: "region",
         subtag: "BY",
         description: "Belarus",
         added: "2005-10-16"}).
iana(region,'BZ',tag{
         type: "region",
         subtag: "BZ",
         description: "Belize",
         added: "2005-10-16"}).
iana(region,'CA',tag{
         type: "region",
         subtag: "CA",
         description: "Canada",
         added: "2005-10-16"}).
iana(region,'CC',tag{
         type: "region",
         subtag: "CC",
         description: "Cocos (Keeling) Islands",
         added: "2005-10-16"}).
iana(region,'CD',tag{
         type: "region",
         subtag: "CD",
         description: "The Democratic Republic of the Congo",
         added: "2005-10-16"}).
iana(region,'CF',tag{
         type: "region",
         subtag: "CF",
         description: "Central African Republic",
         added: "2005-10-16"}).
iana(region,'CG',tag{
         type: "region",
         subtag: "CG",
         description: "Congo",
         added: "2005-10-16"}).
iana(region,'CH',tag{
         type: "region",
         subtag: "CH",
         description: "Switzerland",
         added: "2005-10-16"}).
iana(region,'CI',tag{
         type: "region",
         subtag: "CI",
         description: "Côte d'Ivoire",
         added: "2005-10-16"}).
iana(region,'CK',tag{
         type: "region",
         subtag: "CK",
         description: "Cook Islands",
         added: "2005-10-16"}).
iana(region,'CL',tag{
         type: "region",
         subtag: "CL",
         description: "Chile",
         added: "2005-10-16"}).
iana(region,'CM',tag{
         type: "region",
         subtag: "CM",
         description: "Cameroon",
         added: "2005-10-16"}).
iana(region,'CN',tag{
         type: "region",
         subtag: "CN",
         description: "China",
         added: "2005-10-16"}).
iana(region,'CO',tag{
         type: "region",
         subtag: "CO",
         description: "Colombia",
         added: "2005-10-16"}).
iana(region,'CP',tag{
         type: "region",
         subtag: "CP",
         description: "Clipperton Island",
         added: "2009-07-29"}).
iana(region,'CR',tag{
         type: "region",
         subtag: "CR",
         description: "Costa Rica",
         added: "2005-10-16"}).
iana(region,'CS',tag{
         type: "region",
         subtag: "CS",
         description: "Serbia and Montenegro",
         added: "2005-10-16",
         deprecated: "2006-10-05",
         comments: "see RS for Serbia or ME for Montenegro"}).
iana(region,'CU',tag{
         type: "region",
         subtag: "CU",
         description: "Cuba",
         added: "2005-10-16"}).
iana(region,'CV',tag{
         type: "region",
         subtag: "CV",
         description: ["Cabo Verde","Cape Verde"],
         added: "2005-10-16"}).
iana(region,'CW',tag{
         type: "region",
         subtag: "CW",
         description: "Curaçao",
         added: "2011-01-07"}).
iana(region,'CX',tag{
         type: "region",
         subtag: "CX",
         description: "Christmas Island",
         added: "2005-10-16"}).
iana(region,'CY',tag{
         type: "region",
         subtag: "CY",
         description: "Cyprus",
         added: "2005-10-16"}).
iana(region,'CZ',tag{
         type: "region",
         subtag: "CZ",
         description: ["Czechia","Czech Republic"],
         added: "2005-10-16"}).
iana(region,'DD',tag{
         type: "region",
         subtag: "DD",
         description: "German Democratic Republic",
         added: "2005-10-16",
         deprecated: "1990-10-30",
         preferred_value: "DE"}).
iana(region,'DE',tag{
         type: "region",
         subtag: "DE",
         description: "Germany",
         added: "2005-10-16"}).
iana(region,'DG',tag{
         type: "region",
         subtag: "DG",
         description: "Diego Garcia",
         added: "2009-07-29"}).
iana(region,'DJ',tag{
         type: "region",
         subtag: "DJ",
         description: "Djibouti",
         added: "2005-10-16"}).
iana(region,'DK',tag{
         type: "region",
         subtag: "DK",
         description: "Denmark",
         added: "2005-10-16"}).
iana(region,'DM',tag{
         type: "region",
         subtag: "DM",
         description: "Dominica",
         added: "2005-10-16"}).
iana(region,'DO',tag{
         type: "region",
         subtag: "DO",
         description: "Dominican Republic",
         added: "2005-10-16"}).
iana(region,'DZ',tag{
         type: "region",
         subtag: "DZ",
         description: "Algeria",
         added: "2005-10-16"}).
iana(region,'EA',tag{
         type: "region",
         subtag: "EA",
         description: "Ceuta, Melilla",
         added: "2009-07-29"}).
iana(region,'EC',tag{
         type: "region",
         subtag: "EC",
         description: "Ecuador",
         added: "2005-10-16"}).
iana(region,'EE',tag{
         type: "region",
         subtag: "EE",
         description: "Estonia",
         added: "2005-10-16"}).
iana(region,'EG',tag{
         type: "region",
         subtag: "EG",
         description: "Egypt",
         added: "2005-10-16"}).
iana(region,'EH',tag{
         type: "region",
         subtag: "EH",
         description: "Western Sahara",
         added: "2005-10-16"}).
iana(region,'ER',tag{
         type: "region",
         subtag: "ER",
         description: "Eritrea",
         added: "2005-10-16"}).
iana(region,'ES',tag{
         type: "region",
         subtag: "ES",
         description: "Spain",
         added: "2005-10-16"}).
iana(region,'ET',tag{
         type: "region",
         subtag: "ET",
         description: "Ethiopia",
         added: "2005-10-16"}).
iana(region,'EU',tag{
         type: "region",
         subtag: "EU",
         description: "European Union",
         added: "2009-07-29"}).
iana(region,'EZ',tag{
         type: "region",
         subtag: "EZ",
         description: "Eurozone",
         added: "2016-07-14"}).
iana(region,'FI',tag{
         type: "region",
         subtag: "FI",
         description: "Finland",
         added: "2005-10-16"}).
iana(region,'FJ',tag{
         type: "region",
         subtag: "FJ",
         description: "Fiji",
         added: "2005-10-16"}).
iana(region,'FK',tag{
         type: "region",
         subtag: "FK",
         description: "Falkland Islands (Malvinas)",
         added: "2005-10-16"}).
iana(region,'FM',tag{
         type: "region",
         subtag: "FM",
         description: "Federated States of Micronesia",
         added: "2005-10-16"}).
iana(region,'FO',tag{
         type: "region",
         subtag: "FO",
         description: "Faroe Islands",
         added: "2005-10-16"}).
iana(region,'FR',tag{
         type: "region",
         subtag: "FR",
         description: "France",
         added: "2005-10-16"}).
iana(region,'FX',tag{
         type: "region",
         subtag: "FX",
         description: "Metropolitan France",
         added: "2005-10-16",
         deprecated: "1997-07-14",
         preferred_value: "FR"}).
iana(region,'GA',tag{
         type: "region",
         subtag: "GA",
         description: "Gabon",
         added: "2005-10-16"}).
iana(region,'GB',tag{
         type: "region",
         subtag: "GB",
         description: "United Kingdom",
         added: "2005-10-16",
         comments: "as of 2006-03-29 GB no longer includes the Channel Islands and Isle of Man; see GG, JE, IM"
     }).
iana(region,'GD',tag{
         type: "region",
         subtag: "GD",
         description: "Grenada",
         added: "2005-10-16"}).
iana(region,'GE',tag{
         type: "region",
         subtag: "GE",
         description: "Georgia",
         added: "2005-10-16"}).
iana(region,'GF',tag{
         type: "region",
         subtag: "GF",
         description: "French Guiana",
         added: "2005-10-16"}).
iana(region,'GG',tag{
         type: "region",
         subtag: "GG",
         description: "Guernsey",
         added: "2006-03-29"}).
iana(region,'GH',tag{
         type: "region",
         subtag: "GH",
         description: "Ghana",
         added: "2005-10-16"}).
iana(region,'GI',tag{
         type: "region",
         subtag: "GI",
         description: "Gibraltar",
         added: "2005-10-16"}).
iana(region,'GL',tag{
         type: "region",
         subtag: "GL",
         description: "Greenland",
         added: "2005-10-16"}).
iana(region,'GM',tag{
         type: "region",
         subtag: "GM",
         description: "Gambia",
         added: "2005-10-16"}).
iana(region,'GN',tag{
         type: "region",
         subtag: "GN",
         description: "Guinea",
         added: "2005-10-16"}).
iana(region,'GP',tag{
         type: "region",
         subtag: "GP",
         description: "Guadeloupe",
         added: "2005-10-16"}).
iana(region,'GQ',tag{
         type: "region",
         subtag: "GQ",
         description: "Equatorial Guinea",
         added: "2005-10-16"}).
iana(region,'GR',tag{
         type: "region",
         subtag: "GR",
         description: "Greece",
         added: "2005-10-16"}).
iana(region,'GS',tag{
         type: "region",
         subtag: "GS",
         description: "South Georgia and the South Sandwich Islands",
         added: "2005-10-16"}).
iana(region,'GT',tag{
         type: "region",
         subtag: "GT",
         description: "Guatemala",
         added: "2005-10-16"}).
iana(region,'GU',tag{
         type: "region",
         subtag: "GU",
         description: "Guam",
         added: "2005-10-16"}).
iana(region,'GW',tag{
         type: "region",
         subtag: "GW",
         description: "Guinea-Bissau",
         added: "2005-10-16"}).
iana(region,'GY',tag{
         type: "region",
         subtag: "GY",
         description: "Guyana",
         added: "2005-10-16"}).
iana(region,'HK',tag{
         type: "region",
         subtag: "HK",
         description: "Hong Kong",
         added: "2005-10-16"}).
iana(region,'HM',tag{
         type: "region",
         subtag: "HM",
         description: "Heard Island and McDonald Islands",
         added: "2005-10-16"}).
iana(region,'HN',tag{
         type: "region",
         subtag: "HN",
         description: "Honduras",
         added: "2005-10-16"}).
iana(region,'HR',tag{
         type: "region",
         subtag: "HR",
         description: "Croatia",
         added: "2005-10-16"}).
iana(region,'HT',tag{
         type: "region",
         subtag: "HT",
         description: "Haiti",
         added: "2005-10-16"}).
iana(region,'HU',tag{
         type: "region",
         subtag: "HU",
         description: "Hungary",
         added: "2005-10-16"}).
iana(region,'IC',tag{
         type: "region",
         subtag: "IC",
         description: "Canary Islands",
         added: "2009-07-29"}).
iana(region,'ID',tag{
         type: "region",
         subtag: "ID",
         description: "Indonesia",
         added: "2005-10-16"}).
iana(region,'IE',tag{
         type: "region",
         subtag: "IE",
         description: "Ireland",
         added: "2005-10-16"}).
iana(region,'IL',tag{
         type: "region",
         subtag: "IL",
         description: "Israel",
         added: "2005-10-16"}).
iana(region,'IM',tag{
         type: "region",
         subtag: "IM",
         description: "Isle of Man",
         added: "2006-03-29"}).
iana(region,'IN',tag{
         type: "region",
         subtag: "IN",
         description: "India",
         added: "2005-10-16"}).
iana(region,'IO',tag{
         type: "region",
         subtag: "IO",
         description: "British Indian Ocean Territory",
         added: "2005-10-16"}).
iana(region,'IQ',tag{
         type: "region",
         subtag: "IQ",
         description: "Iraq",
         added: "2005-10-16"}).
iana(region,'IR',tag{
         type: "region",
         subtag: "IR",
         description: "Islamic Republic of Iran",
         added: "2005-10-16"}).
iana(region,'IS',tag{
         type: "region",
         subtag: "IS",
         description: "Iceland",
         added: "2005-10-16"}).
iana(region,'IT',tag{
         type: "region",
         subtag: "IT",
         description: "Italy",
         added: "2005-10-16"}).
iana(region,'JE',tag{
         type: "region",
         subtag: "JE",
         description: "Jersey",
         added: "2006-03-29"}).
iana(region,'JM',tag{
         type: "region",
         subtag: "JM",
         description: "Jamaica",
         added: "2005-10-16"}).
iana(region,'JO',tag{
         type: "region",
         subtag: "JO",
         description: "Jordan",
         added: "2005-10-16"}).
iana(region,'JP',tag{
         type: "region",
         subtag: "JP",
         description: "Japan",
         added: "2005-10-16"}).
iana(region,'KE',tag{
         type: "region",
         subtag: "KE",
         description: "Kenya",
         added: "2005-10-16"}).
iana(region,'KG',tag{
         type: "region",
         subtag: "KG",
         description: "Kyrgyzstan",
         added: "2005-10-16"}).
iana(region,'KH',tag{
         type: "region",
         subtag: "KH",
         description: "Cambodia",
         added: "2005-10-16"}).
iana(region,'KI',tag{
         type: "region",
         subtag: "KI",
         description: "Kiribati",
         added: "2005-10-16"}).
iana(region,'KM',tag{
         type: "region",
         subtag: "KM",
         description: "Comoros",
         added: "2005-10-16"}).
iana(region,'KN',tag{
         type: "region",
         subtag: "KN",
         description: "Saint Kitts and Nevis",
         added: "2005-10-16"}).
iana(region,'KP',tag{
         type: "region",
         subtag: "KP",
         description: "Democratic People's Republic of Korea",
         added: "2005-10-16"}).
iana(region,'KR',tag{
         type: "region",
         subtag: "KR",
         description: "Republic of Korea",
         added: "2005-10-16"}).
iana(region,'KW',tag{
         type: "region",
         subtag: "KW",
         description: "Kuwait",
         added: "2005-10-16"}).
iana(region,'KY',tag{
         type: "region",
         subtag: "KY",
         description: "Cayman Islands",
         added: "2005-10-16"}).
iana(region,'KZ',tag{
         type: "region",
         subtag: "KZ",
         description: "Kazakhstan",
         added: "2005-10-16"}).
iana(region,'LA',tag{
         type: "region",
         subtag: "LA",
         description: "Lao People's Democratic Republic",
         added: "2005-10-16"}).
iana(region,'LB',tag{
         type: "region",
         subtag: "LB",
         description: "Lebanon",
         added: "2005-10-16"}).
iana(region,'LC',tag{
         type: "region",
         subtag: "LC",
         description: "Saint Lucia",
         added: "2005-10-16"}).
iana(region,'LI',tag{
         type: "region",
         subtag: "LI",
         description: "Liechtenstein",
         added: "2005-10-16"}).
iana(region,'LK',tag{
         type: "region",
         subtag: "LK",
         description: "Sri Lanka",
         added: "2005-10-16"}).
iana(region,'LR',tag{
         type: "region",
         subtag: "LR",
         description: "Liberia",
         added: "2005-10-16"}).
iana(region,'LS',tag{
         type: "region",
         subtag: "LS",
         description: "Lesotho",
         added: "2005-10-16"}).
iana(region,'LT',tag{
         type: "region",
         subtag: "LT",
         description: "Lithuania",
         added: "2005-10-16"}).
iana(region,'LU',tag{
         type: "region",
         subtag: "LU",
         description: "Luxembourg",
         added: "2005-10-16"}).
iana(region,'LV',tag{
         type: "region",
         subtag: "LV",
         description: "Latvia",
         added: "2005-10-16"}).
iana(region,'LY',tag{
         type: "region",
         subtag: "LY",
         description: "Libya",
         added: "2005-10-16"}).
iana(region,'MA',tag{
         type: "region",
         subtag: "MA",
         description: "Morocco",
         added: "2005-10-16"}).
iana(region,'MC',tag{
         type: "region",
         subtag: "MC",
         description: "Monaco",
         added: "2005-10-16"}).
iana(region,'MD',tag{
         type: "region",
         subtag: "MD",
         description: "Moldova",
         added: "2005-10-16"}).
iana(region,'ME',tag{
         type: "region",
         subtag: "ME",
         description: "Montenegro",
         added: "2006-10-05"}).
iana(region,'MF',tag{
         type: "region",
         subtag: "MF",
         description: "Saint Martin (French part)",
         added: "2007-11-02"}).
iana(region,'MG',tag{
         type: "region",
         subtag: "MG",
         description: "Madagascar",
         added: "2005-10-16"}).
iana(region,'MH',tag{
         type: "region",
         subtag: "MH",
         description: "Marshall Islands",
         added: "2005-10-16"}).
iana(region,'MK',tag{
         type: "region",
         subtag: "MK",
         description: "North Macedonia",
         added: "2005-10-16"}).
iana(region,'ML',tag{
         type: "region",
         subtag: "ML",
         description: "Mali",
         added: "2005-10-16"}).
iana(region,'MM',tag{
         type: "region",
         subtag: "MM",
         description: "Myanmar",
         added: "2005-10-16"}).
iana(region,'MN',tag{
         type: "region",
         subtag: "MN",
         description: "Mongolia",
         added: "2005-10-16"}).
iana(region,'MO',tag{
         type: "region",
         subtag: "MO",
         description: "Macao",
         added: "2005-10-16"}).
iana(region,'MP',tag{
         type: "region",
         subtag: "MP",
         description: "Northern Mariana Islands",
         added: "2005-10-16"}).
iana(region,'MQ',tag{
         type: "region",
         subtag: "MQ",
         description: "Martinique",
         added: "2005-10-16"}).
iana(region,'MR',tag{
         type: "region",
         subtag: "MR",
         description: "Mauritania",
         added: "2005-10-16"}).
iana(region,'MS',tag{
         type: "region",
         subtag: "MS",
         description: "Montserrat",
         added: "2005-10-16"}).
iana(region,'MT',tag{
         type: "region",
         subtag: "MT",
         description: "Malta",
         added: "2005-10-16"}).
iana(region,'MU',tag{
         type: "region",
         subtag: "MU",
         description: "Mauritius",
         added: "2005-10-16"}).
iana(region,'MV',tag{
         type: "region",
         subtag: "MV",
         description: "Maldives",
         added: "2005-10-16"}).
iana(region,'MW',tag{
         type: "region",
         subtag: "MW",
         description: "Malawi",
         added: "2005-10-16"}).
iana(region,'MX',tag{
         type: "region",
         subtag: "MX",
         description: "Mexico",
         added: "2005-10-16"}).
iana(region,'MY',tag{
         type: "region",
         subtag: "MY",
         description: "Malaysia",
         added: "2005-10-16"}).
iana(region,'MZ',tag{
         type: "region",
         subtag: "MZ",
         description: "Mozambique",
         added: "2005-10-16"}).
iana(region,'NA',tag{
         type: "region",
         subtag: "NA",
         description: "Namibia",
         added: "2005-10-16"}).
iana(region,'NC',tag{
         type: "region",
         subtag: "NC",
         description: "New Caledonia",
         added: "2005-10-16"}).
iana(region,'NE',tag{
         type: "region",
         subtag: "NE",
         description: "Niger",
         added: "2005-10-16"}).
iana(region,'NF',tag{
         type: "region",
         subtag: "NF",
         description: "Norfolk Island",
         added: "2005-10-16"}).
iana(region,'NG',tag{
         type: "region",
         subtag: "NG",
         description: "Nigeria",
         added: "2005-10-16"}).
iana(region,'NI',tag{
         type: "region",
         subtag: "NI",
         description: "Nicaragua",
         added: "2005-10-16"}).
iana(region,'NL',tag{
         type: "region",
         subtag: "NL",
         description: "Netherlands",
         added: "2005-10-16"}).
iana(region,'NO',tag{
         type: "region",
         subtag: "NO",
         description: "Norway",
         added: "2005-10-16"}).
iana(region,'NP',tag{
         type: "region",
         subtag: "NP",
         description: "Nepal",
         added: "2005-10-16"}).
iana(region,'NR',tag{
         type: "region",
         subtag: "NR",
         description: "Nauru",
         added: "2005-10-16"}).
iana(region,'NT',tag{
         type: "region",
         subtag: "NT",
         description: "Neutral Zone",
         added: "2005-10-16",
         deprecated: "1993-07-12"}).
iana(region,'NU',tag{
         type: "region",
         subtag: "NU",
         description: "Niue",
         added: "2005-10-16"}).
iana(region,'NZ',tag{
         type: "region",
         subtag: "NZ",
         description: "New Zealand",
         added: "2005-10-16"}).
iana(region,'OM',tag{
         type: "region",
         subtag: "OM",
         description: "Oman",
         added: "2005-10-16"}).
iana(region,'PA',tag{
         type: "region",
         subtag: "PA",
         description: "Panama",
         added: "2005-10-16"}).
iana(region,'PE',tag{
         type: "region",
         subtag: "PE",
         description: "Peru",
         added: "2005-10-16"}).
iana(region,'PF',tag{
         type: "region",
         subtag: "PF",
         description: "French Polynesia",
         added: "2005-10-16"}).
iana(region,'PG',tag{
         type: "region",
         subtag: "PG",
         description: "Papua New Guinea",
         added: "2005-10-16"}).
iana(region,'PH',tag{
         type: "region",
         subtag: "PH",
         description: "Philippines",
         added: "2005-10-16"}).
iana(region,'PK',tag{
         type: "region",
         subtag: "PK",
         description: "Pakistan",
         added: "2005-10-16"}).
iana(region,'PL',tag{
         type: "region",
         subtag: "PL",
         description: "Poland",
         added: "2005-10-16"}).
iana(region,'PM',tag{
         type: "region",
         subtag: "PM",
         description: "Saint Pierre and Miquelon",
         added: "2005-10-16"}).
iana(region,'PN',tag{
         type: "region",
         subtag: "PN",
         description: "Pitcairn",
         added: "2005-10-16"}).
iana(region,'PR',tag{
         type: "region",
         subtag: "PR",
         description: "Puerto Rico",
         added: "2005-10-16"}).
iana(region,'PS',tag{
         type: "region",
         subtag: "PS",
         description: "State of Palestine",
         added: "2005-10-16"}).
iana(region,'PT',tag{
         type: "region",
         subtag: "PT",
         description: "Portugal",
         added: "2005-10-16"}).
iana(region,'PW',tag{
         type: "region",
         subtag: "PW",
         description: "Palau",
         added: "2005-10-16"}).
iana(region,'PY',tag{
         type: "region",
         subtag: "PY",
         description: "Paraguay",
         added: "2005-10-16"}).
iana(region,'QA',tag{
         type: "region",
         subtag: "QA",
         description: "Qatar",
         added: "2005-10-16"}).
iana(region,'QM',tag{
         type: "region",
         subtag: "QM..QZ",
         description: "Private use",
         added: "2005-10-16"}).
iana(region,'RE',tag{
         type: "region",
         subtag: "RE",
         description: "Réunion",
         added: "2005-10-16"}).
iana(region,'RO',tag{
         type: "region",
         subtag: "RO",
         description: "Romania",
         added: "2005-10-16"}).
iana(region,'RS',tag{
         type: "region",
         subtag: "RS",
         description: "Serbia",
         added: "2006-10-05"}).
iana(region,'RU',tag{
         type: "region",
         subtag: "RU",
         description: "Russian Federation",
         added: "2005-10-16"}).
iana(region,'RW',tag{
         type: "region",
         subtag: "RW",
         description: "Rwanda",
         added: "2005-10-16"}).
iana(region,'SA',tag{
         type: "region",
         subtag: "SA",
         description: "Saudi Arabia",
         added: "2005-10-16"}).
iana(region,'SB',tag{
         type: "region",
         subtag: "SB",
         description: "Solomon Islands",
         added: "2005-10-16"}).
iana(region,'SC',tag{
         type: "region",
         subtag: "SC",
         description: "Seychelles",
         added: "2005-10-16"}).
iana(region,'SD',tag{
         type: "region",
         subtag: "SD",
         description: "Sudan",
         added: "2005-10-16"}).
iana(region,'SE',tag{
         type: "region",
         subtag: "SE",
         description: "Sweden",
         added: "2005-10-16"}).
iana(region,'SG',tag{
         type: "region",
         subtag: "SG",
         description: "Singapore",
         added: "2005-10-16"}).
iana(region,'SH',tag{
         type: "region",
         subtag: "SH",
         description: "Saint Helena, Ascension and Tristan da Cunha",
         added: "2005-10-16"}).
iana(region,'SI',tag{
         type: "region",
         subtag: "SI",
         description: "Slovenia",
         added: "2005-10-16"}).
iana(region,'SJ',tag{
         type: "region",
         subtag: "SJ",
         description: "Svalbard and Jan Mayen",
         added: "2005-10-16"}).
iana(region,'SK',tag{
         type: "region",
         subtag: "SK",
         description: "Slovakia",
         added: "2005-10-16"}).
iana(region,'SL',tag{
         type: "region",
         subtag: "SL",
         description: "Sierra Leone",
         added: "2005-10-16"}).
iana(region,'SM',tag{
         type: "region",
         subtag: "SM",
         description: "San Marino",
         added: "2005-10-16"}).
iana(region,'SN',tag{
         type: "region",
         subtag: "SN",
         description: "Senegal",
         added: "2005-10-16"}).
iana(region,'SO',tag{
         type: "region",
         subtag: "SO",
         description: "Somalia",
         added: "2005-10-16"}).
iana(region,'SR',tag{
         type: "region",
         subtag: "SR",
         description: "Suriname",
         added: "2005-10-16"}).
iana(region,'SS',tag{
         type: "region",
         subtag: "SS",
         description: "South Sudan",
         added: "2011-08-25"}).
iana(region,'ST',tag{
         type: "region",
         subtag: "ST",
         description: "Sao Tome and Principe",
         added: "2005-10-16"}).
iana(region,'SU',tag{
         type: "region",
         subtag: "SU",
         description: "Union of Soviet Socialist Republics",
         added: "2005-10-16",
         deprecated: "1992-08-30"}).
iana(region,'SV',tag{
         type: "region",
         subtag: "SV",
         description: "El Salvador",
         added: "2005-10-16"}).
iana(region,'SX',tag{
         type: "region",
         subtag: "SX",
         description: "Sint Maarten (Dutch part)",
         added: "2011-01-07"}).
iana(region,'SY',tag{
         type: "region",
         subtag: "SY",
         description: "Syrian Arab Republic",
         added: "2005-10-16"}).
iana(region,'SZ',tag{
         type: "region",
         subtag: "SZ",
         description: ["Eswatini","eSwatini","Swaziland"],
         added: "2005-10-16"}).
iana(region,'TA',tag{
         type: "region",
         subtag: "TA",
         description: "Tristan da Cunha",
         added: "2009-07-29"}).
iana(region,'TC',tag{
         type: "region",
         subtag: "TC",
         description: "Turks and Caicos Islands",
         added: "2005-10-16"}).
iana(region,'TD',tag{
         type: "region",
         subtag: "TD",
         description: "Chad",
         added: "2005-10-16"}).
iana(region,'TF',tag{
         type: "region",
         subtag: "TF",
         description: "French Southern Territories",
         added: "2005-10-16"}).
iana(region,'TG',tag{
         type: "region",
         subtag: "TG",
         description: "Togo",
         added: "2005-10-16"}).
iana(region,'TH',tag{
         type: "region",
         subtag: "TH",
         description: "Thailand",
         added: "2005-10-16"}).
iana(region,'TJ',tag{
         type: "region",
         subtag: "TJ",
         description: "Tajikistan",
         added: "2005-10-16"}).
iana(region,'TK',tag{
         type: "region",
         subtag: "TK",
         description: "Tokelau",
         added: "2005-10-16"}).
iana(region,'TL',tag{
         type: "region",
         subtag: "TL",
         description: "Timor-Leste",
         added: "2005-10-16"}).
iana(region,'TM',tag{
         type: "region",
         subtag: "TM",
         description: "Turkmenistan",
         added: "2005-10-16"}).
iana(region,'TN',tag{
         type: "region",
         subtag: "TN",
         description: "Tunisia",
         added: "2005-10-16"}).
iana(region,'TO',tag{
         type: "region",
         subtag: "TO",
         description: "Tonga",
         added: "2005-10-16"}).
iana(region,'TP',tag{
         type: "region",
         subtag: "TP",
         description: "East Timor",
         added: "2005-10-16",
         deprecated: "2002-05-20",
         preferred_value: "TL"}).
iana(region,'TR',tag{
         type: "region",
         subtag: "TR",
         description: ["Türkiye","Turkey"],
         added: "2005-10-16"}).
iana(region,'TT',tag{
         type: "region",
         subtag: "TT",
         description: "Trinidad and Tobago",
         added: "2005-10-16"}).
iana(region,'TV',tag{
         type: "region",
         subtag: "TV",
         description: "Tuvalu",
         added: "2005-10-16"}).
iana(region,'TW',tag{
         type: "region",
         subtag: "TW",
         description: "Taiwan, Province of China",
         added: "2005-10-16"}).
iana(region,'TZ',tag{
         type: "region",
         subtag: "TZ",
         description: "United Republic of Tanzania",
         added: "2005-10-16"}).
iana(region,'UA',tag{
         type: "region",
         subtag: "UA",
         description: "Ukraine",
         added: "2005-10-16"}).
iana(region,'UG',tag{
         type: "region",
         subtag: "UG",
         description: "Uganda",
         added: "2005-10-16"}).
iana(region,'UM',tag{
         type: "region",
         subtag: "UM",
         description: "United States Minor Outlying Islands",
         added: "2005-10-16"}).
iana(region,'UN',tag{
         type: "region",
         subtag: "UN",
         description: "United Nations",
         added: "2016-07-14"}).
iana(region,'US',tag{
         type: "region",
         subtag: "US",
         description: "United States",
         added: "2005-10-16"}).
iana(region,'UY',tag{
         type: "region",
         subtag: "UY",
         description: "Uruguay",
         added: "2005-10-16"}).
iana(region,'UZ',tag{
         type: "region",
         subtag: "UZ",
         description: "Uzbekistan",
         added: "2005-10-16"}).
iana(region,'VA',tag{
         type: "region",
         subtag: "VA",
         description: "Holy See (Vatican City State)",
         added: "2005-10-16"}).
iana(region,'VC',tag{
         type: "region",
         subtag: "VC",
         description: "Saint Vincent and the Grenadines",
         added: "2005-10-16"}).
iana(region,'VE',tag{
         type: "region",
         subtag: "VE",
         description: "Venezuela",
         added: "2005-10-16"}).
iana(region,'VG',tag{
         type: "region",
         subtag: "VG",
         description: "British Virgin Islands",
         added: "2005-10-16"}).
iana(region,'VI',tag{
         type: "region",
         subtag: "VI",
         description: "U.S. Virgin Islands",
         added: "2005-10-16"}).
iana(region,'VN',tag{
         type: "region",
         subtag: "VN",
         description: "Viet Nam",
         added: "2005-10-16"}).
iana(region,'VU',tag{
         type: "region",
         subtag: "VU",
         description: "Vanuatu",
         added: "2005-10-16"}).
iana(region,'WF',tag{
         type: "region",
         subtag: "WF",
         description: "Wallis and Futuna",
         added: "2005-10-16"}).
iana(region,'WS',tag{
         type: "region",
         subtag: "WS",
         description: "Samoa",
         added: "2005-10-16"}).
iana(region,'XA',tag{
         type: "region",
         subtag: "XA..XZ",
         description: "Private use",
         added: "2005-10-16"}).
iana(region,'YD',tag{
         type: "region",
         subtag: "YD",
         description: "Democratic Yemen",
         added: "2005-10-16",
         deprecated: "1990-08-14",
         preferred_value: "YE"}).
iana(region,'YE',tag{
         type: "region",
         subtag: "YE",
         description: "Yemen",
         added: "2005-10-16"}).
iana(region,'YT',tag{
         type: "region",
         subtag: "YT",
         description: "Mayotte",
         added: "2005-10-16"}).
iana(region,'YU',tag{
         type: "region",
         subtag: "YU",
         description: "Yugoslavia",
         added: "2005-10-16",
         deprecated: "2003-07-23",
         comments: "see BA, HR, ME, MK, RS, or SI"}).
iana(region,'ZA',tag{
         type: "region",
         subtag: "ZA",
         description: "South Africa",
         added: "2005-10-16"}).
iana(region,'ZM',tag{
         type: "region",
         subtag: "ZM",
         description: "Zambia",
         added: "2005-10-16"}).
iana(region,'ZR',tag{
         type: "region",
         subtag: "ZR",
         description: "Zaire",
         added: "2005-10-16",
         deprecated: "1997-07-14",
         preferred_value: "CD"}).
iana(region,'ZW',tag{
         type: "region",
         subtag: "ZW",
         description: "Zimbabwe",
         added: "2005-10-16"}).
iana(region,'ZZ',tag{
         type: "region",
         subtag: "ZZ",
         description: "Private use",
         added: "2005-10-16"}).
iana(region,'001',tag{
         type: "region",
         subtag: "001",
         description: "World",
         added: "2005-10-16"}).
iana(region,'002',tag{
         type: "region",
         subtag: "002",
         description: "Africa",
         added: "2005-10-16"}).
iana(region,'003',tag{
         type: "region",
         subtag: "003",
         description: "North America",
         added: "2010-08-16",
         comments: "Includes Northern America (021), Caribbean (029), and Central America (013); see also 021"
     }).
iana(region,'005',tag{
         type: "region",
         subtag: "005",
         description: "South America",
         added: "2005-10-16"}).
iana(region,'009',tag{
         type: "region",
         subtag: "009",
         description: "Oceania",
         added: "2005-10-16"}).
iana(region,'011',tag{
         type: "region",
         subtag: "011",
         description: "Western Africa",
         added: "2005-10-16"}).
iana(region,'013',tag{
         type: "region",
         subtag: "013",
         description: "Central America",
         added: "2005-10-16"}).
iana(region,'014',tag{
         type: "region",
         subtag: "014",
         description: "Eastern Africa",
         added: "2005-10-16"}).
iana(region,'015',tag{
         type: "region",
         subtag: "015",
         description: "Northern Africa",
         added: "2005-10-16"}).
iana(region,'017',tag{
         type: "region",
         subtag: "017",
         description: "Middle Africa",
         added: "2005-10-16"}).
iana(region,'018',tag{
         type: "region",
         subtag: "018",
         description: "Southern Africa",
         added: "2005-10-16"}).
iana(region,'019',tag{
         type: "region",
         subtag: "019",
         description: "Americas",
         added: "2005-10-16"}).
iana(region,'021',tag{
         type: "region",
         subtag: "021",
         description: "Northern America",
         added: "2005-10-16",
         comments: "Does not include Caribbean (029) or Central America (013); see also 003"
     }).
iana(region,'029',tag{
         type: "region",
         subtag: "029",
         description: "Caribbean",
         added: "2005-10-16"}).
iana(region,'030',tag{
         type: "region",
         subtag: "030",
         description: "Eastern Asia",
         added: "2005-10-16"}).
iana(region,'034',tag{
         type: "region",
         subtag: "034",
         description: "Southern Asia",
         added: "2005-10-16"}).
iana(region,'035',tag{
         type: "region",
         subtag: "035",
         description: "South-Eastern Asia",
         added: "2005-10-16"}).
iana(region,'039',tag{
         type: "region",
         subtag: "039",
         description: "Southern Europe",
         added: "2005-10-16"}).
iana(region,'053',tag{
         type: "region",
         subtag: "053",
         description: "Australia and New Zealand",
         added: "2005-10-16"}).
iana(region,'054',tag{
         type: "region",
         subtag: "054",
         description: "Melanesia",
         added: "2005-10-16"}).
iana(region,'057',tag{
         type: "region",
         subtag: "057",
         description: "Micronesia",
         added: "2005-10-16"}).
iana(region,'061',tag{
         type: "region",
         subtag: "061",
         description: "Polynesia",
         added: "2005-10-16"}).
iana(region,'142',tag{
         type: "region",
         subtag: "142",
         description: "Asia",
         added: "2005-10-16"}).
iana(region,'143',tag{
         type: "region",
         subtag: "143",
         description: "Central Asia",
         added: "2005-10-16"}).
iana(region,'145',tag{
         type: "region",
         subtag: "145",
         description: "Western Asia",
         added: "2005-10-16"}).
iana(region,'150',tag{
         type: "region",
         subtag: "150",
         description: "Europe",
         added: "2005-10-16"}).
iana(region,'151',tag{
         type: "region",
         subtag: "151",
         description: "Eastern Europe",
         added: "2005-10-16"}).
iana(region,'154',tag{
         type: "region",
         subtag: "154",
         description: "Northern Europe",
         added: "2005-10-16"}).
iana(region,'155',tag{
         type: "region",
         subtag: "155",
         description: "Western Europe",
         added: "2005-10-16"}).
iana(region,'202',tag{
         type: "region",
         subtag: "202",
         description: "Sub-Saharan Africa",
         added: "2017-04-18"}).
iana(region,'419',tag{
         type: "region",
         subtag: "419",
         description: "Latin America and the Caribbean",
         added: "2005-10-16"}).
iana(variant,'1606nict',tag{
         type: "variant",
         subtag: "1606nict",
         description: "Late Middle French (to 1606)",
         added: "2007-03-20",
         prefix: "frm",
         comments: "16th century French as in Jean Nicot, \"Thresor de la langue francoyse\", 1606, but also including some French similar to that of Rabelais"
     }).
iana(variant,'1694acad',tag{
         type: "variant",
         subtag: "1694acad",
         description: "Early Modern French",
         added: "2007-03-20",
         prefix: "fr",
         comments: "17th century French, as catalogued in the \"Dictionnaire de\", l'académie françoise 4eme ed. 1694; frequently includes elements of Middle French, as this is a transitional period"
     }).
iana(variant,'1901',tag{
         type: "variant",
         subtag: "1901",
         description: "Traditional German orthography",
         added: "2005-10-16",
         prefix: "de"}).
iana(variant,'1959acad',tag{
         type: "variant",
         subtag: "1959acad",
         description: "Academic\" (\"governmental\") variant of Belarusian as codified in 1959",
         added: "2008-09-30",
         prefix: "be"}).
iana(variant,'1994',tag{
         type: "variant",
         subtag: "1994",
         description: "Standardized Resian orthography",
         added: "2007-07-28",
         prefix: ["sl-rozaj","sl-rozaj-biske","sl-rozaj-njiva","sl-rozaj-osojs","sl-rozaj-solba"],
         comments: "For standardized Resian an orthography was published in 1994."}).
iana(variant,'1996',tag{
         type: "variant",
         subtag: "1996",
         description: "German orthography of 1996",
         added: "2005-10-16",
         prefix: "de"}).
iana(variant,abl1943,tag{
         type: "variant",
         subtag: "abl1943",
         description: "Orthographic formulation of 1943 - Official in Brazil (Formulário Ortográfico de 1943 - Oficial no Brasil)",
         added: "2015-05-06",
         prefix: "pt-BR",
         comments: "Denotes conventions established by the Academia Brasileira de Letras in 1943 and generally used in Brazil until 2009"}).
iana(variant,akuapem,tag{
         type: "variant",
         subtag: "akuapem",
         description: "Akuapem Twi",
         added: "2017-06-05",
         prefix: "tw"}).
iana(variant,alalc97,tag{
         type: "variant",
         subtag: "alalc97",
         description: "ALA-LC Romanization, 1997 edition",
         added: "2009-12-09",
         comments: "Romanizations recommended by the American Library Association and the Library of Congress, in \"ALA-LC Romanization Tables: Transliteration Schemes for Non-Roman Scripts\" (1997), ISBN
  978-0-8444-0940-5."
}).
iana(variant,aluku,tag{
         type: "variant",
         subtag: "aluku",
         description: ["Aluku dialect","Boni dialect"],
         added: "2009-09-05",
         prefix: "djk",
         comments: "Aluku dialect of the \"Busi Nenge Tongo\" English-based Creole continuum in Eastern Suriname and Western French Guiana"
}).
iana(variant,ao1990,tag{
         type: "variant",
         subtag: "ao1990",
         description: "Portuguese Language Orthographic Agreement of 1990 (Acordo Ortográfico da Língua Portuguesa de 1990)",
         added: "2015-05-06",
         prefix: ["pt","gl"],
         comments: "Portuguese orthography conventions established in 1990 but not brought into effect until 2009"
}).
iana(variant,aranes,tag{
         type: "variant",
         subtag: "aranes",
         description: "Aranese",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in the Val d'Aran"}).
iana(variant,arevela,tag{
         type: "variant",
         subtag: "arevela",
         description: "Eastern Armenian",
         added: "2006-09-18",
         deprecated: "2018-03-24",
         prefix: "hy",
         comments: "Preferred tag is hy"}).
iana(variant,arevmda,tag{
         type: "variant",
         subtag: "arevmda",
         description: "Western Armenian",
         added: "2006-09-18",
         deprecated: "2018-03-24",
         prefix: "hy",
         comments: "Preferred tag is hyw"}).
iana(variant,arkaika,tag{
         type: "variant",
         subtag: "arkaika",
         description: ["Arcaicam Esperantom","Arkaika Esperanto"],
         added: "2020-12-17",
         prefix: "eo",
         comments: "Archaic Esperanto variant devised by Manuel Halvelik"}).
iana(variant,asante,tag{
         type: "variant",
         subtag: "asante",
         description: ["Asante Twi","Ashanti Twi"],
         added: "2017-06-05",
         prefix: "tw"}).
iana(variant,auvern,tag{
         type: "variant",
         subtag: "auvern",
         description: "Auvergnat",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Auvergne"}).
iana(variant,baku1926,tag{
         type: "variant",
         subtag: "baku1926",
         description: "Unified Turkic Latin Alphabet (Historical)",
         added: "2007-04-18",
         prefix: ["az","ba","crh","kk","krc","ky","sah","tk","tt","uz"],
         comments: "Denotes alphabet used in Turkic republics/regions of the former USSR in late 1920s, and throughout 1930s, which aspired to represent equivalent phonemes in a unified fashion. Also known as: \"New\", Turkic Alphabet; Birlәşdirilmiş Jeni Tyrk Әlifbasь (Birlesdirilmis Jeni Tyrk Elifbasi); Jaŋalif (Janalif)."}).
iana(variant,balanka,tag{
         type: "variant",
         subtag: "balanka",
         description: "The Balanka dialect of Anii",
         added: "2014-02-15",
         prefix: "blo",
         comments: "Balanka is one of 19 Anii dialects."}).
iana(variant,barla,tag{
         type: "variant",
         subtag: "barla",
         description: "The Barlavento dialect group of Kabuverdianu",
         prefix: "kea",
         comments: "Barlavento is one of the two main dialect groups of Kabuverdianu.",
         added: "2013-12-10"}).
iana(variant,basiceng,tag{
         type: "variant",
         subtag: "basiceng",
         description: "Basic English",
         added: "2015-12-29",
         prefix: "en"}).
iana(variant,bauddha,tag{
         type: "variant",
         subtag: "bauddha",
         description: "Buddhist Hybrid Sanskrit",
         added: "2010-07-28",
         prefix: "sa"}).
iana(variant,biscayan,tag{
         type: "variant",
         subtag: "biscayan",
         description: "Biscayan dialect of Basque",
         added: "2010-04-13",
         prefix: "eu"}).
iana(variant,biske,tag{
         type: "variant",
         subtag: "biske",
         description: ["The San Giorgio dialect of Resian","The Bila dialect of Resian"],
         added: "2007-07-05",
         prefix: "sl-rozaj",
         comments: "The dialect of San Giorgio/Bila is one of the four major local dialects of Resian"
}).
iana(variant,bohoric,tag{
         type: "variant",
         subtag: "bohoric",
         description: "Slovene in Bohorič alphabet",
         added: "2012-06-27",
         prefix: "sl",
         comments: "The subtag represents the alphabet codified by Adam Bohorič in 1584 and used from the first printed Slovene book and up to the mid-19th century."
}).
iana(variant,boont,tag{
         type: "variant",
         subtag: "boont",
         description: "Boontling",
         added: "2006-09-18",
         prefix: "en",
         comments: "Jargon embedded in American English"}).
iana(variant,bornholm,tag{
         type: "variant",
         subtag: "bornholm",
         description: "Bornholmsk",
         added: "2019-03-27",
         prefix: "da"}).
iana(variant,cisaup,tag{
         type: "variant",
         subtag: "cisaup",
         description: "Cisalpine",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in northwestern Italy"}).
iana(variant,colb1945,tag{
         type: "variant",
         subtag: "colb1945",
         description: "Portuguese-Brazilian Orthographic Convention of 1945 (Convenção Ortográfica Luso-Brasileira de 1945)",
         added: "2015-05-06",
         prefix: "pt",
         comments: "Portuguese orthography conventions established in 1945, generally in effect until 2009. This reform was not ratified in Brazil."
}).
iana(variant,cornu,tag{
         type: "variant",
         subtag: "cornu",
         description: ["Cornu-English","Cornish English","Anglo-Cornish"],
         added: "2015-12-07",
         prefix: "en"}).
iana(variant,creiss,tag{
         type: "variant",
         subtag: "creiss",
         description: "Occitan variants of the Croissant area",
         added: "2018-04-22",
         prefix: "oc"}).
iana(variant,dajnko,tag{
         type: "variant",
         subtag: "dajnko",
         description: "Slovene in Dajnko alphabet",
         added: "2012-06-27",
         prefix: "sl",
         comments: "The subtag represents the alphabet codified by Peter Dajnko and used from 1824 to 1839 mostly in Styria (in what is now Eastern Slovenia)."
}).
iana(variant,ekavsk,tag{
         type: "variant",
         subtag: "ekavsk",
         description: "Serbian with Ekavian pronunciation",
         prefix: ["sr","sr-Latn","sr-Cyrl"],
         added: "2013-12-02"}).
iana(variant,emodeng,tag{
         type: "variant",
         subtag: "emodeng",
         description: "Early Modern English (1500-1700)",
         added: "2012-02-05",
         prefix: "en"}).
iana(variant,fonipa,tag{
         type: "variant",
         subtag: "fonipa",
         description: "International Phonetic Alphabet",
         added: "2006-12-11"}).
iana(variant,fonkirsh,tag{
         type: "variant",
         subtag: "fonkirsh",
         description: "Kirshenbaum Phonetic Alphabet",
         added: "2018-04-22"}).
iana(variant,fonnapa,tag{
         type: "variant",
         subtag: "fonnapa",
         description: ["North American Phonetic Alphabet","Americanist Phonetic Notation"],
         added: "2016-06-24"}).
iana(variant,fonupa,tag{
         type: "variant",
         subtag: "fonupa",
         description: "Uralic Phonetic Alphabet",
         added: "2006-12-11"}).
iana(variant,fonxsamp,tag{
         type: "variant",
         subtag: "fonxsamp",
         description: "X-SAMPA transcription",
         added: "2010-10-23",
         comments: "Indicates that the content is transcribed according to X-SAMPA"}).
iana(variant,gallo,tag{
         type: "variant",
         subtag: "gallo",
         description: "Gallo",
         added: "2021-08-05",
         prefix: "fr"}).
iana(variant,gascon,tag{
         type: "variant",
         subtag: "gascon",
         description: "Gascon",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Gascony"}).
iana(variant,grclass,tag{
         type: "variant",
         subtag: "grclass",
         description: "Classical Occitan orthography",
         added: "2018-04-22",
         prefix: ["oc","oc-aranes","oc-auvern","oc-cisaup","oc-creiss","oc-gascon","oc-lemosin","oc-lengadoc","oc-nicard","oc-provenc","oc-vivaraup"],
         comments: "Classical written standard for Occitan developed in 1935 by Alibèrt"
     }).
iana(variant,grital,tag{
         type: "variant",
         subtag: "grital",
         description: "Italian-inspired Occitan orthography",
         added: "2018-04-22",
         prefix: ["oc","oc-cisaup","oc-nicard","oc-provenc"]}).
iana(variant,grmistr,tag{
         type: "variant",
         subtag: "grmistr",
         description: "Mistralian or Mistralian-inspired Occitan orthography",
         added: "2018-04-22",
         prefix: ["oc","oc-aranes","oc-auvern","oc-cisaup","oc-creiss","oc-gascon","oc-lemosin","oc-lengadoc","oc-nicard","oc-provenc","oc-vivaraup"],
         comments: "Written standard developed by Romanilha in 1853 and used by Mistral and the Félibres, including derived standards such as Escolo dóu Po, Escolo Gaston Febus, and others"
}).
iana(variant,hepburn,tag{
         type: "variant",
         subtag: "hepburn",
         description: "Hepburn romanization",
         added: "2009-10-01",
         prefix: "ja-Latn"}).
iana(variant,heploc,tag{
         type: "variant",
         subtag: "heploc",
         description: "Hepburn romanization, Library of Congress method",
         added: "2009-10-01",
         deprecated: "2010-02-07",
         preferred_value: "alalc97",
         prefix: "ja-Latn-hepburn",
         comments: "Preferred tag is ja-Latn-alalc97"}).
iana(variant,hognorsk,tag{
         type: "variant",
         subtag: "hognorsk",
         description: "Norwegian in Høgnorsk (High Norwegian) orthography",
         added: "2010-01-02",
         prefix: "nn",
         comments: "Norwegian following Ivar Aasen's orthographical principles, including modern usage."
     }).
iana(variant,hsistemo,tag{
         type: "variant",
         subtag: "hsistemo",
         description: "Standard H-system orthographic fallback for spelling Esperanto",
         added: "2017-03-14",
         prefix: "eo"}).
iana(variant,ijekavsk,tag{
         type: "variant",
         subtag: "ijekavsk",
         description: "Serbian with Ijekavian pronunciation",
         prefix: ["sr","sr-Latn","sr-Cyrl"],
         added: "2013-12-02"}).
iana(variant,itihasa,tag{
         type: "variant",
         subtag: "itihasa",
         description: "Epic Sanskrit",
         added: "2010-07-28",
         prefix: "sa"}).
iana(variant,ivanchov,tag{
         type: "variant",
         subtag: "ivanchov",
         description: "Bulgarian in 1899 orthography",
         added: "2017-12-13",
         prefix: "bg",
         comments: "Bulgarian orthography introduced by Todor Ivanchov in 1899"}).
iana(variant,jauer,tag{
         type: "variant",
         subtag: "jauer",
         description: "Jauer dialect of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "The spoken dialect of the Val Müstair, which has no written standard."
     }).
iana(variant,jyutping,tag{
         type: "variant",
         subtag: "jyutping",
         description: "Jyutping Cantonese Romanization",
         added: "2010-10-23",
         prefix: "yue",
         comments: "Jyutping romanization of Cantonese"}).
iana(variant,kkcor,tag{
         type: "variant",
         subtag: "kkcor",
         description: "Common Cornish orthography of Revived Cornish",
         added: "2008-10-14",
         prefix: "kw"}).
iana(variant,kociewie,tag{
         type: "variant",
         subtag: "kociewie",
         description: "The Kociewie dialect of Polish",
         added: "2014-11-27",
         prefix: "pl",
         comments: "The dialect of Kociewie is spoken in the region around Starogard Gdański, Tczew and Świecie in northern Poland."
     }).
iana(variant,kscor,tag{
         type: "variant",
         subtag: "kscor",
         description: ["Standard Cornish orthography of Revived Cornish","Kernowek Standard"],
         added: "2012-06-27",
         prefix: "kw"}).
iana(variant,laukika,tag{
         type: "variant",
         subtag: "laukika",
         description: "Classical Sanskrit",
         added: "2010-07-28",
         prefix: "sa"}).
iana(variant,lemosin,tag{
         type: "variant",
         subtag: "lemosin",
         description: "Limousin",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Limousin"}).
iana(variant,lengadoc,tag{
         type: "variant",
         subtag: "lengadoc",
         description: "Languedocien",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Languedoc"}).
iana(variant,lipaw,tag{
         type: "variant",
         subtag: "lipaw",
         description: ["The Lipovaz dialect of Resian","The Lipovec dialect of Resian"],
         added: "2007-08-11",
         prefix: "sl-rozaj",
         comments: "The dialect of Lipovaz/Lipovec is one of the minor local dialects of Resian"
}).
iana(variant,ltg1929,tag{
         type: "variant",
         subtag: "ltg1929",
         description: "The Latgalian language orthography codified in 1929",
         added: "2022-08-05",
         prefix: "ltg"}).
iana(variant,ltg2007,tag{
         type: "variant",
         subtag: "ltg2007",
         description: "The Latgalian language orthography codified in the language law in 2007",
         added: "2022-06-23",
         prefix: "ltg"}).
iana(variant,luna1918,tag{
         type: "variant",
         subtag: "luna1918",
         description: "Post-1917 Russian orthography",
         added: "2010-10-10",
         prefix: "ru",
         comments: "Russian orthography as established by the 1917/1918 orthographic reforms"
     }).
iana(variant,metelko,tag{
         type: "variant",
         subtag: "metelko",
         description: "Slovene in Metelko alphabet",
         added: "2012-06-27",
         prefix: "sl",
         comments: "The subtag represents the alphabet codified by Franc Serafin Metelko and used from 1825 to 1833."
}).
iana(variant,monoton,tag{
         type: "variant",
         subtag: "monoton",
         description: "Monotonic Greek",
         added: "2006-12-11",
         prefix: "el"}).
iana(variant,ndyuka,tag{
         type: "variant",
         subtag: "ndyuka",
         description: ["Ndyuka dialect","Aukan dialect"],
         added: "2009-09-05",
         prefix: "djk",
         comments: "Ndyuka dialect of the \"Busi Nenge Tongo\" English-based Creole continuum in Eastern Suriname and Western French Guiana"
     }).
iana(variant,nedis,tag{
         type: "variant",
         subtag: "nedis",
         description: ["Natisone dialect","Nadiza dialect"],
         added: "2005-10-16",
         prefix: "sl"}).
iana(variant,newfound,tag{
         type: "variant",
         subtag: "newfound",
         description: "Newfoundland English",
         added: "2015-11-25",
         prefix: "en-CA"}).
iana(variant,nicard,tag{
         type: "variant",
         subtag: "nicard",
         description: "Niçard",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Nice"}).
iana(variant,njiva,tag{
         type: "variant",
         subtag: "njiva",
         description: ["The Gniva dialect of Resian","The Njiva dialect of Resian"],
         added: "2007-07-05",
         prefix: "sl-rozaj",
         comments: "The dialect of Gniva/Njiva is one of the four major local dialects of Resian"
}).
iana(variant,nulik,tag{
         type: "variant",
         subtag: "nulik",
         description: ["Volapük nulik","Volapük perevidöl","Volapük nulädik","de Jong's Volapük","New Volapük","Revised Volapük","Modern Volapük"],
         added: "2012-01-28",
         prefix: "vo"}).
iana(variant,osojs,tag{
         type: "variant",
         subtag: "osojs",
         description: ["The Oseacco dialect of Resian","The Osojane dialect of Resian"],
         added: "2007-07-05",
         prefix: "sl-rozaj",
         comments: "The dialect of Oseacco/Osojane is one of the four major local dialects of Resian"
     }).
iana(variant,oxendict,tag{
         type: "variant",
         subtag: "oxendict",
         description: "Oxford English Dictionary spelling",
         added: "2015-04-17",
         prefix: "en"}).
iana(variant,pahawh2,tag{
         type: "variant",
         subtag: "pahawh2",
         description: "Pahawh Hmong Second Stage Reduced orthography",
         added: "2017-01-13",
         prefix: ["mww","hnj"]}).
iana(variant,pahawh3,tag{
         type: "variant",
         subtag: "pahawh3",
         description: "Pahawh Hmong Third Stage Reduced orthography",
         added: "2017-01-13",
         prefix: ["mww","hnj"]}).
iana(variant,pahawh4,tag{
         type: "variant",
         subtag: "pahawh4",
         description: "Pahawh Hmong Final Version orthography",
         added: "2017-01-13",
         prefix: ["mww","hnj"]}).
iana(variant,pamaka,tag{
         type: "variant",
         subtag: "pamaka",
         description: "Pamaka dialect",
         added: "2009-09-05",
         prefix: "djk",
         comments: "Pamaka dialect of the \"Busi Nenge Tongo\" English-based Creole continuum in Eastern Suriname and Western French Guiana"
}).
iana(variant,peano,tag{
         type: "variant",
         subtag: "peano",
         description: ["Latino Sine Flexione","Interlingua de API","Interlingua de Peano"],
         prefix: "la",
         comments: "Peano’s Interlingua, created in 1903 by Giuseppe Peano as an international auxiliary language",
         added: "2020-03-12"}).
iana(variant,petr1708,tag{
         type: "variant",
         subtag: "petr1708",
         description: "Petrine orthography",
         added: "2010-10-10",
         prefix: "ru",
         comments: "Russian orthography from the Petrine orthographic reforms of 1708 to the 1917 orthographic reform"
     }).
iana(variant,pinyin,tag{
         type: "variant",
         subtag: "pinyin",
         description: "Pinyin romanization",
         added: "2008-10-14",
         prefix: ["zh-Latn","bo-Latn"]}).
iana(variant,polyton,tag{
         type: "variant",
         subtag: "polyton",
         description: "Polytonic Greek",
         added: "2006-12-11",
         prefix: "el"}).
iana(variant,provenc,tag{
         type: "variant",
         subtag: "provenc",
         description: "Provençal",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in Provence"}).
iana(variant,puter,tag{
         type: "variant",
         subtag: "puter",
         description: "Puter idiom of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Puter is one of the five traditional written standards or \"idioms\" of the Romansh language."
}).
iana(variant,rigik,tag{
         type: "variant",
         subtag: "rigik",
         description: ["Volapük rigik","Schleyer's Volapük","Original Volapük","Classic Volapük"],
         added: "2012-01-28",
         prefix: "vo"}).
iana(variant,rozaj,tag{
         type: "variant",
         subtag: "rozaj",
         description: ["Resian","Resianic","Rezijan"],
         added: "2005-10-16",
         prefix: "sl"}).
iana(variant,rumgr,tag{
         type: "variant",
         subtag: "rumgr",
         description: "Rumantsch Grischun",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Supraregional Romansh written standard"}).
iana(variant,scotland,tag{
         type: "variant",
         subtag: "scotland",
         description: "Scottish Standard English",
         added: "2007-08-31",
         prefix: "en"}).
iana(variant,scouse,tag{
         type: "variant",
         subtag: "scouse",
         description: "Scouse",
         added: "2006-09-18",
         prefix: "en",
         comments: "English Liverpudlian dialect known as 'Scouse'"}).
iana(variant,simple,tag{
         type: "variant",
         subtag: "simple",
         description: "Simplified form",
         added: "2015-12-29"}).
iana(variant,solba,tag{
         type: "variant",
         subtag: "solba",
         description: ["The Stolvizza dialect of Resian","The Solbica dialect of Resian"],
         added: "2007-07-05",
         prefix: "sl-rozaj",
         comments: "The dialect of Stolvizza/Solbica is one of the four major local dialects of Resian"
     }).
iana(variant,sotav,tag{
         type: "variant",
         subtag: "sotav",
         description: "The Sotavento dialect group of Kabuverdianu",
         prefix: "kea",
         comments: "Sotavento is one of the two main dialect groups of Kabuverdianu.",
         added: "2013-12-10"}).
iana(variant,spanglis,tag{
         type: "variant",
         subtag: "spanglis",
         description: "Spanglish",
         added: "2017-02-23",
         prefix: ["en","es"],
         comments: "A variety of contact dialects of English and Spanish"}).
iana(variant,surmiran,tag{
         type: "variant",
         subtag: "surmiran",
         description: "Surmiran idiom of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Surmiran is one of the five traditional written standards or \"idioms\" of the Romansh language."
}).
iana(variant,sursilv,tag{
         type: "variant",
         subtag: "sursilv",
         description: "Sursilvan idiom of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Sursilvan is one of the five traditional written standards or \"idioms\" of the Romansh language."
}).
iana(variant,sutsilv,tag{
         type: "variant",
         subtag: "sutsilv",
         description: "Sutsilvan idiom of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Sutsilvan is one of the five traditional written standards or \"idioms\" of the Romansh language."
}).
iana(variant,synnejyl,tag{
         type: "variant",
         subtag: "synnejyl",
         description: ["Synnejysk","South Jutish"],
         added: "2021-07-17",
         prefix: "da"}).
iana(variant,tarask,tag{
         type: "variant",
         subtag: "tarask",
         description: "Belarusian in Taraskievica orthography",
         added: "2007-04-27",
         prefix: "be",
         comments: "The subtag represents Branislau Taraskievic's Belarusian orthography as published in \"Bielaruski klasycny pravapis\" by Juras Buslakou, Vincuk Viacorka, Zmicier Sanko, and Zmicier Sauka (Vilnia-Miensk 2005)."
}).
iana(variant,tongyong,tag{
         type: "variant",
         subtag: "tongyong",
         description: "Tongyong Pinyin romanization",
         added: "2020-06-08",
         prefix: "zh-Latn",
         comments: "Former official transcription standard for Mandarin Chinese in Taiwan."
}).
iana(variant,tunumiit,tag{
         type: "variant",
         subtag: "tunumiit",
         description: ["Tunumiisiut","East Greenlandic","Østgrønlandsk"],
         added: "2020-07-16",
         prefix: "kl",
         comments: "Also known as Tunumiit oraasiat"}).
iana(variant,uccor,tag{
         type: "variant",
         subtag: "uccor",
         description: "Unified Cornish orthography of Revived Cornish",
         added: "2008-10-14",
         prefix: "kw"}).
iana(variant,ucrcor,tag{
         type: "variant",
         subtag: "ucrcor",
         description: "Unified Cornish Revised orthography of Revived Cornish",
         added: "2008-10-14",
         prefix: "kw"}).
iana(variant,ulster,tag{
         type: "variant",
         subtag: "ulster",
         description: "Ulster dialect of Scots",
         added: "2010-04-10",
         prefix: "sco"}).
iana(variant,unifon,tag{
         type: "variant",
         subtag: "unifon",
         description: "Unifon phonetic alphabet",
         added: "2013-10-02",
         prefix: ["en","hup","kyh","tol","yur"]}).
iana(variant,vaidika,tag{
         type: "variant",
         subtag: "vaidika",
         description: "Vedic Sanskrit",
         added: "2010-07-28",
         prefix: "sa",
         comments: "The most ancient dialect of Sanskrit used in verse and prose composed until about the 4th century B.C.E."
}).
iana(variant,valencia,tag{
         type: "variant",
         subtag: "valencia",
         description: "Valencian",
         added: "2007-03-06",
         prefix: "ca",
         comments: "Variety spoken in the \"Comunidad Valenciana\" region of Spain, where it is co-official with Spanish."
}).
iana(variant,vallader,tag{
         type: "variant",
         subtag: "vallader",
         description: "Vallader idiom of Romansh",
         added: "2010-06-29",
         prefix: "rm",
         comments: "Vallader is one of the five traditional written standards or \"idioms\" of the Romansh language."
}).
iana(variant,vecdruka,tag{
         type: "variant",
         subtag: "vecdruka",
         description: "Latvian orthography used before 1920s (\"vecā druka\")",
         added: "2020-09-26",
         prefix: "lv",
         comments: "The subtag represents the old orthography of the Latvian language used during c. 1600s–1920s."
}).
iana(variant,vivaraup,tag{
         type: "variant",
         subtag: "vivaraup",
         description: "Vivaro-Alpine",
         added: "2018-04-22",
         prefix: "oc",
         comments: "Occitan variant spoken in northeastern Occitania"}).
iana(variant,wadegile,tag{
         type: "variant",
         subtag: "wadegile",
         description: "Wade-Giles romanization",
         added: "2008-10-03",
         prefix: "zh-Latn"}).
iana(variant,xsistemo,tag{
         type: "variant",
         subtag: "xsistemo",
         description: "Standard X-system orthographic fallback for spelling Esperanto",
         added: "2017-03-14",
         prefix: "eo"}).
iana(grandfathered,'art-lojban',tag{
         type: "grandfathered",
         tag: "art-lojban",
         description: "Lojban",
         added: "2001-11-11",
         deprecated: "2003-09-02",
         preferred_value: "jbo"}).
iana(grandfathered,'cel-gaulish',tag{
         type: "grandfathered",
         tag: "cel-gaulish",
         description: "Gaulish",
         added: "2001-05-25",
         deprecated: "2015-03-29",
         comments: "see xcg, xga, xtg"}).
iana(grandfathered,'en-GB-oed',tag{
         type: "grandfathered",
         tag: "en-GB-oed",
         description: "English, Oxford English Dictionary spelling",
         added: "2003-07-09",
         deprecated: "2015-04-17",
         preferred_value: "en-GB-oxendict"}).
iana(grandfathered,'i-ami',tag{
         type: "grandfathered",
         tag: "i-ami",
         description: "Amis",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "ami"}).
iana(grandfathered,'i-bnn',tag{
         type: "grandfathered",
         tag: "i-bnn",
         description: "Bunun",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "bnn"}).
iana(grandfathered,'i-default',tag{
         type: "grandfathered",
         tag: "i-default",
         description: "Default Language",
         added: "1998-03-10"}).
iana(grandfathered,'i-enochian',tag{
         type: "grandfathered",
         tag: "i-enochian",
         description: "Enochian",
         added: "2002-07-03",
         deprecated: "2015-03-29"}).
iana(grandfathered,'i-hak',tag{
         type: "grandfathered",
         tag: "i-hak",
         description: "Hakka",
         added: "1999-01-31",
         deprecated: "2000-01-10",
         preferred_value: "hak"}).
iana(grandfathered,'i-klingon',tag{
         type: "grandfathered",
         tag: "i-klingon",
         description: "Klingon",
         added: "1999-05-26",
         deprecated: "2004-02-24",
         preferred_value: "tlh"}).
iana(grandfathered,'i-lux',tag{
         type: "grandfathered",
         tag: "i-lux",
         description: "Luxembourgish",
         added: "1997-09-19",
         deprecated: "1998-09-09",
         preferred_value: "lb"}).
iana(grandfathered,'i-mingo',tag{
         type: "grandfathered",
         tag: "i-mingo",
         description: "Mingo",
         added: "1997-09-19"}).
iana(grandfathered,'i-navajo',tag{
         type: "grandfathered",
         tag: "i-navajo",
         description: "Navajo",
         added: "1997-09-19",
         deprecated: "2000-02-18",
         preferred_value: "nv"}).
iana(grandfathered,'i-pwn',tag{
         type: "grandfathered",
         tag: "i-pwn",
         description: "Paiwan",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "pwn"}).
iana(grandfathered,'i-tao',tag{
         type: "grandfathered",
         tag: "i-tao",
         description: "Tao",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "tao"}).
iana(grandfathered,'i-tay',tag{
         type: "grandfathered",
         tag: "i-tay",
         description: "Tayal",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "tay"}).
iana(grandfathered,'i-tsu',tag{
         type: "grandfathered",
         tag: "i-tsu",
         description: "Tsou",
         added: "1999-05-25",
         deprecated: "2009-07-29",
         preferred_value: "tsu"}).
iana(grandfathered,'no-bok',tag{
         type: "grandfathered",
         tag: "no-bok",
         description: "Norwegian Bokmal",
         added: "1995-08-23",
         deprecated: "2000-02-18",
         preferred_value: "nb"}).
iana(grandfathered,'no-nyn',tag{
         type: "grandfathered",
         tag: "no-nyn",
         description: "Norwegian Nynorsk",
         added: "1995-08-23",
         deprecated: "2000-02-18",
         preferred_value: "nn"}).
iana(grandfathered,'sgn-BE-FR',tag{
         type: "grandfathered",
         tag: "sgn-BE-FR",
         description: "Belgian-French Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "sfb"}).
iana(grandfathered,'sgn-BE-NL',tag{
         type: "grandfathered",
         tag: "sgn-BE-NL",
         description: "Belgian-Flemish Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "vgt"}).
iana(grandfathered,'sgn-CH-DE',tag{
         type: "grandfathered",
         tag: "sgn-CH-DE",
         description: "Swiss German Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "sgg"}).
iana(grandfathered,'zh-guoyu',tag{
         type: "grandfathered",
         tag: "zh-guoyu",
         description: "Mandarin or Standard Chinese",
         added: "1999-12-18",
         deprecated: "2005-07-15",
         preferred_value: "cmn"}).
iana(grandfathered,'zh-hakka',tag{
         type: "grandfathered",
         tag: "zh-hakka",
         description: "Hakka",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         preferred_value: "hak"}).
iana(grandfathered,'zh-min',tag{
         type: "grandfathered",
         tag: "zh-min",
         description: "Min, Fuzhou, Hokkien, Amoy, or Taiwanese",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         comments: "see cdo, cpx, czo, mnp, nan"}).
iana(grandfathered,'zh-min-nan',tag{
         type: "grandfathered",
         tag: "zh-min-nan",
         description: "Minnan, Hokkien, Amoy, Taiwanese, Southern Min, Southern Fujian, Hoklo, Southern Fukien, Ho-lo",
         added: "2001-03-26",
         deprecated: "2009-07-29",
         preferred_value: "nan"}).
iana(grandfathered,'zh-xiang',tag{
         type: "grandfathered",
         tag: "zh-xiang",
         description: "Xiang or Hunanese",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         preferred_value: "hsn"}).
iana(redundant,'az-Arab',tag{
         type: "redundant",
         tag: "az-Arab",
         description: "Azerbaijani in Arabic script",
         added: "2003-05-30"}).
iana(redundant,'az-Cyrl',tag{
         type: "redundant",
         tag: "az-Cyrl",
         description: "Azerbaijani in Cyrillic script",
         added: "2003-05-30"}).
iana(redundant,'az-Latn',tag{
         type: "redundant",
         tag: "az-Latn",
         description: "Azerbaijani in Latin script",
         added: "2003-05-30"}).
iana(redundant,'be-Latn',tag{
         type: "redundant",
         tag: "be-Latn",
         description: "Belarusian in Latin script",
         added: "2005-01-06"}).
iana(redundant,'bs-Cyrl',tag{
         type: "redundant",
         tag: "bs-Cyrl",
         description: "Bosnian in Cyrillic script",
         added: "2005-02-17"}).
iana(redundant,'bs-Latn',tag{
         type: "redundant",
         tag: "bs-Latn",
         description: "Bosnian in Latin script",
         added: "2005-02-17"}).
iana(redundant,'de-1901',tag{
         type: "redundant",
         tag: "de-1901",
         description: "German, traditional orthography",
         added: "2001-07-17"}).
iana(redundant,'de-1996',tag{
         type: "redundant",
         tag: "de-1996",
         description: "German, orthography of 1996",
         added: "2001-07-17"}).
iana(redundant,'de-AT-1901',tag{
         type: "redundant",
         tag: "de-AT-1901",
         description: "German, Austrian variant, traditional orthography",
         added: "2001-07-17"}).
iana(redundant,'de-AT-1996',tag{
         type: "redundant",
         tag: "de-AT-1996",
         description: "German, Austrian variant, orthography of 1996",
         added: "2001-07-17"}).
iana(redundant,'de-CH-1901',tag{
         type: "redundant",
         tag: "de-CH-1901",
         description: "German, Swiss variant, traditional orthography",
         added: "2001-07-17"}).
iana(redundant,'de-CH-1996',tag{
         type: "redundant",
         tag: "de-CH-1996",
         description: "German, Swiss variant, orthography of 1996",
         added: "2001-07-17"}).
iana(redundant,'de-DE-1901',tag{
         type: "redundant",
         tag: "de-DE-1901",
         description: "German, German variant, traditional orthography",
         added: "2001-07-17"}).
iana(redundant,'de-DE-1996',tag{
         type: "redundant",
         tag: "de-DE-1996",
         description: "German, German variant, orthography of 1996",
         added: "2001-07-17"}).
iana(redundant,'en-boont',tag{
         type: "redundant",
         tag: "en-boont",
         description: "Boontling",
         added: "2003-02-14"}).
iana(redundant,'en-scouse',tag{
         type: "redundant",
         tag: "en-scouse",
         description: "Scouse",
         added: "2000-05-25"}).
iana(redundant,'es-419',tag{
         type: "redundant",
         tag: "es-419",
         description: "Latin American Spanish",
         added: "2005-07-15"}).
iana(redundant,'iu-Cans',tag{
         type: "redundant",
         tag: "iu-Cans",
         description: "Inuktitut in Canadian Aboriginal Syllabic script",
         added: "2005-02-17"}).
iana(redundant,'iu-Latn',tag{
         type: "redundant",
         tag: "iu-Latn",
         description: "Inuktitut in Latin script",
         added: "2005-02-17"}).
iana(redundant,'mn-Cyrl',tag{
         type: "redundant",
         tag: "mn-Cyrl",
         description: "Mongolian in Cyrillic script",
         added: "2005-02-17"}).
iana(redundant,'mn-Mong',tag{
         type: "redundant",
         tag: "mn-Mong",
         description: "Mongolian in Mongolian script",
         added: "2005-02-17"}).
iana(redundant,'sgn-BR',tag{
         type: "redundant",
         tag: "sgn-BR",
         description: "Brazilian Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "bzs"}).
iana(redundant,'sgn-CO',tag{
         type: "redundant",
         tag: "sgn-CO",
         description: "Colombian Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "csn"}).
iana(redundant,'sgn-DE',tag{
         type: "redundant",
         tag: "sgn-DE",
         description: "German Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "gsg"}).
iana(redundant,'sgn-DK',tag{
         type: "redundant",
         tag: "sgn-DK",
         description: "Danish Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "dsl"}).
iana(redundant,'sgn-ES',tag{
         type: "redundant",
         tag: "sgn-ES",
         description: "Spanish Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "ssp"}).
iana(redundant,'sgn-FR',tag{
         type: "redundant",
         tag: "sgn-FR",
         description: "French Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "fsl"}).
iana(redundant,'sgn-GB',tag{
         type: "redundant",
         tag: "sgn-GB",
         description: "British Sign Language",
         added: "2001-03-02",
         deprecated: "2009-07-29",
         preferred_value: "bfi"}).
iana(redundant,'sgn-GR',tag{
         type: "redundant",
         tag: "sgn-GR",
         description: "Greek Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "gss"}).
iana(redundant,'sgn-IE',tag{
         type: "redundant",
         tag: "sgn-IE",
         description: "Irish Sign Language",
         added: "2001-03-02",
         deprecated: "2009-07-29",
         preferred_value: "isg"}).
iana(redundant,'sgn-IT',tag{
         type: "redundant",
         tag: "sgn-IT",
         description: "Italian Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "ise"}).
iana(redundant,'sgn-JP',tag{
         type: "redundant",
         tag: "sgn-JP",
         description: "Japanese Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "jsl"}).
iana(redundant,'sgn-MX',tag{
         type: "redundant",
         tag: "sgn-MX",
         description: "Mexican Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "mfs"}).
iana(redundant,'sgn-NI',tag{
         type: "redundant",
         tag: "sgn-NI",
         description: "Nicaraguan Sign Language",
         added: "2001-03-02",
         deprecated: "2009-07-29",
         preferred_value: "ncs"}).
iana(redundant,'sgn-NL',tag{
         type: "redundant",
         tag: "sgn-NL",
         description: "Dutch Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "dse"}).
iana(redundant,'sgn-NO',tag{
         type: "redundant",
         tag: "sgn-NO",
         description: "Norwegian Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "nsl"}).
iana(redundant,'sgn-PT',tag{
         type: "redundant",
         tag: "sgn-PT",
         description: "Portuguese Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "psr"}).
iana(redundant,'sgn-SE',tag{
         type: "redundant",
         tag: "sgn-SE",
         description: "Swedish Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "swl"}).
iana(redundant,'sgn-US',tag{
         type: "redundant",
         tag: "sgn-US",
         description: "American Sign Language",
         added: "2001-03-02",
         deprecated: "2009-07-29",
         preferred_value: "ase"}).
iana(redundant,'sgn-ZA',tag{
         type: "redundant",
         tag: "sgn-ZA",
         description: "South African Sign Language",
         added: "2001-11-11",
         deprecated: "2009-07-29",
         preferred_value: "sfs"}).
iana(redundant,'sl-nedis',tag{
         type: "redundant",
         tag: "sl-nedis",
         description: "Natisone dialect, Nadiza dialect",
         added: "2004-06-01"}).
iana(redundant,'sl-rozaj',tag{
         type: "redundant",
         tag: "sl-rozaj",
         description: "Resian, Resianic, Rezijan",
         added: "2003-10-09"}).
iana(redundant,'sr-Cyrl',tag{
         type: "redundant",
         tag: "sr-Cyrl",
         description: "Serbian in Cyrillic script",
         added: "2003-05-30"}).
iana(redundant,'sr-Latn',tag{
         type: "redundant",
         tag: "sr-Latn",
         description: "Serbian in Latin script",
         added: "2003-05-30"}).
iana(redundant,'tg-Arab',tag{
         type: "redundant",
         tag: "tg-Arab",
         description: "Tajik in Arabic script",
         added: "2005-02-17"}).
iana(redundant,'tg-Cyrl',tag{
         type: "redundant",
         tag: "tg-Cyrl",
         description: "Tajik in Cyrillic script",
         added: "2005-02-17"}).
iana(redundant,'uz-Cyrl',tag{
         type: "redundant",
         tag: "uz-Cyrl",
         description: "Uzbek in Cyrillic script",
         added: "2003-05-30"}).
iana(redundant,'uz-Latn',tag{
         type: "redundant",
         tag: "uz-Latn",
         description: "Uzbek in Latin script",
         added: "2003-05-30"}).
iana(redundant,'yi-Latn',tag{
         type: "redundant",
         tag: "yi-Latn",
         description: "Yiddish, in Latin script",
         added: "2003-01-07"}).
iana(redundant,'zh-cmn',tag{
         type: "redundant",
         tag: "zh-cmn",
         description: "Mandarin Chinese",
         added: "2005-07-15",
         deprecated: "2009-07-29",
         preferred_value: "cmn"}).
iana(redundant,'zh-cmn-Hans',tag{
         type: "redundant",
         tag: "zh-cmn-Hans",
         description: "Mandarin Chinese (Simplified)",
         added: "2005-07-15",
         deprecated: "2009-07-29",
         preferred_value: "cmn-Hans"}).
iana(redundant,'zh-cmn-Hant',tag{
         type: "redundant",
         tag: "zh-cmn-Hant",
         description: "Mandarin Chinese (Traditional)",
         added: "2005-07-15",
         deprecated: "2009-07-29",
         preferred_value: "cmn-Hant"}).
iana(redundant,'zh-gan',tag{
         type: "redundant",
         tag: "zh-gan",
         description: "Kan or Gan",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         preferred_value: "gan"}).
iana(redundant,'zh-Hans',tag{
         type: "redundant",
         tag: "zh-Hans",
         description: "simplified Chinese",
         added: "2003-05-30"}).
iana(redundant,'zh-Hans-CN',tag{
         type: "redundant",
         tag: "zh-Hans-CN",
         description: "PRC Mainland Chinese in simplified script",
         added: "2005-04-13"}).
iana(redundant,'zh-Hans-HK',tag{
         type: "redundant",
         tag: "zh-Hans-HK",
         description: "Hong Kong Chinese in simplified script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hans-MO',tag{
         type: "redundant",
         tag: "zh-Hans-MO",
         description: "Macao Chinese in simplified script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hans-SG',tag{
         type: "redundant",
         tag: "zh-Hans-SG",
         description: "Singapore Chinese in simplified script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hans-TW',tag{
         type: "redundant",
         tag: "zh-Hans-TW",
         description: "Taiwan Chinese in simplified script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hant',tag{
         type: "redundant",
         tag: "zh-Hant",
         description: "traditional Chinese",
         added: "2003-05-30"}).
iana(redundant,'zh-Hant-CN',tag{
         type: "redundant",
         tag: "zh-Hant-CN",
         description: "PRC Mainland Chinese in traditional script",
         added: "2005-04-13"}).
iana(redundant,'zh-Hant-HK',tag{
         type: "redundant",
         tag: "zh-Hant-HK",
         description: "Hong Kong Chinese in traditional script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hant-MO',tag{
         type: "redundant",
         tag: "zh-Hant-MO",
         description: "Macao Chinese in traditional script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hant-SG',tag{
         type: "redundant",
         tag: "zh-Hant-SG",
         description: "Singapore Chinese in traditional script",
         added: "2005-04-11"}).
iana(redundant,'zh-Hant-TW',tag{
         type: "redundant",
         tag: "zh-Hant-TW",
         description: "Taiwan Chinese in traditional script",
         added: "2005-04-11"}).
iana(redundant,'zh-wuu',tag{
         type: "redundant",
         tag: "zh-wuu",
         description: "Shanghaiese or Wu",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         preferred_value: "wuu"}).
iana(redundant,'zh-yue',tag{
         type: "redundant",
         tag: "zh-yue",
         description: "Cantonese",
         added: "1999-12-18",
         deprecated: "2009-07-29",
         preferred_value: "yue"}).
